#ifndef BSL_h_INCLUDED
#define BSL_h_INCLUDED

/*
 * !!!!	BSL.xcc
 *
 * Copyright 2007 STMicroelectronics.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CDT.h"

/*
 * The BSL (Basic Services Library) contains the following modules:
 * *	<Except Except.html> Exception support, mostly assertion macros for now.
 * *	<Hacker Hacker.html> Implementation of Hacker's Delights algorithms.
 * *	<Memory Memory.html> Memory allocator with =freeAll= and pools.
 * *	<BBTree BBTree.html> Base Binary Tree common code (unused).
 * *	<String String.html> String and StringTable implementations.
 * *	<Buffer Buffer.html> Character buffer for preparing output.
 * *	<REScan REScan.html> Support for the <re2c http://re2c.org> scanner generator.
 * *	<ETable ETable.html> Support for tables of enumerated values.
 * *	<Escape Escape.html> Support for XML escape codes.
 * *	<XMLCan XMLCan.html> XML Canonical pull scanner.
 * *	<XMLGen XMLGen.html> XML Generator.
 * *	<XMLScan XMLScan.html> XML pull Scanner.
 * *	<XMLAuto XMLAuto.html> XML Dump function generator.
 */

/*
 * Include all the exported BSL headers.
 */
#include "BSL/Except.h"
#include "BSL/Hacker.h"
#include "BSL/Memory.h"
#include "BSL/BBTree.h"
#include "BSL/String.h"
#include "BSL/Buffer.h"
#include "BSL/REScan.h"
#include "BSL/ETable.h"
#include "BSL/Escape.h"
#include "BSL/XMLCan.h"
#include "BSL/XMLGen.h"
#include "BSL/XMLScan.h"
#include "BSL/XMLAuto.h"

/*
 * BSL_TRACE --	File handle used to print the BSL traces.
 */
extern FILE *BSL_TRACE;
#ifdef CDT_NTRACE
#define BSL_TRACE NULL
#endif//CDT_NTRACE

/*
 * Variables used to control BSL in non-release mode.
 */
extern int32_t BSL_MEMORY_STATS;
extern int32_t FORCE_POOLED;

/*
 * BSL_INIT --	Initialize the BSL library.
 */
void
BSL_INIT(void);

/*
 * BSL_FINI --	Finalize BSL library.
 */
void
BSL_FINI(void);


#endif//BSL_h_INCLUDED
