#ifndef REScan_h_INCLUDED
#define REScan_h_INCLUDED

/*
 * !!!!	REScan.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
typedef unsigned char byte;

/*
  REScan --	Scan environment for re2c scanners.
*/
struct REScan_ {
  //@args	Memory memory, FILE *file
  Memory MEMORY;		// Where to allocate memory.
  FILE *INPUT;			// Input file.
  byte *TOKEN;			// Start of current token.
  byte *LIMIT;			// For YYLIMIT (past valid data).
  byte *MARKER;			// For YYMARKER.
  byte *CURSOR;			// For YYCURSOR.
  byte *BASE;			// Base of the buffer.
  byte *PAST;			// Past the buffer.
  byte *END;			// End of input.
  int32_t LINE;			// Input line counter.
  int32_t STATE;		// Scanner state.
  //@access LENGTH	(REScan_CURSOR(this) - REScan_TOKEN(this))
};
typedef struct REScan_ REScan_, *REScan;
typedef const struct REScan_ *const_REScan;
typedef struct REScan_ * restrict_REScan;

extern REScan
REScan_Ctor(REScan this, Memory memory, FILE *file);

extern REScan
REScan_Copy(REScan this, const_REScan that);

extern void
REScan_Dtor(REScan this);

extern size_t
REScan_Size(Memory memory, FILE *file);

#define REScan_MEMORY(this) ((this)->MEMORY)
#define REScan__MEMORY(this) (&(this)->MEMORY)
#define REScan_INPUT(this) ((this)->INPUT)
#define REScan__INPUT(this) (&(this)->INPUT)
#define REScan_TOKEN(this) ((this)->TOKEN)
#define REScan__TOKEN(this) (&(this)->TOKEN)
#define REScan_LIMIT(this) ((this)->LIMIT)
#define REScan__LIMIT(this) (&(this)->LIMIT)
#define REScan_MARKER(this) ((this)->MARKER)
#define REScan__MARKER(this) (&(this)->MARKER)
#define REScan_CURSOR(this) ((this)->CURSOR)
#define REScan__CURSOR(this) (&(this)->CURSOR)
#define REScan_BASE(this) ((this)->BASE)
#define REScan__BASE(this) (&(this)->BASE)
#define REScan_PAST(this) ((this)->PAST)
#define REScan__PAST(this) (&(this)->PAST)
#define REScan_END(this) ((this)->END)
#define REScan__END(this) (&(this)->END)
#define REScan_LINE(this) ((this)->LINE)
#define REScan__LINE(this) (&(this)->LINE)
#define REScan_STATE(this) ((this)->STATE)
#define REScan__STATE(this) (&(this)->STATE)
#define REScan_LENGTH(this) ((REScan_CURSOR(this) - REScan_TOKEN(this)))


//
byte *
REScan_fill(REScan this, size_t bsize);

//
static inline void
REScan_incLine(REScan this)
{
  ++*REScan__LINE(this);
}

//
static inline bool
REScan_atEnd(REScan this)
{
  return  REScan_CURSOR(this) == REScan_END(this);
}


#endif//REScan_h_INCLUDED
