#ifndef String__h_INCLUDED
#define String__h_INCLUDED

//
#include "BSL/String.h"

/*
 * String_Ctor --	Given =this= pointer, construct the String.
 */
void
String_Ctor(String this, const void *data, int32_t length, const_StringTableEntry leaf);

//
#ifdef XCC__C
extern int32_t StringTableEntry_cmpCount;
#endif

//
static inline int32_t
StringTableEntry_cmp_(const_StringTableEntry this, const void *that_data, int32_t that_length)
{
  int32_t this_length = StringTableEntry_LENGTH(this);
  const char *this_data = StringTableEntry_DATA(this);
  int32_t cmp = this_length - that_length;
#ifdef XCC__C
#ifdef String_c_INCLUDED
  StringTableEntry_cmpCount++;
#endif
#endif
  if (cmp == 0) {
    cmp = memcmp(this_data, that_data, that_length);
  }
  return cmp;
}

//
static inline int32_t
StringTableEntry_cmp(const_StringTableEntry this, const_StringTableEntry that)
{
  const char *that_data = StringTableEntry_DATA(that);
  int32_t that_length = StringTableEntry_LENGTH(that);
  return StringTableEntry_cmp_(this, that_data, that_length);
}

//
#ifndef StringTable_SPLAY
# define StringTable_insertTreap(this, data, length) \
    StringTable_insert(this, data, length)
# define StringTable_removeTreap(this, data, length) \
        StringTable_remove(this, data, length)
#else
# define StringTable_insertSplay(this, data, length) \
            StringTable_insert(this, data, length)
# define StringTable_removeSplay(this, data, length) \
                StringTable_remove(this, data, length)
#endif//StringTable_SPLAY

/*
 * StringTable_insertSplay --	Insert data into this table, unless it's already there.
 *
 * Adapted from ftp://ftp.cs.cmu.edu/user/sleator/splaying/top-down-splay.c
 * "An implementation of top-down splaying", D. Sleator <sleator@cs.cmu.edu>
 */
String
StringTable_insertSplay(StringTable this, const void *data, int32_t length);

/*
 * StringTable_insertTreap --	Insert data into the tree t, unless it's already there.
 *
 * Adapted from http://www.cs.fiu.edu/~weiss/dsaa_c2e/files.html
 * "Data Structures and Algorithm Analysis in C", Mark Allen Weiss <weiss@fiu.edu>
 */
String
StringTable_insertTreap(StringTable this, const void *data, int32_t length);

//
void
StringTableEntry_map(StringTableEntry this, const_StringTableEntry leaf,
                void (*map)(String, va_list), va_list va);

/*
 * StringTable_INIT --	Initialize the StringTable module.
 */
#define String_INIT()

/*
 * StringTable_FINI --	Finalize the StringTable module.
 */
#define String_FINI()


#endif//String__h_INCLUDED
