#ifndef XMLGen_h_INCLUDED
#define XMLGen_h_INCLUDED

/*
 * !!!!	XMLGen.xcc
 *
 * Laura Hernandez de la Puerta (lauhp09@gmail.com)
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2008 STMicroelectronics.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * UTF-8 strings
 */
typedef char *UTF8;
typedef const char *const_UTF8;
int escape;

//
typedef enum {
  SEQUENCE_NO_DOC,
  SEQUENCE_PRE_DOC,
  SEQUENCE_POST_DOC,
  SEQUENCE_START_TAG,
  SEQUENCE_ATTRIBUTES,
  SEQUENCE_CONTENT,
} writerSequence;

//
typedef enum {
  XMLGenStatus_OK,
  XMLGenStatus_Error,
  XMLGenStatus_BadUTF8,
  XMLGenStatus_NonXML,
  XMLGenStatus_AllocFailed,
  XMLGenStatus_SeqError,
  XMLGenStatus_CommentError,
  XMLGenStatus_PIError,
} XMLGenStatus;

/*
 * XMLGenWriter_rec -- XML Generator object.
 */
struct XMLGenWriter_rec_;
typedef struct XMLGenWriter_rec_ XMLGenWriter_rec_, *XMLGenWriter_rec;
typedef const struct XMLGenWriter_rec_ *const_XMLGenWriter_rec;
typedef struct XMLGenWriter_rec_ * restrict_XMLGenWriter_rec;

//
typedef struct
{
  XMLGenWriter_rec writer;
  int        count;
  int        space;
  void * *   pointers;
} plist;

//
typedef struct
{
  UTF8 buf;
  int  used;
  int  space;
} collector;

//
struct XMLGenElement_rec
{
  //@args const_UTF8 type
   const_UTF8 type;
};

//
struct XMLGenElement_{
  struct XMLGenElement_rec *XMLGENELEMENT;
};
typedef struct XMLGenElement_ XMLGenElement_, *XMLGenElement;
typedef const struct XMLGenElement_ *const_XMLGenElement;
typedef struct XMLGenElement_ * restrict_XMLGenElement;

#define XMLGenElement_XMLGENELEMENT(this) ((this)->XMLGENELEMENT)
#define XMLGenElement__XMLGENELEMENT(this) (&(this)->XMLGENELEMENT)


//
struct XMLGenAttribute_rec_ {
  const_UTF8 name;
  int provided;
  collector value;
};
typedef struct XMLGenAttribute_rec_ XMLGenAttribute_rec_, *XMLGenAttribute_rec;
typedef const struct XMLGenAttribute_rec_ *const_XMLGenAttribute_rec;
typedef struct XMLGenAttribute_rec_ * restrict_XMLGenAttribute_rec;

#define XMLGenAttribute_rec_name(this) ((this)->name)
#define XMLGenAttribute_rec__name(this) (&(this)->name)
#define XMLGenAttribute_rec_provided(this) ((this)->provided)
#define XMLGenAttribute_rec__provided(this) (&(this)->provided)
#define XMLGenAttribute_rec_value(this) ((this)->value)
#define XMLGenAttribute_rec__value(this) (&(this)->value)


//
struct XMLGenAttribute_ {
  struct XMLGenAttribute_rec *XMLGENATTRIBUTE;
};
typedef struct XMLGenAttribute_ XMLGenAttribute_, *XMLGenAttribute;
typedef const struct XMLGenAttribute_ *const_XMLGenAttribute;
typedef struct XMLGenAttribute_ * restrict_XMLGenAttribute;

#define XMLGenAttribute_XMLGENATTRIBUTE(this) ((this)->XMLGENATTRIBUTE)
#define XMLGenAttribute__XMLGENATTRIBUTE(this) (&(this)->XMLGENATTRIBUTE)


/*
 * XMLGen -- XML Generator object.
 */
struct XMLGen_;
typedef struct XMLGen_ XMLGen_, *XMLGen;
typedef const struct XMLGen_ *const_XMLGen;
typedef struct XMLGen_ * restrict_XMLGen;

/*
 * isLetter
 */
static int isLetter(int c)
{
  if (c < 0 || c > 0xffff)
    return 0;
  else
    return 1;
}

/*
 * isNameChar
 */
static int isNameChar(int c)
{
  if (c < 0 || c > 0xffff)
    return 0;
  else
    return 1;
}

/*
 * isXMLChar
 */
static int isXMLChar(int c);

/*
 * NextUnicodeChar
 */
int NextUnicodeChar(const_UTF8 *sp);

/*
 * checkNCName
 */
static XMLGenStatus checkNCName(const_UTF8 name);

/*
 * checkText
 */
XMLGenStatus checkText( const_UTF8 s);

/*
 * initCollector
 */
static XMLGenStatus initCollector(XMLGenWriter_rec w, collector * c);

/*
 * growCollector
 */
static XMLGenStatus growCollector(XMLGenWriter_rec w, collector * c, int size);

/*
 * startCollect
 */
static void startCollect(collector * c);

/*
 * endCollect
 */
static void endCollect(collector * c);

/*
 * collectString
 */
static XMLGenStatus collectString(XMLGenWriter_rec w, collector * c, const_UTF8 string);

/*
 *
 */
#define collectPiece(w,c,d,size) {if (((c)->used+(size))>=(c)->space){if (((w)->status=growCollector(w,c,(c)->used+(size)))!=XMLGenStatus_OK) return (w)->status;}strncpy((char *)(c)->buf+(c)->used,d,size);(c)->used+=size;}

/*
 * initPlist
 */
static XMLGenStatus initPlist(XMLGenWriter_rec w, plist * pl);

/*
 * initElemList
 */
static XMLGenStatus initElemList(XMLGenWriter_rec w, plist * pl);

/*
 * XMLGen_new --	Make a new XMLGen.
 */
XMLGen
XMLGen_new(Memory parent);

/*
 * XMLGen_dispose
 */
void XMLGen_dispose(XMLGenWriter_rec_ *writer);

/*
 * XMLGen_delete --	Delete this XMLGen.
 */
XMLGen
XMLGen_delete(XMLGen this);

/*
 *  allocate
 */
static void * allocate(XMLGenWriter_rec w, int bytes);

/*
 *  deallocate
 */
static void deallocate(XMLGenWriter_rec w, void * data);

/*
 *  copy
 */
static UTF8 copy(XMLGenWriter_rec w, const_UTF8 from);

/*
 *  checkExpand -- make room in a plist
 */
static int checkExpand(plist * pl);

/*
 *  listAppend
 */
static XMLGenStatus listAppend(plist * pl, void * pointer);

/*
 *  findElement
 */
static int findElement(plist * pl, const_UTF8 type);

/*
 *  findAttribute
 */
static int findAttribute(plist * pl, const_UTF8 name);

/*
 * XMLGen_declareElement --	Make a XMLGenElement.
 */
XMLGenElement
XMLGen_declareElement(XMLGen this, const_UTF8 type);

/*
 * XMLGen_declareAttribute --	Make a XMLGenAttribute.
 */
XMLGenAttribute
XMLGen_declareAttribute(XMLGen this, const_UTF8 name);

/*
 * XMLGen_addAttribute --	Add an attribute to the current element.
 */
XMLGenStatus
XMLGen_addAttribute(XMLGen this, XMLGenAttribute attribute, const_UTF8 value);

/*
 * XMLGen_PI
 */
XMLGenStatus XMLGen_PI(XMLGen this, const_UTF8 target, const_UTF8 text);

/*
 * XMLGen_startDocFile --	Start a new document.
 */
XMLGenStatus
XMLGen_startDocFile(XMLGen this, int escapeOpt);

/*
 * XMLGen_endDocument --	End a document.
 */
XMLGenStatus
XMLGen_endDocument(XMLGen this, FILE *file);

/*
 * XMLGen_startElement --	Start an element.
 */
XMLGenStatus
XMLGen_startElement(XMLGen this, XMLGenElement element);

/*
 * XMLGen_startElementLiteral --	Start an element.
 */
XMLGenStatus
XMLGen_startElementLiteral(XMLGen this, const_UTF8 type);

/*
 * XMLGen_addAttributeLiteral --	Add an attribute to the current element.
 */
XMLGenStatus
XMLGen_addAttributeLiteral(XMLGen this, const_UTF8 name, const_UTF8 value);

/*
 * XMLGen_addText --	Add text to the current element.
 */
XMLGenStatus
XMLGen_addText(XMLGen this, const_UTF8 start);

/*
 * XMLGen_comment --  Inserts a comment with the text provided
 */
XMLGenStatus
XMLGen_comment(XMLGen this, const_UTF8 text);

/*
 * addEscape
 */
static int addEscape(XMLGen this, int c);

/*
 * XMLGen_addCharacter --	Add unicode character to the current element.
 */
XMLGenStatus
XMLGen_addCharacter(XMLGen this, int c);

/*
 * writeTag
 */
XMLGenStatus
writeTag (XMLGen this, int empty);

/*
 * XMLGen_endEmptyElement --	End the current element.
 */
XMLGenStatus
XMLGen_endEmptyElement(XMLGen this);

/*
 * XMLGen_endElement --	End the current element.
 */
XMLGenStatus
XMLGen_endElement(XMLGen this);


#endif//XMLGen_h_INCLUDED
