#ifndef XMLGen__h_INCLUDED
#define XMLGen__h_INCLUDED

//
#include "BSL/XMLGen.h"

//
struct XMLGenWriter_rec_
{
  //@args      XMLGenStatus status,  Buffer_ buffer, plist elements, void * userData, void * (* alloc)(void * userData, int bytes), void(* dealloc)(void * userData, void * data);
  XMLGenStatus status;
  Buffer_ *buffer;
  plist elements;
  plist stackElems;
  plist attributes;
  writerSequence sequence;
  int escape;
  void *          	   userData;
  void *       		(* alloc)(void * userData, int bytes);
  void         		(* dealloc)(void * userData, void * data);
};

//
struct XMLGen_ {
  //@args	Memory memory, struct XMLGenWriter_rec *XMLGenWriter
   Memory MEMORY;
  struct XMLGenWriter_rec *XMLGENWRITER;
};

extern XMLGen
XMLGen_Ctor(XMLGen this, Memory memory, struct XMLGenWriter_rec *XMLGenWriter);

extern XMLGen
XMLGen_Copy(XMLGen this, const_XMLGen that);

extern void
XMLGen_Dtor(XMLGen this);

extern size_t
XMLGen_Size(Memory memory, struct XMLGenWriter_rec *XMLGenWriter);

#define XMLGen_MEMORY(this) ((this)->MEMORY)
#define XMLGen__MEMORY(this) (&(this)->MEMORY)
#define XMLGen_XMLGENWRITER(this) ((this)->XMLGENWRITER)
#define XMLGen__XMLGENWRITER(this) (&(this)->XMLGENWRITER)


//
static XMLGenStatus initCollector(XMLGenWriter_rec w, collector * c)
{
  c->space = 20;
  if ((c->buf = allocate(w, c->space)) == NULL)
    return XMLGenStatus_Error;
  c->used = 0;
  return XMLGenStatus_OK;
}

//
static XMLGenStatus growCollector(XMLGenWriter_rec w, collector * c, int size)
{
  UTF8 newSpace;
  c->space = size * 2;
  if ((newSpace = allocate(w, c->space)) == NULL)
    return XMLGenStatus_AllocFailed;
  strncpy((char *) newSpace, (const char *) c->buf, c->used);
  newSpace[c->used] = 0;
  if (c->buf!=NULL){
    deallocate(w, c->buf);
    c->buf = newSpace;
  }
  return XMLGenStatus_OK;
}

//
static void startCollect(collector * c)
{
  c->used = 0;
}

//
static void endCollect(collector * c)
{
  c->buf[c->used] = 0;
}

//
static XMLGenStatus collectString(XMLGenWriter_rec w, collector * c, const_UTF8 string)
{
  int sl = strlen((const char *) string);
  if (sl >= c->space)
    if ((w->status = growCollector(w, c, sl)) != XMLGenStatus_OK)
      return XMLGenStatus_AllocFailed;
  strcpy((char *) c->buf, (const char *) string);
  return XMLGenStatus_OK;
}

//
static void * allocate(XMLGenWriter_rec w, int bytes)
{
  if (w->alloc)
    return (void *) (*w->alloc)(w->userData, bytes);
  else
    return (void *) malloc(bytes);
}

//
static void deallocate(XMLGenWriter_rec w, void * data)
{
  if (w->dealloc)
    (*w->dealloc)(w->userData, data);
  else if (w->alloc == NULL)
    free(data);
}

/*
 * XMLGen_INIT --	Initialize the XMLGen module.
 */
#define XMLGen_INIT()
/*@XCC_.h
 * XMLGen_FINI --	Finalize the XMLGen module.
 */
#define XMLGen_FINI()


#endif//XMLGen__h_INCLUDED
