#ifndef XMLScan_h_INCLUDED
#define XMLScan_h_INCLUDED

/*
 * !!!!	XMLScan.xcc
 *
 * Laura Hernandez de la Puerta (lauhp09@gmail.com)
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
  XMLScan --	Scan environment for re2c scanners.
*/
struct XMLScan_ {
  //@args	Memory memory, FILE *file
  Memory MEMORY;		// Where to allocate memory.
  FILE *INPUT;			// Input file.
  byte *TOKEN;			// Start of current token.
  byte *LIMIT;			// For YYLIMIT (past valid data).
  byte *MARKER;			// For YYMARKER.
  byte *CURSOR;			// For YYCURSOR.
  byte *BASE;			// Base of the buffer.
  byte *PAST;			// Past the buffer.
  byte *END;			// End of input.
  int32_t LINE;			// Input line counter.
  int32_t STATE;		//
  //@access LENGTH	(XMLScan_CURSOR(this) - XMLScan_TOKEN(this))
};
typedef struct XMLScan_ XMLScan_, *XMLScan;
typedef const struct XMLScan_ *const_XMLScan;
typedef struct XMLScan_ * restrict_XMLScan;

extern XMLScan
XMLScan_Ctor(XMLScan this, Memory memory, FILE *file);

extern XMLScan
XMLScan_Copy(XMLScan this, const_XMLScan that);

extern void
XMLScan_Dtor(XMLScan this);

extern size_t
XMLScan_Size(Memory memory, FILE *file);

#define XMLScan_MEMORY(this) ((this)->MEMORY)
#define XMLScan__MEMORY(this) (&(this)->MEMORY)
#define XMLScan_INPUT(this) ((this)->INPUT)
#define XMLScan__INPUT(this) (&(this)->INPUT)
#define XMLScan_TOKEN(this) ((this)->TOKEN)
#define XMLScan__TOKEN(this) (&(this)->TOKEN)
#define XMLScan_LIMIT(this) ((this)->LIMIT)
#define XMLScan__LIMIT(this) (&(this)->LIMIT)
#define XMLScan_MARKER(this) ((this)->MARKER)
#define XMLScan__MARKER(this) (&(this)->MARKER)
#define XMLScan_CURSOR(this) ((this)->CURSOR)
#define XMLScan__CURSOR(this) (&(this)->CURSOR)
#define XMLScan_BASE(this) ((this)->BASE)
#define XMLScan__BASE(this) (&(this)->BASE)
#define XMLScan_PAST(this) ((this)->PAST)
#define XMLScan__PAST(this) (&(this)->PAST)
#define XMLScan_END(this) ((this)->END)
#define XMLScan__END(this) (&(this)->END)
#define XMLScan_LINE(this) ((this)->LINE)
#define XMLScan__LINE(this) (&(this)->LINE)
#define XMLScan_STATE(this) ((this)->STATE)
#define XMLScan__STATE(this) (&(this)->STATE)
#define XMLScan_LENGTH(this) ((XMLScan_CURSOR(this) - XMLScan_TOKEN(this)))


//
byte *
XMLScan_fill(XMLScan this, size_t bsize);

//
static inline void
XMLScan_incLine(XMLScan this)
{
  ++*XMLScan__LINE(this);
}

//
static inline bool
XMLScan_atEnd(XMLScan this)
{
  return  XMLScan_CURSOR(this) == XMLScan_END(this);
}

//
enum tokens {
  EOI,NAME_ATTRIBUTE, START_ELEMENT, CLOSE_ELEMENT,
  VALUE_ATTRIBUTE , CDATA   ,RACINE_DOCTYPE, CHARDATA , 
  STOP_ELEMENT , COMMENT , PROCINSTRUCTION , ELEMNAME_DTD , 
  ELEMCHILD_DTD , ATTELEMNAME_DTD , CATEGORYELEMDTD , 
  DTDPUBLIC_DTD , DTDSYSTEM_DTD , ELEMCHILDEITHER_DTD , 
  ELEMCHILDMIXED_DTD , CHOICESYMBOL , OPENCHILD_DTD , 
  CLOSECHILD_DTD , OCURRENCE , ATTDEFVALUE_DTD , 
  ATTTYPE_DTD , ATTNAME_DTD , ATTENUM_DTD , ENTITYNAME_DTD , 
  ENTITYVALUE_DTD, ENTITYNAMEPARAM_DTD,ATTENTITYPARAM_DTD, 
  ENTITYPUBLIC_DTD, ENTITYSYSTEM_DTD, REFENTITY, END_DTD};
enum states {
  STD, ELEM, STOPELEM, PICE, DECLCE, COMM, DOCTYPE, 
  CDAT, ERROR, ELEMNAMEDTD, ELEMCHILDDTD, ATTLISTDTD, 
  ATTDTD, ELEMCHILDMIXEDDTD, ELEMCHILDEITHERDTD, ATTENUMDTD, 
  ENTITYDTD, ATTLIST, ATTTYPEDTD, DTDPUBLIC, DTDSYSTEM , 
  ENTITYDECL, ENTITYDOUBLE, ENTITYSIMPLE};


#endif//XMLScan_h_INCLUDED
