#ifndef Dominators__h_INCLUDED
#define Dominators__h_INCLUDED

//
#include "CAL/Dominators.h"

//
struct Dominators_ {
  //@args	Memory memory, DGraph dgraph, DGraphNode entry, DGraphNode exit
  Memory MEMORY;
  DGraph DGRAPH;
  DGraphNode ENTRY;
  DGraphNode EXIT;
  int32_t COUNT;		// Number of nodes in the graph.
  DGraphNode *NODEOF;		// Maps node number to node.
  IETree_ *BUCKET;		// Lengauer-Tarjan temporary.
  int32_t *PARENT;		// Lengauer-Tarjan temporary.
  int32_t *ANCESTOR;		// Lengauer-Tarjan temporary.
  int32_t *CHILD;		// Lengauer-Tarjan temporary.
  int32_t *LABEL;		// Lengauer-Tarjan temporary.
  int32_t *SEMI;		// Lengauer-Tarjan temporary.
  int32_t *SIZE;		// Lengauer-Tarjan temporary.
  int32_t NDFS;			// Number of DFS reached nodes.
  int32_t *VERTEX;		// Lengauer-Tarjan result: DFS number to node number.
  int32_t *IDOM;		// Lengauer-Tarjan result: node number to IDOM node
                		// number. Set to 0 for entry and unreachable nodes.
  IETree_ *DOMS;		// Set of nodes that whose IDOM is this node.
  int32_t *DOMN;		// DFS number of node in the dominator tree.
  int32_t *DOML;		// Max DFS number of the node children.
  IETree_ *DOMF;		// The dominance frontier of this node.
};

extern Dominators
Dominators_Ctor(Dominators this, Memory memory, DGraph dgraph, DGraphNode entry, DGraphNode exit);

extern Dominators
Dominators_Copy(Dominators this, const_Dominators that);

extern void
Dominators_Dtor(Dominators this);

extern size_t
Dominators_Size(Memory memory, DGraph dgraph, DGraphNode entry, DGraphNode exit);

#define Dominators_MEMORY(this) ((this)->MEMORY)
#define Dominators__MEMORY(this) (&(this)->MEMORY)
#define Dominators_DGRAPH(this) ((this)->DGRAPH)
#define Dominators__DGRAPH(this) (&(this)->DGRAPH)
#define Dominators_ENTRY(this) ((this)->ENTRY)
#define Dominators__ENTRY(this) (&(this)->ENTRY)
#define Dominators_EXIT(this) ((this)->EXIT)
#define Dominators__EXIT(this) (&(this)->EXIT)
#define Dominators_COUNT(this) ((this)->COUNT)
#define Dominators__COUNT(this) (&(this)->COUNT)
#define Dominators_NODEOF(this) ((this)->NODEOF)
#define Dominators__NODEOF(this) (&(this)->NODEOF)
#define Dominators_BUCKET(this) ((this)->BUCKET)
#define Dominators__BUCKET(this) (&(this)->BUCKET)
#define Dominators_PARENT(this) ((this)->PARENT)
#define Dominators__PARENT(this) (&(this)->PARENT)
#define Dominators_ANCESTOR(this) ((this)->ANCESTOR)
#define Dominators__ANCESTOR(this) (&(this)->ANCESTOR)
#define Dominators_CHILD(this) ((this)->CHILD)
#define Dominators__CHILD(this) (&(this)->CHILD)
#define Dominators_LABEL(this) ((this)->LABEL)
#define Dominators__LABEL(this) (&(this)->LABEL)
#define Dominators_SEMI(this) ((this)->SEMI)
#define Dominators__SEMI(this) (&(this)->SEMI)
#define Dominators_SIZE(this) ((this)->SIZE)
#define Dominators__SIZE(this) (&(this)->SIZE)
#define Dominators_NDFS(this) ((this)->NDFS)
#define Dominators__NDFS(this) (&(this)->NDFS)
#define Dominators_VERTEX(this) ((this)->VERTEX)
#define Dominators__VERTEX(this) (&(this)->VERTEX)
#define Dominators_IDOM(this) ((this)->IDOM)
#define Dominators__IDOM(this) (&(this)->IDOM)
#define Dominators_DOMS(this) ((this)->DOMS)
#define Dominators__DOMS(this) (&(this)->DOMS)
#define Dominators_DOMN(this) ((this)->DOMN)
#define Dominators__DOMN(this) (&(this)->DOMN)
#define Dominators_DOML(this) ((this)->DOML)
#define Dominators__DOML(this) (&(this)->DOML)
#define Dominators_DOMF(this) ((this)->DOMF)
#define Dominators__DOMF(this) (&(this)->DOMF)


/*
 * Dominators_DOMINATORS --	The core Lengauer-Tarjan algorithm.
 *
 * See ``A Fast Algorithm for Finding Dominators in a Flowgraph'' by Lengauer and
 * Tarjan, ACM TOPLAS Vol.1 , No. 1, July 1979. The implementation below is
 * verbatim from the paper, which is needed to understand the algorithm. --	BD3
 */
void
Dominators_DOMINATORS(Dominators this, int32_t r);

/*
 * Dominators_DOMDFS --	Make the DOMN and the DOML indices from the DOMSets.
 */
void
Dominators_DOMDFS(Dominators this, int32_t v, int32_t *ndfs);

/*
 * Dominators_complete --	Complete this Dominators results.
 */
void
Dominators_complete(Dominators this, int32_t r);

/*
 * Dominators_INIT --	Initialize the Dominators module.
 */
#define Dominators_INIT()

/*
 * Dominators_FINI --	Finalize the Dominators module.
 */
#define Dominators_FINI()


#endif//Dominators__h_INCLUDED
