#ifndef CCL_h_INCLUDED
#define CCL_h_INCLUDED

/*
 * !!!!	CCL.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CDT.h"
#include "BSL/BSL.h"

/*
 * The CCL (Basic Services Library) contains the following modules:
 * *	<PQueue PQueue.html> Priority queue implementation.
 * *	<IArray IArray.html> Array of inlined items.
 * *	<IStack IStack.html> Stack of inlined items.
 * *	<IQueue IQueue.html> Queue (double ended) of inlined items.
 * *	<PtrSeq PtrSeq.html> Pointer sequence implemented as stack.
 * *	<IDList IDList.html> List (double linked) of inlined items.
 * *	<IBList IBList.html> List (blocked) of inlined items.
 * *	<TSTree TSTree.html> Ternary search tree of Sedgewick.
 * *	<PtrSet PtrSet.html> Pointer set implement as dense sorted array.
 * *	<BitSet BitSet.html> Integer set implement at bit-vector.
 * *	<Sparse Sparse.html> Integer set implement at sparse set.
 * *	<IETree IETree.html> Integer set implement at interval encoding tree.
 * *	<HTable HTable.html> Hash table implementation.
 * *	<Scoped Scoped.html> Scoped symbol table implementation.
 * *	<PtrMap PtrMap.html> Pointer mapping implemented as binary tree.
 * *	<Indexed Indexed.html> Base class for indexed objects.
 */

/*
 * Include all the exported CCL headers.
 */
#include "CCL/PQueue.h"
#include "CCL/IArray.h"
#include "CCL/IStack.h"
#include "CCL/IQueue.h"
#include "CCL/PtrSeq.h"
#include "CCL/IDList.h"
#include "CCL/IBList.h"
#include "CCL/TSTree.h"
#include "CCL/PtrSet.h"
#include "CCL/BitSet.h"
#include "CCL/Sparse.h"
#include "CCL/IETree.h"
#include "CCL/HTable.h"
#include "CCL/PtrMap.h"
#include "CCL/Scoped.h"
#include "CCL/Indexed.h"

/*
 * CCL_TRACE --	File handle used to print the CCL traces.
 */
extern FILE *CCL_TRACE;
#ifdef CDT_NTRACE
#define CCL_TRACE NULL
#endif//CDT_NTRACE

/*
 * CCL_INIT --	Initialize the CCL library.
 */
void
CCL_INIT(void);

/*
 * CCL_FINI --	Finalize CCL library.
 */
void
CCL_FINI(void);


#endif//CCL_h_INCLUDED
