#ifndef TSTree__h_INCLUDED
#define TSTree__h_INCLUDED

//
#include "CCL/TSTree.h"

/*
 * The TSTreeNode_ structure implements the TSTree tree nodes.
 */
struct TSTreeNode_ {
  //@args	intptr_t split
  intptr_t SPLIT;
  void *LTKID;
  void *EQKID;
  void *GTKID;
};
typedef struct TSTreeNode_ TSTreeNode_, *TSTreeNode;
typedef const struct TSTreeNode_ *const_TSTreeNode;
typedef struct TSTreeNode_ * restrict_TSTreeNode;

extern TSTreeNode
TSTreeNode_Ctor(TSTreeNode this, intptr_t split);

extern TSTreeNode
TSTreeNode_Copy(TSTreeNode this, const_TSTreeNode that);

extern void
TSTreeNode_Dtor(TSTreeNode this);

extern size_t
TSTreeNode_Size(intptr_t split);

#define TSTreeNode_SPLIT(this) ((this)->SPLIT)
#define TSTreeNode__SPLIT(this) (&(this)->SPLIT)
#define TSTreeNode_LTKID(this) ((this)->LTKID)
#define TSTreeNode__LTKID(this) (&(this)->LTKID)
#define TSTreeNode_EQKID(this) ((this)->EQKID)
#define TSTreeNode__EQKID(this) (&(this)->EQKID)
#define TSTreeNode_GTKID(this) ((this)->GTKID)
#define TSTreeNode__GTKID(this) (&(this)->GTKID)


//
struct TSTree_ {
  //@args	Memory memory, intptr_t stop
  //memory:	Where the TSTree nodes are allocated.
  //stop:	The special value that stops the search key.
  Memory MEMORY;
  intptr_t STOP;	// The stop value.
  TSTreeNode ROOT;	// The root node.
  TSTreeNode LAST;	// The last node inserted.
  int32_t COUNT;	// Current count of entries.
  int32_t LENGTH;	//<! Maximum key length.
};

extern TSTree
TSTree_Ctor(TSTree this, Memory memory, intptr_t stop);

extern TSTree
TSTree_Copy(TSTree this, const_TSTree that);

extern void
TSTree_Dtor(TSTree this);

extern size_t
TSTree_Size(Memory memory, intptr_t stop);

#define TSTree_MEMORY(this) ((this)->MEMORY)
#define TSTree__MEMORY(this) (&(this)->MEMORY)
#define TSTree_STOP(this) ((this)->STOP)
#define TSTree__STOP(this) (&(this)->STOP)
#define TSTree_ROOT(this) ((this)->ROOT)
#define TSTree__ROOT(this) (&(this)->ROOT)
#define TSTree_LAST(this) ((this)->LAST)
#define TSTree__LAST(this) (&(this)->LAST)
#define TSTree_COUNT(this) ((this)->COUNT)
#define TSTree__COUNT(this) (&(this)->COUNT)
#define TSTree_LENGTH(this) ((this)->LENGTH)
#define TSTree__LENGTH(this) (&(this)->LENGTH)


//
TSTreeNode
TSTreeNode_insert(TSTreeNode this, intptr_t key[], TSTree tstree);

//
void
TSTreeNode_map(TSTreeNode this, intptr_t stop, intptr_t key[], int i,
               void (*map)(intptr_t [], void *, va_list), va_list va);

/*
 * TSTree_INIT --	Initialize the TSTree module.
 */
#define TSTree_INIT()

/*
 * TSTree_FINI --	Finalize the TSTree module.
 */
#define TSTree_FINI()


#endif//TSTree__h_INCLUDED
