#ifndef CDT_INCLUDED
#define CDT_INCLUDED

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifndef __cplusplus
# if HAVE_STDBOOL_H
#  include <stdbool.h>
# else

/* C99 states this must be defined if true, false and bool macros are defined */
#ifndef __bool_true_false_are_defined
enum { false, true };
# define bool	unsigned char
# define false	false
# define true	true
#define __bool_true_false_are_defined 1
#endif //__bool_true_false_are_defined

# endif//__cplusplus
#endif
#include "CDT_stdint.h"
#include "CDT_public.h"

/*
 * ASSERT -- Always fail if expression is false.
 */
#define ASSERT(expression) \
  do { \
    if (!(expression)) { \
      fprintf(stderr, "%s:%d: assertion failed\t" #expression "\n", __FILE__,  __LINE__); \
      *(long *)(0x1) = 0; \
    } \
  } while (0) \
/* ASSERT */

/*
 * Set CDT_OPTIMIZE according to __OPTIMIZE__ (defined by gcc and icc) and NDEBUG.
 */
#ifdef __OPTIMIZE__
# ifndef CDT_OPTIMIZE
#  ifdef NDEBUG
#    define CDT_OPTIMIZE 3
#  else//!NDEBUG
#    define CDT_OPTIMIZE 2
#  endif//NDEBUG
# endif//CDT_OPTIMIZE
#else//!__OPTIMIZE__
# ifndef CDT_OPTIMIZE
#  ifdef NDEBUG
#    define CDT_OPTIMIZE 1
#  else//!NDEBUG
#    define CDT_OPTIMIZE 0
#  endif//NDEBUG
# endif//CDT_OPTIMIZE
#endif//__OPTIMIZE__

/*
 * GETENV -- Replacement for getenv().
 */
#ifndef GETENV
# define GETENV(string) getenv(string)
#endif//GETENV

/*
 * Fall-back definitions for the MIN / MAX macros.
 */
#ifndef MIN
# define MIN(x,y) ((x) < (y)? (x): (y))
#endif
#ifndef MAX
# define MAX(x,y) ((x) > (y)? (x): (y))
#endif
#ifndef ABS
# define ABS(x) ((x) >= -(x)? (x): -(x))
#endif

/* Maximun alignment required by the basic C types. */
#ifndef CDT_ALIGN_MAX
#define CDT_ALIGN_MAX MEMORY_ALIGN_MAX
#endif//CDT_ALIGN_MAX
/* Align the supplied =size= to the next CDT_ALIGN_MAX multiple. */
#define CDT_ALIGN_NEXT(size) \
  (((size_t)(size) + CDT_ALIGN_MAX - 1) & (size_t)-CDT_ALIGN_MAX)
/* Align the supplied =size= to the previous CDT_ALIGN_MAX multiple. */
#define CDT_ALIGN_PREV(size) \
  ((size_t)(size) & (size_t)-CDT_ALIGN_MAX)
/* True if the supplied =size= is CDT_ALIGN_MAX aligned, else false. */
#define CDT_ALIGNED(size) \
  (((size_t)(size) & (size_t)-CDT_ALIGN_MAX) == (size_t)(size))
/* The distance in bytes between two pointers =a= and =b=. */
#define CDT_DISTANCE(a, b) \
  ((char *)(a) - (char *)(b))

#define CDT_MEMORY_POOLS 1

#define CDT_MEMORY_STATS 0

#endif//CDT_INCLUDED

