dnl @synopsis AC_C_MEMORY_ALIGN_MAX
dnl
dnl Check for the maximum memory alignment of the C basic types.
dnl In case of cross-compilation, the cached value MUST be set.
dnl
dnl @author  Stephane Guiboud-Ribaud <Stephane.Guiboud-Ribaud@st.com>
dnl

dnl Available from the GNU Autoconf Macro Archive at:
dnl http://www.gnu.org/software/ac-archive/htmldoc/ac_compile_check_sizeof.html
dnl
AC_DEFUN([AC_COMPILE_CHECK_SIZEOF],
[changequote(<<, >>)dnl
dnl The name to #define.
define(<<AC_TYPE_NAME>>, translit(sizeof_$1, [a-z *], [A-Z_P]))dnl
dnl The cache variable name.
define(<<AC_CV_NAME>>, translit(ac_cv_sizeof_$1, [ *], [_p]))dnl
changequote([, ])dnl
AC_MSG_CHECKING(size of $1)
AC_CACHE_VAL(AC_CV_NAME,
[for ac_size in 4 8 1 2 16 $2 ; do # List sizes in rough order of prevalence.
  AC_TRY_COMPILE([#include "confdefs.h"
#include <sys/types.h>
$2
], [switch (0) case 0: case (sizeof ($1) == $ac_size):;], AC_CV_NAME=$ac_size)
  if test x$AC_CV_NAME != x ; then break; fi
done
])
if test x$AC_CV_NAME = x ; then
  AC_MSG_ERROR([cannot determine a size for $1])
fi
AC_MSG_RESULT($AC_CV_NAME)
AC_DEFINE_UNQUOTED(AC_TYPE_NAME, $AC_CV_NAME, [The number of bytes in type $1])
undefine([AC_TYPE_NAME])dnl
undefine([AC_CV_NAME])dnl
])


AC_DEFUN([AC_C_MEMORY_ALIGN_MAX],
[
AC_CACHE_CHECK(
  [for host machine memory alignment constraint],
  [ac_cv_c_memory_align_max],
[AC_TRY_RUN([#include <stdio.h>

typedef char conf_char_t;
typedef short conf_short_t;
typedef int conf_int_t;
typedef long conf_long_t;
typedef long long conf_long_long_t;
typedef float conf_float_t;
typedef double conf_double_t;
typedef long double conf_long_double_t;

#define STRUCT_DEFINE(TYPE) struct { char a; TYPE b; } var_ ## TYPE
#define STRUCT_ALIGN(TYPE)  { p1 = &(var_ ## TYPE .a); p2 = &(var_ ## TYPE .b); if (align < (p2-p1)) align = (p2-p1); }

int conf_struct_offset(void *p1, void *p2) { return p2 - p1; }

STRUCT_DEFINE(conf_char_t);
STRUCT_DEFINE(conf_short_t);
STRUCT_DEFINE(conf_int_t);
STRUCT_DEFINE(conf_long_t);
STRUCT_DEFINE(conf_long_long_t);
STRUCT_DEFINE(conf_float_t);
STRUCT_DEFINE(conf_double_t);
STRUCT_DEFINE(conf_long_double_t);

main()
{
  void *p1, *p2;
  int align = 0;

  FILE *f=fopen("conftestval", "w");
  if (!f) exit(1);
  
  STRUCT_ALIGN(conf_char_t);
  STRUCT_ALIGN(conf_short_t);
  STRUCT_ALIGN(conf_int_t);
  STRUCT_ALIGN(conf_long_t);
  STRUCT_ALIGN(conf_long_long_t);
  STRUCT_ALIGN(conf_float_t);
  STRUCT_ALIGN(conf_double_t);
  STRUCT_ALIGN(conf_long_double_t);

  fprintf(f, "%d\n", align);
  fclose(f);

  exit(0);
}],
[ac_cv_c_memory_align_max=`cat conftestval`],
[AC_MSG_ERROR([unknown memory alignment constraint - sorry, please pre-set ac_cv_c_memory_align_max])],

[case ${host} in
    arm-*)
    ac_cv_c_memory_align_max=8
    ac_cv_memory_align_max="set"
    break
    ;;
    st231-*-*|st221-*-*|st220-*-*|st200-*-*)
    ac_cv_c_memory_align_max=8
    ac_cv_memory_align_max="set"
    break
    ;;
    sh4-*-*|st40-*-*)
    ac_cv_c_memory_align_max=8
    ac_cv_memory_align_max="set"
    break
    ;;
 esac

 if test "x${ac_cv_memory_align_max}" = "x"; then
  AC_MSG_ERROR([unknown memory alignment constraint - sorry, please pre-set ac_cv_c_memory_align_max])
 fi])])

AC_DEFINE_UNQUOTED([MEMORY_ALIGN_MAX], [${ac_cv_c_memory_align_max}])
])

dnl @synopsis AC_CREATE_STDINT_H [( HEADER-TO-GENERATE [, HEDERS-TO-CHECK])]
dnl
dnl the "ISO C9X: 7.18 Integer types <stdint.h>" section requires the
dnl existence of an include file <stdint.h> that defines a set of
dnl typedefs, especially uint8_t,int32_t,uintptr_t. Many older
dnl installations will not provide this file, but some will have the
dnl very same definitions in <inttypes.h>. In other enviroments we can
dnl use the inet-types in <sys/types.h> which would define the typedefs
dnl int8_t and u_int8_t respectivly.
dnl
dnl This macros will create a local "_stdint.h" or the headerfile given
dnl as an argument. In many cases that file will just have a singular
dnl "#include <stdint.h>" or "#include <inttypes.h>" statement, while
dnl in other environments it will provide the set of basic 'stdint's
dnl defined:
dnl int8_t,uint8_t,int16_t,uint16_t,int32_t,uint32_t,intptr_t,uintptr_t
dnl int_least32_t.. int_fast32_t.. intmax_t which may or may not rely
dnl on the definitions of other files, or using the
dnl AC_COMPILE_CHECK_SIZEOF macro to determine the actual sizeof each
dnl type.
dnl
dnl if your header files require the stdint-types you will want to
dnl create an installable file mylib-int.h that all your other
dnl installable header may include. So if you have a library package
dnl named "mylib", just use
dnl
dnl      AC_CREATE_STDINT_H(mylib-int.h)
dnl
dnl in configure.in and go to install that very header file in
dnl Makefile.am along with the other headers (mylib.h) - and the
dnl mylib-specific headers can simply use "#include <mylib-int.h>" to
dnl obtain the stdint-types.
dnl
dnl Remember, if the system already had a valid <stdint.h>, the
dnl generated file will include it directly. No need for fuzzy
dnl HAVE_STDINT_H things...
dnl
dnl (note also the newer variant AX_CREATE_STDINT_H of this macro)
dnl
dnl @category C
dnl @author Guido Draheim <guidod@gmx.de>
dnl @version 2003-05-21
dnl @license GPLWithACException

AC_DEFUN([AC_CREATE_STDINT_H],
[# ------ AC CREATE STDINT H -------------------------------------
AC_MSG_CHECKING([for stdint-types....])
ac_stdint_h=`echo ifelse($1, , _stdint.h, $1)`
if test "$ac_stdint_h" = "stdint.h" ; then
 AC_MSG_RESULT("(are you sure you want them in ./stdint.h?)")
elif test "$ac_stdint_h" = "inttypes.h" ; then
 AC_MSG_RESULT("(are you sure you want them in ./inttypes.h?)")
else
 AC_MSG_RESULT("(putting them into $ac_stdint_h)")
fi

inttype_headers=`echo inttypes.h sys/inttypes.h sys/inttypes.h $2 \
| sed -e 's/,/ /g'`

 ac_cv_header_stdint_x="no-file"
 ac_cv_header_stdint_o="no-file"
 ac_cv_header_stdint_u="no-file"
 for i in stdint.h $inttype_headers ; do
   unset ac_cv_type_uintptr_t
   unset ac_cv_type_uint64_t
   _AC_CHECK_TYPE_NEW(uintptr_t,[ac_cv_header_stdint_x=$i],dnl
     continue,[#include <$i>])
   AC_CHECK_TYPE(uint64_t,[and64="(uint64_t too)"],[and64=""],[#include<$i>])
   AC_MSG_RESULT(... seen our uintptr_t in $i $and64)
   break;
 done
 if test "$ac_cv_header_stdint_x" = "no-file" ; then
 for i in stdint.h $inttype_headers ; do
   unset ac_cv_type_uint32_t
   unset ac_cv_type_uint64_t
   AC_CHECK_TYPE(uint32_t,[ac_cv_header_stdint_o=$i],dnl
     continue,[#include <$i>])
   AC_CHECK_TYPE(uint64_t,[and64="(uint64_t too)"],[and64=""],[#include<$i>])
   AC_MSG_RESULT(... seen our uint32_t in $i $and64)
   break;
 done
 if test "$ac_cv_header_stdint_o" = "no-file" ; then
 for i in sys/types.h $inttype_headers ; do
   unset ac_cv_type_u_int32_t
   unset ac_cv_type_u_int64_t
   AC_CHECK_TYPE(u_int32_t,[ac_cv_header_stdint_u=$i],dnl
     continue,[#include <$i>])
   AC_CHECK_TYPE(uint64_t,[and64="(u_int64_t too)"],[and64=""],[#include<$i>])
   AC_MSG_RESULT(... seen our u_int32_t in $i $and64)
   break;
 done
 fi
 fi

# ----------------- DONE inttypes.h checks MAYBE C basic types --------

if test "$ac_cv_header_stdint_x" = "no-file" ; then
   AC_COMPILE_CHECK_SIZEOF(char)
   AC_COMPILE_CHECK_SIZEOF(short)
   AC_COMPILE_CHECK_SIZEOF(int)
   AC_COMPILE_CHECK_SIZEOF(long)
   AC_COMPILE_CHECK_SIZEOF(void*)
   ac_cv_header_stdint_test="yes"
else
   ac_cv_header_stdint_test="no"
fi

# ----------------- DONE inttypes.h checks START header -------------
_ac_stdint_h=AS_TR_CPP(_$ac_stdint_h)
AC_MSG_RESULT(creating $ac_stdint_h : $_ac_stdint_h)
echo "#ifndef" $_ac_stdint_h >$ac_stdint_h
echo "#define" $_ac_stdint_h "1" >>$ac_stdint_h
echo "#ifndef" _GENERATED_STDINT_H >>$ac_stdint_h
echo "#define" _GENERATED_STDINT_H '"'$PACKAGE $VERSION'"' >>$ac_stdint_h
if test "$GCC" = "yes" ; then
  echo "/* generated using a gnu compiler version" `$CC --version` "*/" \
  >>$ac_stdint_h
else
  echo "/* generated using $CC */" >>$ac_stdint_h
fi
echo "" >>$ac_stdint_h

if test "$ac_cv_header_stdint_x" != "no-file" ; then
   ac_cv_header_stdint="$ac_cv_header_stdint_x"
elif  test "$ac_cv_header_stdint_o" != "no-file" ; then
   ac_cv_header_stdint="$ac_cv_header_stdint_o"
elif  test "$ac_cv_header_stdint_u" != "no-file" ; then
   ac_cv_header_stdint="$ac_cv_header_stdint_u"
else
   ac_cv_header_stdint="stddef.h"
fi

# ----------------- See if int_least and int_fast types are present
unset ac_cv_type_int_least32_t
unset ac_cv_type_int_fast32_t
AC_CHECK_TYPE(int_least32_t,,,[#include <$ac_cv_header_stdint>])
AC_CHECK_TYPE(int_fast32_t,,,[#include<$ac_cv_header_stdint>])

if test "$ac_cv_header_stdint" != "stddef.h" ; then
if test "$ac_cv_header_stdint" != "stdint.h" ; then
AC_MSG_RESULT(..adding include stddef.h)
   echo "#include <stddef.h>" >>$ac_stdint_h
fi ; fi
AC_MSG_RESULT(..adding include $ac_cv_header_stdint)
   echo "#include <$ac_cv_header_stdint>" >>$ac_stdint_h
echo "" >>$ac_stdint_h

# ----------------- DONE header START basic int types -------------
if test "$ac_cv_header_stdint_x" = "no-file" ; then
   AC_MSG_RESULT(... need to look at C basic types)
dnl ac_cv_header_stdint_test="yes" # moved up before creating the file
else
   AC_MSG_RESULT(... seen good stdint.h inttypes)
dnl ac_cv_header_stdint_test="no"  # moved up before creating the file
fi

if test "$ac_cv_header_stdint_u" != "no-file" ; then
   AC_MSG_RESULT(... seen bsd/sysv typedefs)
   cat >>$ac_stdint_h <<EOF

/* int8_t int16_t int32_t defined by inet code, redeclare the u_intXX types */
typedef u_int8_t uint8_t;
typedef u_int16_t uint16_t;
typedef u_int32_t uint32_t;
EOF
    cat >>$ac_stdint_h <<EOF

/* glibc compatibility */
#ifndef __int8_t_defined
#define __int8_t_defined
#endif
EOF
fi

ac_cv_sizeof_x="$ac_cv_sizeof_char:$ac_cv_sizeof_short"
ac_cv_sizeof_X="$ac_cv_sizeof_x:$ac_cv_sizeof_int"
ac_cv_sizeof_X="$ac_cv_sizeof_X:$ac_cv_sizeof_voidp:$ac_cv_sizeof_long"
if test "$ac_cv_header_stdint" = "stddef.h" ; then
#   we must guess all the basic types. Apart from byte-adressable system,
# there a few 32-bit-only dsp-systems. nibble-addressable systems are way off.
    cat >>$ac_stdint_h <<EOF
/* ------------ BITSPECIFIC INTTYPES SECTION --------------- */
EOF
    t="typedefs for a"
    case "$ac_cv_sizeof_X" in
     1:2:2:2:4) AC_MSG_RESULT(..adding $t normal 16-bit system)
                cat >>$ac_stdint_h <<EOF
/*              a normal 16-bit system                       */
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned long   uint32_t;
#ifndef __int8_t_defined
#define __int8_t_defined
typedef          char    int8_t;
typedef          short   int16_t;
typedef          long    int32_t;
#endif
EOF
;;
     1:2:2:4:4) AC_MSG_RESULT(..adding $t 32-bit system derived from a 16-bit)
                cat >>$ac_stdint_h <<EOF
/*              a 32-bit system derived from a 16-bit        */
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned int    uint32_t;
#ifndef __int8_t_defined
#define __int8_t_defined
typedef          char    int8_t;
typedef          short   int16_t;
typedef          int     int32_t;
#endif
EOF
;;
     1:2:4:4:4) AC_MSG_RESULT(..adding $t normal 32-bit system)
                cat >>$ac_stdint_h <<EOF
/*              a normal 32-bit system                       */
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned int    uint32_t;
#ifndef __int8_t_defined
#define __int8_t_defined
typedef          char    int8_t;
typedef          short   int16_t;
typedef          int     int32_t;
#endif
EOF
;;
     1:2:4:4:8) AC_MSG_RESULT(..adding $t 32-bit system prepared for 64-bit)
                cat >>$ac_stdint_h <<EOF

/*              a 32-bit system prepared for 64-bit          */
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned int    uint32_t;
#ifndef __int8_t_defined
#define __int8_t_defined
typedef          char    int8_t;
typedef          short   int16_t;
typedef          int     int32_t;
#endif
EOF
;;
     1:2:4:8:8) AC_MSG_RESULT(..adding $t normal 64-bit system)
                cat >>$ac_stdint_h <<EOF

/*              a normal 64-bit system                       */
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned int    uint32_t;
#ifndef __int8_t_defined
#define __int8_t_defined
typedef          char    int8_t;
typedef          short   int16_t;
typedef          int     int32_t;
#endif
EOF
;;
     1:2:4:8:4) AC_MSG_RESULT(..adding $t 64-bit system derived from a 32-bit)
                cat >>$ac_stdint_h <<EOF

/*              a 64-bit system derived from a 32-bit system */
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned int    uint32_t;
#ifndef __int8_t_defined
#define __int8_t_defined
typedef          char    int8_t;
typedef          short   int16_t;
typedef          int     int32_t;
#endif
EOF
;;
  *)
    AC_MSG_ERROR([ $ac_cv_sizeof_X dnl
 what is that a system? contact the author, quick! http://ac-archive.sf.net])
    exit 1
;;
   esac
fi

# ------------- DONE basic int types START int64_t types ------------
if test "$ac_cv_type_uint64_t" = "yes"
then AC_MSG_RESULT(... seen good uint64_t)
     cat >>$ac_stdint_h <<EOF

/* system headers have good uint64_t */
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
#endif
EOF

elif test "$ac_cv_type_u_int64_t" = "yes"
then AC_MSG_RESULT(..adding typedef u_int64_t uint64_t)
     cat >>$ac_stdint_h <<EOF

/* system headers have an u_int64_t */
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef u_int64_t uint64_t;
#endif
EOF
else AC_MSG_RESULT(..adding generic uint64_t runtime checks)
     cat >>$ac_stdint_h <<EOF

/* -------------------- 64 BIT GENERIC SECTION -------------------- */
/* here are some common heuristics using compiler runtime specifics */
#if (defined __STDC_VERSION__ && defined __STDC_VERSION__ > 199901L) || \
    defined __STCC_ST200__

#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef long long int64_t;
typedef unsigned long long uint64_t;
#endif

#elif !defined __STRICT_ANSI__
#if defined _MSC_VER || defined __WATCOMC__ || defined __BORLANDC__

#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef __int64 int64_t;
typedef unsigned __int64 uint64_t;
#endif

#elif defined __GNUC__ || defined __MWERKS__ || defined __ELF__
dnl /* note: all ELF-systems seem to have loff-support which needs 64-bit */

#if !defined _NO_LONGLONG
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef long long int64_t;
typedef unsigned long long uint64_t;
#endif
#endif

#elif defined __alpha || (defined __mips && defined _ABIN32)

#if !defined _NO_LONGLONG
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef long int64_t;
typedef unsigned long uint64_t;
#endif
#endif
  /* compiler/cpu type ... or just ISO C99 */
#endif
#endif
EOF

# plus a default 64-bit for systems that are likely to be 64-bit ready
  case "$ac_cv_sizeof_x:$ac_cv_sizeof_voidp:$ac_cv_sizeof_long" in
    1:2:8:8) AC_MSG_RESULT(..adding uint64_t default, normal 64-bit system)
cat >>$ac_stdint_h <<EOF
/* DEFAULT: */
/* seen normal 64-bit system, CC has sizeof(long and void*) == 8 bytes */
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef long int64_t;
typedef unsigned long uint64_t;
#endif
EOF
;;
    1:2:4:8) AC_MSG_RESULT(..adding uint64_t default, typedef to long)
cat >>$ac_stdint_h <<EOF
/* DEFAULT: */
/* seen 32-bit system prepared for 64-bit, CC has sizeof(long) == 8 bytes */
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef long int64_t;
typedef unsigned long uint64_t;
#endif
EOF
;;
    1:2:8:4) AC_MSG_RESULT(..adding uint64_t default, typedef long long)
cat >>$ac_stdint_h <<EOF
/* DEFAULT: */
/* seen 64-bit derived from a 32-bit, CC has sizeof(long) == 4 bytes */
#ifndef _HAVE_UINT64_T
#define _HAVE_UINT64_T
typedef long long int64_t;
typedef unsigned long long uint64_t;
#endif
EOF
;;
   *)
cat >>$ac_stdint_h <<EOF
/* NOTE: */
/* the configure-checks for the basic types did not make us believe */
/* that we could add a fallback to a 'long long' typedef to int64_t */
EOF
  esac
fi

# ------------- DONE int64_t types START intptr types ------------
if test "$ac_cv_header_stdint_x" = "no-file" ; then
  cat >>$ac_stdint_h <<EOF

/* -------------------------- INPTR SECTION --------------------------- */
EOF
  case "$ac_cv_sizeof_x:$ac_cv_sizeof_voidp" in
  1:2:2)
    a="int16_t" ; cat >>$ac_stdint_h <<EOF
/* we tested sizeof(void*) to be of 2 chars, hence we declare it 16-bit */

typedef uint16_t uintptr_t;
typedef  int16_t  intptr_t;
EOF
;;
  1:2:4)
    a="int32_t" ; cat >>$ac_stdint_h <<EOF
/* we tested sizeof(void*) to be of 4 chars, hence we declare it 32-bit */

typedef uint32_t uintptr_t;
typedef  int32_t  intptr_t;
EOF
;;
  1:2:8)
    a="int64_t" ; cat >>$ac_stdint_h <<EOF
/* we tested sizeof(void*) to be of 8 chars, hence we declare it 64-bit */

typedef uint64_t uintptr_t;
typedef  int64_t  intptr_t;
EOF
;;
  *)
    a="long" ; cat >>$ac_stdint_h <<EOF
/* we tested sizeof(void*) but got no guess, hence we declare it as if long */

typedef unsigned long uintptr_t;
typedef          long  intptr_t;
EOF
;;
  esac
AC_MSG_RESULT(..adding typedef $a intptr_t)
fi

# ------------- DONE intptr types START int_least types ------------
if test "$ac_cv_type_int_least32_t" = "no"; then
AC_MSG_RESULT(..adding generic int_least-types)
     cat >>$ac_stdint_h <<EOF

/* --------------GENERIC INT_LEAST ------------------ */

typedef  int8_t    int_least8_t;
typedef  int16_t   int_least16_t;
typedef  int32_t   int_least32_t;
#ifdef _HAVE_UINT64_T
typedef  int64_t   int_least64_t;
#endif

typedef uint8_t   uint_least8_t;
typedef uint16_t  uint_least16_t;
typedef uint32_t  uint_least32_t;
#ifdef _HAVE_UINT64_T
typedef uint64_t  uint_least64_t;
#endif
EOF
fi

# ------------- DONE intptr types START int_least types ------------
if test "$ac_cv_type_int_fast32_t" = "no"; then
AC_MSG_RESULT(..adding generic int_fast-types)
     cat >>$ac_stdint_h <<EOF

/* --------------GENERIC INT_FAST ------------------ */

typedef  int8_t    int_fast8_t;
typedef  int32_t   int_fast16_t;
typedef  int32_t   int_fast32_t;
#ifdef _HAVE_UINT64_T
typedef  int64_t   int_fast64_t;
#endif

typedef uint8_t   uint_fast8_t;
typedef uint32_t  uint_fast16_t;
typedef uint32_t  uint_fast32_t;
#ifdef _HAVE_UINT64_T
typedef uint64_t  uint_fast64_t;
#endif
EOF
fi

if test "$ac_cv_header_stdint_x" = "no-file" ; then
     cat >>$ac_stdint_h <<EOF

#ifdef _HAVE_UINT64_T
typedef int64_t        intmax_t;
typedef uint64_t      uintmax_t;
#else
typedef long int       intmax_t;
typedef unsigned long uintmax_t;
#endif
EOF
fi

# [CG]------------- START INTx_... defines ------------
# [CG] Extracted from linux redhat /usr/include/stdint.h
# [BD3] Enclose each #define within #ifndef / #endif

AC_MSG_RESULT(..adding limits macros)
cat >>$ac_stdint_h <<EOF
/* The ISO C99 standard specifies that in C++ implementations these
   macros should only be defined if explicitly requested.  */
#if !defined __cplusplus || defined __STDC_LIMIT_MACROS

EOF
case "$ac_sizeof_long" in
  8) cat >>$ac_stdint_h <<EOF
#ifdef _HAVE_UINT64_T
#ifndef __INT64_C
#define __INT64_C(c)	c ## L
#endif/*__INT64_C*/
#ifndef __UINT64_C
#define __UINT64_C(c)	c ## UL
#endif/*__UINT64_C*/
#endif
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifdef _HAVE_UINT64_T
#ifndef __INT64_C
#define __INT64_C(c)		c ## LL
#endif/*__INT64_C*/
#ifndef __UINT64_C
#define __UINT64_C(c)	c ## ULL
#endif/*__UINT64_C*/
#endif
EOF
;;
esac

cat >>$ac_stdint_h <<EOF
/* Limits of integral types.  */

/* Minimum of signed integral types.  */
#ifndef INT8_MIN
#define INT8_MIN		(-128)
#endif/*INT8_MIN*/
#ifndef INT16_MIN
#define INT16_MIN		(-32767-1)
#endif/*INT16_MIN*/
#ifndef INT32_MIN
#define INT32_MIN		(-2147483647-1)
#endif/*INT32_MIN*/
#ifdef _HAVE_UINT64_T
#ifndef INT64_MIN
#define INT64_MIN		(-__INT64_C(9223372036854775807)-1)
#endif/*INT64_MIN*/
#endif

/* Maximum of signed integral types.  */
#ifndef INT8_MAX
#define INT8_MAX		(127)
#endif/*INT8_MAX*/
#ifndef INT16_MAX
#define INT16_MAX		(32767)
#endif/*INT16_MAX*/
#ifndef INT32_MAX
#define INT32_MAX		(2147483647)
#endif/*INT32_MAX*/
#ifdef _HAVE_UINT64_T
#ifndef INT64_MAX
#define INT64_MAX		(__INT64_C(9223372036854775807))
#endif/*INT64_MAX*/
#endif

/* Maximum of unsigned integral types.  */
#ifndef UINT8_MAX
#define UINT8_MAX		(255)
#endif/*UINT8_MAX*/
#ifndef UINT16_MAX
#define UINT16_MAX		(65535)
#endif/*UINT16_MAX*/
#ifndef UINT32_MAX
#define UINT32_MAX		(4294967295U)
#endif/*UINT32_MAX*/
#ifdef _HAVE_UINT64_T
#ifndef UINT64_MAX
#define UINT64_MAX		(__UINT64_C(18446744073709551615))
#endif/*UINT64_MAX*/
#endif

/* Minimum of signed integral types having a minimum size.  */
#ifndef INT_LEAST8_MIN
#define INT_LEAST8_MIN		(-128)
#endif/*INT_LEAST8_MIN*/
#ifndef INT_LEAST16_MIN
#define INT_LEAST16_MIN	(-32767-1)
#endif/*INT_LEAST16_MIN*/
#ifndef INT_LEAST32_MIN
#define INT_LEAST32_MIN	(-2147483647-1)
#endif/*INT_LEAST32_MIN*/
#ifdef _HAVE_UINT64_T
#ifndef INT_LEAST64_MIN
#define INT_LEAST64_MIN	(-__INT64_C(9223372036854775807)-1)
#endif/*INT_LEAST64_MIN*/
#endif

/* Maximum of signed integral types having a minimum size.  */
#ifndef INT_LEAST8_MAX
#define INT_LEAST8_MAX		(127)
#endif/*INT_LEAST8_MAX*/
#ifndef INT_LEAST16_MAX
#define INT_LEAST16_MAX	(32767)
#endif/*INT_LEAST16_MAX*/
#ifndef INT_LEAST32_MAX
#define INT_LEAST32_MAX	(2147483647)
#endif/*INT_LEAST32_MAX*/
#ifdef _HAVE_UINT64_T
#ifndef INT_LEAST64_MAX
#define INT_LEAST64_MAX	(__INT64_C(9223372036854775807))
#endif/*INT_LEAST64_MAX*/
#endif

/* Maximum of unsigned integral types having a minimum size.  */
#ifndef UINT_LEAST8_MAX
#define UINT_LEAST8_MAX	(255)
#endif/*UINT_LEAST8_MAX*/
#ifndef UINT_LEAST16_MAX
#define UINT_LEAST16_MAX	(65535)
#endif/*UINT_LEAST16_MAX*/
#ifndef UINT_LEAST32_MAX
#define UINT_LEAST32_MAX	(4294967295U)
#endif/*UINT_LEAST32_MAX*/
#ifdef _HAVE_UINT64_T
#ifndef UINT_LEAST64_MAX
#define UINT_LEAST64_MAX	(__UINT64_C(18446744073709551615))
#endif/*UINT_LEAST64_MAX*/
#endif

/* Minimum of fast signed integral types having a minimum size.  */
#ifndef INT_FAST8_MIN
#define INT_FAST8_MIN		(-128)
#endif/*INT_FAST8_MIN*/
#ifndef INT_FAST16_MIN
#define INT_FAST16_MIN	(-2147483647-1)
#endif/*INT_FAST16_MIN*/
#ifndef INT_FAST32_MIN
#define INT_FAST32_MIN	(-2147483647-1)
#endif/*INT_FAST32_MIN*/
#ifdef _HAVE_UINT64_T
#ifndef INT_FAST64_MIN
#define INT_FAST64_MIN		(-__INT64_C(9223372036854775807)-1)
#endif/*INT_FAST64_MIN*/
#endif

/* Maximum of fast signed integral types having a minimum size.  */
#ifndef INT_FAST8_MAX
#define INT_FAST8_MAX		(127)
#endif/*INT_FAST8_MAX*/
#ifndef INT_FAST16_MAX
#define INT_FAST16_MAX	(2147483647)
#endif/*INT_FAST16_MAX*/
#ifndef INT_FAST32_MAX
#define INT_FAST32_MAX	(2147483647)
#endif/*INT_FAST32_MAX*/
#ifdef _HAVE_UINT64_T
#ifndef INT_FAST64_MAX
#define INT_FAST64_MAX		(__INT64_C(9223372036854775807))
#endif/*INT_FAST64_MAX*/
#endif

/* Maximum of fast unsigned integral types having a minimum size.  */
#ifndef UINT_FAST8_MAX
#define UINT_FAST8_MAX		(255)
#endif/*UINT_FAST8_MAX*/
#ifndef UINT_FAST16_MAX
#define UINT_FAST16_MAX	(4294967295U)
#endif/*UINT_FAST16_MAX*/
#ifndef UINT_FAST32_MAX
#define UINT_FAST32_MAX	(4294967295U)
#endif/*UINT_FAST32_MAX*/
#ifdef _HAVE_UINT64_T
#ifndef UINT_FAST64_MAX
#define UINT_FAST64_MAX	(__UINT64_C(18446744073709551615))
#endif/*UINT_FAST64_MAX*/
#endif

/* Values to test for integral types holding void * pointer.  */
EOF
case "$ac_sizeof_voidp" in
  8) cat >>$ac_stdint_h <<EOF
#ifndef INTPTR_MIN
#define INTPTR_MIN		(-9223372036854775807L-1)
#endif/*INTPTR_MIN*/
#ifndef INTPTR_MAX
#define INTPTR_MAX		(9223372036854775807L)
#endif/*INTPTR_MAX*/
#ifndef UINTPTR_MAX
#define UINTPTR_MAX		(18446744073709551615UL)
#endif/*UINTPTR_MAX*/
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifndef INTPTR_MIN
#define INTPTR_MIN		(-2147483647-1)
#endif/*INTPTR_MIN*/
#ifndef INTPTR_MAX
#define INTPTR_MAX		(2147483647)
#endif/*INTPTR_MAX*/
#ifndef UINTPTR_MAX
#define UINTPTR_MAX		(4294967295U)
#endif/*UINTPTR_MAX*/
EOF
;;
esac
cat >>$ac_stdint_h <<EOF

/* Minimum for largest signed integral type.  */
#ifdef _HAVE_UINT64_T
#ifndef INTMAX_MIN
#define INTMAX_MIN		(-__INT64_C(9223372036854775807)-1)
#endif/*INTMAX_MIN*/
#else
#ifndef INTMAX_MIN
#define INTMAX_MIN		(-2147483647-1)
#endif/*INTMAX_MIN*/
#endif

/* Maximum for largest signed integral type.  */
#ifdef _HAVE_UINT64_T
#ifndef INTMAX_MAX
#define INTMAX_MAX		(__INT64_C(9223372036854775807))
#endif/*INTMAX_MAX*/
#else
#ifndef INTMAX_MAX
#define INTMAX_MAX		(2147483647)
#endif/*INTMAX_MAX*/
#endif

/* Maximum for largest unsigned integral type.  */
#ifdef _HAVE_UINT64_T
#ifndef UINTMAX_MAX
#define UINTMAX_MAX		(__UINT64_C(18446744073709551615))
#endif/*UINTMAX_MAX*/
#else
#ifndef UINTMAX_MAX
#define UINTMAX_MAX		(4294967295U)
#endif/*UINTMAX_MAX*/
#endif

/* Limits of other integer types.  */

/* Limits of ptrdiff_t type.  */
EOF
case "$ac_sizeof_voidp" in
  8) cat >>$ac_stdint_h <<EOF
#ifndef PTRDIFF_MIN
#define PTRDIFF_MIN		(-9223372036854775807L-1)
#endif/*PTRDIFF_MIN*/
#ifndef PTRDIFF_MAX
#define PTRDIFF_MAX		(9223372036854775807L)
#endif/*PTRDIFF_MAX*/
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifndef PTRDIFF_MIN
#define PTRDIFF_MIN		(-2147483647-1)
#endif/*PTRDIFF_MIN*/
#ifndef PTRDIFF_MAX
#define PTRDIFF_MAX		(2147483647)
#endif/*PTRDIFF_MAX*/
EOF
;;
esac
cat >>$ac_stdint_h <<EOF

/* Limits of sig_atomic_t.  */
#ifndef SIG_ATOMIC_MIN
#define SIG_ATOMIC_MIN		(-2147483647-1)
#endif/*SIG_ATOMIC_MIN*/
#ifndef SIG_ATOMIC_MAX
#define SIG_ATOMIC_MAX		(2147483647)
#endif/*SIG_ATOMIC_MAX*/

/* Limit of size_t type.  */
EOF
case "$ac_sizeof_voidp" in
  8) cat >>$ac_stdint_h <<EOF
#ifndef SIZE_MAX
#define SIZE_MAX		(18446744073709551615UL)
#endif/*SIZE_MAX*/
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifndef SIZE_MAX
#define SIZE_MAX		(4294967295U)
#endif/*SIZE_MAX*/
EOF
;;
esac
cat >>$ac_stdint_h <<EOF

#endif	/* C++ && limit macros */

EOF

cat >>$ac_stdint_h <<EOF
/* The ISO C99 standard specifies that in C++ implementations these
   should only be defined if explicitly requested.  */
#if !defined __cplusplus || defined __STDC_CONSTANT_MACROS

/* Signed.  */
#ifndef INT8_C
#define INT8_C(c)	c
#endif/*INT8_C*/
#ifndef INT16_C
#define INT16_C(c)	c
#endif/*INT16_C*/
#ifndef INT32_C
#define INT32_C(c)	c
#endif/*INT32_C*/
#ifdef _HAVE_UINT64_T
EOF
case "$ac_sizeof_long" in
  8) cat >>$ac_stdint_h <<EOF
#ifndef INT64_C
#define INT64_C(c)	c ## L
#endif/*INT64_C*/
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifndef INT64_C
#define INT64_C(c)	c ## LL
#endif/*INT64_C*/
EOF
;;
esac
cat >>$ac_stdint_h <<EOF
#endif

/* Unsigned.  */
#ifndef UINT8_C
#define UINT8_C(c)	c ## U
#endif/*UINT8_C*/
#ifndef UINT16_C
#define UINT16_C(c)	c ## U
#endif/*UINT16_C*/
#ifndef UINT32_C
#define UINT32_C(c)	c ## U
#endif/*UINT32_C*/
#ifdef _HAVE_UINT64_T
EOF
case "$ac_sizeof_long" in
  8) cat >>$ac_stdint_h <<EOF
#ifndef UINT64_C
#define UINT64_C(c)	c ## UL
#endif/*UINT64_C*/
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifndef UINT64_C
#define UINT64_C(c)	c ## ULL
#endif/*UINT64_C*/
EOF
;;
esac
cat >>$ac_stdint_h <<EOF
#endif

/* Maximal type.  */
EOF
case "$ac_sizeof_long" in
  8) cat >>$ac_stdint_h <<EOF
#ifndef INTMAX_C
#define INTMAX_C(c)	c ## L
#endif/*INTMAX_C*/
#ifndef UINTMAX_C
#define UINTMAX_C(c)	c ## UL
#endif/*UINTMAX_C*/
EOF
;;
  *) cat >>$ac_stdint_h <<EOF
#ifndef INTMAX_C
#define INTMAX_C(c)	c ## LL
#endif/*INTMAX_C*/
#ifndef UINTMAX_C
#define UINTMAX_C(c)	c ## ULL
#endif/*UINTMAX_C*/

EOF
;;
esac

cat >>$ac_stdint_h <<EOF
#endif	/* C++ && constant macros */

EOF

# [CG]------------- DONE INTx_... defines -------------

AC_MSG_RESULT(... DONE $ac_stdint_h)
   cat >>$ac_stdint_h <<EOF

  /* once */
#endif
#endif
EOF
])


dnl @synopsis AX_CFLAGS_GCC_OPTION (optionflag [,[shellvar][,[A][,[NA]]])
dnl
dnl AX_CFLAGS_GCC_OPTION(-fvomit-frame) would show a message as like
dnl "checking CFLAGS for gcc -fvomit-frame ... yes" and adds the
dnl optionflag to CFLAGS if it is understood. You can override the
dnl shellvar-default of CFLAGS of course. The order of arguments stems
dnl from the explicit macros like AX_CFLAGS_WARN_ALL.
dnl
dnl The cousin AX_CXXFLAGS_GCC_OPTION would check for an option to add
dnl to CXXFLAGS - and it uses the autoconf setup for C++ instead of C
dnl (since it is possible to use different compilers for C and C++).
dnl
dnl The macro is a lot simpler than any special AX_CFLAGS_* macro (or
dnl ac_cxx_rtti.m4 macro) but allows to check for arbitrary options.
dnl However, if you use this macro in a few places, it would be great
dnl if you would make up a new function-macro and submit it to the
dnl ac-archive.
dnl
dnl   - $1 option-to-check-for : required ("-option" as non-value)
dnl   - $2 shell-variable-to-add-to : CFLAGS (or CXXFLAGS in the other case)
dnl   - $3 action-if-found : add value to shellvariable
dnl   - $4 action-if-not-found : nothing
dnl
dnl note: in earlier versions, $1-$2 were swapped. We try to detect the
dnl situation and accept a $2=~/-/ as being the old
dnl option-to-check-for.
dnl
dnl also: there are other variants that emerged from the original macro
dnl variant which did just test an option to be possibly added.
dnl However, some compilers accept an option silently, or possibly for
dnl just another option that was not intended. Therefore, we have to do
dnl a generic test for a compiler family. For gcc we check "-pedantic"
dnl being accepted which is also understood by compilers who just want
dnl to be compatible with gcc even when not being made from gcc
dnl sources.
dnl
dnl see also:
dnl
dnl       AX_CFLAGS_SUN_OPTION               AX_CFLAGS_HPUX_OPTION
dnl       AX_CFLAGS_AIX_OPTION               AX_CFLAGS_IRIX_OPTION
dnl
dnl @category C
dnl @author Guido Draheim <guidod@gmx.de>
dnl @version 2003-11-04
dnl @license GPLWithACException

AC_DEFUN([AX_CFLAGS_GCC_OPTION_OLD], [dnl
AS_VAR_PUSHDEF([FLAGS],[CFLAGS])dnl
AS_VAR_PUSHDEF([VAR],[ac_cv_cflags_gcc_option_$2])dnl
AC_CACHE_CHECK([m4_ifval($1,$1,FLAGS) for gcc m4_ifval($2,$2,-option)],
VAR,[VAR="no, unknown"
 AC_LANG_SAVE
 AC_LANG_C
 ac_save_[]FLAGS="$[]FLAGS"
for ac_arg dnl
in "-pedantic  % m4_ifval($2,$2,-option)"  dnl   GCC
   #
do FLAGS="$ac_save_[]FLAGS "`echo $ac_arg | sed -e 's,%%.*,,' -e 's,%,,'`
   AC_TRY_COMPILE([],[return 0;],
   [VAR=`echo $ac_arg | sed -e 's,.*% *,,'` ; break])
done
 FLAGS="$ac_save_[]FLAGS"
 AC_LANG_RESTORE
])
case ".$VAR" in
     .ok|.ok,*) m4_ifvaln($3,$3) ;;
   .|.no|.no,*) m4_ifvaln($4,$4) ;;
   *) m4_ifvaln($3,$3,[
   if echo " $[]m4_ifval($1,$1,FLAGS) " | grep " $VAR " 2>&1 >/dev/null
   then AC_RUN_LOG([: m4_ifval($1,$1,FLAGS) does contain $VAR])
   else AC_RUN_LOG([: m4_ifval($1,$1,FLAGS)="$m4_ifval($1,$1,FLAGS) $VAR"])
                      m4_ifval($1,$1,FLAGS)="$m4_ifval($1,$1,FLAGS) $VAR"
   fi ]) ;;
esac
AS_VAR_POPDEF([VAR])dnl
AS_VAR_POPDEF([FLAGS])dnl
])


dnl the only difference - the LANG selection... and the default FLAGS

AC_DEFUN([AX_CXXFLAGS_GCC_OPTION_OLD], [dnl
AS_VAR_PUSHDEF([FLAGS],[CXXFLAGS])dnl
AS_VAR_PUSHDEF([VAR],[ac_cv_cxxflags_gcc_option_$2])dnl
AC_CACHE_CHECK([m4_ifval($1,$1,FLAGS) for gcc m4_ifval($2,$2,-option)],
VAR,[VAR="no, unknown"
 AC_LANG_SAVE
 AC_LANG_CXX
 ac_save_[]FLAGS="$[]FLAGS"
for ac_arg dnl
in "-pedantic  % m4_ifval($2,$2,-option)"  dnl   GCC
   #
do FLAGS="$ac_save_[]FLAGS "`echo $ac_arg | sed -e 's,%%.*,,' -e 's,%,,'`
   AC_TRY_COMPILE([],[return 0;],
   [VAR=`echo $ac_arg | sed -e 's,.*% *,,'` ; break])
done
 FLAGS="$ac_save_[]FLAGS"
 AC_LANG_RESTORE
])
case ".$VAR" in
     .ok|.ok,*) m4_ifvaln($3,$3) ;;
   .|.no|.no,*) m4_ifvaln($4,$4) ;;
   *) m4_ifvaln($3,$3,[
   if echo " $[]m4_ifval($1,$1,FLAGS) " | grep " $VAR " 2>&1 >/dev/null
   then AC_RUN_LOG([: m4_ifval($1,$1,FLAGS) does contain $VAR])
   else AC_RUN_LOG([: m4_ifval($1,$1,FLAGS)="$m4_ifval($1,$1,FLAGS) $VAR"])
                      m4_ifval($1,$1,FLAGS)="$m4_ifval($1,$1,FLAGS) $VAR"
   fi ]) ;;
esac
AS_VAR_POPDEF([VAR])dnl
AS_VAR_POPDEF([FLAGS])dnl
])

dnl -------------------------------------------------------------------------

AC_DEFUN([AX_CFLAGS_GCC_OPTION_NEW], [dnl
AS_VAR_PUSHDEF([FLAGS],[CFLAGS])dnl
AS_VAR_PUSHDEF([VAR],[ac_cv_cflags_gcc_option_$1])dnl
AC_CACHE_CHECK([m4_ifval($2,$2,FLAGS) for gcc m4_ifval($1,$1,-option)],
VAR,[VAR="no, unknown"
 AC_LANG_SAVE
 AC_LANG_C
 ac_save_[]FLAGS="$[]FLAGS"
for ac_arg dnl
in "-pedantic  % m4_ifval($1,$1,-option)"  dnl   GCC
   #
do FLAGS="$ac_save_[]FLAGS "`echo $ac_arg | sed -e 's,%%.*,,' -e 's,%,,'`
   AC_TRY_COMPILE([],[return 0;],
   [VAR=`echo $ac_arg | sed -e 's,.*% *,,'` ; break])
done
 FLAGS="$ac_save_[]FLAGS"
 AC_LANG_RESTORE
])
case ".$VAR" in
     .ok|.ok,*) m4_ifvaln($3,$3) ;;
   .|.no|.no,*) m4_ifvaln($4,$4) ;;
   *) m4_ifvaln($3,$3,[
   if echo " $[]m4_ifval($2,$2,FLAGS) " | grep " $VAR " 2>&1 >/dev/null
   then AC_RUN_LOG([: m4_ifval($2,$2,FLAGS) does contain $VAR])
   else AC_RUN_LOG([: m4_ifval($2,$2,FLAGS)="$m4_ifval($2,$2,FLAGS) $VAR"])
                      m4_ifval($2,$2,FLAGS)="$m4_ifval($2,$2,FLAGS) $VAR"
   fi ]) ;;
esac
AS_VAR_POPDEF([VAR])dnl
AS_VAR_POPDEF([FLAGS])dnl
])


dnl the only difference - the LANG selection... and the default FLAGS

AC_DEFUN([AX_CXXFLAGS_GCC_OPTION_NEW], [dnl
AS_VAR_PUSHDEF([FLAGS],[CXXFLAGS])dnl
AS_VAR_PUSHDEF([VAR],[ac_cv_cxxflags_gcc_option_$1])dnl
AC_CACHE_CHECK([m4_ifval($2,$2,FLAGS) for gcc m4_ifval($1,$1,-option)],
VAR,[VAR="no, unknown"
 AC_LANG_SAVE
 AC_LANG_CXX
 ac_save_[]FLAGS="$[]FLAGS"
for ac_arg dnl
in "-pedantic  % m4_ifval($1,$1,-option)"  dnl   GCC
   #
do FLAGS="$ac_save_[]FLAGS "`echo $ac_arg | sed -e 's,%%.*,,' -e 's,%,,'`
   AC_TRY_COMPILE([],[return 0;],
   [VAR=`echo $ac_arg | sed -e 's,.*% *,,'` ; break])
done
 FLAGS="$ac_save_[]FLAGS"
 AC_LANG_RESTORE
])
case ".$VAR" in
     .ok|.ok,*) m4_ifvaln($3,$3) ;;
   .|.no|.no,*) m4_ifvaln($4,$4) ;;
   *) m4_ifvaln($3,$3,[
   if echo " $[]m4_ifval($2,$2,FLAGS) " | grep " $VAR " 2>&1 >/dev/null
   then AC_RUN_LOG([: m4_ifval($2,$2,FLAGS) does contain $VAR])
   else AC_RUN_LOG([: m4_ifval($2,$2,FLAGS)="$m4_ifval($2,$2,FLAGS) $VAR"])
                      m4_ifval($2,$2,FLAGS)="$m4_ifval($2,$2,FLAGS) $VAR"
   fi ]) ;;
esac
AS_VAR_POPDEF([VAR])dnl
AS_VAR_POPDEF([FLAGS])dnl
])

AC_DEFUN([AX_CFLAGS_GCC_OPTION],[ifelse(m4_bregexp([$2],[-]),-1,
[AX_CFLAGS_GCC_OPTION_NEW($@)],[AX_CFLAGS_GCC_OPTION_OLD($@)])])

AC_DEFUN([AX_CXXFLAGS_GCC_OPTION],[ifelse(m4_bregexp([$2],[-]),-1,
[AX_CXXFLAGS_GCC_OPTION_NEW($@)],[AX_CXXFLAGS_GCC_OPTION_OLD($@)])])
