#ifndef AIR_h_INCLUDED
#define AIR_h_INCLUDED

/*
 * !!!!	AIR.xcc
 *
 * Author: Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2001 - 2007 STMicroelectronics
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CCL.h"
#include "ECL.h"
#include "MDS/MDS.h"

/*
 * The AIR (Assembler Intermediate Representation) library contains the following modules:
 * *	<Label Label.html> Label for code.
 * *	<Symbol Symbol.html> Symbol for code and data.
 * *	<Constant Constant.html> Constant in the code stream.
 * *	<Literal Literal.html> Literal, should be merged with Constant.
 * *	<Instruction Instruction.html> Instruction in the code stream.
 * *	<Pipeline Pipeline.html> Cycle-Based instruction scheduler.
 * *	<Scoreboard Scoreboard.html> Scoreboard instruction scheduler.
 * *	<NativeBlock NativeBlock.html> Native code block.
 * *	<NativeStream NativeStream.html> Native code stream.
 * *	<InterBlock InterBlock.html> Inter-Block scoreboard scheduler.
 * *	<Trampoline Trampoline.html> Trampoline for JIT (to do).
 */

/*
 * Include all the exported AIR headers.
 */
#include "AIR/Label.h"
#include "AIR/Symbol.h"
#include "AIR/Constant.h"
#include "AIR/Literal.h"
#include "AIR/Instruction.h"
#include "AIR/Pipeline.h"
#include "AIR/Scoreboard.h"
#include "AIR/NativeBlock.h"
#include "AIR/NativeStream.h"
#include "AIR/InterBlock.h"
#include "AIR/Trampoline.h"

/*
 * AIR_TRACE --	File handle used to print the AIR traces.
 */
extern FILE *AIR_TRACE;
#ifdef ECL_NTRACE
#define AIR_TRACE NULL
#endif//NULL

//
extern int FORCE_PIPELINE;
extern int FORCE_ONLYCOUNT;

/*
 * AIR_INIT --	Initialize the AIR library.
 */
void
AIR_INIT(void);

/*
 * AIR_FINI --	Finalize the AIR library.
 */
void
AIR_FINI(void);


#endif//AIR_h_INCLUDED
