#ifndef Instruction__h_INCLUDED
#define Instruction__h_INCLUDED

//
#include "AIR/Instruction.h"

/*
 * Instruction_relocate --	Relocate this Instruction.
 */
void
Instruction_relocate(Instruction this, PlatformUInt nativePC,
                     InstructionRelocationFixup fixup,
                     void* userData, NativeStream stream);

/*
 * Instruction_encode --	Encode this Instruction in the supplied buffer.
 */
#if !Bundle_ISSUE_MAX
static inline void
Instruction_encode(const_Instruction this, void *buffer)
{
  char *encode = buffer;
  Instance instance = Instruction_INSTANCE(this);
  int increment = Instance_increment(instance), i;
  const char *opcodes = (const void *)Instruction_OPCODES(this);
  while (increment--) *encode++ = *opcodes++;
}
#endif//!Bundle_ISSUE_MAX

/*
 * Instruction_decodeOperands --	Decode the Operands of this Instruction.
 */
void
Instruction_decodeOperands(const_Instruction this, OperandsBuffer buffer);

/*
 * InstructionGroup_addInstruction --	Add an Instruction to this InstructionGroup.
 */
static inline void
InstructionGroup_addInstruction(InstructionGroup this, Instruction instruction)
{
#if Bundle_ISSUE_MAX
  int count = InstructionGroup_COUNT(this);
  Bundling bundling = Instruction_bundling(instruction);
  InstructionGroup_INSTRUCTIONS(this)[count] = instruction;
  InstructionGroup_BUNDLINGS(this)[count] = bundling;
  *InstructionGroup__COUNT(this) = count + 1;
  Except_REQUIRE(count < Bundle_ISSUE_MAX);
#else//!Bundle_ISSUE_MAX
  Except_CHECK(InstructionGroup_INSTRUCTION(this) == NULL);
  *InstructionGroup__INSTRUCTION(this) = instruction;
#endif//Bundle_ISSUE_MAX
}

/*
 * InstructionGroup_findTemplate --	Find a Template for this InstructionGroup.
 */
Template
InstructionGroup_findTemplate(InstructionGroup this, PlatformUInt nativePC);

/*
 * InstructionGroup_encode --	Encode this InstructionGroup in the given buffer.
 */
void
InstructionGroup_encode(InstructionGroup this, PlatformUInt nativePC, void *buffer,
                        InstructionRelocationFixup fixup,
                        void* userData, NativeStream stream);

/*
 * Instruction_INIT --	Initialize the Instruction module.
 */
#define Instruction_INIT()

/*
 * Instruction_FINI --	Finalize the Instruction module.
 */
#define Instruction_FINI()


#endif//Instruction__h_INCLUDED
