#ifndef Pipeline_c_INCLUDED
#define Pipeline_c_INCLUDED

#line 15 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
#include "AIR/AIR_.h"

#line 20 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 47 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
PipelineEffect
PipelineEffect_Ctor(PipelineEffect this,
                    Instruction instruction, InstructionAction action,
                    PipelineEffectFlags flags)
{
  *PipelineEffect__FLAGS(this) = flags;
  *PipelineEffect__INSTANCE(this) = Instruction_INSTANCE(instruction);
  *PipelineEffect__INSTRUCTION(this) = instruction;
  *PipelineEffect__ACTION(this) = action;
  *PipelineEffect__NEXT(this) = NULL;
  *PipelineEffect__PRED(this) = NULL;
  *PipelineEffect__SUCC(this) = NULL;
  return this;
}

#line 62 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
size_t
PipelineEffect_Size(Instruction instruction, InstructionAction action,
                    PipelineEffectFlags flags)
{
  return sizeof(PipelineEffect_);
}

#line 80 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
int
PipelineEffect_latencyRAW(const_PipelineEffect this, const_PipelineEffect that)
{
  Instance this_instance = PipelineEffect_INSTANCE(this);
  Instance that_instance = PipelineEffect_INSTANCE(that);
  InstructionAction this_action = PipelineEffect_ACTION(this);
  InstructionAction that_action = PipelineEffect_ACTION(that);
  StorageCell cell = InstructionAction_CELL(this_action);
  int action_RAW = InstructionAction_RAW(this_action);
  int latency = InstructionAction_STAGE(this_action)
              - InstructionAction_STAGE(that_action)
              + action_RAW;
  if (Instance_fixupRAW(this_instance, that_instance)) {
    latency += Instance_latencyRAW(this_instance, that_instance, cell);
  }
  Except_CHECK(PipelineEffect_isWrite(this));
  Except_CHECK(PipelineEffect_isRead(that));
  Except_CHECK(latency >= 0);
  return latency;
}

#line 107 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
int
PipelineEffect_latencyWAR(const_PipelineEffect this, const_PipelineEffect that)
{
  InstructionAction this_action = PipelineEffect_ACTION(this);
  int latency = InstructionAction_WAR(this_action);
  Except_CHECK(PipelineEffect_isRead(this));
  Except_CHECK(PipelineEffect_isWrite(that));
  Except_CHECK(latency >= 0);
  return latency;
}

#line 124 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
int
PipelineEffect_latencyWAW(const_PipelineEffect this, const_PipelineEffect that)
{
  InstructionAction this_action = PipelineEffect_ACTION(this);
  InstructionAction that_action = PipelineEffect_ACTION(that);
  int action_WAW = InstructionAction_WAW(this_action);
  int latency = InstructionAction_STAGE(this_action)
              - InstructionAction_STAGE(that_action);
  if (latency < action_WAW) latency = action_WAW;
  Except_CHECK(PipelineEffect_isWrite(this));
  Except_CHECK(PipelineEffect_isWrite(that));
  Except_CHECK(latency >= 0);
  return latency;
}

#line 139 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 150 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
PipelineEffectList
PipelineEffectList_Ctor(PipelineEffectList this)
{
  *PipelineEffectList__FIRST(this) = NULL;
  *PipelineEffectList__LAST(this) = NULL;
  return this;
}

#line 169 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
bool
PipelineEffectList_isEmpty(const_PipelineEffectList this)
{
  return PipelineEffectList_FIRST(this) == NULL;
}

#line 181 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
PipelineEffect
PipelineEffectList_first(const_PipelineEffectList this)
{
  return PipelineEffectList_FIRST(this);
}

#line 205 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
PipelineEffect
PipelineEffectList_push_(PipelineEffectList this, PipelineEffect effect)
{
  PipelineEffect first = PipelineEffectList_FIRST(this);
  PipelineEffect last = PipelineEffectList_LAST(this);
  if (first == NULL) *PipelineEffectList__FIRST(this) = effect;
  if (last != NULL) *PipelineEffect__NEXT(last) = effect;
  *PipelineEffectList__LAST(this) = effect;
  return effect;
}

#line 216 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 231 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
PipelineADL
PipelineADL_Ctor(PipelineADL this,
                 Memory memory, InstructionStack instructions)
{
  int cellCount = StorageCell__COUNT, cell;
  PipelineEffectList_ *lists = Memory_alloc(memory, sizeof(PipelineEffectList_)*cellCount);
  for (cell = 0; cell < StorageCell__COUNT; cell++) {
    PipelineEffectList_Ctor(lists + cell);
  }
  *PipelineADL__MEMORY(this) = memory;
  *PipelineADL__INSTRUCTIONS(this) = instructions;
  *PipelineADL__LISTS(this) = lists;
  //
  InstructionStack_FOREACH_Instruction(instructions, instruction) {
    InstructionAction_ *actions = Instruction_ACTIONS(instruction);
    int writeCount = Instruction_WRITECOUNT(instruction);
    int readCount = Instruction_READCOUNT(instruction), i;
    size_t effectsSize = (writeCount + readCount)*sizeof(PipelineEffect_);
    PipelineEffect_ *effects = Memory_alloc(memory, effectsSize);
    // Enter the Read effects.
    for (i = writeCount; i < writeCount + readCount; i++) {
      StorageCell cell = InstructionAction_CELL(actions + i);
      PipelineEffect effect = PipelineEffect_Ctor(effects + i, instruction,
                                                  actions + i, 0);
      PipelineEffectList_push_(lists + cell, effect);
    }
    // Enter the Write effects.
    for (i = 0; i < writeCount; i++) {
      StorageCell cell = InstructionAction_CELL(actions + i);
      PipelineEffect effect = PipelineEffect_Ctor(effects + i, instruction,
                                                  actions + i, PipelineEffectFlag_Write);
      PipelineEffectList_push_(lists + cell, effect);
    }
    *Instruction__EFFECTS(instruction) = effects;
  } InstructionStack_ENDEACH_Instruction;
  //
  PipelineADL_linkEffects(this);
  //
  return this;
}

#line 272 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
void
PipelineADL_Dtor(PipelineADL this)
{
  Memory memory = PipelineADL_MEMORY(this);
  Memory_free_(memory, PipelineADL_LISTS(this));
}

#line 290 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
PipelineEffectList
PipelineADL_lists(PipelineADL this)
{
  return PipelineADL_LISTS(this);
}

#line 317 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
void
PipelineADL_linkEffects(PipelineADL this)
{
  PipelineADL_FOREACH_PipelineEffectList(this, list) {
    PipelineEffect write_effect = NULL;
    PipelineEffectList_FOREACH_PipelineEffect(list, effect) {
      // Set PRED to write_effect the Write predecessor.
      *PipelineEffect__PRED(effect) = write_effect;
      if (PipelineEffect_isWrite(effect)) {
        if (write_effect != NULL) {
          // Set SUCC of each predecessor to this effect.
          PipelineEffect pred_effect = write_effect;
          do {
            *PipelineEffect__SUCC(pred_effect) = effect;
            pred_effect = PipelineEffect_NEXT(pred_effect);
          } while (pred_effect != effect);
        }
        write_effect = effect;
      }
    } PipelineEffectList_ENDEACH_PipelineEffect;
  } PipelineADL_ENDEACH_PipelineEffectList;
}

#line 346 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
#ifndef ECL_NTRACE
bool
PipelineADL_pretty(PipelineADL this, FILE *file)
{
  if (file == NULL) file = stderr;
  fprintf(file, "*** PipelineADL:\n");
  PipelineADL_FOREACH_PipelineEffectList(this, list) {
    if (!PipelineEffectList_isEmpty(list)) {
      fprintf(file, "    %s:\n", StorageCell_name(PipelineADL_CELL));
      PipelineEffectList_FOREACH_PipelineEffect(list, effect) {
        PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
        PipelineEffect next_effect = PipelineEffect_NEXT(effect);
        Instruction instruction = PipelineEffect_INSTRUCTION(effect);
        if (PipelineEffect_isWrite(effect)) {
          fprintf(file, "\tWrite\t");
        } else {
          fprintf(file, "\tRead\t");
        }
        Instruction_pretty(instruction, file, NULL);
        if (PipelineEffect_isWrite(effect)) {
          while (next_effect != NULL) {
            if (PipelineEffect_isWrite(next_effect)) {
              fprintf(file, "\tWAW(%d)", PipelineEffect_latencyWAW(effect, next_effect));
              Except_CHECK(PipelineEffect_SUCC(effect) == next_effect);
              Except_CHECK(PipelineEffect_PRED(next_effect) == effect);
            } else {
              fprintf(file, "\tRAW(%d)", PipelineEffect_latencyRAW(effect, next_effect));
            }
            if (next_effect == succ_effect) break;
            next_effect = PipelineEffect_NEXT(next_effect);
          }
        } else /*PipelineEffect_isRead(effect)*/ {
          while (next_effect != NULL) {
            if (PipelineEffect_isWrite(next_effect)) {
              fprintf(file, "\tWAR(%d)", PipelineEffect_latencyWAR(effect, next_effect));
            }
            if (next_effect == succ_effect) break;
            next_effect = PipelineEffect_NEXT(next_effect);
          }
        }
        fprintf(file, "\n");
      } PipelineEffectList_ENDEACH_PipelineEffect;
    }
  } PipelineADL_ENDEACH_PipelineEffectList;
  return true;
}
#endif//ECL_NTRACE

#line 394 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 416 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
#define Pipeline_DUALQUEUE 0

#line 418 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
Pipeline
Pipeline_Ctor(Pipeline this,
              Memory memory, Processor processor, InstructionStack instructions)
{
  size_t tableSize = ResourceTable_Size(processor, 0);
  ResourceTable_ *table = Memory_alloc(memory, tableSize);
  int32_t instructionCount = InstructionStack_count(instructions);
  PipelineADL_Ctor(Pipeline_ADL(this), memory, instructions);
#if Pipeline_DUALQUEUE
  PQueue_Ctor(Pipeline_WAITING(this), memory, instructionCount);
#endif
  PQueue_Ctor(Pipeline_RELEASED(this), memory, instructionCount);
  *Pipeline__TABLE(this) = ResourceTable_Ctor(table, processor, 0);
  return this;
}

#line 434 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
void
Pipeline_Dtor(Pipeline this)
{
  Memory memory = Pipeline_MEMORY(this);
  PipelineADL_Dtor(Pipeline_ADL(this));
#if Pipeline_DUALQUEUE
  PQueue_Dtor(Pipeline_WAITING(this));
#endif
  PQueue_Dtor(Pipeline_RELEASED(this));
  Memory_free_(memory, Pipeline_TABLE(this));
}

#line 457 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
const_InstructionStack
Pipeline_instructions(const_Pipeline this)
{
  return Pipeline_INSTRUCTIONS(this);
}

#line 491 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
void
Pipeline_makeCritical(Pipeline this)
{
  Pipeline_FORBACK_Instruction(this, instruction) {
    int critical = Instruction_CRITICAL(instruction);
    PipelineEffect_ *effects = Instruction_EFFECTS(instruction);
    int writeCount = Instruction_WRITECOUNT(instruction);
    int readCount = Instruction_READCOUNT(instruction), i;
    // Process the Write effects.
    for (i = 0; i < writeCount; i++) {
      PipelineEffect effect = effects + i;
      PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
      PipelineEffect next_effect = PipelineEffect_NEXT(effect);
      while (next_effect != NULL) {
        Instruction next_instruction = PipelineEffect_INSTRUCTION(next_effect);
        if (next_instruction != instruction) {
          int latency = PipelineEffect_isWrite(next_effect)?
                        PipelineEffect_latencyWAW(effect, next_effect):
                        PipelineEffect_latencyRAW(effect, next_effect);
          int new_critical = Instruction_CRITICAL(next_instruction) + latency;
          if (critical < new_critical) critical = new_critical;
          ++*Instruction__PREDCOUNT(next_instruction);
        }
        if (next_effect == succ_effect) break;
        next_effect = PipelineEffect_NEXT(next_effect);
      }
    }
    // Process the Read effects.
    for (i = writeCount; i < writeCount + readCount; i++) {
      PipelineEffect effect = effects + i;
      PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
      PipelineEffect next_effect = PipelineEffect_NEXT(effect);
      while (next_effect != NULL) {
        Instruction next_instruction = PipelineEffect_INSTRUCTION(next_effect);
        if (next_instruction != instruction && PipelineEffect_isWrite(next_effect)) {
          int latency = PipelineEffect_latencyWAR(effect, next_effect);
          int new_critical = Instruction_CRITICAL(next_instruction) + latency;
          if (critical < new_critical) critical = new_critical;
          ++*Instruction__PREDCOUNT(next_instruction);
        }
        if (next_effect == succ_effect) break;
        next_effect = PipelineEffect_NEXT(next_effect);
      }
    }
    // Update the Instruction CRITICAL.
    *Instruction__CRITICAL(instruction) = critical;
  } Pipeline_ENDBACK_Instruction;
}

#line 546 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
#if Pipeline_DUALQUEUE
void
Pipeline_schedule(Pipeline this)
{
  int timeSlot = -1;
  PQueue waiting = Pipeline_WAITING(this);
  PQueue released = Pipeline_RELEASED(this);
  ResourceTable table = Pipeline_TABLE(this);
  AIR_TRACE && fprintf(AIR_TRACE, "*** Pipeline schedule %d instructions\n",
                       InstructionStack_count(Pipeline_INSTRUCTIONS(this)));
  // Enter Instruction(s) with zero PREDCOUNT into the waiting queue.
  Pipeline_FOREACH_Instruction(this, instruction) {
    if (Instruction_PREDCOUNT(instruction) == 0) {
      int release = Instruction_RELEASE(instruction);
      PQueue_addEntry(waiting, instruction, release);
      Except_CHECK(release >= 0);
    }
  } Pipeline_ENDEACH_Instruction;
  PQueue_makeHeap(waiting);
  // Main scheduling loop.
  while (1) {
    // Fill the released queue with the Instruction(s) data-ready at timeSlot.
    while (!PQueue_isEmpty(waiting)) {
      Instruction instruction = PQueue_extractMin(waiting);
      int release = Instruction_RELEASE(instruction);
      if (PQueue_isEmpty(released)) {
        // Move to next timeSlot.
        timeSlot = release;
        ResourceTable_advance(table, timeSlot);
      }
      if (release == timeSlot) {
        // Instruction(s) relased at timeSlot are put into the released queue.
        int critical = Instruction_CRITICAL(instruction);
        PQueue_insert(released, instruction, -critical&0xFFFF);
      } else {
        // Put back instruction into the waiting queue.
        PQueue_insert(waiting, instruction, release);
        break;
      }
    }
    // Try to schedule Instruction(s) in the released queue.
    if (PQueue_isEmpty(released)) return;
    do {
      Instruction instruction = PQueue_extractMin(released);
      Reservation reservation = Instruction_RESERVATION(instruction);
      AIR_TRACE && fprintf(AIR_TRACE, "    time=%d\t", timeSlot);
      if (ResourceTable_fitReservation(table, reservation, timeSlot)) {
        // Enter the instruction successors into the waiting queue.
        PipelineEffect_ *effects = Instruction_EFFECTS(instruction);
        int writeCount = Instruction_WRITECOUNT(instruction);
        int readCount = Instruction_READCOUNT(instruction), i;
        // Process the Read effects.
        for (i = writeCount; i < writeCount + readCount; i++) {
          PipelineEffect effect = effects + i;
          PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
          PipelineEffect next_effect = PipelineEffect_NEXT(effect);
          while (next_effect != NULL) {
            Instruction next_instruction = PipelineEffect_INSTRUCTION(next_effect);
            if (next_instruction != instruction && PipelineEffect_isWrite(next_effect)) {
              int latency = PipelineEffect_latencyWAR(effect, next_effect);
              int next_release = Instruction_RELEASE(next_instruction);
              if (next_release < timeSlot + latency) {
                next_release = timeSlot + latency;
              }
              *Instruction__RELEASE(next_instruction) = next_release;
              if (--*Instruction__PREDCOUNT(next_instruction) == 0) {
                PQueue_insert(waiting, next_instruction, next_release);
              }
            }
            if (next_effect == succ_effect) break;
            next_effect = PipelineEffect_NEXT(next_effect);
          }
        }
        // Process the Write effects.
        for (i = 0; i < writeCount; i++) {
          PipelineEffect effect = effects + i;
          PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
          PipelineEffect next_effect = PipelineEffect_NEXT(effect);
          while (next_effect != NULL) {
            Instruction next_instruction = PipelineEffect_INSTRUCTION(next_effect);
            if (next_instruction != instruction) {
              int latency = PipelineEffect_isWrite(next_effect)?
                            PipelineEffect_latencyWAW(effect, next_effect):
                            PipelineEffect_latencyRAW(effect, next_effect);
              int next_release = Instruction_RELEASE(next_instruction);
              if (next_release < timeSlot + latency) {
                next_release = timeSlot + latency;
              }
              *Instruction__RELEASE(next_instruction) = next_release;
              if (--*Instruction__PREDCOUNT(next_instruction) == 0) {
                PQueue_insert(waiting, next_instruction, next_release);
              }
            }
            if (next_effect == succ_effect) break;
            next_effect = PipelineEffect_NEXT(next_effect);
          }
        }
        // Issue instruction at release.
        *Instruction__STARTDATE(instruction) = timeSlot;
        AIR_TRACE && fprintf(AIR_TRACE, "scheduled\t");
      } else {
        // Put back the instruction in waiting queue at timeSlot + 1.
        *Instruction__RELEASE(instruction) = timeSlot + 1;
        PQueue_insert(waiting, instruction, timeSlot + 1);
        AIR_TRACE && fprintf(AIR_TRACE, "delayed \t");
      }
      AIR_TRACE && Instruction_pretty(instruction, AIR_TRACE, 0)
                && fprintf(AIR_TRACE, "\n");
    } while (!PQueue_isEmpty(released));
  }
}
#else//!Pipeline_DUALQUEUE
void
Pipeline_schedule(Pipeline this)
{
  PQueue released = Pipeline_RELEASED(this);
  ResourceTable table = Pipeline_TABLE(this);
  // Enter Instruction(s) with zero PREDCOUNT into the released queue.
  Pipeline_FOREACH_Instruction(this, instruction) {
    if (Instruction_PREDCOUNT(instruction) == 0) {
      int release = Instruction_RELEASE(instruction);
      int critical = Instruction_CRITICAL(instruction);
      PQueuePriority priority = (-critical&0xFFFF) | (release<<16);
      PQueue_addEntry(released, instruction, priority);
    }
  } Pipeline_ENDEACH_Instruction;
  PQueue_makeHeap(released);
  AIR_TRACE && fprintf(AIR_TRACE, "*** Pipeline schedule %d Instructions\n",
                       InstructionStack_count(Pipeline_INSTRUCTIONS(this)));
  // Instruction scheduling loop.
  while (!PQueue_isEmpty(released)) {
    Instruction instruction = PQueue_extractMin(released);
    Reservation reservation = Instruction_RESERVATION(instruction);
    int timeSlot = Instruction_RELEASE(instruction);
    AIR_TRACE && fprintf(AIR_TRACE, "    time=%d\t", timeSlot);
    ResourceTable_advance(table, timeSlot);
    if (ResourceTable_fitReservation(table, reservation, timeSlot)) {
      // Enter the instruction successors into the released queue.
      PipelineEffect_ *effects = Instruction_EFFECTS(instruction);
      int writeCount = Instruction_WRITECOUNT(instruction);
      int readCount = Instruction_READCOUNT(instruction), i;
      // Process the Read effects.
      for (i = writeCount; i < writeCount + readCount; i++) {
        PipelineEffect effect = effects + i;
        PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
        PipelineEffect next_effect = PipelineEffect_NEXT(effect);
        while (next_effect != NULL) {
          Instruction next_instruction = PipelineEffect_INSTRUCTION(next_effect);
          if (next_instruction != instruction && PipelineEffect_isWrite(next_effect)) {
            int latency = PipelineEffect_latencyWAR(effect, next_effect);
            int next_release = Instruction_RELEASE(next_instruction);
            if (next_release < timeSlot + latency) {
              next_release = timeSlot + latency;
            }
            *Instruction__RELEASE(next_instruction) = next_release;
            if (--*Instruction__PREDCOUNT(next_instruction) == 0) {
              int next_critical = Instruction_CRITICAL(next_instruction);
              PQueuePriority next_priority = (-next_critical&0xFFFF) | (next_release<<16);
              PQueue_insert(released, next_instruction, next_priority);
            }
          }
          if (next_effect == succ_effect) break;
          next_effect = PipelineEffect_NEXT(next_effect);
        }
      }
      // Process the Write effects.
      for (i = 0; i < writeCount; i++) {
        PipelineEffect effect = effects + i;
        PipelineEffect succ_effect = PipelineEffect_SUCC(effect);
        PipelineEffect next_effect = PipelineEffect_NEXT(effect);
        while (next_effect != NULL) {
          Instruction next_instruction = PipelineEffect_INSTRUCTION(next_effect);
          if (next_instruction != instruction) {
            int latency = PipelineEffect_isWrite(next_effect)?
                          PipelineEffect_latencyWAW(effect, next_effect):
                          PipelineEffect_latencyRAW(effect, next_effect);
            int next_release = Instruction_RELEASE(next_instruction);
            if (next_release < timeSlot + latency) {
              next_release = timeSlot + latency;
            }
            *Instruction__RELEASE(next_instruction) = next_release;
            if (--*Instruction__PREDCOUNT(next_instruction) == 0) {
              int next_critical = Instruction_CRITICAL(next_instruction);
              PQueuePriority next_priority = (-next_critical&0xFFFF) | (next_release<<16);
              PQueue_insert(released, next_instruction, next_priority);
            }
          }
          if (next_effect == succ_effect) break;
          next_effect = PipelineEffect_NEXT(next_effect);
        }
      }
      // Issue instruction at timeSlot.
      *Instruction__STARTDATE(instruction) = timeSlot;
      AIR_TRACE && fprintf(AIR_TRACE, "scheduled\t");
    } else {
      // Put back the instruction in the released queue at timeSlot + 1.
      int critical = Instruction_CRITICAL(instruction);
      PQueuePriority priority = (-critical&0xFFFF) | (++timeSlot<<16);
      PQueue_insert(released, instruction, priority);
      *Instruction__RELEASE(instruction) = timeSlot;
      AIR_TRACE && fprintf(AIR_TRACE, "delayed \t");
    }
    AIR_TRACE && Instruction_pretty(instruction, AIR_TRACE, 0) && fprintf(AIR_TRACE, "\n");
  }
}
#endif//Pipeline_DUALQUEUES

#line 759 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
#ifndef ECL_NTRACE
bool
Pipeline_pretty(Pipeline this, FILE *file)
{
  if (file == NULL) file = stderr;
  fprintf(file, "Pipeline:\n");
  Pipeline_FOREACH_Instruction(this, instruction) {
    fprintf(file, "    critical=%d\tpredcount=%d\t",
                  Instruction_CRITICAL(instruction),
                  Instruction_PREDCOUNT(instruction));
    Instruction_pretty(instruction, file, 0);
    fprintf(file, "\n");
  } Pipeline_ENDEACH_Instruction;
  return true;
}
#endif//ECL_NTRACE

#line 777 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 893 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 905 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
#if XCC__C
#include "MDS/MDS_.h"
static void
Pipeline_TEST(void)
{
#include "AIR/Pipeline_.c"
}

#line 913 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Pipeline.xcc"
int
main(int argc, char **argv)
{
  AIR_Test(Pipeline_TEST);
  return 0;
}
#endif


#endif//Pipeline_c_INCLUDED
