#ifndef Pipeline__h_INCLUDED
#define Pipeline__h_INCLUDED

//
#include "AIR/Pipeline.h"

/*
 *
 */
typedef enum {
  PipelineEffectFlag_Write = 0x1,
} enum_PipelineEffectFlag;
typedef uint8_t PipelineEffectFlags;

/*
 * Read or Write effect for the Pipeline Array of Dependency Lists.
 */
struct PipelineEffect_ {
  //@args	Instruction instruction, InstructionAction action,
  //@args	PipelineEffectFlags flags
  PipelineEffectFlags FLAGS;	// This PipelineEffect flags.
  //@access isRead	((PipelineEffect_FLAGS(this) & PipelineEffectFlag_Write) == 0)
  //@access isWrite	((PipelineEffect_FLAGS(this) & PipelineEffectFlag_Write) != 0)
  Instance INSTANCE;		// Cached Instance of INSTRUCTION.
  Instruction INSTRUCTION;		// Owning Instruction of this PipelineEffect.
  InstructionAction ACTION;		// InstructionAction of this PipelineEffect.
  struct PipelineEffect_ *NEXT;		// Next PipelineEffect in the PipelineEffectList.
  struct PipelineEffect_ *PRED;		// Dependence predecessor of this PipelineEffect.
  struct PipelineEffect_ *SUCC;		// Dependence successor of this PipelineEffect.
};
typedef struct PipelineEffect_ PipelineEffect_, *PipelineEffect;
typedef const struct PipelineEffect_ *const_PipelineEffect;
typedef struct PipelineEffect_ * restrict_PipelineEffect;

extern PipelineEffect
PipelineEffect_Ctor(PipelineEffect this, Instruction instruction, InstructionAction action, PipelineEffectFlags flags);

extern PipelineEffect
PipelineEffect_Copy(PipelineEffect this, const_PipelineEffect that);

extern void
PipelineEffect_Dtor(PipelineEffect this);

extern size_t
PipelineEffect_Size(Instruction instruction, InstructionAction action, PipelineEffectFlags flags);

#define PipelineEffect_FLAGS(this) ((this)->FLAGS)
#define PipelineEffect__FLAGS(this) (&(this)->FLAGS)
#define PipelineEffect_INSTANCE(this) ((this)->INSTANCE)
#define PipelineEffect__INSTANCE(this) (&(this)->INSTANCE)
#define PipelineEffect_INSTRUCTION(this) ((this)->INSTRUCTION)
#define PipelineEffect__INSTRUCTION(this) (&(this)->INSTRUCTION)
#define PipelineEffect_ACTION(this) ((this)->ACTION)
#define PipelineEffect__ACTION(this) (&(this)->ACTION)
#define PipelineEffect_NEXT(this) ((this)->NEXT)
#define PipelineEffect__NEXT(this) (&(this)->NEXT)
#define PipelineEffect_PRED(this) ((this)->PRED)
#define PipelineEffect__PRED(this) (&(this)->PRED)
#define PipelineEffect_SUCC(this) ((this)->SUCC)
#define PipelineEffect__SUCC(this) (&(this)->SUCC)
#define PipelineEffect_isRead(this) (((PipelineEffect_FLAGS(this) & PipelineEffectFlag_Write) == 0))
#define PipelineEffect_isWrite(this) (((PipelineEffect_FLAGS(this) & PipelineEffectFlag_Write) != 0))


/*
 * PipelineEffect_latencyRAW --	Compute the RAW latency between two PipelineEffect(s).
 */
int
PipelineEffect_latencyRAW(const_PipelineEffect this, const_PipelineEffect that);

/*
 * PipelineEffect_latencyWAR --	Compute the WAR latency between two PipelineEffect(s).
 */
int
PipelineEffect_latencyWAR(const_PipelineEffect this, const_PipelineEffect that);

/*
 * PipelineEffect_latencyWAW --	Compute the WAW latency between two PipelineEffect(s).
 */
int
PipelineEffect_latencyWAW(const_PipelineEffect this, const_PipelineEffect that);

/*
 * PipelineEffectList --	List of PipelineEffect(s).
 */
struct PipelineEffectList_ {
  //@args	
  struct PipelineEffect_ *FIRST;
  struct PipelineEffect_ *LAST;
};
typedef struct PipelineEffectList_ PipelineEffectList_, *PipelineEffectList;
typedef const struct PipelineEffectList_ *const_PipelineEffectList;
typedef struct PipelineEffectList_ * restrict_PipelineEffectList;

extern PipelineEffectList
PipelineEffectList_Ctor(PipelineEffectList this);

extern PipelineEffectList
PipelineEffectList_Copy(PipelineEffectList this, const_PipelineEffectList that);

extern void
PipelineEffectList_Dtor(PipelineEffectList this);

extern size_t
PipelineEffectList_Size(void);

#define PipelineEffectList_FIRST(this) ((this)->FIRST)
#define PipelineEffectList__FIRST(this) (&(this)->FIRST)
#define PipelineEffectList_LAST(this) ((this)->LAST)
#define PipelineEffectList__LAST(this) (&(this)->LAST)


/*
 * PipelineEffectList_isEmpty --	True iff this PipelineEffectList is empty.
 */
bool
PipelineEffectList_isEmpty(const_PipelineEffectList this);

/*
 * PipelineEffectList_first --	For use by PipelineEffectList_FOREACH_PipelineEffect.
 */
PipelineEffect
PipelineEffectList_first(const_PipelineEffectList this);

/*
 * PipelineEffectList_FOREACH_PipelineEffect --	Iterator
 */
#define PipelineEffectList_FOREACH_PipelineEffect(this, effect) { \
  PipelineEffect PipelineEffectList_NEXT = NULL; \
  PipelineEffect effect = PipelineEffectList_first(this); \
  for (; effect != NULL; effect = PipelineEffectList_NEXT) { \
    PipelineEffectList_NEXT = PipelineEffect_NEXT(effect);
#define PipelineEffectList_ENDEACH_PipelineEffect \
  } \
}

/*
 * PipelineEffectList_push_ --	Push an already allocated PipelineEffect.
 */
PipelineEffect
PipelineEffectList_push_(PipelineEffectList this, PipelineEffect effect);

//
struct PipelineADL_ {
  //@args	Memory memory, InstructionStack instructions
  Memory MEMORY;				// Where to allocate ADLs
  InstructionStack INSTRUCTIONS;		// Instruction(s) to schedule.
  PipelineEffectList_ *LISTS;			// Array of Dependency Lists (ADL).
};

extern PipelineADL
PipelineADL_Ctor(PipelineADL this, Memory memory, InstructionStack instructions);

extern PipelineADL
PipelineADL_Copy(PipelineADL this, const_PipelineADL that);

extern void
PipelineADL_Dtor(PipelineADL this);

extern size_t
PipelineADL_Size(Memory memory, InstructionStack instructions);

#define PipelineADL_MEMORY(this) ((this)->MEMORY)
#define PipelineADL__MEMORY(this) (&(this)->MEMORY)
#define PipelineADL_INSTRUCTIONS(this) ((this)->INSTRUCTIONS)
#define PipelineADL__INSTRUCTIONS(this) (&(this)->INSTRUCTIONS)
#define PipelineADL_LISTS(this) ((this)->LISTS)
#define PipelineADL__LISTS(this) (&(this)->LISTS)


/*
 * PipelineADL_lists --	For use by PipelineADL_FOREACH_PipelineEffectList.
 */
PipelineEffectList_ *
PipelineADL_lists(PipelineADL this);

/*
 * PipelineADL_FOREACH_PipelineEffectList --	Iterator
 */
#define PipelineADL_FOREACH_PipelineEffectList(this, list) { \
  PipelineEffectList PipelineADL_LISTS = PipelineADL_lists(this); \
  int32_t PipelineADL_CELL = 0; \
  for (; PipelineADL_CELL < StorageCell__COUNT; PipelineADL_CELL++) { \
    PipelineEffectList list = PipelineADL_LISTS + PipelineADL_CELL;
#define PipelineADL_ENDEACH_PipelineEffectList \
  } \
}

/*
 * PipelineADL_linkEffects --	Link the PipelineEffect(s) using PRED and SUCC.
 *
 * The linking of PipelineEffect(s) creates the equivalent of a DAG for each StorageCell.
 * Ensure:	The PRED and SUCC point to the predecessor Write and successor Write.
 */
void
PipelineADL_linkEffects(PipelineADL this);

//
struct Pipeline_ {
  //@args	Memory memory, Processor processor, InstructionStack instructions
  PipelineADL_ ADL[1];		// This Pipeline Array of Dependency Lists.
  //@access INSTRUCTIONS	PipelineADL_INSTRUCTIONS(Pipeline_ADL(this))
  //@access MEMORY	PipelineADL_MEMORY(Pipeline_ADL(this))
  PQueue_ WAITING[1];		// Ordered by lower release date first.
  PQueue_ RELEASED[1];		// Ordered by higher priority first.
  ResourceTable_ *TABLE;	// ResourceTable.
};

extern Pipeline
Pipeline_Ctor(Pipeline this, Memory memory, Processor processor, InstructionStack instructions);

extern Pipeline
Pipeline_Copy(Pipeline this, const_Pipeline that);

extern void
Pipeline_Dtor(Pipeline this);

extern size_t
Pipeline_Size(Memory memory, Processor processor, InstructionStack instructions);

#define Pipeline_ADL(this) ((this)->ADL)
#define Pipeline__ADL(this) (&(this)->ADL)
#define Pipeline_WAITING(this) ((this)->WAITING)
#define Pipeline__WAITING(this) (&(this)->WAITING)
#define Pipeline_RELEASED(this) ((this)->RELEASED)
#define Pipeline__RELEASED(this) (&(this)->RELEASED)
#define Pipeline_TABLE(this) ((this)->TABLE)
#define Pipeline__TABLE(this) (&(this)->TABLE)
#define Pipeline_INSTRUCTIONS(this) (PipelineADL_INSTRUCTIONS(Pipeline_ADL(this)))
#define Pipeline_MEMORY(this) (PipelineADL_MEMORY(Pipeline_ADL(this)))


/*
 * Pipeline_INIT --	Initialize the Pipeline module.
 */
#define Pipeline_INIT()

/*
 * Pipeline_FINI --	Finalize the Pipeline module.
 */
#define Pipeline_FINI()


#endif//Pipeline__h_INCLUDED
