#ifndef Scoreboard_c_INCLUDED
#define Scoreboard_c_INCLUDED

#line 15 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
#include "AIR/AIR_.h"

#line 20 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 46 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 70 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
Scoreboard
Scoreboard_Ctor(Scoreboard this,
                Memory memory, Processor processor, int windowSize)
{
  int cellCount = StorageCell__COUNT, cell;
  size_t tableSize = ResourceTable_Size(processor, windowSize);
  ResourceTable_ *table = Memory_alloc(memory, tableSize);
  ScoreboardAction_ *writeActions = Memory_alloc(memory, sizeof(ScoreboardAction_)*2*cellCount);
  ScoreboardAction_ *accessActions = writeActions + cellCount;
  ScoreboardAction_ actions_min_  = { { Instance__UNDEF, INT16_MIN } };
  uint32_t packed_min = ScoreboardAction_PACKED(&actions_min_);
  *Scoreboard__MEMORY(this) = memory;
  *Scoreboard__TABLE(this) = ResourceTable_Ctor(table, processor, windowSize);
  for (cell = 0; cell < StorageCell__COUNT; cell++) {
    *ScoreboardAction__PACKED(writeActions + cell) = packed_min;
  }
  for (cell = 0; cell < StorageCell__COUNT; cell++) {
    *ScoreboardAction__PACKED(accessActions + cell) = packed_min;
  }
  *Scoreboard__WRITEACTIONS(this) = writeActions;
  *Scoreboard__ACCESSACTIONS(this) = accessActions;
  return this;
}

#line 94 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
void
Scoreboard_Dtor(Scoreboard this)
{
  Memory memory = Scoreboard_MEMORY(this);
  Memory_free_(memory, Scoreboard_TABLE(this));
  Memory_free_(memory, Scoreboard_WRITEACTIONS(this));
}

#line 107 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 127 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 148 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
int
Scoreboard_tryInstruction(const_Scoreboard this, const_Instruction instruction)
{
  int windowStart = Scoreboard_WINDOWSTART(this);
  const ScoreboardAction_ *writeActions = Scoreboard_WRITEACTIONS(this);
  const ScoreboardAction_ *accessActions = Scoreboard_ACCESSACTIONS(this);
  const InstructionAction_ *actions = Instruction_ACTIONS(instruction);
  int writeCount = Instruction_WRITECOUNT(instruction);
  int readCount = Instruction_READCOUNT(instruction), i;
  Instance instance = Instruction_INSTANCE(instruction);
  int startDate = Instruction_RELEASE(instruction);
  int savedDate = startDate;
  unsigned fixupRAW = 0;
  if (startDate < windowStart) {
    startDate = savedDate = windowStart;
  }
  // The RAW Register latencies.
  for (i = writeCount; i < writeCount + readCount; i++) {
    InstructionAction_ action_ = actions[i];
    StorageCell cell = action_.__.CELL;
    Instance write_instance = ScoreboardAction_INSTANCE(writeActions + cell);
    int lastWrite = ScoreboardAction_LASTDATE(writeActions + cell);
    int readStage = action_.__.STAGE, latencyRAW = action_.__.RAW;
    int readDate = lastWrite - readStage + latencyRAW;
    if (startDate < readDate) startDate = readDate;
    fixupRAW |= Instance_fixupRAW(write_instance, instance);
  }
  if (fixupRAW) {
    // Special cases for the RAW latencies.
    startDate = savedDate;
    for (i = writeCount; i < writeCount + readCount; i++) {
      InstructionAction_ action_ = actions[i];
      StorageCell cell = action_.__.CELL;
      Instance write_instance = ScoreboardAction_INSTANCE(writeActions + cell);
      int lastWrite = ScoreboardAction_LASTDATE(writeActions + cell);
      int readStage = action_.__.STAGE, latencyRAW = action_.__.RAW;
      int readDate = lastWrite - readStage + latencyRAW
                   + Instance_latencyRAW(write_instance, instance, cell);
      if (startDate < readDate) startDate = readDate;
    }
  }
  // The WAR and WAW register latencies.
  for (i = 0; i < writeCount; i++) {
    InstructionAction_ action_ = actions[i];
    StorageCell cell = action_.__.CELL;
    int lastWrite = ScoreboardAction_LASTDATE(writeActions + cell);
    int lastAccess = ScoreboardAction_LASTDATE(accessActions + cell);
    int writeStage = action_.__.STAGE, latencyWAW = action_.__.WAW;
    int writeDate = lastWrite - writeStage + latencyWAW;
    if (startDate < writeDate) startDate = writeDate;
    if (startDate < lastAccess) startDate = lastAccess;
  }
  // Try the Resource(s).
  startDate = ResourceTable_tryReservation(Scoreboard_TABLE(this),
                                           Instruction_RESERVATION(instruction),
                                           startDate);
  // Return the earliest start date found.
  return startDate;
}

#line 218 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
int
Scoreboard_addInstruction(Scoreboard this, Instruction instruction, int startDate)
{
  int increase = 0;
  int windowSize = Scoreboard_WINDOWSIZE(this);
  int windowStart = Scoreboard_WINDOWSTART(this);
  ScoreboardAction_ *restrict writeActions = Scoreboard_WRITEACTIONS(this);
  ScoreboardAction_ *restrict accessActions = Scoreboard_ACCESSACTIONS(this);
  const InstructionAction_ *actions = Instruction_ACTIONS(instruction);
  int writeCount = Instruction_WRITECOUNT(instruction);
  int readCount = Instruction_READCOUNT(instruction), i;
  Instance instance = Instruction_INSTANCE(instruction);
  Except_CHECK(startDate >= windowStart);
  // Read actions: update the ACCESSACTIONS of register and its aliases.
  for (i = writeCount; i < writeCount + readCount; i++) {
    InstructionAction_ action_ = actions[i];
    StorageCell cell = action_.__.CELL;
    int newAccessDate = startDate + action_.__.WAR;
    if (newAccessDate > ScoreboardAction_LASTDATE(accessActions + cell)) {
      *ScoreboardAction__INSTANCE(accessActions + cell) = instance;
      *ScoreboardAction__LASTDATE(accessActions + cell) = newAccessDate;
    }
  }
  // Write actions: the WRITEACTIONS and ACCESSACTIONS of register and its aliases.
  for (i = 0; i < writeCount; i++) {
    InstructionAction_ action_ = actions[i];
    StorageCell cell = action_.__.CELL;
    int newWriteDate = startDate + action_.__.STAGE;
    int newAccessDate = startDate + action_.__.WAW;
    Except_CHECK(newWriteDate >= ScoreboardAction_LASTDATE(writeActions + cell));
    Except_CHECK(newAccessDate >= ScoreboardAction_LASTDATE(accessActions + cell));
    *ScoreboardAction__INSTANCE(writeActions + cell) = instance;
    *ScoreboardAction__LASTDATE(writeActions + cell) = newWriteDate;
    *ScoreboardAction__INSTANCE(accessActions + cell) = instance;
    *ScoreboardAction__LASTDATE(accessActions + cell) = newAccessDate;
  }
  // Update the Resource(s).
  ResourceTable_addReservation(Scoreboard_TABLE(this),
                               Instruction_RESERVATION(instruction),
                               startDate);
  // Update instruction STARTDATE and compute increase.
  increase = startDate - Instruction_RELEASE(instruction);
  *Instruction__STARTDATE(instruction) = startDate;
  *Instruction__RELEASE(instruction) = startDate;
  return increase;
}

#line 265 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 273 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
void
Scoreboard_schedule(Scoreboard this, InstructionStack instructions)
{
}

#line 301 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
Scoreboard
Scoreboard_reset(Scoreboard this, int windowStart)
{
  int cell = 0;
  ScoreboardAction_ actions_min_  = { { Instance__UNDEF, INT16_MIN } };
  uint32_t packed_min = ScoreboardAction_PACKED(&actions_min_);
  ScoreboardAction_ *writeActions = Scoreboard_WRITEACTIONS(this);
  ScoreboardAction_ *accessActions = Scoreboard_ACCESSACTIONS(this);
  ResourceTable_reset(Scoreboard_TABLE(this), windowStart);
  for (cell = 0; cell < StorageCell__COUNT; cell++) {
    *ScoreboardAction__PACKED(writeActions + cell) = packed_min;
    *ScoreboardAction__PACKED(accessActions + cell) = packed_min;
  }
  return this;
}

#line 327 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
bool
Scoreboard_merge(Scoreboard this, const_Scoreboard that, int delay)
{
  bool changed = false;
  int this_windowStart = Scoreboard_WINDOWSTART(this);
  int that_windowStart = Scoreboard_WINDOWSTART(that) + delay;
  ScoreboardAction_ *this_writeActions = Scoreboard_WRITEACTIONS(this);
  ScoreboardAction_ *this_accessActions = Scoreboard_ACCESSACTIONS(this);
  const ScoreboardAction_ *that_writeActions = Scoreboard_WRITEACTIONS(that);
  const ScoreboardAction_ *that_accessActions = Scoreboard_ACCESSACTIONS(that);
  int delta = that_windowStart - this_windowStart, cell;
  for (cell = 0; cell < StorageCell__COUNT; cell++) {
    ScoreboardAction this_writeAction = this_writeActions + cell;
    ScoreboardAction this_accessAction = this_accessActions + cell;
    int this_lastWrite = ScoreboardAction_LASTDATE(this_writeAction);
    int this_lastAccess = ScoreboardAction_LASTDATE(this_accessAction);
    const_ScoreboardAction that_writeAction = that_writeActions + cell;
    const_ScoreboardAction that_accessAction = that_accessActions + cell;
    int that_lastWrite = ScoreboardAction_LASTDATE(that_writeAction);
    int that_lastAccess = ScoreboardAction_LASTDATE(that_accessAction);
    if (this_lastAccess < that_lastAccess - delta) {
      this_lastAccess = that_lastAccess - delta;
      changed = true;
    }
    if (this_lastWrite < that_lastWrite - delta) {
      this_lastWrite = that_lastWrite - delta;
      changed = true;
    }
    //TODO! *ScoreboardAction__INSTANCE(this_writeAction) = write_instance;
    *ScoreboardAction__LASTDATE(this_writeAction) = this_lastWrite;
    //TODO! *ScoreboardAction__INSTANCE(this_accessAction) = read_instance;
    *ScoreboardAction__LASTDATE(this_accessAction) = this_lastAccess;
  }
  changed |= ResourceTable_merge(Scoreboard_TABLE(this), Scoreboard_TABLE(that), delay);
  return changed;
}

#line 364 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 370 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
#ifndef ECL_NTRACE
bool
Scoreboard_pretty(const_Scoreboard this, FILE *file)
{
  int clp2Horizon = Scoreboard_CLP2HORIZON(this), i;
  const ScoreboardAction_ *writeActions = Scoreboard_WRITEACTIONS(this);
  const ScoreboardAction_ *accessActions = Scoreboard_ACCESSACTIONS(this);
  int windowStart = Scoreboard_WINDOWSTART(this), cell;
  if (file == NULL) file = stderr;
  ResourceTable_pretty(Scoreboard_TABLE(this), file);
  fputs("--------|", file);
  for (i = 0; i < clp2Horizon; i++) fputs("---", file);
  fputc('\n', file);
  for (cell = 0; cell < StorageCell__COUNT; cell++) {
    int lastAccess = ScoreboardAction_LASTDATE(accessActions + cell);
    int lastWrite = ScoreboardAction_LASTDATE(writeActions + cell);
    if (lastAccess >= windowStart || lastWrite >= windowStart) {
      if (cell == StorageCell__Volatile) {
        fprintf(file, "%s|", StorageCell_name(cell));
      } else {
        fprintf(file, "%s\t|", StorageCell_name(cell));
      }
      for (i = 0; i < clp2Horizon; i++) {
        int date = windowStart + i;
        char a = date <= lastAccess? 'a': ' ';
        char w = date <= lastWrite? 'w': ' ';
        fprintf(file, " %c%c", a, w);
      }
      fputs("  (", file);
      if (lastAccess >= windowStart + clp2Horizon) fprintf(file, "%d", lastAccess);
      fputc(',', file);
      if (lastWrite >= windowStart + clp2Horizon) fprintf(file, "%d", lastWrite);
      fputs(")\n", file);
    }
  }
  return true;
}
#endif//ECL_NTRACE

#line 409 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 985 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 997 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
#if XCC__C
#include "MDS/MDS_.h"
static void
Scoreboard_TEST(void)
{
#include "AIR/Scoreboard_.c"
}

#line 1005 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/AIR/Scoreboard.xcc"
int
main(int argc, char **argv)
{
  AIR_Test(Scoreboard_TEST);
  return 0;
}
#endif


#endif//Scoreboard_c_INCLUDED
