#ifndef Scoreboard__h_INCLUDED
#define Scoreboard__h_INCLUDED

//
#include "AIR/Scoreboard.h"

/*
 * Scoreboard_tryInstruction --	Try to schedule an Instruction with this Scoreboard.
 *
 * Return:	The earliest start date of =instruction=.
 *
 * The principle is to check the register dependences, then find the earliest start date
 * that satisfies the dependences and which is also free of resource conflicts. Searching
 * starts at the Instruction_RELEASE.
 *
 * For the RAW dependences, we rely on the write LASTDATE(s) combined with the action STAGE
 * correction form the Instruction actionArray.
 *
 * For the WAR and WAW dependences, we use the access LASTDATE(s) defined as the issue
 * date plus a corrective term defined as follows. For a read issued at date t, access
 * date is t + WAR latency. For a write at date t, access date is t + WAW latency.
 */
int
Scoreboard_tryInstruction(const_Scoreboard this, const_Instruction instruction);

/*
 * Scoreboard_addInstruction --	Add an Instruction to this Scoreboard.
 *
 * Update the TABLE, WRITEACTIONS and ACCESSACTIONS.
 *
 * Return:	the increase of the =instruction= start date.
 */
int
Scoreboard_addInstruction(Scoreboard this, Instruction instruction, int startDate);

/*
 * Scoreboard_INIT --	Initialize the Scoreboard module.
 */
#define Scoreboard_INIT()

/*
 * Scoreboard_FINI --	Finalize the Scoreboard module.
 */
#define Scoreboard_FINI()


#endif//Scoreboard__h_INCLUDED
