#ifndef Allocator_h_INCLUDED
#define Allocator_h_INCLUDED

/*
 * !!!!	Allocator.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * Allocator --	Scalar optimizer for a CodeRegion.
 */
struct Allocator_ {
  //@args	Memory memory, CodeRegion codeRegion
  Memory MEMORY;
  CodeRegion CODEREGION;
};
typedef struct Allocator_ Allocator_, *Allocator;
typedef const struct Allocator_ *const_Allocator;
typedef struct Allocator_ * restrict_Allocator;

extern Allocator
Allocator_Ctor(Allocator this, Memory memory, CodeRegion codeRegion);

extern Allocator
Allocator_Copy(Allocator this, const_Allocator that);

extern void
Allocator_Dtor(Allocator this);

extern size_t
Allocator_Size(Memory memory, CodeRegion codeRegion);

#define Allocator_MEMORY(this) ((this)->MEMORY)
#define Allocator__MEMORY(this) (&(this)->MEMORY)
#define Allocator_CODEREGION(this) ((this)->CODEREGION)
#define Allocator__CODEREGION(this) (&(this)->CODEREGION)


//
Allocator
Allocator_new(Memory parent, CodeRegion codeRegion);

//
Allocator
Allocator_delete(Allocator this);

/*
 * Allocator_memory --	This Allocator Memory.
 */
Memory
Allocator_memory(Allocator this);

/*
 * Allocator_codeRegion --	This Allocator CodeRegion.
 */
CodeRegion
Allocator_codeRegion(Allocator this);

/*
 * Allocator_pretty --	Pretty-print this Allocator.
 */
bool
Allocator_pretty(Allocator this, FILE *file);


#endif//Allocator_h_INCLUDED
