#ifndef ParallelCopy_c_INCLUDED
#define ParallelCopy_c_INCLUDED

#line 17 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
#include "CGO/CGO_.h"

#line 22 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
#ifndef PARALLEL_COPY_POOLED
#   define PARALLEL_COPY_POOLED 128
#endif//PARALLEL_COPY_POOLED

#line 36 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
bool
ParallelCopy_resolveDuplication(CodeRegion codeRegion, RegFile regFile) {
  Memory memory = Memory_new(Procedure_memory(CodeRegion_procedure(codeRegion)),
                             PARALLEL_COPY_POOLED);
  bool result = true;
  // need liveness
  // must update liveness for phi operands
  // in which order should I treat the basic blocks?
  Liveness liveness = Liveness_make(memory, codeRegion,
                                    CodeRegion_globalTable(codeRegion), true);
  // do work
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    result &= ParallelCopy_resolveDuplicationAtBBLevel(basicBlock, regFile);
  } CodeRegion_ENDEACH_BasicBlock;

#line 51 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  Liveness_kill(liveness);
  Memory_delete(memory);
  return result;
}

#line 60 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
ParallelCopy_initRegisterSet(BasicBlock basicBlock,  RegisterSet_ availableReg[][1]);

#line 63 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
bool
ParallelCopy_resolveDuplicationAtBBLevel(BasicBlock basicBlock,
                                         RegFile regFile) {
  RegisterSet_ availableReg[RegFile__COUNT][1];
  bool result = true;
  ParallelCopy_initRegisterSet(basicBlock, availableReg);
  BasicBlock_FORBACK_Operation(basicBlock, operation) {
    if(Operator_isPhi(Operation_operator(operation))
       && (regFile == RegFile__UNDEF ||
           Temporary_regFile(Operation_getResult(operation, 0)) == regFile)) {
      // resolve Phi parallel copy
      result &= ParallelCopy_resolveDuplicationOnPhi(availableReg, operation,
                                                     regFile);
      break;
    }
    Operation_RESULTS_FOREACH_Temporary(operation, resTemp) {
      if(Temporary_isDedicated(resTemp) || Temporary_isAssigned(resTemp)) {
        RegisterSet_insert(availableReg[Temporary_regFile(resTemp)],
                           Temporary_register(resTemp));
      }
    } Operation_RESULTS_ENDEACH_Temporary;
    Operation_ARGUMENTS_FOREACH_Temporary(operation, argTemp) {
      if(Temporary_isDedicated(argTemp) || Temporary_isAssigned(argTemp)) {
        RegisterSet_remove(availableReg[Temporary_regFile(argTemp)],
                           Temporary_register(argTemp));
      }
    } Operation_ARGUMENTS_ENDEACH_Temporary;
    if(Operator_isParCopy(Operation_operator(operation))
       && (regFile == RegFile__UNDEF ||
           Temporary_regFile(Operation_getResult(operation, 0)) == regFile)) {
      // resolve ParallelCopy
      RegFile regFile = Temporary_regFile(Operation_getResult(operation, 0));
      result &= ParallelCopy_resolveDuplicationOnParCopy(availableReg[regFile],
                                                         operation);
    }
  } BasicBlock_ENDBACK_Operation;
  return result;
}

#line 103 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
ParallelCopy_initRegisterSet(BasicBlock basicBlock, RegisterSet_ availableReg[][1]) {
  Convention convention = CodeRegion_convention(BasicBlock_codeRegion(basicBlock));
  RegFile regFile;
  for(regFile = 0; regFile < RegFile__COUNT; ++regFile) {
    availableReg[regFile][0] = Convention_allocatableSet(convention, regFile);
  }
  TemporarySet_FOREACH_Temporary(BasicBlock_liveOutSet(basicBlock), live) {
    if(Temporary_isAssigned(live) || Temporary_isDedicated(live)) {
      RegisterSet_remove(availableReg[Temporary_regFile(live)],
                         Temporary_register(live));
    }
  } TemporarySet_ENDEACH_Temporary;
  TemporarySet_FOREACH_Temporary(BasicBlock_phiOutSet(basicBlock), live) {
    if(Temporary_isAssigned(live) || Temporary_isDedicated(live)) {
      RegisterSet_remove(availableReg[Temporary_regFile(live)],
                         Temporary_register(live));
    }
  } TemporarySet_ENDEACH_Temporary;
}

#line 136 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
PermutationNode
PermutationNode_Ctor(PermutationNode this, Register color, Memory memory) {
  *PermutationNode__COLOR(this) = color;
  IDList_Ctor(PermutationNode_COPIES(this), memory);
  *PermutationNode__PARENT(this) = NULL;
  *PermutationNode__CYCLEIDX(this) = -1;
  return this;
}

#line 145 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
void
PermutationNode_Dtor(PermutationNode this) {
  IDList_Dtor(PermutationNode_COPIES(this));
}

#line 150 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
struct PermutationNodeAndIdx_ {
  //@args PermutationNode node, int32_t idx
  PermutationNode NODE;
  int32_t IDX;
};
typedef struct PermutationNodeAndIdx_ PermutationNodeAndIdx_, *PermutationNodeAndIdx;
typedef const struct PermutationNodeAndIdx_ *const_PermutationNodeAndIdx;
typedef struct PermutationNodeAndIdx_ * restrict_PermutationNodeAndIdx;

extern PermutationNodeAndIdx
PermutationNodeAndIdx_Ctor(PermutationNodeAndIdx this, PermutationNode node, int32_t idx);

extern PermutationNodeAndIdx
PermutationNodeAndIdx_Copy(PermutationNodeAndIdx this, const_PermutationNodeAndIdx that);

extern void
PermutationNodeAndIdx_Dtor(PermutationNodeAndIdx this);

extern size_t
PermutationNodeAndIdx_Size(PermutationNode node, int32_t idx);

#define PermutationNodeAndIdx_NODE(this) ((this)->NODE)
#define PermutationNodeAndIdx__NODE(this) (&(this)->NODE)
#define PermutationNodeAndIdx_IDX(this) ((this)->IDX)
#define PermutationNodeAndIdx__IDX(this) (&(this)->IDX)


#line 156 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
PermutationNodeAndIdx
PermutationNodeAndIdx_Ctor(PermutationNodeAndIdx this, PermutationNode node,
                           int32_t idx) {
  *PermutationNodeAndIdx__NODE(this) = node;
  *PermutationNodeAndIdx__IDX(this) = idx;
  return this;
}

#line 164 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setPcopiesResultsForPhis(Selector selector, Operation phi, RegFile regFile,
                         RegisterSet_ availableReg[][1], Operation pcopies[]);

#line 168 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
materializeParCopyModificationOnPhis(Operation pcopy, Operation phi,
                                     int32_t idx);

#line 177 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
bool
ParallelCopy_resolveDuplicationOnPhi(RegisterSet_ availableReg[][1],
                                     Operation phi, RegFile a_regFile) {
  bool result = true;
  int32_t i;
  Operation pcopies[RegFile__COUNT];
  BasicBlock basicBlock = Operation_basicBlock(phi);
  bool isFirst = true;
  RegisterSet_ curAvailableReg[RegFile__COUNT][1];
  Selector_ selector[1];
  Selector curSelector = BasicBlock_selector(basicBlock);
  Memory memory = Selector_memory(curSelector);
  Procedure procedure = Selector_procedure(curSelector);
  Processor processor = Selector_processor(curSelector);
  RegFile regFile;
  bool hasDuplication[RegFile__COUNT];

#line 194 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  Selector_Ctor(selector, memory, procedure, processor, false);

#line 196 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  memset(hasDuplication, false, sizeof(bool) * RegFile__COUNT);

#line 198 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  for(regFile = 0; regFile < RegFile__COUNT; ++regFile) {
    RegisterSet_Copy(curAvailableReg[regFile], availableReg[regFile]);
  }

#line 202 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  Except_CHECK(Operator_isPhi(Operation_operator(phi)));

#line 204 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  setPcopiesResultsForPhis(selector, phi, a_regFile, curAvailableReg, pcopies);

#line 207 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  for(i = 0; i < Operation_argCount(phi) && result; ++i) {
    // first of all, collect all phi operations
    // List of result
    // List of arguments
    // Build virtual parallel copy
    Operation cur = phi;
    int32_t j = 0;
    BasicBlockEdge edge = BasicBlock_getPhiEnterEdge(basicBlock, i);
    RegisterSet_ usedColors[1];
    int32_t nbArg[RegFile__COUNT];

#line 218 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    //  count the number of argument per class of parallel copy
    memset(nbArg, 0, sizeof(int32_t) * RegFile__COUNT);

#line 221 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    RegisterSet_Ctor(usedColors);

#line 223 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    do {
      Temporary curArg = Operation_getArgument(cur, i);
      RegFile curRegfile = Temporary_regFile(curArg);
      if(a_regFile != RegFile__UNDEF && curRegfile != a_regFile) {
        continue;
      }
      Except_CHECK(pcopies[curRegfile] != NULL);

#line 231 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
      Operation_setArgument(pcopies[curRegfile], nbArg[curRegfile]++, curArg);
      if(!RegisterSet_insert(usedColors, Temporary_register(curArg))) {
        hasDuplication[curRegfile] = true;
      }
      // check whether there are duplications on this edge
    } while((cur = Operation_prev(cur)) != NULL
            && Operator_isPhi(Operation_operator(cur)));

#line 239 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    for(regFile = 0; regFile < RegFile__COUNT && result; ++regFile) {
      if(pcopies[regFile] != NULL && hasDuplication[regFile]) {
        result &=
          ParallelCopy_resolveDuplicationOnParCopyOnEdge(curAvailableReg[regFile],
                                                         pcopies[regFile],
                                                         edge);
        materializeParCopyModificationOnPhis(pcopies[regFile], phi, i);
      }
    }
    RegisterSet_Dtor(usedColors);
  }
  Selector_Dtor(selector);
  return result;
}

#line 254 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
materializeParCopyModificationOnPhis(Operation pcopy, Operation phi,
                                     int32_t idx) {
  Operation cur = phi;
  int32_t argIdx = 0;
  RegFile curRegfile = Temporary_regFile(Operation_getResult(pcopy, 0));
  do {
    Temporary phiRes = Operation_getResult(cur, 0);

#line 263 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    if(Temporary_regFile(phiRes) == curRegfile) {
      Temporary phiCheckArg = Operation_getArgument(cur, idx);
      Register curReg = Temporary_register(phiRes);
      Except_CHECK(phiRes == Operation_getResult(pcopy, argIdx)
                   && "Phis and virtual Par copy do not match");
      // The purpose of this test is to avoid unnecessary call to setArgument
      if(phiCheckArg != Operation_getArgument(pcopy, argIdx)) {
        Operation_setArgument(cur, idx, Operation_getArgument(pcopy, argIdx));
      }
      ++argIdx;
    }
  } while((cur = Operation_prev(cur)) != NULL
          && Operator_isPhi(Operation_operator(cur)));
}

#line 278 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setPcopiesResultsForPhis(Selector selector, Operation phi, RegFile a_regFile,
                         RegisterSet_ availableReg[][1], Operation pcopies[]) {
  BasicBlock basicBlock = Operation_basicBlock(phi);
  int32_t phiCount = BasicBlock_phiCount(basicBlock);
  Operation cur = phi;
  int32_t nbArg[RegFile__COUNT];

#line 286 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  memset(pcopies, 0, sizeof(Operation) * RegFile__COUNT);

#line 288 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  // first of all, collect all phi operations
  // List of result
  // List of arguments
  // Build virtual parallel copy

#line 293 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  // count the number of argument per class of parallel copy
  memset(nbArg, 0, sizeof(int32_t) * RegFile__COUNT);

#line 296 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  do {
    Temporary phiRes = Operation_getResult(cur, 0);
    RegFile curRegfile = Temporary_regFile(phiRes);
    Register curReg = Temporary_register(phiRes);
    if(a_regFile != RegFile__UNDEF && a_regFile != curRegfile) {
      continue;
    }
    Except_CHECK(Temporary_isDedicated(phiRes) || Temporary_isAssigned(phiRes));
    // we set pcopy results only once since at each iteration it is the same
    // as previous
    if(pcopies[curRegfile] == NULL) {
      pcopies[curRegfile] =
        Selector_makeOperationParCopy(selector, Operator_PCOPY, phiCount, false);
    }
    *Operation__RESCOUNT(pcopies[curRegfile]) =
      *Operation__ARGCOUNT(pcopies[curRegfile]) = nbArg[curRegfile] + 1;
    Operation_setResult(pcopies[curRegfile], nbArg[curRegfile]++, phiRes);
    RegisterSet_remove(availableReg[curRegfile], curReg);
  } while((cur = Operation_prev(cur)) != NULL
          && Operator_isPhi(Operation_operator(cur)));
}

#line 327 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
bool
ParallelCopy_resolveDuplicationOnParCopyOnEdge(const_RegisterSet availableReg,
                                               Operation pcopy,
                                               BasicBlockEdge edge) {
  Operation point = NULL;
  BasicBlock destinationBlock = BasicBlockEdge_splitAndAction(edge, true, &point,
                                                              NULL, NULL);
  return ParallelCopy_resolveDuplicationOnParCopy_(availableReg, pcopy,
                                                   destinationBlock, point,
                                                   false);
}

#line 345 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static bool
parCopyHasDuplication(Operation pcopy);

#line 348 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
bool
ParallelCopy_resolveDuplicationOnParCopy(const_RegisterSet availableReg,
                                         Operation parCopy) {
  bool result = true;
  if(parCopyHasDuplication(parCopy)) {
    result =
      ParallelCopy_resolveDuplicationOnParCopy_(availableReg, parCopy,
                                                Operation_basicBlock(parCopy),
                                                parCopy, true);
  }
  return result;
}

#line 361 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
buildPermutationScheme(Operation parCopy, PermutationNode rootNodes[],
                       IDList permutationNodeStorage, Memory memory,
                       RegisterSet leaves, RegisterSet isDuplicated);

#line 366 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static bool
resolvePermutationScheme(Operation parCopy, Memory memory, Selector selector,
                         PermutationNode rootNodes[],
                         const_RegisterSet availableReg, RegisterSet leaves,
                         RegisterSet isDuplicated);

#line 379 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
bool
ParallelCopy_resolveDuplicationOnParCopy_(const_RegisterSet availableReg,
                                          Operation parCopy,
                                          BasicBlock basicBlock,
                                          Operation point, bool before) {
  // We can just allocate the number of registers of the examined register file
  PermutationNode rootNodes[Register__COUNT];
  IDList_ permutationNodeStorage[1];
  RegisterSet_ leaves[1];
  RegisterSet_ isDuplicated[1];
  bool result;
  Memory memory = Memory_new(BasicBlock_MEMORY(basicBlock),
                             PARALLEL_COPY_POOLED);
  Selector_ selector[1];
  Selector curSelector = BasicBlock_selector(basicBlock);
  Procedure procedure = Selector_procedure(curSelector);
  Processor processor = Selector_processor(curSelector);

#line 397 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  Selector_Ctor(selector, Selector_memory(curSelector), procedure, processor,
                false);

#line 400 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  Except_CHECK(Operator_isParCopy(Operation_operator(parCopy)));
  IDList_Ctor(permutationNodeStorage, memory);
  memset(rootNodes, 0, sizeof(PermutationNode) * Register__COUNT);

#line 404 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  RegisterSet_Ctor(leaves);
  RegisterSet_Ctor(isDuplicated);

#line 407 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  // faire un tableau avec l'ensemble des noeuds
  // les relier en fonction de la topographie de la copie
  // en profiter pour mettre ceux qui ont plus de 2 fils dans un ensemble
  // de schma  rsoudre
  // mettre les autres dans les feuilles  (ne sont des feuilles que ceux qui
  // sont disponible avant la copy parallele (nouvelle couleurs))
  // resoudre les schmas de duplications.
  buildPermutationScheme(parCopy, rootNodes, permutationNodeStorage, memory,
                         leaves, isDuplicated);

#line 417 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  result = 
    resolvePermutationScheme(parCopy, memory, selector, rootNodes,
                             availableReg, leaves, isDuplicated);
  if(point != NULL) {
    if(before) {
      BasicBlock_insertSelector(basicBlock, point, selector);
    } else {
      BasicBlock_appendSelector(basicBlock, point, selector);
    }
  } else {
    if(before) {
      BasicBlock_putSelector(basicBlock, selector);
    } else {
      BasicBlock_pushSelector(basicBlock, selector);
    }
  }
#ifndef ECL_NDTOR
  IDList_FOREACH(permutationNodeStorage, PermutationNode_, permNode) {
    PermutationNode_Dtor(permNode);
  } IDList_ENDEACH;
  IDList_Dtor(permutationNodeStorage);
  RegisterSet_Dtor(leaves);
  RegisterSet_Dtor(isDuplicated);
#endif
  Memory_delete(memory);
  return result;
}

#line 445 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static bool
parCopyHasDuplication(Operation pcopy) {
  RegisterSet_ usedColors[1];
  bool result = false;
  Except_CHECK(Operator_isParCopy(Operation_operator(pcopy)));
  RegisterSet_Ctor(usedColors);
  Operation_ARGUMENTS_FOREACH_Temporary(pcopy, curArg) {
    if(!RegisterSet_insert(usedColors, Temporary_register(curArg))) {
      result = true;
      break;
    }
  } Operation_ARGUMENTS_ENDEACH_Temporary;
  RegisterSet_Dtor(usedColors);
  return result;
}

#line 473 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
ReachabilityNode
ReachabilityNode_Ctor(ReachabilityNode this, Memory memory, int32_t size) {
  *ReachabilityNode__MEMORY(this) = memory;
  *ReachabilityNode__SIZE(this) = size;
  RegisterSet_Ctor(ReachabilityNode_REACHES(this));
  // allocate one more space to avoid realloc when connecting a leaf
  ++size;
  if(size > 0) {
    int32_t i;
    *ReachabilityNode__REACHEDGES(this) =
      Memory_alloc(memory, sizeof(RegisterSet_) * size);
    for(i = 0; i < size; ++i) {
      RegisterSet_Ctor(&ReachabilityNode_REACHEDGES(this)[i]);
    }
    *ReachabilityNode__CHILDREN(this) =
      Memory_alloc(memory, sizeof(Register) * size);
    memset(ReachabilityNode_CHILDREN(this), 0, sizeof(Register) * size);
  } else {
    *ReachabilityNode__REACHEDGES(this) = NULL;
    *ReachabilityNode__CHILDREN(this) = NULL;
  }
  return this;
}

#line 497 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
void
ReachabilityNode_Dtor(ReachabilityNode this) {
  RegisterSet_Dtor(ReachabilityNode_REACHES(this));

#line 501 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  if(ReachabilityNode_REACHEDGES(this) != NULL) {
    int32_t i, size = ReachabilityNode_SIZE(this);
#ifndef ECL_NDTOR
    for(i = 0; i < size; ++i) {
      RegisterSet_Dtor(&ReachabilityNode_REACHEDGES(this)[i]);
    }
#endif
    Memory_free(ReachabilityNode_MEMORY(this), ReachabilityNode_REACHEDGES(this));
    Memory_free(ReachabilityNode_MEMORY(this), ReachabilityNode_CHILDREN(this));
  }
}

#line 513 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachability(PermutationNode rootNodes[], ReachabilityNode reachabilityNodes[],
                IDList reachabilityStorage, const_RegisterSet leaves);

#line 517 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static Register
candidateForResolution(RegisterSet leaves, RegisterSet freeRegisters,
                       RegisterSet isDuplicated, PermutationNode cpyPermNode,
                       ReachabilityNode reachabilityNodes[],
                       PermutationNode rootNodes[]);

#line 523 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
materializeCandidate(Selector selector, Operation parCopy,
                     int32_t idx, Register candidate);

#line 527 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
updatePermutationScheme(Register duplicatedRegister, Register candidate,
                        Register dst, int32_t cpyIdx, PermutationNode rootNodes[]);

#line 531 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
updateReachabilityScheme(Register duplicatedRegister, Register candidate,
                         int32_t cpyIdx, PermutationNode rootNodes[],
                         ReachabilityNode reachabilityNodes[]);

#line 537 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static bool
resolvePermutationScheme(Operation parCopy, Memory memory, Selector selector,
                         PermutationNode rootNodes[],
                         const_RegisterSet availableReg, RegisterSet leaves,
                         RegisterSet isDuplicated) {
  bool result = true;
  // free registers are available registers that are not involved in permutation
  // scheme. Available registers involved in permutation scheme are in leaves
  RegisterSet_ freeRegisters[1];
  ReachabilityNode reachabilityNodes[Register__COUNT];
  IDList_ reachabilityStorage[1];
  IDList_Ctor(reachabilityStorage, memory);
  memset(reachabilityNodes, 0, sizeof(ReachabilityNode) * Register__COUNT);
  RegisterSet_Copy(freeRegisters, availableReg);
  RegisterSet_diff(freeRegisters, leaves);
  setReachability(rootNodes, reachabilityNodes, reachabilityStorage, leaves);
  RegisterSet_FOREACH_Register(isDuplicated, duplicatedRegister) {
    // for a duplicated register four strategies are available to materialize
    // the duplication:
    // 1. take an immediate leaf: best, do not create additional copy
    // 2. take a leaf that is not reachable by the destination of the copy:
    //    avoid to create cycle and do not use an additional color
    // 3. take a new register: last choice
    // 4. take a reachable leaf: create a cycle and no color left ->
    //    will need a swap
    PermutationNode duplicatedNode = rootNodes[duplicatedRegister];
    IDList copies = PermutationNode_COPIES(duplicatedNode);
    bool isFirst = true;
    int32_t cpyIdx = IDList_count(copies);
    IDList_FORBACK(copies, PermutationNodeAndIdx_, p_cpyPermNode) {
      PermutationNode cpyPermNode = PermutationNodeAndIdx_NODE(p_cpyPermNode);
      Register candidate;
      --cpyIdx;
      // Renaming is not complicated but is boring because of parallel copy
      // arguments should be removed (to make the duplication disappear)
      // instead of updated. Two ways are possible:
      // * create a move, which represents the renaming
      // * propagate the renaming
      // Renaming is correctly handle in Florent.xcc, discardMoves
      Except_CHECK(PermutationNode_COLOR(cpyPermNode) != duplicatedRegister
                   && "Renaming is not supported in duplications removal");
      if(!RegisterSet_contains(leaves, PermutationNode_COLOR(cpyPermNode))
         && IDList_count(copies) == 2 && isFirst) {
        // first edge of permutation scheme is not simple.
        // Try the next one. If it is not simple too, the function will remove
        // the duplication
        isFirst = false;
        continue;
      }
      // Choose the best solution
      candidate = candidateForResolution(leaves, freeRegisters, isDuplicated,
                                         cpyPermNode, reachabilityNodes,
                                         rootNodes);
      if(candidate != Register__UNDEF) {
        // materialize solution
        // update permutation scheme and reachability
        materializeCandidate(selector, parCopy,
                             PermutationNodeAndIdx_IDX(p_cpyPermNode),
                             candidate);
        updatePermutationScheme(duplicatedRegister, candidate,
                                PermutationNode_COLOR(cpyPermNode), cpyIdx,
                                rootNodes);
        updateReachabilityScheme(duplicatedRegister,
                                 PermutationNode_COLOR(cpyPermNode), cpyIdx,
                                 rootNodes, reachabilityNodes);
      } else {
        result = false;
        break;
      }
    } IDList_ENDBACK;
    if(!result) {
      break;
    }
    RegisterSet_remove(isDuplicated, duplicatedRegister);
  } RegisterSet_ENDEACH_Register;
#ifndef ECL_NDTOR
  RegisterSet_Dtor(freeRegisters);
  IDList_Dtor(reachabilityStorage);
#endif
  return result;
}

#line 619 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
buildPermutationScheme(Operation parCopy, PermutationNode rootNodes[],
                       IDList permutationNodeStorage, Memory memory,
                       RegisterSet leaves, RegisterSet isDuplicated) {
  Operation_ARGUMENTS_FOREACH_Temporary(parCopy, curArg) {
    PermutationNode src;
    PermutationNode dst;
    Temporary dstTemp = Operation_getResult(parCopy, Operation_ARGUMENTS_INDEX);
    PermutationNodeAndIdx nodeAndIdx;
    Except_CHECK(Temporary_isDedicated(curArg) || Temporary_isAssigned(curArg));
    Except_CHECK(Temporary_isDedicated(dstTemp) || Temporary_isAssigned(dstTemp));
    if((src = rootNodes[Temporary_register(curArg)]) == NULL) {
      src = (PermutationNode)IDList_push(permutationNodeStorage,
                                         sizeof(PermutationNode_));
      PermutationNode_Ctor(src, Temporary_register(curArg), memory);
      rootNodes[Temporary_register(curArg)] = src;
    }
    if((dst = rootNodes[Temporary_register(dstTemp)]) == NULL) {
      dst = (PermutationNode)IDList_push(permutationNodeStorage,
                                         sizeof(PermutationNode_));
      PermutationNode_Ctor(dst, Temporary_register(dstTemp), memory);
      rootNodes[Temporary_register(dstTemp)] = dst;
      RegisterSet_insert(leaves, Temporary_register(dstTemp));
    }
    RegisterSet_remove(leaves, Temporary_register(curArg));
    // also insert index of usage to easy updating
    nodeAndIdx =
      (PermutationNodeAndIdx)IDList_push(PermutationNode_COPIES(src),
                                         sizeof(PermutationNodeAndIdx_));
    PermutationNodeAndIdx_Ctor(nodeAndIdx, dst, Operation_ARGUMENTS_INDEX);
    *PermutationNode__PARENT(dst) = src;
    if(IDList_count(PermutationNode_COPIES(src)) > 1) {
      RegisterSet_insert(isDuplicated, Temporary_register(curArg));
    }
  } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 655 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
}

#line 657 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachabilityNode(Register reg, PermutationNode rootNodes[],
                    ReachabilityNode reachabilityNodes[],
                    IDList reachabilityStorage);

#line 662 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachabilityOnLeaves(RegisterSet doneNodes, const_RegisterSet leaves,
                        ReachabilityNode reachabilityNodes[],
                        IDList reachabilityStorage,
                        PermutationNode rootNodes[]);

#line 668 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachabilityOnCycle(RegisterSet doneNodes,
                       ReachabilityNode reachabilityNodes[],
                       IDList reachabilityStorage,
                       PermutationNode rootNodes[]);

#line 674 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static ReachabilityNode
getReachNode(Register reg, ReachabilityNode reachabilityNodes[],
             IDList reachabilityStorage, PermutationNode rootNodes[]);

#line 679 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachability(PermutationNode rootNodes[], ReachabilityNode reachabilityNodes[],
                IDList reachabilityStorage, const_RegisterSet leaves) {
  RegisterSet_ doneNodes[1];
  Memory memory = IDList_memory(reachabilityStorage);

#line 685 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  RegisterSet_Ctor(doneNodes);

#line 687 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  setReachabilityOnLeaves(doneNodes, leaves, reachabilityNodes,
                          reachabilityStorage, rootNodes);
  // No more leaves, it remains only cycles
  setReachabilityOnCycle(doneNodes, reachabilityNodes, reachabilityStorage,
                         rootNodes);
#ifndef ECL_NDTOR
  RegisterSet_Dtor(doneNodes);
#endif
}

#line 697 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachabilityNode(Register reg, PermutationNode rootNodes[],
                    ReachabilityNode reachabilityNodes[],
                    IDList reachabilityStorage) {
  int32_t permutationEdgeIdx = 0;
  ReachabilityNode reachNode =
    getReachNode(reg, reachabilityNodes, reachabilityStorage, rootNodes);
  PermutationNode permNode = rootNodes[reg];
  IDList_FOREACH(PermutationNode_COPIES(permNode), PermutationNodeAndIdx_,
                 p_cpyPermNode) {
    PermutationNode cpyPermNode = PermutationNodeAndIdx_NODE(p_cpyPermNode);
    Register cpyColor = PermutationNode_COLOR(cpyPermNode);
    ReachabilityNode cpyReachNode =
      getReachNode(cpyColor,reachabilityNodes, reachabilityStorage, rootNodes);
    RegisterSet_insert(ReachabilityNode_REACHES(reachNode), cpyColor);
    RegisterSet_insert(&ReachabilityNode_REACHEDGES(reachNode)[permutationEdgeIdx],
                       cpyColor);
    ReachabilityNode_CHILDREN(reachNode)[permutationEdgeIdx] = cpyColor;
    RegisterSet_union(ReachabilityNode_REACHES(reachNode),
                      ReachabilityNode_REACHES(cpyReachNode));
    RegisterSet_union(&ReachabilityNode_REACHEDGES(reachNode)[permutationEdgeIdx],
                      ReachabilityNode_REACHES(cpyReachNode));
    ++permutationEdgeIdx;
  } IDList_ENDEACH;
}

#line 723 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachabilityOnLeaves(RegisterSet doneNodes, const_RegisterSet leaves,
                        ReachabilityNode reachabilityNodes[],
                        IDList reachabilityStorage,
                        PermutationNode rootNodes[]) {
  RegisterSet_ curLeaves[1]; 
  int32_t childrenSeen[Register__COUNT];
  RegisterSet_Copy(curLeaves, leaves);
  memset(childrenSeen, 0, sizeof(int32_t) * Register__COUNT);

#line 733 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  while(!RegisterSet_isEmpty(curLeaves)) {
    RegisterSet_ newLeaves[1];
    RegisterSet_Ctor(newLeaves);
    RegisterSet_FOREACH_Register(curLeaves, reg) {
      setReachabilityNode(reg, rootNodes, reachabilityNodes, reachabilityStorage);
      if(PermutationNode_PARENT(rootNodes[reg]) != NULL) {
        Register parentReg =
          PermutationNode_COLOR(PermutationNode_PARENT(rootNodes[reg]));
        if(++childrenSeen[parentReg] ==
           IDList_count(PermutationNode_COPIES(PermutationNode_PARENT(rootNodes[reg])))) {
          RegisterSet_insert(newLeaves, parentReg);
        }
      }
    } RegisterSet_ENDEACH_Register;
    RegisterSet_union(doneNodes, curLeaves);
    RegisterSet_Copy(curLeaves, newLeaves);
#ifndef ECL_NDTOR
    RegisterSet_Dtor(newLeaves);
#endif
  }

#line 754 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
#ifndef ECL_NDTOR
  RegisterSet_Dtor(curLeaves);
#endif
}

#line 759 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
struct CycleInfo_ {
  //@args int32_t idx, Register color
  int32_t IDX;
  Register COLOR;
};
typedef struct CycleInfo_ CycleInfo_, *CycleInfo;
typedef const struct CycleInfo_ *const_CycleInfo;
typedef struct CycleInfo_ * restrict_CycleInfo;

extern CycleInfo
CycleInfo_Ctor(CycleInfo this, int32_t idx, Register color);

extern CycleInfo
CycleInfo_Copy(CycleInfo this, const_CycleInfo that);

extern void
CycleInfo_Dtor(CycleInfo this);

extern size_t
CycleInfo_Size(int32_t idx, Register color);

#define CycleInfo_IDX(this) ((this)->IDX)
#define CycleInfo__IDX(this) (&(this)->IDX)
#define CycleInfo_COLOR(this) ((this)->COLOR)
#define CycleInfo__COLOR(this) (&(this)->COLOR)


#line 765 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
CycleInfo
CycleInfo_Ctor(CycleInfo this, int32_t idx, Register color) {
  *CycleInfo__IDX(this) = idx;
  *CycleInfo__COLOR(this) = color;
  return this;
}

#line 772 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
setReachabilityOnCycle(RegisterSet doneNodes,
                       ReachabilityNode reachabilityNodes[],
                       IDList reachabilityStorage,
                       PermutationNode rootNodes[]) {
  int32_t i;
  IDList_ cycleInfo[1];
  IDList_Ctor(cycleInfo, IDList_memory(reachabilityStorage));

#line 781 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  for(i = 0; i < Register__COUNT; ++i) {
    if(rootNodes[i] != NULL && !RegisterSet_contains(doneNodes, (Register)i)) {
      // get all reachable nodes in the cycle and assign the result to each
      // node of the cycle
      // Remember: A node is at most in one cycle by construction of a
      // permutation graph
      PermutationNode curNode = rootNodes[i];
      RegisterSet_ reachedColors[1];
      RegisterSet_Ctor(reachedColors);
      while(!RegisterSet_contains(doneNodes, PermutationNode_COLOR(curNode))) {
        int32_t cycleIdx = 0;
        RegisterSet_insert(doneNodes, PermutationNode_COLOR(curNode));
        IDList_FOREACH(PermutationNode_COPIES(curNode), PermutationNodeAndIdx_,
                       p_cpyNode) {
          PermutationNode cpyNode = PermutationNodeAndIdx_NODE(p_cpyNode);
          // if doneNodes does not contain cpyNode, cpyNode is on the cycle
          if(!RegisterSet_contains(doneNodes, PermutationNode_COLOR(cpyNode))) {
            Register color = PermutationNode_COLOR(curNode);
            CycleInfo info = (CycleInfo)IDList_push(cycleInfo, sizeof(CycleInfo_));
            ReachabilityNode node = getReachNode(color, reachabilityNodes,
                                                 reachabilityStorage, rootNodes);

#line 803 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
            CycleInfo_Ctor(info, cycleIdx, color);
            setReachabilityNode(color, rootNodes,
                                reachabilityNodes, reachabilityStorage);
            RegisterSet_union(reachedColors, ReachabilityNode_REACHES(node));
            RegisterSet_insert(reachedColors, PermutationNode_COLOR(cpyNode));
            // set cycle information
            *PermutationNode__CYCLEIDX(curNode) = cycleIdx;
            curNode = cpyNode;
            break;
          }
          ++cycleIdx;
        } IDList_ENDEACH;
      }
      IDList_FOREACH(cycleInfo, CycleInfo_, info) {
        ReachabilityNode node =
          getReachNode(CycleInfo_COLOR(info), reachabilityNodes,
                       reachabilityStorage, rootNodes);
        RegisterSet_union(&ReachabilityNode_REACHEDGES(node)[CycleInfo_IDX(info)],
                          reachedColors);
        RegisterSet_union(ReachabilityNode_REACHES(node), reachedColors);
      } IDList_ENDEACH;
      IDList_empty(cycleInfo, NULL);
    }
  }
  IDList_Dtor(cycleInfo);
}

#line 830 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static ReachabilityNode
getReachNode(Register reg, ReachabilityNode reachabilityNodes[],
             IDList reachabilityStorage, PermutationNode rootNodes[]) {
  ReachabilityNode reachNode = reachabilityNodes[reg];  
  if(reachNode == NULL) {
    reachNode =
      (ReachabilityNode)IDList_push(reachabilityStorage, sizeof(ReachabilityNode_));
    ReachabilityNode_Ctor(reachNode, IDList_memory(reachabilityStorage),
                          IDList_count(PermutationNode_COPIES(rootNodes[reg])));
    reachabilityNodes[reg] = reachNode;
  }
  return reachNode;
}

#line 844 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static Register
candidateForResolution(RegisterSet leaves, RegisterSet freeRegisters,
                       RegisterSet isDuplicated, PermutationNode cpyPermNode,
                       ReachabilityNode reachabilityNodes[],
                       PermutationNode rootNodes[]) {
  // see resolvePermutationScheme comments for the description of cases
  Register candidate = Register__UNDEF;
  // case 1:
  if(RegisterSet_contains(leaves, PermutationNode_COLOR(cpyPermNode))) {
    candidate = PermutationNode_COLOR(cpyPermNode);
    RegisterSet_remove(leaves, candidate);
  } else {
    RegisterSet_ leavesNotReachable[1];
    ReachabilityNode reachNode =
      reachabilityNodes[PermutationNode_COLOR(cpyPermNode)];
    Except_CHECK(reachNode != NULL && "node not treated as leaf or in a"
                 " cycle. Do not understand the construction");
    RegisterSet_Copy(leavesNotReachable, leaves);
    // if node is in a cycle consider edge of the cycle as back edge and
    // ignore reachability from that one
    if(PermutationNode_isInCycle(cpyPermNode)) {
      int32_t i;
      for(i = 0; i < ReachabilityNode_SIZE(reachNode); ++i) {
        if(i != PermutationNode_CYCLEIDX(cpyPermNode)) {
          RegisterSet_diff(leavesNotReachable,
                           &ReachabilityNode_REACHEDGES(reachNode)[i]);
        }
      }
    } else {
      RegisterSet_diff(leavesNotReachable,
                       ReachabilityNode_REACHES(reachNode));
    }
    // case 2:
    if(!RegisterSet_isEmpty(leavesNotReachable)) {
      RegisterSet_FOREACH_Register(leavesNotReachable, leaf) {
        candidate = leaf;
        Except_CHECK(PermutationNode_PARENT(rootNodes[candidate]) != NULL
                     && "This register must be in freeRegisters");
        if(!RegisterSet_contains(isDuplicated,
                                 PermutationNode_COLOR(PermutationNode_PARENT(rootNodes[candidate])))) {
          // find a good candidate to resolve duplication
          // otherwise, we may miss the opportunity to use case 1 to
          // resolve the duplication of the parent of this color
          break;
        }
      } RegisterSet_ENDEACH_Register;
    } else if(!RegisterSet_isEmpty(freeRegisters)) {
      // case 3:
      candidate = RegisterSet_choose(freeRegisters);
    } else {
      // case 4:
      // no more register left, try a reachable leaf
      if(!RegisterSet_isEmpty(leaves)) {
        candidate = RegisterSet_choose(leaves);
      } else {
        Except_CHECK(false && "by construction, we must have the same number "
                     "of leaves and duplication");
      }
    }
    RegisterSet_Dtor(leavesNotReachable);
  }
  return candidate;
}

#line 909 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
materializeCandidate(Selector selector, Operation parCopy,
                     int32_t idx, Register candidate) {
  Temporary tempArg = Operation_getArgument(parCopy, idx);
  Temporary tempRes =
    Procedure_makeAssignedTemporary(Selector_procedure(selector), candidate);
  Selector_makeOperationCOPY(selector, tempRes, tempArg);
  Operation_setArgument(parCopy, idx, tempRes);
}

#line 919 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
// Transform:
//    Dup   candidate
//     |\\_____
//     | \     \              .
//     |  \     \             .
//     |   \     \            .
//     |    \     \           .
//    dst1 dst2... dstN
// Into:
//    Dup  candidate
//     |\\_____|___________
//     | \     |           \       .
//     |  \     \           \      .
//     |   \     \           \     .
//     |    \     \           \    .
//    dst1 dst2...dstCpyIdx... dstN
static void
updatePermutationScheme(Register duplicatedRegister, Register candidate,
                        Register dst, int32_t cpyIdx, PermutationNode rootNodes[]) {
  PermutationNode duplicatedNode = rootNodes[duplicatedRegister];
  PermutationNode candidateNode = rootNodes[candidate];
  PermutationNode dstNode = rootNodes[dst];
  int32_t nbCpies = IDList_count(PermutationNode_COPIES(duplicatedNode));
  // set parent link between dst and candidate
  // candidate
  //    ^
  //    |
  //   dst
  // Note: candidateNode can be NULL but it does not matter
  *PermutationNode__PARENT(dstNode) = candidateNode;
  // remove link between duplicatedNode and dstNode
  // Dup
  //  |
  //  X
  //  |
  // dst
  IDList_FORBACK(PermutationNode_COPIES(duplicatedNode), PermutationNodeAndIdx_,
                 permAndIdx) {
    if(--nbCpies == cpyIdx) {
      Except_CHECK(PermutationNodeAndIdx_NODE(permAndIdx) == dstNode
                   && "We remove the wrong link!");
      IDList_remove(PermutationNode_COPIES(duplicatedNode), permAndIdx, NULL);
      break;
    }
  } IDList_ENDBACK;
  // Update cycle index
  if(PermutationNode_CYCLEIDX(duplicatedNode) >= cpyIdx) {
    *PermutationNode__CYCLEIDX(duplicatedNode) =
      PermutationNode_CYCLEIDX(duplicatedNode) - 1;
  }
  // Create link between candidateNode and dstNode
  // candidate
  //    |
  //    V
  //   dst
  if(candidateNode != NULL) {
    PermutationNodeAndIdx permNodeAndIdx;
    Except_CHECK(IDList_isEmpty(PermutationNode_COPIES(candidateNode))
                 && "candidateNode must be a leaf. Otherwise we create a"
                 " duplication");
    permNodeAndIdx =
      (PermutationNodeAndIdx)IDList_push(PermutationNode_COPIES(candidateNode),
                                         sizeof(PermutationNodeAndIdx_));
    PermutationNodeAndIdx_Ctor(permNodeAndIdx, dstNode, 0);
  }
}

#line 986 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
updateReachabilityLinkOfDuplicatedNode(Register duplicatedRegister,
                                       int32_t cpyIdx,
                                       ReachabilityNode reachabilityNodes[]) {
  ReachabilityNode reachNode = reachabilityNodes[duplicatedRegister];
  RegisterSet reachedColors = ReachabilityNode_REACHES(reachNode);
  int32_t i;
  RegisterSet_empty(reachedColors);
  for(i = 0; i < ReachabilityNode_SIZE(reachNode); ++i) {
    if(i != cpyIdx) {
      RegisterSet_union(reachedColors,
                        &ReachabilityNode_REACHEDGES(reachNode)[i]);
    }
    if(i >= cpyIdx && i != ReachabilityNode_SIZE(reachNode) - 1) {
      ReachabilityNode_REACHEDGES(reachNode)[i] =
        ReachabilityNode_REACHEDGES(reachNode)[i + 1];
      ReachabilityNode_CHILDREN(reachNode)[i] =
        ReachabilityNode_CHILDREN(reachNode)[i + 1];
    }
  }
  *ReachabilityNode__SIZE(reachNode) = ReachabilityNode_SIZE(reachNode) - 1;
}

#line 1009 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
updateAncestor(Register startPoint, PermutationNode rootNodes[],
               ReachabilityNode reachabilityNodes[]) {
  RegisterSet_ prevReachedColors[1];
  PermutationNode startNode = rootNodes[startPoint];
  PermutationNode prevNode = rootNodes[startPoint];
  PermutationNode curNode;
  ReachabilityNode startReachNode = reachabilityNodes[startPoint];
  RegisterSet_ saveStartColors[1];
  RegisterSet_Copy(saveStartColors, ReachabilityNode_REACHES(startReachNode));
  RegisterSet_Copy(prevReachedColors, saveStartColors);

#line 1021 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
  while((curNode = PermutationNode_PARENT(prevNode)) != NULL) {
    ReachabilityNode curReachNode =
      reachabilityNodes[PermutationNode_COLOR(curNode)];
    RegisterSet reachedColors = ReachabilityNode_REACHES(curReachNode);
    int32_t i;

#line 1027 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    RegisterSet_empty(reachedColors);
    for(i = 0; i < ReachabilityNode_SIZE(curReachNode); ++i) {
      if(ReachabilityNode_CHILDREN(curReachNode)[i] ==
         PermutationNode_COLOR(prevNode)) {
        RegisterSet_Copy(&ReachabilityNode_REACHEDGES(curReachNode)[i],
                         prevReachedColors);
        RegisterSet_insert(&ReachabilityNode_REACHEDGES(curReachNode)[i],
                           PermutationNode_COLOR(prevNode));
      }
      RegisterSet_union(ReachabilityNode_REACHES(curReachNode),
                        &ReachabilityNode_REACHEDGES(curReachNode)[i]);
    }

#line 1040 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    if(PermutationNode_COLOR(curNode) == startPoint) {
      if(RegisterSet_equals(reachedColors, saveStartColors)) {
        break;
      } else {
        RegisterSet_Copy(saveStartColors, reachedColors);
      }
    }

#line 1049 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
    // set next
    RegisterSet_Copy(prevReachedColors, reachedColors);
    prevNode = curNode;
  }
  RegisterSet_Dtor(prevReachedColors);
  RegisterSet_Dtor(saveStartColors);
}

#line 1057 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
enlargeParentCpyList(Register dstReg, PermutationNode rootNodes[],
                        ReachabilityNode reachabilityNodes[]) {
  PermutationNode parentNode = PermutationNode_PARENT(rootNodes[dstReg]);
  // if parent node is NULL, it means it is a new node, thus not accessible by
  // anyonelse, hence no need to update!
  if(parentNode != NULL) {
    ReachabilityNode parentReachNode =
      reachabilityNodes[PermutationNode_COLOR(parentNode)];
    Except_CHECK(parentReachNode != NULL && "Reachability should have been "
                 "compute for that node, since it was involved in permutation "
                 "scheme");
    Except_CHECK(ReachabilityNode_SIZE(parentReachNode) == 0
                 && "We insert a duplication!");
    // use the additional slot set purposely
    *ReachabilityNode__SIZE(parentReachNode) = 1;
    ReachabilityNode_CHILDREN(parentReachNode)[0] = dstReg;
  }
}

#line 1077 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
static void
updateReachabilityScheme(Register duplicatedRegister, Register dstReg,
                         int32_t cpyIdx, PermutationNode rootNodes[],
                         ReachabilityNode reachabilityNodes[]) {
  // update all ancestors of duplicatedRegister
  // same for dstReg
  updateReachabilityLinkOfDuplicatedNode(duplicatedRegister, cpyIdx,
                                         reachabilityNodes);
  updateAncestor(duplicatedRegister, rootNodes, reachabilityNodes);
  enlargeParentCpyList(dstReg, rootNodes, reachabilityNodes);
  updateAncestor(dstReg, rootNodes, reachabilityNodes);
}

#line 1090 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1339 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
void
ParallelCopy_INIT(void)
{
}

#line 1350 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
void
ParallelCopy_FINI(void)
{

#line 1354 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
}

#line 1356 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
#if XCC__C
#include "MDS/MDS_.h"
static void
ParallelCopy_TEST(void)
{
#include "CGO/ParallelCopy_.c"
}

#line 1364 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/ParallelCopy.xcc"
int
main(int argc, char **argv)
{
  CGO_Test(ParallelCopy_TEST);
  return 0;
}
#endif


#endif//ParallelCopy_c_INCLUDED
