#ifndef SSAReconstruct_c_INCLUDED
#define SSAReconstruct_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
/*
 * The problem: Given a value and a set of "copies" that are known to
 * represent the same abstract value, rewire all usages of the original value
 * to their closest copy while introducing phis as necessary.
 *
 * Algorithm: Mark all blocks in the iterated dominance frontiers of the value
 * and it's copies. Link the copies ordered by dominance to the blocks.  The
 * we search for each use all all definitions in the current block, if none is
 * found, then we search one in the immediate dominator. If we are in a block
 * of the dominance frontier, create a phi and search do the same search for
 * the phi arguments.
 *
 * A copy in this context means, that you want to introduce several new
 * abstract values (in Firm: nodes) for which you know, that they
 * represent the same concrete value. This is the case if you
 * - copy
 * - spill and reload
 * - re-materialize
 * a value.
 *
 * This function reroutes all uses of the original value to the copies in the
 * corresponding dominance subtrees and creates Phi functions where necessary.
 */
//REMARK: We use OperationResult structure for all manipulations of definitions.
// If we are sure that all definitions are in index 0 of the results array of
// an operation, we can use Operation instead and thus, reduce the footprint of
// this module.

#line 44 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
#include "CGO/CGO_.h"

#line 49 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 62 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
OperationArgumentList
OperationArgumentList_Ctor(OperationArgumentList this,
                           Memory memory)
{
  IDList_Ctor(OperationArgumentList____(this), memory);
  return this;
}

#line 70 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
OperationArgumentList_Dtor(OperationArgumentList this)
{
  IDList_Dtor(OperationArgumentList____(this));
}

#line 76 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
size_t
OperationArgumentList_Size(Memory memory)
{
  return sizeof(OperationArgumentList_);
}

#line 152 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 232 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 245 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
OperationResultList
OperationResultList_Ctor(OperationResultList this, Memory memory)
{
  IDList_Ctor(OperationResultList____(this), memory);
  return this;
}

#line 252 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
OperationResultList_Dtor(OperationResultList this)
{
  IDList_Dtor(OperationResultList____(this));
}

#line 258 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
size_t
OperationResultList_Size(Memory memory)
{
  return sizeof(OperationResultList_);
}

#line 334 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 406 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 434 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 445 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
DefUses
DefUses_Ctor(DefUses this, Memory memory)
{
  OperationResult_Ctor(DefUses__DEF_(this), (struct Operation_*)NULL, 0);
  OperationArgumentList_Ctor(DefUses__USES_(this), memory);
  return this;
}

#line 453 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
DefUses_Dtor(DefUses this)
{
  OperationResult_Dtor(DefUses__DEF_(this));
   OperationArgumentList_Dtor(DefUses__USES_(this));
}

#line 467 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
DefUses_setDef(DefUses this, struct Operation_*def, int32_t index) {
  OperationResult opAndIdx = DefUses_DEF(this);
  Except_ENSURE(this != NULL && "Attempt to set a definition on an invalid object");
  Except_CHECK(def != NULL);
  Except_CHECK(index >= 0 && index < Operation_resCount(def));
  // A check could be made to ensure that all uses and def reference the same
  // temporary
  *OperationResult__OPERATION(opAndIdx) = def;
  *OperationResult__RESINDEX(opAndIdx) = index;
}

#line 484 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
DefUses_setProducer(DefUses this, struct Operation_*def) {
  DefUses_setDef(this, def, 0);
}

#line 497 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
DefUses_addUse(DefUses this, struct Operation_*use, int32_t index) {
  OperationArgument_ opAndIdx[1];
  OperationArgument tmp;
  Except_ENSURE(this != NULL && "Attempt to add a use on an invalid object");
  Except_CHECK(use != NULL);
  Except_CHECK(index >= 0 && index < Operation_argCount(use));
  // A check could be made to ensure that all uses and def reference the same
  // temporary
  Except_CHECK(OperationArgumentList_isEmpty(DefUses_getUses(this)) ||
               ((tmp = OperationArgumentList_firstItem(DefUses_getUses(this)))
                && Operation_getArgument(OperationArgument_OPERATION(tmp),
                                         OperationArgument_ARGINDEX(tmp))
                == Operation_getArgument(use, index)));
  OperationArgument_Ctor(opAndIdx, use, index);
  OperationArgumentList_PUSH(DefUses_getUses(this), *opAndIdx);
}

#line 525 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
bool
DefUses_addUniqueUse(DefUses this, struct Operation_*use, int32_t index) {
  bool result = true;
  OperationArgument_ opAndIdx[1];
  OperationArgument_Ctor(opAndIdx, use, index);
  OperationArgumentList_FOREACH_OperationArgument(DefUses_USES(this), usage) {
    if(OperationArgument_equal(opAndIdx, usage)) {
      result = false;
      break;
    }
  } OperationArgumentList_ENDEACH_OperationArgument;
  if(result) {
    DefUses_addUse(this, use, index);
  }
  OperationArgument_Dtor(opAndIdx);
  return result;
}

#line 547 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
DefUses_removeUseItem(DefUses this, OperationArgument opndAndIdx) {
  OperationArgumentList_remove(DefUses_getUses(this), opndAndIdx,
                               (IDListItemRelease)NULL);
}

#line 562 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
bool
DefUses_removeUse(DefUses this, struct Operation_*use, int32_t index) {
  OperationArgumentList uses = DefUses_getUses(this);
  OperationArgumentList_FOREACH_OperationArgument(uses, it) {
    if(OperationArgument_OPERATION(it) == use
       && OperationArgument_ARGINDEX(it) == index) {
      OperationArgumentList_remove(uses, it, (IDListItemRelease)NULL);
      return true;
    }
  } OperationArgumentList_ENDEACH_OperationArgument;
  return false;
}

#line 579 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
OperationResult
DefUses_getDef(DefUses this) {
  Except_ENSURE(this != NULL && "Attempt to get definition on an invalid object");
  return DefUses_DEF(this);
}

#line 589 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
struct Operation_*
DefUses_getProducer(DefUses this) {
  OperationResult opAndIdx = DefUses_getDef(this);
  struct Operation_*result = NULL;
  if(opAndIdx) {
    result = OperationResult_OPERATION(opAndIdx);
  }
  return result;
}

#line 603 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
OperationArgumentList
DefUses_getUses(DefUses this) {
  Except_ENSURE(this != NULL && "Attempt to get uses on an invalid object");
  return DefUses_USES(this);
}

#line 616 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
#ifndef ECL_NTRACE
bool
DefUses_pretty(DefUses this, FILE* file, bool printArgIdx) {
  OperationArgumentList uses = DefUses_getUses(this);
  OperationResult def = DefUses_getDef(this);
  if(file == NULL) {
    file = stderr;
  }
  if(OperationResult_OPERATION(def)) {
    fprintf(file, "\tProducer");
    if(printArgIdx) {
      fprintf(file, " (res %d): ", OperationResult_RESINDEX(def));
    }
    Operation_pretty(OperationResult_OPERATION(def), file);
  } else {
    fprintf(file, "\tProducer not set\n");
  }

#line 634 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  OperationArgumentList_FOREACH_OperationArgument(uses, use) {
    fprintf(file, "\tConsumer");
    if(printArgIdx) {
      fprintf(file, " (arg %d): ", OperationArgument_ARGINDEX(use));
    }
    Operation_pretty(OperationArgument_OPERATION(use), file);
  } OperationArgumentList_ENDEACH_OperationArgument;
  return true;
}
#endif//ECL_NTRACE

#line 738 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
DefUses
Temporary_defUses(Temporary tempo) {
  Except_CHECK(tempo != NULL && Temporary_isVariable(tempo));
  return Variable_DEFUSES(Temporary_VARIABLE(tempo));
}

#line 748 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
Operation
Temporary_getProducer(Temporary tempo) {
  Except_CHECK(tempo != NULL && Temporary_isVariable(tempo));
  return DefUses_getProducer(Temporary_defUses(tempo));
}

#line 754 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 756 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
#define DBG(msg) // Nothing

#line 779 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
SSAReconstruct
SSAReconstruct_Ctor(SSAReconstruct this,
                    SSAForm ssaForm, Temporary value, Memory memory)
{
  *SSAReconstruct__MEMORY(this) = memory;
  *SSAReconstruct__SSAFORM(this) = ssaForm;
  *SSAReconstruct__VALUE(this) = value;
  *SSAReconstruct__COUNTER(this) =  IndexedFactory_COUNTER(BasicBlock_Factory);
  BitSet_Ctor(SSAReconstruct_MARKEDBASICBLOCK(this),
              SSAReconstruct_MEMORY(this),
              SSAReconstruct_COUNTER(this));
  IQueue_Ctor(SSAReconstruct_WORKLIST(this),
              SSAReconstruct_MEMORY(this), sizeof(BasicBlock),
              SSAReconstruct_COUNTER(this));
  *SSAReconstruct__DEFSPERBASICBLOCK(this) = NULL;
  *SSAReconstruct__MINDOM(this) = INT_MAX;
  *SSAReconstruct__MAXDOM(this) = 0;
  *SSAReconstruct__ISITERATEDFRONTIERCALCULATED(this) = false;
  return this;
}

#line 800 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
SSAReconstruct_Dtor(SSAReconstruct this)
{
  BitSet_Dtor(SSAReconstruct_MARKEDBASICBLOCK(this));
  IQueue_Dtor(SSAReconstruct_WORKLIST(this));
#ifndef ECL_NDTOR
  if(SSAReconstruct_DEFSPERBASICBLOCK(this) != NULL) {
    int32_t i;
    for(i = 0; i < SSAReconstruct_COUNTER(this); ++i) {
      if(SSAReconstruct_DEFSPERBASICBLOCK(this)[i] != NULL) {
         OperationResultList_Dtor(SSAReconstruct_DEFSPERBASICBLOCK(this)[i]);
         Memory_free_(SSAReconstruct_MEMORY(this),
                      SSAReconstruct_DEFSPERBASICBLOCK(this)[i]);
      }
    }
  }
  Memory_free_(SSAReconstruct_MEMORY(this),
               SSAReconstruct_DEFSPERBASICBLOCK(this));
#endif
}

#line 821 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static OperationResult
SSAReconstruct_searchDefEndOfBlock(SSAReconstruct this, BasicBlock block);

#line 824 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static OperationResultList
SSAReconstruct_getBlockDefs(SSAReconstruct this, BasicBlock block) {
  int32_t idx = BasicBlock_identity(block);
  Except_CHECK(SSAReconstruct_COUNTER(this) > idx);
  if(SSAReconstruct_DEFSPERBASICBLOCK(this) == NULL) {
    size_t size = sizeof(SSAReconstruct_DEFSPERBASICBLOCK(this)[0])
      * SSAReconstruct_COUNTER(this);
    *SSAReconstruct__DEFSPERBASICBLOCK(this) =
      Memory_alloc(SSAReconstruct_MEMORY(this), size);
    bzero(SSAReconstruct_DEFSPERBASICBLOCK(this), size);
  }
  if(SSAReconstruct_DEFSPERBASICBLOCK(this)[idx] == NULL) {
    // FIXME: ordering is not more maintained
    BasicBlock_makeOrderings(block, 0, 1);
    SSAReconstruct_DEFSPERBASICBLOCK(this)[idx] =
      Memory_alloc(SSAReconstruct_MEMORY(this),
                   sizeof(*SSAReconstruct_DEFSPERBASICBLOCK(this)[idx]));
    OperationResultList_Ctor(SSAReconstruct_DEFSPERBASICBLOCK(this)[idx],
                SSAReconstruct_MEMORY(this));
  }
  return SSAReconstruct_DEFSPERBASICBLOCK(this)[idx];
}

#line 847 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static bool
OperationResult_dominates(SSAReconstruct this, OperationResult opAndIdx1,
                            OperationResult opAndIdx2) {
  Operation op1 = OperationResult_OPERATION(opAndIdx1);
  Operation op2 = OperationResult_OPERATION(opAndIdx2);
  BasicBlock bb1 = Operation_basicBlock(op1);
  BasicBlock bb2 = Operation_basicBlock(op2);
  if(bb1 != bb2) {
    return Dominance_dominates(SSAReconstruct_DOMINANCE(this), bb1, bb2);
  }
  return Operation_ordering(op1) < Operation_ordering(op2);
}

#line 860 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static void
SSAReconstruct_updateDomzone(SSAReconstruct this, const_BasicBlock bl) {
  const int32_t *preDFN = Dominance_preDFN(SSAReconstruct_DOMINANCE(this));
  const int32_t *lastDFN = Dominance_lastDFN(SSAReconstruct_DOMINANCE(this));
  int32_t start = preDFN[BasicBlock_identity(bl)];
  int32_t end = lastDFN[BasicBlock_identity(bl)];

#line 867 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  *SSAReconstruct__MINDOM(this) = MIN(SSAReconstruct_MINDOM(this), start);
  *SSAReconstruct__MAXDOM(this) = MAX(SSAReconstruct_MAXDOM(this), end);
}

#line 871 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static int
isInside(int32_t what, int32_t low, int32_t hi) {
  return (low <= what) && (what <= hi);
}

#line 881 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
SSAReconstruct_addCopy(SSAReconstruct this, OperationResult copy) {
  BasicBlock block;
  OperationResultList defs;
  OperationResult next = NULL;

#line 887 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  Except_CHECK(!SSAReconstruct_ISITERATEDFRONTIERCALCULATED(this));

#line 889 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  block = Operation_basicBlock(OperationResult_OPERATION(copy));
  defs = SSAReconstruct_getBlockDefs(this, block);

#line 892 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  if(OperationResultList_isEmpty(defs)) {
    IQueue_PUT(SSAReconstruct_WORKLIST(this), BasicBlock, block);
  }
  OperationResultList_FOREACH_OperationResult(defs, iter) {
    if(OperationResult_equal(iter, copy)) {
      // already in defs list
      return;
    }
    if(OperationResult_dominates(this, iter, copy)) {
      next = iter;
      break;
    }
  } OperationResultList_ENDEACH_OperationResult;

#line 906 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  if(next == NULL) {
    OperationResultList_PUSH(defs, *copy);
  } else {
    *OperationResultList_insert(defs, next) = *copy;
  }
  SSAReconstruct_updateDomzone(this, block);
}

#line 914 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static BasicBlock
SSAReconstruct_getBasicBlockIdom(SSAReconstruct this, BasicBlock block) {
  BasicBlock result = Dominance_basicBlockIdom(SSAReconstruct_DOMINANCE(this), block);
  Except_CHECK(result != NULL && "Unable to find immediate dominator");
  return result;
}

#line 921 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static OperationResult
SSAReconstruct_getDefAtIdom(SSAReconstruct this, BasicBlock block) {
  BasicBlock dom = SSAReconstruct_getBasicBlockIdom(this, block);
  return SSAReconstruct_searchDefEndOfBlock(this, dom);
}

#line 927 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static OperationResult
SSAReconstruct_createPhi(SSAReconstruct this, BasicBlock block,
                         OperationResultList defs) {
  Operation phi = BasicBlock_makeOperationPHI(block, NULL);
  OperationResult_ phiAndIdx[1];
  Temporary value = SSAReconstruct_VALUE(this);
  Variable value_variable = Temporary_isVariable(value)
    ? Temporary_VARIABLE(value): NULL;
  SSAForm ssaForm = SSAReconstruct_SSAFORM(this);
  CodeRegion codeRegion = SSAForm_codeRegion(ssaForm);
  Temporary target = CodeRegion_makeVariableTemporary(codeRegion,
                                                      value_variable?Variable_ORIGIN(value_variable):value);
  Variable target_variable = Temporary_VARIABLE(target);
  int32_t argIdx = 0;
  // FIXME: ordering is not more maintained
  BasicBlock_makeOrderings(block, 0, 1);

#line 944 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  Operation_setResult(phi, 0, target);
  DefUses_setDef(Variable_DEFUSES(target_variable), phi, 0);
  OperationResult_Ctor(phiAndIdx, phi, 0);

#line 948 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  DBG((dbg, LEVEL_2, "\tcreating phi %p in %p\n", phi, block));
  OperationResultList_PUSH(defs, *phiAndIdx);

#line 951 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  BasicBlock_ENTER_FOREACH_BasicBlockEdge(block, edge) {
    BasicBlock predBlock = BasicBlockEdge_origBlock(edge);
    OperationResult predDef = SSAReconstruct_searchDefEndOfBlock(this, predBlock);
    Temporary predDefTmp = Operation_getResult(OperationResult_OPERATION(predDef),
                                               OperationResult_RESINDEX(predDef));
    Operation_setArgument(phi, argIdx, predDefTmp);
    if(Temporary_isVariable(predDefTmp)) {
      Except_CHECK(Temporary_isVariable(predDefTmp));
      DefUses_addUse(Variable_DEFUSES(Temporary_VARIABLE(predDefTmp)), phi, argIdx);
    }
    ++argIdx;
  } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
  OperationResult_Dtor(phiAndIdx);
  return OperationResultList_lastItem(defs);
}

#line 967 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static OperationResult
SSAReconstruct_searchDefEndOfBlock(SSAReconstruct this, BasicBlock block) {
  OperationResultList defs = SSAReconstruct_getBlockDefs(this, block);
  OperationResult def = NULL;
  if(!OperationResultList_isEmpty(defs)) {
    def = OperationResultList_firstItem(defs);
  } else if(BitSet_contains(SSAReconstruct_MARKEDBASICBLOCK(this),
                            BasicBlock_identity(block))) {
    def = SSAReconstruct_createPhi(this, block, defs);
  } else {
    def = SSAReconstruct_getDefAtIdom(this, block);
    OperationResultList_PUSH(defs, *def);
  }
  return def;
}

#line 983 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
static OperationResult
SSAReconstruct_searchDef(SSAReconstruct this, Operation point,
                         BasicBlock block) {
  OperationResultList defs = SSAReconstruct_getBlockDefs(this, block);
  
  DBG((dbg, LEVEL_3, "\t...searching def at %p\n", point));

#line 990 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  /* no defs in the current block we can do the normal searching */
  if(OperationResultList_isEmpty(defs)
     && !BitSet_contains(SSAReconstruct_MARKEDBASICBLOCK(this),
                         BasicBlock_identity(block))) {
    DBG((dbg, LEVEL_3, "\t...continue at idom\n"));
    return SSAReconstruct_getDefAtIdom(this, block);
  }

#line 998 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  /* there are defs in the current block, walk the linked list to find
     the one immediately dominating us
  */
  {
    // Dummy operation result built to represent point
    OperationResult_ at[1];
    OperationResult_Ctor(at, point, 0);

#line 1006 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
    OperationResultList_FOREACH_OperationResult(defs, iter) {
      if(OperationResult_OPERATION(at) != OperationResult_OPERATION(iter)
         && !OperationResult_dominates(this, at, iter)) {
        DBG((dbg, LEVEL_3, "\t...found dominating def %p\n", def));
        return iter;
      }
    } OperationResultList_ENDEACH_OperationResult;
    OperationResult_Dtor(at);
  }

#line 1016 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  /* block in dominance frontier? create a phi then */
  if(BitSet_contains(SSAReconstruct_MARKEDBASICBLOCK(this),
                     BasicBlock_identity(block))) {
    DBG((dbg, LEVEL_3, "\t...create phi at block %p\n", block));
    Except_CHECK(OperationResultList_isEmpty(defs)
                 || !Operator_isPhi(Operation_operator(OperationResult_OPERATION(OperationResultList_lastItem(defs)))));
    return SSAReconstruct_createPhi(this, block, defs);
  }

#line 1025 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  DBG((dbg, LEVEL_3, "\t...continue at idom (after checking block)\n"));
  return SSAReconstruct_getDefAtIdom(this, block);
}

#line 1033 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
SSAReconstruct_addCopies(SSAReconstruct this, OperationResultList copies) {
  size_t i;

#line 1037 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  Except_CHECK(!SSAReconstruct_ISITERATEDFRONTIERCALCULATED(this));

#line 1039 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  OperationResultList_FOREACH_OperationResult(copies, opAndIdx) {
    SSAReconstruct_addCopy(this, opAndIdx);
  } OperationResultList_ENDEACH_OperationResult;
}

#line 1044 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
/**
 * Calculates the iterated dominance frontier of a set of blocks. Marks the
 * blocks as visited. Sets the link fields of the blocks in the dominance
 * frontier to the block itself.
 */
static void
SSAReconstruct_markIteratedDominanceFrontiers(SSAReconstruct this) {
  Dominance dominance = SSAReconstruct_DOMINANCE(this);
  DBG((dbg, LEVEL_3, "Dominance Frontier:"));
  while(!IQueue_isEmpty(SSAReconstruct_WORKLIST(this))) {
    BasicBlock block;
    IQueue_POP(SSAReconstruct_WORKLIST(this), BasicBlock, block);

#line 1057 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
    Dominance_BasicBlock_FORFRONTIER_BasicBlock(dominance, block, frontier) {
      if (BitSet_contains(SSAReconstruct_MARKEDBASICBLOCK(this),
                          BasicBlock_identity(frontier))) {
        continue;
      }

#line 1063 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
      /*
       * It makes no sense to add phi-functions to blocks
       * that are not dominated by any definition;
       * all uses are dominated, hence the paths reaching the uses
       * have to stay in the dominance subtrees of the given definitions.
       */

#line 1070 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
      if(!isInside(Dominance_preDFN(dominance)[BasicBlock_index(frontier)],
                   SSAReconstruct_MINDOM(this), SSAReconstruct_MAXDOM(this))) {
        continue;
      }

#line 1075 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
      if(OperationResultList_isEmpty(SSAReconstruct_getBlockDefs(this,
                                                                   frontier))) {
        IQueue_PUT(SSAReconstruct_WORKLIST(this), BasicBlock, frontier);
      }

#line 1080 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
      DBG((dbg, LEVEL_3, " %+F", y));
      BitSet_insert(SSAReconstruct_MARKEDBASICBLOCK(this),
                    BasicBlock_identity(frontier));
    } Dominance_BasicBlock_ENDFRONTIER_BasicBlock;
  }
  DBG((dbg, LEVEL_3, "\n"));
}

#line 1092 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
SSAReconstruct_fixUsersList(SSAReconstruct this, OperationArgumentList uses) {
  // make this test to be able to call this function with dedicated and defined
  // uses list
  DefUses oldDefUses = Temporary_isVariable(SSAReconstruct_VALUE(this))
    ?Temporary_defUses(SSAReconstruct_VALUE(this)): NULL;
  if(!SSAReconstruct_ISITERATEDFRONTIERCALCULATED(this)) {
    SSAReconstruct_markIteratedDominanceFrontiers(this);
    *SSAReconstruct__ISITERATEDFRONTIERCALCULATED(this) = true;
  }

#line 1103 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
  OperationArgumentList_FOREACH_OperationArgument(uses, use) {
    /*
     * Search the valid def for each use and set it.
     */
    OperationArgument_ at[1];
    Operation useOp = OperationArgument_OPERATION(use);
    OperationResult def;
    Temporary newValue;
    DefUses newDefUses;
    BasicBlock block = Operation_basicBlock(useOp);
    OperationArgument_Ctor(at, OperationArgument_OPERATION(use),
                           OperationArgument_ARGINDEX(use));

#line 1116 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
    if(Operator_isPhi(Operation_operator(useOp))) {
      int32_t idx = 0;
      BasicBlock predBlock = NULL;

#line 1120 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
      Except_CHECK((Operation_getArgument(OperationArgument_OPERATION(use),
                                          OperationArgument_ARGINDEX(use))
                    == SSAReconstruct_VALUE(this))
                   && "User is not compatible with defined variables");
      BasicBlock_ENTER_FOREACH_BasicBlockEdge(block, edge) {
        if(idx == OperationArgument_ARGINDEX(use)) {
          predBlock = BasicBlockEdge_origBlock(edge);
          break;
        }
        ++idx;
      } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
      Except_NEVER(predBlock == NULL
                   && "Unable to find the source of the argument of the phi node");
      // The value of ARGINDEX for uses is used only in the previous loop.
      // So, it  does not matter from this point
      def = SSAReconstruct_searchDefEndOfBlock(this, predBlock);
    } else {

#line 1138 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
      def = SSAReconstruct_searchDef(this, OperationArgument_OPERATION(at), block);
    }

#line 1141 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
    Except_NEVER(def == NULL && "no definition found");

#line 1143 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
    newValue = Operation_getResult(OperationResult_OPERATION(def),
                                   OperationResult_RESINDEX(def));
    // synchronize before removal of the usage, since we may did some adding
    OperationArgumentList_FOREACH_SYNC;

#line 1148 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
    if(newValue != SSAReconstruct_VALUE(this)) {
      Operation_setArgument(useOp, OperationArgument_ARGINDEX(use), newValue);
      // Update def use link
      newDefUses = Temporary_defUses(newValue);
      DefUses_addUniqueUse(newDefUses, useOp, OperationArgument_ARGINDEX(use));
      if(oldDefUses != NULL) {
        DefUses_removeUse(oldDefUses, useOp, OperationArgument_ARGINDEX(use));
      }
    }
    OperationArgument_Dtor(at);
  } OperationArgumentList_ENDEACH_OperationArgument;
}

#line 1165 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
void
SSAReconstruct_fixUsers(SSAReconstruct this, Temporary value) {
/*   OperationArgumentList_ defUses[1]; */
  OperationArgumentList valUses;
  Except_CHECK(Temporary_isVariable(value) &&
               "def uses information on no-variable temporary not implemented");
/*   OperationArgumentList_Ctor(defUses, SSAReconstruct_MEMORY(this)); */
  valUses = DefUses_getUses(Temporary_defUses(value));
/*   OperationArgumentList_FOREACH_OperationArgument(valUses, use) { */
/*     OperationArgumentList_PUSH(defUses, *use); */
/*   } OperationArgumentList_ENDEACH_OperationArgument; */
  SSAReconstruct_fixUsersList(this, valUses);
/*   OperationArgumentList_Dtor(defUses); */
}

#line 1186 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1198 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
#if XCC__C
static void
SSAReconstruct_TEST(void)
{
#include "CGO/SSAReconstruct_.c"
}

#line 1205 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSAReconstruct.xcc"
int
main(int argc, char **argv)
{
  CGO_Test(SSAReconstruct_TEST);
  return 0;
}
#endif


#endif//SSAReconstruct_c_INCLUDED
