#ifndef SSASpill_c_INCLUDED
#define SSASpill_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Beladys spill algorithm version 2.
 *
 * The main differences to the original Belady are:
 * - The workset is empty at the start of a block
 *   There is no attempt to fill it with variables which
 *   are not used in the block.
 * - There is a global pass which tries to use the remaining
 *   capacity of the blocks to let global variables live through
 *   them.
 *
 * This implementation is a port of source code written by Sebastian in
 * libFirm of University of Karlsruhe (orginal file: ir/be/bespillbelady2.c).
 * Sources of that libraries are available at
 * http://www.info.uni-karlsruhe.de/software/libfirm/index.php?title=Main_Page
 */

#line 33 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#include "CGO/CGO_.h"

#line 38 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
// Dependence on SSAForm -> OK include by CGO.h which is include by CGO_.h

#line 41 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define DBG_SPILL     1
#define DBG_WSETS     2
#define DBG_FIX       4
#define DBG_DECIDE    8
#define DBG_START    16
#define DBG_SLOTS    32
#define DBG_TRACE    64
#define DBG_WORKSET 128
#define DBG_GLOBAL  256

#line 51 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define ALREADY_SPILLED_FACTOR 2

#line 53 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define DEAD       UINT_MAX
#define LIVE_END   (DEAD-1)
#define REMAT_DIST (DEAD-2)

#line 57 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define HUGE_VAL  1000000

#line 59 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define isPhi(op) ((op) && Operator_isPhi(Operation_operator(op)))

#line 61 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define BitSet_addTempo(set, tempo) BitSet_insert(set, Temporary_identity(tempo))
#define BitSet_containsTempo(set, tempo) BitSet_contains(set, Temporary_identity(tempo))

#line 64 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
// Options
static int already_spilled_factor = 2;
static int remat_live_range_ext   = 1;
static int global_pass_enabled    = 1;
static bool g_originalAlgo = false;

#line 70 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define DBG(a) // nothing

#line 72 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BeladyEnv_;
typedef struct BeladyEnv_ BeladyEnv_, *BeladyEnv;
typedef const struct BeladyEnv_ *const_BeladyEnv;
typedef struct BeladyEnv_ * restrict_BeladyEnv;

#line 74 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct WorkSet_;
typedef struct WorkSet_ WorkSet_, *WorkSet;
typedef const struct WorkSet_ *const_WorkSet;
typedef struct WorkSet_ * restrict_WorkSet;

#line 76 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
resetMark(BeladyEnv env);

#line 79 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
union  LocUnion_ {
  // A use time.
  uint32_t TIME;
  // Version
  uint32_t VER;
};
typedef union LocUnion_ LocUnion_, *LocUnion;

#line 87 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * An association between a node and a point in time.
 */
struct Loc_ {
  Temporary VAR;        /**< A node. */
  LocUnion_ LOCUNION_;
  //@mutate TIME &(Loc_LOCUNION_(this)).TIME
  //@access TIME (Loc_LOCUNION_(this)).TIME
  //@mutate VERSION &(Loc_LOCUNION_(this)).VER
  //@access VERSION (Loc_LOCUNION_(this)).VER
};
typedef struct Loc_ Loc_, *Loc;
typedef const struct Loc_ *const_Loc;
typedef struct Loc_ * restrict_Loc;

#define Loc_VAR(this) ((this)->VAR)
#define Loc__VAR(this) (&(this)->VAR)
#define Loc_LOCUNION_(this) ((this)->LOCUNION_)
#define Loc__LOCUNION_(this) (&(this)->LOCUNION_)
#define Loc_TIME(this) ((Loc_LOCUNION_(this)).TIME)
#define Loc_VERSION(this) ((Loc_LOCUNION_(this)).VER)
#define Loc__TIME(this) (&(Loc_LOCUNION_(this)).TIME)
#define Loc__VERSION(this) (&(Loc_LOCUNION_(this)).VER)


#line 99 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillEnv_;
typedef struct SpillEnv_ SpillEnv_, *SpillEnv;
typedef const struct SpillEnv_ *const_SpillEnv;
typedef struct SpillEnv_ * restrict_SpillEnv;

#line 101 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BeladyEnv_ {
  //@args Memory memory, SSAForm ssaForm, RegFile regFile, bool removePhiMemory
  Memory MEMORY;
  SSAForm SSAFORM;
  RegFile REGFILE;
  // Current instruction
  Operation INSTR;
  // Status on phi with memory operands
  bool REMOVEPHIMEMORY;
  // Number of register in REGFILE
  uint32_t NREGS;
  // Liveness, complex object -> constructed in the auxiliary constructor
  Liveness LIVENESS;
  //@access BLOCKCOUNT CodeRegion_blockCount(BeladyEnv_CODEREGION(this))
  // The main workset used while processing a block. Complex object
  struct WorkSet_ *WORKSET;
  // Spill environment used to manage spill/reload (location, insertion).
  // Complex object
  struct SpillEnv_ *SPILLENV;
  // Bitset to keep all the temporaries' id which have already been spilled.
  // Complex object
  BitSet_ SPILLED_;
  //@access SPILLED BeladyEnv__SPILLED_(this)
  // Shortcut for codeRegion access
  //@access CODEREGION SSAForm_codeRegion(BeladyEnv_SSAFORM(this))
  //@access CONVENTION CodeRegion_convention(BeladyEnv_CODEREGION(this))
};

extern BeladyEnv
BeladyEnv_Ctor(BeladyEnv this, Memory memory, SSAForm ssaForm, RegFile regFile, bool removePhiMemory);

extern BeladyEnv
BeladyEnv_Copy(BeladyEnv this, const_BeladyEnv that);

extern void
BeladyEnv_Dtor(BeladyEnv this);

extern size_t
BeladyEnv_Size(Memory memory, SSAForm ssaForm, RegFile regFile, bool removePhiMemory);

#define BeladyEnv_MEMORY(this) ((this)->MEMORY)
#define BeladyEnv__MEMORY(this) (&(this)->MEMORY)
#define BeladyEnv_SSAFORM(this) ((this)->SSAFORM)
#define BeladyEnv__SSAFORM(this) (&(this)->SSAFORM)
#define BeladyEnv_REGFILE(this) ((this)->REGFILE)
#define BeladyEnv__REGFILE(this) (&(this)->REGFILE)
#define BeladyEnv_INSTR(this) ((this)->INSTR)
#define BeladyEnv__INSTR(this) (&(this)->INSTR)
#define BeladyEnv_REMOVEPHIMEMORY(this) ((this)->REMOVEPHIMEMORY)
#define BeladyEnv__REMOVEPHIMEMORY(this) (&(this)->REMOVEPHIMEMORY)
#define BeladyEnv_NREGS(this) ((this)->NREGS)
#define BeladyEnv__NREGS(this) (&(this)->NREGS)
#define BeladyEnv_LIVENESS(this) ((this)->LIVENESS)
#define BeladyEnv__LIVENESS(this) (&(this)->LIVENESS)
#define BeladyEnv_WORKSET(this) ((this)->WORKSET)
#define BeladyEnv__WORKSET(this) (&(this)->WORKSET)
#define BeladyEnv_SPILLENV(this) ((this)->SPILLENV)
#define BeladyEnv__SPILLENV(this) (&(this)->SPILLENV)
#define BeladyEnv_SPILLED_(this) ((this)->SPILLED_)
#define BeladyEnv__SPILLED_(this) (&(this)->SPILLED_)
#define BeladyEnv_CONVENTION(this) (CodeRegion_convention(BeladyEnv_CODEREGION(this)))
#define BeladyEnv_CODEREGION(this) (SSAForm_codeRegion(BeladyEnv_SSAFORM(this)))
#define BeladyEnv_SPILLED(this) (BeladyEnv__SPILLED_(this))
#define BeladyEnv_BLOCKCOUNT(this) (CodeRegion_blockCount(BeladyEnv_CODEREGION(this)))


#line 129 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct WorkSet_ *
WorkSet_make(struct BeladyEnv_ *env);

#line 132 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillEnv_ *
SpillEnv_make(Memory memory, SSAForm ssaForm);

#line 135 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
BeladyEnv
BeladyEnv_Ctor(BeladyEnv this,
               Memory memory, SSAForm ssaForm, RegFile regFile, bool removePhiMemory)
{
  CodeRegion codeRegion = SSAForm_codeRegion(ssaForm);
  TemporaryTable globalTable = CodeRegion_globalTable(codeRegion);
  Convention convention = CodeRegion_convention(codeRegion);
  RegisterSet_ registerSet[1];
  *registerSet = Convention_allocatableSet(convention, regFile);
  *BeladyEnv__MEMORY(this) = memory;
  *BeladyEnv__SSAFORM(this) = ssaForm;
  *BeladyEnv__REGFILE(this) = regFile;
  *BeladyEnv__INSTR(this) = NULL;
  *BeladyEnv__REMOVEPHIMEMORY(this) = removePhiMemory;
  *BeladyEnv__NREGS(this) = RegisterSet_count(registerSet);
  // Should we force the adding of all SSA variables in the globalTable?
  *BeladyEnv__LIVENESS(this) = Liveness_make(memory, codeRegion, globalTable, true);
  Liveness_multiplex(BeladyEnv_LIVENESS(this));
  *BeladyEnv__WORKSET(this) = WorkSet_make(this);
  *BeladyEnv__SPILLENV(this) = SpillEnv_make(memory, ssaForm);
  BitSet_Ctor(BeladyEnv_SPILLED(this), memory,
              IndexedFactory_COUNTER(Temporary_Factory));
  return this;
}

#line 160 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
WorkSet_kill(struct WorkSet_ *this);

#line 163 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
SpillEnv_kill(struct SpillEnv_ *this);

#line 166 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
BeladyEnv_Dtor(BeladyEnv this) {
  Liveness_kill(BeladyEnv_LIVENESS(this));
  WorkSet_kill(BeladyEnv_WORKSET(this));
  SpillEnv_kill(BeladyEnv_SPILLENV(this));
}

#line 173 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static Operation
skipPhis(Operation op) {
  do {
    op = Operation_next(op);
  } while(op && isPhi(op));
  return op;
}

#line 182 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillPoint_ {
  //@args BasicBlock block, Operation point
  Operation BEFORE;
  Operation AFTER;
  BasicBlock BLOCK;
};
typedef struct SpillPoint_ SpillPoint_, *SpillPoint;
typedef const struct SpillPoint_ *const_SpillPoint;
typedef struct SpillPoint_ * restrict_SpillPoint;

extern SpillPoint
SpillPoint_Ctor(SpillPoint this, BasicBlock block, Operation point);

extern SpillPoint
SpillPoint_Copy(SpillPoint this, const_SpillPoint that);

extern void
SpillPoint_Dtor(SpillPoint this);

extern size_t
SpillPoint_Size(BasicBlock block, Operation point);

#define SpillPoint_BEFORE(this) ((this)->BEFORE)
#define SpillPoint__BEFORE(this) (&(this)->BEFORE)
#define SpillPoint_AFTER(this) ((this)->AFTER)
#define SpillPoint__AFTER(this) (&(this)->AFTER)
#define SpillPoint_BLOCK(this) ((this)->BLOCK)
#define SpillPoint__BLOCK(this) (&(this)->BLOCK)


#line 189 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
SpillPoint
SpillPoint_Ctor(SpillPoint this, BasicBlock block, Operation point)
{
  Operation after = NULL, before = point;
  do {
    after = before;
  } while(before && (before = Operation_next(before)) && isPhi(before));
  *SpillPoint__BEFORE(this) = before;
  *SpillPoint__AFTER(this) = after;
  *SpillPoint__BLOCK(this) = block;
  if(after == NULL) {
    // after is NULL means before the first instruction (I am after nobody, thus, I am first)
    *SpillPoint__BEFORE(this) = BasicBlock_firstOperation(block);
   
  }
  return this;
}

#line 207 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillLoc_ {
  //@args BasicBlock block, Operation point
  // First spill operation
  Operation SPILL;
  SpillPoint_ POINT_;
  //@access POINT SpillLoc__POINT_(this)
  // Location
  Temporary LOC;
};
typedef struct SpillLoc_ SpillLoc_, *SpillLoc;
typedef const struct SpillLoc_ *const_SpillLoc;
typedef struct SpillLoc_ * restrict_SpillLoc;

extern SpillLoc
SpillLoc_Ctor(SpillLoc this, BasicBlock block, Operation point);

extern SpillLoc
SpillLoc_Copy(SpillLoc this, const_SpillLoc that);

extern void
SpillLoc_Dtor(SpillLoc this);

extern size_t
SpillLoc_Size(BasicBlock block, Operation point);

#define SpillLoc_SPILL(this) ((this)->SPILL)
#define SpillLoc__SPILL(this) (&(this)->SPILL)
#define SpillLoc_POINT_(this) ((this)->POINT_)
#define SpillLoc__POINT_(this) (&(this)->POINT_)
#define SpillLoc_LOC(this) ((this)->LOC)
#define SpillLoc__LOC(this) (&(this)->LOC)
#define SpillLoc_POINT(this) (SpillLoc__POINT_(this))


#line 217 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
SpillLoc
SpillLoc_Ctor(SpillLoc this,
              BasicBlock block, Operation point)
{
  *SpillLoc__SPILL(this) = NULL;
  SpillPoint_Ctor(SpillLoc_POINT(this), block, point);
  *SpillLoc__LOC(this) = NULL;
  return this;
}

#line 227 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline Operation
SpillLoc_operationPoint(SpillLoc this) {
  return SpillPoint_BEFORE(SpillLoc_POINT(this)) != NULL?
    SpillPoint_BEFORE(SpillLoc_POINT(this))
    : SpillPoint_AFTER(SpillLoc_POINT(this));
}

#line 234 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline Operation
SpillLoc_operationBefore(SpillLoc this) {
  return SpillPoint_BEFORE(SpillLoc_POINT(this));
}

#line 239 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline Operation
SpillLoc_operationAfter(SpillLoc this) {
  return SpillPoint_AFTER(SpillLoc_POINT(this));
}

#line 244 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillInfo_ {
  //@args Memory memory, Temporary var
  Temporary VAR;
  IDList_ SPILLS_;
  //@access SPILLS SpillInfo__SPILLS_(this)
  IDList_ RELOADS_;
  //@access RELOADS SpillInfo__RELOADS_(this)
  double SPILLCOSTS;
};
typedef struct SpillInfo_ SpillInfo_, *SpillInfo;
typedef const struct SpillInfo_ *const_SpillInfo;
typedef struct SpillInfo_ * restrict_SpillInfo;

extern SpillInfo
SpillInfo_Ctor(SpillInfo this, Memory memory, Temporary var);

extern SpillInfo
SpillInfo_Copy(SpillInfo this, const_SpillInfo that);

extern void
SpillInfo_Dtor(SpillInfo this);

extern size_t
SpillInfo_Size(Memory memory, Temporary var);

#define SpillInfo_VAR(this) ((this)->VAR)
#define SpillInfo__VAR(this) (&(this)->VAR)
#define SpillInfo_SPILLS_(this) ((this)->SPILLS_)
#define SpillInfo__SPILLS_(this) (&(this)->SPILLS_)
#define SpillInfo_RELOADS_(this) ((this)->RELOADS_)
#define SpillInfo__RELOADS_(this) (&(this)->RELOADS_)
#define SpillInfo_SPILLCOSTS(this) ((this)->SPILLCOSTS)
#define SpillInfo__SPILLCOSTS(this) (&(this)->SPILLCOSTS)
#define SpillInfo_SPILLS(this) (SpillInfo__SPILLS_(this))
#define SpillInfo_RELOADS(this) (SpillInfo__RELOADS_(this))


#line 254 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
SpillInfo
SpillInfo_Ctor(SpillInfo this,
               Memory memory, Temporary var) {
  *SpillInfo__VAR(this) = var;
  IDList_Ctor(SpillInfo_SPILLS(this), memory);
  IDList_Ctor(SpillInfo_RELOADS(this), memory);
  *SpillInfo__SPILLCOSTS(this) = -1;
  return this;
}

#line 264 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
SpillInfo_Dtor(SpillInfo this)
{
  IDList_Dtor(SpillInfo_SPILLS(this));
  IDList_Dtor(SpillInfo_RELOADS(this));
}

#line 271 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillEnv_ {
  //@args Memory memory, SSAForm ssaForm
  Memory MEMORY;
  SpillInfo *VARSPILLLOCS;
  uint32_t COUNTER;
  uint32_t LASTIDX;
  TemporarySet_ MEMPHIS_;
  //@access MEMPHIS SpillEnv__MEMPHIS_(this)
  SSAForm SSAFORM;
};

extern SpillEnv
SpillEnv_Ctor(SpillEnv this, Memory memory, SSAForm ssaForm);

extern SpillEnv
SpillEnv_Copy(SpillEnv this, const_SpillEnv that);

extern void
SpillEnv_Dtor(SpillEnv this);

extern size_t
SpillEnv_Size(Memory memory, SSAForm ssaForm);

#define SpillEnv_MEMORY(this) ((this)->MEMORY)
#define SpillEnv__MEMORY(this) (&(this)->MEMORY)
#define SpillEnv_VARSPILLLOCS(this) ((this)->VARSPILLLOCS)
#define SpillEnv__VARSPILLLOCS(this) (&(this)->VARSPILLLOCS)
#define SpillEnv_COUNTER(this) ((this)->COUNTER)
#define SpillEnv__COUNTER(this) (&(this)->COUNTER)
#define SpillEnv_LASTIDX(this) ((this)->LASTIDX)
#define SpillEnv__LASTIDX(this) (&(this)->LASTIDX)
#define SpillEnv_MEMPHIS_(this) ((this)->MEMPHIS_)
#define SpillEnv__MEMPHIS_(this) (&(this)->MEMPHIS_)
#define SpillEnv_SSAFORM(this) ((this)->SSAFORM)
#define SpillEnv__SSAFORM(this) (&(this)->SSAFORM)
#define SpillEnv_MEMPHIS(this) (SpillEnv__MEMPHIS_(this))


#line 282 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
SpillEnv
SpillEnv_Ctor(SpillEnv this, Memory memory, SSAForm ssaForm)
{
  *SpillEnv__MEMORY(this) = memory;
  *SpillEnv__VARSPILLLOCS(this) = NULL;
  *SpillEnv__COUNTER(this) = IndexedFactory_COUNTER(Temporary_Factory);
  *SpillEnv__LASTIDX(this) = 0;
  TemporarySet_Ctor(SpillEnv_MEMPHIS(this), memory, SpillEnv_COUNTER(this));
  *SpillEnv__SSAFORM(this) = ssaForm;
  return this;
}

#line 294 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
SpillEnv_Dtor(SpillEnv this)
{
  TemporarySet_Dtor(SpillEnv_MEMPHIS(this));
#ifndef ECL_NDTOR
  if(SpillEnv_VARSPILLLOCS(this) != NULL) {
    int32_t i;
    for(i = 0; i < SpillEnv_COUNTER(this); ++i) {
      if(SpillEnv_VARSPILLLOCS(this)[i] != NULL) {
         SpillInfo_Dtor(SpillEnv_VARSPILLLOCS(this)[i]);
         Memory_free_(SpillEnv_MEMORY(this), SpillEnv_VARSPILLLOCS(this)[i]);
      }
    }
  }
  Memory_free_(SpillEnv_MEMORY(this), SpillEnv_VARSPILLLOCS(this));
#endif
}

#line 312 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_determineSpillCosts(SpillEnv env, Temporary toSpill);

#line 315 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_spillTemporary(SpillEnv this, Temporary temporary);

#line 318 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static SpillInfo
SpillEnv_getSpillInfo(SpillEnv this, Temporary var) {
  int32_t idx = Temporary_identity(var);
  Except_CHECK(SpillEnv_COUNTER(this) > idx);
  if(SpillEnv_VARSPILLLOCS(this) == NULL) {
    size_t size = sizeof(SpillEnv_VARSPILLLOCS(this)[0])
      * SpillEnv_COUNTER(this);
    *SpillEnv__VARSPILLLOCS(this) = Memory_alloc(SpillEnv_MEMORY(this), size);
    bzero(SpillEnv_VARSPILLLOCS(this), size);
  }
  if(SpillEnv_VARSPILLLOCS(this)[idx] == NULL) {
    SpillEnv_VARSPILLLOCS(this)[idx] =
      Memory_alloc(SpillEnv_MEMORY(this), sizeof(*SpillEnv_VARSPILLLOCS(this)[idx]));
    SpillInfo_Ctor(SpillEnv_VARSPILLLOCS(this)[idx], SpillEnv_MEMORY(this), var);
  }
  return SpillEnv_VARSPILLLOCS(this)[idx];
}

#line 336 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_addReload_(SpillEnv this, Temporary var, BasicBlock block,
                    Operation point, bool isAfter) {
  SpillInfo spillLocs = SpillEnv_getSpillInfo(this, var);
  SpillLoc_ spillRange[1];
  // SpillLoc_Ctor expects that given operation represent an after point
  if(isAfter) {
    SpillLoc_Ctor(spillRange, block, point);
  } else {
    if(point != NULL && Operation_prev(point) != NULL) {
      SpillLoc_Ctor(spillRange, block, Operation_prev(point));
    } else {
      SpillLoc_Ctor(spillRange, block, point);
      *SpillPoint__BEFORE(SpillLoc_POINT(spillRange)) = point;
      *SpillPoint__AFTER(SpillLoc_POINT(spillRange)) = NULL;
    }
  }
  IDList_PUSH(SpillInfo_RELOADS(spillLocs), SpillLoc_, *spillRange);
}

#line 356 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_addReload(SpillEnv this, Temporary var, Operation before) {
  SpillEnv_addReload_(this, var, Operation_basicBlock(before), before, false);
}

#line 361 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_addReloadAtEnd(SpillEnv this, Temporary var, BasicBlock bb) {
  SpillEnv_addReload_(this, var, bb, BasicBlock_lastOperation(bb), true);
}

#line 366 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static bool
Operation_dominates(SpillEnv this, Operation op1, Operation op2) {
  BasicBlock bb1 = Operation_basicBlock(op1);
  BasicBlock bb2 = Operation_basicBlock(op2);
  if(bb1 != bb2) {
    return Dominance_dominates(SSAForm_makeDominance(SpillEnv_SSAFORM(this), 0),
                               bb1, bb2);
  }
  return Operation_ordering(op1) > Operation_ordering(op2);
}

#line 377 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_addSpill(SpillEnv this, Temporary var, BasicBlock block,
                  Operation point) {
  SpillInfo spillLocs = SpillEnv_getSpillInfo(this, var);
  SpillLoc_ spillLoc[1];
  SpillLoc_Ctor(spillLoc, block, point);
  point = SpillLoc_operationPoint(spillLoc);
  IDList_FOREACH(SpillInfo_SPILLS(spillLocs), SpillLoc_, iter) {
    Operation iterPoint = SpillLoc_operationPoint(iter);
    // No need to add this spill if it is dominated by another
    if(Operation_dominates(this, iterPoint, point)) {
      return;
    }
    // Remove spills that we dominate
    if(Operation_dominates(this, point, iterPoint)) {
      IDList_remove_(SpillInfo_SPILLS(spillLocs), iter);
    }
  } IDList_ENDEACH;
  IDList_PUSH(SpillInfo_SPILLS(spillLocs), SpillLoc_, *spillLoc);
}

#line 398 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_spillPhi(SpillEnv this, Operation phi) {
  Except_CHECK(isPhi(phi));
  Except_CHECK(Operation_resCount(phi) == 1);
  TemporarySet_insert(SpillEnv_MEMPHIS(this), Operation_getResult(phi, 0));
  Operation_ARGUMENTS_FOREACH_Temporary(phi, arg) {
    Operation producer = Temporary_getProducer(arg);
    SpillEnv_addSpill(this, arg, Operation_basicBlock(producer), producer);
  } Operation_ARGUMENTS_ENDEACH_Temporary;
}

#line 409 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
SpillEnv
SpillEnv_make(Memory memory, SSAForm ssaForm) {
  SpillEnv res = Memory_alloc(memory, sizeof(*res));
  SpillEnv_Ctor(res, memory, ssaForm);
  return res;
}

#line 416 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
SpillEnv_kill(SpillEnv this) {
  if (this != NULL) {
    SpillEnv_Dtor(this);
#ifndef ECL_NDTOR
    Memory_free_(SpillEnv_MEMORY(this), this);
#endif//ECL_NDTOR
  }
}

#line 426 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_spillPhi_(SpillEnv this, SpillInfo spillInfo) {
  // TODO phi of memory addresses
  Temporary var = SpillInfo_VAR(spillInfo);
  Variable variable = Temporary_VARIABLE(var);
  Operation producer = Temporary_getProducer(var);
  BasicBlock block = Operation_basicBlock(producer);
  SSAForm ssaForm = SpillEnv_SSAFORM(this);
  CodeRegion codeRegion = SSAForm_codeRegion(ssaForm);
  SpillLoc_ spillLoc[1];
  int32_t i = 0;
  Except_CHECK(Temporary_isVariable(var));

#line 439 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  // To have this support, we must have a global handling of def uses chains,
  // i.e. not only on temporary variable but also on symbol, etc.
  //  Except_CHECK(false && "Phi spilling not yet implemented");
  Except_CHECK(isPhi(producer));
  DBG((dbg, LEVEL_1, "spilling Phi %p:\n", producer));

#line 445 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  SpillLoc_Ctor(spillLoc, block, producer);

#line 447 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  *SpillLoc__SPILL(spillLoc) = BasicBlock_makeOperationPHI(block, NULL);
  // FIXME: Instead of setting a variable, we should use a memory location.
  // I.e. we should give which symbol must be used to reload.
  // Currently it is not possibly, since Def uses chains are available only on
  // SSA variable type
  *SpillLoc__LOC(spillLoc) =
      CodeRegion_makeVariableTemporary(codeRegion, Variable_ORIGIN(variable));
//  *SpillLoc__LOC(spillLoc) = Spill_genStackedLocation(SSAVariable_ORIGIN(variable));

#line 456 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  Operation_setResult(SpillLoc_SPILL(spillLoc), 0, SpillLoc_LOC(spillLoc));
  DefUses_setDef(Temporary_defUses(SpillLoc_LOC(spillLoc)),
                 SpillLoc_SPILL(spillLoc), 0);
  IDList_empty(SpillInfo_SPILLS(spillInfo), NULL);
  IDList_PUSH(SpillInfo_SPILLS(spillInfo), SpillLoc_, *spillLoc);

#line 462 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  Operation_ARGUMENTS_FOREACH_Temporary(producer, tempo) {
    Temporary location;
    SpillEnv_determineSpillCosts(this, tempo);
    SpillEnv_spillTemporary(this, tempo);
    Except_CHECK(!IDList_isEmpty(SpillInfo_SPILLS(SpillEnv_getSpillInfo(this, tempo))));
    location = SpillLoc_LOC((SpillLoc)IDList_firstItem
                            (SpillInfo_SPILLS(SpillEnv_getSpillInfo(this, tempo))));
    Except_CHECK(location != NULL);
    Operation_setArgument(SpillLoc_SPILL(spillLoc), i, location);
    // in a specific way, we should describe the fact that tempo is used by
    // this operation, doesn't it?
    if (Temporary_isVariable(location)) {
      DefUses_addUse(Temporary_defUses(location),
                     SpillLoc_SPILL(spillLoc), i);
    }
    ++i;
  } Operation_ARGUMENTS_ENDEACH_Temporary;
  DBG((dbg, LEVEL_1, "... done spilling Phi %p, created PhiM %p\n", producer,
       SpillLoc_SPILL(spillLoc)));
}

#line 483 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct SpillLeaveParam_ {
  //@args Temporary to_spill, SpillLoc spillLoc
  Temporary TOSPILL;
  SpillLoc SPILLLOC;
};
typedef struct SpillLeaveParam_ SpillLeaveParam_, *SpillLeaveParam;
typedef const struct SpillLeaveParam_ *const_SpillLeaveParam;
typedef struct SpillLeaveParam_ * restrict_SpillLeaveParam;

extern SpillLeaveParam
SpillLeaveParam_Ctor(SpillLeaveParam this, Temporary to_spill, SpillLoc spillLoc);

extern SpillLeaveParam
SpillLeaveParam_Copy(SpillLeaveParam this, const_SpillLeaveParam that);

extern void
SpillLeaveParam_Dtor(SpillLeaveParam this);

extern size_t
SpillLeaveParam_Size(Temporary to_spill, SpillLoc spillLoc);

#define SpillLeaveParam_TOSPILL(this) ((this)->TOSPILL)
#define SpillLeaveParam__TOSPILL(this) (&(this)->TOSPILL)
#define SpillLeaveParam_SPILLLOC(this) ((this)->SPILLLOC)
#define SpillLeaveParam__SPILLLOC(this) (&(this)->SPILLLOC)


#line 489 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
SpillLeaveParam
SpillLeaveParam_Ctor(SpillLeaveParam this, Temporary to_spill, SpillLoc spillLoc) {
  *SpillLeaveParam__TOSPILL(this) = to_spill;
  *SpillLeaveParam__SPILLLOC(this) = spillLoc;
  return this;
}

#line 496 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_doSpill(BasicBlock block, Operation after, void *param) {
  SpillLeaveParam spillLeaveParam = (SpillLeaveParam)param;
  Temporary to_spill = SpillLeaveParam_TOSPILL(spillLeaveParam);
  SpillLoc spillLoc = SpillLeaveParam_SPILLLOC(spillLeaveParam);
  // identity of the next created operation
  int32_t identity = IndexedFactory_COUNTER(Operation_Factory);
  Operation addedOperation;
  DefUses defUses = Temporary_defUses(to_spill);
  Spill_appendSpill(block, after, to_spill, to_spill);
  //FIXME: What location should we set on remat variable?
  *SpillLoc__LOC(spillLoc) = Spill_genStackedLocation(to_spill);
  //FIXME: Collect the list of all inserted operations
  if(after == NULL) {
    *SpillLoc__SPILL(spillLoc) = BasicBlock_firstOperation(block);
  } else {
    *SpillLoc__SPILL(spillLoc) = Operation_next(after);
  }
  // Def uses chains
  addedOperation = SpillLoc_SPILL(spillLoc);
  while(addedOperation != NULL
        && Operation_identity(addedOperation) >= identity) {
    int32_t argIdx = 0;
    Operation_ARGUMENTS_FOREACH_Temporary(addedOperation, arg) {
      if(arg == to_spill) {
        DefUses_addUse(defUses, addedOperation, argIdx);
      }
      ++argIdx;
    } Operation_ARGUMENTS_ENDEACH_Temporary;
    addedOperation = Operation_next(addedOperation);
  }
}

#line 529 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_spillTemporary_(SpillEnv this, SpillInfo spillInfo) {
  Temporary to_spill = SpillInfo_VAR(spillInfo);

#line 533 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  // determine_spill_costs must have been run before
  Except_CHECK(SpillInfo_SPILLCOSTS(spillInfo) >= 0);

#line 536 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  // some backends have virtual noreg/unknown nodes that are not scheduled
  // and simply always available. 

#line 539 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, LEVEL_1, "spilling %p ... ", to_spill));
  IDList_FOREACH(SpillInfo_SPILLS(spillInfo), SpillLoc_, spillLoc) {
    Operation after = SpillLoc_operationAfter(spillLoc);
    BasicBlock block  = SpillPoint_BLOCK(SpillLoc_POINT(spillLoc));
    SpillLeaveParam_ spillLeaveParam[1];
    SpillLeaveParam_Ctor(spillLeaveParam, to_spill, spillLoc);
    if(after != NULL && after == BasicBlock_lastOperation(block)) {
      BasicBlock_splitLeaveEdgeAndAction(block, true, (void*)spillLeaveParam,
                                         SpillEnv_doSpill);
    } else {
      SpillEnv_doSpill(block, after, (void*)spillLeaveParam);
    }
   

#line 553 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    DBG((dbg, LEVEL_1, "\t%p after %p\n", SpillLoc_SPILL(spillLoc), after));
  } IDList_ENDEACH;
  DBG((dbg, LEVEL_1, "\n"));
}

#line 558 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_spillTemporary(SpillEnv this, Temporary temporary) {
  SpillInfo spillInfo = SpillEnv_getSpillInfo(this, temporary);

#line 562 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  // node is already spilled
  if(!IDList_isEmpty(SpillInfo_SPILLS(spillInfo)) &&
     SpillLoc_SPILL(((SpillLoc)IDList_firstItem(SpillInfo_SPILLS(spillInfo))))
     != NULL) {
    return;
  }

#line 569 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  {
    Operation producer = Temporary_getProducer(temporary);
    if(isPhi(producer) && TemporarySet_contains(SpillEnv_MEMPHIS(this),
                                                temporary)) {
      SpillEnv_spillPhi_(this, spillInfo);
    } else {
      SpillEnv_spillTemporary_(this, spillInfo);
    }
  }
}

#line 580 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * analyzes how to best spill a node and determine costs for that
 */
static void
SpillEnv_determineSpillCosts(SpillEnv env, Temporary toSpill) {
  SpillInfo spillInfo = SpillEnv_getSpillInfo(env, toSpill);
  Operation producer = Temporary_getProducer(toSpill);
  BasicBlock spillBlock = Operation_basicBlock(producer);
  SpillLoc_ spillLoc[1];
  double spillExecFreq = BasicBlock_frequency(spillBlock);
  double spillCost = Spill_getSpillCost(toSpill);

#line 592 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* already calculated? */
  if(SpillInfo_SPILLCOSTS(spillInfo) >= 0) {
    return;
  }

#line 597 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  //assert(!be_is_Reload(to_spill));

#line 599 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if (isPhi(producer) && TemporarySet_contains(SpillEnv_MEMPHIS(env), toSpill)) {
    *SpillInfo__SPILLCOSTS(spillInfo) = spillCost * spillExecFreq;
    return;
  }

#line 604 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if(!IDList_isEmpty(SpillInfo_SPILLS(spillInfo))) {
    double  spillsExecFreq;

#line 607 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* calculate sum of executaion frequencies of individual spills */
    spillsExecFreq = 0;
    IDList_FOREACH(SpillInfo_SPILLS(spillInfo), SpillLoc_, spillLoc) {
      spillBlock = SpillPoint_BLOCK(SpillLoc_POINT(spillLoc));
      spillsExecFreq += BasicBlock_frequency(spillBlock);
    } IDList_ENDEACH;

#line 614 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    DBG((dbg, LEVEL_1, "%p: latespillcosts %f after def: %f\n", toSpill,
         spillsExecFreq * spillCost, spillExecFreq * spillCost));

#line 617 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* multi-/latespill is advantageous -> return*/
    if(spillsExecFreq < spillExecFreq) {
      DBG((dbg, LEVEL_1, "use latespills for %p\n", toSpill));
      *SpillInfo__SPILLCOSTS(spillInfo) = spillsExecFreq * spillCost;
      return;
    }
  }

#line 625 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  IDList_empty(SpillInfo_SPILLS(spillInfo), NULL);
  /* override spillinfos or create a new SpillLoc */
  SpillLoc_Ctor(spillLoc, spillBlock, producer);
  IDList_PUSH(SpillInfo_SPILLS(spillInfo), SpillLoc_, *spillLoc);
  *SpillInfo__SPILLCOSTS(spillInfo) = spillExecFreq * spillCost;
  DBG((dbg, LEVEL_1, "spill %p after definition\n", toSpill));
}

#line 633 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct ReloadParam_ {
  //@args SpillEnv spillEnv, Temporary toSpill
  SpillEnv SPILLENV;
  Temporary TOSPILL;
};
typedef struct ReloadParam_ ReloadParam_, *ReloadParam;
typedef const struct ReloadParam_ *const_ReloadParam;
typedef struct ReloadParam_ * restrict_ReloadParam;

extern ReloadParam
ReloadParam_Ctor(ReloadParam this, SpillEnv spillEnv, Temporary toSpill);

extern ReloadParam
ReloadParam_Copy(ReloadParam this, const_ReloadParam that);

extern void
ReloadParam_Dtor(ReloadParam this);

extern size_t
ReloadParam_Size(SpillEnv spillEnv, Temporary toSpill);

#define ReloadParam_SPILLENV(this) ((this)->SPILLENV)
#define ReloadParam__SPILLENV(this) (&(this)->SPILLENV)
#define ReloadParam_TOSPILL(this) ((this)->TOSPILL)
#define ReloadParam__TOSPILL(this) (&(this)->TOSPILL)


#line 639 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
ReloadParam
ReloadParam_Ctor(ReloadParam this, SpillEnv spillEnv, Temporary toSpill) {
  *ReloadParam__SPILLENV(this) = spillEnv;
  *ReloadParam__TOSPILL(this) = toSpill;
  return this;
}

#line 646 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
ReloadParam
ReloadParam_Copy(ReloadParam this, const_ReloadParam that) {
  return this;
}

#line 651 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
ReloadParam_Dtor(ReloadParam this) {
}

#line 655 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_actionReload(BasicBlock block, Operation pointInsertionAfter,
                      void* param) {
  ReloadParam reloadParam = (ReloadParam)param;
  SpillEnv_addReload_(ReloadParam_SPILLENV(reloadParam),
                      ReloadParam_TOSPILL(reloadParam), block,
                      pointInsertionAfter, true);
}

#line 664 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
SpillEnv_insertSpillsReloads(SpillEnv this) {
  SSAForm ssaForm = SpillEnv_SSAFORM(this);
  CodeRegion codeRegion = SSAForm_codeRegion(ssaForm);
  int32_t i;
  *SpillEnv__LASTIDX(this) = IndexedFactory_COUNTER(Operation_Factory);
  TemporarySet_FOREACH_Temporary(SpillEnv_MEMPHIS(this), temporary) {
    SpillEnv_spillTemporary(this, temporary);
  } TemporarySet_ENDEACH_Temporary;

#line 674 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if(SpillEnv_VARSPILLLOCS(this) == NULL) {
    // Nothing to do
    return;
  }

#line 679 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* process each spilled node */
  for(i = 0; i < SpillEnv_COUNTER(this); ++i) {
    SpillInfo spillInfo = SpillEnv_VARSPILLLOCS(this)[i];
    Temporary toSpill;
    OperationResultList_ copies[1];
    double    all_remat_costs = 0; /** costs when we would remat all nodes */
    int       force_remat     = 0;

#line 687 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    if(spillInfo == NULL) {
      continue;
    }
    
    toSpill = SpillInfo_VAR(spillInfo);

#line 693 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    OperationResultList_Ctor(copies, SpillEnv_MEMORY(this));
    DBG((dbg, LEVEL_1, "\nhandling all reloaders of %p:\n", toSpill));

#line 696 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    SpillEnv_determineSpillCosts(this, toSpill);

#line 698 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    // Spill interface automatically handle rematerialization and is not
    // controlable on that point. So, this part is currently useless. 
#if 0
    /* determine possibility of rematerialisations */
    if(be_do_remats) {
      /* calculate cost savings for each indivial value when it would
         be rematted instead of reloaded */
      IDList_FOREACH(SpillInfo_RELOADS(spillInfo), SpillLoc_, rld) {
        double   freq;
        int      remat_cost;
        int      remat_cost_delta;
        ir_node *block;
        ir_node *reloader = rld->reloader;

#line 712 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        if(rld->rematted_node != NULL) {
          DBG((dbg, LEVEL_2, "\tforced remat %+F before %+F\n",
               rld->rematted_node, reloader));
          continue;
        }
        if(rld->remat_cost_delta >= REMAT_COST_INFINITE) {
          DBG((dbg, LEVEL_2, "\treload before %+F is forbidden\n",
               reloader));
          all_remat_costs = REMAT_COST_INFINITE;
          continue;
        }

#line 724 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        remat_cost  = check_remat_conditions_costs(this, to_spill,
                                                   reloader, 0);
        if(remat_cost >= REMAT_COST_INFINITE) {
          DBG((dbg, LEVEL_2, "\tremat before %+F not possible\n",
               reloader));
          rld->remat_cost_delta = REMAT_COST_INFINITE;
          all_remat_costs       = REMAT_COST_INFINITE;
          continue;
        }

#line 734 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        remat_cost_delta      = remat_cost - this->reload_cost;
        rld->remat_cost_delta = remat_cost_delta;
        block                 = is_Block(reloader) ? reloader : get_nodes_block(reloader);
        freq                  = get_block_execfreq(exec_freq, block);
        all_remat_costs      += remat_cost_delta * freq;
        DBG((dbg, LEVEL_2, "\tremat costs delta before %+F: "
             "%d (rel %f)\n", reloader, remat_cost_delta,
             remat_cost_delta * freq));
      } IDList_ENDEACH;
      if(all_remat_costs < REMAT_COST_INFINITE) {
        /* we don't need the costs for the spill if we can remat
           all reloaders */
        all_remat_costs -= si->spill_costs;

#line 748 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        DBG((dbg, LEVEL_2, "\tspill costs %d (rel %f)\n",
             this->spill_cost, si->spill_costs));
      }

#line 752 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      if(all_remat_costs < 0) {
        DBG((dbg, LEVEL_1, "\nforcing remats of all reloaders (%f)\n",
             all_remat_costs));
        force_remat = 1;
      }
    }
#endif // if 0
    /* go through all reloads for this spill */
    IDList_FOREACH (SpillInfo_RELOADS(spillInfo), SpillLoc_, rld) {
      OperationResult_ copyAndIdx[1]; /* a reload is a "copy" of the original value */
      Operation copy = NULL;
      // Same remark as previous 'if 0'
#if 0
      if (rld->rematted_node != NULL) {
        copy = rld->rematted_node;
        sched_add_before(rld->reloader, copy);
      } else if (be_do_remats &&
                 (force_remat || rld->remat_cost_delta < 0)) {
        copy = do_remat(this, to_spill, rld->reloader);
      } else {
#endif

#line 774 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        Operation before = SpillLoc_operationBefore(rld);
        BasicBlock block  = SpillPoint_BLOCK(SpillLoc_POINT(rld));
        Variable variable = Temporary_VARIABLE(toSpill);
        Temporary dst = CodeRegion_makeVariableTemporary(codeRegion, Variable_ORIGIN(variable));
        bool unsafe = false;
        Except_CHECK(Temporary_isVariable(toSpill));
        /* make sure we have a spill */
        SpillEnv_spillTemporary(this, toSpill);

#line 783 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        /* create a reload, use the first spill for now SSA
         * reconstruction for memory comes below */
        Except_CHECK(!IDList_isEmpty(SpillInfo_SPILLS(spillInfo)));

#line 787 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        if(before == NULL) {
          Operation spillPoint = Spill_findLastInsertPoint(block, before);
          while(spillPoint != NULL && !unsafe) {
            Operation_ARGUMENTS_FOREACH_Temporary(spillPoint, temp) {
              if(Temporary_regFile(temp) == Temporary_regFile(toSpill)) {
                unsafe = true;
                break;
              }
            } Operation_ARGUMENTS_ENDEACH_Temporary;
            Operation_RESULTS_FOREACH_Temporary(spillPoint, temp) {
              if(Temporary_regFile(temp) == Temporary_regFile(toSpill)) {
                unsafe = true;
                break;
              }
            } Operation_RESULTS_ENDEACH_Temporary;
            if(Operation_hasClobber(spillPoint)) {
              RegisterSet_FOREACH_Register(Operation_clobberSet(spillPoint),
                                           reg) {
                if(Register_regFile(reg) == Temporary_regFile(toSpill)) {
                  unsafe = true;
                  break;
                }
              } RegisterSet_ENDEACH_Register;
            }
            spillPoint = Operation_next(spillPoint);
          }
        }

#line 815 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        //TODO: Should return the list of newly inserted operation (at least
        // the bounds (first and last operation))
        if(before == NULL && unsafe) {
          // There is a great probability that toSpill will not be reloaded at
          // the very end of the basic block. (control operation or call)
          // So, to avoid problems when increasing register pressure by move up
          // the reload operation, we insert in each succ with only one
          // predecessor or create an intermediate block
          ReloadParam_ reloadParam[1];
          ReloadParam_Ctor(reloadParam, this, toSpill);
          BasicBlock_splitLeaveEdgeAndAction(block, true, (void*)reloadParam,
                                             SpillEnv_actionReload);
          SSAForm_killDominance(SpillEnv_SSAFORM(this));
          // go to next
          IDList_FOREACH_SYNC;
          IDList_remove(SpillInfo_RELOADS(spillInfo), rld, NULL);
          continue;
        } else {
          Spill_insertReload(block, before, dst, toSpill);
        }
        //FIXME: Collect the list of all inserted operations
        if(before == NULL) {
          before = BasicBlock_lastOperation(block);
        } else {
          before = Operation_prev(before);
        }
        do {
          copy = before;
        } while((Operation_resCount(copy) < 1
                 || Operation_getResult(copy, 0) != dst)
                && (before = Operation_prev(before)));
        Except_CHECK(copy != NULL && Operation_getResult(copy, 0) == dst );
        DefUses_setDef(Temporary_defUses(dst),
                       copy, 0);
        *SpillLoc__LOC(rld) = dst;
        *SpillLoc__SPILL(rld) = copy;
#if 0
      }
#endif
      // FIXME: Is it always the right index?
      OperationResult_Ctor(copyAndIdx, copy, 0);
      DBG((dbg, LEVEL_1, " %p of %p before %p\n", copy, toSpill, rld));
      OperationResultList_PUSH(copies, *copyAndIdx);
      OperationResult_Dtor(copyAndIdx);
    } IDList_ENDEACH;

#line 861 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* if we had any reloads or remats, then we need to reconstruct the
     * SSA form for the spilled value */
    if (!OperationResultList_isEmpty(copies)) {
      SSAReconstruct_ ssaReconst[1];

#line 866 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      SSAReconstruct_Ctor(ssaReconst, SpillEnv_SSAFORM(this), toSpill,
                          SpillEnv_MEMORY(this));
      SSAReconstruct_addCopy(ssaReconst, DefUses_getDef(Temporary_defUses(toSpill)));
      SSAReconstruct_addCopies(ssaReconst, copies);
      SSAReconstruct_fixUsers(ssaReconst, toSpill);
      SSAReconstruct_Dtor(ssaReconst);
    }
    
    /* need to reconstruct SSA form if we had multiple spills */
    if (!IDList_isSingle(SpillInfo_SPILLS(spillInfo))) {
      SSAReconstruct_ ssaReconst[1];
      int spill_count = 0;

#line 879 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      SSAReconstruct_Ctor(ssaReconst, SpillEnv_SSAFORM(this), toSpill,
                          SpillEnv_MEMORY(this));
      IDList_FOREACH(SpillInfo_SPILLS(spillInfo), SpillLoc_, spillLoc) {
        /* maybe we rematerialized the value and need no spill */
        if(SpillLoc_SPILL(spillLoc) == NULL) {
          continue;
        }
#if 0
        SSAReconstruct_addCopy(ssaReconst, SpillLoc_SPILL(spillLoc));
#endif
        spill_count++;
      } IDList_ENDEACH;
      if(spill_count > 1) {
        /* all reloads are attached to the first spill, fix them now */
        Except_NEVER(false && "Support for multiple spill should not be "
                     "requiered. We do not support phi with memory operands,"
                     "thus, all spills should be at the same spill slot");
#if 0
        SSAReconstruct_fixUsers(ssaReconst, si->spills->spill);
#endif
      }
      SSAReconstruct_Dtor(ssaReconst);
    }

#line 903 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    OperationResultList_Dtor(copies);
  }
#if 0
        /* Matze: In theory be_ssa_construction should take care of the liveness...
         * try to disable this again in the future */
        be_liveness_invalidate(env->birg->lv);
        be_remove_dead_nodes_from_schedule(env->birg);
#endif
}

#line 913 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct WorkSet_ {
  //@args BeladyEnv env
  Memory MEMORY;
  int32_t LEN;			/**< current length */
  Loc_ VARS_[0];		/**< inlined array of the variables/distances in this working set */
};

extern WorkSet
WorkSet_Ctor(WorkSet this, BeladyEnv env);

extern WorkSet
WorkSet_Copy(WorkSet this, const_WorkSet that);

extern void
WorkSet_Dtor(WorkSet this);

extern size_t
WorkSet_Size(BeladyEnv env);

#define WorkSet_MEMORY(this) ((this)->MEMORY)
#define WorkSet__MEMORY(this) (&(this)->MEMORY)
#define WorkSet_LEN(this) ((this)->LEN)
#define WorkSet__LEN(this) (&(this)->LEN)
#define WorkSet_VARS_(this) ((this)->VARS_)
#define WorkSet__VARS_(this) (&(this)->VARS_)


#line 920 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
WorkSet
WorkSet_Ctor(WorkSet this,
             BeladyEnv env) {
  bzero(this, sizeof(*this));
  *WorkSet__MEMORY(this) = BeladyEnv_MEMORY(env);
  return this;
}

#line 928 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define WorkSet_getLoc(workSet, i) (&(WorkSet_VARS_(workSet)[i]))
#define WorkSet_setLoc(workSet, i, loc) ((WorkSet_VARS_(workSet)[i])) = *(loc)
#define WorkSet_setTime(workSet, i, t) (*Loc__TIME(WorkSet_getLoc(workSet, i)))=(t)
#define WorkSet_getTime(workSet, i) Loc_TIME(WorkSet_getLoc(workSet, i))
#define WorkSet_setVersion(workSet, i, t) *Loc__VERSION(WorkSet_getLoc(workSet, i)) = (t)
#define WorkSet_getVersion(workSet, i) Loc_VERSION(WorkSet_getLoc(workSet, i))
#define WorkSet_setLength(workSet, length) *WorkSet__LEN(workSet) = (length)
#define WorkSet_getLength(workSet) WorkSet_LEN(workSet)
#define WorkSet_getVar(workSet, i) Loc_VAR(WorkSet_getLoc(workSet, i))
#define WorkSet_setVar(workSet, i, tempo) *Loc__VAR(WorkSet_getLoc(workSet, i)) = tempo
#define WorkSet_sort(workSet) qsort(WorkSet_VARS_(workSet), WorkSet_getLength(workSet), sizeof(WorkSet_VARS_(workSet)[0]), Loc_compare);
#define WorkSet_contains(workSet, n) (WorkSet_getIndex(workSet, n) >= 0)

#line 942 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline void
WorkSet_print(const_WorkSet w) {
  int i;
  for(i = 0; i < WorkSet_getLength(w); ++i) {
    fprintf(stderr, "%p %d\n", WorkSet_getVar(w, i), WorkSet_getTime(w, i));
  }
}

#line 950 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Alloc a new workset on obstack @p ob with maximum size @p max
 */
WorkSet
WorkSet_make(BeladyEnv env) {
  WorkSet res;
  size_t size = sizeof(*res) + BeladyEnv_NREGS(env) * sizeof(*WorkSet_getLoc(res, 0));
  res = Memory_alloc(BeladyEnv_MEMORY(env), size);
  WorkSet_Ctor(res, env);
  return res;
}

#line 962 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
WorkSet_kill(WorkSet this) {
  if (this != NULL) {
#ifndef ECL_NDTOR
    Memory_free_(WorkSet_MEMORY(this), this);
#endif//ECL_NDTOR
  }
}

#line 971 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Do NOT alloc anything. Make @param tgt equal to @param src.
 * returns @param tgt for convenience
 */
static inline WorkSet
WorkSet_copy(BeladyEnv env, WorkSet tgt, WorkSet src) {
  size_t size = sizeof(*src) + BeladyEnv_NREGS(env) * sizeof(*WorkSet_getLoc(src, 0));
  memcpy(tgt, src, size);
  return tgt;
}

#line 982 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Alloc a new instance on obstack and make it equal to @param ws
 */
static inline WorkSet
WorkSet_clone(BeladyEnv env, WorkSet workSet) {
  WorkSet res = WorkSet_make(env);
  WorkSet_copy(env, res, workSet);
  return res;
}

#line 992 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Inserts the value @p val into the workset, iff it is not
 * already contained. The workset must not be full.
 */
static inline void
WorkSet_insert(BeladyEnv env, WorkSet workSet, Temporary var) {
  int32_t i;
  /* check for current regclass */
  //FIXME: does it exist an equivalent test in LAO?
  /* 	if (!arch_irn_consider_in_reg_alloc(env->arch, env->cls, val)) { */
  /* 		// DBG((dbg, DBG_WORKSET, "Skipped %p\n", val)); */
     /* 		return; */
     /* 	} */

#line 1006 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
     /* check if val is already contained */
     for(i=0; i<WorkSet_getLength(workSet); ++i)
     if (WorkSet_getVar(workSet, i) == var)
       return;

#line 1011 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* insert val */
  Except_CHECK(WorkSet_getLength(workSet) < BeladyEnv_NREGS(env) && "WorkSet already full!");
  WorkSet_setVar(workSet, WorkSet_getLength(workSet), var);
  (*WorkSet__LEN(workSet))++;
}

#line 1017 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Removes all entries from this workset
 */
static inline void WorkSet_clear(WorkSet workSet) {
  WorkSet_setLength(workSet, 0);
}

#line 1024 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Removes the value @p val from the workset if present.
 */
static inline void
WorkSet_remove(WorkSet workSet, Temporary var) {
  int32_t i;
  for(i=0; i<WorkSet_getLength(workSet); ++i) {
    if (WorkSet_getVar(workSet, i) == var) {
      WorkSet_setLoc(workSet, i, WorkSet_getLoc(workSet,
                                                --(*WorkSet__LEN(workSet))));
      return;
    }
  }
}

#line 1039 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline int32_t
WorkSet_getIndex(const_WorkSet workSet, Temporary var) {
  int32_t i;
  for(i=0; i<WorkSet_getLength(workSet); ++i) {
    if (WorkSet_getVar(workSet, i) == var)
      return i;
  }

#line 1047 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  return -1;
}

#line 1050 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Iterates over all values in the working set.
 * @p workSet The workset to iterate
 * @p v  A variable to put the current value in
 * @p i  An integer for internal use
 */
#define WorkSet_FOREACH(workSet, v, i)	{ \
    int32_t i; \
    Temporary v; \
    for(i=0; v=(i < WorkSet_getLength(workSet)) ? WorkSet_getVar(workSet, i) : NULL, \
          i < WorkSet_getLength(workSet); ++i) {
#define WorkSet_ENDEACH \
  } \
}

#line 1065 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct ListHead_ {
  struct ListHead_ *next, *prev;
};
typedef struct ListHead_ ListHead_, *ListHead;
typedef const struct ListHead_ *const_ListHead;
typedef struct ListHead_ * restrict_ListHead;

#define ListHead_next(this) ((this)->next)
#define ListHead__next(this) (&(this)->next)
#define ListHead_prev(this) ((this)->prev)
#define ListHead__prev(this) (&(this)->prev)


#line 1069 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define INIT_LIST_HEAD(ptr) do { \
  (ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while (0)

#line 1073 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/*
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline void
__list_add(ListHead new_node, ListHead prev, ListHead next) {
  next->prev = new_node;
  new_node->next = next;
  new_node->prev = prev;
  prev->next = new_node;
}

#line 1087 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * list_add_tail - add a new entry
 * @param new_node   new entry to be added
 * @param head       list head to add it before
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 */
static inline void
list_add_tail(ListHead new_node, ListHead head) {
  __list_add(new_node, head->prev, head);
}

#line 1100 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * list_for_each_entry	-	iterate over list of given type
 * @param type    the type of the struct where the listhead is embedded in
 * @param pos     the type * to use as a loop counter.
 * @param head    the head for your list.
 * @param member  the name of the list_struct within the struct.
 */
#define list_for_each_entry(type, pos, head, member)    \
  for (pos = list_entry((head)->next, type, member);        \
       &pos->member != (head);                              \
       pos = list_entry(pos->member.next, type, member))

#line 1112 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * list_for_each_entry_reverse - iterate backwards over list of given type.
 * @param type    the type of the struct where the listhead is embedded in
 * @param pos     the type * to use as a loop counter.
 * @param head    the head for your list.
 * @param member  the name of the list_struct within the struct.
 */
#define list_for_each_entry_reverse(type, pos, head, member)    \
  for (pos = list_entry((head)->prev, type, member);            \
       &pos->member != (head);                                  \
       pos = list_entry(pos->member.prev, type, member))

#line 1124 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * list_entry - get the struct for this entry
 * @param ptr     the &struct list_head pointer.
 * @param type    the type of the struct this is embedded in.
 * @param member  the name of the list_struct within the struct.
 */
#define list_entry(ptr, type, member) \
  _list_container_of(ptr, type, member)

#line 1133 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define _list_offsetof(type,member) \
  ((char *) &(((type *) 0)->member) - (char *) 0)

#line 1136 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define _list_container_of(ptr, type, member)                   \
  ((type *) ((char *) (ptr) - _list_offsetof(type, member)))

#line 1140 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static int
Loc_compare(const void *a, const void *b) {
  const_Loc p = a;
  const_Loc q = b;
  return (Loc_TIME(p) > Loc_TIME(q)) - (Loc_TIME(p) < Loc_TIME(q));
}

#line 1147 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BlockInfo_;
typedef struct BlockInfo_ BlockInfo_, *BlockInfo;
typedef const struct BlockInfo_ *const_BlockInfo;
typedef struct BlockInfo_ * restrict_BlockInfo;

#line 1149 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
BasicBlock_setBlockInfo(BasicBlock basicBlock, BlockInfo info);

#line 1152 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BlockInfo_ {
  //@args BeladyEnv bel, BasicBlock bb
  BeladyEnv BELADYENV;
  //@access MEMORY BeladyEnv_MEMORY(BlockInfo_BELADYENV(this))
  BasicBlock BASICBLOCK;
  //@access FREQUENCY BasicBlock_frequency(BlockInfo_BASICBLOCK(this))
  // Take the index as identifier, since we will not insert or remove bbs
  //@access ID BasicBlock_index(BlockInfo_BASICBLOCK(this))
  // List of uses in that basic block
  IDList *NEXTUSES;
  // The end set after the local belady pass.
  WorkSet WORKSETEND;
  // Cost of a reload in this block.
  double RELOADCOST;
  //FIXME: temporary reload cost before better integration:
  //i.e. move Spill.xcc file here
  // First node in block which is not a phi.
  Operation FIRSTNONIN;
  // The instruction before which end of block reloads will be inserted.
  Operation LASTINS;
  // The amount of registers which remain free in this block. This capacity can
  // be used to let global variables, transported into other blocks, live
  // through this block.
  int32_t PRESSURE;
  // The pressure right before the first real (non-phi) node. At the beginning
  // of the global pass, this is 0.
  int32_t FRONTPRESSURE;
  // List head for all bring_in variables.
  ListHead_ BRHEAD_;
  //@access BRHEAD BlockInfo__BRHEAD_(this)
  // registers free at jump.
  int32_t FREEATJUMP;
  int32_t POSTNUM;
};

extern BlockInfo
BlockInfo_Ctor(BlockInfo this, BeladyEnv bel, BasicBlock bb);

extern BlockInfo
BlockInfo_Copy(BlockInfo this, const_BlockInfo that);

extern void
BlockInfo_Dtor(BlockInfo this);

extern size_t
BlockInfo_Size(BeladyEnv bel, BasicBlock bb);

#define BlockInfo_BELADYENV(this) ((this)->BELADYENV)
#define BlockInfo__BELADYENV(this) (&(this)->BELADYENV)
#define BlockInfo_BASICBLOCK(this) ((this)->BASICBLOCK)
#define BlockInfo__BASICBLOCK(this) (&(this)->BASICBLOCK)
#define BlockInfo_NEXTUSES(this) ((this)->NEXTUSES)
#define BlockInfo__NEXTUSES(this) (&(this)->NEXTUSES)
#define BlockInfo_WORKSETEND(this) ((this)->WORKSETEND)
#define BlockInfo__WORKSETEND(this) (&(this)->WORKSETEND)
#define BlockInfo_RELOADCOST(this) ((this)->RELOADCOST)
#define BlockInfo__RELOADCOST(this) (&(this)->RELOADCOST)
#define BlockInfo_FIRSTNONIN(this) ((this)->FIRSTNONIN)
#define BlockInfo__FIRSTNONIN(this) (&(this)->FIRSTNONIN)
#define BlockInfo_LASTINS(this) ((this)->LASTINS)
#define BlockInfo__LASTINS(this) (&(this)->LASTINS)
#define BlockInfo_PRESSURE(this) ((this)->PRESSURE)
#define BlockInfo__PRESSURE(this) (&(this)->PRESSURE)
#define BlockInfo_FRONTPRESSURE(this) ((this)->FRONTPRESSURE)
#define BlockInfo__FRONTPRESSURE(this) (&(this)->FRONTPRESSURE)
#define BlockInfo_BRHEAD_(this) ((this)->BRHEAD_)
#define BlockInfo__BRHEAD_(this) (&(this)->BRHEAD_)
#define BlockInfo_FREEATJUMP(this) ((this)->FREEATJUMP)
#define BlockInfo__FREEATJUMP(this) (&(this)->FREEATJUMP)
#define BlockInfo_POSTNUM(this) ((this)->POSTNUM)
#define BlockInfo__POSTNUM(this) (&(this)->POSTNUM)
#define BlockInfo_BRHEAD(this) (BlockInfo__BRHEAD_(this))
#define BlockInfo_ID(this) (BasicBlock_index(BlockInfo_BASICBLOCK(this)))
#define BlockInfo_FREQUENCY(this) (BasicBlock_frequency(BlockInfo_BASICBLOCK(this)))
#define BlockInfo_MEMORY(this) (BeladyEnv_MEMORY(BlockInfo_BELADYENV(this)))


#line 1187 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
BlockInfo
BlockInfo_Ctor(BlockInfo this, BeladyEnv bel, BasicBlock bb)
{
  *BlockInfo__BELADYENV(this) = bel;
  *BlockInfo__BASICBLOCK(this) = bb;
  *BlockInfo__NEXTUSES(this) = NULL;
  *BlockInfo__WORKSETEND(this) = NULL;
  *BlockInfo__RELOADCOST(this)  = 4.0 * BlockInfo_FREQUENCY(this);
  *BlockInfo__FIRSTNONIN(this) = NULL;
  *BlockInfo__LASTINS(this) = NULL;
  *BlockInfo__PRESSURE(this) = 0;
  *BlockInfo__FRONTPRESSURE(this) = 0;
  INIT_LIST_HEAD(BlockInfo_BRHEAD(this));
  *BlockInfo__FREEATJUMP(this) = BeladyEnv_NREGS(bel);
  *BlockInfo__POSTNUM(this) = -1;
  BasicBlock_setBlockInfo(bb, this);
  return this;
}

#line 1206 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BlockInfo *g_blockInfos = NULL;
static BasicBlock *g_dfsPostOrder = NULL;

#line 1209 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
Belady_init(BeladyEnv env) {
  size_t size = sizeof(BlockInfo) * BeladyEnv_BLOCKCOUNT(env);
  g_blockInfos = Memory_alloc(BeladyEnv_MEMORY(env), size);
  bzero(g_blockInfos, size);

#line 1215 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  size = sizeof(int32_t) * BeladyEnv_BLOCKCOUNT(env);
  g_dfsPostOrder = Memory_alloc(BeladyEnv_MEMORY(env), size);
  bzero(g_dfsPostOrder, size);
  Spill_initCodeRegion(BeladyEnv_CODEREGION(env));
}

#line 1221 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
Belady_fini(BeladyEnv env) {
#ifndef ECL_NDTOR
  Memory_free_(BeladyEnv_MEMORY(env), g_blockInfos);
  Memory_free_(BeladyEnv_MEMORY(env), g_dfsPostOrder);
#endif//ECL_NDTOR
  g_blockInfos = NULL;
  g_dfsPostOrder = NULL;
  resetMark(env);
  Spill_finiCodeRegion(BeladyEnv_CODEREGION(env));
}

#line 1233 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
BasicBlock_setBlockInfo(BasicBlock basicBlock, BlockInfo info) {
  Except_CHECK(g_blockInfos != NULL && "Block info mapper not initialized");
  g_blockInfos[BasicBlock_index(basicBlock)] = info;
}

#line 1239 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BlockInfo
BasicBlock_getBlockInfo(const_BasicBlock basicBlock) {
  BlockInfo res = NULL;
  if(g_blockInfos != NULL) {
    res = g_blockInfos[BasicBlock_index(basicBlock)];
    Except_CHECK(res != NULL && "Want a not set information");
  }
  return res;
}

#line 1249 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BlockInfo
BlockInfo_make(BeladyEnv bel, BasicBlock bb) {
  BlockInfo this = Memory_alloc(BeladyEnv_MEMORY(bel), sizeof(BlockInfo_));
  BlockInfo_Ctor(this, bel, bb);
  return this;
}

#line 1256 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline Operation
BlockInfo_getLastIns(BlockInfo bi) {
  if (BlockInfo_LASTINS(bi) == NULL) {
    *BlockInfo__LASTINS(bi) =
      BasicBlock_lastOperation(BlockInfo_BASICBLOCK(bi));
  }

#line 1263 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  return BlockInfo_LASTINS(bi);
}

#line 1266 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct NextUse_ {
  bool ISFIRSTUSE; /**< Indicate that this use is the first
                      in the block. Needed to identify
                      transport in values for the global
                      pass. */
  uint32_t STEP;
  Operation OPERATION;
  // Must be an element of a IDList
  //@access NEXT IDListItemCell_NEXT(IDListItem_CELL(this))
};
typedef struct NextUse_ NextUse_, *NextUse;
typedef const struct NextUse_ *const_NextUse;
typedef struct NextUse_ * restrict_NextUse;

#define NextUse_ISFIRSTUSE(this) ((this)->ISFIRSTUSE)
#define NextUse__ISFIRSTUSE(this) (&(this)->ISFIRSTUSE)
#define NextUse_STEP(this) ((this)->STEP)
#define NextUse__STEP(this) (&(this)->STEP)
#define NextUse_OPERATION(this) ((this)->OPERATION)
#define NextUse__OPERATION(this) (&(this)->OPERATION)
#define NextUse_NEXT(this) (IDListItemCell_NEXT(IDListItem_CELL(this)))


#line 1277 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline IDList
BlockInfo_getNextUsesForTempo(BlockInfo bi, Temporary var) {
  IDList res;
  Except_CHECK(BlockInfo_NEXTUSES(bi) != NULL);
  res = BlockInfo_NEXTUSES(bi)[Temporary_identity(var)];
  if(res == NULL) {
    res = Memory_alloc(BlockInfo_MEMORY(bi), sizeof(*res));
    IDList_Ctor(res, BlockInfo_MEMORY(bi));
    BlockInfo_NEXTUSES(bi)[Temporary_identity(var)] = res;
  }
  return res;
}

#line 1290 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline NextUse
BlockInfo_getNextUseForTempo(BlockInfo bi, Temporary var) {
  IDList nextUses = BlockInfo_getNextUsesForTempo(bi, var);
  NextUse res = IDList_isEmpty(nextUses)? NULL: (NextUse)IDList_firstItem(nextUses);
  return res;
}

#line 1298 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static bool
isMostProbablePredecessor(BasicBlock curBB, BasicBlock potentialSucc) {
  float weight = 0.0;
  bool result = false;
  BasicBlock_ENTER_FOREACH_BasicBlockEdge(potentialSucc, edge) {
    if(weight < BasicBlockEdge_probability(edge)) {
      weight = BasicBlockEdge_probability(edge);
      if(BasicBlockEdge_origBlock(edge) == curBB) {
        result = true;
      } else {
        result = false;
      }
    }
  } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
  return result;
}

#line 1315 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
buildNextUses(BlockInfo bi) {
  BeladyEnv env = BlockInfo_BELADYENV(bi);
  size_t size = sizeof(IDList) * IndexedFactory_COUNTER(Temporary_Factory);
  IDList_ continusBasicBlocks[1];
  BasicBlock curBB = BlockInfo_BASICBLOCK(bi);
  uint32_t nbContinusOps = 0;
  *BlockInfo__NEXTUSES(bi) = Memory_alloc(BlockInfo_MEMORY(bi), size);
  bzero(BlockInfo_NEXTUSES(bi), size);
  IDList_Ctor(continusBasicBlocks, BlockInfo_MEMORY(bi));

#line 1326 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  do {
    IDList_PUSH(continusBasicBlocks, BasicBlock, curBB);
    *BasicBlock__MARK(curBB) = 1;
    nbContinusOps += Selector_countOperations(BasicBlock_selector(curBB));
    if(BasicBlock_leaveCount(curBB) == 1) {
      curBB = BasicBlockEdge_destBlock(BasicBlock_leaveFirst(curBB));
    } else if(BasicBlock_next(curBB)
              && isMostProbablePredecessor(curBB, BasicBlock_next(curBB))) {
      curBB = BasicBlock_next(curBB);
    } else {
      curBB = NULL;
    }
  } while(curBB != NULL && BasicBlock_MARK(curBB) == 0
          && !BasicBlock_isExit(curBB));
  
  IDList_FORBACK(continusBasicBlocks, BasicBlock, pBB) {
    BasicBlock bb = *pBB;
    // remove mark
    *BasicBlock__MARK(bb) = 0;
    // maybe block has not been proceed yet, thus ordering is wrong
    BasicBlock_makeOrderings(bb, 0, 1);
    nbContinusOps -= Selector_countOperations(BasicBlock_selector(bb));
    BasicBlock_FORBACK_Operation(bb, operation) {
      int i;
      if (isPhi(operation)) {
        break;
      }

#line 1354 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      Operation_ARGUMENTS_FOREACH_Temporary(operation, tempo) {
        IDList nextUses;
        NextUse_ use[1];
        if(Temporary_regFile(tempo) != BeladyEnv_REGFILE(env)) {
          continue;
        }
        nextUses = BlockInfo_getNextUsesForTempo(bi, tempo);

#line 1362 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        *NextUse__ISFIRSTUSE(use) = 1;
        *NextUse__OPERATION(use) = operation;
        *NextUse__STEP(use) = Operation_ordering(operation) + nbContinusOps;
      
        if (!IDList_isEmpty(nextUses)) {
          NextUse oldFirst = (NextUse)IDList_firstItem(nextUses);
          *NextUse__ISFIRSTUSE(oldFirst) = 0;
          Except_CHECK(NextUse_STEP(oldFirst) >= NextUse_STEP(use));
        }
      
        IDList_PUT(nextUses, NextUse_, *use);
      } Operation_ARGUMENTS_ENDEACH_Temporary;
    } BasicBlock_ENDBACK_Operation;
  } IDList_ENDBACK;
  Except_CHECK(nbContinusOps == 0);
}

#line 1379 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
killNextUses(BlockInfo this) {
#ifndef ECL_NDTOR
  int32_t index;
  for(index = 0; index < IndexedFactory_COUNTER(Temporary_Factory);
      ++index) {
    IDList list = BlockInfo_NEXTUSES(this)[index];
    if(list != NULL) {
      IDList_Dtor(list);
      Memory_free_(BlockInfo_MEMORY(this), list);
    }
  }
  Memory_free_(BlockInfo_MEMORY(this), BlockInfo_NEXTUSES(this));
#endif
}

#line 1395 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline void
advanceCurrentUse(BlockInfo bi, Temporary var) {
  IDList nextUses = BlockInfo_getNextUsesForTempo(bi, var);
  Except_CHECK(nextUses != NULL && !IDList_isEmpty(nextUses));
  IDList_drop(nextUses, NULL);
}

#line 1402 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/*
   ____         _                 ___
  | __ )  _ __ (_) _ __    __ _  |_ _|_ __
  |  _ \ | '__|| || '_ \  / _` |  | || '_ \
  | |_) || |   | || | | || (_| |  | || | | |
  |____/ |_|   |_||_| |_| \__, | |___|_| |_|
                           |___/

#line 1410 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  Data structures to represent bring in variables.
*/

#line 1413 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BringIn_ {
  //@args BlockInfo blockInfo, Temporary var, Operation use
  // node to bring in.
  Temporary TEMPORARY;
  // The block to which bring in should happen.
  BlockInfo BLOCKINFO;
  // The maximal pressure till the first use of irn in bl.
  int32_t PRESSURESOFAR;
  // The first user of TEMPORARY in BLOCKINFO.
  Operation FIRSTUSE;
  // Schedule step of the first use.
  //@access USESTEP Operation_ordering(BringIn_FIRSTUSE(this))
  //@access ISREMAT Temporary_isRemater(BringIn_TEMPORARY(this))
  // Offset to maximum pressure in block.
  int32_t SECTPRESSURE;
  float GLOBALCOST;
  ListHead_ LIST_;
  //@access LIST BringIn__LIST_(this)
};
typedef struct BringIn_ BringIn_, *BringIn;
typedef const struct BringIn_ *const_BringIn;
typedef struct BringIn_ * restrict_BringIn;

extern BringIn
BringIn_Ctor(BringIn this, BlockInfo blockInfo, Temporary var, Operation use);

extern BringIn
BringIn_Copy(BringIn this, const_BringIn that);

extern void
BringIn_Dtor(BringIn this);

extern size_t
BringIn_Size(BlockInfo blockInfo, Temporary var, Operation use);

#define BringIn_TEMPORARY(this) ((this)->TEMPORARY)
#define BringIn__TEMPORARY(this) (&(this)->TEMPORARY)
#define BringIn_BLOCKINFO(this) ((this)->BLOCKINFO)
#define BringIn__BLOCKINFO(this) (&(this)->BLOCKINFO)
#define BringIn_PRESSURESOFAR(this) ((this)->PRESSURESOFAR)
#define BringIn__PRESSURESOFAR(this) (&(this)->PRESSURESOFAR)
#define BringIn_FIRSTUSE(this) ((this)->FIRSTUSE)
#define BringIn__FIRSTUSE(this) (&(this)->FIRSTUSE)
#define BringIn_SECTPRESSURE(this) ((this)->SECTPRESSURE)
#define BringIn__SECTPRESSURE(this) (&(this)->SECTPRESSURE)
#define BringIn_GLOBALCOST(this) ((this)->GLOBALCOST)
#define BringIn__GLOBALCOST(this) (&(this)->GLOBALCOST)
#define BringIn_LIST_(this) ((this)->LIST_)
#define BringIn__LIST_(this) (&(this)->LIST_)
#define BringIn_ISREMAT(this) (Temporary_isRemater(BringIn_TEMPORARY(this)))
#define BringIn_LIST(this) (BringIn__LIST_(this))
#define BringIn_USESTEP(this) (Operation_ordering(BringIn_FIRSTUSE(this)))


#line 1433 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BasicBlock
BringIn_getTemporaryBasicBlock(const_BringIn bringIn);

#line 1436 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BasicBlock
Temporary_getDefBlock(Temporary tempo, BasicBlock usageBlock);

#line 1439 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static float
Temporary_getGlobalSpillCost(Temporary var, BasicBlock usageBlock);

#line 1442 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
BringIn
BringIn_Ctor(BringIn this,
             BlockInfo blockInfo, Temporary var, Operation use)
{
  *BringIn__TEMPORARY(this) = var;
  *BringIn__BLOCKINFO(this) = blockInfo;
  *BringIn__PRESSURESOFAR(this) = BlockInfo_PRESSURE(blockInfo);
  *BringIn__FIRSTUSE(this) = use;
  *BringIn__SECTPRESSURE(this) = BlockInfo_FRONTPRESSURE(blockInfo);
  *BringIn__GLOBALCOST(this) = 0.0;
  INIT_LIST_HEAD(BringIn_LIST(this));
  list_add_tail(BringIn_LIST(this), BlockInfo_BRHEAD(blockInfo));
  if(getenv("SPILL_EVERYWHERE_ORACLE") != NULL) {
    *BringIn__GLOBALCOST(this) =
      Temporary_getGlobalSpillCost(var, Operation_basicBlock(use));
  } else {
    *BringIn__GLOBALCOST(this) =
      g_originalAlgo?
      BlockInfo_FREQUENCY(blockInfo)
      : Spill_getReloadCost(var) * BlockInfo_FREQUENCY(blockInfo)
      + Spill_getSpillCost(var) * BasicBlock_frequency(BringIn_getTemporaryBasicBlock(this));
  }
  return this;
}

#line 1467 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BasicBlock
Temporary_getDefBlock(Temporary tempo, BasicBlock usageBlock) {
  BasicBlock result;
  if(Temporary_isDedicated(tempo)) {
    // trivial algorithm, no check
    result = usageBlock;
    if(BasicBlock_enterCount(result) == 1) {
      // FIXME: maybe it comes from upper!
      result = BasicBlockEdge_origBlock(BasicBlock_enterFirst(result));
    }
  } else {
    result =  Operation_basicBlock(Temporary_getProducer(tempo));
  }
  return result;
}

#line 1483 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
// Get the basic block which contains the defintion
// of the bring in temporary
static BasicBlock
BringIn_getTemporaryBasicBlock(const_BringIn bringIn) {
  return Temporary_getDefBlock(BringIn_TEMPORARY(bringIn),
                               Operation_basicBlock(BringIn_FIRSTUSE(bringIn)));
}

#line 1491 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BringIn
BringIn_make(BlockInfo bi, Temporary val, Operation use) {
  BringIn this;
  Except_CHECK((Temporary_isDedicated(val) || Temporary_isVariable(val))
               && "Must be a temporary variable since we need def uses information");
  this = Memory_alloc(BeladyEnv_MEMORY(BlockInfo_BELADYENV(bi)), sizeof(BringIn_));
  BringIn_Ctor(this, bi, val, use);
  return this;
}

#line 1501 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static float
Temporary_getGlobalSpillCost(Temporary var, BasicBlock usageBlock) {
  float cost = 0.0;
  BasicBlock defBlock = Temporary_getDefBlock(var, usageBlock);
  // Cost of the spill of the definition
  cost = Spill_getSpillCost(var)* BasicBlock_frequency(defBlock);
  if(Temporary_isDedicated(var)) {
    cost += Spill_getReloadCost(var) * BasicBlock_frequency(usageBlock);
  } else {
    OperationArgumentList_FOREACH_OperationArgument(DefUses_getUses(Temporary_defUses(var)),
                                                    opArg) {
      float weight = 0.0;
      Operation use = OperationArgument_OPERATION(opArg);
      if(Operator_isPhi(Operation_operator(use))) {
        Operation_ARGUMENTS_FOREACH_Temporary(use, phiArg) {
          BasicBlockEdge edge = BasicBlock_getPhiEnterEdge(Operation_basicBlock(use),
                                                           Operation_ARGUMENTS_INDEX);
          if (phiArg == var) {
            weight = BasicBlock_frequency(BasicBlockEdge_origBlock(edge));
            break;
          }
        } Operation_ARGUMENTS_ENDEACH_Temporary;
      } else {
        weight = BasicBlock_frequency(Operation_basicBlock(use));
      }
      cost += Spill_getReloadCost(var) * weight;
    } OperationArgumentList_ENDEACH_OperationArgument;
  }
  return cost;
}

#line 1532 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
// Modify this function to keep the transitive property of the cmp when we are
// not in SEB_ORIG mode.
static int
BringIn_cmp(const_IDListItem a, const_IDListItem b) {
  const_BringIn p = *(const_BringIn*) a;
  const_BringIn q = *(const_BringIn*) b;
  double fp, fq;

#line 1540 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* if one of both is a remat node, it will be done after the other. */
  if (BringIn_ISREMAT(p) != BringIn_ISREMAT(q))
    return BringIn_ISREMAT(p) - BringIn_ISREMAT(q);

#line 1544 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* in the same block, the one further in the front has to be processed first!
   * Otherwise the front_pressure 'trick' is not exact. */
  if (BringIn_BLOCKINFO(p) == BringIn_BLOCKINFO(q)  && g_originalAlgo)
    return BringIn_USESTEP(p) - BringIn_USESTEP(q);

#line 1549 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  fp = BringIn_GLOBALCOST(p);
  fq = BringIn_GLOBALCOST(q);

#line 1552 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* if both have the same frequency, inspect the frequency of the definition */
  if (fp == fq && g_originalAlgo) {
    double fdp = BasicBlock_frequency(BringIn_getTemporaryBasicBlock(p));
    double fdq = BasicBlock_frequency(BringIn_getTemporaryBasicBlock(q));

#line 1557 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* if the defs of both have the same freq, we go for reverse dfs post order.*/
    if (fdp == fdq) {
      int pp = BlockInfo_POSTNUM(BringIn_BLOCKINFO(p));
      int pq = BlockInfo_POSTNUM(BringIn_BLOCKINFO(q));
      return pq - pp;
    }
    return (fdq > fdp) - (fdq < fdp);
  }
  return (fq > fp) - (fq < fp);
}

#line 1568 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline uint32_t
getCurrDistance(BlockInfo bi, Temporary var, bool is_usage)
{
  BeladyEnv env = BlockInfo_BELADYENV(bi);
  int32_t curr_step = Operation_ordering(BeladyEnv_INSTR(env));
  NextUse use = BlockInfo_getNextUseForTempo(bi, var);
  
/* 	int flags                  = arch_irn_get_flags(env->arch, irn); */

#line 1577 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/* 	assert(!(flags & arch_irn_flags_ignore)); */

#line 1579 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        /* We have to keep nonspillable nodes in the workingset */
  if(Temporary_isDedicated(var)) {
    return 0;
  }

#line 1584 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if (!is_usage && use && NextUse_STEP(use) == curr_step) {
    use = NextUse_NEXT(use);
  }

#line 1588 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if (use) {
    uint32_t res  = NextUse_STEP(use) - curr_step;

#line 1591 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    Except_CHECK(NextUse_STEP(use) >= curr_step);

#line 1593 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    if (res != 0) {
      if (remat_live_range_ext && Temporary_isRemater(var))
        res = REMAT_DIST;
      else if (BitSet_containsTempo(BeladyEnv_SPILLED(env), var))
        res *= already_spilled_factor;
    }
    return res;
  }
  {
    uint32_t liveEnd = LIVE_END;
    if(getenv("TEST_HEURISTIC") != NULL) {
      liveEnd -=
        (uint32_t)Temporary_getGlobalSpillCost(var, Operation_basicBlock(BeladyEnv_INSTR(env)));
    }
    return BasicBlock_isLiveOut(BlockInfo_BASICBLOCK(bi), var) ? liveEnd: DEAD;
  }
}

#line 1611 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline bool
isLocalPhi(const_BasicBlock bl, const_Operation op) {
  return op != NULL && isPhi(op) && Operation_basicBlock(op) == bl;
}

#line 1616 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static Operation
findTrivialBackwardDefinition(Operation startPoint, const_BasicBlock bl, Temporary tempo) {
  Operation point = startPoint != NULL? Operation_prev(startPoint): BasicBlock_lastOperation(bl);
  Operation result = NULL;
  while(point != NULL && result == NULL) {
    Operation_RESULTS_FOREACH_Temporary(point, tmpRes) {
      if(tmpRes == tempo) {
        result = point;
        break;
      }
    } Operation_RESULTS_ENDEACH_Temporary;
    point = Operation_prev(point);
  }
  if(result == NULL && BasicBlock_enterCount(bl) == 1) {
    BasicBlock prev = BasicBlockEdge_origBlock(BasicBlock_enterFirst(bl));
    if(prev != CodeRegion_rootBlock(BasicBlock_codeRegion(prev))) {
      result = findTrivialBackwardDefinition(NULL, prev, tempo);
    }
  }
  return result;
}

#line 1638 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Check, if the value is something that is transported into a block.
 * That is, the value is defined elsewhere or defined by a Phi in the block.
 * @param env  The belady environment.
 * @param bl   The block in question.
 * @param irn  The node in question.
 * @return     1, if node is something transported into @p bl, 0 if not.
 * @note       The function will only give correct answers in the case
 *             where @p irn is unsed in the block @p bl which is always
 *             the case in our usage scenario.
 */
static inline bool
isTransportIn(BeladyEnv env, const_BasicBlock bl, Temporary temporary)
{
  bool result = true;
  if(!Temporary_isVariable(temporary)) {
    Operation producer = findTrivialBackwardDefinition(BeladyEnv_INSTR(env), bl, temporary);
    Except_CHECK(Temporary_hasRegister(temporary));
    result = producer == NULL || Operation_basicBlock(producer) != bl;
  } else {
    Operation producer;
    Except_CHECK(Temporary_isVariable(temporary));
    producer = Temporary_getProducer(temporary);
    result = (Operation_basicBlock(producer) != bl) || isPhi(producer);
  }
  return result;
}

#line 1666 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Performs the actions necessary to grant the request that:
 * - new_vals can be held in registers
 * - as few as possible other values are disposed
 * - the worst values get disposed
 *
 * @p is_usage indicates that the values in new_vals are used (not defined)
 * In this case reloads must be performed
 */
static void
displace(BlockInfo bi, TemporarySet new_vals, bool is_usage, int32_t clobberSize) {
  BeladyEnv env = BlockInfo_BELADYENV(bi);
  WorkSet workSet = BeladyEnv_WORKSET(env);
  Temporary  *to_insert = alloca(BeladyEnv_NREGS(env) * sizeof(to_insert[0]));

#line 1681 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  int32_t i, len, max_allowed, demand;

#line 1683 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
    1. Identify the number of needed slots and the values to reload
  */
  demand = 0;
  TemporarySet_FOREACH_Temporary(new_vals, val) {
    /* mark value as used */

#line 1690 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    if (! WorkSet_contains(workSet, val)) {
      DBG((dbg, DBG_DECIDE, "\t\tinsert %p\n", val));
      to_insert[demand++] = val;
      if (is_usage) {
        NextUse use = BlockInfo_getNextUseForTempo(bi, val);

#line 1696 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        /*
         * if we use a value which is transported in this block, i.e. a
         * phi defined here or a live in, for the first time, we check
         * if there is room for that guy to survive from the block's
         * entrance to here or not.
         */
        Except_CHECK(use != NULL);
        Except_CHECK(Operation_ordering(BeladyEnv_INSTR(env)) == NextUse_STEP(use));
        if (isTransportIn(env, BlockInfo_BASICBLOCK(bi), val)
            && NextUse_ISFIRSTUSE(use)) {
          BringIn bri = BringIn_make(bi, val, NextUse_OPERATION(use));
          
          /* reset the section pressure, since a new section starts. */
          *BlockInfo__FRONTPRESSURE(bi) = 0;

#line 1711 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
          DBG((dbg, DBG_DECIDE, "\t\tbring in node %p, pressure %d:\n", val, bi->pressure));
          DBG((dbg, DBG_DECIDE, "\t\tno reload. must be considered at block start\n"));
        } else {
          BitSet_addTempo(BeladyEnv_SPILLED(env), val);
          DBG((dbg, DBG_SPILL, "\t\tReload %p before %p\n", val, env->instr));
          SpillEnv_addReload(BeladyEnv_SPILLENV(env), val, BeladyEnv_INSTR(env));
        }
      }
    } else {
      assert((is_usage || Temporary_isDedicated(val)) && "Defined value already in workset?!?");
      DBG((dbg, DBG_DECIDE, "\t\tskip %p\n", val));
    }
  } TemporarySet_ENDEACH_Temporary;
  DBG((dbg, DBG_DECIDE, "\t\tdemand = %d\n", demand));

#line 1726 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
    2. Make room for at least 'demand' slots
  */
  len = WorkSet_getLength(workSet);
  max_allowed = BeladyEnv_NREGS(env) - demand - clobberSize;

#line 1732 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* Only make more free room if we do not have enough */
  if (len > max_allowed) {
    DBG((dbg, DBG_DECIDE, "\t\tdisposing %d values\n", len - max_allowed));

#line 1736 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* get current next-use distance */
    for (i = 0; i < WorkSet_getLength(workSet); ++i) {
      Temporary var  = WorkSet_getVar(workSet, i);
      uint32_t dist = getCurrDistance(bi, var, is_usage);
      WorkSet_setTime(workSet, i, dist);
    }

#line 1743 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* sort entries by increasing nextuse-distance*/
    WorkSet_sort(workSet);
    
    Except_CHECK(max_allowed >= 0 && "We need more registers than possible");
    /* kill the last 'demand' entries in the array */
    WorkSet_setLength(workSet, max_allowed);
  }

#line 1751 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
    3. Insert the new values into the workset
    Also, we update the pressure in the block info.
    That is important for the global pass to decide
    how many values can live through the block.
  */
  for (i = 0; i < demand; ++i) {
    WorkSet_insert(env, BeladyEnv_WORKSET(env), to_insert[i]);
  }

#line 1761 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* TODO: simplify this expression? */
  *BlockInfo__PRESSURE(bi) = MAX(BlockInfo_PRESSURE(bi),
                                 WorkSet_getLength(BeladyEnv_WORKSET(env))
                                 + clobberSize);
  *BlockInfo__FRONTPRESSURE(bi) = MAX(BlockInfo_FRONTPRESSURE(bi),
                                      WorkSet_getLength(BeladyEnv_WORKSET(env))
                                      + clobberSize);
}

#line 1770 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static bool
Temporary_isAllocatableDedicated(const_Temporary tempo, Convention convention) {
  bool result = false;
  if(Temporary_hasRegister(tempo)) {
    RegisterSet_ allocatableSet[1];
    *allocatableSet = Convention_allocatableSet(convention, Temporary_regFile(tempo));
    result = RegisterSet_contains(allocatableSet, Temporary_register(tempo));
  }
  return result;
}

#line 1781 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * For the given block @p block, decide for each values
 * whether it is used from a register or is reloaded
 * before the use.
 */
static void
Belady(BeladyEnv env, BasicBlock basicBlock) {
  Convention convention = BeladyEnv_CONVENTION(env);
  BlockInfo block_info = BlockInfo_make(env, basicBlock);
  BasicBlock block = BlockInfo_BASICBLOCK(block_info);

#line 1792 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  TemporarySet_ new_vals[1];

#line 1794 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  TemporarySet_Ctor(new_vals, BeladyEnv_MEMORY(env), BeladyEnv_NREGS(env));

#line 1796 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, DBG_WSETS, "Belady on %p\n", block));
  WorkSet_clear(BeladyEnv_WORKSET(env));

#line 1799 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* build the next use information for this block. */
  buildNextUses(block_info);

#line 1802 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* process the block from start to end */
  BasicBlock_FOREACH_Operation(block, op) {
    int32_t i, arity;
    RegisterSet_ clobberSet[1];
    RegisterSet_Ctor(clobberSet);
    Except_CHECK(WorkSet_getLength(BeladyEnv_WORKSET(env)) <= BeladyEnv_NREGS(env) && "Too much values in workset!");

#line 1809 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /*
     * Phis are no real instr (see insert_starters())
     * instr_nr does not increase */
    if (isPhi(op))
      continue;
    DBG((dbg, DBG_DECIDE, "\t%p\n", op));

#line 1816 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    if (BlockInfo_FIRSTNONIN(block_info) == NULL) {
      *BlockInfo__FIRSTNONIN(block_info) = op;
    }

#line 1820 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* set instruction in the workset */
    *BeladyEnv__INSTR(env) = op;

#line 1823 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* allocate all values _used_ by this instruction */
    TemporarySet_empty(new_vals);
    Operation_ARGUMENTS_FOREACH_Temporary(op, var) {
      if(Temporary_regFile(var) != BeladyEnv_REGFILE(env)
         || ((Temporary_hasRegister(var))
             && !Temporary_isAllocatableDedicated(var, convention))
         /*|| !Temporary_isVariable(var)*/) {
        continue;
      }
      TemporarySet_insert(new_vals, var);
    } Operation_ARGUMENTS_ENDEACH_Temporary;
    DBG((dbg, DBG_DECIDE, "\t* uses\n"));
    displace(block_info, new_vals, 1, 0);

#line 1837 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    TemporarySet_empty(new_vals);
    // before killing all dead variables add the list of clobber register to
    // have a good estimation of register that can go through
    if(Operation_hasClobber(op)) {
      int32_t idx = 0;
      WorkSet workSet = BeladyEnv_WORKSET(env);

#line 1844 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      // Kill clobbers that are in workset
      while(idx < WorkSet_getLength(workSet)) {
        Temporary var = WorkSet_getVar(workSet, idx);
        if((Temporary_hasRegister(var))
           && RegisterSet_contains(Operation_clobberSet(op),
                                   Temporary_register(var))) {
          // Here we are sure var is allocatable by construction.
          // Take the last element and put it here.
          // Reduce the length of the set
          WorkSet_setLoc(workSet, idx, WorkSet_getLoc(workSet,
                                                      --(*WorkSet__LEN(workSet))));
        } else {
          ++idx;
        }
      }
      {
        RegisterSet_ allocatableSet[1];
        *allocatableSet = Convention_allocatableSet(convention,
                                                    BeladyEnv_REGFILE(env));
        RegisterSet_FOREACH_Register(Operation_clobberSet(op), reg) {
          if(Register_regFile(reg) == BeladyEnv_REGFILE(env)
             && RegisterSet_contains(allocatableSet, reg)) {
            RegisterSet_insert(clobberSet, reg);
          }
        } RegisterSet_ENDEACH_Register;
      }
      // set usage to true to dispose not the argument of the operation.
      // this must be done because of register constraint that may be on
      // arguments
      displace(block_info, new_vals, true, RegisterSet_count(clobberSet));
    }

#line 1876 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /*
     * set all used variables to the next use in their NextUse list
     * Also, kill all dead variables from the workset. They are only
     * augmenting the pressure. Note, that a variable is dead
     * if it has no further use in this block and is *not* live end
     */
    Operation_ARGUMENTS_FOREACH_Temporary(op, tempo) {
      NextUse use;
      if(Temporary_regFile(tempo) != BeladyEnv_REGFILE(env)) {
        continue;
      }
      use = BlockInfo_getNextUseForTempo(block_info, tempo);

#line 1889 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      Except_CHECK(use);
      if (NextUse_NEXT(use) == NULL && !BasicBlock_isLiveOut(block, tempo)) {
        WorkSet_remove(BeladyEnv_WORKSET(env), tempo);
      }

#line 1894 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      advanceCurrentUse(block_info, tempo);
    } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 1897 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* allocate all values _defined_ by this instruction */
    Operation_RESULTS_FOREACH_Temporary(op, var) {
      if(Temporary_regFile(var) != BeladyEnv_REGFILE(env)
         || ((Temporary_hasRegister(var))
             && !Temporary_isAllocatableDedicated(var, convention))
         /*|| !Temporary_isVariable(var)*/) {
        continue;
      } else if(Temporary_hasRegister(var)) {
        RegisterSet_remove(clobberSet, Temporary_register(var));
      }
      TemporarySet_insert(new_vals, var);
    } Operation_RESULTS_ENDEACH_Temporary;
    DBG((dbg, DBG_DECIDE, "\t* defs\n"));
    // We count registerSet only here to materialize available bringIn slot
    displace(block_info, new_vals, 0, RegisterSet_count(clobberSet));
    RegisterSet_Dtor(clobberSet);
    //FIXME: does such operation exist in LAO?
#if 0
    if ( is_op_forking(get_irn_op(env->instr))) {
      for (i = get_irn_arity(env->instr) - 1; i >= 0; --i) {
        ir_node *op = get_irn_n(env->instr, i);
        //FIXME: does it exist an equivalent test in LAO?
        // block_info->free_at_jump -= arch_irn_consider_in_reg_alloc(env->arch, env->cls, op);
      }
    }
#endif
  } BasicBlock_ENDEACH_Operation;

#line 1925 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* kill the next use information for this block. */
  killNextUses(block_info);

#line 1928 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* Remember end-workset for this block */
  *BlockInfo__WORKSETEND(block_info) = WorkSet_clone(env, BeladyEnv_WORKSET(env));
  DBG((dbg, DBG_WSETS, "End workset for %p:\n", block));
  WorkSet_FOREACH(BlockInfo_WORKSETEND(block_info), var, iter) {
    DBG((dbg, DBG_WSETS, "  %p (%u)\n", var,
         WorkSet_getTime(BlockInfo_WORKSETEND(block_info), iter)));
  } WorkSet_ENDEACH;
  DBG((dbg, DBG_WSETS, "Max pressure in block: %d\n", BlockInfo_PRESSURE(block_info)));

#line 1937 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* now, initialize the front pressure to 0. */
  *BlockInfo__FRONTPRESSURE(block_info) = 0;
}

#line 1941 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/*
 _____ _                  _       _           _   ____            _
|_   _| |__   ___    __ _| | ___ | |__   __ _| | |  _ \ __ _ _ __| |_
  | | | '_ \ / _ \  / _` | |/ _ \| '_ \ / _` | | | |_) / _` | '__| __|
  | | | | | |  __/ | (_| | | (_) | |_) | (_| | | |  __/ (_| | |  | |_
  |_| |_| |_|\___|  \__, |_|\___/|_.__/ \__,_|_| |_|   \__,_|_|   \__|
                    |___/

#line 1949 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
*/

#line 1951 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define ver_oldest                        (0)
#define ver_youngest                      ((uint32_t) -1)
#define ver_make_newer(v)                 ((v) + 1)
#define ver_is_older(v, w)                ((v) < (w))
#define ver_is_younger(v, w)              ((v) > (w))

#line 1957 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
enum {
  tempo_act_none = 0,
  tempo_act_reload,
  tempo_act_live_through
};

#line 1963 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
// Forward declaration
struct TemporaryAction_;
typedef struct TemporaryAction_ TemporaryAction_, *TemporaryAction;
typedef const struct TemporaryAction_ *const_TemporaryAction;
typedef struct TemporaryAction_ * restrict_TemporaryAction;

#line 1966 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BlockState_;
typedef struct BlockState_ BlockState_, *BlockState;
typedef const struct BlockState_ *const_BlockState;
typedef struct BlockState_ * restrict_BlockState;

#line 1968 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct GlobalEndState_;
typedef struct GlobalEndState_ GlobalEndState_, *GlobalEndState;
typedef const struct GlobalEndState_ *const_GlobalEndState;
typedef struct GlobalEndState_ * restrict_GlobalEndState;

#line 1970 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct RollbackInfo_;
typedef struct RollbackInfo_ RollbackInfo_, *RollbackInfo;
typedef const struct RollbackInfo_ *const_RollbackInfo;
typedef struct RollbackInfo_ * restrict_RollbackInfo;

#line 1972 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct GlobalEndState_ {
  //@args BeladyEnv env
  BeladyEnv BELADYENV;
  BitSet_ SUCCPHIS_;
  //@access SUCCPHIS GlobalEndState__SUCCPHIS_(this)
  BitSet_ COMMITTED_;
  //@access COMMITTED GlobalEndState__COMMITTED_(this)
  //@access MEMORY BeladyEnv_MEMORY(GlobalEndState_BELADYENV(this))
  uint32_t VER;
  //@access VERSION GlobalEndState_VER(this)
  //@mutate VERSION GlobalEndState__VER(this)
  uint32_t *BSTOPSVERS;
  BlockState *BSTOPS;
  BlockState BSTOP;
  TemporaryAction IATOP;
};

extern GlobalEndState
GlobalEndState_Ctor(GlobalEndState this, BeladyEnv env);

extern GlobalEndState
GlobalEndState_Copy(GlobalEndState this, const_GlobalEndState that);

extern void
GlobalEndState_Dtor(GlobalEndState this);

extern size_t
GlobalEndState_Size(BeladyEnv env);

#define GlobalEndState_BELADYENV(this) ((this)->BELADYENV)
#define GlobalEndState__BELADYENV(this) (&(this)->BELADYENV)
#define GlobalEndState_SUCCPHIS_(this) ((this)->SUCCPHIS_)
#define GlobalEndState__SUCCPHIS_(this) (&(this)->SUCCPHIS_)
#define GlobalEndState_COMMITTED_(this) ((this)->COMMITTED_)
#define GlobalEndState__COMMITTED_(this) (&(this)->COMMITTED_)
#define GlobalEndState_VER(this) ((this)->VER)
#define GlobalEndState__VER(this) (&(this)->VER)
#define GlobalEndState_BSTOPSVERS(this) ((this)->BSTOPSVERS)
#define GlobalEndState__BSTOPSVERS(this) (&(this)->BSTOPSVERS)
#define GlobalEndState_BSTOPS(this) ((this)->BSTOPS)
#define GlobalEndState__BSTOPS(this) (&(this)->BSTOPS)
#define GlobalEndState_BSTOP(this) ((this)->BSTOP)
#define GlobalEndState__BSTOP(this) (&(this)->BSTOP)
#define GlobalEndState_IATOP(this) ((this)->IATOP)
#define GlobalEndState__IATOP(this) (&(this)->IATOP)
#define GlobalEndState_VERSION(this) (GlobalEndState_VER(this))
#define GlobalEndState_MEMORY(this) (BeladyEnv_MEMORY(GlobalEndState_BELADYENV(this)))
#define GlobalEndState_SUCCPHIS(this) (GlobalEndState__SUCCPHIS_(this))
#define GlobalEndState_COMMITTED(this) (GlobalEndState__COMMITTED_(this))
#define GlobalEndState__VERSION(this) (GlobalEndState__VER(this))


#line 1989 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
GlobalEndState
GlobalEndState_Ctor(GlobalEndState this,
                    BeladyEnv env)
{
  *GlobalEndState__BELADYENV(this) = env;
  BitSet_Ctor(GlobalEndState_SUCCPHIS(this), GlobalEndState_MEMORY(this),
              IndexedFactory_COUNTER(Operation_Factory));
  BitSet_Ctor(GlobalEndState_COMMITTED(this), GlobalEndState_MEMORY(this),
              BeladyEnv_BLOCKCOUNT(env));
  *GlobalEndState__VERSION(this) = ver_make_newer(ver_oldest);
  *GlobalEndState__BSTOPSVERS(this) =
        Memory_alloc(GlobalEndState_MEMORY(this),
                     sizeof(uint32_t)*BeladyEnv_BLOCKCOUNT(env));
  *GlobalEndState__BSTOPS(this) =
        Memory_alloc(GlobalEndState_MEMORY(this),
                     sizeof(BlockState)*BeladyEnv_BLOCKCOUNT(env));
  *GlobalEndState__BSTOP(this) = NULL;
  *GlobalEndState__IATOP(this) = NULL;
  return this;
}

#line 2010 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
GlobalEndState_Dtor(GlobalEndState this)
{
#ifndef ECL_NDTOR
  Memory_free_(GlobalEndState_MEMORY(this),
               GlobalEndState_BSTOPSVERS(this));
#endif
#ifndef ECL_NDTOR
  Memory_free_(GlobalEndState_MEMORY(this),
               GlobalEndState_BSTOPS(this));
#endif
}

#line 2023 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct BlockState_ {
  //@args GlobalEndState ges, BlockInfo bi
  struct BlockState_ *NEXT;
  struct BlockState_ *NEXTINTERN;
  BlockInfo BLOCKINFO;
  int32_t PRESSURE;
  WorkSet ENDSTATE;
  Memory MEMORY;
};

extern BlockState
BlockState_Ctor(BlockState this, GlobalEndState ges, BlockInfo bi);

extern BlockState
BlockState_Copy(BlockState this, const_BlockState that);

extern void
BlockState_Dtor(BlockState this);

extern size_t
BlockState_Size(GlobalEndState ges, BlockInfo bi);

#define BlockState_NEXT(this) ((this)->NEXT)
#define BlockState__NEXT(this) (&(this)->NEXT)
#define BlockState_NEXTINTERN(this) ((this)->NEXTINTERN)
#define BlockState__NEXTINTERN(this) (&(this)->NEXTINTERN)
#define BlockState_BLOCKINFO(this) ((this)->BLOCKINFO)
#define BlockState__BLOCKINFO(this) (&(this)->BLOCKINFO)
#define BlockState_PRESSURE(this) ((this)->PRESSURE)
#define BlockState__PRESSURE(this) (&(this)->PRESSURE)
#define BlockState_ENDSTATE(this) ((this)->ENDSTATE)
#define BlockState__ENDSTATE(this) (&(this)->ENDSTATE)
#define BlockState_MEMORY(this) ((this)->MEMORY)
#define BlockState__MEMORY(this) (&(this)->MEMORY)


#line 2033 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
BlockState
GlobalEndState_getBlockState(GlobalEndState ges, const_BlockInfo bi);
void
GlobalEndState_setBlockState(GlobalEndState ges, const_BlockInfo bi, BlockState bs);
void
GlobalEndState_setBlockVersion(GlobalEndState ges, const_BlockInfo bi, uint32_t version);

#line 2040 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
BlockState
BlockState_Ctor(BlockState this,
                GlobalEndState ges, BlockInfo bi)
{
  BlockState bs = GlobalEndState_getBlockState(ges, bi);
  *BlockState__MEMORY(this) = GlobalEndState_MEMORY(ges);
  *BlockState__NEXT(this) = GlobalEndState_BSTOP(ges);
  *BlockState__NEXTINTERN(this) = bs;
  *BlockState__BLOCKINFO(this) = bi; 
  *GlobalEndState__BSTOP(ges) = this;
  GlobalEndState_setBlockState(ges, bi, this);
  GlobalEndState_setBlockVersion(ges, bi, GlobalEndState_VERSION(ges));
  if (bs) {
    *BlockState__PRESSURE(this) = BlockState_PRESSURE(bs);
    *BlockState__ENDSTATE(this) = WorkSet_clone(GlobalEndState_BELADYENV(ges),
                                                BlockState_ENDSTATE(bs));
  } else {
    *BlockState__PRESSURE(this) = BlockInfo_PRESSURE(bi);
    *BlockState__ENDSTATE(this) = WorkSet_clone(GlobalEndState_BELADYENV(ges),
                                                BlockInfo_WORKSETEND(bi));
  }
  return this;
}

#line 2064 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct TemporaryAction_ {
  //@args GlobalEndState ges, Temporary var, const BasicBlock bl
  struct TemporaryAction_ *NEXT;
  Temporary VAR;
  BasicBlock BASICBLOCK;
  int32_t ACT;
  Memory MEMORY;
};

extern TemporaryAction
TemporaryAction_Ctor(TemporaryAction this, GlobalEndState ges, Temporary var, const BasicBlock bl);

extern TemporaryAction
TemporaryAction_Copy(TemporaryAction this, const_TemporaryAction that);

extern void
TemporaryAction_Dtor(TemporaryAction this);

extern size_t
TemporaryAction_Size(GlobalEndState ges, Temporary var, const BasicBlock bl);

#define TemporaryAction_NEXT(this) ((this)->NEXT)
#define TemporaryAction__NEXT(this) (&(this)->NEXT)
#define TemporaryAction_VAR(this) ((this)->VAR)
#define TemporaryAction__VAR(this) (&(this)->VAR)
#define TemporaryAction_BASICBLOCK(this) ((this)->BASICBLOCK)
#define TemporaryAction__BASICBLOCK(this) (&(this)->BASICBLOCK)
#define TemporaryAction_ACT(this) ((this)->ACT)
#define TemporaryAction__ACT(this) (&(this)->ACT)
#define TemporaryAction_MEMORY(this) ((this)->MEMORY)
#define TemporaryAction__MEMORY(this) (&(this)->MEMORY)


#line 2073 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
TemporaryAction
TemporaryAction_Ctor(TemporaryAction this,
                     GlobalEndState ges, Temporary var, const BasicBlock bl)
{
  *TemporaryAction__NEXT(this) = GlobalEndState_IATOP(ges);
  *GlobalEndState__IATOP(ges) = this;
  *TemporaryAction__VAR(this)  = var;
  *TemporaryAction__BASICBLOCK(this) = bl;
  *TemporaryAction__ACT(this) = tempo_act_none;
  *TemporaryAction__MEMORY(this) = GlobalEndState_MEMORY(ges);
  return this;
}

#line 2086 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
struct RollbackInfo_ {
  //@args GlobalEndState ges
  BlockState BSTOP;
  TemporaryAction IATOP;
};

extern RollbackInfo
RollbackInfo_Ctor(RollbackInfo this, GlobalEndState ges);

extern RollbackInfo
RollbackInfo_Copy(RollbackInfo this, const_RollbackInfo that);

extern void
RollbackInfo_Dtor(RollbackInfo this);

extern size_t
RollbackInfo_Size(GlobalEndState ges);

#define RollbackInfo_BSTOP(this) ((this)->BSTOP)
#define RollbackInfo__BSTOP(this) (&(this)->BSTOP)
#define RollbackInfo_IATOP(this) ((this)->IATOP)
#define RollbackInfo__IATOP(this) (&(this)->IATOP)


#line 2092 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
RollbackInfo
RollbackInfo_Ctor(RollbackInfo this,
                  GlobalEndState ges)
{
  *RollbackInfo__BSTOP(this) = GlobalEndState_BSTOP(ges);
  *RollbackInfo__IATOP(this) = GlobalEndState_IATOP(ges);
  return this;
}

#line 2102 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#define GlobalEndState_getBSTopsVers(ges, blockInfo) \
  GlobalEndState_BSTOPSVERS(ges)[BlockInfo_ID(blockInfo)]
#define GlobalEndState_getBSTops(ges, blockInfo) \
  GlobalEndState_BSTOPS(ges)[BlockInfo_ID(blockInfo)]

#line 2107 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
GlobalEndState_setBlockState(GlobalEndState ges, const_BlockInfo bi, BlockState bs) {
  GlobalEndState_getBSTops(ges, bi) = bs;
}

#line 2112 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
GlobalEndState_setBlockVersion(GlobalEndState ges, const_BlockInfo bi, uint32_t version) {
  GlobalEndState_getBSTopsVers(ges, bi) = version;
}

#line 2117 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
BlockState
GlobalEndState_getBlockState(GlobalEndState ges, const_BlockInfo bi) {
  Except_CHECK(!ver_is_younger(GlobalEndState_getBSTopsVers(ges, bi),
                               GlobalEndState_VERSION(ges)));
  return ver_is_older(GlobalEndState_getBSTopsVers(ges, bi),
                      GlobalEndState_VERSION(ges))? NULL:
    GlobalEndState_getBSTops(ges, bi);
}

#line 2126 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline const_WorkSet
GlobalEndState_getEndState(GlobalEndState ges, BlockInfo bi) {
  BlockState bs = GlobalEndState_getBlockState(ges, bi);
  return bs? BlockState_ENDSTATE(bs): BlockInfo_WORKSETEND(bi);
}

#line 2132 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static BlockState
BlockState_make(GlobalEndState ges, BlockInfo bi) {
  BlockState nw = Memory_alloc(GlobalEndState_MEMORY(ges), sizeof(*nw));
  BlockState_Ctor(nw, ges, bi);
  return nw;
}

#line 2139 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
BlockState_kill(BlockState this) {
  if (this != NULL) {
#ifndef ECL_NDTOR
    Memory_free_(BlockState_MEMORY(this), this);
#endif//ECL_NDTOR
  }
}

#line 2148 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static TemporaryAction
TemporaryAction_make(GlobalEndState ges, Temporary var, BasicBlock bl) {
  TemporaryAction ia = Memory_alloc(GlobalEndState_MEMORY(ges), sizeof(*ia));
  TemporaryAction_Ctor(ia, ges, var, bl);
  return ia;
}

#line 2155 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
TemporaryAction_kill(TemporaryAction this) {
  if (this != NULL) {
#ifndef ECL_NDTOR
    Memory_free_(TemporaryAction_MEMORY(this), this);
#endif//ECL_NDTOR
  }
}

#line 2164 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline RollbackInfo_
transBegin(GlobalEndState ges) {
  RollbackInfo_ rb[1];
  RollbackInfo_Ctor(rb, ges);
  return *rb;
}

#line 2171 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static inline void
transRollback(GlobalEndState ges, RollbackInfo rb) {
  BlockState bs = NULL;

#line 2175 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* unwind all the stacks indiced with the block number */
  for(bs = GlobalEndState_BSTOP(ges); bs != RollbackInfo_BSTOP(rb);
      bs = BlockState_NEXT(bs)) {
    GlobalEndState_setBlockState(ges, BlockState_BLOCKINFO(bs),
                                 BlockState_NEXTINTERN(bs));
  }

#line 2182 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  *GlobalEndState__IATOP(ges) = RollbackInfo_IATOP(rb);
  *GlobalEndState__BSTOP(ges) = RollbackInfo_BSTOP(rb);
}

#line 2187 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static double
canBringIn(GlobalEndState ges, BasicBlock bl, Temporary var, double limit,
           int32_t level);

#line 2191 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static double
canMakeAvailableAtEnd(GlobalEndState ges, BasicBlock bl, Temporary var,
                      double limit, int32_t level) {
  BlockInfo bi = BasicBlock_getBlockInfo(bl);
  const_WorkSet end = GlobalEndState_getEndState(ges, bi);
  double res;
  int32_t index;

#line 2199 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, DBG_GLOBAL, "\t%2Dcan make avail %p at end of %p\n", level, var, bl));

#line 2201 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * to make the value available at end,
   * we have several cases here.
   *
   * - we already visited that block.
   * - If the value is in the final end set, return 0.
   *   somebody else already allocated it there.
   * - If not and the final end set is already full,
   *   we cannot make the value available at the end
   *   of this block. return INFINITY.
   * - Else (value not in final end set and there is room):
   *   1) The value is in a register at the end of the local Belady pass.
   *      Allocate a slot in  the final end set and return 0.
   *   2) The value is not in the Belady end set:
   *      If the block's capacity is < k then check what it costs
   *      to transport the value from upper blocks to this block.
   *      Compare that against the reload cost in this block. If
   *      cheaper, do the other thing. If not, reload it here.
   */

#line 2221 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* if the end set contains it already, it is in a reg and it costs nothing
   * to load it to one. */
  index = WorkSet_getIndex(end, var);
  if (index >= 0) {
    uint32_t ver = WorkSet_getVersion(end, index);
    DBG((dbg, DBG_GLOBAL, "\t%2Dnode is in the end set and is %s fixed\n",
         level, ver_is_older(ver, GlobalEndState_VERSION(ges)) ? "already" : "not yet"));

#line 2229 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /*
     * if the version is older, the value is already fixed
     * and cannot be removed from the end set.
     *
     * If not, we will create a new block state for that block since
     * we modify it by giving the end state a new version.
     */
    if (ver_is_younger(ver, GlobalEndState_VERSION(ges))) {
      BlockState bs = BlockState_make(ges, bi);
      WorkSet_setVersion(BlockState_ENDSTATE(bs), index, GlobalEndState_VERSION(ges));
    }

#line 2241 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    res = 0.0;
    goto end;
  }

#line 2245 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * Now we have two options:
   * 1) Reload the value at the end of the block.
   *    Therefore, perhaps, we have to erase another one from the workset.
   *    This may only be done if it has not been fixed.
   *    Since fixed means that a previous pass has decided that that value
   *    *has* to stay in the end set.
   * 2) we can try, if the capacity of the block allows it, to let
   *    the value live through the block and make it available at
   *    the entrance.
   *
   * First, we test the local (reload in this block) alternative
   * and compare against the other alternative.
   * Of course, we chose the cheaper one.
   */

#line 2261 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  {
    int32_t n_regs = BlockInfo_FREEATJUMP(bi);
    int32_t len  = WorkSet_getLength(end);
    int32_t slot = -1;
    int32_t i;

#line 2267 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    res = limit + 1;

#line 2269 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /*
     * look if there is room in the end array
     * for the variable. Note that this does not
     * mean that the var can live through the block.
     * There is just room at the *end*
     */
    if (len < n_regs) {
      DBG((dbg, DBG_GLOBAL, "\t%2Dthe end set has %d free slots\n", level, n_regs - len));
      slot = len;
    } else {
      for (i = 0; i < len; ++i) {
        uint32_t ver = WorkSet_getVersion(end, i);
        if (ver_is_younger(ver, GlobalEndState_VERSION(ges)))
          break;
      }

#line 2285 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      if (i < len) {
        DBG((dbg, DBG_GLOBAL, "\t%2D%p (slot %d) can be erased from the end set\n",
             level, WorkSet_getVar(end, i), i));
        slot = i;
      }
    }

#line 2292 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /*
     * finally there is some room. we can at least reload the value.
     * but we will try to let ot live through anyhow.
     */
    if (slot >= 0) {
      TemporaryAction vs = TemporaryAction_make(ges, var, BlockInfo_BASICBLOCK(bi));
      BlockState bs = BlockState_make(ges, bi);
      WorkSet end = BlockState_ENDSTATE(bs);
      Operation ins_before = BlockInfo_getLastIns(bi);
      double reload_here = Spill_getReloadCost(var) * BasicBlock_frequency(BlockInfo_BASICBLOCK(bi));
      bool pressure_ok = BlockState_PRESSURE(bs) < n_regs;
      // Originaly avoid to keep a room for remat variable
/*       if (reload_here < BlockInfo_RELOADCOST(bi)) */
/*         reload_here = 0.0; */

#line 2307 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      /*
       * No matter what we do, the value will be in the end set
       * if the block from now on (of course only regarding the
       * current state). Enter it and set the new length
       * appropriately.
       */
      WorkSet_setVar(end, slot, var);
      WorkSet_setVersion(end, slot, GlobalEndState_VERSION(ges));
      WorkSet_setLength(end, MAX(WorkSet_getLength(end), slot + 1));

#line 2317 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      *TemporaryAction__ACT(vs) = tempo_act_reload;
      res = reload_here;

#line 2320 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      DBG((dbg, DBG_GLOBAL, "\t%2Dthere is a free slot. capacity=%d, reload here=%f, pressure %s\n",
           level, n_regs - BlockState_PRESSURE(bs), reload_here, pressure_ok ? "ok" : "insufficient"));

#line 2324 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      /* look if we can bring the value in. */
      if (pressure_ok && reload_here > 0.0) {
        RollbackInfo_ rb[1];
        double new_limit = MIN(reload_here, limit);
        rb[0] = transBegin(ges);

#line 2330 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        *TemporaryAction__ACT(vs) = tempo_act_live_through;
        *BlockState__PRESSURE(bs) += 1;
        res = canBringIn(ges, bl, var, new_limit, level + 1);

#line 2334 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        /*
         * if bring in is too expensive re-adjust the pressure
         * and roll back the state
         */
        if (res >= reload_here) {
          *BlockState__PRESSURE(bs) -= 1;
          *TemporaryAction__ACT(vs) = tempo_act_reload;
          transRollback(ges, rb);
          res = reload_here;
        }
      }

#line 2347 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      DBG((dbg, DBG_GLOBAL, "\t%2D%s\n", level,
           TemporaryAction_ACT(vs) == tempo_act_reload ? "reloading" : "bringing in"));
    }
  }

#line 2352 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
 end:
  DBG((dbg, DBG_GLOBAL, "\t%2D-> %f\n", level, res));
  return res;
}

#line 2357 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static double
canBringIn(GlobalEndState ges, BasicBlock bl, Temporary var, double limit,
           int32_t level) {
  BeladyEnv env = GlobalEndState_BELADYENV(ges);
  double glob_costs = (Temporary_hasRegister(var))
    ? 0.0: limit + 1;

#line 2364 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, DBG_GLOBAL, "\t%2Dcan bring in (max %f) for %p at block %p\n", level, limit, var, bl));

#line 2366 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if (isTransportIn(env, bl, var) && !Temporary_isDedicated(var)) {
    RollbackInfo_ rb[1];
    int32_t argIdx = 0;
    Operation producer = Temporary_getProducer(var);
    bool is_local_phi = isLocalPhi(bl, producer);
    rb[0] = transBegin(ges);

#line 2373 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    glob_costs = 0.0;
    // Check that we can bring in the variable or its representations (for phi
    // nodes) from all its predecessors. 
    BasicBlock_ENTER_FOREACH_BasicBlockEdge(bl, edge) {
      BasicBlock pred = BasicBlockEdge_origBlock(edge);
      Temporary curTempo = is_local_phi?
        Operation_getArgument(producer, argIdx++): var;
      double c;

#line 2382 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      /*
       * there might by unknwons as operands of phis in that case
       * we set the costs to zero, since they won't get spilled.
       */
      //FIXME: does it exist an equivalent test in LAO? (arch part)
      // if (arch_irn_consider_in_reg_alloc(env->arch, env->cls, op))
      c = canMakeAvailableAtEnd(ges, pred, curTempo, limit - glob_costs,
                                level + 1);
      // else
      //	c = 0.0;

#line 2393 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      glob_costs += c;

#line 2395 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      if (glob_costs >= limit) {
        glob_costs = limit + 1;
        transRollback(ges, rb);
        goto end;
      }
    } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
  }

#line 2403 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
 end:
  DBG((dbg, DBG_GLOBAL, "\t%2D-> %f\n", level, glob_costs));
  return glob_costs;
}

#line 2408 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
materializeAndCommitEndState(GlobalEndState ges) {
  BeladyEnv env = GlobalEndState_BELADYENV(ges);
  TemporaryAction ia;
  BlockState bs;

#line 2414 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, DBG_GLOBAL, "\tmaterializing\n"));

#line 2416 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * Perform all the variable actions.
   */
  for (ia = GlobalEndState_IATOP(ges); ia != NULL; ia = TemporaryAction_NEXT(ia)) {
    switch (TemporaryAction_ACT(ia)) {
    case tempo_act_live_through:
      {
        BlockInfo bi = BasicBlock_getBlockInfo(TemporaryAction_BASICBLOCK(ia));
        BringIn iter;
        Operation producer = Temporary_getProducer(TemporaryAction_VAR(ia));
        if (isLocalPhi(TemporaryAction_BASICBLOCK(ia), producer)) {
          BitSet_insert(GlobalEndState_SUCCPHIS(ges), Operation_identity(producer));
          DBG((dbg, DBG_GLOBAL, "\t\tlive through phi kept alive: %p\n", ia->irn));
        }

#line 2431 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
        list_for_each_entry_reverse(BringIn_, iter, BlockInfo_BRHEAD(bi), LIST_) {
          ++(*BringIn__SECTPRESSURE(iter));
        }
        ++(*BlockInfo__PRESSURE(bi));
      }
      break;
    case tempo_act_reload:
      SpillEnv_addReloadAtEnd(BeladyEnv_SPILLENV(env), TemporaryAction_VAR(ia),
                              TemporaryAction_BASICBLOCK(ia));
      DBG((dbg, DBG_GLOBAL, "\t\tadding reload of %p at end of %p\n", ia->irn, ia->bl));
      break;
    default:
      DBG((dbg, DBG_GLOBAL, "\t\t%p is in the end set of %p\n", ia->irn, ia->bl));
    }
  }

#line 2447 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * Commit the block end states
   */
  for (bs = GlobalEndState_BSTOP(ges); bs != NULL; bs = BlockState_NEXT(bs)) {
    BlockInfo bi = BlockState_BLOCKINFO(bs);

#line 2453 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    if (!BitSet_contains(GlobalEndState_COMMITTED(ges), BlockInfo_ID(bi))) {
      DBG((dbg, DBG_GLOBAL, "\t\tcommiting workset of %p with version %x\n", BlockInfo_BASICBLOCK(bi), GlobalEndState_VERSION(ges)));
      // bes->bs->end_state->vars[idx].version = ges->version;
      WorkSet_copy(env, BlockInfo_WORKSETEND(bi), BlockState_ENDSTATE(bs));
      DBG((dbg, DBG_GLOBAL, "\t\told pressure: %d, new pressure: %d, end length: %d\n",
           BlockInfo_PRESSURE(bi), BlockState_PRESSURE(bs), WorkSet_getLength(BlockState_ENDSTATE(bs))));
      *BlockInfo__PRESSURE(bi) = BlockState_PRESSURE(bs);
      BitSet_insert(GlobalEndState_COMMITTED(ges), BlockInfo_ID(bi));
    }
  }

#line 2464 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* clear the committed bitset. the next call is expecting it. */
  BitSet_empty(GlobalEndState_COMMITTED(ges));
}

#line 2468 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static Operation
betterSpilledHere(const_BringIn br) {
  const_BlockInfo bi = BringIn_BLOCKINFO(br);
  double spill_ef = BasicBlock_frequency(BringIn_getTemporaryBasicBlock(br));

#line 2473 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * If the bring in node is a phi in the bring in block,
   * we look at all definitions and sum up their execution frequencies,
   * since spills will be placed there.
   * (except for the case where an operand is also a phi which is spilled :-( )
   * If that cost is higher than spilling the phi in that block, we opt for
   * bringing the phi into the block and spill it there.
   */
  Operation producer = Temporary_getProducer(BringIn_TEMPORARY(br));
  if (isLocalPhi(BlockInfo_BASICBLOCK(bi), producer)) {
    spill_ef = 0.0;
    BasicBlock_ENTER_FOREACH_BasicBlockEdge(BlockInfo_BASICBLOCK(bi), edge) {
      BasicBlock pred = BasicBlockEdge_origBlock(edge);
      spill_ef += BasicBlock_frequency(pred);
    } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
  }

#line 2490 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  return BlockInfo_FREQUENCY(bi) < spill_ef?
    Operation_prev(BlockInfo_FIRSTNONIN(bi)) : NULL;
}

#line 2494 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static int32_t
getMaxPressureSoFar(const_BlockInfo bi, BringIn br) {
  ListHead l;
  int32_t res = INT_MIN;

#line 2499 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  Except_CHECK(BringIn_BLOCKINFO(br) == bi);
  for (l = &br->LIST_; l != BlockInfo_BRHEAD(bi); l = l->prev) {
    br  = list_entry(l, BringIn_, LIST_);
    res = MAX(res, BringIn_SECTPRESSURE(br));
  }

#line 2505 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* finally consider the front pressure distance and add the reference line (the global block pressure) */
  return MAX(res, BlockInfo_FRONTPRESSURE(bi));
}

#line 2509 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/**
 * Try to bring a variable into a block.
 * @param ges   The state of all end sets.
 * @param block The block.
 * @param irn   The variable.
 */
static void
optimizeVariable(GlobalEndState ges, BringIn br) {
  BlockInfo bi = BringIn_BLOCKINFO(br);
  Temporary var = BringIn_TEMPORARY(br);
  BasicBlock bl = BlockInfo_BASICBLOCK(bi);
  BeladyEnv env = GlobalEndState_BELADYENV(ges);
  int32_t k = BeladyEnv_NREGS(env);
  int32_t pressure_upto_use = getMaxPressureSoFar(bi, br);
  int32_t front_pressure = BlockInfo_FRONTPRESSURE(bi);
  Operation better_spill_loc = NULL;

#line 2526 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  Except_CHECK(front_pressure <= k);
  Except_CHECK(pressure_upto_use <= k);

#line 2529 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, DBG_GLOBAL, "fixing %p at %p (%f), front pr: %d, pr to use: %d, first use: %x\n", 
       var, bl, BlockInfo_FREQUENCY(bi), front_pressure, pressure_upto_use,
       BringIn_FIRSTUSE(br)));

#line 2533 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * if we cannot bring the value to the use, let's see ifit would be worthwhile
   * to bring the value to the beginning of the block to have a better spill
   * location.
   *
   * better _spilled_here will return a node where the value can be spilled after
   * or NULL if this block does not provide a better spill location.
   */
  if (pressure_upto_use >= k && front_pressure < k &&
      !BitSet_containsTempo(BeladyEnv_SPILLED(env), var)) {
    better_spill_loc = betterSpilledHere(br);
  }

#line 2546 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * If either we can bring the value to the use or we should try
   * to bring it here to do the spill here, let's try to bring it in.
   */
  if (better_spill_loc || pressure_upto_use < k) {
    BlockState bs;
    double bring_in_costs, local_costs;
    RollbackInfo_ trans[1];
    int pressure_inc;

#line 2556 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* process all variables which shall be in a reg at
     * the beginning of the block in the order of the next use. */
    local_costs  = Spill_getReloadCost(var) * BasicBlock_frequency(Operation_basicBlock(BringIn_FIRSTUSE(br)));

#line 2560 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* reset the lists */
    *GlobalEndState__BSTOP(ges)  = NULL;
    *GlobalEndState__IATOP(ges)  = NULL;

#line 2564 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* if the variable will live into this block, we must adapt the pressure.
     * The new pressure is the MAX of:
     * 1) the total block pressure
     * 2) the pressure so far + the front pressure increase + 1
     *
     * If the second is larger than the first,
     * we have to increment the total block pressure and hence
     * save the old pressure to restire it in case of failing to
     * bring the variable into the block in a register.
     */
    trans[0] = transBegin(ges);
    bs = BlockState_make(ges, bi);
    pressure_inc = MAX(BlockState_PRESSURE(bs), better_spill_loc ? front_pressure : pressure_upto_use + 1);
    *BlockState__PRESSURE(bs) = pressure_inc;

#line 2579 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    Except_CHECK(BlockInfo_PRESSURE(bi) <= k);
    DBG((dbg, DBG_GLOBAL, "\ttrans in var %p, version %x\n", var,
         GlobalEndState_VERSION(ges)));
    bring_in_costs = canBringIn(ges, bl, var, local_costs, 1);
    DBG((dbg, DBG_GLOBAL, "\tbring in: %f, local: %f\n", bring_in_costs, local_costs));

#line 2585 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /*
     * Following cases can now occur:
     * 1) There is room and costs ok
     * 2) Cannot bring to use but can spill at begin and costs are ok
     * 3) neither of both worked.
     *
     * following actions can be taken:
     * a) commit changes
     * b) mark phi as succeded if node was phi
     * c) insert reload at use location
     * d) give a spill location hint
     *
     * this is the case/action matrix
     *   | abcd
     * --+----------
     * 1 | XX
     * 2 | XXXX
     * 3 |   X
     */

#line 2605 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* the costs were acceptable... */
    if (bring_in_costs < local_costs) {
      int32_t check = 0;
      BringIn iter;
      Operation producer = Temporary_isDedicated(var)?
        findTrivialBackwardDefinition(BringIn_FIRSTUSE(br), bl, var):
        Temporary_getProducer(var);

#line 2613 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      /*
       * case 1 and first part of case 2:
       * commit all the changes done. this manifests the bring-in action.
       * if the transport-in was a phi (that is actually used in block)
       * mark it in the succ_phis set to *not* phi spill it.
       */
      materializeAndCommitEndState(ges);
      if (isLocalPhi(bl, producer)) {
        BitSet_insert(GlobalEndState_SUCCPHIS(ges), Operation_identity(producer));
      }

#line 2624 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      DBG((dbg, DBG_GLOBAL, "\t-> bring it in.", pressure_inc));

#line 2626 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      /* second half of case 2 */
      if (pressure_upto_use >= k) {
        DBG((dbg, DBG_GLOBAL, "\t-> use blocked. local reload: %p, try spill at: %p\n",
             BringIn_FIRSTUSE(br), better_spill_loc));
        SpillEnv_addReload(BeladyEnv_SPILLENV(env), var, BringIn_FIRSTUSE(br));
        SpillEnv_addSpill(BeladyEnv_SPILLENV(env), var,
                          Operation_basicBlock(better_spill_loc),
                          Operation_next((better_spill_loc)));
      }

#line 2636 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
      /*
       * go from the last bring in use to the first and add all the variabled
       * which additionally live through the block to their pressure.
       * at the point were the actually treated use is, we have to increase
       * the pressure by one more as the nrought in value starts to count.
       * Finally, adjust the front pressure as well.
       */
      pressure_inc = 0;
      list_for_each_entry_reverse(BringIn_, iter, BlockInfo_BRHEAD(bi), LIST_) {
        if (iter == br) {
          pressure_inc += pressure_upto_use < k;
        }
        *BringIn__SECTPRESSURE(iter) += pressure_inc;
        check = MAX(check, BringIn_SECTPRESSURE(iter));
        DBG((dbg, DBG_GLOBAL, "\tinc section pressure of %p by %d to %d\n",
             BringIn_FIRSTUSE(iter), pressure_inc, BringIn_SECTPRESSURE(iter)));
      }
      *BlockInfo__FRONTPRESSURE(bi) += pressure_inc;
      Except_CHECK(MAX(check, BlockInfo_FRONTPRESSURE(bi)) <= BlockInfo_PRESSURE(bi));
      DBG((dbg, DBG_GLOBAL, "\t-> result: p: %d, fp: %d\n",
           BlockInfo_PRESSURE(bi), BlockInfo_FRONTPRESSURE(bi)));
    }
    
    /* case 3: nothing worked. insert normal reload and rollback. */
    else {
      DBG((dbg, DBG_GLOBAL, "\t-> bring in was too expensive. local reload: %p\n", BringIn_FIRSTUSE(br)));
      SpillEnv_addReload(BeladyEnv_SPILLENV(env), var, BringIn_FIRSTUSE(br));
      BitSet_addTempo(BeladyEnv_SPILLED(env), var);
      transRollback(ges, trans);
    }
  }

#line 2668 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* there was no opportunity for optimization at all. reload and be sad ...  */
  else {
    DBG((dbg, DBG_GLOBAL, "\t-> can\'t do anything but reload before %p\n",
         BringIn_FIRSTUSE(br)));
    SpillEnv_addReload(BeladyEnv_SPILLENV(env), var, BringIn_FIRSTUSE(br));
    BitSet_addTempo(BeladyEnv_SPILLED(env), var);
  }

#line 2676 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  DBG((dbg, DBG_GLOBAL, "\n"));

#line 2678 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* reset the obstack and create a new version. */
  *GlobalEndState__VERSION(ges) = ver_make_newer(GlobalEndState_VERSION(ges));
}

#line 2682 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
determineGlobalOrder(IDList res, BeladyEnv env) {
  BringIn elm;
  int32_t i;

#line 2687 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  for (i = 0; i < BeladyEnv_BLOCKCOUNT(env); ++i) {
    BlockInfo bi = g_blockInfos[i];
    if(bi != NULL) {
      list_for_each_entry(BringIn_, elm, BlockInfo_BRHEAD(bi), LIST_) {
        IDList_PUSH(res, BringIn, elm);
      }
    }
  }

#line 2696 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  IDList_sort(res, BringIn_cmp);
}

#line 2699 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
resetMark(BeladyEnv env) {
  CodeRegion_FOREACH_BasicBlock(BeladyEnv_CODEREGION(env), basicBlock) {
    *BasicBlock__MARK(basicBlock) = 0;
  } CodeRegion_ENDEACH_BasicBlock;
}

#line 2706 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
dfsPerform(int32_t *postNum, BasicBlock bb) {
  *BasicBlock__MARK(bb) = 1;
  
  BasicBlock_LEAVE_FOREACH_BasicBlockEdge(bb, edge) {
    BasicBlock succ = BasicBlockEdge_destBlock(edge);
    if (!BasicBlock_MARK(succ)) {
      dfsPerform(postNum, succ);
    }
  } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;

#line 2717 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  g_dfsPostOrder[*postNum] = bb;
  *BlockInfo__POSTNUM(BasicBlock_getBlockInfo(bb)) = (*postNum)++;
}

#line 2721 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
globalAssign(BeladyEnv env) {
  GlobalEndState_ ges[1];
  IDList_ br[1];
  int32_t i = 0, j;
  IDList_Ctor(br, BeladyEnv_MEMORY(env));

#line 2728 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  GlobalEndState_Ctor(ges, env);

#line 2730 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  dfsPerform(&i, CodeRegion_rootBlock(BeladyEnv_CODEREGION(env)));

#line 2732 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* invalidate all state stack pointer versions */
  for (i = 0; i < BeladyEnv_BLOCKCOUNT(env); ++i) {
    BlockInfo bi = g_blockInfos[i];
    GlobalEndState_setBlockVersion(ges, bi, ver_oldest);

#line 2737 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    /* Set all block end sets entries to the youngest version */
    for (j = WorkSet_getLength(BlockInfo_WORKSETEND(bi)) - 1; j >= 0; --j) {
      WorkSet_setVersion(BlockInfo_WORKSETEND(bi), j, ver_youngest);
    }
  }

#line 2743 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* determine ordeer and optimize them */
  determineGlobalOrder(br, env);
  IDList_FOREACH(br, BringIn, bringIn) {
    optimizeVariable(ges, *bringIn);
  } IDList_ENDEACH;

#line 2749 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /*
   * Now we spill phis which cannot be kept since they were replaced
   * by reloads at the block entrances.
   */
  for (i = 0; i < BeladyEnv_BLOCKCOUNT(env); ++i) {
    BasicBlock bl = g_dfsPostOrder[i];
    if(bl == NULL) {
      continue;
    }
    BasicBlock_FOREACH_Operation(bl, op) {
      if (!isPhi(op)) {
        break;
      }
      if(Temporary_regFile(Operation_getResult(op, 0)) == BeladyEnv_REGFILE(env)
         && !BitSet_contains(GlobalEndState_SUCCPHIS(ges),
                             Operation_identity(op))) {
        SpillEnv_spillPhi(BeladyEnv_SPILLENV(env), op);
      }
    } BasicBlock_ENDEACH_Operation;
  }
}

#line 2771 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
cleanUpDefUses(BeladyEnv env) {
  // Remove references of usage in memory phis.
  // We do that to avoid updating the memory phis during reconstruct phase,
  // since it may causes some problems (like no spill location for argument of
  // memory phis) when removing them.
  TemporarySet_FOREACH_Temporary(SpillEnv_MEMPHIS(BeladyEnv_SPILLENV(env)),
                                 temporary) {
    Operation producer = Temporary_getProducer(temporary);
    int32_t argIdx = 0;
    Operation_ARGUMENTS_FOREACH_Temporary(producer, arg) {
      if(Temporary_isVariable(arg)) {
        DefUses_removeUse(Temporary_defUses(arg), producer, argIdx);
      }
      ++argIdx;
    } Operation_ARGUMENTS_ENDEACH_Temporary;
  } TemporarySet_ENDEACH_Temporary;
}

#line 2790 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static void
updateSpillInfo(IDList spillLocs, Temporary tempo, Temporary location) {
  IDList_FOREACH(spillLocs, SpillLoc_, spillLoc) {
    Temporary spillAddr = SpillLoc_LOC(spillLoc);
    Operation spill = SpillLoc_SPILL(spillLoc);
    int32_t i = 0;
    if(Temporary_isVariable(spillAddr)) {
      spillAddr = Spill_genStackedLocation(tempo);
    }
    // This check is not valid, since memPhi argument may be placed in ssa
    // variable when their source is a phi (still the problem of def uses
    // chains)
    //        Except_CHECK(spillAddr == Operation_getArgument(memPhi, argIdx));
    Except_CHECK(Temporary_isSymbol(spillAddr));
    Operation_ARGUMENTS_FOREACH_Temporary(spill, spillArg) {
      if(Temporary_isSymbol(spillArg) &&
         TemporaryInvariant_SYMBOL((TemporaryInvariant)spillAddr)
         == TemporaryInvariant_SYMBOL((TemporaryInvariant)spillArg)) {
        Operation_setArgument(spill, i, location);
      }
      ++i;
    } Operation_ARGUMENTS_ENDEACH_Temporary;
    *SpillLoc__LOC(spillLoc) = location;
  } IDList_ENDEACH;
}

#line 2816 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
/*
 * pre-condition: ssaForm must be in CSSA
 */
static void
removePhiMemory(BeladyEnv env) {
  SpillEnv this = BeladyEnv_SPILLENV(env);
  TemporarySet_FOREACH_Temporary(SpillEnv_MEMPHIS(this), temporary) {
    SpillLoc spillLoc =
      (SpillLoc)IDList_firstItem(SpillInfo_SPILLS(SpillEnv_getSpillInfo(this, temporary)));
    // Modify spill slot
    Operation memPhi = SpillLoc_SPILL(spillLoc);
    Operation phi = Temporary_getProducer(temporary);
    //FIXME: Should use the result of memPhi for this assignment, but not
    // possible due to def uses chains constraints (see spillPhi_)
    Temporary location = Spill_genStackedLocation(temporary);
    Except_CHECK(isPhi(memPhi) && "Temporary should not be in memphis");
    
    Operation_ARGUMENTS_FOREACH_Temporary(phi, tempo) {
      SpillInfo argInfo = SpillEnv_getSpillInfo(this, tempo);
      updateSpillInfo(SpillInfo_SPILLS(argInfo), tempo, location);
      updateSpillInfo(SpillInfo_RELOADS(argInfo), tempo, location);
    } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 2839 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
    // Remove memPhi operation
    BasicBlock_detachOperation(Operation_basicBlock(memPhi), memPhi);
    // Do a simple dead code elimination on phi we spilled
    if(OperationArgumentList_isEmpty(DefUses_getUses(Temporary_defUses(temporary)))) {
      BasicBlock_detachOperation(Operation_basicBlock(phi), phi);
    }
  } TemporarySet_ENDEACH_Temporary;
}

#line 2848 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#ifndef Belady_POOLED
#   define Belady_POOLED 128
#endif//Belady_POOLED

#line 2865 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
void
SSASpill_spillBelady(SSAForm ssaForm, RegFile regFile, bool removeMemPhi) {
  BeladyEnv_ env[1];
  WorkSet_ emptySet[1];
  Memory memory = Memory_new(SSAForm_memory(ssaForm), Belady_POOLED);
  BlockInfo bi;

#line 2872 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  g_originalAlgo = getenv("SEB_ORIG") != NULL;

#line 2874 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* init belady env */
  BeladyEnv_Ctor(env, memory, ssaForm, regFile, removeMemPhi);
  Belady_init(env);
  // Create a dummy block information for a dummy block to avoid special
  // cases.
  bi = BlockInfo_make(env, CodeRegion_rootBlock(BeladyEnv_CODEREGION(env)));
  WorkSet_Ctor(emptySet, env);
  *BlockInfo__WORKSETEND(bi) = emptySet;

#line 2883 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  /* Fix high register pressure in blocks with belady algorithm */
  CodeRegion_FOREACH_BasicBlock(BeladyEnv_CODEREGION(env), basicBlock) {
    Except_CHECK(BasicBlock_MARK(basicBlock) == 0 && "mark not reset");
    // Why this call is necessary now!
    BasicBlock_makeOrderings(basicBlock, 0, 1);
    Belady(env, basicBlock);
  } CodeRegion_ENDEACH_BasicBlock;

#line 2891 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  globalAssign(env);

#line 2893 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if(BeladyEnv_REMOVEPHIMEMORY(env) == true) {
    cleanUpDefUses(env);
  }

#line 2897 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  // Insert spill/reload nodes into the graph and fix usages
  SpillEnv_insertSpillsReloads(BeladyEnv_SPILLENV(env));

#line 2900 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  if(BeladyEnv_REMOVEPHIMEMORY(env) == true) {
    removePhiMemory(env);
  }

#line 2904 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  Belady_fini(env);
#ifndef ECL_NDTOR
  BeladyEnv_Dtor(env);
#endif//ECL_NDTOR
  Memory_delete(memory);
}

#line 2911 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
static TemporarySet
TemporarySet_cloneRegFile(TemporarySet src, Convention convention, RegFile regFile,
                          Memory memory, int32_t estimate) {
  TemporarySet result = TemporarySet_make(memory, estimate);
  TemporarySet_FOREACH_Temporary(src, tempo) {
    if(Temporary_regFile(tempo) == regFile
       && (!Temporary_hasRegister(tempo)
           || Temporary_isAllocatableDedicated(tempo, convention))) {
      TemporarySet_insert(result, tempo);
    }
  } TemporarySet_ENDEACH_Temporary;
  return result;
}

#line 2929 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
int32_t
SSASpill_maxLive(SSAForm ssaForm, RegFile regFile) {
  int32_t result = 0;
  int32_t nbRegs = RegFile_highRegister(regFile) - RegFile_lowRegister(regFile) + 1;
  Memory memory = Memory_new(SSAForm_memory(ssaForm), Belady_POOLED);
  CodeRegion codeRegion = SSAForm_codeRegion(ssaForm);
  Procedure procedure = CodeRegion_procedure(codeRegion);
  Convention convention = CodeRegion_convention(codeRegion);
  TemporaryTable globalTable =
    CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
  Liveness liveness = Liveness_make(memory, codeRegion, globalTable, true);
  RegisterSet_ allocatableSet[1];
  *allocatableSet = Convention_allocatableSet(convention, regFile);
  Liveness_multiplex(liveness);

#line 2944 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    TemporarySet temporarySet = TemporarySet_cloneRegFile(BasicBlock_liveOutSet(basicBlock),
                                                          convention, regFile, memory, nbRegs);
    result = MAX(result, TemporarySet_count(temporarySet));
    BasicBlock_FORBACK_Operation(basicBlock, operation) {
      int32_t tmp = 0;
      RegisterSet_ seenColor[1];
      RegisterSet_Ctor(seenColor);
      Operation_RESULTS_FOREACH_Temporary(operation, res) {
        if(Temporary_regFile(res) == regFile) {
          TemporarySet_remove(temporarySet, res);
          if(Temporary_hasRegister(res)) {
            TemporarySet_remove(temporarySet,
                                Procedure_makeDedicatedTemporary(procedure,
                                                                 Temporary_register(res)));
            if(Temporary_isAllocatableDedicated(res, convention)) {
              RegisterSet_insert(seenColor, Temporary_register(res));
              ++tmp;
            }
          } else {
            ++tmp;
          }
        }
      } Operation_RESULTS_ENDEACH_Temporary;
      if(Operation_clobberSet(operation) != NULL) {
        RegisterSet_FOREACH_Register(Operation_clobberSet(operation), reg) {
          if(Register_regFile(reg) == regFile) {
            if(!RegisterSet_contains(seenColor, reg)
               && RegisterSet_contains(allocatableSet, reg)) {
              ++tmp;
            }
          }
        } RegisterSet_ENDEACH_Register;
      }
      result = MAX(result, TemporarySet_count(temporarySet) + tmp);
      tmp = 0;
      RegisterSet_empty(seenColor);
      if(!isPhi(operation)) {
        Operation_ARGUMENTS_FOREACH_Temporary(operation, arg) {
          if(Temporary_regFile(arg) == regFile
             && (!Temporary_hasRegister(arg)
                 || Temporary_isAllocatableDedicated(arg, convention))) {
            TemporarySet_insert(temporarySet, arg);
            if(Temporary_hasRegister(arg)) {
              RegisterSet_insert(seenColor, Temporary_register(arg));
            }
          }
        } Operation_ARGUMENTS_ENDEACH_Temporary;
      }
      if(Operation_clobberSet(operation) != NULL) {
        RegisterSet_FOREACH_Register(Operation_clobberSet(operation), reg) {
          if(Register_regFile(reg) == regFile) {
            if(!RegisterSet_contains(seenColor, reg)
               && RegisterSet_contains(allocatableSet, reg)) {
              ++tmp;
            }
          }
        } RegisterSet_ENDEACH_Register;
      }
      result = MAX(result, TemporarySet_count(temporarySet) + tmp);
      RegisterSet_Dtor(seenColor);
    } BasicBlock_ENDBACK_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
  Liveness_kill(liveness);
  return result;
}

#line 3011 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 3243 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 3255 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
#if XCC__C
static void
SSASpill_TEST(void)
{
#include "CGO/SSASpill_.c"
}

#line 3262 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/SSASpill.xcc"
int
main(int argc, char **argv)
{
  CGO_Test(SSASpill_TEST);
  return 0;
}
#endif


#endif//SSASpill_c_INCLUDED
