#ifndef Scoreboarder__h_INCLUDED
#define Scoreboarder__h_INCLUDED

//
#include "CGO/Scoreboarder.h"

/*
 * ScoreboarderState --	Scoreboarder state for a BasicBlock.
 */
struct ScoreboarderState_ {
  //@args	Memory memory, Processor processor, int windowSize
  BasicBlock BASICBLOCK;
  //@access FREQUENCY	BasicBlock_frequency(ScoreboarderState_BASICBLOCK(this))
  bool ISLISTED;		// This ScoreboarderState is in Scoreboarder worklist.
  Processor PROCESSOR;
  struct ScoreboarderState_ *WORKNEXT;
  uint32_t STAMP;		// Time stamp for this ScoreboarderState.
  Scoreboard_ SCOREBOARD[1];	// Scoreboard at the end of BASICBLOCK.
  //@access MEMORY	Scoreboard_memory(ScoreboarderState_SCOREBOARD(this))
};
typedef struct ScoreboarderState_ ScoreboarderState_, *ScoreboarderState;
typedef const struct ScoreboarderState_ *const_ScoreboarderState;
typedef struct ScoreboarderState_ * restrict_ScoreboarderState;

extern ScoreboarderState
ScoreboarderState_Ctor(ScoreboarderState this, Memory memory, Processor processor, int windowSize);

extern ScoreboarderState
ScoreboarderState_Copy(ScoreboarderState this, const_ScoreboarderState that);

extern void
ScoreboarderState_Dtor(ScoreboarderState this);

extern size_t
ScoreboarderState_Size(Memory memory, Processor processor, int windowSize);

#define ScoreboarderState_BASICBLOCK(this) ((this)->BASICBLOCK)
#define ScoreboarderState__BASICBLOCK(this) (&(this)->BASICBLOCK)
#define ScoreboarderState_ISLISTED(this) ((this)->ISLISTED)
#define ScoreboarderState__ISLISTED(this) (&(this)->ISLISTED)
#define ScoreboarderState_PROCESSOR(this) ((this)->PROCESSOR)
#define ScoreboarderState__PROCESSOR(this) (&(this)->PROCESSOR)
#define ScoreboarderState_WORKNEXT(this) ((this)->WORKNEXT)
#define ScoreboarderState__WORKNEXT(this) (&(this)->WORKNEXT)
#define ScoreboarderState_STAMP(this) ((this)->STAMP)
#define ScoreboarderState__STAMP(this) (&(this)->STAMP)
#define ScoreboarderState_SCOREBOARD(this) ((this)->SCOREBOARD)
#define ScoreboarderState__SCOREBOARD(this) (&(this)->SCOREBOARD)
#define ScoreboarderState_FREQUENCY(this) (BasicBlock_frequency(ScoreboarderState_BASICBLOCK(this)))
#define ScoreboarderState_MEMORY(this) (Scoreboard_memory(ScoreboarderState_SCOREBOARD(this)))


/*
 * ScoreboarderState_schedule --	Local scheduling for this ScoreboarderState.
 */
bool
ScoreboarderState_schedule(ScoreboarderState this, Scoreboard temp_scoreboard);

//
struct Scoreboarder_ {
  //@args	Memory memory, CodeRegion codeRegion
  Memory MEMORY;
  CodeRegion CODEREGION;
  Scoreboard_ SCOREBOARD[1];		// Used to merge entering ScoreboarderState.
  uint32_t STAMP;
  int32_t ITERCOUNT;
  int32_t BLOCKCOUNT;
  ScoreboarderState_ *BASESTATE;	// Base of ScoreboarderState array.
  ScoreboarderState_ *WORKTAIL;		// First ScoreboarderState in worklist.
  ScoreboarderState_ *WORKHEAD;		// Last ScoreboarderState in worklist.
};

extern Scoreboarder
Scoreboarder_Ctor(Scoreboarder this, Memory memory, CodeRegion codeRegion);

extern Scoreboarder
Scoreboarder_Copy(Scoreboarder this, const_Scoreboarder that);

extern void
Scoreboarder_Dtor(Scoreboarder this);

extern size_t
Scoreboarder_Size(Memory memory, CodeRegion codeRegion);

#define Scoreboarder_MEMORY(this) ((this)->MEMORY)
#define Scoreboarder__MEMORY(this) (&(this)->MEMORY)
#define Scoreboarder_CODEREGION(this) ((this)->CODEREGION)
#define Scoreboarder__CODEREGION(this) (&(this)->CODEREGION)
#define Scoreboarder_SCOREBOARD(this) ((this)->SCOREBOARD)
#define Scoreboarder__SCOREBOARD(this) (&(this)->SCOREBOARD)
#define Scoreboarder_STAMP(this) ((this)->STAMP)
#define Scoreboarder__STAMP(this) (&(this)->STAMP)
#define Scoreboarder_ITERCOUNT(this) ((this)->ITERCOUNT)
#define Scoreboarder__ITERCOUNT(this) (&(this)->ITERCOUNT)
#define Scoreboarder_BLOCKCOUNT(this) ((this)->BLOCKCOUNT)
#define Scoreboarder__BLOCKCOUNT(this) (&(this)->BLOCKCOUNT)
#define Scoreboarder_BASESTATE(this) ((this)->BASESTATE)
#define Scoreboarder__BASESTATE(this) (&(this)->BASESTATE)
#define Scoreboarder_WORKTAIL(this) ((this)->WORKTAIL)
#define Scoreboarder__WORKTAIL(this) (&(this)->WORKTAIL)
#define Scoreboarder_WORKHEAD(this) ((this)->WORKHEAD)
#define Scoreboarder__WORKHEAD(this) (&(this)->WORKHEAD)


/*
 * Scoreboarder_blockCount --	For use in Scoreboarder_FOREACH.
 */
static inline int32_t
Scoreboarder_blockCount(Scoreboarder this)
{
  return Scoreboarder_BLOCKCOUNT(this);
}

/*
 * Scoreboarder_baseState --	For use in Scoreboarder_FOREACH.
 */
static inline ScoreboarderState_ *
Scoreboarder_baseState(Scoreboarder this)
{
  return Scoreboarder_BASESTATE(this);
}

/*
 * Scoreboarder_FOREACH_ScoreboarderState --	Iterate over this Scoreboarder ScoreboarderState.
 */
#define Scoreboarder_FOREACH_ScoreboarderState(this, state) { \
  int32_t Scoreboarder_BLOCKCOUNT = Scoreboarder_blockCount(this); \
  ScoreboarderState Scoreboarder_BASESTATE = Scoreboarder_baseState(this), state; \
  ScoreboarderState Scoreboarder_PASTSTATE = Scoreboarder_BASESTATE + Scoreboarder_BLOCKCOUNT; \
  for (state = Scoreboarder_BASESTATE; state < Scoreboarder_PASTSTATE; ++state) { \
    if (ScoreboarderState_BASICBLOCK(state) != NULL) {
#define Scoreboarder_ENDEACH_ScoreboarderState \
    } \
  } \
}

//
#define Scoreboarder_FORWORK_ScoreboarderState(this, state) { \
  ScoreboarderState Scoreboarder_WORKNEXT = NULL; \
  ScoreboarderState state = Scoreboarder_WORKHEAD(this); \
  for (; state != NULL; state = Scoreboarder_WORKNEXT) { \
    Scoreboarder_WORKNEXT = ScoreboarderState_WORKNEXT(state);
#define Scoreboarder_ENDWORK_ScoreboarderState \
  } \
}

/*
 * Scoreboarder_activateState --	Activate ScoreboarderState for a BasicBlock.
 *
 * We push last on the worklist the blocks in post-order, so they are pop(ed) in
 * reverse post-order. This is a recommended ordering for forward data-flow problems.
 */
void
Scoreboarder_activateState(Scoreboarder this, BasicBlock curr_block);

/*
 * Scoreboarder_doIterate --	Iterate this Scoreboarder problem until fixpoint.
 *
 * This is the worklist iterative algorithm to solve forward data-flow problems.
 */
void
Scoreboarder_doIterate(Scoreboarder this);

/*
 * Scoreboarder_doFixup --	Fixup pending latencies at BasicBlock boundaries.
 */
void
Scoreboarder_doFixup(Scoreboarder this);

/*
 * Scoreboarder_writeBack --	Write-back results into the CodeRegion.
 */
void
Scoreboarder_writeBack(Scoreboarder this);

/*
 * Scoreboarder_INIT --	Initialize the Scoreboarder module.
 */
#define Scoreboarder_INIT()

/*
 * Scoreboarder_FINI --	Finalize the Scoreboarder module.
 */
#define Scoreboarder_FINI()


#endif//Scoreboarder__h_INCLUDED
