#ifndef Spill_c_INCLUDED
#define Spill_c_INCLUDED

#line 15 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*
 * This module provide support function for the generation
 * of spill symbol and the generation of spill code.
 */

#line 20 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
#include "CGO/CGO_.h"

#line 25 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 104 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*
  Default spill cost values. 
  TODO: Should be defined by architecture.
  The cost is compute as:
  cost = latency+issue/max_issue.
  The default costs are for a machine with:
  issue of remat: 1/4
  issue of load or store: 1/1
  latency of load: 3
  latency of store: 0
  This is the ST200 model.
*/
#define SPILL_COST_SPILL_REMAT 0.0
#define SPILL_COST_RELOAD_REMAT 1.25
#define SPILL_COST_SPILL 1.0
#define SPILL_COST_RELOAD 4.0

#line 121 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*
 * Helper macros
 */
/* Defined here, but should be available in Operator.xcc. */
#define Operator_loadedIndex(operator) 0      // Assume loaded value is at result index 0.

#line 127 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*!
  Code region currently processed
*/
static CodeRegion spill_coderegion;

#line 132 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*!
  Table of Temporary -> spill symbol (Symbol type) mapping
  And access functions
*/
static HTable_ spill_table_;
static HTable spill_table = (HTable)&spill_table_;

#line 139 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
static Symbol 
Spill_getSymbol(Temporary temporary)
{
  void *pointer = HTable_search(spill_table, temporary);
  if (pointer != NULL) return *(Symbol *)pointer;
  return NULL;
}

#line 147 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
static void
Spill_setSymbol(Temporary temporary, Symbol symbol)
{
  Symbol *_symbol = HTable_lookup(spill_table, temporary, sizeof(Symbol));
  *_symbol = symbol;
}

#line 154 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
void
Spill_initCodeRegion(CodeRegion codeRegion)
{
  spill_coderegion = codeRegion;
  HTable_Ctor(spill_table, Procedure_memory(CodeRegion_procedure(codeRegion)), 1024);
}

#line 161 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
void
Spill_finiCodeRegion(CodeRegion codeRegion)
{
}

#line 166 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
Temporary
Spill_genStackedLocation(Temporary spilled_temporary)
{
  static uint32_t spill_temp;
  static char buffer[16+10+1]; // prefix_len + maxuint len + 1
  Procedure procedure = CodeRegion_procedure(spill_coderegion);
  Program program = Procedure_program(procedure);
  Symbol symbol = Spill_getSymbol(spilled_temporary);
  if (symbol == NULL) {
    int lenght = sprintf(buffer, ".lao_spill_temp_%u", spill_temp++);
    Except_ENSURE(lenght < sizeof(buffer));
    symbol = SymbolTable_makeSymbol(Program_symbolTable(program), buffer);  
    Symbol_setClasses(symbol, SymbolClass_VAR, SymbolStore_AUTO, SymbolExport_LOCAL);
    Symbol_setNativeType(symbol, RegFile_getNativeType(Temporary_regFile(spilled_temporary)));
    Spill_setSymbol(spilled_temporary, symbol);
  }
  return Procedure_makeSymbolTemporary(procedure, symbol, 0, Relocation__UNDEF);
}

#line 185 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
Temporary
Spill_getRematerializeValue(Temporary spilled_temporary)
{
  Procedure procedure = CodeRegion_procedure(spill_coderegion);
  Temporary value = Procedure_getRemater(procedure, spilled_temporary);
  Except_REQUIRE(Temporary_isRemater(spilled_temporary));
  Except_ENSURE(Temporary_isInvariant(value));
  return value;
}

#line 195 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
Temporary
Spill_getHomeLocation(Temporary spilled_temporary)
{
  Procedure procedure = CodeRegion_procedure(spill_coderegion);
  Temporary home = Procedure_getHomeable(procedure, spilled_temporary);
  Except_REQUIRE(Temporary_isHomeable(spilled_temporary));
  Except_ENSURE(Temporary_isSymbol(home));
  return home;
}

#line 205 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
static Operation
Spill_startOperation(BasicBlock bb)
{
  Operation last_phi = NULL, tmp;
  Operation start;
  Operation spill_start;
  bool phi_check = false, start_check = false;

#line 213 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  // Find last phi operation
  tmp = BasicBlock_firstOperation(bb);
  while (tmp != NULL && Operation_operator(tmp) == Operator_PHI) {
    last_phi = tmp;
    tmp = Operation_next(tmp);
  }

#line 220 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  start = NULL;
  if (BasicBlock_isEntry(bb)) {
    BasicBlock_FOREACH_Operation(bb, operation) {
      if (Operation_isEntryCode(operation)) 
        start = Operation_next(operation);
	  else break;
    } BasicBlock_ENDEACH_Operation;
  }

#line 229 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (last_phi == NULL) phi_check = true;
  if (start == NULL) start_check = true;
  if (phi_check && start_check) return NULL;
  
  spill_start = BasicBlock_firstOperation(bb);
  while (spill_start != NULL) {
    if (last_phi == spill_start) phi_check = true;
    if (start == spill_start) start_check = true;
    if (phi_check && start_check) break;
    spill_start = Operation_next(spill_start);
  }
  return spill_start;
}

#line 243 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
static Operation
Spill_stopOperation(BasicBlock bb)
{
  Operation control, stop;
  bool control_check = false, stop_check = false;
  Operation spill_stop;
  
  control = BasicBlock_getControlOperation(bb);
  stop = NULL;

#line 253 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (BasicBlock_isExit(bb)) {
    BasicBlock_FORBACK_Operation(bb, operation) {
	  if (Operation_isExitCode(operation)) 
        stop = Operation_prev(operation);
	  else break;
    } BasicBlock_ENDBACK_Operation;
  }

#line 261 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (control == NULL) control_check = true;
  if (stop == NULL) stop_check = true;
  if (control_check && stop_check) return NULL;

#line 265 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  spill_stop = BasicBlock_lastOperation(bb);
  while (spill_stop != NULL) {
    if (control == spill_stop) control_check = true;
    if (stop == spill_stop) stop_check = true;
    if (control_check && stop_check) break;
    spill_stop = Operation_prev(spill_stop);
  }
  return spill_stop;
}

#line 279 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*!
  Find forward the first safe operation in the entry basic block for
  insertion of spill code.
*/
Operation
Spill_findFirstAppendPoint(BasicBlock bb, Operation where)
{
  Operation start;
  Operation safe;
  bool where_found = false;

#line 290 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  start = Spill_startOperation(bb);
  if (start == NULL) return where;

#line 293 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (where == NULL) where_found = true;
  safe = BasicBlock_firstOperation(bb);
  while (safe != NULL) {
    if (safe == where) where_found = true;
    if (safe == start) {
      break;
    }
    safe = Operation_next(safe);
  }

#line 303 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  return where_found? safe: where;
}

#line 310 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
/*!
  Find backward the last safe operation in the exit basic block for
  insertion of spill code.
  If where is NULL, consider that the insertion point is at the end of the
  basic block.
  Returns where if it is a safe insertion point.
  Returns NULL if the insetion point is at the end of the basic block.
*/
Operation
Spill_findLastInsertPoint(BasicBlock bb, Operation where)
{
  Operation stop;
  Operation safe;
  bool where_found = false;

#line 325 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  stop = Spill_stopOperation(bb);
  if (stop == NULL) return where;

#line 328 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (where == NULL) where_found = true;
  safe = BasicBlock_lastOperation(bb);
  while (safe != NULL) {
    if (safe == where) where_found = true;
    if (safe == stop) {
      break;
    }
    safe = Operation_prev(safe);
  }

#line 338 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  return where_found? safe: where;
}

#line 341 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
void
Spill_insertReload(BasicBlock bb, Operation where, Temporary dst, Temporary spilled_temporary)
{
  Operation point;
  Selector_ selector_;
  Selector selector = &selector_;
  
  Selector_Ctor(selector, Selector_memory(BasicBlock_selector(bb)),
		Selector_procedure(BasicBlock_selector(bb)),
		Selector_processor(BasicBlock_selector(bb)), false);

#line 352 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  point = Spill_findLastInsertPoint(bb, where);
  
  if (Temporary_isRemater(spilled_temporary)) {
    Temporary value;

#line 357 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
    value = Spill_getRematerializeValue(spilled_temporary);

#line 359 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
    if (Temporary_isAbsolute(value)) {
      TemporaryInvariant invariant = Temporary_INVARIANT(value);
      Selector_selectMAKEAbsolute(selector,
                                  RegFile_getNativeType(Temporary_regFile(dst)), dst,
                                  TemporaryInvariant_VALUE(invariant));
    } else if (Temporary_isSymbol(value)) {
      TemporaryInvariant invariant = Temporary_INVARIANT(value);
      Selector_selectMAKEAddress(selector,
                                 RegFile_getNativeType(Temporary_regFile(dst)), dst,
                                 TemporaryInvariant_SYMBOL(invariant),
                                 TemporaryInvariant_OFFSET(invariant));
    } else Except_NEVER(true);
    
    if (CGO_TRACE) {
      fprintf(CGO_TRACE, "Insert rematerialization: ");
      Selector_FOREACH_Operation(selector, operation) {
	Operation_pretty(operation, CGO_TRACE);
      } Selector_ENDEACH_Operation;
    }
    
    if (point == NULL) BasicBlock_pushSelector(bb, selector);
    else BasicBlock_insertSelector(bb, point, selector);

#line 382 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  } else {
    Temporary locate = NULL;
    TemporaryInvariant locate_invariant = NULL;
    Operation load_op = NULL;
    if (Temporary_isHomeable(spilled_temporary)) {
      locate = Spill_getHomeLocation(spilled_temporary);
    } else {
      locate = Spill_genStackedLocation(spilled_temporary);
    }
    Except_ENSURE(Temporary_isSymbol(locate));
    locate_invariant = Temporary_INVARIANT(locate);

#line 394 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
    // If the insertion point is a store at the same location,
    // we remove the sequence: a = load @L, store @L = a.
    if (point != NULL &&
	Operator_isMemoryWrite(Operation_operator(point)) &&
	(Operation_flags(point) & OperationFlag_SpillCode)) {
      Operator point_operator = Operation_operator(point);
      int offsetIndex = Operator_offsetIndex(point_operator);
      int storedIndex = Operator_storedIndex(point_operator);
      Temporary offset = Operation_getArgument(point, offsetIndex);
      Temporary stored = Operation_getArgument(point, storedIndex);
      if (stored == dst && Temporary_isSymbol(offset)) {
        TemporaryInvariant offset_invariant = Temporary_INVARIANT(offset);
        if (      TemporaryInvariant_SYMBOL(offset_invariant)
               == TemporaryInvariant_SYMBOL(locate_invariant)
            &&    TemporaryInvariant_OFFSET(offset_invariant)
               == TemporaryInvariant_OFFSET(locate_invariant)) {
          CGO_TRACE && fprintf(CGO_TRACE, "Remove reload/store sequence: ")
	            && Operation_pretty(point, CGO_TRACE);
          BasicBlock_detachOperation(Operation_basicBlock(point), point);
          return;
        }
      }
    }
    
    Selector_selectDirectLOAD(selector,
                              RegFile_getNativeType(Temporary_regFile(dst)), dst,
		              TemporaryInvariant_SYMBOL(locate_invariant),
                              TemporaryInvariant_OFFSET(locate_invariant));
    /* Merk load operation as spill operation. */
    load_op = Selector_firstOperation(selector);
    while(load_op != NULL) {
      if (Operator_isMemoryRead(Operation_operator(load_op))) {
	Operation_setFlags(load_op, OperationFlag_SpillCode);
	break;
      }
      load_op = Operation_next(load_op);
    }
    
    if (CGO_TRACE) {
      fprintf(CGO_TRACE, "Insert reload: ");
      Selector_FOREACH_Operation(selector, operation) {
	Operation_pretty(operation, CGO_TRACE);
      } Selector_ENDEACH_Operation;
    }
    
    if (point == NULL) BasicBlock_pushSelector(bb, selector);
    else BasicBlock_insertSelector(bb, point, selector);
    
  }

#line 444 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  Selector_Dtor(selector);
}

#line 447 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
void
Spill_appendSpill(BasicBlock bb, Operation where, Temporary src, Temporary spilled_temporary)
{
  Selector_ selector_;
  Selector selector = &selector_;
  Temporary locate = NULL;
  TemporaryInvariant locate_invariant = NULL;
  Operation store_op, point, next;
  
  if (Temporary_isRemater(spilled_temporary))
    /* No need to store a rematerializable value. */
    return;

#line 460 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  point =  Spill_findFirstAppendPoint(bb, where);
  
  if (Temporary_isHomeable(spilled_temporary)) {
    locate = Spill_getHomeLocation(spilled_temporary);
  } else {
    locate = Spill_genStackedLocation(spilled_temporary);
  }
  Except_ENSURE(Temporary_isSymbol(locate));
  locate_invariant = Temporary_INVARIANT(locate);

#line 470 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  // If the insertion point is a load at the same location,
  // we remove the sequence: a = load @L, store @L = a.
  if (point != NULL &&
      Operator_isMemoryRead(Operation_operator(point)) &&
      (Operation_flags(point) & OperationFlag_SpillCode)) {
    Operator point_operator = Operation_operator(point);
    int offsetIndex = Operator_offsetIndex(point_operator);
    int loadedIndex = Operator_loadedIndex(point_operator);
    Temporary offset = Operation_getArgument(point, offsetIndex);
    Temporary loaded = Operation_getResult(point, loadedIndex);
    if (loaded == src && Temporary_isSymbol(offset)) {
      TemporaryInvariant offset_invariant = Temporary_INVARIANT(offset);
      if (      TemporaryInvariant_SYMBOL(offset_invariant)
             == TemporaryInvariant_SYMBOL(locate_invariant)
          &&    TemporaryInvariant_OFFSET(offset_invariant)
             == TemporaryInvariant_OFFSET(locate_invariant)) {
        CGO_TRACE && fprintf(CGO_TRACE, "Remove load/spill sequence: ")
                  && Operation_pretty(point, CGO_TRACE);
        BasicBlock_detachOperation(Operation_basicBlock(point), point);
        return;
      }
    }
  } 
  // If the operation following the insertion point is already a spill
  // of the same variable, we skip the spill
  if (point != NULL &&
      (next = Operation_next(point)) != NULL &&
      Operator_isMemoryWrite(Operation_operator(next)) &&
      (Operation_flags(next) & OperationFlag_SpillCode)) {
    Operator next_operator = Operation_operator(next);
    int offsetIndex = Operator_offsetIndex(next_operator);
    int storedIndex = Operator_storedIndex(next_operator);
    Temporary offset = Operation_getArgument(next, offsetIndex);
    Temporary stored = Operation_getArgument(next, storedIndex);
    if (stored == src && Temporary_isSymbol(offset)) {
      TemporaryInvariant offset_invariant = Temporary_INVARIANT(offset);
      if (      TemporaryInvariant_SYMBOL(offset_invariant)
             == TemporaryInvariant_SYMBOL(locate_invariant)
          &&    TemporaryInvariant_OFFSET(offset_invariant)
             == TemporaryInvariant_OFFSET(locate_invariant)) {
        CGO_TRACE && fprintf(CGO_TRACE, "Skip spill redundant with store: ")
	          && Operation_pretty(next, CGO_TRACE);
        return;
      }
    }
  } 
  
  
  Selector_Ctor(selector, Selector_memory(BasicBlock_selector(bb)),
		Selector_procedure(BasicBlock_selector(bb)),
		Selector_processor(BasicBlock_selector(bb)), false);
  
  Selector_selectDirectSTORE(selector,
                             RegFile_getNativeType(Temporary_regFile(src)), src,
		             TemporaryInvariant_SYMBOL(locate_invariant),
                             TemporaryInvariant_OFFSET(locate_invariant));
  /* Mark store operation as spill operation. */
  store_op = Selector_firstOperation(selector);
  while(store_op != NULL) {
    if (Operator_isMemoryWrite(Operation_operator(store_op))) {
      Operation_setFlags(store_op, OperationFlag_SpillCode);
      break;
    }
    store_op = Operation_next(store_op);
  }

#line 536 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (CGO_TRACE) {
    fprintf(CGO_TRACE, "Insert spill: ");
    Selector_FOREACH_Operation(selector, operation) {
      Operation_pretty(operation, CGO_TRACE);
    } Selector_ENDEACH_Operation;
  }

#line 543 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
  if (point == NULL) BasicBlock_putSelector(bb, selector);
  else BasicBlock_appendSelector(bb, point, selector);
  
  Selector_Dtor(selector);
}

#line 550 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
float
Spill_getSpillCost(Temporary spilled_temporary)
{
  if (Temporary_isRemater(spilled_temporary)) {
    return SPILL_COST_SPILL_REMAT;
  } else {
    return SPILL_COST_SPILL;
  }
}

#line 560 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
float
Spill_getReloadCost(Temporary spilled_temporary)
{
  if (Temporary_isRemater(spilled_temporary)) {
    return SPILL_COST_RELOAD_REMAT;
  } else {
    return SPILL_COST_RELOAD;
  }
}

#line 570 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 582 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
#if XCC__C
static void
Spill_TEST(void)
{
#include "Spill_.c"
}

#line 589 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/CGO/Spill.xcc"
int
main(int argc, char **argv)
{
  CGO_Test(Spill_TEST);
  return 0;
}
#endif


#endif//Spill_c_INCLUDED
