#ifndef BasicBlock__h_INCLUDED
#define BasicBlock__h_INCLUDED

//
#include "LIR/BasicBlock.h"
#include "PFA/PFA.h"

//
static inline bool
BasicBlockEdge_gt(const_BasicBlockEdge this, const_BasicBlockEdge that)
{
  BasicBlock this_block = BasicBlockEdge_ORIGBLOCK(this);
  BasicBlock that_block = BasicBlockEdge_ORIGBLOCK(that);
  int32_t this_index = BasicBlock_INDEX(this_block);
  int32_t that_index = BasicBlock_INDEX(that_block);
  return this_index > that_index;
}

/*
 * BasicBlock_sortEnterEdges --	Sort this BasicBlock ENTEREDGES array.
 *
 * See the en.wikibook.org heapsort.
 */
void
BasicBlock_sortEnterEdges(BasicBlock this);

/*
 * BasicBlock_flagEntryCode --	Flag the Procedure entry code of this BasicBlock
 *
 * The Procedure entry code sequence includes:
 * - any copy from a dedicated register that is not constant.
 * - copy, add or adjust that defines the frame pointer
 * - first copy, add or adjust that defines the stack pointer
 */
void
BasicBlock_flagEntryCode(BasicBlock this, Convention convention);

/*
 * BasicBlock_flagExitCode --	Flag the Procedure exit code of this BasicBlock.
 *
 * The Procedure exit code sequence includes:
 * - last control operation
 * - any copy to a dedicated register
 * - last copy, add or adjust of stack pointer
 */
void
BasicBlock_flagExitCode(BasicBlock this, Convention convention);

/*
 * BasicBlock_INIT --	Initialize BasicBlock.
 */
void
BasicBlock_INIT(void);

/*
 * BasicBlock_FINI --	Finalize BasicBlock.
 */
#ifndef ECL_NTRACE
void
BasicBlock_FINI(void);
#else//ECL_NTRACE
#define BasicBlock_FINI()
#endif//ECL_NTRACE


#endif//BasicBlock__h_INCLUDED
