#ifndef Convention_h_INCLUDED
#define Convention_h_INCLUDED

/*
 * !!!!	Convention.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * ConventionRole --	Enumerate the Convention Register roles.
 */
typedef enum {
  ConventionRole_Reserved,		// Reserved (non allocatable) Register(s).
  ConventionRole_Argument,		// Function call argument Register(s).
  ConventionRole_Result,		// Function call result Register(s).
  ConventionRole_Struct,		// Pointer to result structure Register(s).
  ConventionRole_Caller,		// Caller saved (scratch) Register(s).
  ConventionRole_Callee,		// Callee saved (preserved) Register(s).
  ConventionRole_Program,		// Program counter Register(s).
  ConventionRole_Return,		// Return address (link) Register(s).
  ConventionRole_Stack,			// Stack pointer Register(s)
  ConventionRole_Frame,			// Frame pointer Register(s).
  ConventionRole_Static,		// Static link pointer Register(s).
  ConventionRole_Global,		// Global data or GOT pointer Register(s).
  ConventionRole_Local,			// Local data or TLS pointer Register(s).
  ConventionRole_Wired,			// Wired (constant or bucket) Register(s).
  ConventionRole_Zero,			// Constant zero value Register(s).
  ConventionRole_One,			// Constant one value Register(s).
  ConventionRole__
} enum_ConventionRole;
typedef uint8_t ConventionRole;
extern const char *
ConventionRole_name_(ConventionRole this);

/*
 * Convention_name_ --	Names for the Convention enumeration.
 */
extern const char *
Convention_name_(Convention this);

/*
 * Convention_roleRegisterSet --	RegisterSet(s) indexed by ConventionRole(s).
 */
const_RegisterSet
Convention_roleRegisterSet(Convention this, ConventionRole role);

//
#define Convention_reservedRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Reserved)
#define Convention_argumentRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Argument)
#define Convention_resultRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Result)
#define Convention_structRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Struct)
#define Convention_callerRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Caller)
#define Convention_calleeRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Callee)
#define Convention_programRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Program)
#define Convention_returnRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Return)
#define Convention_stackRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Stack)
#define Convention_frameRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Frame)
#define Convention_staticRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Static)
#define Convention_globalRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Global)
#define Convention_localRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Local)
#define Convention_wiredRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Wired)
#define Convention_zeroRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_Zero)
#define Convention_oneRegisterSet(this) \
  Convention_roleRegisterSet(this, ConventionRole_One)

//
#define Convention_isReservedRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Reserved), registre)
#define Convention_isArgumentRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Argument), registre)
#define Convention_isResultRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Result), registre)
#define Convention_isStructRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Struct), registre)
#define Convention_isCallerRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Caller), registre)
#define Convention_isCalleeRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Callee), registre)
#define Convention_isProgramRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Program), registre)
#define Convention_isReturnRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Return), registre)
#define Convention_isStackRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Stack), registre)
#define Convention_isFrameRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Frame), registre)
#define Convention_isStaticRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Static), registre)
#define Convention_isGlobalRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Global), registre)
#define Convention_isLocalRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Local), registre)
#define Convention_isWiredRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Wired), registre)
#define Convention_isZeroRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_Zero), registre)
#define Convention_isOneRegister(this, registre) \
  RegisterSet_contains(Convention_roleRegisterSet(this, ConventionRole_One), registre)

/*
 * Convention_structRegister --	This Convention Struct pointer Register.
 */
Register
Convention_structRegister(Convention this);

/*
 * Convention_programRegister --	This Convention Program pointer Register.
 */
Register
Convention_programRegister(Convention this);

/*
 * Convention_returnRegister --	This Convention Return pointer Register.
 */
Register
Convention_returnRegister(Convention this);

/*
 * Convention_stackRegister --	This Convention Stack pointer Register.
 */
Register
Convention_stackRegister(Convention this);

/*
 * Convention_frameRegister --	This Convention Frame pointer Register.
 */
Register
Convention_frameRegister(Convention this);

/*
 * Convention_staticRegister --	This Convention Static pointer Register.
 */
Register
Convention_staticRegister(Convention this);

/*
 * Convention_globalRegister --	This Convention Global pointer Register.
 */
Register
Convention_globalRegister(Convention this);

/*
 * Convention_localRegister --	This Convention Local pointer Register.
 */
Register
Convention_localRegister(Convention this);

/*
 * Convention_alignText --	This Convention Text Alignment Base.
 */
unsigned
Convention_alignText(Convention this);

/*
 * Convention_alignData --	This Convention Data Alignment Base.
 */
unsigned
Convention_alignData(Convention this);

/*
 * Convention_alignHeap --	This Convention Heap Alignment Base.
 */
unsigned
Convention_alignHeap(Convention this);

/*
 * Convention_alignStack --	This Convention Stack Alignment Base.
 */
unsigned
Convention_alignStack(Convention this);

/*
 * Convention_nativeInt --	This Convention Native Int.
 */
NativeType
Convention_nativeInt(Convention this);

/*
 * Convention_nativeUInt --	This Convention Native UInt.
 */
NativeType
Convention_nativeUInt(Convention this);

/*
 * Convention_nativeFloat --	This Convention Native Float.
 */
NativeType
Convention_nativeFloat(Convention this);

/*
 * Convention_nativePtr --	This Convention Native Ptr.
 */
NativeType
Convention_nativePtr(Convention this);

//
extern RegisterSet_
Convention_allocatableSet(Convention this, RegFile regFile);

//
bool
Convention_DTD(FILE *file);

//
bool
Convention_XML(FILE *file);


#endif//Convention_h_INCLUDED
