#ifndef Instance_h_INCLUDED
#define Instance_h_INCLUDED

/*
 * !!!!	Instance.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * Instance_name_ --	Names for the Instance enumeration.
 */
extern const char *
Instance_name_(Instance this);

/*
 * Instance_scheduling --	This Instance Scheduling class.
 */
Scheduling
Instance_scheduling(Instance this);

/*
 * Instance_encoding --	This Instance Encoding.
 */
Encoding
Instance_encoding(Instance this);

//
struct InstanceOpcodes_ {
  //@args	
  uint32_t WORDS[2];
};
typedef struct InstanceOpcodes_ InstanceOpcodes_, *InstanceOpcodes;
typedef const struct InstanceOpcodes_ *const_InstanceOpcodes;
typedef struct InstanceOpcodes_ * restrict_InstanceOpcodes;

extern InstanceOpcodes
InstanceOpcodes_Ctor(InstanceOpcodes this);

extern InstanceOpcodes
InstanceOpcodes_Copy(InstanceOpcodes this, const_InstanceOpcodes that);

extern void
InstanceOpcodes_Dtor(InstanceOpcodes this);

extern size_t
InstanceOpcodes_Size(void);

#define InstanceOpcodes_WORDS(this) ((this)->WORDS)
#define InstanceOpcodes__WORDS(this) (&(this)->WORDS)


/*
 * Instance_opcodes --	This Instance opcodes.
 */
const_InstanceOpcodes
Instance_opcodes(Instance this);

/*
 * Instance_operands --	This Instance Operands.
 */
Operands
Instance_operands(Instance this);

/*
 * Instance_increment --	This Instance PC Increment.
 */
int
Instance_increment(Instance this);

/*
 * Instance_mnemonic --	This Instance mnemonic string.
 */
const char *
Instance_mnemonic(Instance this);

/*
 * Instance_syntax --	This Instance syntax string.
 */
const char *
Instance_syntax(Instance this);

/*
 * Instance_unitCost --	This Instance unit cost estimate.
 */
int
Instance_unitCost(Instance this);

/*
 * Instance_actionArray --	This Instance RegisterActionArray.
 */
const_RegisterActionArray
Instance_actionArray(Instance this);

/*
 * Instance_operator --	This Instance Operator class.
 */
Operator
Instance_operator(Instance this);

/*
 * Instance_makeRegisterActionItems --	Make this Instance RegisterActionArray_ITEMS.
 *
 * Only used to replace the MDS-initialized ITEMS with a writeable copy for patching.
 */
RegisterAction_ *
Instance_makeRegisterActionItems(Instance this, Memory memory);

//
typedef enum {
  InstanceAttribute_Control,
  InstanceAttribute_MemoryRead,
  InstanceAttribute_MemoryWrite,
  InstanceAttribute_FixRegisterRAW,
} InstanceAttribute;
typedef uint8_t InstanceAttributes;

//
#define Instance_isControl(this) \
  (Instance_attributes(this)>>InstanceAttribute_Control & 1)
#define Instance_isMemoryRead(this) \
  (Instance_attributes(this)>>InstanceAttribute_MemoryRead & 1)
#define Instance_isMemoryWrite(this) \
  (Instance_attributes(this)>>InstanceAttribute_MemoryWrite & 1)
#define Instance_isFixRegisterRAW(this) \
  (Instance_attributes(this)>>InstanceAttribute_FixRegisterRAW & 1)

/*
 * Instance_attributes --	Access this Instance attributes.
 */
InstanceAttributes
Instance_attributes(Instance this);

//
unsigned
Instance_fixupRAW(Instance this, Instance that);

//
int
Instance_latencyRAW(Instance this, Instance that, StorageCell cell);

/*
 * !!!!	st200-Instance.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */


#endif//Instance_h_INCLUDED
