#ifndef Resource_c_INCLUDED
#define Resource_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#include "MDS/MDS_.h"

#line 31 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
const int Resource__COUNT = Resource__;

#line 33 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 67 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#ifdef NSIMDIZE
void
ResourceVector_clear(ResourceVector this)
{
  int resource = 0;
  for (resource = 0; resource < Resource__; resource++) {
    ResourceVector_UNITS(this)[resource] = 0;
  }
}
#endif//NSIMDIZE

#line 92 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#ifdef NSIMDIZE
void
ResourceVector_accumulate(ResourceVector this, const_ResourceVector that)
{
  int resource = 0;
  for (resource = 0; resource < Resource__; resource++) {
    ResourceVector_UNITS(this)[resource] += ResourceVector_UNITS(that)[resource];
  }
}
#endif//NSIMDIZE

#line 127 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#ifdef NSIMDIZE
bool
ResourceVector_accept(const_ResourceVector this,
                      const_ResourceVector increase,
                      const_ResourceVector limit)
{
  int resource = 0;
  for (resource = 0; resource < Resource__; resource++) {
    int required = ResourceVector_UNITS(increase)[resource];
    int available = ResourceVector_UNITS(limit)[resource];
    int unit = ResourceVector_UNITS(this)[resource];
    int new_unit = unit + required;
    if (new_unit > available) return false;
  }
  return true;
}
#endif//NSIMDIZE

#line 173 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#ifdef NSIMDIZE
bool
ResourceVector_maxMerge(ResourceVector this, const_ResourceVector that)
{
  int resource = 0;
  bool changed = false;
  for (resource = 0; resource < Resource__; resource++) {
    int this_unit = ResourceVector_UNITS(this)[resource];
    int that_unit = ResourceVector_UNITS(that)[resource];
    if (this_unit < that_unit) {
      ResourceVector_UNITS(this)[resource] = that_unit;
      changed = true;
    }
  }
  return changed;
}
#endif//NSIMDIZE

#line 191 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 208 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
ResourceTable
ResourceTable_Ctor(ResourceTable this,
                   Processor processor, int windowSize)
{
  int horizon = windowSize+Reservation_COLUMNS_COUNT_MAX;
  int clp2Horizon = HackerU32_clp2(horizon);
  size_t vectorsSize = clp2Horizon*sizeof(ResourceVector_);
  Except_REQUIRE(windowSize >= 0 && clp2Horizon > 0);
  *ResourceTable__WINDOWSIZE(this) = windowSize;
  *ResourceTable__CLP2HORIZON(this) = clp2Horizon;
  *ResourceTable__WINDOWSTART(this) = 0;
  *ResourceTable__AVAILABILITY(this) = Processor_availability(processor);
  memset(ResourceTable_VECTORS(this), 0, vectorsSize);
  return this;
}

#line 224 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
size_t
ResourceTable_Size(Processor processor, int windowSize)
{
  int horizon = windowSize+Reservation_COLUMNS_COUNT_MAX;
  int clp2Horizon = HackerU32_clp2(horizon);
  size_t vectorsSize = clp2Horizon*sizeof(ResourceVector_);
  Except_REQUIRE(windowSize >= 0 && clp2Horizon > 0);
  return sizeof(ResourceTable_) + vectorsSize;
}

#line 238 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
void
ResourceTable_reset(ResourceTable this, int windowStart)
{
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  size_t vectorsSize = clp2Horizon*sizeof(ResourceVector_);
  *ResourceTable__WINDOWSTART(this) = windowStart;
  memset(ResourceTable_VECTORS(this), 0, vectorsSize);
}

#line 251 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
bool
ResourceTable_fitReservation(ResourceTable this, Reservation reservation, int startDate)
{
  int windowStart = ResourceTable_WINDOWSTART(this);
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  ResourceVector_ *vectors = ResourceTable_VECTORS(this);
  const_ResourceVector availability = ResourceTable_AVAILABILITY(this);
  const_ReservationTable table = Reservation_table(reservation);
  const_ResourceVector columns = ReservationTable_COLUMNS(table);	// FIXME!
  unsigned moduloDate = startDate & (clp2Horizon - 1);
  ResourceVector vector = vectors + moduloDate;
  Except_CHECK(startDate >= windowStart && startDate < windowStart + clp2Horizon);
  if (ResourceVector_accept(vector, columns, availability)) {
    ResourceVector_accumulate(vector, columns);
    return true;
  }
  return false;
}

#line 274 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
bool
ResourceTable_isFull(const_ResourceTable this, const_ResourceVector column, int startDate)
{
  int windowStart = ResourceTable_WINDOWSTART(this);
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  ResourceVector_ *vectors = ResourceTable_VECTORS(this);
  const_ResourceVector availability = ResourceTable_AVAILABILITY(this);
  unsigned moduloDate = startDate & (clp2Horizon - 1);
  ResourceVector vector = vectors + moduloDate;
  Except_CHECK(startDate >= windowStart && startDate < windowStart + clp2Horizon);
  return !ResourceVector_accept(vector, column, availability);
}

#line 291 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
int
ResourceTable_tryReservation(ResourceTable this, Reservation reservation, int startDate)
{
  int windowStart = ResourceTable_WINDOWSTART(this);
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  ResourceVector_ *vectors = ResourceTable_VECTORS(this);
  const_ResourceVector availability = ResourceTable_AVAILABILITY(this);
  const_ReservationTable table = Reservation_table(reservation);
  const_ResourceVector columns = ReservationTable_COLUMNS(table);	// FIXME!
  while (startDate < windowStart + clp2Horizon) {
    unsigned moduloDate = startDate & (clp2Horizon - 1);
    ResourceVector vector = vectors + moduloDate;
    if (ResourceVector_accept(vector, columns, availability)) break;
    startDate++;
  }
  return startDate;
}

#line 313 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
void
ResourceTable_addReservation(ResourceTable this, Reservation reservation, int startDate)
{
  int windowSize = ResourceTable_WINDOWSIZE(this);
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  ResourceVector_ *vectors = ResourceTable_VECTORS(this);
  const_ResourceVector availability = ResourceTable_AVAILABILITY(this);
  const_ReservationTable table = Reservation_table(reservation);
  const_ResourceVector columns = ReservationTable_COLUMNS(table);	// FIXME!
  unsigned moduloDate = startDate & (clp2Horizon - 1);
  // Advance windowStart to startDate - windowSize.
  ResourceTable_advance(this, startDate - windowSize);
  // Accumulate the resource uses.
  ResourceVector_accumulate(vectors + moduloDate, columns);
}

#line 333 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
void
ResourceTable_advance(ResourceTable this, int newDate)
{
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  int windowStart = ResourceTable_WINDOWSTART(this);
  ResourceVector_ *vectors = ResourceTable_VECTORS(this);
  if (newDate > windowStart) {
    // Clear the past entries of VECTORS.
    if (newDate >= windowStart + clp2Horizon) {
      size_t tableSize = clp2Horizon*sizeof(ResourceVector_);
      memset(vectors, 0, tableSize);
    } else {
      for (; windowStart < newDate; windowStart++) {
        unsigned moduloDate = windowStart & (clp2Horizon - 1);
        ResourceVector_clear(vectors + moduloDate);
      }
    }
    *ResourceTable__WINDOWSTART(this) = newDate;
  }
}

#line 358 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
bool
ResourceTable_merge(ResourceTable this, const_ResourceTable that, int delay)
{
  bool changed = false;
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  int this_windowStart = ResourceTable_WINDOWSTART(this), i;
  int that_windowStart = ResourceTable_WINDOWSTART(that) + delay;
  ResourceVector_ *this_vectors = ResourceTable_VECTORS(this);
  const ResourceVector_ *that_vectors = ResourceTable_VECTORS(that);
  Except_CHECK(clp2Horizon == ResourceTable_CLP2HORIZON(that));
  for (i = 0; i < clp2Horizon - delay; i++) {
    int this_date = this_windowStart + i;
    int that_date = that_windowStart + i;
    unsigned this_moduloDate = this_date & (clp2Horizon - 1);
    unsigned that_moduloDate = that_date & (clp2Horizon - 1);
    ResourceVector this_vector = this_vectors + this_moduloDate;
    const_ResourceVector that_vector = that_vectors + that_moduloDate;
    changed |= ResourceVector_maxMerge(this_vector, that_vector);
  }
  return changed;
}

#line 384 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#ifndef ECL_NTRACE
bool
ResourceTable_pretty(const_ResourceTable this, FILE *file)
{
  int clp2Horizon = ResourceTable_CLP2HORIZON(this);
  int windowStart = ResourceTable_WINDOWSTART(this), resource, i;
  const_ResourceVector availability = ResourceTable_AVAILABILITY(this);
  ResourceVector_ *vectors = ResourceTable_VECTORS(this);
  if (file == NULL) file = stderr;
  fprintf(file, "[%d]\t|", windowStart);
  for (i = 0; i < clp2Horizon; i++) {
    fprintf(file, "%+3d", i);
  }
  fputc('\n', file);
  fputs("--------|", file);
  for (i = 0; i < clp2Horizon; i++) fputs("---", file);
  fputc('\n', file);
  for (resource = Resource__ISSUE; resource < Resource__; resource++) {
    fprintf(file, "%s\t|", Resource_name(resource));
    for (i = 0; i < clp2Horizon; i++) {
      int date = windowStart + i;
      unsigned moduloDate = date & (clp2Horizon - 1);
      unsigned use = ResourceVector_UNITS(vectors + moduloDate)[resource];
      if (use != 0) fprintf(file, "%3d", use);
      else fputs("   ", file);
    }
    fprintf(file,"  (%d)\n", ResourceVector_UNITS(availability)[resource]);
  }
  return true;
}
#endif//ECL_NTRACE

#line 416 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 426 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#ifndef ECL_NTRACE
const char *
Resource_name_(Resource this)
{
  static const char *_[] = {
#define Resource(NAME, AVAILABILITY) \
    #NAME,
    "_UNDEF",
#include "Resource.enum"
  };
  Except_REQUIRE((unsigned)this < Resource__);
  return _[this];
}
const char *
Resource_name(Resource this)
{
  return this? Resource_name_(this) + sizeof(MDS_FAMILY): "_UNDEF";
}
#endif//ECL_NTRACE

#line 454 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 456 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
const uint8_t
Resource_Availability[][Processor__] = {
#define AVAILABILITY(items) { 1/*Processor_*/, items }
#define AVAILABLE(processor,available) /*[processor] =*/ available,
#define Resource(NAME, AVAILABILITY) \
  AVAILABILITY,
  AVAILABILITY(0),
#include "Resource.enum"
};

#line 472 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
int
Resource_availability(Resource this, Processor processor)
{
  Except_CHECK((unsigned)this < Resource__);
  Except_CHECK((unsigned)processor < Processor__);
  return Resource_Availability[this][processor];
}

#line 498 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 510 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
#if XCC__C
static void
Resource_TEST(void)
{
#include "MDS/Resource_.c"
}

#line 517 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/MDS/Resource.xcc"
int
main(int argc, char **argv)
{
  MDS_Test(Resource_TEST);
  return 0;
}
#endif


#endif//Resource_c_INCLUDED
