#ifndef Boissinot_c_INCLUDED
#define Boissinot_c_INCLUDED

#line 19 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
#include "PFA/PFA_.h"

#line 24 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 51 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
Boissinot
Boissinot_Ctor(Boissinot this,
               Memory memory, CodeRegion codeRegion, SSAForm ssaForm)
{
  Optimize optimize = CodeRegion_optimize(codeRegion);
  OptimizeFlags coalescingFlags = Optimize_COALESCING(optimize);
  TemporaryTable temporaryTable = CodeRegion_globalTable(codeRegion);
  int32_t phiCount = TemporaryTable_count(temporaryTable);
  Dominance dominance = SSAForm_DOMINANCE(ssaForm);
  Coalescer coalescer = Boissinot_COALESCER(this);
  Coalescer_Ctor(coalescer, memory, codeRegion, temporaryTable, dominance, coalescingFlags);
  if (Coalescer_CONSUMING(coalescer)) {
    CodeRegion_makeConsumers(codeRegion, temporaryTable);
  }
  OperationStack_Ctor(Boissinot_PHIOPERATIONS(this), memory, phiCount);
  Boissinot_fillPhiOperations(this);
  PFA_TRACE && fprintf(PFA_TRACE, "*** Before Boissinot CSSA\n")
            && CodeRegion_pretty(codeRegion, PFA_TRACE)
            && SSAForm_pretty(ssaForm, PFA_TRACE);
  Except_CHECK(coalescingFlags & OptimizeCoalescing_Boissinot);
  if (coalescingFlags & OptimizeCoalescing_Virtualize) {
    Boissinot_doBoissinot_III_Coalescer(this);
  } else {
    Boissinot_doBoissinot_I_Coalescer(this);
  }
  if (coalescingFlags & OptimizeCoalescing_Congruence) {
    Coalescer_coalesceCongruences(coalescer);
  }
  PFA_TRACE && (GETENV("SSA_TRACE"))
            && fprintf(PFA_TRACE, " *** After Boissinot CSSA\n")
            && CodeRegion_pretty(codeRegion, PFA_TRACE);
  return this;
}

#line 85 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
void
Boissinot_Dtor(Boissinot this)
{
  Memory memory = Boissinot_MEMORY(this);
  Coalescer_Dtor(Boissinot_COALESCER(this));
  OperationStack_Dtor(Boissinot_PHIOPERATIONS(this));
}

#line 93 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
size_t
Boissinot_Size(Memory memory, CodeRegion codeRegion, SSAForm ssaForm)
{
  return sizeof(Boissinot_);
}

#line 108 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
#ifndef Boissinot_POOLED
#define Boissinot_POOLED 128
#endif//Boissinot_POOLED

#line 112 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
Boissinot
Boissinot_new(Memory parent, CodeRegion codeRegion, SSAForm ssaForm)
{
  Memory memory = Memory_new(parent, Boissinot_POOLED);
  size_t size = Boissinot_Size(memory, codeRegion, ssaForm);
  void *alloc = Memory_alloc(memory, size);
  return Boissinot_Ctor(alloc, memory, codeRegion, ssaForm);
}

#line 125 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
Boissinot
Boissinot_delete(Boissinot this)
{
  if (this != NULL) {
    Memory memory = Boissinot_MEMORY(this);
    //#ifndef ECL_NDTOR
    Boissinot_Dtor(this);
    Memory_free_(memory, this);
    //#endif//ECL_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 139 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 145 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
int32_t
Boissinot_fillPhiOperations(Boissinot this)
{
  int32_t maxPhiArity = 0;
  CodeRegion codeRegion = Boissinot_CODEREGION(this);
  TemporaryTable temporaryTable = Boissinot_TEMPORARYTABLE(this);
  OperationStack phiOperations = Boissinot_PHIOPERATIONS(this);
  // Fill the sortedBlocks.
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      if (Operator_isPhi(Operation_operator(operation))) {
        int32_t operandCount = Operation_operandCount(operation);
        if (maxPhiArity < operandCount) maxPhiArity = operandCount;
        if (!ECL_OPTIMIZE) {
          Operation_OPERANDS_FOREACH_Temporary(operation, operand) {
            Except_ALWAYS(TemporaryTable_contains(temporaryTable, operand));
          } Operation_OPERANDS_ENDEACH_Temporary;
        }
        OperationStack_push(phiOperations, operation);
      } else break;
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
  if (PFA_TRACE) {
    BasicBlock prev_block = NULL;
    fprintf(PFA_TRACE, "*** PHI Operations\tmaxPhiArity=%d\n", maxPhiArity);
    OperationStack_FOREACH_Operation(phiOperations, operation) {
      BasicBlock basicBlock = Operation_basicBlock(operation);
      if (basicBlock != prev_block) {
        fprintf(PFA_TRACE, "    Block_%d\tfrequency=%g\n",
                BasicBlock_index(basicBlock), BasicBlock_frequency(basicBlock));
        prev_block = basicBlock;
      }
      Operation_pretty(operation, PFA_TRACE);
    } OperationStack_ENDEACH_Operation;
  }
  return maxPhiArity;
}

#line 183 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 186 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
typedef struct PhiBlocks {
  BasicBlock basicBlock;
  int32_t i;
} PhiBlocks;

#line 191 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
static inline int
comparePhiBlocks(const void *a, const void *b) {
  uint32_t freq_a = BasicBlock_frequency(((PhiBlocks *) a)->basicBlock);
  uint32_t freq_b = BasicBlock_frequency(((PhiBlocks *) b)->basicBlock);
  return (freq_a > freq_b) - (freq_a < freq_b);
}

#line 208 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
void
Boissinot_doBoissinot_III(Boissinot this)
{
  CodeRegion codeRegion = Boissinot_CODEREGION(this);
  Memory memory = Boissinot_MEMORY(this);
  Coalescer coalescer = Boissinot_COALESCER(this);
  LiveCheck livecheck = Coalescer_LIVECHECK(coalescer);
  IBList congruences = Boissinot_CONGRUENCES(this);
  //
  CodeRegion_makeConsumers(codeRegion, Boissinot_TEMPORARYTABLE(this));
  //foreach phi: B0:a0=phi(B1:a1,...,Bn:an)
  CodeRegion_FOREACH_BasicBlock(codeRegion, B0) {
    1 && PFA_TRACE && fprintf(PFA_TRACE, "Boissinot_III treating BB %d\n", BasicBlock_INDEX(B0));
    BasicBlock_FOREACH_Operation(B0, phi) {
      if (!Operator_isPhi(Operation_operator(phi))) break;
      1 && PFA_TRACE && fprintf(PFA_TRACE, "\tBoissinot_III treating ")
              && Operation_pretty(phi, PFA_TRACE) && fprintf(PFA_TRACE,"\n");

#line 226 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
      //congruence class of the phi variable (variables merged with the phi) initially empty
      Except_CHECK(Operation_COLOR(phi)==NULL);
      //foreach Bi in order of decreasing frequency
      {
        int32_t n=Operation_operandCount(phi);
        int32_t i, k;
        PhiBlocks BBList[n];
        BBList[0].basicBlock=B0;
        BBList[0].i=0;
        for (i=1 ; i<n ; i++) {
          BBList[i].basicBlock=BasicBlockEdge_origBlock(BasicBlock_getPhiEnterEdge(B0, i-1));
          BBList[i].i=i;
        }
        qsort(BBList, n, sizeof(PhiBlocks), comparePhiBlocks);
        for (k=n-1 ; k>=0 ; k--) {
          bool merge = true;
          int32_t i = BBList[k].i;
          //let ai be the i-th argument
          Variable ai = Temporary_VARIABLE(Operation_OPERANDS(phi)[i]);
          VariableClass congruence_ai = Variable_CONGRUENCE(ai);
          PFA_TRACE && fprintf(PFA_TRACE,"\t\ttreating argument %d",i)
                    && fprintf(PFA_TRACE,", ai=") && Variable_pretty(ai, PFA_TRACE, 1)
                    && fprintf(PFA_TRACE,", Ai=") &&
                    VariableClass_pretty(Variable_CONGRUENCE(ai), PFA_TRACE, false)
                    && Operation_COLOR(phi) && fprintf(PFA_TRACE,", PHIVAR=")
                    && VariableClass_pretty(Operation_COLOR(phi), PFA_TRACE, false)
                    && fprintf(PFA_TRACE, "\n");
          //if class(ai) interfere phi
          //then goto split
          if (Coalescer_interferesCP(coalescer, congruence_ai, phi)) {
            PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tAi interferes with phi!\n");
            merge = false;
          }
          //if class(ai) interfere phimerged
          //then goto split
#if 1
          else {
            VariableClass color = Operation_COLOR(phi);
            if (color != NULL) {
              if (color != congruence_ai) {
                if (Coalescer_interferesCC(coalescer, congruence_ai, color)) {
                  PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tAi interferes with PHIVAR!\n");
                } else {
                  PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tno interference. Merge.!\n");
                  color = Coalescer_mergeCongruencesCC(coalescer, congruence_ai, color);
                }
              } else {
                PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tno interference. Merge.!\n");
              }
            } else {
              PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tno interference. Merge.!\n");
              color = congruence_ai;
            }
            *Operation__COLOR(phi) = color;
          }
#else
          if (   merge && Operation_COLOR(phi) != NULL
              && Coalescer_interferesCC(coalescer, congruence_ai, Operation_COLOR(phi))) {
 	    PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tAi interferes with PHIVAR!\n");
            merge = false;
          }
          //merge class(ai) & phimerged
          if (merge) {
            VariableClass color = Operation_COLOR(phi);
 	    PFA_TRACE && fprintf(PFA_TRACE, "\t\t\tno interference. Merge.!\n");
            if (color == NULL) color = congruence_ai;
            else if (color != congruence_ai)
              color =  Coalescer_mergeCongruencesCC(coalescer, congruence_ai, color);
            *Operation__COLOR(phi) = color;
          }
#endif
        // endforeach Bi
        }
      }
      Coalescer_explicitPhi(coalescer, phi);
      Operation_setFlags(phi, OperationFlag_Processed);
      // endforeach phi
    }  BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock ;
}

#line 313 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
/*
 * Boissinot_doBoissinot_I --	Main processing of Boissinot CSSA method
 *
 * Require:  empty CONGRUENCES
 * Output: Generated move operations around PHIs
 *         All newly created variables put in the same phi congruence class
 */
static void
Boissinot_doBoissinot_I(Boissinot this);

#line 323 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
static void
Boissinot_doBoissinot_I(Boissinot this)
{
  Memory memory = Boissinot_MEMORY(this);
  SSAForm ssaForm = Boissinot_SSAFORM(this);
  Coalescer coalescer = Boissinot_COALESCER(this);
  IBList congruences = Boissinot_CONGRUENCES(this);
  CodeRegion codeRegion = Boissinot_CODEREGION(this);
  OperationStack phiOperations = Boissinot_PHIOPERATIONS(this);
  TemporaryTable temporaryTable = Boissinot_TEMPORARYTABLE(this);
  //
  // Step 1. Insert copies for each phi operand (result and arguments)
  // and merge all newly created variables in the same congruence class
  OperationStack_FOREACH_Operation(phiOperations, phi_operation) {
    int resIndex = 0;
    Operation copy_operation = NULL;
    BasicBlock block_0 = Operation_basicBlock(phi_operation);
    Temporary resource_0 = Operation_getResult(phi_operation, 0);
    Variable variable_0 = Temporary_VARIABLE(resource_0);
    Temporary origin = Variable_ORIGIN(variable_0);
    Temporary new_resource_0 = CodeRegion_makeVariableTemporary(codeRegion, origin);
    Variable new_variable_0 = Temporary_VARIABLE(new_resource_0);
    VariableClass congruence_0 = IBList_push(congruences);
    VariableClass_Ctor(congruence_0, new_variable_0);
    TemporaryTable_insert(temporaryTable, new_resource_0);
    // Insert a copy x0 = xnew_0 in the PHI block.
    if (!Coalescer_SEQCOPIES) {
      copy_operation = BasicBlock_makeOperationECOPY(block_0);
      resIndex = Operation_addtoParCopy(copy_operation, resource_0, new_resource_0);
    } else {
      copy_operation = BasicBlock_makeEarlyOperationCOPY(block_0, resource_0, new_resource_0);
    }
    // Replace x0 with xnew_0 as the target.
    // Add xnew_0 in phiCongruenceClass of xnew_0.
    Variable_setOperationResult(variable_0, copy_operation, resIndex);
    Variable_updateCongruence(variable_0);
    Variable_setOperationArgument(new_variable_0, copy_operation, resIndex);
    Variable_setOperationResult(new_variable_0, phi_operation, 0);
    if (Coalescer_CONSUMING(coalescer)) Variable_raiseConsuming(new_variable_0);
    //
    Operation_ARGUMENTS_FOREACH_Temporary(phi_operation, resource_i) {
      Variable variable_i = Temporary_VARIABLE(resource_i);
      Temporary origin = Variable_ORIGIN(variable_i);
      BasicBlockEdge edge_k = BasicBlock_getPhiEnterEdge(block_0, Operation_ARGUMENTS_INDEX);
      BasicBlock block_k = BasicBlockEdge_origBlock(edge_k);
      Temporary new_resource_i = CodeRegion_makeVariableTemporary(codeRegion, origin);
      Variable new_variable_i = Temporary_VARIABLE(new_resource_i);
      TemporaryTable_insert(temporaryTable, new_resource_i);
      // Insert a copy x_newi = xi in the corresponding block.
      if (!Coalescer_SEQCOPIES) {
        copy_operation = BasicBlock_makeOperationLCOPY(block_k);
        resIndex = Operation_addtoParCopy(copy_operation, new_resource_i, resource_i);
      } else {
        copy_operation = BasicBlock_makeLateOperationCOPY(block_k, new_resource_i, resource_i);
      }
      Variable_setOperationArgument(variable_i, copy_operation, resIndex);
      Variable_setOperationArgument(new_variable_i, phi_operation, Operation_ARGUMENTS_INDEX);
      Variable_setOperationResult(new_variable_i, copy_operation, resIndex);
      VariableClass_insert(congruence_0, new_variable_i);
      if (Coalescer_CONSUMING(coalescer)) Variable_raiseConsuming(new_variable_i);
    } Operation_ARGUMENTS_ENDEACH_Temporary;
  } OperationStack_ENDEACH_Operation;
  //
  if (PFA_TRACE) {
    TemporaryTable temporaryTable = Boissinot_TEMPORARYTABLE(this);
    fprintf(PFA_TRACE, "*** SSA Congruences\n");
    TemporaryTable_FOREACH_Temporary(temporaryTable, temporary) {
      Variable variable = Temporary_VARIABLE(temporary);
      VariableClass phiCongruence = Variable_CONGRUENCE(variable);
      PFA_TRACE && Temporary_pretty(temporary, PFA_TRACE)
        && fprintf(PFA_TRACE, " => ")
        && VariableClass_pretty(phiCongruence, PFA_TRACE, false)
        && fprintf(PFA_TRACE, "\n");
    } TemporaryTable_ENDEACH_Temporary;
  }
}

#line 406 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
void
Boissinot_doBoissinot_III_Coalescer(Boissinot this)
{
  Coalescer coalescer = Boissinot_COALESCER(this);
  Coalescer_complete(coalescer);
  Boissinot_doBoissinot_III(this);
  if (PFA_TRACE) {
    Memory memory = Boissinot_MEMORY(this);
    CodeRegion codeRegion = Boissinot_CODEREGION(this);
    TemporaryTable temporaryTable = Boissinot_TEMPORARYTABLE(this);
    Liveness liveness = Liveness_make(memory, codeRegion, temporaryTable, true);
    fprintf(PFA_TRACE, " *** After Boissinot III\n");
    CodeRegion_pretty(Boissinot_CODEREGION(this), PFA_TRACE);
    Liveness_kill(liveness);
    fprintf(PFA_TRACE, "*** SSA Congruences\n");
    TemporaryTable_FOREACH_Temporary(temporaryTable, temporary) {
      Variable variable = Temporary_VARIABLE(temporary);
      VariableClass phiCongruence = Variable_CONGRUENCE(variable);
      PFA_TRACE && Temporary_pretty(temporary, PFA_TRACE)
        && fprintf(PFA_TRACE, " => ")
        && VariableClass_pretty(phiCongruence, PFA_TRACE, false)
        && fprintf(PFA_TRACE, "\n");
    } TemporaryTable_ENDEACH_Temporary;
  }
}

#line 432 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
void
Boissinot_doBoissinot_I_Coalescer(Boissinot this)
{
  String name = Procedure_name(CodeRegion_procedure(Boissinot_CODEREGION(this)));
  //Coalescer build in OptimizeCoalescing_Dominance
  PFA_TRACE && fprintf(PFA_TRACE, "*** Exact Method:\n")
            && fprintf(PFA_TRACE, "*** Exact COPY insertion:\n");
  // Step 1. Insert copies for each phi operand (result and arguments)
  Boissinot_doBoissinot_I(this);
  PFA_TRACE && fprintf(PFA_TRACE, " *** After Exact Step I\n")
            && CodeRegion_pretty(Boissinot_CODEREGION(this), PFA_TRACE);
#ifndef ECL_NTRACE
  PFA_STATS && SSAForm_dumpAppelGraph(Boissinot_SSAFORM(this), PFA_STATS, name);
#endif
  // Step 2.Interference Graph building
  PFA_TRACE && fprintf(PFA_TRACE, "*** Exact Interference Build:\n");
  {
    Coalescer coalescer = Boissinot_COALESCER(this);
    Coalescer_complete(coalescer);
    if (!ECL_OPTIMIZE) {
      VariableClass class = Coalescer_findIntersectingClass(coalescer);
      Except_CHECK(class == NULL);
    }
    if (PFA_TRACE && Coalescer_INTERFERENCE(coalescer)) {
      Interference interference = Coalescer_INTERFERENCE(coalescer);
      TemporaryTable temporaryTable = Coalescer_TEMPORARYTABLE(coalescer);
      fprintf(PFA_TRACE, " *** After Exact Step 2\n");
      TemporaryTable_FOREACH_Temporary(temporaryTable, temporary_1) {
        TemporaryTable_FOREACH_Temporary(temporaryTable, temporary_2) {
          if (Interference_exists(interference, temporary_1, temporary_2)) {
            if (Temporary_IDENTITY(temporary_1) > Temporary_IDENTITY(temporary_2)) {
              fprintf(PFA_TRACE, "Intersection ");
              if (   Coalescer_EQUIVALUE > 0 &&
                         Variable_EQUIVALUE(Temporary_VARIABLE(temporary_1))
                      == Variable_EQUIVALUE(Temporary_VARIABLE(temporary_2))) {
                fprintf(PFA_TRACE, "But Same Value ");
              }
              Temporary_pretty(temporary_1, PFA_TRACE)
                && fprintf(PFA_TRACE, " ")
                && Temporary_pretty(temporary_2, PFA_TRACE)
                && fprintf(PFA_TRACE, "\n");
            }
          }
        } TemporaryTable_ENDEACH_Temporary;
      } TemporaryTable_ENDEACH_Temporary;
    }
  }
}

#line 481 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1006 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1018 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
#if XCC__C
static void
Boissinot_TEST(void)
{
#include "PFA/Boissinot_.c"
}

#line 1025 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Boissinot.xcc"
int
main(int argc, char **argv)
{
  PFA_Test(Boissinot_TEST);
  return 0;
}
#endif


#endif//Boissinot_c_INCLUDED
