#ifndef Boissinot__h_INCLUDED
#define Boissinot__h_INCLUDED

//
#include "PFA/Boissinot_.h"

//
#define Boissinot_TIMER_T1 0
#define Boissinot_TIMER_T2 1
#define Boissinot_TIMER_T3 2
#define Boissinot_TIMER_COUNT 3

//
struct Boissinot_ {
  //@args       Memory memory, CodeRegion codeRegion, SSAForm ssaForm
  Coalescer_ COALESCER_;     		// SSA Coalescer.
  //@access COALESCER Boissinot__COALESCER_(this)
  //@access CODEREGION		Coalescer_CODEREGION(Boissinot_COALESCER(this))
  //@access TEMPORARYTABLE	Coalescer_TEMPORARYTABLE(Boissinot_COALESCER(this))
  //@access CONGRUENCES		Coalescer_CONGRUENCES(Boissinot_COALESCER(this))
  //@access SSAFORM		CodeRegion_ssaForm(Boissinot_CODEREGION(this))
  OperationStack_ PHIOPERATIONS_;	// Phi-Operations to process.
  //@access PHIOPERATIONS	Boissinot__PHIOPERATIONS_(this)
  //@access MEMORY      OperationStack_memory(Boissinot_PHIOPERATIONS(this))
};

extern Boissinot
Boissinot_Ctor(Boissinot this, Memory memory, CodeRegion codeRegion, SSAForm ssaForm);

extern Boissinot
Boissinot_Copy(Boissinot this, const_Boissinot that);

extern void
Boissinot_Dtor(Boissinot this);

extern size_t
Boissinot_Size(Memory memory, CodeRegion codeRegion, SSAForm ssaForm);

#define Boissinot_COALESCER_(this) ((this)->COALESCER_)
#define Boissinot__COALESCER_(this) (&(this)->COALESCER_)
#define Boissinot_PHIOPERATIONS_(this) ((this)->PHIOPERATIONS_)
#define Boissinot__PHIOPERATIONS_(this) (&(this)->PHIOPERATIONS_)
#define Boissinot_COALESCER(this) (Boissinot__COALESCER_(this))
#define Boissinot_CONGRUENCES(this) (Coalescer_CONGRUENCES(Boissinot_COALESCER(this)))
#define Boissinot_CODEREGION(this) (Coalescer_CODEREGION(Boissinot_COALESCER(this)))
#define Boissinot_TEMPORARYTABLE(this) (Coalescer_TEMPORARYTABLE(Boissinot_COALESCER(this)))
#define Boissinot_MEMORY(this) (OperationStack_memory(Boissinot_PHIOPERATIONS(this)))
#define Boissinot_SSAFORM(this) (CodeRegion_ssaForm(Boissinot_CODEREGION(this)))
#define Boissinot_PHIOPERATIONS(this) (Boissinot__PHIOPERATIONS_(this))


//
int32_t
Boissinot_fillPhiOperations(Boissinot this);

/*
 * Boissinot_doBoissinot_I_Coalescer --	Main processing of Boissinot CSSA method
 */
void
Boissinot_doBoissinot_I_Coalescer(Boissinot this);

/*
 * Boissinot_doBoissinot_III_Coalescer --
 */
void
Boissinot_doBoissinot_III_Coalescer(Boissinot this);

/*
 * Boissinot_INIT --	Initialize the Boissinot module.
 */
#define Boissinot_INIT()

/*
 * Boissinot_FINI --	Finalize the Boissinot module.
 */
#define Boissinot_FINI()


#endif//Boissinot__h_INCLUDED
