#ifndef Decoalescer_c_INCLUDED
#define Decoalescer_c_INCLUDED

#line 18 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
#include "PFA/PFA_.h"

#line 24 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 38 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Decoalescer
Decoalescer_Ctor(Decoalescer this,
                 Memory memory, CodeRegion codeRegion, SSAForm ssaForm)
{
  Optimize optimize = CodeRegion_optimize(codeRegion);
  OptimizeFlags coalescingFlags = Optimize_COALESCING(optimize);
  TemporaryTable temporaryTable = CodeRegion_globalTable(codeRegion);
  int32_t phiCount = TemporaryTable_count(temporaryTable);
  Dominance dominance = SSAForm_DOMINANCE(ssaForm);
  Coalescer coalescer = Decoalescer_COALESCER(this);
  Coalescer_Ctor(coalescer, memory, codeRegion, temporaryTable, dominance, coalescingFlags);
  if (Coalescer_CONSUMING(coalescer)) {
    CodeRegion_makeConsumers(codeRegion, temporaryTable);
  }
  Coalescer_complete(coalescer);
  if (!ECL_OPTIMIZE) {
    LiveCheck_checkLiveness(Coalescer_LIVECHECK(coalescer));
  }
  PFA_TRACE && fprintf(PFA_TRACE, "*** Before Decoalescer\n")
            && CodeRegion_pretty(codeRegion, PFA_TRACE)
            && SSAForm_pretty(ssaForm, PFA_TRACE);
  // TBD: Decoalescer_Budimlic(this, codeRegion);
  // FdF 20071116: While Budimlic is not implemented, just create SSA webs.
  // FdF 20090218: Apply minimum filtering on strong interferences.
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operator operator = Operation_operator(operation);
      if (Operator_isPhi(operator) || Operator_isSeqCopy(operator) || Operator_isParCopy(operator)) {
	Decoalescer_coalesceFiltering(this, operation, Decoalescer_filteringStrongInterference);
      }
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operator operator = Operation_operator(operation);
      if (Operator_isPhi(operator)) {
	Temporary *operands = Operation_operands(operation);
	*Operation__COLOR(operation) = Variable_CONGRUENCE(Temporary_VARIABLE(operands[0]));
      } else break;
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
  // Use Decoalescing algorithm.
  Decoalescer_decoalesceWebs(this);
  Decoalescer_generateCopies(this);
  if (coalescingFlags & OptimizeCoalescing_Congruence) {
    Coalescer_coalesceCongruences(coalescer);
  } else {
    Coalescer_markCoalescedCopies(coalescer);
  }
  PFA_TRACE && (GETENV("SSA_TRACE"))
            && fprintf(PFA_TRACE, " *** After Decoalescer CSSA\n")
            && CodeRegion_pretty(codeRegion, PFA_TRACE);
  return this;
}

#line 93 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
void
Decoalescer_Dtor(Decoalescer this)
{
  Memory memory = Decoalescer_MEMORY(this);
  Coalescer_Dtor(Decoalescer_COALESCER(this));
}

#line 100 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
size_t
Decoalescer_Size(Memory memory, CodeRegion codeRegion, SSAForm ssaForm)
{
  return sizeof(Decoalescer_);
}

#line 115 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
#ifndef Decoalescer_POOLED
#define Decoalescer_POOLED 128
#endif//Decoalescer_POOLED

#line 119 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Decoalescer
Decoalescer_new(Memory parent, CodeRegion codeRegion, SSAForm ssaForm)
{
  Memory memory = Memory_new(parent, Decoalescer_POOLED);
  size_t size = Decoalescer_Size(memory, codeRegion, ssaForm);
  void *alloc = Memory_alloc(memory, size);
  return Decoalescer_Ctor(alloc, memory, codeRegion, ssaForm);
}

#line 132 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Decoalescer
Decoalescer_delete(Decoalescer this)
{
  if (this != NULL) {
    Memory memory = Decoalescer_MEMORY(this);
#ifndef ECL_NDTOR
    Decoalescer_Dtor(this);
    Memory_free_(memory, this);
#endif//ECL_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 146 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 152 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
static VariableClass
Variable_color(Variable V)
{
  return Variable_CONGRUENCE(V);
}

#line 159 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
/*
weigth(op_t Phi)
  let Phi: a0=phi(a1,...,an)
  let L0 be the bbloc of Phi
  let Li be the i-th predecessor of L0
  res:=0
  for i=0:n
       if col(ai)!=col(Phi) then res+=frequency(Li)
  endfor
  return res
*/

#line 171 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
static float
Operation_WEIGHT(Operation phi)
{
  float weight = 0;
  VariableClass colPhi= Operation_COLOR(phi);
  if (colPhi != NULL) {
    Temporary result = Operation_getResult(phi, 0);
    Variable a0 = Temporary_VARIABLE(result);
    VariableClass colArg0 = Variable_color(a0);
    BasicBlock phi_block = Operation_basicBlock(phi);
    if (colArg0 == colPhi) {
      weight += BasicBlock_FREQUENCY(phi_block);
    }
    // foreach Lj, j-th predecessor of L
    Operation_ARGUMENTS_FOREACH_Temporary(phi, argument) {
      Variable aj = Temporary_VARIABLE(argument);
      BasicBlockEdge edge_j = BasicBlock_getPhiEnterEdge(phi_block, Operation_ARGUMENTS_INDEX);
      BasicBlock aj_block = BasicBlockEdge_origBlock(edge_j);
      VariableClass colArgj = Variable_color(aj);
      if (colArgj == colPhi) {
        weight += BasicBlock_FREQUENCY(aj_block);
      }
    } Operation_ARGUMENTS_ENDEACH_Temporary;
  }
  return weight;
}

#line 198 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
/* Variable_Init_Weight()
  foreach_variable(V)
    w(V) := 0
  foreach_phi(Phi)
    let Phi: a0=phi(a1,...,an)
    let L0 be the bbloc of Phi
    let Li be the i-th predecessor of L0
    if col(ai)==col(Phi)
      w(ai) = w(ai) + frequency(Li)
*/

#line 209 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
static void
Variable_Init_Weight (CodeRegion codeRegion) {

#line 212 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
  CodeRegion_FOREACH_BasicBlock(codeRegion, block) {
    BasicBlock_FOREACH_Operation(block, operation) {

#line 215 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
      VariableClass colPhi;
      Temporary result;
      Variable a0;
      VariableClass colArg0;

#line 220 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
      if (!Operator_isPhi(Operation_operator(operation)))
        break;

#line 223 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
      colPhi= Operation_COLOR(operation);
      result = Operation_getResult(operation, 0);
      a0 = Temporary_VARIABLE(result);
      colArg0 = Variable_color(a0);

#line 228 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
      if (colArg0 == colPhi)
        *Variable__WEIGHT(a0) += BasicBlock_FREQUENCY(block);

#line 231 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
      // foreach Lj, j-th predecessor of L
      Operation_ARGUMENTS_FOREACH_Temporary(operation, argument) {
        Variable aj = Temporary_VARIABLE(argument);
        BasicBlockEdge edge_j = BasicBlock_getPhiEnterEdge(block, Operation_ARGUMENTS_INDEX);
        BasicBlock aj_block = BasicBlockEdge_origBlock(edge_j);
        VariableClass colArgj = Variable_color(aj);

#line 238 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
        if (colArgj == colPhi)
          *Variable__WEIGHT(aj) += BasicBlock_FREQUENCY(aj_block);

#line 241 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
      } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 243 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock
}

#line 247 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
/* Variable_Update_Weight(op_t Phi)
  let Phi: a0=phi(a1,...,an)
  let L0 be the bbloc of Phi
  let Li be the i-th predecessor of L0
  for i=0:n
    if col(ai)==col(Phi)
    then w(ai) = w(ai)-frequency(Li)
    endif
  endfor
*/

#line 258 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
static void
Variable_Update_Weight(Operation phi) {

#line 261 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
  VariableClass colPhi= Operation_COLOR(phi);
  Temporary result = Operation_getResult(phi, 0);
  Variable a0 = Temporary_VARIABLE(result);
  VariableClass colArg0 = Variable_color(a0);
  BasicBlock phi_block = Operation_basicBlock(phi);

#line 267 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
  if (colArg0 == colPhi)
    *Variable__WEIGHT(a0) -= BasicBlock_FREQUENCY(phi_block);

#line 270 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
  // foreach Lj, j-th predecessor of L
  Operation_ARGUMENTS_FOREACH_Temporary(phi, argument) {
    Variable aj = Temporary_VARIABLE(argument);
    BasicBlockEdge edge_j = BasicBlock_getPhiEnterEdge(phi_block, Operation_ARGUMENTS_INDEX);
    BasicBlock aj_block = BasicBlockEdge_origBlock(edge_j);
    VariableClass colArgj = Variable_color(aj);

#line 277 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
    if (colArgj == colPhi)
      *Variable__WEIGHT(aj) -= BasicBlock_FREQUENCY(aj_block);

#line 280 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
  } Operation_ARGUMENTS_ENDEACH_Temporary;
}

#line 284 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 293 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
bool
Decoalescer_filteringStrongInterference(Variable variable_0, VariableClass congruence_0,
				       Variable variable_1, VariableClass congruence_1)
{
  // Just check for strong interferences.
  return !VariableClass_hasStrongInterference(congruence_0, congruence_1);
}

#line 308 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
void
Decoalescer_coalesceFiltering(Decoalescer this, Operation operation, CoalescerFilter filter)
{
  Operator operator = Operation_operator(operation);
  Coalescer coalescer = Decoalescer_COALESCER(this);
  if (Operator_isPhi(operator)) {
    Coalescer_mergePhiCongruences(coalescer, operation, filter);
  }
  else if (Operator_isSeqCopy(operator) || Operator_isParCopy(operator)) {
    Coalescer_mergeCopyCongruences(coalescer, operation, filter);
  }
}

#line 328 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
/* Pseudo Code from F.Rastello.

#line 330 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Repair D-Forest:
================

#line 333 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Goal:
----
The aim of this phase is to traverse the dominance forest (notice that
the dominance forest does not have to be explicitely built) so as to
check that within each congruence class (represented by a color) there
remains no interference. If there remains any then one of the
interfering variable is removed from the congruence class (uncolored).

#line 341 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Notations:
---------
- type var_t: the type of a variable
- set VARS: set of variables
- type col_t: the type of a color. colorless (bottom) is represented by NULL
- set COLORS: set of congruence classes (colors)
- col_t col(VAR): to each variable V a color col(V)
- var_t cur_dom(COLORS): for each color C cur_dom(C) represents
  the variable of color C that dominates the current operation. Can be equal to NULL.
- var_r idom(VAR): for each variable V idom(V) represents the variable
  of color col(V) which definition immediatly dominates the definition
  of V (NULL if none exists).
- uint w(VAR): to each variable V a weight w(V) that relates the cost of
uncoloring it (the number of uses roughly gives the number of moves)

#line 356 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
Color of a Variable:
-------------------
- Variables have been colored during the coalescing phase. The dforest
phase is not allowed to give another color to a variable other than NULL.
- We need to define properly what is the color of a Phi. The color of
a Phi is the color (different than NULL) of its operands.

#line 363 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
colt_t col(op_t Phi)
  let Phi: a0=phi(a1,...,an)
  colPhi:=NULL
  for i:=1:n
    if col(ai)!=NULL
      colPhi:=col(ai)
      break
    endif
  endfor
  return colPhi
  

#line 375 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
uncoloring variables that interfere:
------------------------------------
 dforest:
 { Initialisation }
 foreach c in COLORS cur_dom(c) := NULL
 
 foreach L in BBLOCS in DFS pre-order of the dominance tree
   { Traversal of all definitions in L }
   foreach Op in Operations(L) in topological order (including phis)
     foreach V in DefinedVariables(Op)
       if ( col(c:=col(V))==NULL ) then next { V }

#line 387 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
       { update cur_dom if necessary }
       U := cur_dom(c)
       while ( (U != NULL) && (!dominates(U,V) || col(U)==NULL) )
         U := idom (U)
       endwhile
       cur_dom(c):=U

#line 394 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
       { See if U or V should be removed; update cur_dom & idom if necessary }
       if U==NULL || !interfere(U,V)
         { No interference: none removed; update cur_dom & idom}
         idom(V):=U
         cur_dom(c):=V
       elseif (w(U) >= w(V))
         { Interference & V is preferable for removal: remove V }
         col(V):=NULL
       else
         { Interference & U is preferable for removal: remove U; update cur_dom & idom }
         col(U):=NULL
         cur_dom(c):=V
         idom(V):=idom(U)
       endif
     endforeach { V }
   endforeach { OP }

#line 411 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
   { Initialize cur_phi }
   foreach c in COLORS cur_phi(c) := NULL

#line 414 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
   { Traversal of all implicit definitions related to phi in successor blocs }
   foreach L' successor of L
     let j such that L is the j-th predecessor of L'

#line 418 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
     foreach Phi: a0=phi(a1,...,an) in Operations(L')
       let aj be the operand coming from L
       colPhi:=col(Phi)
       if colPhi==NULL next

#line 423 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
       { update cur_dom if necessary }
       U := cur_dom(colPhi)
       while ( (U != NULL) && (!dominates(bbloc(def(U)),L) || col(U)==NULL) )
         U := idom (U)
       endwhile
       cur_dom(colPhi):=U

#line 430 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
       { Check if Phi interferes with its related cur_dom U }
       if U != NULL && U in liveOut(L) && U != aj
         { U interfere with Phi }
         if w(U)>=w(Phi)
           { Interference & Phi is preferable for removal }
           col(Phi):=NULL
           next
         else
           { Interference & U is preferable for removal }
           col(U):=NULL
         endif
       endif

#line 443 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
       { Check if Phi interferes with another Phi }
       let Phi'=cur_phi(colPhi)
       if (Phi' != NULL)
         if (L is a predecessor of bbloc(Phi') )
           let a'k := the operand of Phi' coming from L
         else
           cur_phi(colPhi) := NULL;
           Phi' := NULL
         endif
       endif
       if (Phi' != NULL && a'k != aj)
         { Interfere; isolate Phi }
         col(Phi):=NULL
       else { update cur_phi }
         cur_phi(colPhi) := Phi
       endif
     endforeach
   endforeach
 endforeach
*/

#line 464 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
void
Decoalescer_decoalesceWebs(Decoalescer this)
{
  Memory memory = Decoalescer_MEMORY(this);
  SSAForm ssaForm = Decoalescer_SSAFORM(this);
  Dominance dominance = SSAForm_DOMINANCE(ssaForm);
  Coalescer coalescer = Decoalescer_COALESCER(this);
  LiveCheck liveCheck = Coalescer_LIVECHECK(coalescer);
  TemporarySet_ *liveOutSets = Coalescer_LIVEOUTSETS(coalescer);
  //
  // foreach L in BBLOCS in DFS pre-order of the dominance tree
  Variable_Init_Weight(Decoalescer_CODEREGION(this));
  Dominance_FORDOMDFS_BasicBlock(dominance, L) {
    // { Traversal of all definitions in L }
    // foreach Op in Operations(L) in topological order (including phis)
    PFA_TRACE && fprintf(PFA_TRACE, "Decoalescing: Traversing BB %d\n", BasicBlock_INDEX(L));
    BasicBlock_FOREACH_Operation(L, operation) {
      // foreach V in DefinedVariables(Op)
      Operation_RESULTS_FOREACH_Temporary(operation, result) {
        Variable V;
        Variable U = NULL;
        VariableClass c;
        if (!Temporary_isVariable(result)) continue;
        V = Temporary_VARIABLE(result);
        PFA_TRACE && fprintf(PFA_TRACE, "Testing interference of ")
                  && Variable_pretty(V, PFA_TRACE, false)
                  && fprintf(PFA_TRACE,"\n");
        // if ( col(c:=col(V))==NULL ) then next { V }
        c = Variable_color(V);
        if (c == NULL) continue;
        PFA_TRACE && fprintf(PFA_TRACE, "... in ")
                  && VariableClass_pretty(c, PFA_TRACE, false)
                  && fprintf(PFA_TRACE, "\n");
        // { update cur_dom if necessary }
        // U := cur_dom(c)
        U = VariableClass_CURDOM(c);
        PFA_TRACE && fprintf(PFA_TRACE, "... with curdom  ");
        // while ( (U != NULL) && (!dominates(U,V) || col(U)==NULL) )
        while (   (U != NULL)
               && (   (Variable_color(U) == NULL)
                   || !Coalescer_dominatesVV(coalescer, U, V))) {
          // U := idom (U)
          PFA_TRACE && Variable_pretty(U, PFA_TRACE, false)
                    && fprintf(PFA_TRACE, " -> ");
          U = Variable_IDOM(U);
        } // endwhile
        PFA_TRACE && (U ? Variable_pretty(U, PFA_TRACE, false) : fprintf(PFA_TRACE, "NULL"))
                       && fprintf(PFA_TRACE,"\n");
        //cur_dom(c):=U
        *VariableClass__CURDOM(c) = U;
        //
        // { See if U or V should be removed; update cur_dom & idom if necessary }
        // if U==NULL || !interfere(U,V)
        if (   (U == NULL)
            || !Coalescer_interferesVV(coalescer, U, V)) {
          // { No interference: none removed; update cur_dom & idom}
          // idom(V):=U
          *Variable__IDOM(V) = U;
          // cur_dom(c):=V
          *VariableClass__CURDOM(c) = V;
        }
        else {
          bool isolateV;
          // See which one of U or V is preferable to isolate
          Except_CHECK( (Variable_PINNED(U) == Register__UNDEF) ||
                        (Variable_PINNED(V) == Register__UNDEF)    );
          // if V one is pinned, isolate the other. Otherwise make the choice using weight
          isolateV = ( ( Variable_PINNED(U) != Register__UNDEF ) ||
                       ( ( Variable_PINNED(V) == Register__UNDEF ) &&
                         ( Variable_WEIGHT(U) >= Variable_WEIGHT(V) ) ) );
          if (isolateV) {
            // { Interference & V is preferable for removal: remove V }
            // col(V):=NULL
            VariableClass_remove(Variable_CONGRUENCE(V), V);
            PFA_TRACE && fprintf(PFA_TRACE, "Decoalescer_isolateVariable for variable: ");
            PFA_TRACE && Temporary_pretty(Variable_TEMPORARY(V), PFA_TRACE);
            PFA_TRACE && fprintf(PFA_TRACE, "\n");
          }
          // else
          else {
            // { Interference & U is preferable for removal: remove U;
            //   update cur_dom & idom }
            // col(U):=NULL
            VariableClass_remove(Variable_CONGRUENCE(U), U);
            PFA_TRACE && fprintf(PFA_TRACE, "Decoalescer_isolateVariable for variable: ");
            PFA_TRACE && Temporary_pretty(Variable_TEMPORARY(U), PFA_TRACE);
            PFA_TRACE && fprintf(PFA_TRACE, "\n");
            // cur_dom(c):=V
            *VariableClass__CURDOM(c) = V;
            // idom(V):=idom(U)
            *Variable__IDOM(V) = Variable_IDOM(U);
          } // endif
        }
        // endforeach { V }
      } Operation_RESULTS_ENDEACH_Temporary;
      // endforeach { OP }
    } BasicBlock_ENDEACH_Operation;
    //
    // { Initialize cur_phi }
    // foreach c in COLORS cur_phi(c) := NULL
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(L, edge) {
      BasicBlock Lp = BasicBlockEdge_destBlock(edge);
      if (BasicBlock_index(Lp) == 0) continue;
      BasicBlock_FOREACH_Operation(Lp, phi) {
        VariableClass colPhi;
        if (!Operator_isPhi(Operation_operator(phi)))
          break;
        colPhi = Operation_COLOR(phi);
        if (colPhi != NULL)
          *VariableClass__CURPHI(colPhi) = NULL;
      } BasicBlock_ENDEACH_Operation;
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
    // { Traversal of all implicit definitions related to phi in successor blocs }
    // foreach L' successor of L
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(L, edge) {
      BasicBlock Lp = BasicBlockEdge_destBlock(edge);
      // let j such that L is the j-th predecessor of L'
      int32_t j = BasicBlockEdge_ARGINDEX(edge);
      if (BasicBlock_index(Lp) == 0) continue;
      PFA_TRACE && fprintf(PFA_TRACE, "    Analysing PHIs in successor %d\n",
                                      BasicBlock_INDEX(Lp));
      //
      // foreach Phi: a0=phi(a1,...,an) in Operations(L')
      BasicBlock_FOREACH_Operation(Lp, phi) {
        Variable U = NULL;
        BasicBlock bb_U = NULL;
        Temporary argument_j;
        Variable aj, a_pk;
        Operation phi_p;
        VariableClass colPhi;
        if (!Operator_isPhi(Operation_operator(phi))) break;
        // let aj be the operand coming from L
        argument_j = Operation_getArgument(phi, j);
        aj = Temporary_VARIABLE(argument_j);
        // colPhi:=col(Phi)
        colPhi = Operation_COLOR(phi);
        // if colPhi==NULL next
        if (colPhi == NULL) continue;
        //
        // { update cur_dom if necessary }
        // U := cur_dom(colPhi)
        U = VariableClass_CURDOM(colPhi);
        // while ( (U != NULL) && (!dominates(bbloc(def(U)),L) || col(U)==NULL) )
        while (   (U != NULL)
               && (   (Variable_color(U) == NULL)
                   || !Coalescer_dominatesVB(coalescer, U, L))) {
          // U := idom (U)
          U = Variable_IDOM(U);
        } // endwhile
        // cur_dom(colPhi):=U
        *VariableClass__CURDOM(colPhi) = U;
        //
        // { Check if Phi interferes with its related cur_dom U }
        // if U != NULL && U in liveOut(L) && U != aj
        if (   (U != NULL) && (U != aj)
               && LiveCheck_isLiveOutBasicBlock(liveCheck, L, U, LiveCheckMode_Copy)) {
          // { U interfere with Phi }
          bool isolatePhi;
          // if U is pinned, isolate Phi. Otherwise make the choice using weight
          isolatePhi = ( (Variable_PINNED(U) != Register__UNDEF) ||
                         (Variable_WEIGHT(U) >= Operation_WEIGHT(phi)) );
          if (isolatePhi) {
            // { Interference & Phi is preferable for removal }
            // col(Phi):=NULL
            Variable_Update_Weight(phi);
            *Operation__COLOR(phi) = NULL;
            PFA_TRACE && fprintf(PFA_TRACE, "Decoalescer_isolatePHI for operation: ");
            PFA_TRACE && Operation_pretty(phi, PFA_TRACE);
            // next
            continue;
          }
          // else
          else {
            // { Interference & U is preferable for removal }
            // col(U):=NULL
            VariableClass_remove(Variable_CONGRUENCE(U), U);
            PFA_TRACE && fprintf(PFA_TRACE, "Decoalescer_isolateVariable for variable: ");
            PFA_TRACE && Temporary_pretty(Variable_TEMPORARY(U), PFA_TRACE);
            PFA_TRACE && fprintf(PFA_TRACE, "\n");
          } // endif
        } // endif
        //
        // { Check if Phi interferes with another Phi }
        // let Phi'= cur_phi(colPhi)
        // if (Phi' != NULL)
        //   if (L is a predecessor of bbloc(Phi') )
        //     let a'k = the operand of Phi' coming from L
        //   else
        //     cur_phi(colPhi) = NULL;
        //     Phi' = NULL
        //   endif
        // endif
        phi_p = VariableClass_CURPHI(colPhi);
        a_pk = NULL;
        if (phi_p != NULL) {
          BasicBlock_LEAVE_FOREACH_BasicBlockEdge(L, edge_p) {
            BasicBlock Lp = BasicBlockEdge_destBlock(edge_p);
            if (BasicBlock_index(Lp) == 0) continue;
            if (Lp == Operation_BASICBLOCK(phi_p)) {
              int32_t k = BasicBlockEdge_ARGINDEX(edge_p);
              Temporary argument_pk = Operation_getArgument(phi_p, k);
              a_pk = Temporary_VARIABLE(argument_pk);
              break;
            }
          } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
          if (a_pk == NULL) {
            *VariableClass__CURPHI(colPhi) = NULL;
            phi_p = NULL;
          }
        }
        // if (Phi' != NULL && a'k != aj)
        if ((phi_p != NULL) && (a_pk != aj)) {
          // { Interfere; isolate Phi }
          // col(Phi):=NULL
          Variable_Update_Weight(phi);
          *Operation__COLOR(phi) = NULL;
          PFA_TRACE && fprintf(PFA_TRACE, "Decoalescer_isolatePHI for operation: ");
          PFA_TRACE && Operation_pretty(phi, PFA_TRACE);
        }
        // else { update cur_phi }
        else {
          // cur_phi(colPhi) := Phi
          *VariableClass__CURPHI(colPhi) = phi;
        } // endif
        // endforeach
      } BasicBlock_ENDEACH_Operation;
      // endforeach
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
    // endforeach
  } Dominance_ENDDOMDFS_BasicBlock;
}

#line 704 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
void
Decoalescer_generateCopies(Decoalescer this)
{
  CodeRegion codeRegion = Decoalescer_CODEREGION(this);
  Coalescer coalescer = Decoalescer_COALESCER(this);
  CodeRegion_FOREACH_BasicBlock(codeRegion, L) {
    BasicBlock_FOREACH_Operation(L, phi) {
      if (Operator_isPhi(Operation_operator(phi))) {
	Coalescer_explicitPhi(coalescer, phi);
	Operation_setFlags(phi, OperationFlag_Processed);
      }
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
}

#line 719 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 731 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
#if XCC__C
static void
Decoalescer_TEST(void)
{
#include "PFA/Decoalescer_.c"
}

#line 738 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Decoalescer.xcc"
int
main(int argc, char **argv)
{
  PFA_Test(Decoalescer_TEST);
  return 0;
}
#endif


#endif//Decoalescer_c_INCLUDED
