#ifndef Interference_c_INCLUDED
#define Interference_c_INCLUDED

#line 18 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
#include "PFA/PFA_.h"

#line 23 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 41 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
Interference
Interference_Ctor(Interference this,
                  Memory memory, CodeRegion codeRegion, Liveness liveness, bool chaintin)
{
  IStack istack = Interference_ISTACK(this);
  TemporaryTable temporaryTable = Liveness_TEMPORARYTABLE(liveness);
  TemporarySet_ *liveOutSets = Liveness_LIVEOUTSETS(liveness);
  TemporarySet_ *phiOutSets = Liveness_PHIOUTSETS(liveness);
  TemporarySparse liveSet = Liveness_SPARSESET(liveness);
  int32_t temporaryCount = TemporaryTable_count(temporaryTable), count = 0;
  *Interference__CODEREGION(this) = codeRegion;
  *Interference__TEMPORARYTABLE(this) = temporaryTable;
  IStack_Ctor(istack, memory, sizeof(BitSet_), temporaryCount);
  // Initialize the ISTACK.
  while (count < temporaryCount) {
    BitSet bitSet = IStack_push(istack);
    BitSet_Ctor(bitSet, memory, ++count);
    Except_CHECK(count == IStack_count(istack));
  }
  /*
   * Adapted from L. George, A. Appel "Iterated Register Coalescing".
   *
   * procedure Build()
   *   forall b in 'blocks in program'
   *     let liveSet = liveOutSet(b)
   *     forall I in instructions(b) in reverse order
   *       if isMove(I) then
   *         liveSet <- liveSet\use(I)
   *       liveSet <- liveSet union def(I)
   *       forall d in def(I)
   *         forall l in liveSet
   *           AddEdge(l, d)
   *       liveSet <- Use(I) union (liveSet\def(I))
   */
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    // let liveSet = liveOutSet(b)
    int32_t index = BasicBlock_index(basicBlock);
    TemporarySparse_empty(liveSet);
    TemporarySparse_insertSet(liveSet, liveOutSets + index);
    /* /!\ This corresponds to the mplex model!
     * For the copy model, the phiOutSet should be unioned only
     * after having traversed the jump. Indeed, the parallel copy can be
     * placed only before the jump.
     */
    if (phiOutSets != NULL) {
      TemporarySparse_insertSet(liveSet, phiOutSets + index);
    }
    // forall I in instructions(b) in reverse order
    BasicBlock_FORBACK_Operation(basicBlock, operation) {
      Operator operator = Operation_operator(operation);
      // if isMove(I) then (Chaitin's trick)
      if (chaintin && Operator_isSeqCopy(operator)) {
        // liveSet <- liveSet\use(I)
        Operation_ARGUMENTS_FOREACH_Temporary(operation, argument) {
          if (TemporaryTable_contains(temporaryTable, argument)) {
            TemporarySparse_remove(liveSet, argument);
          }
        } Operation_ARGUMENTS_ENDEACH_Temporary;
      }
      // liveSet <- liveSet union def(I)
      Operation_RESULTS_FOREACH_Temporary(operation, result) {
        if (TemporaryTable_contains(temporaryTable, result)) {
          TemporarySparse_insert(liveSet, result);
        }
      } Operation_RESULTS_ENDEACH_Temporary;
      // forall d in def(I)
      Operation_RESULTS_FOREACH_Temporary(operation, result) {
        if (TemporaryTable_contains(temporaryTable, result)) {
          // forall l in liveSet
          TemporarySparse_FOREACH_Temporary(liveSet, temporary) {
            if (temporary != result) {
              Interference_add(this, result, temporary);
            }
          } TemporarySparse_ENDEACH_Temporary;
        }
      } Operation_RESULTS_ENDEACH_Temporary;
      // liveSet <- Use(I) union (liveSet\def(I))
      if (!Operator_isPhi(Operation_operator(operation))) {
        // Rastello: do not remove PHI targets from liveSet, as PHI are parallel
        // and their results should interfere.
        Operation_RESULTS_FOREACH_Temporary(operation, result) {
          if (TemporaryTable_contains(temporaryTable, result)) {
            TemporarySparse_remove(liveSet, result);
          }
        } Operation_RESULTS_ENDEACH_Temporary;
        // From: cliffc <cliff.click@Eng.Sun.COM> 6 Jan 1998 22:42:51 -0500
        // (2) Build the InterFerence Graph (IFG) based on this SSA-LIVE info.
        // Same basic technique as with normal LIVE: roll backwards through
        // the block DEFd things interference with what is currently LIVE.
        // Normal USEs are added to the current LIVE set; Phi uses are not.
        Operation_ARGUMENTS_FOREACH_Temporary(operation, argument) {
          if (TemporaryTable_contains(temporaryTable, argument)) {
            TemporarySparse_insert(liveSet, argument);
          }
        } Operation_ARGUMENTS_ENDEACH_Temporary;
      }
    } BasicBlock_ENDBACK_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
  return this;
}

#line 142 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
void
Interference_Dtor(Interference this)
{
  IStack_empty(Interference_ISTACK(this), (IStackItemRelease)BitSet_Dtor);
  IStack_Dtor(Interference_ISTACK(this));
}

#line 149 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
size_t
Interference_Size(Memory memory, CodeRegion codeRegion, Liveness liveness, bool chaintin)
{
  return sizeof(Interference_);
}

#line 164 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
#ifndef Interference_POOLED
#define Interference_POOLED 128
#endif//Interference_POOLED

#line 168 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
Interference
Interference_make(Memory parent, CodeRegion codeRegion, Liveness liveness, bool chaintin)
{
  Memory memory = Memory_new(parent, Interference_POOLED);
  size_t size = Interference_Size(memory, codeRegion, liveness, chaintin);
  Interference this = Memory_alloc(memory, size);
  return Interference_Ctor(this, memory, codeRegion, liveness, chaintin);
}

#line 181 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
Interference
Interference_kill(Interference this)
{
  if (this != NULL) {
    Memory memory = Interference_MEMORY(this);
#ifndef ECL_NDTOR
    Interference_Dtor(this);
    Memory_free_(memory, this);
#endif//ECL_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 195 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 203 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
Memory
Interference_memory(const_Interference this)
{
  return Interference_MEMORY(this);
}

#line 215 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
#ifndef ECL_NTRACE
bool
Interference_pretty(const_Interference this, FILE *file)
{
  const_TemporaryTable temporaryTable = Interference_TEMPORARYTABLE(this);
  if (file == NULL) file = stderr;
  TemporaryTable_FOREACH_Temporary(temporaryTable, t_u) {
    fprintf(file, "\t");
    Temporary_pretty(t_u, file);
  } TemporaryTable_ENDEACH_Temporary;
  fprintf(file, "\n");
  TemporaryTable_FOREACH_Temporary(temporaryTable, t_u) {
    Temporary_pretty(t_u, file);
    TemporaryTable_FOREACH_Temporary(temporaryTable, t_v) {
      fprintf(file, "\t");
      if (Interference_exists(this, t_u, t_v)) {
        fprintf(file, "1");
      }
    } TemporaryTable_ENDEACH_Temporary;
    fprintf(file, "\n");
  } TemporaryTable_ENDEACH_Temporary;
  return true;
}
#endif//ECL_NTRACE

#line 240 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 248 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
void
Interference_add(Interference this, const_Temporary u, const_Temporary v)
{
  int32_t count = Interference_COUNT(this);
  int32_t index_u = Temporary_INDEX(u);
  int32_t index_v = Temporary_INDEX(v);
  int32_t row = MAX(index_u, index_v);
  int32_t col = MIN(index_u, index_v);
  int32_t increase = row - count + 1;
  struct BitSet_ *baseBitSet_ = Interference_BASE(this);
  Except_CHECK(TemporaryTable_contains(Interference_TEMPORARYTABLE(this), u));
  Except_CHECK(TemporaryTable_contains(Interference_TEMPORARYTABLE(this), v));
  Except_CHECK(index_u != index_v);
  if (increase > 0) {
    Memory memory = Interference_MEMORY(this);
    IStack istack = Interference_ISTACK(this);
    while (increase--) {
      // WARNING! the BitSet_(s) may be moved in memory by IStack_push2.
      BitSet bitSet = IStack_push2(istack);
      BitSet_Ctor(bitSet, memory, ++count);
      Except_CHECK(count == IStack_count(istack));
    }
    baseBitSet_ = IStack_base(istack);
  }
  BitSet_insert(baseBitSet_ + row, col);
0&&
  PFA_TRACE && fprintf(PFA_TRACE, "\tInterference_add([%d]", index_u)
            && Temporary_pretty(u, PFA_TRACE) && fprintf(PFA_TRACE, ", [%d]", index_v)
            && Temporary_pretty(v, PFA_TRACE) && fprintf(PFA_TRACE, ")\n");
}

#line 285 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
bool
Interference_exists(const_Interference this, const_Temporary u, const_Temporary v)
{
  int32_t count = Interference_COUNT(this);
  int32_t index_u = Temporary_INDEX(u);
  int32_t index_v = Temporary_INDEX(v);
  int32_t row = MAX(index_u, index_v);
  int32_t col = MIN(index_u, index_v);
  struct BitSet_ *baseBitSet_ = Interference_BASE(this);
  Except_CHECK(TemporaryTable_contains(Interference_TEMPORARYTABLE(this), u));
  Except_CHECK(TemporaryTable_contains(Interference_TEMPORARYTABLE(this), v));
  if (row < count) {
    return BitSet_contains(baseBitSet_ + row, col);
  }
  return false;
}

#line 352 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 364 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
#if XCC__C
static void
Interference_TEST(void)
{
#include "PFA/Interference_.c"
}

#line 371 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/Interference.xcc"
int
main(int argc, char **argv)
{
  PFA_Test(Interference_TEST);
  return 0;
}
#endif


#endif//Interference_c_INCLUDED
