#ifndef LiveCheck_c_INCLUDED
#define LiveCheck_c_INCLUDED

#line 38 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#include "PFA/PFA_.h"

#line 54 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef LiveCheck_LIVENESS
int LiveCheck_LIVENESS = 1;
#endif//LiveCheck_LIVENESS

#line 58 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 109 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
LiveCheck
LiveCheck_Ctor(LiveCheck this,
               Memory memory, Dominance dominance, Liveness liveness)
{
  *LiveCheck__MEMORY(this) = memory;
  if (LiveCheck_LIVENESS==0 || LiveCheck_LIVENESS==2) {
    int32_t *preDfn = Dominance_PREDFN(dominance);
    const BasicBlock *blockMap = Dominance_BLOCKMAP(dominance);
    int32_t blockCount = Dominance_BLOCKCOUNT(dominance), index;
    BasicBlock rootBlock = blockMap[0];
    uint8_t *dfsFlags = Memory_alloc(memory, blockCount*sizeof(uint8_t));
    int32_t *dfnMap = Memory_alloc(memory, blockCount*sizeof(int32_t));
    BitSet_ *rSets = Memory_alloc(memory, 3*blockCount*sizeof(BitSet_));
    BitSet_ *hSets = rSets + blockCount;
    BitSet_ *tSets = hSets + blockCount;
    Except_CHECK(Dominance_isForward(dominance));
    *LiveCheck__BLOCKCOUNT(this) = blockCount;
    *LiveCheck__DOMINANCE(this) = dominance;
    *LiveCheck__DFSFLAGS(this) = dfsFlags;
    *LiveCheck__DFNMAP(this) = dfnMap;
    BitSet_Ctor(LiveCheck_BACKHEADS(this), memory, blockCount);
    *LiveCheck__RSETS(this) = rSets;
    *LiveCheck__HSETS(this) = hSets;
    *LiveCheck__TSETS(this) = tSets;
    // Compute all information used for fast liveness checking.
    // PREDFN, LASTDFN, RSETS and TSETS will be used.
    // Compute dfnMap = preDfn^(-1)
    for (index = 0; index < 3*blockCount; index++) {
      BitSet_Ctor(rSets + index, memory, blockCount);
    }
    memset(dfnMap, 0, blockCount*sizeof(int32_t));
    for (index = 0; index < blockCount; index++) {
      dfnMap[preDfn[index]] = index;
    }
    // Compute backedge targets, Rv & Tv^
    memset(dfsFlags, 0, blockCount*sizeof(uint8_t));
    firstDFS4precompute(this, rootBlock);
    // Compute Tv
    memset(dfsFlags, 0, blockCount*sizeof(uint8_t));
    secondDFS4precompute(this, rootBlock);
    for (index = 0; index < blockCount; index++) {
      BitSet_insert(tSets + index, preDfn[index]);
    }
  }
  if (LiveCheck_LIVENESS>0) {
    *LiveCheck__LIVEINSETS(this) = Liveness_LIVEINSETS(liveness);
    *LiveCheck__LIVEOUTSETS(this) = Liveness_LIVEOUTSETS(liveness);
    *LiveCheck__PHIINSETS(this) = Liveness_PHIINSETS(liveness);
    *LiveCheck__PHIOUTSETS(this) = Liveness_PHIOUTSETS(liveness);
    *LiveCheck__BLOCKCOUNT(this) = Liveness_BLOCKCOUNT(liveness);
  }
  return this;
}

#line 163 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
void
LiveCheck_Dtor(LiveCheck this)
{
  if (LiveCheck_LIVENESS==0 || LiveCheck_LIVENESS==2) {
    int32_t blockCount = LiveCheck_BLOCKCOUNT(this), index;
    Memory memory = LiveCheck_MEMORY(this);
    Memory_free(memory, LiveCheck_DFSFLAGS(this));
    Memory_free(memory, LiveCheck_DFNMAP(this));
    BitSet_Dtor(LiveCheck_BACKHEADS(this));
    for (index = 0; index < 3*blockCount; index++) {
      BitSet_Dtor(LiveCheck_RSETS(this) + index);
    }
    Memory_free(memory, LiveCheck_RSETS(this));
  }
}

#line 179 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
size_t
LiveCheck_Size(Memory memory, Dominance dominance, Liveness liveness)
{
  return sizeof(LiveCheck_);
}

#line 196 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef LiveCheck_POOLED
#define LiveCheck_POOLED 128
#endif//LiveCheck_POOLED

#line 200 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
LiveCheck
LiveCheck_make(Memory parent, Dominance dominance, Liveness liveness)
{
  Memory memory = Memory_new(parent, LiveCheck_POOLED);
  size_t size = LiveCheck_Size(memory, dominance, liveness);
  LiveCheck this = Memory_alloc(memory, size);
  LiveCheck_Ctor(this, memory, dominance, liveness);
  return this;
}

#line 214 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
LiveCheck
LiveCheck_kill(LiveCheck this)
{
  if (this!=NULL) {
    Memory memory = LiveCheck_MEMORY(this);
#ifndef ECL_NDTOR
    LiveCheck_Dtor(this);
    Memory_free_(memory, this);
#endif//ECL_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 228 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 236 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
Memory
LiveCheck_memory(const_LiveCheck this)
{
  return LiveCheck_MEMORY(this);
}

#line 246 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
static void
firstDFS4precompute(LiveCheck this, BasicBlock block)
{
  // Perform Depth First Search post order traversal
  // and update RSETS & HSETS & BACKHEADS & BasicBlockEdge_ISBACK
  BitSet backHeads = LiveCheck_BACKHEADS(this);
  uint8_t *dfsFlags = LiveCheck_DFSFLAGS(this);
  int32_t *preDfn = LiveCheck_PREDFN(this);
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  BitSet_ *rSets = LiveCheck_RSETS(this);
  BitSet_ *hSets = LiveCheck_HSETS(this);
  int32_t block_index = BasicBlock_index(block);
  dfsFlags[block_index] = 1;
  // Now visit all the other children (fall through is guaranteed to be the first one in the list)
  BitSet_insert(rSets + block_index, preDfn[block_index]);
  BasicBlock_LEAVE_FOREACH_BasicBlockEdge(block, successor) {
    BasicBlock succ = BasicBlockEdge_destBlock(successor);
    int32_t succ_index = BasicBlock_index(succ);
    if (dfsFlags[succ_index] == 0) {
      firstDFS4precompute(this, succ);
    }
    assert(dfsFlags[succ_index] != 0);
    if (dfsFlags[succ_index] == 2) { // processed: tree-edge, cross-edge or forward-edge
      BitSet_union(rSets + block_index, rSets + succ_index);  // propagates rSets of successors
      BitSet_union(hSets + block_index, hSets + succ_index); // propagates hSets
    }
    if (dfsFlags[succ_index] == 1 && succ_index != 0) { // currently processing it: succ is an ancestor
      // This is a back edge to a block different than r
      BitSet_insert(hSets + block_index, preDfn[succ_index]);
      *BasicBlockEdge__ISBACK(successor) = true;
      BitSet_insert(backHeads, succ_index);
    }
  } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  BitSet_diff(hSets + block_index, rSets + block_index); // remove Rt from Tt^
  dfsFlags[block_index] = 2;
}

#line 287 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
static void
secondDFS4precompute(LiveCheck this, BasicBlock block)
{
  // using a preorder traversal, propagate Tv on back edge targets
  // using a postorder traversal, propagate Tv\{v} on other nodes
  // We could have simply propagate Tv on all nodes using a preorder traversal
  //    using equation 1. But we suppose that Tv^ is ususally larger  than succ(v)
  // /!\ There remains to add {v} to each Tv for non back-edge targets
  BitSet backHeads = LiveCheck_BACKHEADS(this);
  uint8_t *dfsFlags = LiveCheck_DFSFLAGS(this);
  BitSet_ *rSets = LiveCheck_RSETS(this);
  BitSet_ *tSets = LiveCheck_TSETS(this);
  BitSet_ *hSets = LiveCheck_HSETS(this);
  int32_t *preDfn = LiveCheck_PREDFN(this);
  int32_t *dfnMap = LiveCheck_DFNMAP(this);
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  int32_t block_index = BasicBlock_index(block);
  // DFS flag
  if (dfsFlags[block_index]) return;
  dfsFlags[block_index] = 1;
  // back-edge target. propagate along a DFS pre-order traversal.
  // using equation 1 of the paper:  Tv = {v} union Union_{w\in Tv^} Tw
  if (BitSet_contains(backHeads, block_index)) {
    int32_t v_index = block_index;
    BitSet_insert(tSets + v_index, preDfn[v_index]);
    {
      BitSet_FOREACH(hSets + v_index, w) {
        Except_CHECK(!BitSet_contains(rSets + v_index, w));
        BitSet_union(tSets + v_index, tSets + dfnMap[w]);
      } BitSet_ENDEACH;
      dfsFlags[v_index] = 2;
    }
  }
  // DFS traversal
  BasicBlock_LEAVE_FOREACH_BasicBlockEdge(block, successor) {
    BasicBlock succ = BasicBlockEdge_destBlock(successor);
    int32_t succ_index = BasicBlock_index(succ);
    if (succ_index == 0 /* rootblock */) continue;
    secondDFS4precompute(this, succ);
  } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  // not a back-edge target. propagate along a DFS postorder traversal.
  // using equation: Tv \ {v}= Union_{w\in succ(v)} Tw \ Rv
  if (!BitSet_contains(backHeads, block_index)) {
    int32_t v_index = block_index;
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(block, outedge) {
      BasicBlock w = BasicBlockEdge_destBlock(outedge);
      int32_t w_index = BasicBlock_index(w);
      if (w_index == 0 /* rootblock */) continue;
      BitSet_union(tSets + block_index, tSets + w_index);
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
    BitSet_diff(tSets + block_index, rSets + v_index);
    dfsFlags[block_index] = 2;
  }
  // /!\ we need to add {v} to each Tv for each non back-edge target
}

#line 343 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 357 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSSALiveInBasicBlock(LiveCheck this, BasicBlock Q, Variable variable, uint32_t mode)
// we check for livein at basic block Q (the query) for variable defined at defA and used at u \in uses
// mode can be Mplex or Copy (see above)
// We use the following convention for notations:
//  BasicBlocks are represented in upper case, operations in lower case.
//  dfn numbering prefixed by _dfn. index numbering prefixed by _index.
// To activate trace during self-test do:
// change 0 && LiveCheck_TRACE... into 1 && LiveCheck_TRACE...
// run env  PFA_LiveCheck_TRACE=1 ./LiveCheck.tst
{
  int32_t *dfnMap = LiveCheck_DFNMAP(this);
  int32_t *preDfn = LiveCheck_PREDFN(this);
  int32_t *lastDfn = LiveCheck_LASTDFN(this);
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  BitSet_ *rSets = LiveCheck_RSETS(this);
  BitSet_ *tSets = LiveCheck_TSETS(this);
  BitSet TQ = NULL; // BitSet of t from which reachability should be checked
  Operation defa = OperationResult_OPERATION(Variable_PRODUCER(variable));
  BasicBlock defA = Operation_basicBlock(defa);
  int32_t defA_index = BasicBlock_index(defA);
  int32_t Q_index = BasicBlock_index(Q);
  int32_t PRE_dfn = preDfn[defA_index]; // tests can be restricted to some dominance subtree which have indexes between PRE_dfn and LAST_dfn
  int32_t LAST_dfn = lastDfn[defA_index];
  int32_t NEXT_dfn;
  bool ret = false;
  // Q == defA, defa is a phi, mode == LiveCheckMode_Mplex
  if (   Q == defA
      && Operator_isPhi(Operation_operator(defa))
      && (   (mode == LiveCheckMode_Mplex)
          || (mode == LiveCheckMode_Mixed && Operation_isProcessed(defa)))) {
    ret = true;
    goto end;
  }
  // Q not strictly dominated by defA
  if (preDfn[Q_index] <= PRE_dfn || LAST_dfn < preDfn[Q_index]) {
    ret = false;
    goto end;
  }
  // Tq
  TQ = tSets + Q_index;
  0 && LiveCheck_TRACE && fprintf(LiveCheck_TRACE, "Q_index:%d, PRE_dfn:%d, LAST_dfn:%d, NEXT_dfn:%d\n", Q_index, PRE_dfn, LAST_dfn, NEXT_dfn);
  // Tqa traversal
  while (   PRE_dfn < LAST_dfn
         && (NEXT_dfn = BitSet_nextMember(TQ, PRE_dfn)) >= 0
         && NEXT_dfn <= LAST_dfn) {
    int32_t T_index = dfnMap[NEXT_dfn];
    PRE_dfn = lastDfn[T_index];
    0 && LiveCheck_TRACE && fprintf(LiveCheck_TRACE, "T_index:%d, Q_index:%d, PRE_dfn:%d, LAST_dfn:%d, NEXT_dfn:%d\n", T_index, Q_index, PRE_dfn, LAST_dfn, NEXT_dfn);
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      // find U
      Operation u = OperationArgument_OPERATION(oparg);
      BasicBlock U = Operation_basicBlock(u);
      if (Operator_isPhi(Operation_operator(u))) {
        int32_t i=OperationArgument_ARGINDEX(oparg);
        BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(U, i);
        BasicBlock block_i = BasicBlockEdge_origBlock(edge_i);
        U = block_i;
      }
      if (BitSet_contains(rSets + T_index, preDfn[BasicBlock_index(U)])) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
  }
end:
  Except_CHECK(ret == LiveCheck_isSSALiveInOrdering(this, Q, BasicBlock_firstOrdering(Q), variable, mode));
  return ret;
}

#line 442 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSetLiveInBasicBlock(LiveCheck this, BasicBlock Q, Variable variable, uint32_t mode)
{
  bool ret;
  int32_t index = BasicBlock_index(Q);
  Temporary temporary = Variable_TEMPORARY(variable);
  Except_CHECK(LiveCheck_LIVENESS>0);
  if (mode == LiveCheckMode_Mixed) {
    ret = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary);
  } else {
    ret = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary)
        || (   (mode == LiveCheckMode_Mplex)
            && TemporarySet_contains(LiveCheck_PHIINSETS(this) + index, temporary));
  }
  Except_CHECK(ret == LiveCheck_isSetLiveInOrdering(this, Q, BasicBlock_firstOrdering(Q), variable, mode));
  return ret;
}

#line 476 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 490 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSSALiveOutBasicBlock(LiveCheck this, BasicBlock Q, Variable variable, uint32_t mode)
// we check for liveout at basic block Q (the query) for variable defined at defA and used at u \in uses
{
  int32_t *dfnMap = LiveCheck_DFNMAP(this);
  int32_t *preDfn = LiveCheck_PREDFN(this);
  int32_t *lastDfn = LiveCheck_LASTDFN(this);
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  BitSet_ *rSets = LiveCheck_RSETS(this);
  BitSet_ *tSets = LiveCheck_TSETS(this);
  BitSet backHeads = LiveCheck_BACKHEADS(this);
  BitSet TQ = NULL; // BitSet of t from which reachability should be checked
  BasicBlock defA = Operation_basicBlock(OperationResult_OPERATION(Variable_PRODUCER(variable)));
  int32_t defA_index = BasicBlock_index(defA);
  int32_t Q_index = BasicBlock_index(Q);
  int32_t PRE_dfn = preDfn[defA_index]; // tests can be restricted to some dominance subtree which have indexes between PRE_dfn and last
  int32_t LAST_dfn = lastDfn[defA_index];
  int32_t NEXT_dfn;
  bool ret = false;
  // Q not dominated by defA
  if (preDfn[Q_index] < PRE_dfn || LAST_dfn < preDfn[Q_index]) {
    ret = false;
    goto end;
  }
  if (defA == Q) {
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      Operation u = OperationArgument_OPERATION(oparg);
      BasicBlock U;
      if (Operator_isPhi(Operation_operator(u))) {
        if (   (mode == LiveCheckMode_Mplex)
            || (mode == LiveCheckMode_Mixed && Operation_isProcessed(u))) {
          // MPLEXmodel, used in a phi => liveout of defA
          ret = true;
          goto end;
        }
        { // COPY mode, U is the predecessor block
          int32_t i = OperationArgument_ARGINDEX(oparg);
          BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(Operation_basicBlock(u), i);
          U = BasicBlockEdge_origBlock(edge_i);
        }
      } else {
        U = Operation_basicBlock(u);
      }
      if (U != defA) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
    ret = false;
    goto end;
  }
  // Tq
  TQ = tSets + Q_index;
  // Tqa traversal
  while (   PRE_dfn < LAST_dfn
         && (NEXT_dfn = BitSet_nextMember(TQ, PRE_dfn)) >= 0
         && NEXT_dfn <= LAST_dfn) {
    int32_t T_index = dfnMap[NEXT_dfn];
    bool t_equal_q_is_notbackedgetarget = (T_index == Q_index && !BitSet_contains(backHeads, Q_index));
    PRE_dfn = lastDfn[T_index];
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      // find u
      Operation u = OperationArgument_OPERATION(oparg);
      BasicBlock U = Operation_basicBlock(u);
      if (Operator_isPhi(Operation_operator(u))) { // For a phi, U is the predecessor's block
        int32_t i = OperationArgument_ARGINDEX(oparg);
        BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(U, i);
        U = BasicBlockEdge_origBlock(edge_i);
        if (U == Q && (    (mode == LiveCheckMode_Mplex)
                        || (mode == LiveCheckMode_Mixed && Operation_isProcessed(u)))){
        // MPLEX mode: a use is liveout of the predecessor's block
          ret = true;
          goto end;
        }
      }
      if (t_equal_q_is_notbackedgetarget && U == Q) continue;
      if (BitSet_contains(rSets + T_index, preDfn[BasicBlock_index(U)])) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
  }
end:
  Except_CHECK(ret == LiveCheck_isSSALiveOutOrdering(this, Q, BasicBlock_lastOrdering(Q), variable, mode));
  return ret;
}

#line 593 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSetLiveOutBasicBlock(LiveCheck this, BasicBlock Q, Variable variable, uint32_t mode)
{
  bool ret;
  int32_t index = BasicBlock_index(Q);
  Temporary temporary = Variable_TEMPORARY(variable);
  Except_CHECK(LiveCheck_LIVENESS>0);
  if (mode == LiveCheckMode_Mixed) {
    ret = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary);
  }
  else {
    ret = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary)
        || (   mode == LiveCheckMode_Mplex
            && TemporarySet_contains(LiveCheck_PHIOUTSETS(this) + index, temporary));
  }
  Except_CHECK(ret == LiveCheck_isSetLiveOutOrdering(this, Q, BasicBlock_lastOrdering(Q), variable, mode));
  return ret;
}

#line 628 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 642 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
// A definir ailleur:
#define IMPLIES(a,b) (!(a) || (b))
#define IFF(a,b) (((a) && (b)) || (!(a) && !(b)))

#line 646 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSSALiveInOperation(LiveCheck this, Operation q, Variable variable, uint32_t mode)
// we check for livein at operation q (the query) for variable defined at defa and used at u \in uses
// /!\ If q is a phi function then livein(q) is undefined
// We use the following convention for notations:
//  BasicBlocks are represented in upper case, operations in lower case.
//  dfn numbering prefixed by _dfn. index numbering prefixed by _index.
// To activate trace during self-test do:
// change 0 && LiveCheck_TRACE... into 1 && LiveCheck_TRACE...
// run env  PFA_LiveCheck_TRACE=1 ./LiveCheck.tst
{
  Operation defa = OperationResult_OPERATION(Variable_PRODUCER(variable));
  int32_t defa_ordering = Operation_ORDERING(defa);
  BasicBlock Q = Operation_basicBlock(q);
  int32_t Q_index = BasicBlock_index(Q);
  int32_t q_ordering = Operation_ORDERING(q);
  bool ret = false;
  Except_CHECK(defa_ordering != -1 && q_ordering != -1);
  Except_CHECK(!Operator_isPhi(Operation_operator(q)));
  ret = LiveCheck_isSSALiveInOrdering(this, Q, q_ordering, variable, mode);
  return ret;
}

#line 684 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSetLiveInOperation(LiveCheck this, Operation q, Variable variable, uint32_t mode)
{
  BasicBlock block_q = Operation_basicBlock(q);
  bool ret = LiveCheck_isSetLiveInOrdering(this, block_q, Operation_ordering(q), variable, mode);
  return ret;
}

#line 692 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 706 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSSALiveOutOperation(LiveCheck this, Operation q, Variable variable, uint32_t mode)
// we check for liveout at operation q (the query) for variable defined at defq and used at u \in uses
// The mode makes no difference. Indeed, for the copy mode, the variable use in a phi is not
//   liveout of the corresponding predecessor basicBlock but liveout of its last instruction.
// We use the following convention for notations:
//  BasicBlocks are represented in upper case, operations in lower case.
//  dfn numbering prefixed by _dfn. index numbering prefixed by _index.
// To activate trace during self-test do:
// change 0 && LiveCheck_TRACE... into 1 && LiveCheck_TRACE...
// run env  PFA_LiveCheck_TRACE=1 ./LiveCheck.tst
{
  int32_t q_ordering = Operation_ORDERING(q);
  BasicBlock Q = Operation_basicBlock(q);
  bool ret = LiveCheck_isSSALiveOutOrdering(this, Q, q_ordering, variable, mode);
  return ret;
}

#line 738 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSetLiveOutOperation(LiveCheck this, Operation q, Variable variable, uint32_t mode)
{
  BasicBlock Q = Operation_basicBlock(q);
  int32_t q_ordering = Operation_ordering(q);
  bool ret  = LiveCheck_isSetLiveOutOrdering(this, Q, q_ordering, variable, mode);
  return ret;
}

#line 747 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 761 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSSALiveInOrdering(LiveCheck this, BasicBlock Q, int32_t ordering, Variable variable, uint32_t mode)
// we check for livein at operation q (the query) for variable defined at defa and used at u \in uses
// Q:=basicBlock(q); ordering = ordering(q)
// we must have min_ordering(Q) <= ordering <= max_ordering(Q)
// if ordering = min_ordering(Q) provides the liveincheck for basicBlock Q.
// the semantic for the case ordering > max_ordering(Q) is undefined for livein check.
// We use the following convention for notations:
//  BasicBlocks are represented in upper case, operations in lower case.
//  dfn numbering prefixed by _dfn. index numbering prefixed by _index.
// To activate trace during self-test do:
// change 0 && LiveCheck_TRACE... into 1 && LiveCheck_TRACE...
// run env  PFA_LiveCheck_TRACE=1 ./LiveCheck.tst
{
  int32_t *dfnMap = LiveCheck_DFNMAP(this);
  int32_t *preDfn = LiveCheck_PREDFN(this);
  int32_t *lastDfn = LiveCheck_LASTDFN(this);
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  BitSet_ *rSets = LiveCheck_RSETS(this);
  BitSet_ *tSets = LiveCheck_TSETS(this);
  BitSet backHeads = LiveCheck_BACKHEADS(this);
  BitSet TQ = NULL; // BitSet of t from which reachability should be checked
  Operation defa = OperationResult_OPERATION(Variable_PRODUCER(variable));
  int32_t defa_ordering;
  BasicBlock defA = Operation_basicBlock(defa);
  int32_t defA_index = BasicBlock_index(defA);
  int32_t Q_index = BasicBlock_index(Q);
  int32_t q_ordering = ordering;
  int32_t PRE_dfn = preDfn[defA_index]; // tests can be restricted to some dominance subtree which have indexes between PRE_dfn and LAST_dfn
  int32_t LAST_dfn = lastDfn[defA_index];
  int32_t NEXT_dfn;
  bool ret = false;
  if (Operator_isPhi(Operation_operator(defa))) {
      if (   (mode == LiveCheckMode_Copy)
          || (mode == LiveCheckMode_Mixed && !Operation_isProcessed(defa)))
        defa_ordering = BasicBlock_earlyOrdering(defA);
      else  defa_ordering = BasicBlock_firstOrdering(defA)-1;
  } else defa_ordering = Operation_ordering(defa);
  Except_CHECK(q_ordering != -1 && defa_ordering != -1);
  Except_CHECK(BasicBlock_firstOrdering(Q) <= ordering && ordering <= BasicBlock_lastOrdering(Q));
  // q not strictly dominated by defa
  if (   preDfn[Q_index] < PRE_dfn
      || LAST_dfn < preDfn[Q_index]
      || (Q == defA && q_ordering <= defa_ordering)) {
    ret = false;
    goto end;
  }
  if (defA == Q) {
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      Operation u = OperationArgument_OPERATION(oparg);
      bool isPhi = Operator_isPhi(Operation_operator(u));
      BasicBlock U;
      int32_t u_ordering;
      if (isPhi) {
        if (   (mode == LiveCheckMode_Mplex)
            || (mode == LiveCheckMode_Mixed && Operation_isProcessed(u))) {
          ret = true;
          goto end;
        } else {
          int32_t i=OperationArgument_ARGINDEX(oparg);
          BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(Operation_basicBlock(u), i);
          BasicBlock block_i = BasicBlockEdge_origBlock(edge_i);
          U = block_i;
          u_ordering = BasicBlock_lateOrdering(U);
        }
      } else {
        U = Operation_basicBlock(u);
        u_ordering = Operation_ORDERING(u);
      }
      Except_CHECK(u_ordering != -1);
      if (U != defA || q_ordering <= u_ordering) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
    ret = false;
    goto end;
  }
  // Tq
  TQ = tSets + Q_index;
  // Tqa traversal
  while (   PRE_dfn < LAST_dfn
         && (NEXT_dfn = BitSet_nextMember(TQ, PRE_dfn)) >= 0
         && NEXT_dfn <= LAST_dfn) {
    int32_t T_index = dfnMap[NEXT_dfn];
    bool T_equal_Q_isnot_backedgetarget = (T_index == Q_index && !BitSet_contains(backHeads, Q_index));
    PRE_dfn = lastDfn[T_index];
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      // find u
      Operation u = OperationArgument_OPERATION(oparg);
      int32_t u_ordering;
      BasicBlock U = Operation_basicBlock(u);
      bool isPhi = Operator_isPhi(Operation_operator(u));
      if (isPhi) {
        // For a phi, U is the predecessor's block
        int32_t i = OperationArgument_ARGINDEX(oparg);
        BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(U, i);
        U = BasicBlockEdge_origBlock(edge_i);
        if (   (mode == LiveCheckMode_Copy)
            || (mode == LiveCheckMode_Mixed && !Operation_isProcessed(u))) {
          u_ordering = BasicBlock_lateOrdering(U);
        } else {
          u_ordering = BasicBlock_lastOrdering(U) + 1;
        }
        Except_CHECK(u_ordering != -1);
      } else {
        u_ordering = Operation_ORDERING(u);
      }
      Except_CHECK(u_ordering != -1);
      if (T_equal_Q_isnot_backedgetarget && U == Q && u_ordering < q_ordering) continue;
      if (BitSet_contains(rSets + T_index, preDfn[BasicBlock_index(U)])) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
  }
end:
  return ret;
}

#line 899 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSetLiveInOrdering(LiveCheck this, BasicBlock Q, int32_t ordering, Variable variable, uint32_t mode)
{
  Temporary temporary = Variable_TEMPORARY(variable);
  int32_t index = BasicBlock_index(Q);
  bool isbblivein, isbbliveout;
  bool ret = true;
  Except_CHECK(LiveCheck_LIVENESS>0);
  if (mode == LiveCheckMode_Copy) {
    isbblivein = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary);
    isbbliveout = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary);
  } else if (mode == LiveCheckMode_Mplex) {
    isbblivein = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary)
               || TemporarySet_contains(LiveCheck_PHIINSETS(this) + index, temporary);
    isbbliveout = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary)
                || TemporarySet_contains(LiveCheck_PHIOUTSETS(this) + index, temporary);
  } else {
    Except_CHECK(mode == LiveCheckMode_Mixed);
    isbblivein = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary);
    isbbliveout = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary);
  }
  Except_CHECK(BasicBlock_firstOrdering(Q) <= ordering && ordering <= BasicBlock_lastOrdering(Q));
  if (!isbblivein) {
    // No (strict) reaching definition
    Operation def = OperationResult_OPERATION(Variable_PRODUCER(variable));
    int32_t def_ordering;
    BasicBlock DEF = Operation_basicBlock(def);
    if ( DEF != Q ||
         ordering == BasicBlock_firstOrdering(Q)) {
      ret = false;
      goto end;
    }
    if (Operator_isPhi(Operation_operator(def))) {
      def_ordering = BasicBlock_earlyOrdering(DEF);
    } else def_ordering = Operation_ordering(def);
    if (def_ordering >= ordering) {
      ret = false;
      goto end;
    }
  }
  if (!isbbliveout) {
    // Not upward exposed use
    bool isusedfurther = false;
    // Search for a use u in uses(variable) and in block(q)
    Except_CHECK(Variable_isConsuming(variable));
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, argument) {
        Operation u = OperationArgument_OPERATION(argument);
        int32_t u_ordering;
        BasicBlock U = Operation_basicBlock(u);
        if (Operator_isPhi(Operation_operator(u))) {
            int32_t argIndex=OperationArgument_ARGINDEX(argument);
            BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(U, argIndex);
            U = BasicBlockEdge_origBlock(edge_i);
	    u_ordering = BasicBlock_lateOrdering(U);
        } else {
          u_ordering = Operation_ordering(u);
        }
        if (U == Q && ordering <= u_ordering) {
          isusedfurther = true;
          break;
        }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
    if (!isusedfurther) {
      ret = false;
      goto end;
    }
  }
  // Strict reaching def and updward exposed use
 end:
  return ret;
}

#line 987 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1001 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSSALiveOutOrdering(LiveCheck this, BasicBlock Q, int32_t ordering, Variable variable, uint32_t mode)
// we check for liveout at operation q (the query) for variable defined at defa and used at u \in uses
// Q = BasicBlock(q)
// ordering = ordering(a)
// We must have min_ordering(Q) <= ordering <= max_ordering(Q)
// if ordering == max_ordering(Q) provides the liveout query for basicBlock Q
// We use the following convention for notations:
//  BasicBlocks are represented in upper case, operations in lower case.
//  dfn numbering prefixed by _dfn. index numbering prefixed by _index.
// To activate trace during self-test do:
// change 0 && LiveCheck_TRACE... into 1 && LiveCheck_TRACE...
// run env  PFA_LiveCheck_TRACE=1 ./LiveCheck.tst
{
  int32_t *dfnMap = LiveCheck_DFNMAP(this);
  int32_t *preDfn = LiveCheck_PREDFN(this);
  int32_t *lastDfn = LiveCheck_LASTDFN(this);
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  BitSet_ *rSets = LiveCheck_RSETS(this);
  BitSet_ *tSets = LiveCheck_TSETS(this);
  BitSet backHeads = LiveCheck_BACKHEADS(this);
  BitSet TQ = NULL; // BitSet of t from which reachability should be checked
  Operation defa = OperationResult_OPERATION(Variable_PRODUCER(variable));
  int32_t defa_ordering;
  BasicBlock defA = Operation_basicBlock(defa);
  int32_t defA_index = BasicBlock_index(defA);
  int32_t Q_index = BasicBlock_index(Q);
  int32_t q_ordering = ordering;
  int32_t PRE_dfn = preDfn[defA_index]; // tests can be restricted to some dominance subtree which have indexes between PRE_dfn and LAST_dfn
  int32_t LAST_dfn = lastDfn[defA_index];
  int32_t NEXT_dfn;
  bool ret = false;
  if (Operator_isPhi(Operation_operator(defa))) {
      if (   (mode == LiveCheckMode_Copy)
          || (mode == LiveCheckMode_Mixed && !Operation_isProcessed(defa)))
        defa_ordering = BasicBlock_earlyOrdering(defA);
      else  defa_ordering = BasicBlock_firstOrdering(defA)-1;
  } else defa_ordering = Operation_ordering(defa);
  Except_CHECK(q_ordering != -1 && defa_ordering != -1);
  Except_CHECK(BasicBlock_firstOrdering(Q) <= ordering && ordering <= BasicBlock_lastOrdering(Q));
  // q not dominated by defa
  if (   preDfn[Q_index] < PRE_dfn
      || LAST_dfn < preDfn[Q_index]
      || (Q == defA && q_ordering < defa_ordering)) {
    ret = false;
    goto end;
  }
  if (defA == Q) {
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      Operation u = OperationArgument_OPERATION(oparg);
      bool isPhi = Operator_isPhi(Operation_operator(u));
      BasicBlock U;
      int32_t u_ordering;
      if (isPhi) {
        if (   (mode == LiveCheckMode_Mplex)
            || (mode == LiveCheckMode_Mixed && Operation_isProcessed(u))) {
          ret = true;
          goto end;
        } else {
          int32_t i=OperationArgument_ARGINDEX(oparg);
          BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(Operation_basicBlock(u), i);
          BasicBlock block_i = BasicBlockEdge_origBlock(edge_i);
          U = block_i;
          u_ordering = BasicBlock_lateOrdering(U);
        }
      } else {
        U = Operation_basicBlock(u);
        u_ordering = Operation_ORDERING(u);
      }
      Except_CHECK(u_ordering != -1);
      if (U != defA || q_ordering < u_ordering) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
    ret = false;
    goto end;
  }
  // Tq
  TQ = tSets + Q_index;
  // Tqa traversal
  while (   PRE_dfn < LAST_dfn
         && (NEXT_dfn = BitSet_nextMember(TQ, PRE_dfn)) >= 0
         && NEXT_dfn <= LAST_dfn) {
    int32_t T_index = dfnMap[NEXT_dfn];
    bool T_equal_Q_isnot_backedgetarget = (T_index == Q_index && !BitSet_contains(backHeads, Q_index));
    PRE_dfn = lastDfn[T_index];
    // traverse uses
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, oparg) {
      // find u
      Operation u = OperationArgument_OPERATION(oparg);
      int32_t u_ordering;
      BasicBlock U = Operation_basicBlock(u);
      bool isPhi = Operator_isPhi(Operation_operator(u));
      if (isPhi) {
        // For a phi, U is the predecessor's block
        int32_t i = OperationArgument_ARGINDEX(oparg);
        BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(U, i);
        U = BasicBlockEdge_origBlock(edge_i);
        if (   (mode == LiveCheckMode_Copy)
            || (mode == LiveCheckMode_Mixed && !Operation_isProcessed(u))) {
          u_ordering = BasicBlock_lateOrdering(U);
        } else {
          u_ordering = BasicBlock_lastOrdering(U) + 1;
        }
        Except_CHECK(u_ordering != -1);
      } else {
        u_ordering = Operation_ORDERING(u);
      }
      Except_CHECK(u_ordering != -1);
      if (T_equal_Q_isnot_backedgetarget && U == Q && u_ordering <= q_ordering) continue;
      if (BitSet_contains(rSets + T_index, preDfn[BasicBlock_index(U)])) {
        ret = true;
        goto end;
      }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
  }
end:
  return ret;
}

#line 1137 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
bool
LiveCheck_isSetLiveOutOrdering(LiveCheck this, BasicBlock Q, int32_t ordering, Variable variable, uint32_t mode)
{
  Temporary temporary = Variable_TEMPORARY(variable);
  int32_t index = BasicBlock_index(Q);
  bool isbblivein, isbbliveout;
  bool ret = true;
  Except_CHECK(LiveCheck_LIVENESS>0);
  if (mode == LiveCheckMode_Copy) {
    isbblivein = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary);
    isbbliveout = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary);
  } else if (mode == LiveCheckMode_Mplex) {
    isbblivein = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary)
               || TemporarySet_contains(LiveCheck_PHIINSETS(this) + index, temporary);
    isbbliveout = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary)
                || TemporarySet_contains(LiveCheck_PHIOUTSETS(this) + index, temporary);
  } else {
    Except_CHECK(mode == LiveCheckMode_Mixed);
    isbblivein = TemporarySet_contains(LiveCheck_LIVEINSETS(this) + index, temporary);
    isbbliveout = TemporarySet_contains(LiveCheck_LIVEOUTSETS(this) + index, temporary);
  }
  if (!isbblivein) {
    // No reaching definition
    Operation def = OperationResult_OPERATION(Variable_PRODUCER(variable));
    int32_t def_ordering;
    BasicBlock DEF = Operation_basicBlock(def);
    if (DEF != Q) {
      ret = false;
      goto end;
    }
    if (Operator_isPhi(Operation_operator(def))) {
      def_ordering = BasicBlock_earlyOrdering(DEF);
    } else def_ordering = Operation_ordering(def);
    if (def_ordering > ordering) {
      ret = false;
      goto end;
    }
  }
  if (!isbbliveout) {
    // Not (strict) upward exposed use
    bool isusedfurther = false;
    int32_t ordering_q = ordering;
    // search for a use u (not a phi use) in use(variable) and in block(q)
    Except_CHECK(Variable_isConsuming(variable));
    Variable_CONSUMERS_FOREACH_OperationArgument(variable, argument) {
        Operation u = OperationArgument_OPERATION(argument);
        int32_t u_ordering;
        BasicBlock U = Operation_basicBlock(u);
        if (Operator_isPhi(Operation_operator(u))) {
            int32_t argIndex=OperationArgument_ARGINDEX(argument);
            BasicBlockEdge edge_i = BasicBlock_getPhiEnterEdge(U, argIndex);
            U = BasicBlockEdge_origBlock(edge_i);
	    u_ordering = BasicBlock_lateOrdering(U);
        } else {
          u_ordering = Operation_ordering(u);
        }
        if (U == Q && ordering < u_ordering) {
          isusedfurther = true;
          break;
        }
    } Variable_CONSUMERS_ENDEACH_OperationArgument;
    if (!isusedfurther) {
      ret = false;
      goto end;
    }
  }
  // Reaching def and strict updward exposed use
 end:
  return ret;
}

#line 1224 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1230 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
static void
prettySet(LiveCheck this, BitSet bitSet, FILE *file)
{
  int32_t *dfnMap = LiveCheck_DFNMAP(this);
  BitSet_FOREACH(bitSet, preDfn) {
    if (1) { // Dump preDfn index
      fprintf(file, "%d ", preDfn);
    } else { // Dump block_index
      fprintf(file, "%d ", dfnMap[preDfn]);
    }
  } BitSet_ENDEACH;
}
#endif//ECL_NTRACE

#line 1249 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
static void
prettyBasicBlock(LiveCheck this, BasicBlock block, uint32_t dfsIndex, FILE *file)
// Pretty print the dotty node of the basic block block.
// Provides:
// - its DFS number (/!\ this is not the label of the node)
// - its basic block index (this is the dotty label of the node)
// - The set of reduced reachable nodes Rq (rset)
// - The set of targets of reduced reachable back-edges Tq^ (hSets)
// - The set of reachable backedge targets not reduced reachable Tq (tset)
// If the BasicBlock is a backedge target, draw in red.
{
  uint32_t block_index = BasicBlock_index(block);
  int32_t block_preDfn = LiveCheck_PREDFN(this)[block_index];
  BitSet block_rset = LiveCheck_RSETS(this) + block_index;
  BitSet block_tset = LiveCheck_TSETS(this) + block_index;
  BitSet block_hSets = LiveCheck_HSETS(this) + block_index;
  fprintf(file, "%d ", block_index); // dotty label is block_index
  fprintf(file, "[label=\""); // what will be dumped in the box
  fprintf(file, "%d (dfs:%d) (idx:%d)\\n", block_preDfn, dfsIndex, BasicBlock_index(block)); // preDfn and dfsindex labels
  fprintf(file, "R="); prettySet(this, block_rset, file); fprintf(file, "\\l");
  fprintf(file, "T="); prettySet(this, block_tset, file); fprintf(file, "\\l");
  fprintf(file, "T^="); prettySet(this, block_hSets, file); fprintf(file, "\\l");
  fprintf(file, "\""); // End of content
  if (BitSet_contains(LiveCheck_BACKHEADS(this), block_index)) {
    fprintf(file, ",color=red");
  }
  fprintf(file, "]\n");
}
#endif//ECL_NTRACE

#line 1284 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
static void
prettyEdge(LiveCheck this, BasicBlock source, BasicBlock dest, const char *attributes, FILE *file)
{
  uint32_t source_index = BasicBlock_index(source); // The dotty label of a node is its BasicBlock_index
  uint32_t dest_index = BasicBlock_index(dest);
  fprintf(file, "%d -> %d [%s]\n", source_index, dest_index, attributes);
}
#endif//ECL_NTRACE

#line 1298 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
static void
DFS4pretty(LiveCheck this, BasicBlock block, uint32_t *_dfsIndex, FILE *file)
// Here we perform the same DFS traversal as for the LiveCheck precomputation
// We print in dotty format edges, and nodes of the CFG.
// We differentiate tree-edges, cross-edges & forward, and back-edges.
{
  int32_t block_index = BasicBlock_index(block);
  uint8_t *dfsFlags = LiveCheck_DFSFLAGS(this);
  uint32_t dfsIndex = *_dfsIndex; // Actual DFS for node
  dfsFlags[block_index] = 1;
  (*_dfsIndex)++; // increments the DFS numbering of nodes.
  prettyBasicBlock(this, block, dfsIndex, file);
  BasicBlock_LEAVE_FOREACH_BasicBlockEdge(block, successor) {
    BasicBlock succ = BasicBlockEdge_destBlock(successor);
    int32_t succ_index = BasicBlock_index(succ);
    if (dfsFlags[succ_index] == 0) {
      DFS4pretty(this, succ, _dfsIndex, file);
      prettyEdge(this, block, succ, "style=bold,color=red,weight=100", file);
      continue;
    }
    if (dfsFlags[succ_index] == 2) {
      // processed: cross-edge or forward-edge
      prettyEdge(this, block, succ, "style=filled", file);
      continue;
    }
    if (dfsFlags[succ_index] == 1 && succ_index != 0) {
      // currently processing it: succ is an ancestor
      // This is a back edge to a block different than r
      prettyEdge(this, succ, block, "style=dotted,dir=back", file);
    }
  } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  dfsFlags[block_index] = 2; // mark processed
}
#endif//ECL_NTRACE

#line 1340 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
void
LiveCheck_prettyCFG(LiveCheck this, FILE *file)
{
  int32_t blockCount = LiveCheck_BLOCKCOUNT(this);
  BasicBlock rootBlock = LiveCheck_ROOTBLOCK(this);
  uint32_t dfsIndex = 0;
  // There are 3 indexing for basic blocks:
  //  - "PREDFN" corresponds to a DFS numbering of the dominance tree
  //  - "dfsIndex" corresponds to a DFS numbering of the CFG. It corresponds to the DFS traversal
  //    used in LiveCheck precomputation. There is no special reason why it should be equal to PREDFN.
  //  - "BasicBlock_index" corresponds to an indexing of basic blocks as nodes of the CFG. This
  //    indexing probably corresponds to the one given by Open64...?
  //  - For each BasicBlock, rset[block_index], tset[block_index] & hSets[block_index] are BitSet(s) indexed using
  //    PREDFN.
  fprintf(file, "digraph G {\n");
  memset(LiveCheck_DFSFLAGS(this), 0, blockCount*sizeof(uint8_t));
  DFS4pretty(this, rootBlock, &dfsIndex, file);
  fprintf(file, "}\n");
}
#endif//ECL_NTRACE

#line 1368 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
void
LiveCheck_dumpCFG(LiveCheck this)
{
  FILE *file;
  CodeRegion codeRegion = LiveCheck_CODEREGION(this);
  Procedure procedure = CodeRegion_procedure(codeRegion);
  const char *procname = Procedure_name(procedure);
  char *filename = (char *)Memory_alloc(Memory_Root, strlen(procname) + 4 + 1);
  strcpy(filename, procname); strcat(filename, ".dot");
  LiveCheck_TRACE && fprintf(LiveCheck_TRACE, "Output dotty file into %s\n", filename);
  file = fopen(filename, "w");
  Except_ALWAYS(file != NULL);
  LiveCheck_prettyCFG(this, file);
  fclose(file);
  Memory_free(Memory_Root, filename);
}
#endif//ECL_NTRACE

#line 1387 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
/////////////////////////////////////////////////////////////////////////////////////////
//@XCC.h
//
// Verify correctness of LiveCheck with respect to the given precomputed liveness.
//
// Preconditions:
// - CodeRegion_globalTable() must be initialized to the set of variables to test.
// - Consumers must be setup (for instance by a call to SSAForm_makeConsumers()) for
// all variables in the global table.
// - Liveness sets in CodeRegion must be updated for all variables in the global table.

#line 1402 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#ifndef ECL_NTRACE
void
LiveCheck_checkLiveness(LiveCheck this)
{
  CodeRegion codeRegion = LiveCheck_CODEREGION(this);
  TemporaryTable globalTable = CodeRegion_globalTable(codeRegion);
  TemporarySet_ *liveInSets = LiveCheck_LIVEINSETS(this);
  TemporarySet_ *liveOutSets = LiveCheck_LIVEOUTSETS(this);
  TemporarySet_ *phiOutSets = LiveCheck_PHIOUTSETS(this);
  TemporarySet_ *phiInSets = LiveCheck_PHIINSETS(this);
  if (PFA_TRACE) {
    CodeRegion_pretty(codeRegion, PFA_TRACE);
    fprintf(PFA_TRACE, "Variables and Consumers\n");
    TemporaryTable_FOREACH_Temporary(globalTable, temporary) {
      if (Temporary_isVariable(temporary)) {
        Temporary_pretty(temporary, PFA_TRACE);
        fprintf(PFA_TRACE, "\n producer: ");
        Operation_pretty(OperationResult_OPERATION(Variable_PRODUCER(Temporary_VARIABLE(temporary))), PFA_TRACE);
        fprintf(PFA_TRACE, " consumer(s): ");
        Variable_CONSUMERS_FOREACH_OperationArgument(Temporary_VARIABLE(temporary), oparg) {
          Operation op = OperationArgument_OPERATION(oparg);
          Operation_pretty(op, PFA_TRACE);
        } Variable_CONSUMERS_ENDEACH_OperationArgument;
      }
    } TemporaryTable_ENDEACH_Temporary;
    fprintf(PFA_TRACE, "\n");
  }
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    int32_t index = BasicBlock_index(basicBlock);
    PFA_TRACE && fprintf(PFA_TRACE, "BB: %d\nlive-in:", index)
              && TemporarySet_pretty(liveInSets + index,PFA_TRACE)
              && fprintf(PFA_TRACE, "\nphi-in:")
              && TemporarySet_pretty(phiInSets + index,PFA_TRACE)
              && fprintf(PFA_TRACE, "\nlive-out:")
              && TemporarySet_pretty(liveOutSets + index,PFA_TRACE)
              && fprintf(PFA_TRACE, "\nphi-out:")
              && TemporarySet_pretty(phiOutSets + index,PFA_TRACE)
              && fprintf(PFA_TRACE, "\n");
    TemporaryTable_FOREACH_Temporary(globalTable, temporary) {
      if (Temporary_isVariable(temporary)) {
        Variable variable = Temporary_VARIABLE(temporary);
        //0 && PFA_TRACE && fprintf(PFA_TRACE, "BB: %d:", index) && Temporary_pretty(temporary, PFA_TRACE) && fprintf(PFA_TRACE, "\n");
        {
          /* Livein and Liveout check on basic block boundaries */
          bool isLiveincheck_copy = LiveCheck_isSSALiveInBasicBlock_copy(this, basicBlock, variable);
          bool isLiveoutcheck_copy = LiveCheck_isSSALiveOutBasicBlock_copy(this, basicBlock, variable);
          bool isLiveincheck_mixed = LiveCheck_isSSALiveInBasicBlock_mixed(this, basicBlock, variable);
          bool isLiveoutcheck_mixed = LiveCheck_isSSALiveOutBasicBlock_mixed(this, basicBlock, variable);
          bool isLiveincheck_mplex = LiveCheck_isSSALiveInBasicBlock_mplex(this, basicBlock, variable);
          bool isLiveoutcheck_mplex = LiveCheck_isSSALiveOutBasicBlock_mplex(this, basicBlock, variable);
          bool isLiveinset_copy = LiveCheck_isSetLiveInBasicBlock_copy(this, basicBlock, variable);
          bool isLiveoutset_copy = LiveCheck_isSetLiveOutBasicBlock_copy(this, basicBlock, variable);
          bool isLiveinset_mixed = LiveCheck_isSetLiveInBasicBlock_mixed(this, basicBlock, variable);
          bool isLiveoutset_mixed = LiveCheck_isSetLiveOutBasicBlock_mixed(this, basicBlock, variable);
          bool isLiveinset_mplex = LiveCheck_isSetLiveInBasicBlock_mplex(this, basicBlock, variable);
          bool isLiveoutset_mplex = LiveCheck_isSetLiveOutBasicBlock_mplex(this, basicBlock, variable);
	  bool correct = true;
          if (isLiveoutset_copy != isLiveoutcheck_copy) {
            fprintf(stderr, "Live out mismatch (liveset_copy: %d, liveCheck_copy: %d) in block %d, temporary: ", isLiveoutset_copy, isLiveoutcheck_copy, index);
            Temporary_pretty(temporary, stderr); fprintf(stderr, "\n");
	    correct = false;
          }
          if (isLiveoutset_mixed != isLiveoutcheck_mixed) {
            fprintf(stderr, "Live out mismatch (liveset_mixed: %d, liveCheck_mixed: %d) in block %d, temporary: ", isLiveoutset_mixed, isLiveoutcheck_mixed, index);
            Temporary_pretty(temporary, stderr); fprintf(stderr, "\n");
	    correct = false;
          }
          if (isLiveoutset_mplex != isLiveoutcheck_mplex) {
            fprintf(stderr, "Live out mismatch (liveset_mplex: %d, liveCheck_mplex: %d) in block %d, temporary: ", isLiveoutset_mplex, isLiveoutcheck_mplex, index);
            Temporary_pretty(temporary, stderr); fprintf(stderr, "\n");
	    correct = false;
          }
          if (isLiveinset_copy != isLiveincheck_copy) {
            fprintf(stderr, "Live in mismatch (liveset_copy: %d, liveCheck_copy: %d) in block %d, temporary: ", isLiveinset_copy, isLiveincheck_copy, index);
            Temporary_pretty(temporary, stderr); fprintf(stderr, "\n");
	    correct = false;
          }
          if (isLiveinset_mixed != isLiveincheck_mixed) {
            fprintf(stderr, "Live in mismatch (liveset_mixed: %d, liveCheck_mixed: %d) in block %d, temporary: ", isLiveinset_mixed, isLiveincheck_mixed, index);
            Temporary_pretty(temporary, stderr); fprintf(stderr, "\n");
	    correct = false;
          }
          if (isLiveinset_mplex != isLiveincheck_mplex) {
            fprintf(stderr, "Live in mismatch (liveset_mplex: %d, liveCheck_mplex: %d) in block %d, temporary: ", isLiveinset_mplex, isLiveincheck_mplex, index);
            Temporary_pretty(temporary, stderr); fprintf(stderr, "\n");
	    correct = false;
          }
	  Except_CHECK(correct);
        }
        {
          /* Livein and Liveout check at operations boundaries */
          BasicBlock_FOREACH_Operation(basicBlock, operation) {
	    {
	      bool isLiveoutcheck_mplex = LiveCheck_isSSALiveOutOperation_mplex(this, operation, variable);
	      bool isLiveoutset_mplex = LiveCheck_isSetLiveOutOperation_mplex(this, operation, variable);
	      bool isLiveoutcheck_copy = LiveCheck_isSSALiveOutOperation_copy(this, operation, variable);
	      bool isLiveoutset_copy = LiveCheck_isSetLiveOutOperation_copy(this, operation, variable);
	      bool isLiveoutcheck_mixed = LiveCheck_isSSALiveOutOperation_mixed(this, operation, variable);
	      bool isLiveoutset_mixed = LiveCheck_isSetLiveOutOperation_mixed(this, operation, variable);
	      bool correct = true;
	      if (isLiveoutcheck_mplex != isLiveoutset_mplex) {
		fprintf(stderr, "Live out mismatch (liveoutcheck_mplex: %d, liveoutset_mplex: %d) for Temporary: ", \
			isLiveoutcheck_mplex, isLiveoutset_mplex);
		Temporary_pretty(temporary, stderr);
		fprintf(stderr, " in block %d, operation: [", index);
		Operation_pretty(operation, stderr);
		fprintf(stderr, "]\n");
		correct = false;
	      }
	      if (isLiveoutcheck_copy != isLiveoutset_copy) {
		fprintf(stderr, "Live out mismatch (liveoutcheck_copy: %d, liveoutset_copy: %d) for Temporary: ", \
			isLiveoutcheck_copy, isLiveoutset_copy);
		Temporary_pretty(temporary, stderr);
		fprintf(stderr, " in block %d, operation: [", index);
		Operation_pretty(operation, stderr);
		fprintf(stderr, "]\n");
		correct = false;
	      }
	      if (isLiveoutcheck_mixed != isLiveoutset_mixed) {
		fprintf(stderr, "Live out mismatch (liveoutcheck_mixed: %d, liveoutset_mixed: %d) for Temporary: ", \
			isLiveoutcheck_mixed, isLiveoutset_mixed);
		Temporary_pretty(temporary, stderr);
		fprintf(stderr, " in block %d, operation: [", index);
		Operation_pretty(operation, stderr);
		fprintf(stderr, "]\n");
		correct = false;
	      }
	      if (Operator_isPhi(Operation_operator(operation))) continue; // Livein for phi is undefined
	      Except_CHECK(!Operator_isSigma(Operation_operator(operation))); // SSI is not supported

#line 1532 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
	      {
		bool isLiveincheck_mplex = LiveCheck_isSSALiveInOperation_mplex(this, operation, variable);
		bool isLiveinset_mplex = LiveCheck_isSetLiveInOperation_mplex(this, operation, variable);
		bool isLiveincheck_copy = LiveCheck_isSSALiveInOperation_copy(this, operation, variable);
		bool isLiveinset_copy = LiveCheck_isSetLiveInOperation_copy(this, operation, variable);
		bool isLiveincheck_mixed = LiveCheck_isSSALiveInOperation_mixed(this, operation, variable);
		bool isLiveinset_mixed = LiveCheck_isSetLiveInOperation_mixed(this, operation, variable);
		bool correct = true;
		if (isLiveincheck_mplex != isLiveinset_mplex) {
		  fprintf(stderr, "Live in mismatch (liveincheck_mplex: %d, liveinset_mplex: %d) for Temporary: ", \
			  isLiveincheck_mplex, isLiveinset_mplex);
		  Temporary_pretty(temporary, stderr);
		  fprintf(stderr, " in block %d, operation: [", index);
		  Operation_pretty(operation, stderr);
		  fprintf(stderr, "]\n");
		  correct = false;
		}
		if (isLiveincheck_copy != isLiveinset_copy) {
		  fprintf(stderr, "Live in mismatch (liveincheck_copy: %d, liveinset_copy: %d) for Temporary: ", \
			  isLiveincheck_copy, isLiveinset_copy);
		  Temporary_pretty(temporary, stderr);
		  fprintf(stderr, " in block %d, operation: [", index);
		  Operation_pretty(operation, stderr);
		  fprintf(stderr, "]\n");
		  correct = false;
		}
		if (isLiveincheck_mixed != isLiveinset_mixed) {
		  fprintf(stderr, "Live in mismatch (liveincheck_mixed: %d, liveinset_mixed: %d) for Temporary: ", \
			  isLiveincheck_mixed, isLiveinset_mixed);
		  Temporary_pretty(temporary, stderr);
		  fprintf(stderr, " in block %d, operation: [", index);
		  Operation_pretty(operation, stderr);
		  fprintf(stderr, "]\n");
		  correct = false;
		}
		Except_CHECK(correct);
	      }
            }
          } BasicBlock_ENDBACK_Operation;
        }
      } /*endif (Temporary_isVariable(temporary)) */
    } TemporaryTable_ENDEACH_Temporary;
  } CodeRegion_ENDEACH_BasicBlock;
}
#else//ECL_NTRACE
void
LiveCheck_checkLiveness(LiveCheck this) { }
#endif//ECL_NTRACE

#line 1581 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1864 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1881 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
void
LiveCheck_INIT(void)
{
#ifndef LiveCheck_LIVENESS
  if (GETENV("LiveCheck_LIVENESS")) {
    sscanf(GETENV("LiveCheck_LIVENESS"), "%d", &LiveCheck_LIVENESS);
    fprintf(stderr, "*** LiveCheck_LIVENESS=%d\n", LiveCheck_LIVENESS);
  }
#endif//LiveCheck_LIVENESS
#ifndef ECL_NTRACE
  if (GETENV("PFA_LiveCheck_TRACE")) {
    int file = 0;
    sscanf(GETENV("PFA_LiveCheck_TRACE"), "%d", &file);
    if (file == 1) LiveCheck_TRACE = stdout;
    else if (file == 2) LiveCheck_TRACE = stderr;
  }
#endif
}

#line 1906 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
void
LiveCheck_FINI(void)
{
}

#line 1911 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
#if XCC__C
static void
LiveCheck_TEST(void)
{
#include "PFA/LiveCheck_.c"
}

#line 1918 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/ECL/PFA/LiveCheck.xcc"
int
main(int argc, char **argv)
{
  PFA_Test(LiveCheck_TEST);
  return 0;
}
#endif


#endif//LiveCheck_c_INCLUDED
