#ifndef LiveCheck__h_INCLUDED
#define LiveCheck__h_INCLUDED

//
#include "PFA/LiveCheck.h"

//
struct LiveCheck_ {
  //@args	Memory memory, Dominance dominance, Liveness liveness
  Memory MEMORY;
  int32_t BLOCKCOUNT;
  Dominance DOMINANCE;
  //@access CODEREGION	Dominance_CODEREGION(LiveCheck_DOMINANCE(this))
  //@access PREDFN	Dominance_PREDFN(LiveCheck_DOMINANCE(this))
  //@access LASTDFN	Dominance_LASTDFN(LiveCheck_DOMINANCE(this))
  //@access BLOCKMAP	Dominance_BLOCKMAP(LiveCheck_DOMINANCE(this))
  //@access ROOTBLOCK	(LiveCheck_BLOCKMAP(this)[0])
  uint8_t *DFSFLAGS;	// 0: not processed yet; 1: currently processing it; 2: processed.
  int32_t *DFNMAP;	// Map PREDFN to BasicBlock index: DFNMAP = PREDFN^(-1).
  BitSet_ BACKHEADS[1];	// Heads of back-edges.
  BitSet_ *RSETS;	// Reduced reachable nodes (as opposed to the paper a node is not reduced-reachable from itself)
  BitSet_ *HSETS;	// Heads of reduced reachable back-edges: Tt^ (as opposed to the paper Tt^ can contain t)
  BitSet_ *TSETS;	// Sets of highest nodes: Tq
  TemporarySet_ *LIVEINSETS;	// Live-in sets from Liveness.
  TemporarySet_ *LIVEOUTSETS;	// Live-out sets from Liveness.
  TemporarySet_ *PHIINSETS;	// Phi-in sets from Liveness.
  TemporarySet_ *PHIOUTSETS;	// Phi-out sets from Liveness.
};

extern LiveCheck
LiveCheck_Ctor(LiveCheck this, Memory memory, Dominance dominance, Liveness liveness);

extern LiveCheck
LiveCheck_Copy(LiveCheck this, const_LiveCheck that);

extern void
LiveCheck_Dtor(LiveCheck this);

extern size_t
LiveCheck_Size(Memory memory, Dominance dominance, Liveness liveness);

#define LiveCheck_MEMORY(this) ((this)->MEMORY)
#define LiveCheck__MEMORY(this) (&(this)->MEMORY)
#define LiveCheck_BLOCKCOUNT(this) ((this)->BLOCKCOUNT)
#define LiveCheck__BLOCKCOUNT(this) (&(this)->BLOCKCOUNT)
#define LiveCheck_DOMINANCE(this) ((this)->DOMINANCE)
#define LiveCheck__DOMINANCE(this) (&(this)->DOMINANCE)
#define LiveCheck_DFSFLAGS(this) ((this)->DFSFLAGS)
#define LiveCheck__DFSFLAGS(this) (&(this)->DFSFLAGS)
#define LiveCheck_DFNMAP(this) ((this)->DFNMAP)
#define LiveCheck__DFNMAP(this) (&(this)->DFNMAP)
#define LiveCheck_BACKHEADS(this) ((this)->BACKHEADS)
#define LiveCheck__BACKHEADS(this) (&(this)->BACKHEADS)
#define LiveCheck_RSETS(this) ((this)->RSETS)
#define LiveCheck__RSETS(this) (&(this)->RSETS)
#define LiveCheck_HSETS(this) ((this)->HSETS)
#define LiveCheck__HSETS(this) (&(this)->HSETS)
#define LiveCheck_TSETS(this) ((this)->TSETS)
#define LiveCheck__TSETS(this) (&(this)->TSETS)
#define LiveCheck_LIVEINSETS(this) ((this)->LIVEINSETS)
#define LiveCheck__LIVEINSETS(this) (&(this)->LIVEINSETS)
#define LiveCheck_LIVEOUTSETS(this) ((this)->LIVEOUTSETS)
#define LiveCheck__LIVEOUTSETS(this) (&(this)->LIVEOUTSETS)
#define LiveCheck_PHIINSETS(this) ((this)->PHIINSETS)
#define LiveCheck__PHIINSETS(this) (&(this)->PHIINSETS)
#define LiveCheck_PHIOUTSETS(this) ((this)->PHIOUTSETS)
#define LiveCheck__PHIOUTSETS(this) (&(this)->PHIOUTSETS)
#define LiveCheck_BLOCKMAP(this) (Dominance_BLOCKMAP(LiveCheck_DOMINANCE(this)))
#define LiveCheck_ROOTBLOCK(this) ((LiveCheck_BLOCKMAP(this)[0]))
#define LiveCheck_CODEREGION(this) (Dominance_CODEREGION(LiveCheck_DOMINANCE(this)))
#define LiveCheck_PREDFN(this) (Dominance_PREDFN(LiveCheck_DOMINANCE(this)))
#define LiveCheck_LASTDFN(this) (Dominance_LASTDFN(LiveCheck_DOMINANCE(this)))


//
static void
firstDFS4precompute(LiveCheck this, BasicBlock block);

//
static void
secondDFS4precompute(LiveCheck this, BasicBlock block);

//
void
LiveCheck_checkLiveness(LiveCheck this);

/*
 * LiveCheck_TRACE -- File handle used to print the PFA_LiveCheck traces.
 */
#ifndef ECL_NTRACE
static FILE *LiveCheck_TRACE;
#else
#define LiveCheck_TRACE NULL
#endif//ECL_NTRACE

/*
 * LiveCheck_INIT --	Initialize the LiveCheck module.
 */
void
LiveCheck_INIT(void);

/*
 * LiveCheck_FINI --	Finalize the LiveCheck module.
 */
void
LiveCheck_FINI(void);


#endif//LiveCheck__h_INCLUDED
