#ifndef Liveness__h_INCLUDED
#define Liveness__h_INCLUDED

//
#include "PFA/Liveness.h"

/*
 * LivenessState --	Liveness state for a BasicBlock.
 */
struct LivenessState_ {
  //@args	Memory memory, int32_t temporaryCount,
  //@args	TemporarySet liveInSet, TemporarySet liveOutSet
  BasicBlock BASICBLOCK;	// BasicBlock of this LivenessState.
  bool ISLISTED;		// This LivenessState is in Liveness worklist.
  struct LivenessState_ *WORKNEXT;
  TemporarySet_ USEINSET[1];	// The use-in TemporarySet.
  TemporarySet_ KILLEDSET[1];	// The killed TemporarySet.
  TemporarySet LIVEINSET;	// The live-in TemporarySet.
  TemporarySet LIVEOUTSET;	// The live-out TemporarySet.
};
typedef struct LivenessState_ LivenessState_, *LivenessState;
typedef const struct LivenessState_ *const_LivenessState;
typedef struct LivenessState_ * restrict_LivenessState;

extern LivenessState
LivenessState_Ctor(LivenessState this, Memory memory, int32_t temporaryCount, TemporarySet liveInSet, TemporarySet liveOutSet);

extern LivenessState
LivenessState_Copy(LivenessState this, const_LivenessState that);

extern void
LivenessState_Dtor(LivenessState this);

extern size_t
LivenessState_Size(Memory memory, int32_t temporaryCount, TemporarySet liveInSet, TemporarySet liveOutSet);

#define LivenessState_BASICBLOCK(this) ((this)->BASICBLOCK)
#define LivenessState__BASICBLOCK(this) (&(this)->BASICBLOCK)
#define LivenessState_ISLISTED(this) ((this)->ISLISTED)
#define LivenessState__ISLISTED(this) (&(this)->ISLISTED)
#define LivenessState_WORKNEXT(this) ((this)->WORKNEXT)
#define LivenessState__WORKNEXT(this) (&(this)->WORKNEXT)
#define LivenessState_USEINSET(this) ((this)->USEINSET)
#define LivenessState__USEINSET(this) (&(this)->USEINSET)
#define LivenessState_KILLEDSET(this) ((this)->KILLEDSET)
#define LivenessState__KILLEDSET(this) (&(this)->KILLEDSET)
#define LivenessState_LIVEINSET(this) ((this)->LIVEINSET)
#define LivenessState__LIVEINSET(this) (&(this)->LIVEINSET)
#define LivenessState_LIVEOUTSET(this) ((this)->LIVEOUTSET)
#define LivenessState__LIVEOUTSET(this) (&(this)->LIVEOUTSET)


/*
 * LivenessState_analyzeOperations --	Local analysis for this LivenessState.
 *
 * Analyze this LivenessState BasicBlock in order to fill USEINSET and KILLEDSET.
 *
 * Require:	useInSparse is empty except perhaps for PHI arguments of successors.
 */
void
LivenessState_analyzeOperations(LivenessState this,
                                TemporarySparse useInSparse,
                                struct Temporary_ dedicated[]);

/*
 * LivenessState_pretty --	Pretty-print this LivenessState.
 */
bool
LivenessState_pretty(LivenessState this, FILE *file);

//
struct Liveness_ {
  //@args	Memory memory, CodeRegion codeRegion,
  //@args	TemporaryTable temporaryTable, bool phiSets
  Memory MEMORY;
  CodeRegion CODEREGION;
  Temporary_ *DEDICATED;
  int32_t ITERCOUNT;
  int32_t BLOCKCOUNT;
  LivenessState_ *BASESTATE;		// Base of LivenessState array.
  TemporarySet_ *LIVEINSETS;		// LiveInSets array.
  TemporarySet_ *LIVEOUTSETS;		// LiveOutSets array.
  TemporarySet_ *PHIINSETS;		// PhiInSets array.
  TemporarySet_ *PHIOUTSETS;		// PhiOutSets array.
  LivenessState_ *WORKTAIL;		// First LivenessState in worklist.
  LivenessState_ *WORKHEAD;		// Last LivenessState in worklist.
  TemporarySparse_ SPARSESET[1];	// Sparse set for computing BasicBlock use-in.
  //@access TEMPORARYTABLE	TemporarySparse_table(Liveness_SPARSESET(this))
};

extern Liveness
Liveness_Ctor(Liveness this, Memory memory, CodeRegion codeRegion, TemporaryTable temporaryTable, bool phiSets);

extern Liveness
Liveness_Copy(Liveness this, const_Liveness that);

extern void
Liveness_Dtor(Liveness this);

extern size_t
Liveness_Size(Memory memory, CodeRegion codeRegion, TemporaryTable temporaryTable, bool phiSets);

#define Liveness_MEMORY(this) ((this)->MEMORY)
#define Liveness__MEMORY(this) (&(this)->MEMORY)
#define Liveness_CODEREGION(this) ((this)->CODEREGION)
#define Liveness__CODEREGION(this) (&(this)->CODEREGION)
#define Liveness_DEDICATED(this) ((this)->DEDICATED)
#define Liveness__DEDICATED(this) (&(this)->DEDICATED)
#define Liveness_ITERCOUNT(this) ((this)->ITERCOUNT)
#define Liveness__ITERCOUNT(this) (&(this)->ITERCOUNT)
#define Liveness_BLOCKCOUNT(this) ((this)->BLOCKCOUNT)
#define Liveness__BLOCKCOUNT(this) (&(this)->BLOCKCOUNT)
#define Liveness_BASESTATE(this) ((this)->BASESTATE)
#define Liveness__BASESTATE(this) (&(this)->BASESTATE)
#define Liveness_LIVEINSETS(this) ((this)->LIVEINSETS)
#define Liveness__LIVEINSETS(this) (&(this)->LIVEINSETS)
#define Liveness_LIVEOUTSETS(this) ((this)->LIVEOUTSETS)
#define Liveness__LIVEOUTSETS(this) (&(this)->LIVEOUTSETS)
#define Liveness_PHIINSETS(this) ((this)->PHIINSETS)
#define Liveness__PHIINSETS(this) (&(this)->PHIINSETS)
#define Liveness_PHIOUTSETS(this) ((this)->PHIOUTSETS)
#define Liveness__PHIOUTSETS(this) (&(this)->PHIOUTSETS)
#define Liveness_WORKTAIL(this) ((this)->WORKTAIL)
#define Liveness__WORKTAIL(this) (&(this)->WORKTAIL)
#define Liveness_WORKHEAD(this) ((this)->WORKHEAD)
#define Liveness__WORKHEAD(this) (&(this)->WORKHEAD)
#define Liveness_SPARSESET(this) ((this)->SPARSESET)
#define Liveness__SPARSESET(this) (&(this)->SPARSESET)
#define Liveness_TEMPORARYTABLE(this) (TemporarySparse_table(Liveness_SPARSESET(this)))


/*
 * Liveness_baseState --	For use in Liveness_FOREACH.
 */
LivenessState_ *
Liveness_baseState(const_Liveness this);

/*
 * Liveness_FOREACH_LivenessState --	Iterate over this Liveness LivenessState.
 */
#define Liveness_FOREACH_LivenessState(this, state) { \
  int32_t Liveness_BLOCKCOUNT = Liveness_blockCount(this); \
  LivenessState Liveness_BASESTATE = Liveness_baseState(this), state; \
  LivenessState Liveness_PASTSTATE = Liveness_BASESTATE + Liveness_BLOCKCOUNT; \
  for (state = Liveness_BASESTATE; state < Liveness_PASTSTATE; ++state) { \
    if (LivenessState_BASICBLOCK(state) != NULL) {
#define Liveness_ENDEACH_LivenessState \
    } \
  } \
}

//
#define Liveness_FORWORK_LivenessState(this, state) { \
  LivenessState Liveness_WORKNEXT = NULL; \
  LivenessState state = Liveness_WORKHEAD(this); \
  for (; state != NULL; state = Liveness_WORKNEXT) { \
    Liveness_WORKNEXT = LivenessState_WORKNEXT(state);
#define Liveness_ENDWORK_LivenessState \
  } \
}

/*
 * Liveness_activateState --	Activate Liveness for a BasicBlock.
 *
 * We put first on the worklist the block in post-order, so they are pop(ed) in
 * post-order. This is a recommended ordering for backward data-flow problems.
 */
void
Liveness_activateState(Liveness this, BasicBlock curr_block, bool phiSets);

/*
 * Liveness_doIterate --	Iterate this Liveness problem until fixpoint.
 *
 * This is the worklist iterative algorithm to solve backward data-flow problems.
 * The activated blocks are pushed on top of the worklist, and the worklist is
 * pop(ed) to get the current block. This enables the data-flow information to
 * stabilize in loops before proceeding further in the flow graph.
 */
void
Liveness_doIterate(Liveness this);

/*
 * Liveness_INIT --	Initialize the Liveness module.
 */
#define Liveness_INIT()

/*
 * Liveness_FINI --	Finalize the Liveness module.
 */
#define Liveness_FINI()


#endif//Liveness__h_INCLUDED
