#ifndef LoopForest_h_INCLUDED
#define LoopForest_h_INCLUDED

/*
 * !!!!	LoopForest
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * LoopForestNodeType --	Enumerate the LoopForestNode types.
 */
typedef enum {
  LoopForestNodeType_Block,
  LoopForestNodeType_Root,
  LoopForestNodeType_SCC,
  LoopForestNodeType_Loop,
  LoopForestNodeType__
} LoopForestNodeType;
typedef uint8_t short_LoopForestNodeType;

/*
 * LoopForestNode --	Region in a LoopForest of a DGraph.
 */
struct LoopForestNode_ {
  //@args	LoopForestNodeType type, BasicBlock block
  short_LoopForestNodeType TYPE;	// This LoopForestNode type.
  int16_t DEPTH;			// Nesting depth of this LoopForestNode.
  BasicBlock BLOCK;			// BasicBlock for this LoopForestNode.
  struct LoopForestNode_ *CHILD;	// First child of this LoopForestNode.
  struct LoopForestNode_ *NEXT;		// Next child in parent LoopForestNode.
};
typedef struct LoopForestNode_ LoopForestNode_, *LoopForestNode;
typedef const struct LoopForestNode_ *const_LoopForestNode;
typedef struct LoopForestNode_ * restrict_LoopForestNode;

extern LoopForestNode
LoopForestNode_Ctor(LoopForestNode this, LoopForestNodeType type, BasicBlock block);

extern LoopForestNode
LoopForestNode_Copy(LoopForestNode this, const_LoopForestNode that);

extern void
LoopForestNode_Dtor(LoopForestNode this);

extern size_t
LoopForestNode_Size(LoopForestNodeType type, BasicBlock block);

#define LoopForestNode_TYPE(this) ((this)->TYPE)
#define LoopForestNode__TYPE(this) (&(this)->TYPE)
#define LoopForestNode_DEPTH(this) ((this)->DEPTH)
#define LoopForestNode__DEPTH(this) (&(this)->DEPTH)
#define LoopForestNode_BLOCK(this) ((this)->BLOCK)
#define LoopForestNode__BLOCK(this) (&(this)->BLOCK)
#define LoopForestNode_CHILD(this) ((this)->CHILD)
#define LoopForestNode__CHILD(this) (&(this)->CHILD)
#define LoopForestNode_NEXT(this) ((this)->NEXT)
#define LoopForestNode__NEXT(this) (&(this)->NEXT)


//
static inline LoopForestNodeType
LoopForestNode_type(const_LoopForestNode this)
{
  return LoopForestNode_TYPE(this);
}

//
static inline int
LoopForestNode_depth(const_LoopForestNode this)
{
  return LoopForestNode_DEPTH(this);
}

//
static inline BasicBlock
LoopForestNode_block(const_LoopForestNode this)
{
  return LoopForestNode_BLOCK(this);
}

//
static inline LoopForestNode
LoopForestNode_child(const_LoopForestNode this)
{
  return LoopForestNode_CHILD(this);
}

//
static inline LoopForestNode
LoopForestNode_next(const_LoopForestNode this)
{
  return LoopForestNode_NEXT(this);
}

//
#define LoopForestNode_CONTENTS_FOREACH_LoopForestNode(this, node) { \
  LoopForestNode node = LoopForestNode_child(this); \
  for (; node != NULL; node = LoopForestNode_next(node)) {
#define LoopForestNode_CONTENTS_ENDEACH_LoopForestNode \
  } \
}

//
bool
LoopForestNode_pretty(const_LoopForestNode this, FILE *file,
                      bool (*block_pretty)(BasicBlock, FILE*));

/*
 * LoopForest --	Loop forest for a CodeRegion.
 */
struct LoopForest_;
typedef struct LoopForest_ LoopForest_, *LoopForest;
typedef const struct LoopForest_ *const_LoopForest;
typedef struct LoopForest_ * restrict_LoopForest;

//
LoopForest
LoopForest_new(Memory parent, CodeRegion codeRegion);

//
LoopForest
LoopForest_delete(LoopForest this);

/*
 * LoopForest_memory --	This LoopForest Memory.
 */
Memory
LoopForest_memory(LoopForest this);

/*
 * LoopForest_codeRegion --	This LoopForest CodeRegion.
 */
CodeRegion
LoopForest_codeRegion(LoopForest this);

//
LoopForestNode
LoopForest_rootNode(LoopForest this);

//
bool
LoopForest_pretty(const_LoopForest this, FILE *file,
                  bool (*block_pretty)(BasicBlock, FILE*));


#endif//LoopForest_h_INCLUDED
