#ifndef LoopForest__h_INCLUDED
#define LoopForest__h_INCLUDED

//
#include "PFA/LoopForest.h"

//
struct LoopForest_ {
  //@args	Memory memory, CodeRegion codeRegion
  CodeRegion CODEREGION;	// This LoopForest CodeRegion.
  //@access BLOCKCOUNT	CodeRegion_blockCount(LoopForest_CODEREGION(this))
  Configure CONFIGURE;
  int32_t *PREDFN;		// Array from BasicBlock INDEX to preorder DFS numbers.
  BasicBlock *PAST;		// Array from BasicBlock INDEX to loop head BasicBlock.
  LoopForestNode_ ROOTNODE[1];	// Root of this LoopForest.
  IBList_ NODES[1];		// Container for the LoopForestNode(s).
  //@access MEMORY	IBList_memory(LoopForest_NODES(this))
};

extern LoopForest
LoopForest_Ctor(LoopForest this, Memory memory, CodeRegion codeRegion);

extern LoopForest
LoopForest_Copy(LoopForest this, const_LoopForest that);

extern void
LoopForest_Dtor(LoopForest this);

extern size_t
LoopForest_Size(Memory memory, CodeRegion codeRegion);

#define LoopForest_CODEREGION(this) ((this)->CODEREGION)
#define LoopForest__CODEREGION(this) (&(this)->CODEREGION)
#define LoopForest_CONFIGURE(this) ((this)->CONFIGURE)
#define LoopForest__CONFIGURE(this) (&(this)->CONFIGURE)
#define LoopForest_PREDFN(this) ((this)->PREDFN)
#define LoopForest__PREDFN(this) (&(this)->PREDFN)
#define LoopForest_PAST(this) ((this)->PAST)
#define LoopForest__PAST(this) (&(this)->PAST)
#define LoopForest_ROOTNODE(this) ((this)->ROOTNODE)
#define LoopForest__ROOTNODE(this) (&(this)->ROOTNODE)
#define LoopForest_NODES(this) ((this)->NODES)
#define LoopForest__NODES(this) (&(this)->NODES)
#define LoopForest_MEMORY(this) (IBList_memory(LoopForest_NODES(this)))
#define LoopForest_BLOCKCOUNT(this) (CodeRegion_blockCount(LoopForest_CODEREGION(this)))


/*
 * LoopForest_makeSCC --	Make a LoopForest of Strongly Connected Components.
 */
void
LoopForest_makeSCC(LoopForest this);

/*
 * LoopForest_makeNSCC --	Make a LoopForest of Nested Strongly Connected Components.
 *
 * The Nested Strongly Connected Components are a for of Loop Forest as defined by
 * G. Ramalingam, ``Identifying loops in almost linear time'', Transactions on
 * Programming Languages and Systems, Volume 21, Number 2, March 1999. F.
 * Bourdoncle introduced this decomposition in his Ph. D. thesis in order to
 * minimize the number of widening points in a control-flow graph.
 */
void
LoopForest_makeNSCC(LoopForest this);

/*
 * LoopForest_INIT --	Initialize the LoopForest module.
 */
#define LoopForest_INIT()

/*
 * LoopForest_FINI --	Finalize the LoopForest module.
 */
#define LoopForest_FINI()


#endif//LoopForest__h_INCLUDED
