#ifndef SSAConstruct__h_INCLUDED
#define SSAConstruct__h_INCLUDED

//
#include "PFA/SSAConstruct.h"

//
//#define SSAConstruct_SPILLABLE 0
#ifndef SSAConstruct_SPILLABLE
extern int SSAConstruct_SPILLABLE;
#endif//SSAConstruct_SPILLABLE

//
struct SSAConstruct_ {
  //@args	Memory memory, SSAForm ssaForm
  CodeRegion CODEREGION;		// The CodeRegion this SSAConstruct is for.
  //@access SSAFORM	CodeRegion_ssaForm(SSAConstruct_CODEREGION(this))
  TemporaryTable GLOBALTABLE;		// The global Temporary(ies) considered.
  Dominance DOMINANCE;			// Pre-dominance with frontiers or merge sets.
  BasicBlockStack_ *DEFBLOCKSMAP;	// Global Temporary INDEX to def BasicBlock(s).
  TemporaryStack_ *GLOBALRENAMES;	// Stacks to rename global SSA Temporary(ies).
  TemporaryStack_ LOCALRENAMES[1];	// Maps Temporary_INDEX to local SSA Temporary.
  TemporaryStack_ GLOBALPUSHES[1];	// Temporary_INDEX to local SSA Temporary.
  bool *APHISET;			// Array indexed by BasicBlock_index.
  bool **KILLSETS;			// Array of array indexed by BasicBlock_index.
  OperationStack_ WORKLIST[1];		// Operation worklist for dead code elimination.
  //@access MEMORY	OperationStack_memory(SSAConstruct_WORKLIST(this))
};

extern SSAConstruct
SSAConstruct_Ctor(SSAConstruct this, Memory memory, SSAForm ssaForm);

extern SSAConstruct
SSAConstruct_Copy(SSAConstruct this, const_SSAConstruct that);

extern void
SSAConstruct_Dtor(SSAConstruct this);

extern size_t
SSAConstruct_Size(Memory memory, SSAForm ssaForm);

#define SSAConstruct_CODEREGION(this) ((this)->CODEREGION)
#define SSAConstruct__CODEREGION(this) (&(this)->CODEREGION)
#define SSAConstruct_GLOBALTABLE(this) ((this)->GLOBALTABLE)
#define SSAConstruct__GLOBALTABLE(this) (&(this)->GLOBALTABLE)
#define SSAConstruct_DOMINANCE(this) ((this)->DOMINANCE)
#define SSAConstruct__DOMINANCE(this) (&(this)->DOMINANCE)
#define SSAConstruct_DEFBLOCKSMAP(this) ((this)->DEFBLOCKSMAP)
#define SSAConstruct__DEFBLOCKSMAP(this) (&(this)->DEFBLOCKSMAP)
#define SSAConstruct_GLOBALRENAMES(this) ((this)->GLOBALRENAMES)
#define SSAConstruct__GLOBALRENAMES(this) (&(this)->GLOBALRENAMES)
#define SSAConstruct_LOCALRENAMES(this) ((this)->LOCALRENAMES)
#define SSAConstruct__LOCALRENAMES(this) (&(this)->LOCALRENAMES)
#define SSAConstruct_GLOBALPUSHES(this) ((this)->GLOBALPUSHES)
#define SSAConstruct__GLOBALPUSHES(this) (&(this)->GLOBALPUSHES)
#define SSAConstruct_APHISET(this) ((this)->APHISET)
#define SSAConstruct__APHISET(this) (&(this)->APHISET)
#define SSAConstruct_KILLSETS(this) ((this)->KILLSETS)
#define SSAConstruct__KILLSETS(this) (&(this)->KILLSETS)
#define SSAConstruct_WORKLIST(this) ((this)->WORKLIST)
#define SSAConstruct__WORKLIST(this) (&(this)->WORKLIST)
#define SSAConstruct_MEMORY(this) (OperationStack_memory(SSAConstruct_WORKLIST(this)))
#define SSAConstruct_SSAFORM(this) (CodeRegion_ssaForm(SSAConstruct_CODEREGION(this)))


/*
 * SSAConstruct_makeDefBlocksMap --	Make the DEFBLOCKSMAP and other structures.
 */
void
SSAConstruct_makeDefBlocksMap(SSAConstruct this);

/*
 * SSAConstruct_placePhiFunctions --	SSA placement of PHI functions.
 *
 * See Appel's book "Modern Compiler Implementation in C" for !SSAForm_MERGESETS.
 *
 * See "A Practical and Fast Iterative Algorithm for PHI-Function Computation Using
 * DJ Graphs", D. Das, U. Ramakrishna, TOPLAS 27, 3 for SSAForm_MERGESETS.
 */
void
SSAConstruct_placePhiFunctions(SSAConstruct this);

/*
 * SSAConstruct_renameBasicBlockTemporaries --	Replace the SSA variables for a BasicBlock.
 *
 * This is the traditional SSA renaming with segregation between global and local
 * Temporary(ies), as rename stacks are only needed for global Temporary(ies).
 */
void
SSAConstruct_renameBasicBlockTemporaries(SSAConstruct this,
                                         BasicBlock basicBlock,
                                         OptimizeFlags conversion);

/*
 * SSAConstruct_removeDeadCode --	Remove dead Operation(s) with a worklist algorithm.
 */
void
SSAConstruct_removeDeadCode(SSAConstruct this);

/*
 * SSAConstruct_INIT --	Initialize the SSAConstruct module.
 */
#if defined(SSAConstruct_SPILLABLE)
#define SSAConstruct_INIT()
#else
void
SSAConstruct_INIT(void);
#endif

/*
 * SSAConstruct_FINI --	Finalize the SSAConstruct module.
 */
#define SSAConstruct_FINI()


#endif//SSAConstruct__h_INCLUDED
