#ifndef SSAForm_h_INCLUDED
#define SSAForm_h_INCLUDED

/*
 * !!!!	SSAForm.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 * Fabrice Rastello (Fabrice.Rastello@ens-lyon.fr).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * SSAForm --	SSA Form for a CodeRegion.
 */
struct SSAForm_;
typedef struct SSAForm_ SSAForm_, *SSAForm;
typedef const struct SSAForm_ *const_SSAForm;
typedef struct SSAForm_ * restrict_SSAForm;

//
SSAForm
SSAForm_construct(Memory parent, CodeRegion codeRegion, bool sortEnterEdges);

//
SSAForm
SSAForm_destruct(SSAForm this);

/*
 * SSAForm_makeConventional --	Make this SSAForm conventional.
 */
void
SSAForm_makeConventional(SSAForm this, TemporaryTable temporaryTable);

/*
 * SSAForm_memory --	This SSAForm Memory.
 */
Memory
SSAForm_memory(SSAForm this);

/*
 * SSAForm_codeRegion --	This SSAForm CodeRegion.
 */
CodeRegion
SSAForm_codeRegion(SSAForm this);

/*
 * SSAForm_makeOrderings --	Make Operation ordering for this SSAForm.
 */
void
SSAForm_makeOrderings(SSAForm this, int32_t increment);

/*
 * SSAForm_makeDominance --	Make pre-dominance for this SSAForm.
 */
Dominance
SSAForm_makeDominance(SSAForm this, DominanceFlags flags);

/*
 * SSAForm_killDominance --	Kill this SSAForm pre-dominance.
 *
 * Must be called to invalidate DOMINANCE after the control-flow graph is changed.
 */
void
SSAForm_killDominance(SSAForm this);

/*
 * SSAForm_variables --	For use by SSAForm_FOREACH_Variable.
 */
IBList
SSAForm_variables(SSAForm this);

/*
 * SSAForm_FOREACH_Variable --	Iterate this SSAForm Variable(s).
 */
#define SSAForm_FOREACH_Variable(this, variable) \
  IBList_FOREACH(SSAForm_variables(this), Variable_, variable)
#define SSAForm_ENDEACH_Variable \
  IBList_ENDEACH;

/*
 * SSAForm_isStrict --	Verify that this SSAForm is strict.
 *
 * Require:	SSAForm_makeDominance updated.
 */
bool
SSAForm_isStrict(const_SSAForm this);

/*
 * SSAForm_fillTemporaryTable --	Fill the TemporaryTable with Variable(s).
*/
void
SSAForm_fillTemporaryTable(SSAForm this, TemporaryTable table);

/*
 * SSAForm_checkProducerConsumers --	Check validity Variable PRODUCER and CONSUMER.
 */
bool
SSAForm_checkOrderings(SSAForm this);

/*
 * SSAForm_checkProducerConsumers --	Check validity Variable PRODUCER and CONSUMER.
 */
bool
SSAForm_checkProducerConsumers(SSAForm this);

/*
 * SSAForm_pretty --	Pretty-print this SSAForm.
 */
bool
SSAForm_pretty(SSAForm this, FILE *file);

/*
 * SSAForm_display --	Display this SSAForm in VCG format.
 */
bool
SSAForm_display(SSAForm this, FILE *file, const char *name);


#endif//SSAForm_h_INCLUDED
