#ifndef SSAForm__h_INCLUDED
#define SSAForm__h_INCLUDED

//
#include "PFA/SSAForm.h"

//
#define SSAForm_MERGESETS 0
#ifndef SSAForm_MERGESETS
extern int SSAForm_MERGESETS;
#endif//SSAForm_MERGESETS

//
struct SSAForm_ {
  //@args	Memory memory, CodeRegion codeRegion
  Memory MEMORY;		// This SSAForm Memory.
  CodeRegion CODEREGION;	// The CodeRegion this SSAForm is for.
  //@access VARIABLES	CodeRegion_variables(SSAForm_CODEREGION(this))
  Dominance DOMINANCE;		// The pre-dominance for this SSAForm.
};

extern SSAForm
SSAForm_Ctor(SSAForm this, Memory memory, CodeRegion codeRegion);

extern SSAForm
SSAForm_Copy(SSAForm this, const_SSAForm that);

extern void
SSAForm_Dtor(SSAForm this);

extern size_t
SSAForm_Size(Memory memory, CodeRegion codeRegion);

#define SSAForm_MEMORY(this) ((this)->MEMORY)
#define SSAForm__MEMORY(this) (&(this)->MEMORY)
#define SSAForm_CODEREGION(this) ((this)->CODEREGION)
#define SSAForm__CODEREGION(this) (&(this)->CODEREGION)
#define SSAForm_DOMINANCE(this) ((this)->DOMINANCE)
#define SSAForm__DOMINANCE(this) (&(this)->DOMINANCE)
#define SSAForm_VARIABLES(this) (CodeRegion_variables(SSAForm_CODEREGION(this)))


//
void
SSAForm_checkConventional(SSAForm this, TemporaryTable globalTable);

/*
 * SSAForm_splitAfterCall --	if successor block of a call has several predecessors, split the edge
 * This is a requirement for:
 *  - SSADestruct when dealing with dedicated
 *  - Out of colored SSA
 */
void 
SSAForm_splitAfterCall(SSAForm this);

/*
 * SSAForm_prepareDedicated --	Add parallel copies to materialize dedicated constraints.
 */
void
SSAForm_prepareDedicated(SSAForm this);

/*
 * SSAForm_emptyVariableWebs --	Empty all variable web link of the global table.
 */
void 
SSAForm_emptyVariableWebs(SSAForm this, TemporaryTable globalTable);

/*
 * SSAForm_weblinksInit --	Initialize web links from pinned variables
 */
void
SSAForm_weblinksInit(SSAForm this, TemporaryTable temporaryTable);

/*
 * SSAForm_weblinksFromSSAwebs --	Make the SSAForm web links as union-find trees.
 *   
 * Requires: 
 *   - all phi operands are supposed to be coalesced.
 *   - coalesced copies have =Variable_isCoalesced()=
*/
void
SSAForm_weblinksFromSSAwebs(SSAForm this, TemporaryTable temporaryTable);

/*
 * SSAForm_eliminate --	Eliminate the PHI, PSI and SIGMA Operation(s) and SSA rename-back.
 *
 * Require:	the Variable webs are constructed with =SSAForm_weblinksFromSSAwebs()=.
 */
void
SSAForm_eliminate(SSAForm this);

/*
 * SSAForm_expandParCopyOperation --	Expand a parallel copy Operation.
 *
 * Note that in the case of multiple definitions of the same temporary,
 * only the first definition is kept.
 * This means that:
 * - the case PCOPY a,...,a = b,...,b is correctly handled
 * - the case PCOPY a,...,a = b,...,c is undefined
 *
 */
void
SSAForm_expandParCopyOperation(SSAForm this, Operation pcopy);

/*
 * SSAForm_dumpAppelGraph --	Dumps the Appel Graph.
 */
bool
SSAForm_dumpAppelGraph(SSAForm this, FILE *file, const char *name);

/*
 * SSAForm_dumpStats --	Dump the statistics on the SSA form.
 */
static bool
SSAForm_dumpStats(const_SSAForm this, FILE *file);

/*
 * SSAForm_INIT --	Initialize the SSAForm module.
 */
#if defined(SSAForm_MERGESETS)
#define SSAForm_INIT()
#else
void
SSAForm_INIT(void);
#endif

/*
 * SSAForm_FINI --	Finalize the SSAForm module.
 */
#define SSAForm_FINI()


#endif//SSAForm__h_INCLUDED
