#ifndef Florent_c_INCLUDED
#define Florent_c_INCLUDED

#line 22 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
#include "GRA/GRA_.h"

#line 27 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
#include "MDS/MDS_.h" // FIXME: to have access to swap operator definition

#line 29 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_renamePhiArgument(PtrSet operationToBeDetached, CodeRegion codeRegion,
                          InterfGraphBuilder interfGraphBuilder, RegFile regFile) {
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      if(Operator_isPhi(Operation_operator(operation))) {
        Temporary temp = Operation_getResult(operation, 0);
        if(Temporary_regFile(temp) == regFile) {
          Register reg = Temporary_register(temp);
          InterfNode representant =
            InterfGraphBuilder_getNode(interfGraphBuilder, temp);
          int32_t argIdx = 0;
          Operation_ARGUMENTS_FOREACH_Temporary(operation, arg) {
            if(reg == Temporary_register(arg)
               // maybe we should take the dedicated if it is one of the
               // argument instead of ignoring it.
               && arg != temp && !Temporary_isDedicated(arg)) {
              // Replace uses and def of arg
              InterfNode interfNode = InterfGraphBuilder_getNode(interfGraphBuilder, arg);
              IDList_FOREACH(InterfNode_uses(interfNode), Operation, use_p) {
                int32_t idx = 0;
                Operation_ARGUMENTS_FOREACH_Temporary(*use_p, temporary) {
                  if(temporary == arg) {
                    // Update the argument.
                    Operation_setArgument(*use_p, idx, temp);
                    InterfNode_addUse(representant, *use_p);
                  }
                  ++idx;
                } Operation_ARGUMENTS_ENDEACH_Temporary;
              } IDList_ENDEACH;
              IDList_FOREACH(InterfNode_defs(interfNode), Operation, def_p) {
                int32_t idx = 0;
                Operation_RESULTS_FOREACH_Temporary(*def_p, temporary) {
                  if(temporary == arg) {
                    // Update the result.
                    Operation_setResult(*def_p, idx, temp);
                    InterfNode_addDef(representant, *def_p);
                  }
                  ++idx;
                } Operation_RESULTS_ENDEACH_Temporary;
              } IDList_ENDEACH;
            }
            ++argIdx;
          } Operation_ARGUMENTS_ENDEACH_Temporary;
        }
      } else {
        break;
      }
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
}

#line 81 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_setRemoveOp(PtrSet operationToBeDetached, IDList pcopiesToBeUpdated,
                    Operation operation, int32_t idx) {
  if(!Operator_isParCopy(Operation_operator(operation))) {
    PtrSet_insert(operationToBeDetached, operation);
  } else {
    PCOPYUpdate_insert(pcopiesToBeUpdated, operation, idx);
  }
}

#line 91 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_discardMoves(Memory memory, PtrSet operationToBeDetached,
                     CodeRegion codeRegion, InterfGraphBuilder interfGraphBuilder,
                     RegFile regFile) {
  InterfGraph interfGraph;
  IDList_ pcopiesToBeUpdated[1];

#line 98 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  interfGraph = InterfGraphBuilder_INTERFGRAPH(interfGraphBuilder);
  IDList_Ctor(pcopiesToBeUpdated, memory);

#line 101 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  // Seek for repair to be done
  IDList_FOREACH(InterfGraphBuilder_NODES(interfGraphBuilder), InterfNode_, node) {
    IDListItem iter;
    int32_t srcIdx = InterfNode_index(node);
    Temporary srcTemp = InterfNode_temporary(node);
    IDList_FOREACH(InterfNode_movelist(node), InterfMove, pmove) {
      InterfMove move = *pmove;
      // Do propagation only when srcIdx is rhs.
      if(InterfNode_index(InterfMove_use(move)) == srcIdx
         // virtual node, nothing to simplify
         /* && InterfMove_defIdx(move) != -1 */) {
        Operation operation = InterfMove_operation(move);
        if(Operator_isSeqCopy(Operation_operator(operation))
           || Operator_isParCopy(Operation_operator(operation))
           || InterfMove_defIdx(move) == -1) {
          Temporary alias = InterfNode_temporary(InterfMove_def(move));
          //  Operation_getResult(operation, InterfMove_defIdx(move));
          if(Temporary_register(alias) == Temporary_register(srcTemp)) {
            if(alias != srcTemp) {
              Temporary swap = NULL;
              InterfNode swapNode = NULL;
              InterfNode aliasNode;
              // if alias is a dedicated, to have a better readable code,
              // remplace the variable by the dedicated.
              // We do that to avoid such simplification:
              // V1(R1) <- ldw @addr
              // R1 <- V1(R1)
              // return R1, ...
              // Without this handling, we get a code valid only regarding color:
              // V(R1) <- ldw @addr
              // return R1, ...
              // Thus we propagate the destination in the other direction:
              // R1 <- ldw @addr
              // return R1, ...
              if(Temporary_isDedicated(alias)) {
                swap = srcTemp;
                swapNode = node;
                srcTemp = alias;
                alias = swap;
                node = InterfGraphBuilder_getNode(interfGraphBuilder, srcTemp);
              }
              Except_CHECK(!Temporary_isDedicated(alias));
              aliasNode = InterfGraphBuilder_getNode(interfGraphBuilder, alias);
              Except_CHECK(IDList_count(InterfNode_defs(aliasNode)) <= 1
                           && "Current proccessing works only under SSA");
              IDList_FOREACH(InterfNode_uses(aliasNode), Operation, aliasUse_p) {
                int32_t idx = 0;
                IDList_ useIdxs[1];
                IDList_Ctor(useIdxs, memory);
                Operation_ARGUMENTS_FOREACH_Temporary(*aliasUse_p, temporary) {
                  if(temporary == alias) {
                    // Update the argument.
                    Operation_setArgument(*aliasUse_p, idx, srcTemp);
                    InterfNode_addUse(node, *aliasUse_p);
                    IDList_PUSH(useIdxs, int32_t, idx);
                  }
                  ++idx;
                } Operation_ARGUMENTS_ENDEACH_Temporary;
                if((Operator_isSeqCopy(Operation_operator(*aliasUse_p))
                    || Operator_isParCopy(Operation_operator(*aliasUse_p)))
                   && InterfMove_defIdx(move) != -1) {
                  IDList_FOREACH(useIdxs, int32_t, pIdx) {
                    if(Temporary_regFile(Operation_getResult(*aliasUse_p, *pIdx))
                       == Temporary_regFile(Operation_getArgument(*aliasUse_p, *pIdx))) {
                      InterfGraphBuilder_addPCOPYAffinity(interfGraphBuilder,
                                                          (*aliasUse_p), *pIdx);
                    }
                  } IDList_ENDEACH;
                }
                IDList_Dtor(useIdxs);
              } IDList_ENDEACH;
              if(InterfMove_defIdx(move) != -1) {
                Florent_setRemoveOp(operationToBeDetached, pcopiesToBeUpdated,
                                    operation, InterfMove_defIdx(move));
              } else {
                // do not remove elements in special nodes.
                // Otherwise we will not see these colors in the related rpc,
                // thus, we will think it is available
#if 0
                // remove from special nodes
                IDList specialNodes =
                  InterfGraphBuilder_SPECIALNODES(interfGraphBuilder);
                IDList_FOREACH(specialNodes, SplitInfo_, splitInfo) {
                  if(operation == SplitInfo_POINT(splitInfo)
                     && SplitInfo_ORIGVAR(splitInfo) == srcTemp
                     && SplitInfo_SPLITVAR(splitInfo) == alias) {
                    IDList_remove(specialNodes, splitInfo, NULL);
                    break;
                  }
                } IDList_ENDEACH;
#endif
              }
              if(swap != NULL) {
                IDList_FOREACH(InterfNode_defs(aliasNode), Operation, aliasDef_p) {
                  int32_t idx = 0;
                  Operation_RESULTS_FOREACH_Temporary(*aliasDef_p, temporary) {
                    if(temporary == alias) {
                      // Update the argument.
                      Operation_setResult(*aliasDef_p, idx, srcTemp);
                      InterfNode_addDef(node, *aliasDef_p);
                    }
                    ++idx;
                  } Operation_RESULTS_ENDEACH_Temporary;
                  if(Operator_isSeqCopy(Operation_operator(*aliasDef_p))
                   && Temporary_regFile(Operation_getResult(*aliasDef_p, 0))
                   == Temporary_regFile(Operation_getArgument(*aliasDef_p, 0))) {
                    InterfGraphBuilder_addMoveAffinity(interfGraphBuilder, (*aliasDef_p));
                  }
                } IDList_ENDEACH;
                srcTemp = swap;
                node = swapNode;
              }

#line 214 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
              // we may add some elts in movelist
              IDList_FOREACH_SYNC;
            } else {
              // defIdx == -1 means virtual operation -> nothing to remove
              if(InterfMove_defIdx(move) != -1) {
                // source and destination are the same temporary.
                // Remove the operation
                Florent_setRemoveOp(operationToBeDetached, pcopiesToBeUpdated,
                                    operation, InterfMove_defIdx(move));
              }
            }
          }
        }
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;
  GRA_updatePCOPIES(operationToBeDetached, pcopiesToBeUpdated);
  IDList_Dtor(pcopiesToBeUpdated);
}

#line 234 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_cleanCode(Memory memory, CodeRegion codeRegion,
                  InterfGraphBuilder interfGraphBuilder, RegFile regFile) {
  PtrSet_ operationToBeDetached[1];
  PtrSet_Ctor(operationToBeDetached, memory,
              IndexedFactory_COUNTER(Operation_Factory));

#line 241 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  Florent_discardMoves(memory, operationToBeDetached, codeRegion,
                       interfGraphBuilder, regFile);
/*   Florent_renamePhiArgument(operationToBeDetached, codeRegion, */
/*                             interfGraphBuilder, regFile); */

#line 246 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  PtrSet_FOREACH(operationToBeDetached, Operation, operation) {
    Except_CHECK(Operator_isParCopy(Operation_operator(operation))
                 || Operator_isSeqCopy(Operation_operator(operation)));
    BasicBlock_detachOperation(Operation_basicBlock(operation), operation);
  } PtrSet_ENDEACH;
  PtrSet_Dtor(operationToBeDetached);
}

#line 255 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static bool
Florent_allocCodeRegion_doIt(Memory memory, const char* prgName,
                            CodeRegion codeRegion, RegFile regFile,
                            InterfGraphBuilder interfGraphBuilder) {
  Register firstReg = RegFile_lowRegister(regFile);
  // Cannot use tmpfile function, since we need the name of the file to
  // pass it to florent tool (which performs florent algorithm)
  char graphFilenameBuff[L_tmpnam];
  char* graphFilename;
  char solutionFilenameBuff[L_tmpnam];
  char* solutionFilename;
  FILE* graph;
  int32_t errno;
  char* cmd = NULL;
  bool result = false;
  //FIXME: Might be unsafe, since name can be used by another process
  //before we open the file
  graphFilename = tmpnam(graphFilenameBuff);
  //FIXME: keep these errors for production compiler
  Except_CHECK(graphFilename != NULL && "Unable to generate a name for graph file");
  graph = fopen(graphFilename, "w");
  Except_CHECK(graph != NULL && "Unable to create a graph file");
  // export graph for Florent algorithm
  InterfGraphBuilder_exportGraph(interfGraphBuilder, graph, false, false,
                                 false);
  fclose(graph);
  solutionFilename = tmpnam(solutionFilenameBuff);
  Except_CHECK(solutionFilename != NULL && "Unable to generate a name for solution file");

#line 284 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  cmd = (char*)malloc((strlen(prgName) + 1 /* space */
                       + strlen(graphFilename) + 1 /* space */
                       + strlen(solutionFilename) + 1) * sizeof(char));
  Except_CHECK(cmd != NULL && "No space left in memory");
  (void)strcpy(cmd, prgName);
  (void)strcat(cmd, " ");
  (void)strcat(cmd, graphFilename);
  (void)strcat(cmd, " ");
  (void)strcat(cmd, solutionFilename);
  errno = system(cmd);

#line 295 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  if(errno != 0) {
    fprintf(stderr, "Failed to execute: '%s'\n", cmd);
    result = false;
  } else {
    // Apply result to program
    FILE* solution = fopen(solutionFilename, "r");
    int32_t nodes, color = 0, colors, idx = 0;
    bool start = false;
    Temporary* tmpArray = (Temporary*)TemporaryTable_base(InterfGraphBuilder_TEMPORARIES(interfGraphBuilder));
    Except_CHECK(solution != NULL && "Unable to open solution file");
    // Read line with number of nodes and colors
    fscanf(solution, "%d %d\n", &nodes, &colors);

#line 308 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
    while(!feof(solution)) {
      // read color and assign it
      int32_t c = fgetc(solution);
      if(c >= '0' && c <= '9') {
        start = true;
        color *= 10;
        color += c - '0';
      } else if(start) {
        if(!Temporary_isDedicated(tmpArray[idx])) {
          Temporary_assignRegister(tmpArray[idx], color + firstReg);
        } else {
          Except_CHECK(Temporary_register(tmpArray[idx]) == color + firstReg);
        }
        color = 0;
        start = false;
        ++idx;
      }
    }
    GRA_exportSolution(InterfGraphBuilder_TEMPORARIES(interfGraphBuilder),
                       GRA_DISP_COLORING, regFile);
    // file ends with newline, so we do not need to assign the last color.
    Except_CHECK(color == 0 && "Solution file does not ends with newline!");
    fclose(solution);
    result = true;
  }
  // purge environment
  (void)remove(graphFilename);
  (void)remove(solutionFilename);
  free(cmd);
  return result;
}

#line 348 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
RepairInfo
RepairInfo_Ctor(RepairInfo this, Temporary temp, int32_t opndIdx, Operation operation) {
  *RepairInfo__TEMPORARY(this) = temp;
  *RepairInfo__OPNDIDX(this) = opndIdx;
  *RepairInfo__OPERATION(this) = operation;
  return this;
}

#line 356 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
void
RepairInfo_Dtor(RepairInfo this) {
}

#line 362 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
// Return true, if it exists at least one temporary which interfers with an
// other and has the same color.
// toBeRepaired will contain the pair of temporaries which interfer
static bool
Florent_seekForRepair(Memory memory,   InterfGraphBuilder interfGraphBuilder,
                     RegFile regFile, IDList toBeRepaired) {
  CodeRegion codeRegion = InterfGraphBuilder_CODEREGION(interfGraphBuilder);
  bool result = false;
  InterfGraph interfGraph;
  interfGraph = InterfGraphBuilder_INTERFGRAPH(interfGraphBuilder);
  // Seek for repair to be done
  IDList_FOREACH(InterfGraphBuilder_NODES(interfGraphBuilder), InterfNode_, node) {
    IDListItem iter;
    int32_t srcIdx = InterfNode_index(node);
    int32_t dstIdx = -1;
    Temporary srcTemp = InterfNode_temporary(node);
    
    IDList_FOREACH(InterfNode_movelist(node), InterfMove, pmove) {
      InterfMove move = *pmove;
      InterfNode consideredNode =
        InterfNode_index(InterfMove_def(move)) == srcIdx
        ? InterfMove_use(move): InterfMove_def(move);
      Temporary dstTemp = InterfNode_temporary(consideredNode);
      dstIdx = InterfNode_index(consideredNode);

#line 387 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      // Negative cost == negative affinity == interference
      if(srcIdx < dstIdx && InterfMove_cost(move) < 0.0
         // By construction we are sure both temporaries are registers
         && Temporary_register(srcTemp) == Temporary_register(dstTemp)) {
        // negative affinities mean we assigned a temporary to a clobber
        // register or a wrong register class. So, only one of the temporaries
        // is  part of the operation operand. Take this one.
        Operation operation = InterfMove_operation(move);
        Temporary toRepair = NULL;
        RepairInfo_ pair;
        int32_t opndIdx = 0;
        Operation_OPERANDS_FOREACH_Temporary(operation, temporary) {
          if(temporary == srcTemp) {
            toRepair = srcTemp;
            break;
          }
          if(temporary == dstTemp) {
            toRepair = dstTemp;
            break;
          }
          ++opndIdx;
        } Operation_OPERANDS_ENDEACH_Temporary;
        if(toRepair == NULL) {
          // interference may be with a clobber or a register class constraint
          opndIdx = -1;
          if(Temporary_isDedicated(srcTemp)) {
            toRepair = dstTemp;
          } else {
            Except_CHECK(Temporary_isDedicated(dstTemp)
                         && "destination is not a clobber register");
            toRepair = srcTemp;
          }
        }
        Except_CHECK(toRepair != NULL && "Unable to find the temporary to be repaired");
        RepairInfo_Ctor(&pair, toRepair, opndIdx, operation);
        IDList_PUSH(toBeRepaired, RepairInfo_, pair);
        result = true;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;
  return result;
}

#line 430 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static int 
comparePair(const_IDListItem elt1, const_IDListItem elt2) {
  RepairInfo a = (RepairInfo)elt1;
  RepairInfo b = (RepairInfo)elt2;
  return Operation_identity(RepairInfo_OPERATION(a))
    - Operation_identity(RepairInfo_OPERATION(b));
}

#line 446 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
ActionParam
ActionParam_Ctor(ActionParam this, Procedure proc, Temporary dst, Temporary src) {
  *ActionParam__PROC(this) = proc;
  *ActionParam__DST(this) = dst;
  *ActionParam__SRC(this) = src;
  return this;
}

#line 454 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
void
ActionParam_Dtor(ActionParam this) {
}

#line 458 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_insertCopy(BasicBlock block, Operation pointInsertionAfter,
                   void* structParam) {
  ActionParam copyParam = (ActionParam) structParam;
  Selector_ selectorAfter[1];
  Selector blockSelector = BasicBlock_selector(block);
  Selector_Ctor(selectorAfter, Selector_memory(blockSelector),
                ActionParam_PROC(copyParam), Selector_processor(blockSelector),
                false);
  Selector_selectCOPY(selectorAfter, ActionParam_DST(copyParam),
                      ActionParam_SRC(copyParam));
  if(pointInsertionAfter == NULL) {
    BasicBlock_putSelector(block, selectorAfter);
  } else {
    BasicBlock_appendSelector(block, pointInsertionAfter, selectorAfter);
  }
  BasicBlock_makeOrderings(block, 0, 1);
}

#line 477 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_insertSwap(BasicBlock block, Operation pointInsertionAfter,
                   void* structParam) {
  ActionParam swapParam = (ActionParam) structParam;
  Selector_ selectorAfter[1];
  Selector blockSelector = BasicBlock_selector(block);
  Selector_Ctor(selectorAfter, Selector_memory(blockSelector),
                ActionParam_PROC(swapParam), Selector_processor(blockSelector),
                false);
  Selector_selectSWAP(selectorAfter, ActionParam_DST(swapParam),
                      ActionParam_SRC(swapParam));
  if(pointInsertionAfter == NULL) {
    BasicBlock_putSelector(block, selectorAfter);
  } else {
    BasicBlock_appendSelector(block, pointInsertionAfter, selectorAfter);
  }
  BasicBlock_makeOrderings(block, 0, 1);
}

#line 496 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_repairResults(RegisterSet availableRegs, IDList temporaryRes,
                     Operation operation, InterfGraphBuilder interfGraphBuilder) {
  RegisterSet_ alreadyTaken[1];
  RegisterSet_ argumentSet[1];
  Procedure proc =
    CodeRegion_procedure(InterfGraphBuilder_CODEREGION(interfGraphBuilder));
  BasicBlock block = Operation_basicBlock(operation);
  Selector blockSelector = BasicBlock_selector(block);

#line 506 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  RegisterSet_Ctor(alreadyTaken);
  RegisterSet_Ctor(argumentSet);

#line 509 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  Operation_ARGUMENTS_FOREACH_Temporary(operation, temp) {
    if(Temporary_hasRegFile(temp)) {
      RegisterSet_insert(argumentSet, Temporary_register(temp));
    }
  } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 515 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  if(Operation_hasClobber(operation)) {
    RegisterSet_FOREACH_Register(Operation_clobberSet(operation), reg) {
      RegisterSet_insert(alreadyTaken, reg);
    } RegisterSet_ENDEACH_Register;
  }

#line 521 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  IDList_FOREACH(temporaryRes, RepairInfo, prepairInfo) {
    RegisterSet_ supportedRegs[1];
    RegisterSet_ tmp[1];
    Temporary result;
    Temporary repairTemp = RepairInfo_TEMPORARY(*prepairInfo);
    InterfNode node =
      InterfGraphBuilder_getNode(interfGraphBuilder, repairTemp);
    RegisterSet_Ctor(supportedRegs);
    RegisterSet_Copy(supportedRegs,
                     RegFile_registerSet(InterfGraphBuilder_REGFILE(interfGraphBuilder)));
    IDList_FOREACH(InterfNode_movelist(node), InterfMove, pmove) {
      InterfMove move = *pmove;
      InterfNode consideredNode =
        InterfNode_index(InterfMove_def(move)) == InterfNode_index(node)
        ? InterfMove_use(move): InterfMove_def(move);
      // Negative cost == negative affinity == interference
      if(InterfMove_cost(move) < 0.0  && InterfMove_operation(move) == operation) {
        RegisterSet_remove(supportedRegs, Temporary_register(InterfNode_temporary(consideredNode)));
      }
    } IDList_ENDEACH;
    RegisterSet_Ctor(tmp);
    RegisterSet_Copy(tmp, supportedRegs);
    RegisterSet_inter(tmp, availableRegs);
    // Seek for a good candidate
    // Trivial repairing
    if(RegisterSet_count(tmp) != 0) {
      Register reg = RegisterSet_first(availableRegs);
      result =
        Procedure_makeVirtualTemporary(proc,
                                       InterfGraphBuilder_REGFILE(interfGraphBuilder));
      Temporary_assignRegister(result, reg);
      RegisterSet_remove(availableRegs, reg);

#line 554 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      if(operation == BasicBlock_lastOperation(block)) {
        ActionParam_ copyParam[1];
        ActionParam_Ctor(copyParam, proc, repairTemp, result);
        // if this operation is the last operation of block, this is for
        // good reasons. So, do not break that.
        BasicBlock_splitLeaveEdgeAndAction(block, true, (void*)copyParam,
                                           Florent_insertCopy);
      } else {
        Selector_ selectorAfter[1];
        Selector_Ctor(selectorAfter, Selector_memory(blockSelector), proc,
                      Selector_processor(blockSelector), false);
        Selector_selectCOPY(selectorAfter, repairTemp, result);
        BasicBlock_appendSelector(block, operation,  selectorAfter);
        BasicBlock_makeOrderings(block, 0, 1);
      }
    } else {
      RegisterSet_ possibleRegs[1];
      RegisterSet_ withoutUsingArgument[1];
      Register reg;
      Selector_ selectorBefore[1];
      RegisterSet_Ctor(possibleRegs);
      RegisterSet_Ctor(withoutUsingArgument);

#line 577 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      RegisterSet_Copy(possibleRegs, supportedRegs);
      RegisterSet_diff(possibleRegs, alreadyTaken);

#line 580 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      Except_REQUIRE(RegisterSet_count(possibleRegs) != 0
                     && "Trivial repair does not work");

#line 583 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      RegisterSet_Copy(withoutUsingArgument, possibleRegs);
      RegisterSet_diff(withoutUsingArgument, argumentSet);

#line 586 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      // try to not take a register used in argument
      if(RegisterSet_count(withoutUsingArgument) != 0) {
        reg = RegisterSet_first(withoutUsingArgument);
      } else {
        int32_t argIdx = 0;
        reg = RegisterSet_first(possibleRegs);
        // Update arguments
        // FIXME: ensure that modifying an argument will not create a negative
        // affinity.
        // This piece of code is needed to handle this case:
        // Rx op Rx, Ry
        // Without it we will generate:
        // swap Rz, Rx <-- put Rx value in Rz
        // Rz op Rx, Ry <-- wrong, Rx value is no more here
        // swap Rz, Rx
        // With it:
        // swap Rz, Rx
        // Rz op Rz, Ry <-- OK, Rx value is in Rz
        // swap Rz, Rx
        Operation_ARGUMENTS_FOREACH_Temporary(operation, argument) {
          if(Temporary_hasRegFile(argument)
             && Temporary_register(argument) == Temporary_register(repairTemp)) {
            Temporary newArg = Procedure_makeVirtualTemporary(proc,
                                                              InterfGraphBuilder_REGFILE(interfGraphBuilder));
            Temporary_assignRegister(newArg, reg);
            // FIXME: Check if there is not already a negative affinity on
            // this argument
            Operation_setArgument(operation, argIdx, newArg);
          }
          ++argIdx;
        } Operation_ARGUMENTS_ENDEACH_Temporary;
      }
      result =
        Procedure_makeVirtualTemporary(proc,
                                       InterfGraphBuilder_REGFILE(interfGraphBuilder));
      Temporary_assignRegister(result, reg);
      // no need to remove form availableRegs

#line 624 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      Selector_Ctor(selectorBefore, Selector_memory(blockSelector), proc,
                    Selector_processor(blockSelector), false);
      Selector_selectSWAP(selectorBefore, RepairInfo_TEMPORARY(*prepairInfo), result);
      BasicBlock_insertSelector(block, operation, selectorBefore);
      // insert after code
      if(operation == BasicBlock_lastOperation(block)) {
        ActionParam_ swapParam[1];
        ActionParam_Ctor(swapParam, proc, repairTemp, result);
        // if this operation is the last operation of block, this is for
        // good reasons. So, do not break that.
        BasicBlock_splitLeaveEdgeAndAction(block, true, (void*)swapParam,
                                           Florent_insertSwap);
      } else {
        Selector_ selectorAfter[1];
        Selector_Ctor(selectorAfter, Selector_memory(blockSelector), proc,
                      Selector_processor(blockSelector), false);
        Selector_selectSWAP(selectorAfter, repairTemp, result);
        BasicBlock_appendSelector(block, operation,  selectorAfter);
      }
      BasicBlock_makeOrderings(block, 0, 1);
    }
    {
      int32_t resIdx = 0;
      Operation_RESULTS_FOREACH_Temporary(operation, resTemp) {
        if(resTemp == repairTemp) {
          Operation_setResult(operation, resIdx, result);
        }
        ++resIdx;
      } Operation_RESULTS_ENDEACH_Temporary;
    }
  } IDList_ENDEACH;

#line 656 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
}

#line 658 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_repairArgs(RegisterSet availableRegs, IDList temporaryArgs,
                  Operation operation, InterfGraphBuilder interfGraphBuilder) {
  Procedure proc =
    CodeRegion_procedure(InterfGraphBuilder_CODEREGION(interfGraphBuilder));
  BasicBlock block = Operation_basicBlock(operation);
  Selector blockSelector = BasicBlock_selector(block);
  IDList_FOREACH(temporaryArgs, RepairInfo, prepairInfo) {
    RegisterSet_ supportedRegs[1];
    Temporary repairTemp = RepairInfo_TEMPORARY(*prepairInfo);
    InterfNode node =
      InterfGraphBuilder_getNode(interfGraphBuilder, repairTemp);
    RegisterSet_Ctor(supportedRegs);
    RegisterSet_Copy(supportedRegs, availableRegs);
    IDList_FOREACH(InterfNode_movelist(node), InterfMove, pmove) {
      InterfMove move = *pmove;
      InterfNode consideredNode =
        InterfNode_index(InterfMove_def(move)) == InterfNode_index(node)
        ? InterfMove_use(move): InterfMove_def(move);
      // Negative cost == negative affinity == interference
      if(InterfMove_cost(move) < 0.0 && InterfMove_operation(move) == operation) {
        RegisterSet_remove(supportedRegs, Temporary_register(InterfNode_temporary(consideredNode)));
      }
    } IDList_ENDEACH;
    Operation_ARGUMENTS_FOREACH_Temporary(operation, temp) {
      if(Temporary_hasRegFile(temp)) {
        RegisterSet_remove(supportedRegs, Temporary_register(temp));
      }
    } Operation_ARGUMENTS_ENDEACH_Temporary;
    Except_CHECK(RegisterSet_count(supportedRegs) != 0
                 && "Not trivial repairing for arguments is not supported");
    // op Rz, ... <-- Rz is not good for this argument (Rz can cross op and be
    //                in the clobber!)
    // Repair code is:
    // move Rx, Rz
    // op Rx, ...
    // move Rz, Rx
    {
      Register reg = RegisterSet_first(supportedRegs);
      Selector_ selectorBefore[1];
      Temporary result =
        Procedure_makeVirtualTemporary(proc,
                                       InterfGraphBuilder_REGFILE(interfGraphBuilder));
      RegisterSet_remove(availableRegs, reg);
      Temporary_assignRegister(result, reg);

#line 704 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      Selector_Ctor(selectorBefore, Selector_memory(blockSelector), proc,
                    Selector_processor(blockSelector), false);
      Selector_selectCOPY(selectorBefore, result, repairTemp);
      BasicBlock_insertSelector(block, operation,  selectorBefore);

#line 709 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      // insert after code
      if(operation == BasicBlock_lastOperation(block)) {
        ActionParam_ copyParam[1];
        ActionParam_Ctor(copyParam, proc, repairTemp, result);
        // if this operation is the last operation of block, this is for
        // good reasons. So, do not break that.
        BasicBlock_splitLeaveEdgeAndAction(block, true, (void*)copyParam,
                                           Florent_insertCopy);
      } else {
        Selector_ selectorAfter[1];
        Selector_Ctor(selectorAfter, Selector_memory(blockSelector), proc,
                      Selector_processor(blockSelector), false);
        Selector_selectCOPY(selectorAfter, repairTemp, result);

#line 723 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
        BasicBlock_appendSelector(block, operation,  selectorAfter);
      }
      BasicBlock_makeOrderings(block, 0, 1);
    }
  } IDList_ENDEACH;
}

#line 730 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_repairInfo(IDList temporaryRes, IDList temporaryArgs,
                  Operation operation, RegFile regFile,
                  InterfGraphBuilder interfGraphBuilder) {
  BasicBlock block = Operation_basicBlock(operation);
  Convention convention = InterfGraphBuilder_CONVENTION(interfGraphBuilder);
  RegisterSet_ availableRegs[1];
  *availableRegs = Convention_allocatableSet(convention, regFile);
  TemporarySet_FOREACH_Temporary(BasicBlock_liveOutSet(block), temporary) {
    if(Temporary_hasRegFile(temporary)) {
      RegisterSet_remove(availableRegs, Temporary_register(temporary));
    }
  } TemporarySet_ENDEACH_Temporary;
  BasicBlock_FORBACK_Operation(block, currOperation) {
    if(currOperation == operation) {
      // Repair results
      Florent_repairResults(availableRegs, temporaryRes, operation,
                           interfGraphBuilder);
    }
    Operation_RESULTS_FOREACH_Temporary(operation, temporary) {
      if(Temporary_hasRegFile(temporary)) {
        RegisterSet_insert(availableRegs, Temporary_register(temporary));
      }
    } Operation_RESULTS_ENDEACH_Temporary;
    Operation_ARGUMENTS_FOREACH_Temporary(operation, temporary) {
      if(Temporary_hasRegFile(temporary)) {
        RegisterSet_remove(availableRegs, Temporary_register(temporary));
      }
    } Operation_ARGUMENTS_ENDEACH_Temporary;
    if(Operation_hasClobber(operation)) {
      RegisterSet_FOREACH_Register(Operation_clobberSet(operation), clobber) {
        RegisterSet_remove(availableRegs, clobber);
      } RegisterSet_ENDEACH_Register;
    }
    if(currOperation == operation) {
      // Repair arguments
      Florent_repairArgs(availableRegs, temporaryArgs, operation, interfGraphBuilder);
    }
  } BasicBlock_ENDBACK_Operation;
}

#line 771 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_doRepair(Memory memory, IDList toBeRepaired,
                InterfGraphBuilder interfGraphBuilder) {
  Operation operation = NULL;
  IDList_ temporaryArgs[1];
  IDList_ temporaryRes[1];
  RegFile regFile = InterfGraphBuilder_REGFILE(interfGraphBuilder);
  IDList_Ctor(temporaryArgs, memory);
  IDList_Ctor(temporaryRes, memory);
  IDList_sort(toBeRepaired, comparePair);
  IDList_FOREACH(toBeRepaired, RepairInfo_, repairInfo) {
    if(operation != RepairInfo_OPERATION(repairInfo)) {
      if(operation != NULL) {
        Florent_repairInfo(temporaryRes, temporaryArgs, operation, regFile,
                          interfGraphBuilder);
      }
      operation = RepairInfo_OPERATION(repairInfo);
      IDList_empty(temporaryArgs, NULL);
      IDList_empty(temporaryRes, NULL);
    }
    if(RepairInfo_OPNDIDX(repairInfo) < Operation_resCount(operation)
       // -1 means cross operation
       && RepairInfo_OPNDIDX(repairInfo) != -1) {
      IDList_PUSH(temporaryRes, RepairInfo, repairInfo);
    } else {
      IDList_PUSH(temporaryArgs, RepairInfo, repairInfo);
    }
  } IDList_ENDEACH;
  if(operation != NULL) {
    Florent_repairInfo(temporaryRes, temporaryArgs, operation, regFile,
                      interfGraphBuilder);
  }
}

#line 806 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_repairCodeRegion(Memory memory, CodeRegion codeRegion, RegFile regFile,
                        InterfGraphBuilder interfGraphBuilder) {
  IDList_ toBeRepaired[1];
  bool result;  
  IDList_Ctor(toBeRepaired, memory);
  result = Florent_seekForRepair(memory, interfGraphBuilder, regFile, toBeRepaired);
  if(result) {
    // Do repair
    Florent_doRepair(memory, toBeRepaired, interfGraphBuilder);
  }
  IDList_Dtor(toBeRepaired);
}

#line 820 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static int
compareSplitInfo(const_IDListItem a, const_IDListItem b) {
  return Operation_identity(SplitInfo_POINT((SplitInfo)a))
    - Operation_identity(SplitInfo_POINT((SplitInfo)b));
}

#line 826 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_mergeRpcEvolved(RpcEvolved dst, const_RpcEvolved src) {
  int32_t i;
  for(i = 0; i < MAX_RPC; ++i) {
    if(RpcEvolved_TO(src)[i] != -1) {
      if(RpcEvolved_TO(dst)[i] != -1
         && RpcEvolved_TO(dst)[i] != i) {
        // ensure someone use this color because it must be lived
        Except_CHECK(RpcEvolved_INV(dst)[RpcEvolved_TO(src)[i]] != -1
                     && "duplication not supported");
      }
      if(RpcEvolved_INV(dst)[RpcEvolved_TO(src)[i]] != -1) {
        Except_CHECK(RpcEvolved_TO(src)[i] == i);
        continue;
      }
      RpcEvolved_TO(dst)[i] = RpcEvolved_TO(src)[i];
      RpcEvolved_INV(dst)[RpcEvolved_TO(src)[i]] =
        RpcEvolved_INV(src)[RpcEvolved_TO(src)[i]];
      RpcEvolved_BEFORE(dst)[i] = RpcEvolved_BEFORE(src)[i];
      RpcEvolved_AFTER(dst)[RpcEvolved_TO(src)[i]] =
        RpcEvolved_AFTER(src)[RpcEvolved_TO(src)[i]];
    }
  }
}

#line 851 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_translateRpcEvolved(RpcEvolved crossRpc, RpcEvolved useRpc) {
  int32_t Rx;
  bool touched[MAX_RPC];
  memset(touched, false, MAX_RPC * sizeof(bool));
  for(Rx = 0; Rx < MAX_RPC; ++Rx) {
    // Purpose is to remplace
    // if Rx -> Ry in crossRpc
    // and Rx -> Rk in useRpc
    // then Ry -> Rk in useRpc (useRpc is executed after crossRpc)
    if(RpcEvolved_TO(useRpc)[Rx] != -1 && !touched[Rx]) {
      int32_t Rk = RpcEvolved_TO(useRpc)[Rx];
      int32_t Ry = RpcEvolved_TO(crossRpc)[Rx];
      if(Ry == -1) {
        Except_CHECK((Rx == Rk && RpcEvolved_INV(crossRpc)[Rx] < Rx)
                     || RpcEvolved_INV(crossRpc)[Rx] == -1);
        continue;
      }
      // ensure nobody go on the new color
      Except_CHECK((RpcEvolved_INV(useRpc)[Ry] == -1
                    || ((RpcEvolved_INV(useRpc)[Ry] == Rx) && Ry == Rk))
                   && "Coloring is wrong, we overwrite a crossing variable "
                   "in use");
      RpcEvolved_TO(useRpc)[Rx] = -1;
      RpcEvolved_TO(useRpc)[Ry] = Rk;
      RpcEvolved_INV(useRpc)[Rk] = Ry;
      touched[Ry] = true;
    }
  }
}

#line 882 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_generateRpc(Operation currOperation, RpcEvolved crossInRpc,
                    RpcEvolved crossOutRpc, RpcEvolved useRpc,
                    RpcEvolved defRpc, int32_t idxToReg[]) {

#line 887 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  // sources of use rpc may contain live variable before their split
  // thus, translate the color of useRpc and merge it with cross
  Florent_translateRpcEvolved(crossInRpc, useRpc);
  // We must merge them, otherwise use rpc will not see live
  // variables from cross
  Florent_mergeRpcEvolved(crossInRpc, useRpc);

#line 894 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  RpcEvolved_generateRpc(currOperation, crossInRpc, idxToReg, true);
  // def rpc contains only variables that do not cross
  // thus, we can merge crossOut and def rpcs.
  // Moreover merge is mandatory to have correct liveness on colors
  Florent_mergeRpcEvolved(crossOutRpc, defRpc);
  RpcEvolved_generateRpc(currOperation, crossOutRpc, idxToReg, false);
}

#line 902 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
Florent_newRepairCodeRegion(Memory memory, CodeRegion codeRegion, RegFile regFile,
                            InterfGraphBuilder interfGraphBuilder) {
  Operation currOperation = NULL;
  IDList specialNodes = InterfGraphBuilder_SPECIALNODES(interfGraphBuilder);
  // {splitLive} parallel copy {live} // <- crossInRpc
  // {splitUse} parallel copy {use} // <- useRpc
  // {splitDef} op {splitUse}
  // {def} parallel copy {splitDef} // <- defRpc
  // {live} parallel copy {splitLive} // <- crossOutRpc
  RpcEvolved_ crossInRpc[1];
  // In fact, crossOutRpc is a swap of INV and TO of crossInRpc
  RpcEvolved_ crossOutRpc[1];
  RpcEvolved_ defRpc[1];
  RpcEvolved_ useRpc[1];
  RegisterSet_ allocatableSet[1];
  int32_t *regToIdx = alloca(sizeof(int32_t)*Register__COUNT);
  int32_t *idxToReg = alloca(sizeof(int32_t)*Register__COUNT);
  memset(regToIdx, -1, sizeof(int32_t)*Register__COUNT);
  memset(idxToReg, -1, sizeof(int32_t)*Register__COUNT);
  *allocatableSet = Convention_allocatableSet(InterfGraphBuilder_CONVENTION(interfGraphBuilder),
                                              InterfGraphBuilder_REGFILE(interfGraphBuilder));
  RpcEvolved_initRegMap(regToIdx, idxToReg, allocatableSet,
                        InterfGraphBuilder_REGFILE(interfGraphBuilder));
  RpcEvolved_Ctor(crossInRpc, RegisterSet_count(allocatableSet));
  RpcEvolved_Ctor(crossOutRpc, RegisterSet_count(allocatableSet));
  RpcEvolved_Ctor(defRpc, RegisterSet_count(allocatableSet));
  RpcEvolved_Ctor(useRpc, RegisterSet_count(allocatableSet));

#line 931 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  RegisterSet_Dtor(allocatableSet);

#line 933 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  Except_CHECK(!IDList_isEmpty(specialNodes));
  IDList_sort(specialNodes, compareSplitInfo);
  currOperation = SplitInfo_POINT((SplitInfo)IDList_firstItem(specialNodes));
  IDList_FOREACH(specialNodes, SplitInfo_, splitInfo) {
    if(currOperation != SplitInfo_POINT(splitInfo)) {
      Florent_generateRpc(currOperation, crossInRpc, crossOutRpc,
                          useRpc, defRpc, idxToReg);
      currOperation = SplitInfo_POINT(splitInfo);
      // rpc
      RpcEvolved_clear(crossInRpc);
      RpcEvolved_clear(crossOutRpc);
      RpcEvolved_clear(defRpc);
      RpcEvolved_clear(useRpc);
    }
    {
      int32_t regOrig, regSplit;
      regOrig = regToIdx[Temporary_register(SplitInfo_ORIGVAR(splitInfo))];
      regSplit = regToIdx[Temporary_register(SplitInfo_SPLITVAR(splitInfo))];
      Except_CHECK(regOrig != -1 && "A non allocatable variable has been splitted");
      Except_CHECK(regSplit != -1 && "Split variable assigned to an invalid color");

#line 954 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      if(SplitInfo_OPERANDIDX(splitInfo) == -1) {
        // Cross variable
        // Check if there is a duplication, because it is not handle currently
        Except_CHECK(RpcEvolved_TO(crossInRpc)[regOrig] == -1);
        Except_CHECK(RpcEvolved_INV(crossInRpc)[regSplit] == -1);
        Except_CHECK(RpcEvolved_TO(crossOutRpc)[regSplit] == -1);
        Except_CHECK(RpcEvolved_INV(crossOutRpc)[regOrig] == -1);
        // Orig to split
        RpcEvolved_TO(crossInRpc)[regOrig] = regSplit;
        RpcEvolved_INV(crossInRpc)[regSplit] = regOrig;
        RpcEvolved_BEFORE(crossInRpc)[regOrig] = SplitInfo_ORIGVAR(splitInfo);
        RpcEvolved_AFTER(crossInRpc)[regSplit] = SplitInfo_SPLITVAR(splitInfo);
        // Split to orig
        RpcEvolved_TO(crossOutRpc)[regSplit] = regOrig;
        RpcEvolved_INV(crossOutRpc)[regOrig] = regSplit;
        RpcEvolved_BEFORE(crossOutRpc)[regSplit] = SplitInfo_SPLITVAR(splitInfo);
        RpcEvolved_AFTER(crossOutRpc)[regOrig] = SplitInfo_ORIGVAR(splitInfo);
      } else if(SplitInfo_OPERANDIDX(splitInfo) < Operation_resCount(currOperation)) {
        // Definition
        // Check if there is a duplication, because it is not handle currently
        Except_CHECK(RpcEvolved_TO(defRpc)[regSplit] == -1);
        Except_CHECK(RpcEvolved_INV(defRpc)[regOrig] == -1);
        // Split to orig
        RpcEvolved_TO(defRpc)[regSplit] = regOrig;
        RpcEvolved_INV(defRpc)[regOrig] = regSplit;
        RpcEvolved_BEFORE(defRpc)[regSplit] = SplitInfo_SPLITVAR(splitInfo);
        RpcEvolved_AFTER(defRpc)[regOrig] = SplitInfo_ORIGVAR(splitInfo);
        // Update operation result
        Except_CHECK(Operation_getResult(SplitInfo_POINT(splitInfo),
                                         SplitInfo_OPERANDIDX(splitInfo))
                     == SplitInfo_ORIGVAR(splitInfo));
        Operation_setResult(SplitInfo_POINT(splitInfo),
                            SplitInfo_OPERANDIDX(splitInfo),
                            SplitInfo_SPLITVAR(splitInfo));
      } else {
        // Usage
        // Check if there is a duplication, because it is not handle currently
        Except_CHECK(RpcEvolved_TO(useRpc)[regOrig] == -1);
        Except_CHECK(RpcEvolved_INV(useRpc)[regSplit] == -1);
        // Orig to split
        RpcEvolved_TO(useRpc)[regOrig] = regSplit;
        RpcEvolved_INV(useRpc)[regSplit] = regOrig;
        RpcEvolved_BEFORE(useRpc)[regOrig] = SplitInfo_ORIGVAR(splitInfo);
        RpcEvolved_AFTER(useRpc)[regSplit] = SplitInfo_SPLITVAR(splitInfo);
        // Update operation argument
        Except_CHECK(Operation_getArgument(SplitInfo_POINT(splitInfo),
                                           SplitInfo_OPERANDIDX(splitInfo)
                                           - Operation_resCount(SplitInfo_POINT(splitInfo)))
                     == SplitInfo_ORIGVAR(splitInfo));
        Operation_setArgument(SplitInfo_POINT(splitInfo),
                              SplitInfo_OPERANDIDX(splitInfo)
                              - Operation_resCount(SplitInfo_POINT(splitInfo)),
                              SplitInfo_SPLITVAR(splitInfo));
      }
    }
  } IDList_ENDEACH;
  // We do not proceed the last operation, do it
  Florent_generateRpc(currOperation, crossInRpc, crossOutRpc, useRpc, defRpc,
                      idxToReg);

#line 1014 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  RpcEvolved_Dtor(crossInRpc);
  RpcEvolved_Dtor(crossOutRpc);
  RpcEvolved_Dtor(defRpc);
  RpcEvolved_Dtor(useRpc);
}

#line 1020 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
checkAllocatedCode(CodeRegion codeRegion) {
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operation_RESULTS_FOREACH_Temporary(operation, temp) {
        if(!Temporary_hasRegFile(temp)) {
          Except_CHECK(0 && "Not a register");
        }
      } Operation_RESULTS_ENDEACH_Temporary;
      Operation_ARGUMENTS_FOREACH_Temporary(operation, temp) {
        if(!Temporary_hasRegFile(temp)) {
          fprintf(stderr, "KO Id %d\tType %d\n", Temporary_identity(temp),
                  Temporary_type(temp));
        } else {
          fprintf(stderr, "OK Id %d\tType %d\n", Temporary_identity(temp),
                  Temporary_type(temp));
        }
      } Operation_ARGUMENTS_ENDEACH_Temporary;
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
}

#line 1042 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static bool
shouldCountCopy(Temporary res, Temporary arg, RegFile regFile) {
  return (Temporary_regFile(res) == regFile)
    && (Temporary_regFile(arg) == regFile)
    && (Temporary_register(arg) != Temporary_register(res));
}

#line 1049 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
struct MoveCost_ {
  uint32_t NBMOVES;
  float MOVESCOST;
  uint32_t NBMOVESONEDGE;
  float MOVESONEDGECOST;
  uint32_t NBMOVESATMAXPRESSURE;
  float MOVESCOSTATMAXPRESSURE;
  uint32_t NBMOVESATMAXPRESSUREONEDGE;
  float MOVESCOSTATMAXPRESSUREONEDGE;
  uint32_t NBOPERATIONS;
  float OPERATIONSCOST;
  uint32_t NBEDGESWITHMOVES;
  float COSTNBEDGESWITHMOVES;
  uint32_t NBJUMPFROMEDGE;
  float JUMPFROMEDGECOST;
  uint32_t NBBBS;
  float BBSCOST;
};
typedef struct MoveCost_ MoveCost_, *MoveCost;
typedef const struct MoveCost_ *const_MoveCost;
typedef struct MoveCost_ * restrict_MoveCost;

#define MoveCost_NBMOVES(this) ((this)->NBMOVES)
#define MoveCost__NBMOVES(this) (&(this)->NBMOVES)
#define MoveCost_MOVESCOST(this) ((this)->MOVESCOST)
#define MoveCost__MOVESCOST(this) (&(this)->MOVESCOST)
#define MoveCost_NBMOVESONEDGE(this) ((this)->NBMOVESONEDGE)
#define MoveCost__NBMOVESONEDGE(this) (&(this)->NBMOVESONEDGE)
#define MoveCost_MOVESONEDGECOST(this) ((this)->MOVESONEDGECOST)
#define MoveCost__MOVESONEDGECOST(this) (&(this)->MOVESONEDGECOST)
#define MoveCost_NBMOVESATMAXPRESSURE(this) ((this)->NBMOVESATMAXPRESSURE)
#define MoveCost__NBMOVESATMAXPRESSURE(this) (&(this)->NBMOVESATMAXPRESSURE)
#define MoveCost_MOVESCOSTATMAXPRESSURE(this) ((this)->MOVESCOSTATMAXPRESSURE)
#define MoveCost__MOVESCOSTATMAXPRESSURE(this) (&(this)->MOVESCOSTATMAXPRESSURE)
#define MoveCost_NBMOVESATMAXPRESSUREONEDGE(this) ((this)->NBMOVESATMAXPRESSUREONEDGE)
#define MoveCost__NBMOVESATMAXPRESSUREONEDGE(this) (&(this)->NBMOVESATMAXPRESSUREONEDGE)
#define MoveCost_MOVESCOSTATMAXPRESSUREONEDGE(this) ((this)->MOVESCOSTATMAXPRESSUREONEDGE)
#define MoveCost__MOVESCOSTATMAXPRESSUREONEDGE(this) (&(this)->MOVESCOSTATMAXPRESSUREONEDGE)
#define MoveCost_NBOPERATIONS(this) ((this)->NBOPERATIONS)
#define MoveCost__NBOPERATIONS(this) (&(this)->NBOPERATIONS)
#define MoveCost_OPERATIONSCOST(this) ((this)->OPERATIONSCOST)
#define MoveCost__OPERATIONSCOST(this) (&(this)->OPERATIONSCOST)
#define MoveCost_NBEDGESWITHMOVES(this) ((this)->NBEDGESWITHMOVES)
#define MoveCost__NBEDGESWITHMOVES(this) (&(this)->NBEDGESWITHMOVES)
#define MoveCost_COSTNBEDGESWITHMOVES(this) ((this)->COSTNBEDGESWITHMOVES)
#define MoveCost__COSTNBEDGESWITHMOVES(this) (&(this)->COSTNBEDGESWITHMOVES)
#define MoveCost_NBJUMPFROMEDGE(this) ((this)->NBJUMPFROMEDGE)
#define MoveCost__NBJUMPFROMEDGE(this) (&(this)->NBJUMPFROMEDGE)
#define MoveCost_JUMPFROMEDGECOST(this) ((this)->JUMPFROMEDGECOST)
#define MoveCost__JUMPFROMEDGECOST(this) (&(this)->JUMPFROMEDGECOST)
#define MoveCost_NBBBS(this) ((this)->NBBBS)
#define MoveCost__NBBBS(this) (&(this)->NBBBS)
#define MoveCost_BBSCOST(this) ((this)->BBSCOST)
#define MoveCost__BBSCOST(this) (&(this)->BBSCOST)


#line 1068 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
MoveCost
MoveCost_Ctor(MoveCost this) {
  memset(this, 0, sizeof(*this));
  return this;
}

#line 1074 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
void
MoveCost_Dtor(MoveCost this) {
}

#line 1078 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
TemporarySet_selectRegFile(TemporarySet dst, const_TemporarySet src,
                           RegFile regFile) {
  TemporarySet_FOREACH_Temporary(src, tmp) {
    if(Temporary_regFile(tmp) == regFile) {
      TemporarySet_insert(dst, tmp);
    }
  } TemporarySet_ENDEACH_Temporary;
}

#line 1088 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
computeStatistic(CodeRegion codeRegion, Liveness liveness, RegFile regFile,
                 MoveCost moveCost, uint32_t counter) {
  Convention convention = CodeRegion_convention(codeRegion);
  RegisterSet_ availableReg[1];
  int32_t nbAvailableRegs;
  TemporarySet_ liveVar[1];

#line 1096 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  availableReg[0] = Convention_allocatableSet(convention, regFile);
  nbAvailableRegs = RegisterSet_count(availableReg);
  TemporarySet_Ctor(liveVar, Liveness_memory(liveness), nbAvailableRegs);
  RegisterSet_Dtor(availableReg);
  // count moves and rpc
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    float cost = BasicBlock_frequency(basicBlock);
    uint32_t id = BasicBlock_identity(basicBlock);
    uint32_t liveInCount;
    BitSet_ bitSet[1];
    bool first = true;
    BitSet_Ctor(bitSet, Liveness_memory(liveness),
                BasicBlock_enterCount(basicBlock));
    
    TemporarySet_empty(liveVar);
    TemporarySet_selectRegFile(liveVar, BasicBlock_phiInSet(basicBlock),
                               regFile);
    TemporarySet_selectRegFile(liveVar, BasicBlock_liveInSet(basicBlock),
                               regFile);
    liveInCount = TemporarySet_count(liveVar);

#line 1117 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
    TemporarySet_empty(liveVar);
    TemporarySet_selectRegFile(liveVar, BasicBlock_liveOutSet(basicBlock),
                               regFile);
    TemporarySet_selectRegFile(liveVar, BasicBlock_phiOutSet(basicBlock),
                               regFile);
    BasicBlock_FORBACK_Operation(basicBlock, operation) {
      Operator operator = Operation_operator(operation);
      if(Operator_isPhi(operator)) {
        int32_t idx;
        Temporary res = Operation_getResult(operation, 0);
        if(Temporary_regFile(res) == regFile) {
          for(idx = 0; idx < Operation_argCount(operation); ++idx) {
            Temporary arg = Operation_getArgument(operation, idx);
            if(shouldCountCopy(res, arg, regFile)) {
              BasicBlockEdge edge = BasicBlock_getPhiEnterEdge(basicBlock, idx);
              BasicBlock origBB = BasicBlockEdge_origBlock(edge);
              float edgeCost = BasicBlock_frequency(origBB)
                * BasicBlockEdge_probability(edge);
              if(BitSet_insert(bitSet, idx) == true) {
                ++(*MoveCost__NBEDGESWITHMOVES(moveCost));
                *MoveCost__COSTNBEDGESWITHMOVES(moveCost)
                  += edgeCost;
              }
              if(liveInCount == nbAvailableRegs) {
                ++(*MoveCost__NBMOVESATMAXPRESSUREONEDGE(moveCost));
                *MoveCost__MOVESCOSTATMAXPRESSUREONEDGE(moveCost)
                  += edgeCost;
              } else {
                ++(*MoveCost__NBMOVESONEDGE(moveCost));
                *MoveCost__MOVESONEDGECOST(moveCost)
                  += edgeCost;
              }
            }
          }
        }
      } else if(Operator_isSeqCopy(operator) || Operator_isParCopy(operator)) {
        int32_t idx;
        for(idx = 0; idx < Operation_resCount(operation); ++idx) {
          Temporary res = Operation_getResult(operation, idx);
          Temporary arg = Operation_getArgument(operation, idx);
          if(shouldCountCopy(res, arg, regFile)) {
            if(id >= counter) {
              if(TemporarySet_count(liveVar) == nbAvailableRegs) {
                ++(*MoveCost__NBMOVESATMAXPRESSUREONEDGE(moveCost));
                *MoveCost__MOVESCOSTATMAXPRESSUREONEDGE(moveCost) += cost;
              } else {
                ++(*MoveCost__NBMOVESONEDGE(moveCost));
                *MoveCost__MOVESONEDGECOST(moveCost) += cost;
              }
              if(first == true) {
                first = false;
                ++(*MoveCost__NBEDGESWITHMOVES(moveCost));
                *MoveCost__COSTNBEDGESWITHMOVES(moveCost)
                  += cost;
              }
            } else {
              if(TemporarySet_count(liveVar) == nbAvailableRegs) {
                ++(*MoveCost__NBMOVESATMAXPRESSURE(moveCost));
                *MoveCost__MOVESCOSTATMAXPRESSURE(moveCost) += cost;
              } else {
                ++(*MoveCost__NBMOVES(moveCost));
                *MoveCost__MOVESCOST(moveCost) += cost;
              }
            }
          }
        }
        // swap counted as move
      } else
// FIXME: Have a more general method to compute swap statistic
#ifdef MDS_TARGET_st200
             if(Operation_operator(operation)
                == Operator_st200_xor_1general_2general_3general
                && Temporary_regFile(Operation_getResult(operation, 0)) == regFile) {
        if(id >= counter) {
          if(TemporarySet_count(liveVar) == nbAvailableRegs) {
            ++(*MoveCost__NBMOVESATMAXPRESSUREONEDGE(moveCost));
            *MoveCost__MOVESCOSTATMAXPRESSUREONEDGE(moveCost) += cost;
          } else {
            ++(*MoveCost__NBMOVESONEDGE(moveCost));
            *MoveCost__MOVESONEDGECOST(moveCost) += cost;
          }
          if(first == true) {
            first = false;
            ++(*MoveCost__NBEDGESWITHMOVES(moveCost));
            *MoveCost__COSTNBEDGESWITHMOVES(moveCost)
              += cost;
          }
        } else {
          if(TemporarySet_count(liveVar) == nbAvailableRegs) {
            ++(*MoveCost__NBMOVESATMAXPRESSURE(moveCost));
            *MoveCost__MOVESCOSTATMAXPRESSURE(moveCost) += cost;
          } else {
            ++(*MoveCost__NBMOVES(moveCost));
            *MoveCost__MOVESCOST(moveCost) += cost;
          }
        }
      }  else
#endif
               if(Operator_isControl(Operation_operator(operation))
                 && id >= counter) {
        ++(*MoveCost__NBJUMPFROMEDGE(moveCost));
        *MoveCost__JUMPFROMEDGECOST(moveCost) += cost;
      } else {
        ++(*MoveCost__NBOPERATIONS(moveCost));
        *MoveCost__OPERATIONSCOST(moveCost) += cost;
      }
      Operation_RESULTS_FOREACH_Temporary(operation, temp) {
        TemporarySet_remove(liveVar, temp);
      } Operation_RESULTS_ENDEACH_Temporary;
      Operation_ARGUMENTS_FOREACH_Temporary(operation, temp) {
        TemporarySet_insert(liveVar, temp);
      } Operation_ARGUMENTS_ENDEACH_Temporary;
    } BasicBlock_ENDBACK_Operation;
    BitSet_Dtor(bitSet);
    if(id < counter) {
      ++(*MoveCost__NBBBS(moveCost));
      *MoveCost__BBSCOST(moveCost) += cost;
    }
  } CodeRegion_ENDEACH_BasicBlock;
  TemporarySet_Dtor(liveVar);
}

#line 1239 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
dispStatistic(CodeRegion codeRegion, FILE* file, MoveCost moveCost,
              char* message) {
  Procedure procedure = CodeRegion_procedure(codeRegion);

#line 1244 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  Except_CHECK(file != NULL);
  fprintf(file, "------- %s: %s -------\n", Procedure_name(procedure),
          message);
    fprintf(file, "                       Count\tCost\n");
    fprintf(file, "Moves:                 %u\t%f\n", MoveCost_NBMOVES(moveCost),
            MoveCost_MOVESCOST(moveCost));
    fprintf(file, "Moves at max:          %u\t%f\n",
            MoveCost_NBMOVESATMAXPRESSURE(moveCost),
            MoveCost_MOVESCOSTATMAXPRESSURE(moveCost));
    fprintf(file, "Moves on edges:        %u\t%f\n", MoveCost_NBMOVESONEDGE(moveCost),
            MoveCost_MOVESONEDGECOST(moveCost));
    fprintf(file, "Moves on edges at max: %u\t%f\n",
            MoveCost_NBMOVESATMAXPRESSUREONEDGE(moveCost),
            MoveCost_MOVESCOSTATMAXPRESSUREONEDGE(moveCost));
    fprintf(file, "Other operations: %u\t%f\n",
            MoveCost_NBOPERATIONS(moveCost),
            MoveCost_OPERATIONSCOST(moveCost));
    fprintf(file, "Edges with moves: %u\t%f\n",
            MoveCost_NBEDGESWITHMOVES(moveCost),
            MoveCost_COSTNBEDGESWITHMOVES(moveCost));
    fprintf(file, "Jump from edges: %u\t%f\n",
            MoveCost_NBJUMPFROMEDGE(moveCost),
            MoveCost_JUMPFROMEDGECOST(moveCost));
    fprintf(file, "Total basic blocks: %u\t%f\n",
            MoveCost_NBBBS(moveCost),
            MoveCost_BBSCOST(moveCost));
    fprintf(file, "--------------\n");
}

#line 1274 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
static void
statisticOnPermutation(FILE* output, CodeRegion codeRegion, Liveness liveness,
                       RegFile regFile, char*message, uint32_t counter) {
  MoveCost_ moveCost[1];
  MoveCost_Ctor(moveCost);
  computeStatistic(codeRegion, liveness, regFile, moveCost, counter);
  dispStatistic(codeRegion, output, moveCost, message);
  MoveCost_Dtor(moveCost);
}

#line 1288 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
bool
Florent_allocCodeRegion(SSAForm ssaForm) {
#ifndef Disp_POOLED
#   define Disp_POOLED 128
#endif//Disp_POOLED

#line 1294 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  int32_t regFile;
  CodeRegion codeRegion = SSAForm_codeRegion(ssaForm);
  Procedure procedure = CodeRegion_procedure(codeRegion);
  Memory memory = Memory_new(Procedure_memory(procedure), Disp_POOLED);
  const char* prgName = getenv("FLORENT_ALLOC");
  TemporaryTable globalTable;
  Liveness liveness;
  bool globalResult = true;
  bool addMoveInterference = getenv("ADD_MOVE_INTERFERENCE") != NULL;
  bool noMoveAffinity = getenv("NO_MOVE_AFFINITY") != NULL;
  uint32_t maxIdBB;
  Except_CHECK(prgName != NULL);

#line 1307 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  // Initialize all options used by graph coloring and related stuff (like out
  // of colored ssa
  GraphColoring_initialize();

#line 1311 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  codeRegion_renameVariablesToVirtual(codeRegion);
  globalTable = CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);

#line 1315 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
  for (regFile = 0; regFile < RegFile__COUNT; ++regFile) {
    InterfGraphBuilder_ interfGraphBuilder[1];
    bool result;
    if(regFile == RegFile__UNDEF) {
      continue;
    }
    InterfGraphBuilder_Ctor(interfGraphBuilder, memory,
                            codeRegion, regFile, false, false,
                            addMoveInterference, noMoveAffinity, true);

#line 1325 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
    result = Florent_allocCodeRegion_doIt(memory, prgName, codeRegion,
                                         regFile, interfGraphBuilder);
    if(result == false) {
      InterfGraphBuilder_Dtor(interfGraphBuilder);
      globalTable = CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
      InterfGraphBuilder_Ctor(interfGraphBuilder, memory,
                              codeRegion, regFile, true,
                              getenv("NEW_HANDLING_OF_NEG_AFF") != NULL,
                              addMoveInterference, noMoveAffinity, true);

#line 1335 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      result = Florent_allocCodeRegion_doIt(memory, prgName, codeRegion,
                                            regFile, interfGraphBuilder);
      Except_CHECK(result == true
                   && "Alloc must work at this point, otherwise spilling is wrong");
      if(IDList_isEmpty(InterfGraphBuilder_SPECIALNODES(interfGraphBuilder))) {
        Florent_repairCodeRegion(memory, codeRegion, regFile, interfGraphBuilder);
      } else {
        // Special nodes require new repairing method
        Florent_newRepairCodeRegion(memory, codeRegion, regFile, interfGraphBuilder);
      }
    }
    if(result == true) {
      Florent_cleanCode(memory, codeRegion, interfGraphBuilder, regFile);
    }

#line 1350 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
    globalResult &= result;
    GRA_statisticOnCodeRegion(codeRegion, "Before discard phi", GRA_DISP_COST);
    // give a correct liveness to discard PHI function.
    // the liveness changed when we repaired the allocation.
    Liveness_kill(InterfGraphBuilder_LIVENESS(interfGraphBuilder));
    InterfGraphBuilder_LIVENESS(interfGraphBuilder) =
      Liveness_make(InterfGraphBuilder_MEMORY(interfGraphBuilder), codeRegion,
                    CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0),
                    true);

#line 1360 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
    if(GRA_DISP_PCOPY_COST) {
      char message[256];
      maxIdBB = IndexedFactory_COUNTER(BasicBlock_Factory);
      strcpy(message, "Permutation cost before discard phi (");
      Except_CHECK(strlen(RegFile_name(regFile))
                   + strlen("Permutation cost before discard phi ()") + 1 < 256);
      strcat(message, RegFile_name(regFile));
      strcat(message, ")");
      statisticOnPermutation(GRA_DISP_PCOPY_COST, codeRegion,
                             InterfGraphBuilder_LIVENESS(interfGraphBuilder),
                             regFile, message, maxIdBB);
    }
    OutOfColoredSSA_discardPHIs(codeRegion, regFile,
                                InterfGraphBuilder__LIVENESS(interfGraphBuilder),
                                memory);

#line 1376 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
    if(GRA_DISP_PCOPY_COST) {
      char message[256];
      strcpy(message, "Permutation cost after discard phi (");
      Except_CHECK(strlen(RegFile_name(regFile))
                   + strlen("Permutation cost before discard phi ()") + 1 < 256);
      strcat(message, RegFile_name(regFile));
      strcat(message, ")");

#line 1384 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      Liveness_kill(InterfGraphBuilder_LIVENESS(interfGraphBuilder));
      InterfGraphBuilder_LIVENESS(interfGraphBuilder) =
        Liveness_make(InterfGraphBuilder_MEMORY(interfGraphBuilder), codeRegion,
                      CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0),
                      true);

#line 1390 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
      statisticOnPermutation(GRA_DISP_PCOPY_COST, codeRegion,
                             InterfGraphBuilder_LIVENESS(interfGraphBuilder),
                             regFile, message, maxIdBB);
    }
    InterfGraphBuilder_Dtor(interfGraphBuilder);
  }
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_setFlags(basicBlock, BasicBlockFlag_Allocated);
  } CodeRegion_ENDEACH_BasicBlock;
  Memory_delete(memory);
  return globalResult;
}

#line 1403 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1415 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
#if XCC__C
static void
Florent_TEST(void)
{
#include "GRA/Florent_.c"
}

#line 1422 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/Florent.xcc"
int
main(int argc, char **argv)
{
  GRA_Test(Florent_TEST);
  return 0;
}
#endif


#endif//Florent_c_INCLUDED
