#ifndef CycleBased_c_INCLUDED
#define CycleBased_c_INCLUDED

#line 17 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
#include "SCD/SCD_.h"

#line 22 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
#include <float.h>

#line 24 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 26 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
//#define DETAILED_TRACE

#line 39 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
#define INFINITY FLT_MAX

#line 45 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
int
CycleBased_ScheduleItemPriorityCompare(const_IDListItem a, const_IDListItem b)
{
  CycleBasedScheduleItem listItemA = (CycleBasedScheduleItem)a;
  CycleBasedScheduleItem listItemB = (CycleBasedScheduleItem)b;
  return *listItemA->Priority - *listItemB->Priority;
} // end of CycleBased_ScheduleItemPriorityCompare

#line 108 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_CreateScheduleContext(int32_t nodeCount, int32_t arcCount, 
                                 CycleBasedScheduleContext context)
{
  CycleBasedNodeId nodeId;
  IDList inArcs;
  IDList outArcs;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  IDList_Ctor(context->ReadyList, Memory_Root);
  IDList_Ctor(context->ReadyAtCurrCycleList, Memory_Root);
  IDList_Ctor(context->ReadyNextCycleList, Memory_Root);
  IDList_Ctor(context->ScheduledAtCurrCycleList, Memory_Root);
  IDList_Ctor(context->UnScheduledAtCurrCycleList, Memory_Root);
  context->NodeCount = nodeCount;
  context->ArcCount = arcCount;
  context->Latency = Memory_alloc(Memory_Root, arcCount*sizeof(int));
  context->Src = Memory_alloc(Memory_Root, arcCount*sizeof(CycleBasedNodeId));
  context->Dst = Memory_alloc(Memory_Root, arcCount*sizeof(CycleBasedNodeId));
  context->InArcs = Memory_alloc(Memory_Root, nodeCount*sizeof(IDList_));
  inArcs = context->InArcs;
  context->OutArcs = Memory_alloc(Memory_Root, nodeCount*sizeof(IDList_));
  outArcs = context->OutArcs;
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    IDList_Ctor((inArcs + nodeId), Memory_Root);
    IDList_Ctor((outArcs + nodeId), Memory_Root);
  } // end for: nodeId
  resourceContext->ReservationsMap = Memory_alloc(Memory_Root,
                                                  nodeCount*Resource__COUNT*sizeof(int));
  memset(resourceContext->ReservationsMap, 0, nodeCount*Resource__COUNT*sizeof(int));
  context->Priorities = Memory_alloc(Memory_Root, nodeCount*sizeof(int));
  context->IsScheduled = Memory_alloc(Memory_Root, nodeCount*sizeof(bool));
  context->ScheduledAt = Memory_alloc(Memory_Root, nodeCount*sizeof(CycleBasedCycle));
  context->PredecessorsNotScheduled = Memory_alloc(Memory_Root, nodeCount*sizeof(int));
} // end of CycleBased_CreateScheduleContext

#line 149 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_ReleaseScheduleContext(CycleBasedScheduleContext context)
{
  CycleBasedNodeId nodeId;
  IDList inArcs;
  IDList outArcs;
  int32_t nodeCount = context->NodeCount;
  ASSERT(IDList_isEmpty(context->ReadyList));
  IDList_Dtor(context->ReadyList);
  ASSERT(IDList_isEmpty(context->ReadyAtCurrCycleList));
  IDList_Dtor(context->ReadyAtCurrCycleList);
  ASSERT(IDList_isEmpty(context->ReadyNextCycleList));
  IDList_Dtor(context->ReadyNextCycleList);
  ASSERT(IDList_isEmpty(context->ScheduledAtCurrCycleList));
  IDList_Dtor(context->ScheduledAtCurrCycleList);
  ASSERT(IDList_isEmpty(context->UnScheduledAtCurrCycleList));
  IDList_Dtor(context->UnScheduledAtCurrCycleList);
  inArcs = context->InArcs;
  outArcs = context->OutArcs;
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    IDList_Dtor((outArcs + nodeId));
    IDList_Dtor((inArcs + nodeId));
  } // end for: nodeId
} // end of CycleBased_ReleaseScheduleContext

#line 183 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
bool
CycleBased_NoResourceConstraints(CycleBasedNodeId nodeId, int resourceUsageMap[],
                                 CycleBasedResourceContext context)
{
  int resource;
  int *reservationsMap = context->ReservationsMap;
  for (resource = 0; resource < Resource__COUNT; ++resource) {
    if (resourceUsageMap[resource] - reservationsMap[resource + nodeId*Resource__COUNT] < 0) {
      return false;
    } // end if
  } // end for: resource
  return true;
} // end of CycleBased_NoResourceConstraints

#line 204 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_AddResourcesUsage(CycleBasedNodeId nodeId, int resourceUsageMap[],
                             CycleBasedResourceContext context)
{
  int resource;
  int *reservationsMap = context->ReservationsMap;
  for (resource = 0; resource < Resource__COUNT; ++resource) {
    resourceUsageMap[resource] -= reservationsMap[resource + nodeId*Resource__COUNT];
    Except_CHECK(resourceUsageMap[resource] >= 0);
  } // end for: resource
} // end of CycleBased_AddResourcesUsage

#line 223 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_ReleaseResourcesUsage(CycleBasedNodeId nodeId, int resourceUsageMap[],
                                 CycleBasedResourceContext context)
{
  int resource;
  int *reservationsMap = context->ReservationsMap;
  int *resourceAvailabilityMap = context->ResourceAvailabilityMap;
  for (resource = 0; resource < Resource__COUNT; ++resource) {
    resourceUsageMap[resource] += reservationsMap[resource + nodeId*Resource__COUNT];
    Except_CHECK(resourceUsageMap[resource] <= resourceAvailabilityMap[resource]);
  } // end for: resource
} // end of CycleBased_ReleaseResourcesUsage

#line 242 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_ClearResourcesUsage(int resourceUsageMap[], CycleBasedResourceContext context)
{
  int resource;
  int *resourceAvailabilityMap = context->ResourceAvailabilityMap;
  for (resource = 0; resource < Resource__COUNT; ++resource) {
    resourceUsageMap[resource] = resourceAvailabilityMap[resource];
  } // end for: resource
} // end of CycleBased_ClearResourcesUsage

#line 258 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void 
CycleBased_ResourceAvailabilityInit(Scheduler scheduler, CycleBasedResourceContext resourceContext)
{
  int resource;
  BlockTrace blockTrace = Scheduler_BLOCKTRACE(scheduler);
  Processor processor = BlockTrace_PROCESSOR(blockTrace);
  int *resourceAvailabilityMap = resourceContext->ResourceAvailabilityMap;
  SCD_TRACE && fprintf(SCD_TRACE, "Resource availability: ");
  for (resource = 0; resource < Resource__COUNT; resource++) {
    SCD_TRACE && fprintf(SCD_TRACE, "%d ", Resource_availability(resource, processor));
    resourceAvailabilityMap[resource] = Resource_availability(resource, processor);
  }  // end for: resource
  SCD_TRACE && fprintf(SCD_TRACE, "\n");
} // end of CycleBased_ResourceAvailabilityInit

#line 281 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void 
CycleBased_ReservationTableInit(DDGraphNode node,
                                CycleBasedNodeId nodeId, 
                                int reservationsMap[])
{
  Operation operation = DDGraphNode_OPERATION(node);
  IssueItem issueItem = Operation_issueItem(operation);
  Scheduling scheduling = IssueItem_SCHEDULING(issueItem);
  Reservation reservation = Scheduling_reservation(scheduling);
  const_ReservationTable reservationTable = Reservation_table(reservation);
  const ResourceVector_ *columns = ReservationTable_columns(reservationTable);
  int columnCount = ReservationTable_columnCount(reservationTable), i, resource;
  int *entryp, num_entries = Resource__COUNT*columnCount;
  Except_CHECK(columnCount <= Reservation_COLUMNS_COUNT_MAX);
  ASSERT(columnCount == 0 || columnCount == 1); // TODO!!! will work only with st200, one cycle execution assumption!
  entryp = (reservationsMap + nodeId*Resource__COUNT);
  for (i = 0; i < columnCount; i++) {
    SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d\treservation cycle: %d\tresources: ", nodeId, i);
    for (resource = 0; resource < Resource__COUNT; resource++) {
      *(entryp + resource) = ResourceVector_UNITS(columns + i)[resource];
      SCD_TRACE && fprintf(SCD_TRACE, "%d ", *(entryp + resource));
    } // end for: resource
    SCD_TRACE && fprintf(SCD_TRACE, "\n");
  } // end for: i
  Except_CHECK(Resource__COUNT <= num_entries || (num_entries == 0 && columnCount == 0));
} // end of CycleBased_ReservationTableInit

#line 314 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_GetSinkNodes(IDList bfsList, CycleBasedScheduleContext context)
{
  CycleBasedNodeId nodeId;
  int32_t nodeCount = context->NodeCount;
  IDList outArcs = context->OutArcs;
  SCD_TRACE && fprintf(SCD_TRACE, "*** SINK NODES\n");
  Except_CHECK(IDList_isEmpty(bfsList));
  for (nodeId = nodeCount-1; nodeId >= 0; --nodeId) {
    if (IDList_isEmpty(outArcs + nodeId)) {
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d is a sink node\n", nodeId);
      *(CycleBasedNodeId *)IDList_put(bfsList, sizeof(CycleBasedNodeId)) = nodeId;
    } // end if
  } // end for: nodeId
} // end of CycleBased_GetSinkNodes

#line 338 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_CalculatePriorities(void (*getSinkNodes)(IDList bfsList,
                                                    CycleBasedScheduleContext context),
                               CycleBasedScheduleContext context)
{
  IDList_ bfsList[1];
  CycleBasedNodeId nodeId;
  CycleBasedArcId arcId;
  int32_t nodeCount = context->NodeCount;
  int32_t arcCount = context->ArcCount;
  int *latency = context->Latency;
  IDList inArcs = context->InArcs;
  IDList outArcs = context->OutArcs;
  CycleBasedNodeId *src = context->Src;
  CycleBasedNodeId *dst = context->Dst;
  int *priorities = context->Priorities;
  bool *visitedArcs = Memory_alloc(Memory_Root, arcCount*sizeof(bool));
  memset(visitedArcs, 0, arcCount*sizeof(bool));
  memset(priorities, 0, nodeCount*sizeof(int));
  IDList_Ctor(bfsList, Memory_Root);
  // calculation of critical path using modified BFS algorithm
  // at the beginning a sink nodes should be put at the head of a walktrough list as start points
  (*getSinkNodes)(bfsList, context);
  Except_CHECK(!IDList_isEmpty(bfsList));
  // BFS loop
  // walk from start points to every other node and update priorities
  while (!IDList_isEmpty(bfsList)) {
    CycleBasedNodeId prevNodeId;
    int priority = 0;
    nodeId = *(CycleBasedNodeId *)IDList_lastItem(bfsList);
    prevNodeId = nodeId;
    IDList_pop(bfsList, NULL);
    IDList_FOREACH((inArcs + nodeId), CycleBasedArcId, pArcId) {
      arcId = *pArcId;
      if (visitedArcs[arcId] == false) {
        visitedArcs[arcId] = true;
        *(CycleBasedNodeId *)IDList_put(bfsList, sizeof(CycleBasedNodeId)) = src[arcId];
      } // end if
    } IDList_ENDEACH;
    IDList_FOREACH((outArcs + nodeId), CycleBasedArcId, pArcId) {
      CycleBasedNodeId dstNodeId; 
      int newPriority;
      arcId = *pArcId;
      dstNodeId = dst[arcId];
      newPriority = priorities[dstNodeId] + latency[arcId];
      if (newPriority > priority ) {
        priority = newPriority;
      } // end if
    } IDList_ENDEACH;
    priorities[nodeId] = priority;
  } // end while
  IDList_Dtor(bfsList);
} // end of CycleBased_CalculatePriorities

#line 398 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_PickReadyAtCurrCycle(CycleBasedCycle cycle, CycleBasedScheduleContext context)
{
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  IDList inArcs = context->InArcs;
  CycleBasedNodeId *src = context->Src;
  int *latency = context->Latency;
  int *priorities = context->Priorities;
  IDList readyList = context->ReadyList;
  IDList readyAtCurrCycleList = context->ReadyAtCurrCycleList;
  IDList_FOREACH(readyList, CycleBasedNodeId, pNodeId) {
    bool ready = true;
    IDList_FOREACH((inArcs + *pNodeId), CycleBasedArcId, pArcId) {
      if (cycle < scheduledAt[src[*pArcId]] + latency[*pArcId]) {
        ready = false;
        break;
      } // end if
    } IDList_ENDEACH;
    if (ready == true) {
      CycleBasedScheduleItem scheduleItem;
      // and add to list
      scheduleItem = IDList_push(readyAtCurrCycleList, sizeof(CycleBasedScheduleItem_));
      scheduleItem->NodeId = *pNodeId;
      scheduleItem->Priority = (priorities + *pNodeId);
#ifdef DETAILED_TRACE
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d ", *pNodeId);
      SCD_TRACE && fprintf(SCD_TRACE, "is ready to be scheduled at this cycle\n");
#endif
      IDList_remove(readyList, pNodeId, NULL);
    } // end if
  } IDList_ENDEACH;
} // end of CycleBased_PickReadyAtCurrCycle

#line 441 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
bool
CycleBased_callbackPreScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                                   CycleBasedScheduleContext context)
{
  int *resourceUsageMap = context->ResourceUsageMap;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  bool canBeScheduled = CycleBased_NoResourceConstraints(nodeId, resourceUsageMap, resourceContext);
  if (canBeScheduled == true) {
    CycleBased_AddResourcesUsage(nodeId, resourceUsageMap, resourceContext);
#ifdef DETAILED_TRACE
  } else {
    SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d CANNOT be scheduled ", nodeId);
    SCD_TRACE && fprintf(SCD_TRACE, "because there are not enough resources\n");
#endif
  } // end if
  return canBeScheduled;
} // end of CycleBased_callbackPreScheduleNode

#line 466 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_callbackPreUnScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                                     CycleBasedScheduleContext context)
{
  int *resourceUsageMap = context->ResourceUsageMap;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  CycleBased_ReleaseResourcesUsage(nodeId, resourceUsageMap, resourceContext);
} // end of CycleBased_callbackPreunScheduleNode

#line 482 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_callbackPostScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                                    CycleBasedScheduleContext context)
{
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d has been scheduled at cycle: %d\n",
                                  nodeId, cycle);
#endif
} // end of CycleBased_callbackPostScheduleNode

#line 498 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_callbackNewCycle(CycleBasedCycle cycle, CycleBasedScheduleContext context)
{
  int *resourceUsageMap = context->ResourceUsageMap;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  CycleBased_ClearResourcesUsage(resourceUsageMap, resourceContext);
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "*** CYCLE=%d\n", cycle);
#endif
} // end of CycleBased_callbackNewCycle

#line 509 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
/*
 * CycleBased_ScheduleNode -- Schedules an operation and updates readyList with proper successors.
 *
 * Return: boolean success if an operation has been scheduled.
 */
static bool
CycleBased_ScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                        CycleBasedScheduleContext context, CycleBasedScheduleInterface interface)
{
  bool *isScheduled = context->IsScheduled;
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  IDList outArcs = context->OutArcs;
  IDList inArcs = context->InArcs;
  CycleBasedNodeId *src = context->Src;
  CycleBasedNodeId *dst = context->Dst;
  int *latency = context->Latency;
  int *priorities = context->Priorities;
  int *predecessorsNotScheduled = context->PredecessorsNotScheduled;
  IDList readyList = context->ReadyList;
  IDList scheduledAtCurrCycleList = context->ScheduledAtCurrCycleList;
  CycleBasedArcId arcId;
  CycleBasedScheduleItem scheduleItem;
  Except_CHECK(isScheduled[nodeId] == false);
  IDList_FOREACH((inArcs + nodeId), CycleBasedArcId, pArcId) {
    CycleBasedNodeId srcNodeId = src[*pArcId];
    Except_CHECK(isScheduled[srcNodeId] == true);
    Except_CHECK(scheduledAt[srcNodeId] + latency[*pArcId] <= cycle);
  } IDList_ENDEACH;
  if ((*interface->callbackPreScheduleNode)(nodeId, cycle, context) == false) {
    return false;
  } // end if
  // update schedule flag
  isScheduled[nodeId] = true;
  // update cycle number in which task was scheduled
  scheduledAt[nodeId] = cycle;
  IDList_FOREACH((outArcs + nodeId), CycleBasedArcId, pArcId) {
    CycleBasedNodeId dstNodeId = dst[*pArcId];
    if (--predecessorsNotScheduled[dstNodeId] == 0) {
      Except_CHECK(isScheduled[dstNodeId] == false);
      *(CycleBasedNodeId *)
      IDList_push(readyList, sizeof(CycleBasedNodeId)) = dstNodeId;
#ifdef DETAILED_TRACE
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d is ready to be scheduled\n", dstNodeId);
#endif
    } // end if
    Except_CHECK(predecessorsNotScheduled[dstNodeId] >= 0);
  } IDList_ENDEACH;
  scheduleItem = IDList_push(scheduledAtCurrCycleList, sizeof(CycleBasedScheduleItem_));
  scheduleItem->NodeId = nodeId;
  scheduleItem->Priority = priorities + nodeId;
  (*interface->callbackPostScheduleNode)(nodeId, cycle, context);
  return true;
} // end of CycleBased_ScheduleNode

#line 563 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
/*
 * CycleBased_UnScheduleNode -- Unschedules an operation.
 */
static void
CycleBased_UnScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                          CycleBasedScheduleContext context, CycleBasedScheduleInterface interface)
{
  bool *isScheduled = context->IsScheduled;
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  IDList outArcs = context->OutArcs;
  IDList inArcs = context->InArcs;
  CycleBasedNodeId *src = context->Src;
  CycleBasedNodeId *dst = context->Dst;
  int *latency = context->Latency;
  int *predecessorsNotScheduled = context->PredecessorsNotScheduled;
  IDList unScheduledAtCurrCycleList = context->UnScheduledAtCurrCycleList;
  CycleBasedArcId arcId;
  Except_CHECK(isScheduled[nodeId] == true);
  IDList_FOREACH((inArcs + nodeId), CycleBasedArcId, pArcId) {
    CycleBasedNodeId srcNodeId = src[*pArcId];
    Except_CHECK(isScheduled[srcNodeId] == true);
  } IDList_ENDEACH;
  IDList_FOREACH((outArcs + nodeId), CycleBasedArcId, pArcId) {
    CycleBasedNodeId dstNodeId = dst[*pArcId];
    Except_CHECK(isScheduled[dstNodeId] == false);
    ++predecessorsNotScheduled[dstNodeId];
    Except_CHECK(predecessorsNotScheduled[dstNodeId] >= 0);
  } IDList_ENDEACH;
  (*interface->callbackPreUnScheduleNode)(nodeId, cycle, context);
  // update schedule flag
  isScheduled[nodeId] = false;
  *(CycleBasedNodeId *)IDList_push(unScheduledAtCurrCycleList, 
                                   sizeof(CycleBasedNodeId)) = nodeId;
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d has been unscheduled at cycle: %d\n",
                                  nodeId, cycle);
#endif
} // end of CycleBased_UnScheduleNode

#line 602 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
/*
 * CycleBased_ReScheduleNode -- Reschedules operations.
 *
 * Return: boolean success if a demanded operation has been scheduled.
 */
static bool
CycleBased_ReScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                          CycleBasedScheduleContext context, CycleBasedScheduleInterface interface)
{
  bool nodeIsScheduled;
  int *priorities = context->Priorities;
  bool *isScheduled = context->IsScheduled;
  CycleBasedNodeId *dst = context->Dst;
  IDList scheduledAtCurrCycleList = context->ScheduledAtCurrCycleList;
  IDList unScheduledAtCurrCycleList = context->UnScheduledAtCurrCycleList;
  IDList outArcs = context->OutArcs;
  IDList readyList = context->ReadyList;
  IDList readyAtCurrCycleList = context->ReadyAtCurrCycleList;
  IDList readyNextCycleList = context->ReadyNextCycleList;
  int *resourceUsageMap = context->ResourceUsageMap;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  // check for reschedule
  IDList_sort(scheduledAtCurrCycleList, CycleBased_ScheduleItemPriorityCompare);        
  IDList_FOREACH(scheduledAtCurrCycleList, CycleBasedScheduleItem_, toUnScheduleItem) {
    bool canBeUnScheduled = true;
    CycleBasedNodeId toUnScheduleNodeId = toUnScheduleItem->NodeId;
    // nodes has been sorted so check for possible reschedule only for nodes with lower priorities
    if (priorities[toUnScheduleNodeId] >= priorities[nodeId]) break;
    IDList_FOREACH((outArcs + toUnScheduleNodeId), CycleBasedArcId, pArcId) {
      // if one its successors is scheduled then it can't be rescheduled
      if (isScheduled[dst[*pArcId]] == true) {
        canBeUnScheduled = false;
        break;
        // if it is predecessor of current node then it can't be rescheduled
      } else if (dst[*pArcId] == nodeId) {
        canBeUnScheduled = false;
        break;
      } // end if
    } IDList_ENDEACH;
    if (canBeUnScheduled == true) {
      CycleBased_UnScheduleNode(toUnScheduleNodeId, cycle, context, interface);
      IDList_remove(scheduledAtCurrCycleList, toUnScheduleItem, NULL);
      IDList_FOREACH((outArcs + toUnScheduleNodeId), CycleBasedArcId, pArcId) {
        CycleBasedScheduleItem toRemoveItem = NULL;
        // remove every successor node from 'ready at current cycle' list
        // and put it again to unscheduled list
        IDList_FOREACH(readyAtCurrCycleList, CycleBasedScheduleItem_, scheduleItem) {
          if (scheduleItem->NodeId == dst[*pArcId]) {
            toRemoveItem = scheduleItem;
            break;
          } // end if
        } IDList_ENDEACH;
        if (toRemoveItem != NULL) {
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "\t\tnodeId: %d removed from readyAtCurrCycleList\n", 
                                  toRemoveItem->NodeId);
#endif
          IDList_remove(readyAtCurrCycleList, toRemoveItem, NULL);    
        } else {
          CycleBasedNodeId *toRemoveNodeId = NULL;
          // do the same with ready list 
          IDList_FOREACH(readyList, CycleBasedNodeId, pNodeId) {
            if (*pNodeId == dst[*pArcId]) {
              toRemoveNodeId = pNodeId; 
              break;
            } // end if
          } IDList_ENDEACH;
          if (toRemoveNodeId != NULL) {
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "\t\tnodeId: %d removed from readyList\n", 
                                  *toRemoveNodeId);
#endif
            IDList_remove(readyList, toRemoveNodeId, NULL);
          } // end if
          toRemoveNodeId = NULL;
          // and ready next cycle list 
          IDList_FOREACH(readyNextCycleList, CycleBasedNodeId, pNodeId) {
            if (*pNodeId == dst[*pArcId]) {
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "\t\tnodeId: %d removed from readyNextCycleList\n", 
                                  *pNodeId);
#endif
              toRemoveNodeId = pNodeId; 
              break;
            } // end if
          } IDList_ENDEACH;
          if (toRemoveNodeId != NULL) {
            IDList_remove(readyNextCycleList, toRemoveNodeId, NULL);
          } // end if
        } // end if
      } IDList_ENDEACH;
      nodeIsScheduled = CycleBased_ScheduleNode(nodeId, cycle, context, interface);
      if (nodeIsScheduled == true) {
        while (!IDList_isEmpty(unScheduledAtCurrCycleList)) {
          CycleBasedScheduleItem scheduleItem;
          nodeId = *(CycleBasedNodeId *)IDList_lastItem(unScheduledAtCurrCycleList);
          scheduleItem = IDList_put(readyAtCurrCycleList, sizeof(CycleBasedScheduleItem_));
          scheduleItem->NodeId = nodeId;
          scheduleItem->Priority = priorities + nodeId;
          IDList_pop(unScheduledAtCurrCycleList, NULL);
        } // end while
        return true;
      } // end if
    } // end if
  } IDList_ENDEACH;
  while (!IDList_isEmpty(unScheduledAtCurrCycleList)) {
    CycleBasedScheduleItem scheduleItem;
    nodeId = *(CycleBasedNodeId *)IDList_lastItem(unScheduledAtCurrCycleList);
    IDList_FOREACH(readyList, CycleBasedNodeId, pNodeId) {
      if (nodeId == *pNodeId) {
        IDList_remove(readyList, pNodeId, NULL);
        break;
      } // end if
    } IDList_ENDEACH;
    nodeIsScheduled = CycleBased_ScheduleNode(nodeId, cycle, context, interface);
    Except_CHECK(nodeIsScheduled == true);
    IDList_pop(unScheduledAtCurrCycleList, NULL);
  } // end while
  return false;
} // end of CycleBased_ReScheduleNode

#line 729 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
bool
CycleBased_Schedule(CycleBasedScheduleContext context, CycleBasedScheduleInterface interface)
{
  CycleBasedNodeId nodeId;
  CycleBasedNodeId arcId;
  int resource;
  int cycle;
  int32_t nodeCount = context->NodeCount;
  int32_t arcCount = context->ArcCount;
  int32_t cycleCount;
  int *latency = context->Latency;
  int *priorities = context->Priorities;
  IDList inArcs = context->InArcs;
  IDList outArcs = context->OutArcs;
  IDList readyList = context->ReadyList;
  IDList readyAtCurrCycleList = context->ReadyAtCurrCycleList;
  IDList readyNextCycleList = context->ReadyNextCycleList;
  IDList scheduledAtCurrCycleList = context->ScheduledAtCurrCycleList;
  IDList unScheduledAtCurrCycleList = context->UnScheduledAtCurrCycleList;
  CycleBasedNodeId *src = context->Src;
  CycleBasedNodeId *dst = context->Dst;
  int *resourceUsageMap = context->ResourceUsageMap;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  bool *isScheduled = context->IsScheduled;
  int *predecessorsNotScheduled;
  predecessorsNotScheduled = context->PredecessorsNotScheduled;
  context->InfeasibleSchedule = false;
  IDList_empty(readyList, NULL);
  IDList_empty(readyNextCycleList, NULL);
  IDList_empty(readyAtCurrCycleList, NULL);
  SCD_TRACE && fprintf(SCD_TRACE, "*** INITIAL NODES\n");
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    // count number of predecessors
    int predNum = IDList_count(inArcs + nodeId);
    predecessorsNotScheduled[nodeId] = predNum;
    // initialize flags because no nodes are scheduled
    isScheduled[nodeId] = false;
    if (predNum == 0) {
      *(CycleBasedNodeId *)IDList_push(readyList, sizeof(CycleBasedNodeId)) = nodeId;
#ifdef DETAILED_TRACE
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d is ready to be scheduled\n",
                                      nodeId);
#endif
    } // end if
  } // end for: nodeId
  SCD_TRACE && fprintf(SCD_TRACE, "*** PRIORITIES & PREDECESSORS NOT SCHEDULED\n");
#ifdef DETAILED_TRACE
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d\tpriority: %d\tpredNum: %d\n",
                                    nodeId, priorities[nodeId], predecessorsNotScheduled[nodeId]);
  } // end for: nodeId
  SCD_TRACE && fprintf(SCD_TRACE, "*** SCHEDULING\n");
#endif
  if (IDList_isEmpty(readyList)) {
    SCD_TRACE && fprintf(SCD_TRACE, "\tSchedule not found\n");
    return false;
  } // end if
  // loop when there are not all of instructions scheduled
  cycle = 0;
  while (   !IDList_isEmpty(readyList)
         || !IDList_isEmpty(readyAtCurrCycleList)) {
    IDList_empty(scheduledAtCurrCycleList, NULL);
    IDList_empty(unScheduledAtCurrCycleList, NULL);
    (*interface->callbackNewCycle)(cycle, context);
    Except_CHECK(IDList_isEmpty(readyNextCycleList));
    (*interface->PickReadyAtCurrCycle)(cycle, context);
    while (!IDList_isEmpty(readyAtCurrCycleList)) {
      bool nodeIsScheduled;
      int priority;
      CycleBasedScheduleItem nowScheduleItem;
      // sort ready list by priorites
      IDList_sort(readyAtCurrCycleList, CycleBased_ScheduleItemPriorityCompare);
#ifdef DETAILED_TRACE
      SCD_TRACE && fprintf(SCD_TRACE, "\tReady at current cycle:\n");
      IDList_FOREACH(readyAtCurrCycleList, CycleBasedScheduleItem_, scheduleItem) {
        SCD_TRACE && fprintf(SCD_TRACE, "\t\tnodeId: %d\tpriority: %d\n",
                                        scheduleItem->NodeId, *scheduleItem->Priority);
      } IDList_ENDEACH;
#endif
      // get node with highest priority
      nowScheduleItem = IDList_lastItem(readyAtCurrCycleList);
      nodeId = nowScheduleItem->NodeId;
      IDList_pop(readyAtCurrCycleList, NULL);
      nodeIsScheduled = CycleBased_ScheduleNode(nodeId, cycle, context, interface);
      if (nodeIsScheduled == true) {
        (*interface->PickReadyAtCurrCycle)(cycle, context);
      } else {
        nodeIsScheduled = CycleBased_ReScheduleNode(nodeId, cycle, context, interface);
        // add again to ready but resources full list, node not scheduled
        if (nodeIsScheduled == false) {
          if (context->InfeasibleSchedule == true) {
            SCD_TRACE && fprintf(SCD_TRACE, "\tSchedule not found\n");
            return false;
          } // end if
          *(CycleBasedNodeId *)
          IDList_put(readyNextCycleList,
                     sizeof(CycleBasedNodeId)) = nodeId;
        } // end if
      } // end if
    } // end while
    // copy all ready nodes to ready list again
    while (!IDList_isEmpty(readyNextCycleList)) {
      CycleBasedScheduleItem scheduleItem = IDList_put(readyAtCurrCycleList,
                                                       sizeof(CycleBasedScheduleItem_));
      nodeId = *(CycleBasedNodeId *)IDList_lastItem(readyNextCycleList);
      scheduleItem->NodeId = nodeId;
      scheduleItem->Priority = (priorities + nodeId);
      IDList_pop(readyNextCycleList, NULL);
    } // end while
    // increase cycle
    ++cycle;
  } // end while
  context->CycleCount = cycle;
  cycleCount = cycle;
  IDList_empty(scheduledAtCurrCycleList, NULL);
  Except_CHECK(IDList_isEmpty(readyList));
  Except_CHECK(IDList_isEmpty(readyNextCycleList));
  Except_CHECK(IDList_isEmpty(readyAtCurrCycleList));
  SCD_TRACE && fprintf(SCD_TRACE, "*** END OF SCHEDULING\n");
  // validation
  for (cycle = 0;
       cycle < cycleCount;
       ++cycle) {
    CycleBased_ClearResourcesUsage(resourceUsageMap, resourceContext);
    for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
      if (scheduledAt[nodeId] == cycle) {
        Except_CHECK(isScheduled[nodeId] == true);
        CycleBased_AddResourcesUsage(nodeId, resourceUsageMap, resourceContext);
        IDList_FOREACH((inArcs + nodeId), CycleBasedArcId, pArcId) {
          CycleBasedNodeId srcNodeId = src[*pArcId];
          Except_CHECK(isScheduled[srcNodeId] == true);
          Except_CHECK(scheduledAt[srcNodeId] + latency[*pArcId] <= cycle);
        } IDList_ENDEACH;
      } // end if
    } // end for: nodeId
  } // end for: cycle
  return true;
} // end of CycleBased_Schedule

#line 876 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
int
CycleBased_NotInnerLoopSchedule(Scheduler scheduler, int scheduleDates[], DDGraph mcrlGraph,
                                int32_t firstIndex, int32_t lastIndex)
{
  int resource;
  bool scheduleValid;
  CycleBasedArcId arcCounter = 0;
  CycleBasedNodeId nodeId;
  CycleBasedCycle cycle;
  int32_t cycleCount;
  int32_t nodeCount = 0;
  int32_t arcCount = 0;
  int *latency;
  CycleBasedNodeId *src;
  CycleBasedNodeId *dst;
  IDList inArcs;
  IDList outArcs;
  CycleBasedScheduleContext_ context[1];
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  CycleBasedScheduleInterface_ interface[1];
  int *resourceAvailabilityMap = resourceContext->ResourceAvailabilityMap;
  int *reservationsMap;
  CycleBasedCycle *scheduledAt;
#ifdef DETAILED_TRACE
  bool *isScheduled;
#endif
  Except_REQUIRE(Scheduler_isInnerLoop(scheduler) == false);
  Except_REQUIRE(lastIndex >= firstIndex);
  CycleBased_ResourceAvailabilityInit(scheduler, resourceContext);
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          if (DDGraphArc_OMEGA(arc) == 0) {
            ++arcCount;
          } // end if
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  nodeCount = lastIndex - firstIndex + 1;
  CycleBased_CreateScheduleContext(nodeCount, arcCount, context);
  latency = context->Latency;
  src = context->Src;
  dst = context->Dst;
  inArcs = context->InArcs;
  outArcs = context->OutArcs;
  reservationsMap = resourceContext->ReservationsMap;
  scheduledAt = context->ScheduledAt;
#ifdef DETAILED_TRACE
  isScheduled = context->IsScheduled;
#endif
  SCD_TRACE && fprintf(SCD_TRACE, "*** RESERVATION TABLE INIT\n");
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      Operation operation = DDGraphNode_OPERATION(node);
      IssueItem issueItem = Operation_issueItem(operation);
      Scheduling scheduling = IssueItem_SCHEDULING(issueItem);
      nodeId = index - firstIndex;
      CycleBased_ReservationTableInit(node, nodeId, reservationsMap);
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  SCD_TRACE && fprintf(SCD_TRACE, "*** ARC EXTRACTION\n");
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          int distanceValue = DDGraphArc_OMEGA(arc);
          if (distanceValue == 0) {
            CycleBasedNodeId srcNodeId = tail_index - firstIndex;
            CycleBasedNodeId dstNodeId = head_index - firstIndex;
            CycleBasedArcId arcId = arcCounter;
            *(CycleBasedArcId *)IDList_push(inArcs + dstNodeId,
                                            sizeof(CycleBasedArcId)) = arcId;
            *(CycleBasedArcId *)IDList_push(outArcs + srcNodeId,
                                            sizeof(CycleBasedArcId)) = arcId;
            src[arcId] = srcNodeId;
            dst[arcId] = dstNodeId;
            latency[arcId] = DDGraphArc_THETA(arc);
            SCD_TRACE && fprintf(SCD_TRACE, "\tarcId: %d\tsrc: %d\tdst: %d",
                                            arcId, src[arcId], dst[arcId]);
            SCD_TRACE && fprintf(SCD_TRACE, "\tlatency: %d\n", latency[arcId]);
            ++arcCounter;
          } // end if
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  CycleBased_CalculatePriorities(CycleBased_GetSinkNodes, context);
  interface->callbackPreScheduleNode = CycleBased_callbackPreScheduleNode;
  interface->callbackPostScheduleNode = CycleBased_callbackPostScheduleNode;
  interface->callbackPreUnScheduleNode = CycleBased_callbackPreUnScheduleNode;
  interface->PickReadyAtCurrCycle = CycleBased_PickReadyAtCurrCycle;
  interface->callbackNewCycle = CycleBased_callbackNewCycle;
  scheduleValid = CycleBased_Schedule(context, interface);
  Except_CHECK(scheduleValid == true);
  cycleCount = context->CycleCount;
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "*** SCHEDULE\n");
#endif
  for (cycle = 0;
       cycle < cycleCount;
       ++cycle) {
#ifdef DETAILED_TRACE
    SCD_TRACE && fprintf(SCD_TRACE, "\tcycle: %d", cycle);
#endif
    for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
      if (scheduledAt[nodeId] == cycle) {
        scheduleDates[nodeId+firstIndex] = cycle;
#ifdef DETAILED_TRACE
        SCD_TRACE && fprintf(SCD_TRACE, "\t%d", nodeId);
        Except_CHECK(isScheduled[nodeId]);
#endif
      } // end if
    } // end for: nodeId
#ifdef DETAILED_TRACE
    SCD_TRACE && fprintf(SCD_TRACE, "\n");
#endif
  } // end for: cycle
  CycleBased_ReleaseScheduleContext(context);
  SCD_TRACE && fprintf(SCD_TRACE, "STATS_PERF: cycleCount: %d\n", cycleCount);
  return cycleCount;
} // end of CycleBased_NotInnerLoopSchedule

#line 1009 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1045 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_CreateMCRContext(int32_t nodeCount, int32_t arcCount, 
                            CycleBasedMCRContext context)
{
  IDList_Ctor(context->TreeNodeList, Memory_Root);
  IDList_Ctor(context->TreeSearchNodeList, Memory_Root);
  IDList_Ctor(context->TreeArcList, Memory_Root);
  IDList_Ctor(context->ArcKeyList, Memory_Root);
  context->NodeCount = nodeCount;
  context->ArcCount = arcCount;
  context->Src = Memory_alloc(Memory_Root, arcCount*sizeof(int));
  context->Dst = Memory_alloc(Memory_Root, arcCount*sizeof(int));
  context->Latency = Memory_alloc(Memory_Root, arcCount*sizeof(int));
  context->Distance = Memory_alloc(Memory_Root, arcCount*sizeof(int));
  context->ArcKey = Memory_alloc(Memory_Root, arcCount*sizeof(float));
  context->T = Memory_alloc(Memory_Root, nodeCount*sizeof(float));
  context->D = Memory_alloc(Memory_Root, nodeCount*sizeof(float));
  context->VisitedNodes = Memory_alloc(Memory_Root, nodeCount*sizeof(bool));
  context->Parent = Memory_alloc(Memory_Root, nodeCount*sizeof(CycleBasedNodeId));
  context->NodeKey = Memory_alloc(Memory_Root, nodeCount*sizeof(CycleBasedArcId));
  context->NodeKeyNotNIL = Memory_alloc(Memory_Root, nodeCount*sizeof(bool));
} // end of CycleBased_CreateMCRContext

#line 1074 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
void
CycleBased_ReleaseMCRContext(CycleBasedMCRContext context)
{
  IDList_Dtor(context->TreeNodeList);
  IDList_Dtor(context->TreeSearchNodeList);
  IDList_Dtor(context->TreeArcList);
  IDList_Dtor(context->ArcKeyList);
} // end of CycleBased_ReleaseMCRContext

#line 1083 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
/*
 * CycleBased_FindArcKey -- Calculates a key value for each arc which is used in MCR algorithm.
 *
 * Return: float arcKey.
 */
static float
CycleBased_FindArcKey(CycleBasedArcId arcId, int II, CycleBasedMCRContext context)
{
  float dT, dD;
  float *t = context->T;
  float *d = context->D;
  CycleBasedNodeId *src = context->Src;
  CycleBasedNodeId *dst = context->Dst;
  int *latency = context->Latency;
  int *distance = context->Distance;
  dT = t[src[arcId]] + distance[arcId] - t[dst[arcId]];
  if ( dT > 0.f ) {
    dD = d[src[arcId]] - (latency[arcId] - II*distance[arcId]) - d[dst[arcId]];
    return dD/dT;
  } // end if
  return INFINITY;
} // end of CycleBased_FindArcKey

#line 1106 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
/*
 * CycleBased_MCR -- Finds minimum cycle ratio.
 *
 * Return: float arcKey which equals minimum cycle ratio.
 */
static float
CycleBased_MCR(int II, CycleBasedMCRContext context)
{
  int32_t nodeCount = context->NodeCount;
  int32_t arcCount = context->ArcCount;
  IDList treeNodeList = context->TreeNodeList;
  IDList treeSearchNodeList = context->TreeSearchNodeList;
  IDList treeArcList = context->TreeArcList;
  IDList arcKeyList = context->ArcKeyList;
  CycleBasedNodeId *src = context->Src;
  CycleBasedNodeId *dst = context->Dst;
  float *arcKey = context->ArcKey;
  bool *nodeKeyNotNIL = context->NodeKeyNotNIL;
  CycleBasedArcId *nodeKey = context->NodeKey;
  int *latency = context->Latency;
  int *distance = context->Distance;
  float *d = context->D;
  float *t = context->T;
  bool *visitedNodes = context->VisitedNodes;
  CycleBasedNodeId *parent = context->Parent;
  CycleBasedArcId arcId;
  CycleBasedNodeId nodeId;
  CycleBasedNodeId rootNodeId;
  CycleBasedArcKeyItem minArcKeyItem;
  float dSrc, dDst;
  float tSrc, tDst;
  float minArcKey = INFINITY;
  bool found;
  int iter = 0;
  SCD_TRACE && fprintf(SCD_TRACE, "CycleBased_MCR(%d)\n", II);
  IDList_empty(treeNodeList, NULL);
  IDList_empty(treeSearchNodeList, NULL);
  IDList_empty(treeArcList, NULL);
  IDList_empty(arcKeyList, NULL);
  memset(nodeKeyNotNIL, 0, nodeCount*sizeof(bool));
  memset(d, 0, nodeCount*sizeof(float));
  memset(t, 0, nodeCount*sizeof(float));
  for (arcId = 0; arcId < arcCount; ++arcId) {
    CycleBasedArcKeyItem arcKeyItem;
    // build an arcKey list but
    // reject all arcs with src equal to dst because they may cause such cycles
    // which are suitable for minimum cost cycle search criteria but
    // they are not suitable for lambda search
    if (src[arcId] != dst[arcId]) {
      arcKeyItem = IDList_push(arcKeyList, sizeof(CycleBasedArcKeyItem_));
      arcKeyItem->ArcId = arcId;
      arcKeyItem->ArcKey = (arcKey + arcId);
    } // end if
    arcKey[arcId] = INFINITY;
  } // end for: arcId
  // #step 1
  // calculate arcKeys and assign nodeKeys
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "Graph's updated costs and arcKeys\n");
#endif
  for (arcId = 0; arcId < arcCount; ++arcId) {
    arcKey[arcId] = CycleBased_FindArcKey(arcId, II, context);
    nodeId = dst[arcId];
    // a nodeKey is an arcId of a node's ingoing arc with highest arcKey
    if (nodeKeyNotNIL[nodeId] == false) {
      nodeKeyNotNIL[nodeId] = true;
      nodeKey[nodeId] = arcId;
    } else {
      if (arcKey[arcId] < arcKey[nodeKey[nodeId]]) {
        nodeKey[nodeId] = arcId;
      } // end if
    } // end if
#ifdef DETAILED_TRACE
    SCD_TRACE && fprintf(SCD_TRACE, "\tarc: %d->%d\t", src[arcId], dst[arcId]);
    SCD_TRACE && fprintf(SCD_TRACE, "upCost: %d\t", -(latency[arcId] - II*distance[arcId]));
    if (arcKey[arcId] == INFINITY) {
      SCD_TRACE && fprintf(SCD_TRACE, "arcKey: Inf.\n");
    } else {
      SCD_TRACE && fprintf(SCD_TRACE, "arcKey: %f\n", arcKey[arcId]);
    } // end if
#endif
  } // end for: arcId
  // Main Loop
  while (true) {
    found = false;
    minArcKey = INFINITY;
#ifdef DETAILED_TRACE
    SCD_TRACE && fprintf(SCD_TRACE, "Nodes d&t data:\n");
    for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d\td:%f\tt:%f\n", nodeId,
                                      d[nodeId], t[nodeId]);
    } // end for: nodeId
#endif
    // #step 2
    // find an arc with minimum arcKey
    minArcKeyItem = NULL;
    IDList_FOREACH(arcKeyList, CycleBasedArcKeyItem_, arcKeyItem) {
      if (*arcKeyItem->ArcKey < minArcKey) {
        minArcKey = *arcKeyItem->ArcKey;
        minArcKeyItem = arcKeyItem;
      }
    } IDList_ENDEACH;
    Except_CHECK(minArcKeyItem != NULL);
    // #step 3
    // if an arcKey of chosen arc is INIFITY then the graph is acyclic and MCR cannot be found
    if (*minArcKeyItem->ArcKey == INFINITY) {
      SCD_TRACE && fprintf(SCD_TRACE, "Found MCR: r=Inf.\n");
      minArcKey == INFINITY;
      found = true;
      break;
    } // end if
    // #step 4
    // otherwise let the destination node of chosen arc be denoted as v
    // it is necessary to find a 'v'-rooted subtree of an 's'-rooted tree
    rootNodeId = dst[minArcKeyItem->ArcId];
#ifdef DETAILED_TRACE
    SCD_TRACE && fprintf(SCD_TRACE, "Arc with minimum arcKey selected: %d->%d\n",
                                    src[minArcKeyItem->ArcId], rootNodeId);
#endif
    // it can be done by a BFS algorithm
    IDList_empty(treeSearchNodeList, NULL);
    IDList_empty(treeNodeList, NULL);
    memset(visitedNodes, 0, nodeCount*sizeof(bool));
    // initialize algorithm by putting start nodeId to the beginning of search list
    *(CycleBasedNodeId *)IDList_put(treeSearchNodeList, 
                                    sizeof(CycleBasedNodeId)) = rootNodeId;
    // BFS loop
    while (!IDList_isEmpty(treeSearchNodeList)) {
      // get a nodeId from the end of search list
      nodeId = *(CycleBasedNodeId *)IDList_lastItem(treeSearchNodeList);
      // mark it if it was not visited
      if (visitedNodes[nodeId] == false) {
        visitedNodes[nodeId] = true;
        // and push it at the end of list
        *(CycleBasedNodeId *)IDList_push(treeNodeList, 
                                         sizeof(CycleBasedNodeId)) = nodeId;
        // but all its children put at the beginnig
        IDList_FOREACH(treeArcList, CycleBasedArcId, pArcId) {
          if (src[*pArcId] == nodeId) {
            *(CycleBasedNodeId *)IDList_put(treeSearchNodeList,
                                            sizeof(CycleBasedNodeId)) = dst[*pArcId];
          }
        } IDList_ENDEACH;
      } // end if: node was not previously visited
      IDList_pop(treeSearchNodeList, NULL);
    } // end while
    // if the source node of an arc with minimum arcKey belongs to a subtree then
    // MCR is equal to this arcKey
    IDList_FOREACH(treeNodeList, CycleBasedNodeId, pNodeId) {
      if (*pNodeId == src[minArcKeyItem->ArcId]) {
        SCD_TRACE && fprintf(SCD_TRACE, "Found MCR: r=%f in %d iteration\n",
                                        *minArcKeyItem->ArcKey, iter);
        found = true;
        minArcKey = *minArcKeyItem->ArcKey;
        break;
      } // end if
    } IDList_ENDEACH;    
    if (found == true) {
#ifdef DETAILED_TRACE
      int cycle = 0;
      nodeId = src[minArcKeyItem->ArcId];
      SCD_TRACE && fprintf(SCD_TRACE, "Minimum Cost Cycle:\n");
      do {
        SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d\n", nodeId);
        for (arcId = 0; arcId < arcCount; ++arcId) {
          if (src[arcId] == parent[nodeId] && dst[arcId] == nodeId) {
            SCD_TRACE && fprintf(SCD_TRACE, "\t\tcost: %d\n",
                                  - (latency[arcId] - II*distance[arcId]));
            cycle = cycle - (latency[arcId] - II*distance[arcId]);
            break;
          } // end if: arc is between node and it's parent
        } // end for: arcId
        nodeId = parent[nodeId];
      } while (nodeId != dst[minArcKeyItem->ArcId]);
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d\n", nodeId);
      arcId = minArcKeyItem->ArcId;
      SCD_TRACE && fprintf(SCD_TRACE, "\t\tcost: %d\n", - (latency[arcId]
                                       - II*distance[arcId]));
      cycle = cycle - (latency[arcId] - II*distance[arcId]);
      SCD_TRACE && fprintf(SCD_TRACE, "\tnodeId: %d ...\n", src[arcId]);
      SCD_TRACE && fprintf(SCD_TRACE, "Cycle cost: %d\n", cycle);
      SCD_TRACE && fprintf(SCD_TRACE, "Lambda: %d\tminArcKey: %f\n", II, 
                                      *minArcKeyItem->ArcKey);
#endif
      break;
    } // end if
    // remove minArcKeyItem
    IDList_remove(arcKeyList, minArcKeyItem, NULL);
    // #step 5
    // update d & t parameters for every node which belongs to subtree
    arcId = minArcKeyItem->ArcId;
    tSrc = t[src[arcId]];
    tDst = t[dst[arcId]];
    dSrc = d[src[arcId]];
    dDst = d[dst[arcId]];
    IDList_FOREACH(treeNodeList, CycleBasedNodeId, pNodeId) {
      t[*pNodeId] = t[*pNodeId] + tSrc + distance[arcId] - tDst;
      d[*pNodeId] = d[*pNodeId] + dSrc - (latency[arcId] - II*distance[arcId]) - dDst;
    } IDList_ENDEACH;
    // #step 6
    // replace predecessor arc of destinantion node (v) by an arc with minimum arcKey in an 's'-rooted tree
    found = false;
    IDList_FOREACH(treeArcList, RCMSArcId, pArcId) {
      if (dst[*pArcId] == rootNodeId) {
        CycleBasedArcKeyItem arcKeyItem;
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "Replacing predecessor arc: %d->%d with: %d->%d\n",
                                  src[*pArcId], dst[*pArcId], src[arcId], dst[arcId]);
#endif
        IDList_remove_(treeArcList, pArcId);
        parent[dst[arcId]] = src[arcId];
        *(CycleBasedArcId *)IDList_push(treeArcList, sizeof(CycleBasedArcId)) = arcId;
        arcKeyItem = IDList_push(arcKeyList, sizeof(CycleBasedArcKeyItem_));
        arcKeyItem->ArcId = *pArcId;
        arcKeyItem->ArcKey = (arcKey + *pArcId);
        found = true;
        break;
      } // end if
    } IDList_ENDEACH;
    if (found == false) {
        CycleBasedArcKeyItem arcKeyItem;
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "Inserting predecessor arc: %d->%d\n",
                                  src[arcId], dst[arcId]);
#endif
        parent[dst[arcId]] = src[arcId];
        *(CycleBasedArcId *)IDList_push(treeArcList, sizeof(CycleBasedArcId)) = arcId;
    } // end if
    // #step 7
    // update arcKeys for all arcs which do not belong to an 's'-rooted tree and nodeKeys for all nodes
    for (arcId = 0; arcId < arcCount; ++arcId) {
      found = false;
      IDList_FOREACH(treeArcList, RCMSArcId, pArcId) {
        if (arcId == *pArcId) {
          found = true;
          break;
        } // end if
      } IDList_ENDEACH;
      if (found == false) {
        arcKey[arcId] = CycleBased_FindArcKey(arcId, II, context);
        nodeId = dst[arcId];
        Except_CHECK(nodeKeyNotNIL[nodeId] == true);
        if (arcKey[arcId] < arcKey[nodeKey[nodeId]]) {
          nodeKey[nodeId] = arcId;
        } // end if
      } // end if: an arc was not found in an 's'-rooted tree
    } // end for: arcId
    // ... begin new iteration
    ++iter;
  } // end while
  return minArcKey;
} // end of CycleBased_MCR

#line 1367 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
int
CycleBased_FindMinII(DDGraph mcrlGraph, int32_t firstIndex, int32_t lastIndex) 
{
  float MCR;
  CycleBasedMCRContext_ mcrContext[1];
  int II = 1;
  CycleBasedArcId arcCounter = 0;
  int32_t nodeCount = 0;
  int32_t arcCount = 0;
  int *latency;
  int *distance;
  CycleBasedNodeId *src;
  CycleBasedNodeId *dst;
  SCD_TRACE && fprintf(SCD_TRACE, "CycleBased_FindMinII()\n");
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          ++arcCount;
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  nodeCount = lastIndex - firstIndex + 1;
  CycleBased_CreateMCRContext(nodeCount, arcCount, mcrContext);
  latency = mcrContext->Latency;
  distance = mcrContext->Distance;
  src = mcrContext->Src;
  dst = mcrContext->Dst;
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          CycleBasedArcId arcId = arcCounter;
          src[arcId] = tail_index - firstIndex;
          dst[arcId] = head_index - firstIndex;
          latency[arcId] = DDGraphArc_THETA(arc);
          distance[arcId] = DDGraphArc_OMEGA(arc);
          SCD_TRACE && fprintf(SCD_TRACE, "arcId: %d\tsrc: %d\tdst: %d",
                                          arcId, src[arcId], dst[arcId]);
          SCD_TRACE && fprintf(SCD_TRACE, "\tlatency: %d\tdistance: %d\n",
                                          latency[arcId], distance[arcId]);
          ++arcCounter;
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  Except_CHECK(arcCounter == arcCount);
  while ((MCR = CycleBased_MCR(II, mcrContext)) < 0.f) {
    if (MCR > -1.f) {
      ++II;
    } else {
      II = II - MCR;
    } // end if
  } // end while
  CycleBased_ReleaseMCRContext(mcrContext);
  return II;
} // end of CycleBased_FindMinII

#line 1435 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1447 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
#if XCC__C
void
CycleBased_TEST(void)
{
#include "SCD/CycleBased_.c"
}

#line 1454 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/CycleBased.xcc"
int
main(int argc, char **argv)
{
  SCD_Test(CycleBased_TEST);
  return 0;
}
#endif


#endif//CycleBased_c_INCLUDED
