#ifndef Formulation__h_INCLUDED
#define Formulation__h_INCLUDED

//
#include "SCD/Formulation.h"

/*
 * Formulation --	Time-indexed formulation for optimal modulo scheduling.
 */
struct Formulation_ {
  //@args	Memory memory, Scheduler scheduler
  Memory MEMORY;		// This Formulation Memory.
  Scheduler SCHEDULER;		// The instruction scheduler.
  //@access LAMBDA2	Scheduler_LAMBDAS(Formulation_SCHEDULER(this))[2]
  //@access LINKARCS	Scheduler_LINKARCS(Formulation_SCHEDULER(this))
  //@access HEADNAME	Scheduler_HEADNAME(Formulation_SCHEDULER(this))
  //@access ISSUETABLE	Scheduler_ISSUETABLE(Formulation_SCHEDULER(this))
  //@access SOLVING	Scheduler_SOLVING(Formulation_SCHEDULER(this))
  //@access TIMEOUT	Scheduler_TIMEOUT(Formulation_SCHEDULER(this))
  GrahamList GRAHAMLIST;	// The GrahamList for this Formulation.
  //@access INDEX2TASK	GrahamList_INDEX2TASK(Formulation_GRAHAMLIST(this))
  DDGraph MCRLGRAPH;		// The reference DDGraph for this Formulation.
  //@access INDEX2NODE	DDGraph_nodes(Formulation_MCRLGRAPH(this))
  int32_t FIRSTINDEX;		// The first GrahamListTask INDEX.
  int32_t LASTINDEX;		// The last GrahamListTask INDEX.
  int16_t *EARLYDATES;		// Array of the DDGraphNode(s) early dates.
  int16_t *LATEDATES;		// Array of the DDGraphNode(s) late dates.
  int16_t **DELAYMATRIX;	// Matrix of longest path delays.
  DDGraphArc **EDGEMATRIX;	// Matrix of longest path edges.
  PtrSeq_ NODESTACK[1];		// DDGraphNode stack for constraint propagation.
  bool *INSTACK;		// Flag the DDGraphNode(s) that are on NODESTACK;
  //@access isDinechin	((Formulation_SOLVING(this) & OptimizeRCMSSolving_Dinechin) != 0)
  //@access isLNSScan1	((Formulation_SOLVING(this) & OptimizeRCMSSolving_LNSScan1) != 0)
  //@access isLNSScan2	((Formulation_SOLVING(this) & OptimizeRCMSSolving_LNSScan2) != 0)
  //@access isEicDavAb	((Formulation_SOLVING(this) & OptimizeRCMSSolving_EicDavAb) != 0)
  //@access isAggrDeps	((Formulation_SOLVING(this) & OptimizeRCMSSolving_AggrDeps) != 0)
  //@access isRational	((Formulation_SOLVING(this) & OptimizeRCMSSolving_Rational) != 0)
  //@access isPostPass	((Formulation_SOLVING(this) & OptimizeRCMSSolving_PostPass) != 0)
  //@access isAcyclic	((Formulation_SOLVING(this) & OptimizeRCMSSolving_Acyclic) != 0)
  int8_t SCANNING;		// Max number of scanning steps in Scanning heuristic.
  LProgram PROGRAM;		// The LProgram corresponding to this Formulation.
  //@access STATUS	LProgram_STATUS(Formulation_PROGRAM(this))
  //@access VARIABLECOUNT	LProgram_VARIABLECOUNT(Formulation_PROGRAM(this))  
  //@access CONSTRAINTCOUNT	LProgram_CONSTRAINTCOUNT(Formulation_PROGRAM(this))  
  //@access SOLUTION	LProgram_solution(Formulation_PROGRAM(this))  
  int16_t LAMBDA;		// The current LAMBDA.
  int16_t HORIZON;		// The current HORIZON.
  unsigned EXITWEIGHT;		// Weight of exit operations in the schedule.
  double TIMER;			// Timer that accumulates time spent on failure.
};
typedef struct Formulation_ Formulation_, *Formulation;
typedef const struct Formulation_ *const_Formulation;
typedef struct Formulation_ * restrict_Formulation;

extern Formulation
Formulation_Ctor(Formulation this, Memory memory, Scheduler scheduler);

extern Formulation
Formulation_Copy(Formulation this, const_Formulation that);

extern void
Formulation_Dtor(Formulation this);

extern size_t
Formulation_Size(Memory memory, Scheduler scheduler);

#define Formulation_MEMORY(this) ((this)->MEMORY)
#define Formulation__MEMORY(this) (&(this)->MEMORY)
#define Formulation_SCHEDULER(this) ((this)->SCHEDULER)
#define Formulation__SCHEDULER(this) (&(this)->SCHEDULER)
#define Formulation_GRAHAMLIST(this) ((this)->GRAHAMLIST)
#define Formulation__GRAHAMLIST(this) (&(this)->GRAHAMLIST)
#define Formulation_MCRLGRAPH(this) ((this)->MCRLGRAPH)
#define Formulation__MCRLGRAPH(this) (&(this)->MCRLGRAPH)
#define Formulation_FIRSTINDEX(this) ((this)->FIRSTINDEX)
#define Formulation__FIRSTINDEX(this) (&(this)->FIRSTINDEX)
#define Formulation_LASTINDEX(this) ((this)->LASTINDEX)
#define Formulation__LASTINDEX(this) (&(this)->LASTINDEX)
#define Formulation_EARLYDATES(this) ((this)->EARLYDATES)
#define Formulation__EARLYDATES(this) (&(this)->EARLYDATES)
#define Formulation_LATEDATES(this) ((this)->LATEDATES)
#define Formulation__LATEDATES(this) (&(this)->LATEDATES)
#define Formulation_DELAYMATRIX(this) ((this)->DELAYMATRIX)
#define Formulation__DELAYMATRIX(this) (&(this)->DELAYMATRIX)
#define Formulation_EDGEMATRIX(this) ((this)->EDGEMATRIX)
#define Formulation__EDGEMATRIX(this) (&(this)->EDGEMATRIX)
#define Formulation_NODESTACK(this) ((this)->NODESTACK)
#define Formulation__NODESTACK(this) (&(this)->NODESTACK)
#define Formulation_INSTACK(this) ((this)->INSTACK)
#define Formulation__INSTACK(this) (&(this)->INSTACK)
#define Formulation_SCANNING(this) ((this)->SCANNING)
#define Formulation__SCANNING(this) (&(this)->SCANNING)
#define Formulation_PROGRAM(this) ((this)->PROGRAM)
#define Formulation__PROGRAM(this) (&(this)->PROGRAM)
#define Formulation_LAMBDA(this) ((this)->LAMBDA)
#define Formulation__LAMBDA(this) (&(this)->LAMBDA)
#define Formulation_HORIZON(this) ((this)->HORIZON)
#define Formulation__HORIZON(this) (&(this)->HORIZON)
#define Formulation_EXITWEIGHT(this) ((this)->EXITWEIGHT)
#define Formulation__EXITWEIGHT(this) (&(this)->EXITWEIGHT)
#define Formulation_TIMER(this) ((this)->TIMER)
#define Formulation__TIMER(this) (&(this)->TIMER)
#define Formulation_SOLVING(this) (Scheduler_SOLVING(Formulation_SCHEDULER(this)))
#define Formulation_STATUS(this) (LProgram_STATUS(Formulation_PROGRAM(this)))
#define Formulation_LAMBDA2(this) (Scheduler_LAMBDAS(Formulation_SCHEDULER(this))[2])
#define Formulation_isAcyclic(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_Acyclic) != 0))
#define Formulation_isRational(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_Rational) != 0))
#define Formulation_INDEX2TASK(this) (GrahamList_INDEX2TASK(Formulation_GRAHAMLIST(this)))
#define Formulation_isAggrDeps(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_AggrDeps) != 0))
#define Formulation_isLNSScan2(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_LNSScan2) != 0))
#define Formulation_CONSTRAINTCOUNT(this) (LProgram_CONSTRAINTCOUNT(Formulation_PROGRAM(this))  )
#define Formulation_ISSUETABLE(this) (Scheduler_ISSUETABLE(Formulation_SCHEDULER(this)))
#define Formulation_VARIABLECOUNT(this) (LProgram_VARIABLECOUNT(Formulation_PROGRAM(this))  )
#define Formulation_isEicDavAb(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_EicDavAb) != 0))
#define Formulation_SOLUTION(this) (LProgram_solution(Formulation_PROGRAM(this))  )
#define Formulation_isPostPass(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_PostPass) != 0))
#define Formulation_INDEX2NODE(this) (DDGraph_nodes(Formulation_MCRLGRAPH(this)))
#define Formulation_isLNSScan1(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_LNSScan1) != 0))
#define Formulation_TIMEOUT(this) (Scheduler_TIMEOUT(Formulation_SCHEDULER(this)))
#define Formulation_LINKARCS(this) (Scheduler_LINKARCS(Formulation_SCHEDULER(this)))
#define Formulation_isDinechin(this) (((Formulation_SOLVING(this) & OptimizeRCMSSolving_Dinechin) != 0))
#define Formulation_HEADNAME(this) (Scheduler_HEADNAME(Formulation_SCHEDULER(this)))


//
Formulation
Formulation_new(Memory parent, Scheduler scheduler);

//
Formulation
Formulation_delete(Formulation this);

/*
 * Formulation_firstIndex --	For use by Formulation_FOREACH_DDGraphNode.
 */
static inline int32_t
Formulation_firstIndex(Formulation this)
{
  return Formulation_FIRSTINDEX(this);
}

/*
 * Formulation_lastIndex --	For use by Formulation_FOREACH_DDGraphNode.
 */
static inline int32_t
Formulation_lastIndex(Formulation this)
{
  return Formulation_LASTINDEX(this);
}

/*
 * Formulation_index2node --	For use by Formulation_FOREACH_DDGraphNode.
 */
static inline DDGraphNode *
Formulation_index2node(Formulation this)
{
  return Formulation_INDEX2NODE(this);
}

/*
 * Formulation_FOREACH_DDGraphNode --	Iterate over this Formulation DDGraphNode(s).
 */
#define Formulation_FOREACH_DDGraphNode(this, node) { \
  DDGraphNode *Formulation_INDEX2NODE = Formulation_index2node(this); \
  int32_t Formulation_FIRSTINDEX = Formulation_firstIndex(this); \
  int32_t Formulation_LASTINDEX = Formulation_lastIndex(this); \
  int32_t Formulation_NODEINDEX = Formulation_FIRSTINDEX; \
  for (; Formulation_NODEINDEX <= Formulation_LASTINDEX; Formulation_NODEINDEX++) { \
    DDGraphNode node = Formulation_INDEX2NODE[Formulation_NODEINDEX];
#define Formulation_ENDEACH_DDGraphNode \
  } \
}

/*
 * Formulation_index2task --	For use by Formulation_FOREACH_GrahamListTask.
 */
static inline struct GrahamListTask_ **
Formulation_index2task(Formulation this)
{
  return Formulation_INDEX2TASK(this);
}

/*
 * Formulation_FOREACH_GrahamListTask --	Iterate over this Formulation GrahamListTask(s).
 */
#define Formulation_FOREACH_GrahamListTask(this, task) { \
  GrahamListTask_ **Formulation_INDEX2TASK = Formulation_index2task(this); \
  int32_t Formulation_FIRSTINDEX = Formulation_firstIndex(this); \
  int32_t Formulation_LASTINDEX = Formulation_lastIndex(this); \
  int32_t Formulation_TASKINDEX = Formulation_FIRSTINDEX; \
  for (; Formulation_TASKINDEX <= Formulation_LASTINDEX; Formulation_TASKINDEX++) { \
    GrahamListTask task = Formulation_INDEX2TASK[Formulation_TASKINDEX];
#define Formulation_ENDEACH_GrahamListTask \
  } \
}

/*
 * Formulation_LINKARCS_FOREACH_DDGraphArc --	Iterate over this Formulation LINKARCS.
 */
#define Formulation_LINKARCS_FOREACH_DDGraphArc(this, regFile, arc) \
  Scheduler_LINKARCS_FOREACH_DDGraphArc(Formulation_SCHEDULER(this), regFile, arc)
#define Formulation_LINKARCS_ENDEACH_DDGraphArc \
  Scheduler_LINKARCS_ENDEACH_DDGraphArc;

//
const char *
Formulation_saveName(Formulation this, const char *name);

//
#define Formulation_KEY1(regFile)	(void *)(regFile - RegFile__)
#define Formulation_KEY2(index, date)	(void *)(((index)<<16) | ((date)&0xffff))

//
bool
Formulation_pretty(Formulation this, FILE *file);

/*
 * Formulation_Schedule --	Schedule by solving a Formulation.
 *
 * We use a dichotomy search to find the minimum delta such that the relaxation
 * succeeds a initiation interval lambda + delta.
 *
 * Require:	The problem is feasible at lambda = lambda_max.
 * Return:	A feasible lambda if found else INT_MAX.
 */
int
Formulation_Schedule(Scheduler scheduler, int lambda_min, int lambda_max, int length);

/*
 * Formulation_trySchedule --	Try to schedule with Formulation at the given lambda.
 */
bool
Formulation_trySchedule(Formulation this, int lambda, int scanStep, bool feasibility);

/*
 * Formulation_buildProgram --	Build the LProgram for this Formulation.
 */
LProgram
Formulation_buildProgram(Formulation this, int lambda, int scanStep, bool feasibility);

/*
 * Formulation_makeAllPairsLongestPaths --	All-pairs longest-paths.
 */
void
Formulation_makeAllPairsLongestPaths(Formulation this);

/*
 * Formulation_adjustEarlyLateDates --	Adjust EARLYDATES, LATEDATES, HORIZON.
 */
void
Formulation_adjustEarlyLateDates(Formulation this, int scanStep, bool feasibility);

/*
 * Formulation_commitSchedule --	Commit the schedule of this Formulation LProgram.
 */
int
Formulation_commitSchedule(Formulation this);

//
void
Formulation_verifyDependences(Formulation this);

/*
 * Formulation_buildVariablesDinechin --	Build this Formulation ProgramVariables.
 */
void
Formulation_buildVariablesDinechin(Formulation this);

/*
 * Formulation_buildObjectiveDinechin --	
 *
 * We list all the problem variables, with a zero coefficient when not used in
 * the objective function. This is because CPLEX orders its variables by the
 * occurrence of first reference, and we want CPLEX to keep our order when
 * displaying solutions.
 */
void
Formulation_buildObjectiveDinechin(Formulation this);

/*
 * Formulation_buildAssignmentsDinechin --	
 */
void
Formulation_buildAssignmentsDinechin(Formulation this);

/*
 * Formulation_buildDependencesDinechin --	
 */
void
Formulation_buildDependencesDinechin(Formulation this);

/*
 * Formulation_buildResourcesDinechin --	
 *
 * \sum_{i=1}^{n} \sum{k=0}^{\div{D_i-p_i}{\lambda}} x_i^{t+k\lambda} b_i^r \le B_i^r
 *
 * The \sum{s=t+k\lambda-p_i+1}^{t+k\lambda} x_i^s b_i^r has been replaced by
 * x_i^{t+k\lambda} b_i^r because p_i is assumed to be 1 (UET operations).
 */
void
Formulation_buildResourcesDinechin(Formulation this);

/*
 * Formulation_buildBoundsDinechin --	
 */
void
Formulation_buildBoundsDinechin(Formulation this);

/*
 * Formulation_buildVariablesEicDavAb --	Build this Formulation ProgramVariables.
 */
void
Formulation_buildVariablesEicDavAb(Formulation this);

/*
 * Formulation_buildObjectiveEicDavAb --	
 *
 * We list all the problem variables, with a zero coefficient when not used in
 * the objective function. This is because CPLEX orders its variables by the
 * occurrence of first reference, and we want CPLEX to keep our order when
 * displaying solutions.
 */
void
Formulation_buildObjectiveEicDavAb(Formulation this);

/*
 * Formulation_buildAssignmentsEicDavAb --	
 */
void
Formulation_buildAssignmentsEicDavAb(Formulation this);

/*
 * Formulation_buildDependencesEicDavAb --	
 */
void
Formulation_buildDependencesEicDavAb(Formulation this);

/*
 * Formulation_buildResourcesEicDavAb --	\sum_{i=1}^{n}
 *   \sum{s=0}^{p_i-1} y_i^{\mod{t-s}{\lambda}} b_i^r \le B_i^r \forall t, \forall r
 *
 * The \sum{s=0}^{p_i-1} y_i^{\mod{t-s}{\lambda}} b_i^r has been replaced by
 * y_i^{t} b_i^r because p_i is assumed to be 1 (UET operations).
 */
void
Formulation_buildResourcesEicDavAb(Formulation this);

/*
 * Formulation_buildBoundsEicDavAb --	y_i^t \in \{ 0, 1 \} \forall i, \forall t
 */
void
Formulation_buildBoundsEicDavAb(Formulation this);

/*
 * Formulation_INIT --	Initialize Formulation.
 */
#define Formulation_INIT()

/*
 * Formulation_FINI --	Finalize Formulation.
 */
#define Formulation_FINI()


#endif//Formulation__h_INCLUDED
