#ifndef GrahamList__h_INCLUDED
#define GrahamList__h_INCLUDED

//
#include "SCD/GrahamList.h"

/*
 * GrahamListDates --	GrahamList early date and late date.
 */
struct GrahamListDates_ {
  //@args	
  int16_t EARLYDATE;		// The early schedule date as computed by Relaxation.
  int16_t LATEDATE;		// The late schedule date as computed by Relaxation.
};
typedef struct GrahamListDates_ GrahamListDates_, *GrahamListDates;
typedef const struct GrahamListDates_ *const_GrahamListDates;
typedef struct GrahamListDates_ * restrict_GrahamListDates;

extern GrahamListDates
GrahamListDates_Ctor(GrahamListDates this);

extern GrahamListDates
GrahamListDates_Copy(GrahamListDates this, const_GrahamListDates that);

extern void
GrahamListDates_Dtor(GrahamListDates this);

extern size_t
GrahamListDates_Size(void);

#define GrahamListDates_EARLYDATE(this) ((this)->EARLYDATE)
#define GrahamListDates__EARLYDATE(this) (&(this)->EARLYDATE)
#define GrahamListDates_LATEDATE(this) ((this)->LATEDATE)
#define GrahamListDates__LATEDATE(this) (&(this)->LATEDATE)


/*
 * GrahamListTask --	GrahamList task.
 */
struct GrahamListTask_ {
  //@args	IssueItem issueItem, int unwinding
  IssueItem ITEM;		// The IssueItem of this GrahamListTask.
  //@access INDEX	IssueItem_INDEX(GrahamListTask_ITEM(this))
  //@access OPERATION	IssueItem_OPERATION(GrahamListTask_ITEM(this))
  //@access ORDERING	IssueItem_ORDERING(GrahamListTask_ITEM(this))
  int16_t PRIORITY1;		// First priority value for this task.
  int16_t PRIORITY2;		// Second priority value for this task.
  int16_t TAU;			// DATE MOD LAMBDA.
  int16_t PHI;			// DATE DIV LAMBDA.
  GrahamListDates_ DATES[1];	// GrahamListDates for each unwinded iteration.
  //@access EARLYDATE	GrahamListDates_EARLYDATE(GrahamListTask_DATES(this))
  //@access LATEDATE	GrahamListDates_LATEDATE(GrahamListTask_DATES(this))
};
typedef struct GrahamListTask_ GrahamListTask_, *GrahamListTask;
typedef const struct GrahamListTask_ *const_GrahamListTask;
typedef struct GrahamListTask_ * restrict_GrahamListTask;

extern GrahamListTask
GrahamListTask_Ctor(GrahamListTask this, IssueItem issueItem, int unwinding);

extern GrahamListTask
GrahamListTask_Copy(GrahamListTask this, const_GrahamListTask that);

extern void
GrahamListTask_Dtor(GrahamListTask this);

extern size_t
GrahamListTask_Size(IssueItem issueItem, int unwinding);

#define GrahamListTask_ITEM(this) ((this)->ITEM)
#define GrahamListTask__ITEM(this) (&(this)->ITEM)
#define GrahamListTask_PRIORITY1(this) ((this)->PRIORITY1)
#define GrahamListTask__PRIORITY1(this) (&(this)->PRIORITY1)
#define GrahamListTask_PRIORITY2(this) ((this)->PRIORITY2)
#define GrahamListTask__PRIORITY2(this) (&(this)->PRIORITY2)
#define GrahamListTask_TAU(this) ((this)->TAU)
#define GrahamListTask__TAU(this) (&(this)->TAU)
#define GrahamListTask_PHI(this) ((this)->PHI)
#define GrahamListTask__PHI(this) (&(this)->PHI)
#define GrahamListTask_DATES(this) ((this)->DATES)
#define GrahamListTask__DATES(this) (&(this)->DATES)
#define GrahamListTask_LATEDATE(this) (GrahamListDates_LATEDATE(GrahamListTask_DATES(this)))
#define GrahamListTask_INDEX(this) (IssueItem_INDEX(GrahamListTask_ITEM(this)))
#define GrahamListTask_OPERATION(this) (IssueItem_OPERATION(GrahamListTask_ITEM(this)))
#define GrahamListTask_EARLYDATE(this) (GrahamListDates_EARLYDATE(GrahamListTask_DATES(this)))
#define GrahamListTask_ORDERING(this) (IssueItem_ORDERING(GrahamListTask_ITEM(this)))


/*
 * GrahamListTask_compare --	Compare function for use by Insertion_selectTask.
 */
int
GrahamListTask_compare(const void *this_pointer, const void *that_pointer);

//
struct GrahamList_ {
  //@args	Memory memory, DDGraph ddgraph, const uint8_t *availability
  DDGraph DDGRAPH;		// The dependence graph.
  //@access LAMBDA	DDGraph_LAMBDA(GrahamList_DDGRAPH(this))
  //@mutate LAMBDA	DDGraph__LAMBDA(GrahamList_DDGRAPH(this))
  IssueTable ISSUETABLE;	// This GrahamList IssueTable.
  //@access AVAILABILITY	IssueTable_availability(GrahamList_ISSUETABLE(this))
  IStack_ TASKS_;		// All the GrahamListTask(s).
    //@access TASKS	GrahamList__TASKS_(this)
  //@access INDEX2TASK	((GrahamListTask *)IStack_base(GrahamList_TASKS(this)) - 1)
  //@access FIRSTTASK	(*(GrahamListTask *)IStack_firstItem(GrahamList_TASKS(this)))
  //@access LASTTASK	(*(GrahamListTask *)IStack_lastItem(GrahamList_TASKS(this)))
  //@access MEMORY	IStack_memory(GrahamList_TASKS(this))
  //@access COUNT	(IStack_usedSize(GrahamList_TASKS(this))/sizeof(GrahamListTask))
  IDList_ TOISSUE_;		// The GrahamListTask(s) to issue.
  //@access TOISSUE	GrahamList__TOISSUE_(this)
  IDList_ NOISSUE_;		// The GrahamListTask(s) that dont'need issue.
  //@access NOISSUE	GrahamList__NOISSUE_(this)
  IDList_ ISSUED_;		// The GrahamListTask(s) issued.
  //@access ISSUED	GrahamList__ISSUED_(this)
};

extern GrahamList
GrahamList_Ctor(GrahamList this, Memory memory, DDGraph ddgraph, const uint8_t *availability);

extern GrahamList
GrahamList_Copy(GrahamList this, const_GrahamList that);

extern void
GrahamList_Dtor(GrahamList this);

extern size_t
GrahamList_Size(Memory memory, DDGraph ddgraph, const uint8_t *availability);

#define GrahamList_DDGRAPH(this) ((this)->DDGRAPH)
#define GrahamList__DDGRAPH(this) (&(this)->DDGRAPH)
#define GrahamList_ISSUETABLE(this) ((this)->ISSUETABLE)
#define GrahamList__ISSUETABLE(this) (&(this)->ISSUETABLE)
#define GrahamList_TASKS_(this) ((this)->TASKS_)
#define GrahamList__TASKS_(this) (&(this)->TASKS_)
#define GrahamList_TOISSUE_(this) ((this)->TOISSUE_)
#define GrahamList__TOISSUE_(this) (&(this)->TOISSUE_)
#define GrahamList_NOISSUE_(this) ((this)->NOISSUE_)
#define GrahamList__NOISSUE_(this) (&(this)->NOISSUE_)
#define GrahamList_ISSUED_(this) ((this)->ISSUED_)
#define GrahamList__ISSUED_(this) (&(this)->ISSUED_)
#define GrahamList_FIRSTTASK(this) ((*(GrahamListTask *)IStack_firstItem(GrahamList_TASKS(this))))
#define GrahamList_LAMBDA(this) (DDGraph_LAMBDA(GrahamList_DDGRAPH(this)))
#define GrahamList_NOISSUE(this) (GrahamList__NOISSUE_(this))
#define GrahamList_ISSUED(this) (GrahamList__ISSUED_(this))
#define GrahamList_INDEX2TASK(this) (((GrahamListTask *)IStack_base(GrahamList_TASKS(this)) - 1))
#define GrahamList_TASKS(this) (GrahamList__TASKS_(this))
#define GrahamList_MEMORY(this) (IStack_memory(GrahamList_TASKS(this)))
#define GrahamList_COUNT(this) ((IStack_usedSize(GrahamList_TASKS(this))/sizeof(GrahamListTask)))
#define GrahamList_AVAILABILITY(this) (IssueTable_availability(GrahamList_ISSUETABLE(this)))
#define GrahamList_LASTTASK(this) ((*(GrahamListTask *)IStack_lastItem(GrahamList_TASKS(this))))
#define GrahamList_TOISSUE(this) (GrahamList__TOISSUE_(this))
#define GrahamList__LAMBDA(this) (DDGraph__LAMBDA(GrahamList_DDGRAPH(this)))


/*
 * GrahamList_tasks --	For use by GrahamList_FOREACH_GrahamListTask.
 */
static inline IStack
GrahamList_tasks(GrahamList this)
{
  return GrahamList_TASKS(this);
}

/*
 * GrahamList_FOREACH_GrahamListTask --	Iterate over this GrahamListTask(s) in topological sort.
 */
#define GrahamList_FOREACH_GrahamListTask(this, task) { \
  IStack GrahamList_TASKS = GrahamList_tasks(this); \
  IStack_FOREACH(GrahamList_TASKS, GrahamListTask, GrahamList__TASK) { \
    GrahamListTask task = *GrahamList__TASK;
#define GrahamList_ENDEACH_GrahamListTask \
  } IStack_ENDEACH \
}

/*
 * GrahamList_FORBACK_GrahamListTask --	Iterate over this GrahamListTask(s) in topological sort.
 */
#define GrahamList_FORBACK_GrahamListTask(this, task) { \
  IStack GrahamList_TASKS = GrahamList_tasks(this); \
  IStack_FORBACK(GrahamList_TASKS, GrahamListTask, GrahamList__TASK) { \
    GrahamListTask task = *GrahamList__TASK;
#define GrahamList_ENDBACK_GrahamListTask \
  } IStack_ENDBACK \
}

/*
 * GrahamList_toIssue --	Needed by GrahamList_FOREACH_TOISSUE_GrahamListTask.
 */
static inline IDList
GrahamList_toIssue(GrahamList this)
{
  return GrahamList_TOISSUE(this);
}

/*
 * GrahamList_FOREACH_TOISSUE_GrahamListTask --	Iterate over this TOISSUE GrahamListTask(s).
 */
#define GrahamList_FOREACH_TOISSUE_GrahamListTask(this, task) \
  IDList_FOREACH(GrahamList_toIssue(this), GrahamListTask_ , task)
#define GrahamList_ENDEACH_TOISSUE_GrahamListTask \
  IDList_ENDEACH;

/*
 * GrahamList_noIssue --	Needed by GrahamList_FOREACH_NOISSUE_GrahamListTask.
 */
static inline IDList
GrahamList_noIssue(GrahamList this)
{
  return GrahamList_NOISSUE(this);
}

/*
 * GrahamList_FOREACH_NOISSUE_GrahamListTask --	Iterate over this NOISSUE GrahamListTask(s).
 */
#define GrahamList_FOREACH_NOISSUE_GrahamListTask(this, task) \
  IDList_FOREACH(GrahamList_noIssue(this), GrahamListTask_ , task)
#define GrahamList_ENDEACH_NOISSUE_GrahamListTask \
  IDList_ENDEACH;

/*
 * GrahamList_issued --	Needed by GrahamList_FOREACH_ISSUED_GrahamListTask.
 */
static inline IDList
GrahamList_issued(GrahamList this)
{
  return GrahamList_ISSUED(this);
}

/*
 * GrahamList_FOREACH_ISSUED_GrahamListTask --	Iterate over this ISSUED GrahamListTask(s).
 */
#define GrahamList_FOREACH_ISSUED_GrahamListTask(this, task) \
  IDList_FOREACH(GrahamList_issued(this), GrahamListTask_ , task)
#define GrahamList_ENDEACH_ISSUED_GrahamListTask \
  IDList_ENDEACH;

/*
 * GrahamList_enterIssueItem --	Enter an IssueItem in this GrahamList.
 */
void
GrahamList_enterIssueItem(GrahamList this, IssueItem issueItem, int unwinding);

/*
 * GrahamList_scheduleNoIssue --	Schedule the NOISSUE tasks.
 */
void
GrahamList_scheduleNoIssue(GrahamList this);

/*
 * GrahamList_INIT --	Initialize the GrahamList module.
 */
void
GrahamList_INIT(void);

/*
 * GrahamList_FINI --	Finalize the GrahamList module.
 */
void
GrahamList_FINI(void);


#endif//GrahamList__h_INCLUDED
