#ifndef Insertion_c_INCLUDED
#define Insertion_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
#include "SCD/SCD_.h"

#line 21 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 38 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
InsertionSlot
InsertionSlot_Ctor(InsertionSlot this,
                   const uint8_t *availability, int tau, int phi)
{
  IssueGroup_Ctor(InsertionSlot_GROUP_M(this), availability);
  IssueGroup_Ctor(InsertionSlot_GROUP_P(this), availability);
  *InsertionSlot__TAU(this) = tau;
  *InsertionSlot__PHI(this) = phi;
  *InsertionSlot__DELTA_M(this) = 0;
  *InsertionSlot__DELTA_P(this) = 0;
  return this;
}

#line 56 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 73 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
Insertion
Insertion_Ctor(Insertion this,
               Memory memory, Scheduler scheduler, int maxRange)
{
  *Insertion__MEMORY(this) = memory;
  *Insertion__SCHEDULER(this) = scheduler;
  *Insertion__MAXRANGE(this) = maxRange;
  *Insertion__SLOTCOUNT(this) = 0;
  memset(Insertion_SLOTS(this), 0, maxRange*sizeof(InsertionSlot_));
  return this;
}

#line 85 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
void
Insertion_Dtor(Insertion this)
{
}

#line 90 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
size_t
Insertion_Size(Memory memory, Scheduler scheduler, int maxRange)
{
  return (sizeof(Insertion_) + (maxRange - 2)*sizeof(InsertionSlot_));
}

#line 105 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
#ifndef Insertion_POOLED
#define Insertion_POOLED 128
#endif//Insertion_POOLED

#line 109 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
Insertion
Insertion_new(Memory parent, Scheduler scheduler, int maxRange)
{
  Memory memory = Memory_new(parent, Insertion_POOLED);
  size_t size = Insertion_Size(memory, scheduler, maxRange);
  Insertion this = Memory_alloc(memory, size);
  return Insertion_Ctor(this, memory, scheduler, maxRange);
}

#line 122 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
Insertion
Insertion_delete(Insertion this)
{
  if (this != NULL) {
    Memory memory = Insertion_MEMORY(this);
#ifndef PRO_NDTOR
    Insertion_Dtor(this);
    Memory_free_(memory, this);
#endif//PRO_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 142 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
GrahamListTask
Insertion_selectTask(Insertion this)
{
  Scheduler scheduler = Insertion_SCHEDULER(this);
  GrahamList grahamList = Scheduler_GRAHAMLIST(scheduler);
  int16_t lambda = Scheduler_LAMBDA(scheduler);
  bool isPrePass = Scheduler_isPrePass(scheduler);
  GrahamList_FOREACH_TOISSUE_GrahamListTask(grahamList, TASK) {
    IssueItem ITEM = GrahamListTask_ITEM(TASK);
    DDGraphNode mcrl_node = IssueItem_MCRLNODE(ITEM);
    DDGraphNode early_node = IssueItem_EARLYNODE(ITEM);
    DDGraphNode late_node = IssueItem_LATENODE(ITEM);
    Operation operation = IssueItem_OPERATION(ITEM);
    bool isInductive = Operation_isInductive(operation);
    int mcrlDate = DDGraphNode_DATE(mcrl_node);
    int earlyDate = DDGraphNode_DATE(early_node);
    int lateDate = DDGraph_EXTENT - DDGraphNode_DATE(late_node);
    int slackCycles = lateDate - earlyDate;
    if (isPrePass) {
      // Previous best PRIORITY1(TASK) = mcrlDate;
      // Absolute best PRIORITY1(TASK) = mcrlDate + slackCycles/lambda;
      *GrahamListTask__PRIORITY1(TASK) = isInductive? lateDate + slackCycles/lambda:
                                                      mcrlDate + slackCycles/lambda;
      *GrahamListTask__PRIORITY2(TASK) = GrahamListTask_LATEDATE(TASK) + slackCycles;
    } else {
      *GrahamListTask__PRIORITY1(TASK) = earlyDate + (slackCycles + 1)/2;
      *GrahamListTask__PRIORITY2(TASK) = GrahamListTask_LATEDATE(TASK);
    }
  } GrahamList_ENDEACH_TOISSUE_GrahamListTask;
  IDList_sort(GrahamList_TOISSUE(grahamList), GrahamListTask_compare);
  return (GrahamListTask)IDList_firstItem(GrahamList_TOISSUE(grahamList));
}

#line 175 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
InsertionSlot
Insertion_makeBestSlot(Insertion this, IssueItem item)
{
  int slotCount = Insertion_SLOTCOUNT(this);
  InsertionSlot slots = Insertion_SLOTS(this);
  Scheduler scheduler = Insertion_SCHEDULER(this);
  IssueTable issueTable = Scheduler_ISSUETABLE(scheduler);
  GrahamList grahamList = Scheduler_GRAHAMLIST(scheduler);
  GrahamListTask *index2task = GrahamList_INDEX2TASK(grahamList);
  int delta = INT_MAX, bestDelta = INT_MAX, bestIndex = 0, i;
  SCD_TRACE && fprintf(SCD_TRACE, "deltas:");
  for (i = 0; i < slotCount; i++) {
    InsertionSlot slot = slots + i;
    int16_t tau = InsertionSlot_TAU(slot);
    IssueGroup group = IssueTable_access(issueTable, tau);
    if (IssueGroup_accept(group, item)) {
      *InsertionSlot__DELTA_M(slot) = 0;
      *InsertionSlot__DELTA_P(slot) = 0;
    } else {
      // Split the group according to the insertion order.
      IssueGroup_FOREACH_IssueItem(group, ITEM) {
        GrahamListTask TASK = index2task[IssueItem_INDEX(ITEM)];
        Except_CHECK(GrahamListTask_ITEM(TASK) == ITEM);
        if (   GrahamListTask_PHI(TASK) > InsertionSlot_PHI(slot)
            || (   GrahamListTask_PHI(TASK) == InsertionSlot_PHI(slot)
                && IssueItem_ORDERING(ITEM) < IssueItem_ORDERING(item))) {
          IssueGroup_insert(InsertionSlot_GROUP_M(slot), ITEM);
        } else {
          IssueGroup_insert(InsertionSlot_GROUP_P(slot), ITEM);
        }
      } IssueGroup_ENDEACH_IssueItem;
      if (IssueGroup_accept(InsertionSlot_GROUP_M(slot), item)) {
        *InsertionSlot__DELTA_M(slot) = 0;
        *InsertionSlot__DELTA_P(slot) = 1;
      } else if (IssueGroup_accept(InsertionSlot_GROUP_P(slot), item)) {
        *InsertionSlot__DELTA_M(slot) = 1;
        *InsertionSlot__DELTA_P(slot) = 0;
      } else {
        *InsertionSlot__DELTA_M(slot) = 1;
        *InsertionSlot__DELTA_P(slot) = 1;
      }
    }
    delta = InsertionSlot_DELTA_M(slot) + InsertionSlot_DELTA_P(slot);
    SCD_TRACE && fprintf(SCD_TRACE, "%2d", delta);
    if (delta < bestDelta) {
      bestDelta = delta, bestIndex = i;
      if (bestDelta == 0) break;
    }
  }
  Except_CHECK(slotCount > 0 && bestIndex < slotCount);
  SCD_TRACE && fprintf(SCD_TRACE, "\n");
  return slots + bestIndex;
}

#line 235 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
InsertionSlot
Insertion_selectSlot(Insertion this, GrahamListTask task)
{
  Scheduler scheduler = Insertion_SCHEDULER(this);
  IssueTable issueTable = Scheduler_ISSUETABLE(scheduler);
  const uint8_t *availability = IssueTable_availability(issueTable);
  InsertionSlot slots = Insertion_SLOTS(this);
  IssueItem item = GrahamListTask_ITEM(task);
  int16_t lambda = Scheduler_LAMBDA(scheduler);
  int16_t slotCount = 1, tau1, tau2, phi1, phi2;
  DDGraphNode mcrl_node = IssueItem_MCRLNODE(item);
  DDGraphNode early_node = IssueItem_EARLYNODE(item);
  DDGraphNode late_node = IssueItem_LATENODE(item);
  int mcrlDate = DDGraphNode_DATE(mcrl_node);
  int earlyDate = DDGraphNode_DATE(early_node);
  int lateDate = DDGraph_EXTENT - DDGraphNode_DATE(late_node);
  int16_t tau = DDGraphNode_DATE(mcrl_node)%lambda;
  int16_t phi = DDGraphNode_DATE(mcrl_node)/lambda;
  int range0 = MIN(Insertion_MAXRANGE(this) - 1, lambda - 1), i;
  int range1 = MIN(lateDate - mcrlDate, range0);
  int range2 = MIN(mcrlDate - earlyDate, range0 - range1);
  while (tau >= lambda) tau -= lambda, phi += 1;
  while (tau < 0) tau += lambda, phi -= 1;
  InsertionSlot_Ctor(slots, availability, tau, phi);
  tau1 = tau + 1, tau2 = tau - 1, phi1 = phi, phi2 = phi;
  for (i = 0; i < range1; i++, tau1++, slotCount++) {
    if (tau1 >= lambda) tau1 -= lambda, phi1 += 1;
    InsertionSlot_Ctor(slots + slotCount, availability, tau1, phi1);
  }
  for (i = 0; i < range2; i++, tau2--, slotCount++) {
    if (tau2 < 0) tau2 += lambda, phi2 -= 1;
    InsertionSlot_Ctor(slots + slotCount, availability, tau2, phi2);
  }
  if (slotCount == 0) {
    InsertionSlot_Ctor(slots + slotCount, availability, tau, phi);
    slotCount++;
  }
  Except_CHECK(slotCount <= Insertion_MAXRANGE(this));
  *Insertion__SLOTCOUNT(this) = slotCount;
  SCD_TRACE && fprintf(SCD_TRACE, "    %s_%d\t",
      Scheduling_name(IssueItem_SCHEDULING(item)), IssueItem_INDEX(item));
  SCD_TRACE && fprintf(SCD_TRACE, "phi=%d\ttau=%d\tdates:%3d+[%d,%d] \t",
      phi, tau, mcrlDate, earlyDate - mcrlDate, lateDate - mcrlDate);
  return Insertion_makeBestSlot(this, item);
}

#line 287 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
int
Insertion_scheduleTaskAtSlot(Insertion this, GrahamListTask task, InsertionSlot slot)
{
  IssueItem item = GrahamListTask_ITEM(task);
  Scheduler scheduler = Insertion_SCHEDULER(this);
  GrahamList grahamList = Scheduler_GRAHAMLIST(scheduler);
  IssueTable issueTable = Scheduler_ISSUETABLE(scheduler);
  int16_t delta_m = InsertionSlot_DELTA_M(slot);
  int16_t delta_p = InsertionSlot_DELTA_P(slot);
  int16_t lambda = Scheduler_LAMBDA(scheduler);
  int32_t ordering = IssueItem_ORDERING(item);
  int16_t tau = InsertionSlot_TAU(slot);
  int16_t phi = InsertionSlot_PHI(slot);
  int16_t delta = delta_m + delta_p;
  int16_t increase = delta;
  *GrahamListTask__TAU(task) = tau;
  *GrahamListTask__PHI(task) = phi;
  Except_CHECK(phi >= 0 && tau >= 0 && tau < lambda);
  Scheduler_freezeTask(scheduler, task, delta_m);
  if (delta != 0) {
    IssueGroup group_m = NULL, group_p = NULL;
    // Grow the IssueTable at tau.
    IssueTable_growAt(issueTable, tau, delta);
    // Split the IssueGroup at tau into the IssueGroups at tau and tau + delta.
    group_m = IssueTable_assign(issueTable, tau, InsertionSlot_GROUP_M(slot));
    group_p = IssueTable_assign(issueTable, tau + delta, InsertionSlot_GROUP_P(slot));
    // Freeze the issued tasks that are after the task to be scheduled.
    GrahamList_FOREACH_ISSUED_GrahamListTask(grahamList, TASK) {
      IssueItem ITEM = GrahamListTask_ITEM(TASK);
      if (GrahamListTask_TAU(TASK) < tau) continue;
      if (GrahamListTask_TAU(TASK) == tau) {
        if (GrahamListTask_PHI(TASK) > phi) continue;
        if (GrahamListTask_PHI(TASK) == phi) {
          if (IssueItem_ORDERING(ITEM) < ordering) continue;
        }
      }
      Scheduler_freezeTask(scheduler, TASK, delta);
    } GrahamList_ENDEACH_ISSUED_GrahamListTask;
    if (delta > 0 && tau + delta < lambda) {
      // Try to reduce the amount of LAMBDA increase.
      int increaseLambda = Scheduler_increaseLambda(scheduler, delta);
      IssueTable_shrinkTo(issueTable, lambda + increaseLambda);
    }
    // Get the new LAMBDA from the issueTable.
    increase = IssueTable_lambda(issueTable) - lambda;
    *Scheduler__LAMBDA(scheduler) = lambda + increase;
  }
  SCD_TRACE && fprintf(SCD_TRACE,
      "\t\tphi=%d\ttau=%d\tdate=%d  \tincrease=%d\n",
      phi, tau, tau + delta_m + lambda*phi, increase);
  Scheduler_optimizeFlows(scheduler, true, delta != 0);
  return tau + delta_m;
}

#line 350 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
void
Insertion_schedule(Insertion this)
{
  Scheduler scheduler = Insertion_SCHEDULER(this);
  GrahamList grahamList = Scheduler_GRAHAMLIST(scheduler);
  IssueTable issueTable = Scheduler_ISSUETABLE(scheduler);
  IssueItem firstItem = Scheduler_FIRSTITEM(scheduler);
  GrahamListTask firstTask = GrahamList_FIRSTTASK(grahamList);
  Except_REQUIRE(GrahamListTask_ITEM(firstTask) == firstItem);
  SCD_TRACE && fprintf(SCD_TRACE, "*** INSERTION SCHEDULE:\n");
  Scheduler_initializeFlows(scheduler, true);
  *GrahamListTask__TAU(firstTask) = 0;
  *GrahamListTask__PHI(firstTask) = 0;
  IssueTable_ready(issueTable, Scheduler_LAMBDA(scheduler));
  IssueTable_insert(issueTable, firstItem, GrahamListTask_TAU(firstTask));
  // Insertion scheduling main loop.
  while (!IDList_isEmpty(GrahamList_TOISSUE(grahamList))) {
    GrahamListTask task = Insertion_selectTask(this);
    InsertionSlot slot = Insertion_selectSlot(this, task);
    int tau = Insertion_scheduleTaskAtSlot(this, task, slot);
    // Move task to the issued state.
    IssueTable_insert(issueTable, GrahamListTask_ITEM(task), tau);
    Except_CHECK(task == IDList_firstItem(GrahamList_TOISSUE(grahamList)));
    IDList_drop_(GrahamList_TOISSUE(grahamList));
    IDList_push_(GrahamList_ISSUED(grahamList), task);
  }
  // Schedule the NOISSUE tasks using the MCRL dates.
  GrahamList_scheduleNoIssue(grahamList);
}

#line 386 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
int
Insertion_Schedule(Scheduler scheduler, int min_lambda, int max_lambda, int length)
{
  Insertion insertion = Insertion_new(Scheduler_MEMORY(scheduler), scheduler, 32);
  Insertion_schedule(insertion);
  Insertion_delete(insertion);
  return Scheduler_LAMBDA(scheduler);
}

#line 395 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 403 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
void
Insertion_INIT(void)
{
  Scheduler_SetSchedule(OptimizeScheduling_Insertion, Insertion_Schedule);
}

#line 415 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
void
Insertion_FINI(void)
{
}

#line 420 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
#if XCC__C
void
Insertion_TEST(void)
{
#include "SCD/Insertion_.c"
}

#line 427 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Insertion.xcc"
int
main(int argc, char **argv)
{
  SCD_Test(Insertion_TEST);
  return 0;
}
#endif


#endif//Insertion_c_INCLUDED
