#ifndef Iterative_c_INCLUDED
#define Iterative_c_INCLUDED

#line 17 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
#include "SCD/SCD_.h"

#line 22 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
//#define TIME_STATS

#line 24 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
#ifdef TIME_STATS
#include <sys/timeb.h>
#endif

#line 28 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 30 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
static int32_t IterativeInnerLoopCounter = 0;
static int32_t IterativeNotInnerLoopCounter = 0;

#line 33 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
#define DETAILED_TRACE
//#define BINARY_II_SEARCH

#line 46 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_callbackPreScheduleNode -- Iterative modulo schedulin interface function invoked before scheduling an operation.
 *                                      Depending on its result an operation may be scheduled or not.
 *
 * Return: bool success if an operation can be scheduled.
 */
static bool
Iterative_callbackPreScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                                  CycleBasedScheduleContext context)
{
  IterativeExternalContext externalContext = (IterativeExternalContext)context->External;
  int II = externalContext->II;
  int *resourceUsageCycleMap =   externalContext->ResourceUsageCycleMap 
                               + Resource__COUNT*(cycle % II);
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  bool canBeScheduled = CycleBased_NoResourceConstraints(nodeId, resourceUsageCycleMap,
                                                         resourceContext);
  if (canBeScheduled == true) {
    CycleBased_AddResourcesUsage(nodeId, resourceUsageCycleMap, resourceContext);
  } else {
    int *preScheduleRetry = externalContext->PreScheduleRetry;
    if (preScheduleRetry[nodeId]++ > II) {
      context->InfeasibleSchedule = true;
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "nodeId: %d maximum resource check exceeded, ", 
                                  nodeId);
  SCD_TRACE && fprintf(SCD_TRACE, "schedule is infeasible!\n");
    } else {
  SCD_TRACE && fprintf(SCD_TRACE, "nodeId: %d CANNOT be ", nodeId);
  SCD_TRACE && fprintf(SCD_TRACE, "scheduled because there are not enough resources\n");
#endif
    } // end if
  } // end if
  return canBeScheduled;
} // end Iterative_callbackPreScheduleNode

#line 82 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_callbackPreUnScheduleNode -- Iterative modulo scheduling interface function invoked before unscheduling an operation.
 */
static void
Iterative_callbackPreUnScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                                    CycleBasedScheduleContext context)
{
  IterativeExternalContext externalContext = (IterativeExternalContext)context->External;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  int II = externalContext->II;
  int *resourceUsageCycleMap =   externalContext->ResourceUsageCycleMap 
                               + Resource__COUNT*(cycle % II);
  CycleBased_ReleaseResourcesUsage(nodeId, resourceUsageCycleMap, resourceContext);
} // end of Iterative_callbackPreunScheduleNode

#line 97 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_callbackPostScheduleNode -- Iterative modulo scheduling interface function invoked after scheduling an operation.
 */
static void
Iterative_callbackPostScheduleNode(CycleBasedNodeId nodeId, CycleBasedCycle cycle,
                                   CycleBasedScheduleContext context)
{
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "nodeId: %d has been scheduled at cycle: %d\n",
                                  nodeId, cycle);
#endif
} // end of Iterative_callbackPostScheduleNode

#line 110 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_callbackNewCycle -- Iterative modulo scheduling interface function invoked at the beginning of each new cycle.
 */
static void
Iterative_callbackNewCycle(CycleBasedCycle cycle,
                           CycleBasedScheduleContext context)
{
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "cycle: %d\n", cycle);
#endif
} // end Iterative_callbackNewCycle

#line 122 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_PickReadyAtCurrCycle -- Iterative modulo scheduling interface function which picks all ready operations from readyList that can be scheduled at current cycle.
 */
static void
Iterative_PickReadyAtCurrCycle(int cycle, CycleBasedScheduleContext context)
{
  CycleBasedArcId arcId;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  IterativeExternalContext externalContext = (IterativeExternalContext)context->External;
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  bool *isScheduled = context->IsScheduled;
  IDList inArcs = context->InArcs;
  IDList additionalInArcs = externalContext->AdditionalInArcs;
  CycleBasedNodeId *src = context->Src;
  int *latency = context->Latency;
  int *distance = externalContext->Distance;
  int *priorities = context->Priorities;
  int II = externalContext->II;
  IDList readyList = context->ReadyList;
  IDList readyAtCurrCycleList = context->ReadyAtCurrCycleList;
  IDList_FOREACH(readyList, CycleBasedNodeId, pNodeId) {
    int maxCycle;
    bool ready = true;
    CycleBasedCycle minStartCycle = 0; 
    IDList_FOREACH((inArcs + *pNodeId), CycleBasedArcId, pArcId) {
      arcId = *pArcId;
      if (isScheduled[src[arcId]] == true) {
        Except_CHECK(distance[arcId] == 0);
        maxCycle = scheduledAt[src[arcId]] + latency[arcId];
        if (maxCycle > minStartCycle) {
          minStartCycle = maxCycle;
        } // end if
      } // end if
    } IDList_ENDEACH;
    IDList_FOREACH((additionalInArcs + *pNodeId), CycleBasedArcId, pArcId) {
      arcId = *pArcId;
      if (isScheduled[src[arcId]] == true) {
        maxCycle = scheduledAt[src[arcId]] + latency[arcId] - II * distance[arcId];
        if (maxCycle > minStartCycle) {
          minStartCycle = maxCycle;
        } // end if
      } // end if
    } IDList_ENDEACH;
    Except_CHECK(cycle < minStartCycle + II);
    if (cycle < minStartCycle) {
      ready = false;
    } // end if
    if (ready == true) {
      CycleBasedScheduleItem scheduleItem;
      // and add to list
      scheduleItem = IDList_push(readyAtCurrCycleList,
                                 sizeof(CycleBasedScheduleItem_));
      scheduleItem->NodeId = *pNodeId;
      scheduleItem->Priority = (priorities + *pNodeId);
#ifdef DETAILED__UNWINDING_SCHEDULE_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "nodeId: %d ", *pNodeId);
  SCD_TRACE && fprintf(SCD_TRACE, "is ready to be scheduled at this cycle\n");
#endif
      IDList_remove(readyList, pNodeId, NULL);
    } // end if
  } IDList_ENDEACH;
} // end of Iterative_PickReadyAtCurrCycle

#line 186 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
static bool
Iterative_Validate(CycleBasedScheduleContext context) {
  CycleBasedNodeId nodeId;
  int32_t nodeCount = context->NodeCount;
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  IterativeExternalContext externalContext = (IterativeExternalContext)context->External;
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  bool *isScheduled = context->IsScheduled;
  IDList inArcs = context->InArcs;
  IDList additionalInArcs = externalContext->AdditionalInArcs;
  CycleBasedNodeId *src = context->Src;
  int *latency = context->Latency;
  int *distance = externalContext->Distance;
  int *priorities = context->Priorities;
  int II = externalContext->II;
  if (context->InfeasibleSchedule == true) {
    return false;
  } // end if
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    Except_CHECK(isScheduled[nodeId] == true);
    IDList_FOREACH((inArcs + nodeId), CycleBasedArcId, pArcId) {
      Except_CHECK(distance[*pArcId] == 0);
      if (scheduledAt[src[*pArcId]] + latency[*pArcId] > scheduledAt[nodeId]) {
        return false;
      } // end if
    } IDList_ENDEACH;
    IDList_FOREACH((additionalInArcs + nodeId), CycleBasedArcId, pArcId) {
      Except_CHECK(distance[*pArcId] != 0);
      if (scheduledAt[src[*pArcId]] + latency[*pArcId] - II * distance[*pArcId] > scheduledAt[nodeId]) {
        return false;
      } // end if
    } IDList_ENDEACH;
  } // end for: nodeId
  return true;
} // end of Iterative_Validate

#line 222 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 236 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
Iterative
Iterative_Ctor(Iterative this,
                 Memory memory, Scheduler scheduler)
{
  DDGraph mcrlGraph = Scheduler_MCRLGRAPH(scheduler);
  GrahamList grahamList = Scheduler_GRAHAMLIST(scheduler);
  GrahamListTask firstTask = GrahamList_FIRSTTASK(grahamList);
  GrahamListTask lastTask = GrahamList_LASTTASK(grahamList);
  int32_t firstIndex = GrahamListTask_INDEX(firstTask);
  int32_t lastIndex = GrahamListTask_INDEX(lastTask);
  *Iterative__MEMORY(this) = memory;
  *Iterative__SCHEDULER(this) = scheduler;
  *Iterative__FIRSTINDEX(this) = firstIndex;
  *Iterative__LASTINDEX(this) = lastIndex;
   //
   return this;
}

#line 254 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
void
Iterative_Dtor(Iterative this)
{
}

#line 259 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
size_t
Iterative_Size(Memory memory, Scheduler scheduler)
{
  return sizeof(Iterative_);
}

#line 274 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
#ifndef Iterative_POOLED
#define Iterative_POOLED 128
#endif//Iterative_POOLED

#line 278 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
Iterative
Iterative_new(Memory parent, Scheduler scheduler)
{
  Memory memory = Memory_new(parent, Iterative_POOLED);
  size_t size = Iterative_Size(memory, scheduler);
  Iterative this = Memory_alloc(memory, size);
  return Iterative_Ctor(this, memory, scheduler);
}

#line 291 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
Iterative
Iterative_delete(Iterative this)
{
  if (this != NULL) {
    Memory memory = Iterative_MEMORY(this);
#ifndef PRO_NDTOR
    Iterative_Dtor(this);
    Memory_free_(memory, this);
#endif//PRO_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 305 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_InnerLoopSchedule -- Prepares data for a an inner loop schedule creation and afterwards creates iterative modulo schedule.
 *
 * Return: int II for which a schedule is a valid modulo schedule.
 */
static int
Iterative_InnerLoopSchedule(Iterative this, int scheduleDates[], int lambda_min, int lambda_max)
{
  int II = lambda_min;
  CycleBasedArcId arcCounter = 0;
  CycleBasedNodeId nodeId;
  CycleBasedCycle cycle;
  int32_t cycleCount;
  Scheduler scheduler = Unwinding_SCHEDULER(this);
  int32_t firstIndex = Unwinding_FIRSTINDEX(this);
  int32_t lastIndex = Unwinding_LASTINDEX(this);
  DDGraph mcrlGraph = Unwinding_MCRLGRAPH(this);
  int32_t nodeCount = 0;
  int32_t arcCount = 0;
  int *latency;
  int *distance;
  CycleBasedNodeId *src;
  CycleBasedNodeId *dst;
  IDList inArcs;
  IDList outArcs;
  IDList additionalInArcs;
  IDList additionalOutArcs;
  bool validSchedule = false;
  int minII = II;
  int maxII = 2*minII;
  CycleBasedScheduleContext_ context[1];
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  IterativeExternalContext_ externalContext[1];
  CycleBasedScheduleInterface_ interface[1];
  int *resourceUsageCycleMap;
  int *resourceAvailabilityMap = resourceContext->ResourceAvailabilityMap;
  int *reservationsMap;
#ifdef DETAILED_TRACE
  CycleBasedCycle *scheduledAt;
  bool *isScheduled;
#endif
  Except_REQUIRE(Scheduler_isInnerLoop(scheduler) == true);
  CycleBased_ResourceAvailabilityInit(scheduler, resourceContext);
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          ++arcCount;
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  nodeCount = lastIndex - firstIndex + 1;
  CycleBased_CreateScheduleContext(nodeCount, arcCount, context);
  externalContext->Distance = Memory_alloc(Memory_Root, arcCount*sizeof(int));
  externalContext->PreScheduleRetry = Memory_alloc(Memory_Root, nodeCount*sizeof(int));
  externalContext->AdditionalInArcs = Memory_alloc(Memory_Root, 
                                                   nodeCount*sizeof(IDList_));
  externalContext->AdditionalOutArcs = Memory_alloc(Memory_Root, 
                                                    nodeCount*sizeof(IDList_));
  reservationsMap = resourceContext->ReservationsMap;
  latency = context->Latency;
  distance = externalContext->Distance;
  src = context->Src;
  dst = context->Dst;
  inArcs = context->InArcs;
  outArcs = context->OutArcs;
  additionalInArcs = externalContext->AdditionalInArcs;
  additionalOutArcs = externalContext->AdditionalOutArcs;
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    IDList_Ctor((additionalInArcs + nodeId), Memory_Root);
    IDList_Ctor((additionalOutArcs + nodeId), Memory_Root);
  } // end for: nodeId
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      nodeId = index - firstIndex;
      SCD_TRACE && fprintf(SCD_TRACE, "nodeId: %d\n", nodeId);
      CycleBased_ReservationTableInit(node, nodeId, reservationsMap);
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          CycleBasedNodeId srcNodeId = tail_index - firstIndex;
          CycleBasedNodeId dstNodeId = head_index - firstIndex;
          CycleBasedArcId arcId = arcCounter;
          int distanceValue = DDGraphArc_OMEGA(arc);
          if (distanceValue == 0) {
            *(CycleBasedArcId *)IDList_push(inArcs + dstNodeId,
                                            sizeof(CycleBasedArcId)) = arcId;
            *(CycleBasedArcId *)IDList_push(outArcs + srcNodeId,
                                            sizeof(CycleBasedArcId)) = arcId;
          } else {
            *(CycleBasedArcId *)IDList_push(additionalInArcs + dstNodeId,
                                            sizeof(CycleBasedArcId)) = arcId;
            *(CycleBasedArcId *)IDList_push(additionalOutArcs + srcNodeId,
                                            sizeof(CycleBasedArcId)) = arcId;
          } // end if
          src[arcId] = srcNodeId;
          dst[arcId] = dstNodeId;
          latency[arcId] = DDGraphArc_THETA(arc);
          distance[arcId] = distanceValue;
          SCD_TRACE && fprintf(SCD_TRACE, "arcId: %d\tsrc: %d\tdst: %d",
                                          arcId, src[arcId], dst[arcId]);
          SCD_TRACE && fprintf(SCD_TRACE, "\tlatency: %d\tdistance: %d\n", 
                                          latency[arcId], distance[arcId]);
          ++arcCounter;
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    }
  } DDGraph_ENDEACH_DDGraphNode;
  /*
  SCD_TRACE && fprintf(SCD_TRACE, "STATS: LAO_II: %d, CycleBased_FindMinII: %d\n", 
                                  II, CycleBased_FindMinII(mcrlGraph, firstIndex,
                                                           lastIndex));
  */
  interface->callbackPreScheduleNode = Iterative_callbackPreScheduleNode;
  interface->callbackPostScheduleNode = CycleBased_callbackPostScheduleNode;
  interface->callbackPreUnScheduleNode = Iterative_callbackPreUnScheduleNode;
  interface->PickReadyAtCurrCycle = Iterative_PickReadyAtCurrCycle;
  interface->callbackNewCycle = Iterative_callbackNewCycle;
  CycleBased_CalculatePriorities(CycleBased_GetSinkNodes, context);
#ifdef BINARY_II_SEARCH
  // BEGIN: binary search for modulo schedule
  II = minII;
  externalContext->II = II;
  externalContext->ResourceUsageCycleMap = Memory_alloc(Memory_Root, 
                                                        Resource__COUNT*II*sizeof(int));
  memset(externalContext->PreScheduleRetry, 0, nodeCount*sizeof(int));
  resourceUsageCycleMap = externalContext->ResourceUsageCycleMap;
  for (cycle = 0; cycle < II; ++cycle) {
    CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                   resourceContext);      
  } // end for: cycle
  context->External = externalContext;
  validSchedule = CycleBased_Schedule(context, interface);
  validSchedule &= Iterative_Validate(context);
  if (validSchedule == false) {
    II = minII + 1;
    externalContext->II = II;
    externalContext->ResourceUsageCycleMap = Memory_alloc(Memory_Root, 
                                                          Resource__COUNT*II*sizeof(int));
    memset(externalContext->PreScheduleRetry, 0, nodeCount*sizeof(int));
    resourceUsageCycleMap = externalContext->ResourceUsageCycleMap;
    for (cycle = 0; cycle < II; ++cycle) {
      CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                     resourceContext);      
    } // end for: cycle
    context->External = externalContext;
    validSchedule = CycleBased_Schedule(context, interface);
    validSchedule &= Iterative_Validate(context);
    if (validSchedule == false) {
      int j = 1;
      II=minII+2*j;
      while (II < maxII+1) {
        externalContext->II = II;
        externalContext->ResourceUsageCycleMap = Memory_alloc(Memory_Root, 
                                                              Resource__COUNT*II*sizeof(int));
        memset(externalContext->PreScheduleRetry, 0, nodeCount*sizeof(int));
        resourceUsageCycleMap = externalContext->ResourceUsageCycleMap;
        for (cycle = 0; cycle < II; ++cycle) {
          CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                          resourceContext);      
        } // end for: cycle
        context->External = externalContext;
        validSchedule = CycleBased_Schedule(context, interface);
        validSchedule &= Iterative_Validate(context);
        if (validSchedule == true) {
          break;
        } // end if
        ++j;
        II=minII+2*j;
      } // end while
      if (validSchedule == true) {
        if (II > minII+2) {
          validSchedule = false;
          for (II = minII + 2*(j-1); II <= minII + 2*j; ++II) {
            externalContext->II = II;
            externalContext->ResourceUsageCycleMap = Memory_alloc(Memory_Root, 
                                                                  Resource__COUNT*II*sizeof(int));
            memset(externalContext->PreScheduleRetry, 0, nodeCount*sizeof(int));
            resourceUsageCycleMap = externalContext->ResourceUsageCycleMap;
            for (cycle = 0; cycle < II; ++cycle) {
              CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                              resourceContext);      
            } // end for: cycle
            context->External = externalContext;
            validSchedule = CycleBased_Schedule(context, interface);
            validSchedule &= Iterative_Validate(context);
            if (validSchedule == true) {
              break;
            } // end if
          }
        }
      } else {
        while (true) {
          externalContext->II = II;
          externalContext->ResourceUsageCycleMap = Memory_alloc(Memory_Root, 
                                                                Resource__COUNT*II*sizeof(int));
          memset(externalContext->PreScheduleRetry, 0, nodeCount*sizeof(int));
          resourceUsageCycleMap = externalContext->ResourceUsageCycleMap;
          for (cycle = 0; cycle < II; ++cycle) {
            CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                           resourceContext);      
          } // end for: cycle
          context->External = externalContext;
          validSchedule = CycleBased_Schedule(context, interface);
          validSchedule &= Iterative_Validate(context);
          if (validSchedule == true) {
            SCD_TRACE && fprintf(SCD_TRACE, "Kernel: \n");
            for (cycle = 0;
                 cycle < II;
                 ++cycle) {
              SCD_TRACE && fprintf(SCD_TRACE, "cycle: %d", cycle);
              CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                              resourceContext);
              for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
                if (context->ScheduledAt[nodeId] % II == cycle) {
                  SCD_TRACE && fprintf(SCD_TRACE, "\t%d", nodeId);
                  CycleBased_AddResourcesUsage(nodeId, 
                                               (resourceUsageCycleMap + Resource__COUNT*cycle), 
                                               resourceContext);
                  Except_CHECK(context->IsScheduled[nodeId]);
                } // end if
              } // end for: nodeid
              SCD_TRACE && fprintf(SCD_TRACE, "\n");
            } // end for: cycle
            SCD_TRACE && fprintf(SCD_TRACE, "End of Kernel: \n");
            break;
          } else {
            ++II;
          } // end if
        } // end while
      } // end if
    } // end if
  } // end if
  Except_CHECK(validSchedule);
  // END: binary search for modulo schedule
#else
  while (true) {
    bool scheduleValid;
    externalContext->II = II;
    externalContext->ResourceUsageCycleMap = Memory_alloc(Memory_Root, 
                                                          Resource__COUNT*II*sizeof(int));
    memset(externalContext->PreScheduleRetry, 0, nodeCount*sizeof(int));
    resourceUsageCycleMap = externalContext->ResourceUsageCycleMap;
    for (cycle = 0; cycle < II; ++cycle) {
      CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                     resourceContext);      
    } // end for: cycle
    context->External = externalContext;
    scheduleValid = CycleBased_Schedule(context, interface);
    scheduleValid &= Iterative_Validate(context);
    if (scheduleValid == true) {
      SCD_TRACE && fprintf(SCD_TRACE, "Kernel: \n");
      for (cycle = 0;
           cycle < II;
           ++cycle) {
        SCD_TRACE && fprintf(SCD_TRACE, "cycle: %d", cycle);
        CycleBased_ClearResourcesUsage((resourceUsageCycleMap + Resource__COUNT*cycle), 
                                       resourceContext);
        for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
          if (context->ScheduledAt[nodeId] % II == cycle) {
            SCD_TRACE && fprintf(SCD_TRACE, "\t%d", nodeId);
            CycleBased_AddResourcesUsage(nodeId, 
                                         (resourceUsageCycleMap + Resource__COUNT*cycle), 
                                         resourceContext);
            Except_CHECK(context->IsScheduled[nodeId]);
          } // end if
        } // end for: nodeid
        SCD_TRACE && fprintf(SCD_TRACE, "\n");
      } // end for: cycle
      SCD_TRACE && fprintf(SCD_TRACE, "End of Kernel: \n");
      break;
    } else {
      ++II;
    } // end if
  } // end while
#endif
#ifdef DETAILED_TRACE
  scheduledAt = context->ScheduledAt;
  isScheduled = context->IsScheduled;
  cycleCount = context->CycleCount;
  for (cycle = 0;
       cycle < cycleCount;
       ++cycle) {
    SCD_TRACE && fprintf(SCD_TRACE, "cycle: %d", cycle);
    for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
      if (scheduledAt[nodeId] == cycle) {
        SCD_TRACE && fprintf(SCD_TRACE, "\t%d", nodeId);
        Except_CHECK(isScheduled[nodeId] == true);
      } // end if
    } // end for: nodeId
    SCD_TRACE && fprintf(SCD_TRACE, "\n");
  } // end for: cycle
#endif
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    scheduleDates[nodeId+firstIndex] = scheduledAt[nodeId];
  } // end for: nodeId
  CycleBased_ReleaseScheduleContext(context);
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    IDList_Dtor((additionalInArcs + nodeId));
    IDList_Dtor((additionalOutArcs + nodeId));
  } // end for: nodeId
  SCD_TRACE && fprintf(SCD_TRACE, "STATS_PERF: InnerLoop loop: %d cycleCount: %d II: %d\n",
                                  IterativeInnerLoopCounter, cycleCount, II);
  return II;
} // end of Iterative_InnerLoopSchedule

#line 624 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
/*
 * Iterative_Schedule -- Creates a schedule for a particular part of code.
 *
 * Return: int lambda.
 */
static int
Iterative_Schedule(Scheduler scheduler, int lambda_min, int lambda_max, int length)
{
#ifdef TIME_STATS
  struct timeb tStart;
  struct timeb tStop;
#endif
  int lambda = INT_MAX;
  bool isInnerLoop = Scheduler_isInnerLoop(scheduler);
  int pipelining = Scheduler_PIPELINING(scheduler);
  DDGraph mcrlGraph = Scheduler_MCRLGRAPH(scheduler);
  int32_t nodeCount = DDGraph_nodeCount(mcrlGraph);
  int* scheduleDates = Memory_alloc(Memory_Root, nodeCount*sizeof(int));
  int32_t firstIndex = Scheduler_FIRSTINDEX(scheduler);
  int32_t lastIndex = Scheduler_LASTINDEX(scheduler), nodeIndex;
  GrahamListTask *index2task = Scheduler_INDEX2TASK(scheduler);
  int32_t i;
  if (isInnerLoop == true) {
    Iterative this = Iterative_new(Scheduler_MEMORY(scheduler), scheduler);
    SCD_TRACE && fprintf(SCD_TRACE, "*** INNER_LOOP=%d\n", IterativeInnerLoopCounter);
  SCD_TRACE && fprintf(SCD_TRACE, "Iterative_Schedule(%d,%d)\n", lambda_min, lambda_max);
#ifdef TIME_STATS
    ftime(&tStart);
#endif
    lambda = Iterative_InnerLoopSchedule(this, scheduleDates, lambda_min, lambda_max);
#ifdef TIME_STATS
    ftime(&tStop);
    SCD_TRACE && fprintf(SCD_TRACE, "STATS_TIME: InnerLoop loop: %d execTime: %ld [ms]\n",
                                    IterativeInnerLoopCounter,
                                    1000 * (tStop.time - tStart.time) + tStop.millitm
                                    - tStart.millitm);
#endif
    ++IterativeInnerLoopCounter;
    Iterative_delete(this);
  } else {
    SCD_TRACE && fprintf(SCD_TRACE, "*** NOT_INNER_LOOP=%d\n", IterativeNotInnerLoopCounter);
    SCD_TRACE && fprintf(SCD_TRACE, "CycleBased_Schedule(%d,%d)\n", lambda_min, lambda_max);
#ifdef TIME_STATS
    ftime(&tStart);
#endif
    lambda = CycleBased_NotInnerLoopSchedule(scheduler, scheduleDates, mcrlGraph, firstIndex, lastIndex);
#ifdef TIME_STATS
    ftime(&tStop);
    SCD_TRACE && fprintf(SCD_TRACE, "STATS_TIME: NotInnerLoop execTime: %ld [ms]\n",
                                    1000 * (tStop.time - tStart.time) + tStop.millitm
                                    - tStart.millitm);
#endif
    ++IterativeNotInnerLoopCounter;
  } // end if
  // Commit back the schedule dates.
  if (lambda < INT_MAX) {
    for (nodeIndex = firstIndex; nodeIndex <= lastIndex; nodeIndex++) {
      int date = scheduleDates[nodeIndex];
      GrahamListTask task = index2task[nodeIndex];
      *GrahamListTask__TAU(task) = date%lambda;
      *GrahamListTask__PHI(task) = date/lambda;
    } // end for: nodeIndex
  } // end if
  return lambda;
} // end of Iterative_Schedule

#line 690 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 698 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
void
Iterative_INIT(void)
{
  Scheduler_SetSchedule(OptimizeScheduling_Iterative, Iterative_Schedule);
}

#line 709 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
#if XCC__C
void
Iterative_TEST(void)
{
#include "SCD/Iterative_.c"
}

#line 716 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Iterative.xcc"
int
main(int argc, char **argv)
{
  SCD_Test(Iterative_TEST);
  return 0;
}
#endif


#endif//Iterative_c_INCLUDED
