#ifndef RCMS__h_INCLUDED
#define RCMS__h_INCLUDED

//
#include "SCD/RCMS.h"

//
RCMSStatus
RCMS_OpenStub(const char *procedure_name, const char *label_name, const char *options);

//
RCMSResourceId
RCMS_MakeResourceStub(const char *name, int num_available);

//
RCMSReservationId
RCMS_MakeReservationStub(const char *name, int num_entries, int entries[]);

//
RCMSRegFileId
RCMS_MakeRegFileStub(const char *name, int num_allocatable);

//
RCMSNodeId
RCMS_MakeNodeStub(RCMSReservationId reservationId, RCMSNodeType nodeType,
                  int num_write, RCMSRegFileId regFileIds[], const char *mnemonic);

//
RCMSArcId
RCMS_MakeArcStub(RCMSNodeId source_nodeId, RCMSNodeId target_nodeId, RCMSArcType arcType,
                 RCMSRegFileId regFileId, int latency, int distance, const char *temporary_name);

//
int
RCMS_ProcessStub(int minII, int latency, int distance, RCMSCallback callback);

//
RCMSStatus
RCMS_CloseStub(void);

//
typedef struct RCMSInterfaceContext_ {
  OptimizeFlags RCMSProblem;
  Processor Processor;
  const_RegisterSet Reserved;
  int32_t FirstIndex;
  int32_t LastIndex;
  Scheduler Scheduler;
  DDGraph DDGraph;
  RCMSResourceId ResourceMap[Resource__FIXME];
  RCMSReservationId ReservationMap[Reservation__FIXME];
  RCMSRegFileId RegFileMap[RegFile__FIXME];
  int32_t NodeCount;
  RCMSNodeId *NodeMap;
} RCMSInterfaceContext_, *RCMSInterfaceContext;

/*
 * RCMSInterface_Schedule --	Pre-Schedule using RCMS through a RCMSInterface.
 */
int
RCMSInterface_Schedule(Scheduler scheduler, int lambda_min);

/*
 * RCMS_XML --	File handle used to dump XML.
 */
extern FILE *RCMS_XML;
#ifdef ECL_NTRACE
#define RCMS_XML NULL
#endif//ECL_NTRACE

//
extern char *
RCMS_LIBRARIES;

/*
 * RCMS_INIT --	Initialize the RCMS module.
 */
void
RCMS_INIT(void);

/*
 * RCMS_FINI --	Finalize the RCMS module.
 */
#define RCMS_FINI()


#endif//RCMS__h_INCLUDED
