#ifndef Relaxation_c_INCLUDED
#define Relaxation_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#include "SCD/SCD_.h"

#line 21 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 46 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
RelaxationItem
RelaxationItem_Ctor(RelaxationItem this,
                    const uint8_t *requirements, DDGraphNode node, int iteration)
{
  *RelaxationItem__NODE(this) = node;
  *RelaxationItem__REQUIREMENTS(this) = requirements;
  *RelaxationItem__RELEASE(this) = -1;
  *RelaxationItem__DEADLINE(this) = -1;
  *RelaxationItem__LATENCY(this) = INT16_MIN;
  *RelaxationItem__ITERATION(this) = iteration;
  *RelaxationItem__TASK(this) = NULL;
  *RelaxationItem__DATES(this) = NULL;
  return this;
}

#line 72 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifndef ECL_NTRACE
bool
RelaxationItem_pretty(RelaxationItem this, FILE *file)
{
  int index = RelaxationItem_INDEX(this);
  int release = RelaxationItem_RELEASE(this);
  int deadline = RelaxationItem_DEADLINE(this);
  int iteration = RelaxationItem_ITERATION(this);
  int latency = RelaxationItem_LATENCY(this);
  int earlydate = RelaxationItem_EARLYDATE(this);
  int latedate = RelaxationItem_LATEDATE(this);
  if (file == NULL) file = stderr;
  if (iteration == 0) fprintf(file, "    _%d  ", index);
  else fprintf(file, "    _%d[%d]", index, iteration);
  fprintf(file, "\tearlydate=%d\tlatedate=%d", earlydate, latedate);
  if (latency >= 0) fprintf(file, "\tlatency=%d", latency);
  else fprintf(file, "\t\t");
  fprintf(file, "\trelease=%d\tdeadline=%d", release, deadline);
  return true;
}
#endif//ECL_NTRACE

#line 94 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 119 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
RelaxationItemTable
RelaxationItemTable_Ctor(RelaxationItemTable this,
                         Memory memory, int unwinding,
                         int32_t listCount, int16_t *_lambda)
{
  int32_t tableCount = listCount*(unwinding + 1);
  int32_t arrayCount = (listCount + 1)*(unwinding + 1);
  Except_REQUIRE(unwinding >= 0);
  IStack_Ctor(RelaxationItemTable_ITEMS(this),
              memory, sizeof(RelaxationItem_), tableCount);
  IArray_Ctor(RelaxationItemTable_IARRAY(this), memory,
               sizeof(RelaxationItem), arrayCount);
  *RelaxationItemTable__LISTCOUNT(this) = listCount;
  *RelaxationItemTable___LAMBDA(this) = _lambda;
  return this;
}

#line 136 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
RelaxationItemTable_Dtor(RelaxationItemTable this)
{
  IStack_Dtor(RelaxationItemTable_ITEMS(this));
  IArray_Dtor(RelaxationItemTable_IARRAY(this));
}

#line 225 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
RelaxationItem
RelaxationItemTable_makeItem(RelaxationItemTable this, const uint8_t *requirements,
                             DDGraphNode node, int iteration)
{
  RelaxationItem newItem = IStack_push(RelaxationItemTable_ITEMS(this));
  RelaxationItem *mapItem = RelaxationItemTable_mapItem(this, DDGraphNode_INDEX(node),
                                                        iteration);
  *mapItem = RelaxationItem_Ctor(newItem, requirements, node, iteration);
  return newItem;
}

#line 243 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
RelaxationItemTable_computeLatencies(RelaxationItemTable this,
                                     RelaxationItem firstItem, int unwinding)
{
  int16_t lambda = RelaxationItemTable_LAMBDA(this);
  RelaxationItem pastItem = RelaxationItemTable_PASTITEM(this), tail_item;
  // Initialize the RelaxationItem LATENCY fields.
  for (tail_item = firstItem; tail_item < pastItem; tail_item++) {
    *RelaxationItem__LATENCY(tail_item) = INT16_MIN;
  }
  *RelaxationItem__LATENCY(firstItem) = 0;
  // Search the firstItem successors to make their transitive LATENCY.
  for (tail_item = firstItem; tail_item < pastItem; tail_item++) {
    int tail_latency = RelaxationItem_LATENCY(tail_item);
    if (tail_latency >= 0) {
      // This node is a transitive successor of firstItem.
      int tail_iteration = RelaxationItem_ITERATION(tail_item);
      DDGraphNode tail_node = RelaxationItem_NODE(tail_item);
      DDGraphNode_FORLEAVE_DDGraphArc(tail_node, arc) {
        int delay = DDGraphArc_DELAY(arc);
        DDGraphNode head_node = DDGraphArc_HEAD(arc);
        int32_t head_index = DDGraphNode_INDEX(head_node);
        RelaxationItem head_item = *RelaxationItemTable_mapItem(this, head_index, tail_iteration);
        while (head_item != NULL && delay < 0) {
          head_item = RelaxationItemTable_nextItem(this, head_item);
          delay += lambda;
        }
        if (head_item != NULL && delay >= 0) {
          int new_latency = tail_latency + delay;
          if (RelaxationItem_LATENCY(head_item) < new_latency) {
            *RelaxationItem__LATENCY(head_item) = new_latency;
            Except_CHECK(new_latency >= 0);
          }
        }
      } DDGraphNode_ENDLEAVE_DDGraphArc;
      if (tail_iteration < unwinding) {
        // Include the regularizing constraint.
        RelaxationItem next_item = RelaxationItemTable_nextItem(this, tail_item);
        if (next_item != NULL) {
          int new_latency = tail_latency + lambda;
          if (RelaxationItem_LATENCY(next_item) < new_latency) {
            *RelaxationItem__LATENCY(next_item) = new_latency;
          }
        }
      }
    }
  }
}

#line 298 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
RelaxationItemTable_updateDeadlines(RelaxationItemTable this, RelaxationItem firstItem, int deadline)
{
  int16_t lambda = RelaxationItemTable_LAMBDA(this);
  RelaxationItem baseItem = RelaxationItemTable_BASEITEM(this), head_item;
  *RelaxationItem__DEADLINE(firstItem) = deadline;
  // Search the firstItem predecessors to make their DEADLINE consistent.
  for (head_item = firstItem; head_item >= baseItem; --head_item) {
    int head_deadline = RelaxationItem_DEADLINE(head_item);
    DDGraphNode head_node = RelaxationItem_NODE(head_item);
    DDGraphNode_FORENTER_DDGraphArc(head_node, arc) {
      int delay = DDGraphArc_DELAY(arc);
      DDGraphNode tail_node = DDGraphArc_TAIL(arc);
      int tail_index = DDGraphNode_INDEX(tail_node);
      RelaxationItem tail_item = *RelaxationItemTable_mapItem(this, tail_index, 0);
      if (tail_item != NULL && delay >= 0) {
        int tail_deadline = RelaxationItem_DEADLINE(tail_item);
        if (tail_deadline > head_deadline - delay) {
          *RelaxationItem__DEADLINE(tail_item) = head_deadline - delay;
        }
      }
    } DDGraphNode_ENDENTER_DDGraphArc;
  }
}

#line 329 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
RelaxationItemTable_completeDeadlines(RelaxationItemTable this, RelaxationItem firstItem)
{
  int16_t lambda = RelaxationItemTable_LAMBDA(this);
  RelaxationItem baseItem = RelaxationItemTable_BASEITEM(this), item;
  for (item = firstItem; item >= baseItem; --item) {
    int prev_deadline = RelaxationItem_DEADLINE(item) - lambda;
    RelaxationItem prev_item = RelaxationItemTable_prevItem(this, item);
    if (prev_item != NULL) *RelaxationItem__DEADLINE(prev_item) = prev_deadline;
    else break;
  }
}

#line 348 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifndef ECL_NTRACE
bool
RelaxationItemTable_pretty(RelaxationItemTable this, FILE *file)
{
  if (file == NULL) file = stderr;
  RelaxationItemTable_FOREACH_RelaxationItem(this, item) {
    RelaxationItem_pretty(item, file);
    fprintf(file, "\n");
  } RelaxationItemTable_ENDEACH_RelaxationItem;
  return true;
}
#endif//ECL_NTRACE

#line 361 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 405 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
RelaxationStep
RelaxationStep_Ctor(RelaxationStep this,
                    Memory memory, RelaxationItemTable itemTable,
                    RelaxationItem firstItem, const uint8_t *availability)
{
  int32_t itemCount = RelaxationItemTable_PASTITEM(itemTable) - firstItem;
  *RelaxationStep__ITEMTABLE(this) = itemTable;
  *RelaxationStep__FIRSTITEM(this) = firstItem;
  FastList_Ctor(RelaxationStep_FASTLIST(this), memory, itemCount, availability);
  *RelaxationStep__RESOURCE(this) = Resource__UNDEF;
  *RelaxationStep__FLAGS(this) = 0;
  return this;
}

#line 419 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
RelaxationStep_Dtor(RelaxationStep this)
{
  FastList_Dtor(RelaxationStep_FASTLIST(this));
}

#line 425 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
size_t
RelaxationStep_Size(Memory memory, RelaxationItemTable itemTable,
                    RelaxationItem firstItem, const uint8_t *availability) {
  return sizeof(RelaxationStep_);
}

#line 441 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifndef RelaxationStep_POOLED
#define RelaxationStep_POOLED 128
#endif//RelaxationStep_POOLED

#line 445 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
RelaxationStep
RelaxationStep_new(Memory parent, RelaxationItemTable itemTable,
                   RelaxationItem firstItem, const uint8_t *availability)
{
  Memory memory = Memory_new(parent, RelaxationStep_POOLED);
  size_t size = RelaxationStep_Size(memory, itemTable, firstItem, availability);
  RelaxationStep this = Memory_alloc(memory, size);
  return RelaxationStep_Ctor(this, memory, itemTable, firstItem, availability);
}

#line 459 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
RelaxationStep
RelaxationStep_delete(RelaxationStep this)
{
  if (this != NULL) {
    Memory memory = RelaxationStep_MEMORY(this);
#ifndef PRO_NDTOR
    RelaxationStep_Dtor(this);
    Memory_free_(memory, this);
#endif//PRO_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 508 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
bool
RelaxationStep_enterItems(RelaxationStep this)
{
  FastList fastList = RelaxationStep_FASTLIST(this);
  RelaxationStep_FOREACH_RelaxationItem(this, item) {
    int release = RelaxationItem_RELEASE(item);
    int deadline = RelaxationItem_DEADLINE(item);
    if (release >= deadline) return false;
    else {
      const uint8_t *requirements = RelaxationItem_REQUIREMENTS(item);
      FastListTask task = FastList_makeTask(fastList, requirements, item);
      *RelaxationItem__TASK(item) = task;
    }
  } RelaxationStep_ENDEACH_RelaxationItem;
  return true;
}

#line 531 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifdef XCC__C
#define SCD_TRACE stdout
#endif

#line 535 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
bool
RelaxationStep_trySchedule(RelaxationStep this, int p, int q)
{
  FastList fastList = RelaxationStep_FASTLIST(this);
  Resource resource = RelaxationStep_RESOURCE(this);
  Except_REQUIRE(resource != Resource__UNDEF);
  if (RelaxationStep_isSorted(this)) {
    // Update the release dates of the FastListTasks.
    FastList_updateRelease(fastList, p, q);
    // Update the deadline dates of the FastListTasks.
    FastList_updateDeadline(fastList, p, q);
  } else {
    FastListTask lists = FastList_LISTS(fastList);
    RelaxationItem firstItem = RelaxationStep_FIRSTITEM(this);
    // Initialize the release and the deadline dates of the FastListTasks.
    RelaxationStep_FOREACH_RelaxationItem(this, item) {
      FastListTask task = RelaxationItem_TASK(item);
      int release = RelaxationItem_RELEASE(item);
      int deadline = RelaxationItem_DEADLINE(item);
      int latency = RelaxationItem_LATENCY(item);
      if (item == firstItem) {
        release = p, deadline = q + 1;
      } else if (release < latency + q) {
        release = latency + q;
        Except_CHECK(latency >= 0);
      }
      if (FastListTask_RELEASE(task) != release) {
        *FastListTask__RELEASE(task) = release;
      }
      if (FastListTask_DEADLINE(task) != deadline) {
        *FastListTask__DEADLINE(task) = deadline;
      }
    } RelaxationStep_ENDEACH_RelaxationItem;
    FastList_sortTasks(fastList, FastListType_Release, lists);
    FastList_sortTasks(fastList, FastListType_Deadline, lists);
    *RelaxationStep__FLAGS(this) |= RelaxationStepFlag_Sorted;
  }
  // FastList schedule each resource.
  if (!FastList_schedule(fastList, resource)) return false;
#ifdef XCC__C
  SCD_TRACE && fprintf(SCD_TRACE, "\n"), FastList_pretty(fastList, SCD_TRACE), fprintf(SCD_TRACE, "\t");
#endif
  return true;
}

#line 587 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
int
RelaxationStep_modifyDeadline(RelaxationStep this, Resource resource)
{
  RelaxationItem firstItem = RelaxationStep_FIRSTITEM(this);
  FastListTask first_task = RelaxationItem_TASK(firstItem);
  int first_iteration = RelaxationItem_ITERATION(firstItem);
  const uint8_t *first_requirements = FastListTask_REQUIREMENTS(first_task);
  static uint8_t zero_requirements[Resource__FIXME];
  int p = RelaxationItem_RELEASE(firstItem);
  int q = RelaxationItem_DEADLINE(firstItem) - 1;
  *RelaxationStep__RESOURCE(this) = resource;
  *RelaxationStep__FLAGS(this) &= ~RelaxationStepFlag_Sorted;
  SCD_TRACE && !first_iteration && fprintf(SCD_TRACE, "    %s_%d\t",
      Resource_name(resource), RelaxationItem_INDEX(firstItem));
  SCD_TRACE && first_iteration && fprintf(SCD_TRACE, "    %s_%d[%d]\t",
      Resource_name(resource), RelaxationItem_INDEX(firstItem), first_iteration);
  {
    // Phase 1: binary search for the maximal q without the firstTask.
    int l = 0, r = q - p, s = -1;
    *FastListTask__REQUIREMENTS(first_task) = zero_requirements;
    while (r >= l) {
      int i = (l + r) >> 1;
      bool feasible = RelaxationStep_trySchedule(this, p, p + i);
      if (feasible) {
        s = p + i;
        l = i + 1;
        SCD_TRACE && fprintf(SCD_TRACE, "OK1[%d,%d] ", p, p + i);
      } else {
        r = i - 1;
        SCD_TRACE && fprintf(SCD_TRACE, "KO1[%d,%d] ", p, p + i);
      }
    }
    *FastListTask__REQUIREMENTS(first_task) = first_requirements;
    SCD_TRACE && fprintf(SCD_TRACE, s < 0? "INFEASIBLE\n": "");
    if (s < 0) return s;
    q = s;
  }
  if (first_requirements[resource] != 0) {
    // Phase 2: binary search for the maximal p with the firstTask.
    int l = 0, r = q - p, s = -1;
    while (r >= l) {
      int i = (l + r) >> 1;
      bool feasible = RelaxationStep_trySchedule(this, p + i, q);
      if (feasible) {
        s = p + i;
        l = i + 1;
        SCD_TRACE && fprintf(SCD_TRACE, "OK2[%d,%d] ", p + i, q);
      } else {
        r = i - 1;
        SCD_TRACE && fprintf(SCD_TRACE, "KO2[%d,%d] ", p + i, q);
      }
    }
    SCD_TRACE && fprintf(SCD_TRACE, s < 0? "INFEASIBLE\n": "");
    if (s < 0) return s;
    p = s;
  } else {
    p = q;
  }
  SCD_TRACE && fprintf(SCD_TRACE, "%s=%d\n",
      p + 1 == RelaxationItem_DEADLINE(firstItem)? "modified": "MODIFIED", p + 1);
  return p + 1;
}

#line 656 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
bool
RelaxationStep_isStationary(RelaxationStep this, int distance)
{
  int32_t itemCount = RelaxationStep_ITEMCOUNT(this);
  RelaxationItem firstItem = RelaxationStep_FIRSTITEM(this);
  RelaxationItemTable itemTable = RelaxationStep_ITEMTABLE(this);
  RelaxationItem baseItem = RelaxationItemTable_BASEITEM(itemTable);
  RelaxationItem pastItem = RelaxationItemTable_PASTITEM(itemTable);
  int32_t listCount = RelaxationItemTable_LISTCOUNT(itemTable);
  int16_t lambda = RelaxationItemTable_LAMBDA(itemTable);
  int32_t maxCount = listCount*distance;
  if (maxCount > itemCount) return false;
  RelaxationStep_FOREACH_RelaxationItem(this, item) {
    int next_deadline = RelaxationItem_DEADLINE(item) + lambda;
    RelaxationItem next_item = RelaxationItemTable_nextItem(itemTable, item);
    if (next_item != NULL && RelaxationItem_DEADLINE(next_item) == next_deadline) {
      if (--maxCount <= 0) return true;
    } else return false;
  } RelaxationStep_ENDEACH_RelaxationItem;
  return false;
}

#line 684 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifndef ECL_NTRACE
bool
RelaxationStep_pretty(RelaxationStep this, FILE *file)
{
  if (file == NULL) file = stderr;
  fprintf(file, "RelaxationStep\n");
  RelaxationStep_FOREACH_RelaxationItem(this, item) {
    RelaxationItem_pretty(item, file);
    fprintf(file, "\n");
  } RelaxationStep_ENDEACH_RelaxationItem;
  return true;
}
#endif//ECL_NTRACE

#line 698 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 712 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
const char *
RelaxationKind_name_(RelaxationKind this)
{
  static const char *_[] = {
    "Forward",
    "Backward",
  };
  Except_REQUIRE((unsigned)this < RelaxationKind__);
  return _[this];
}

#line 753 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
Relaxation
Relaxation_Ctor(Relaxation this,
                Memory memory, RelaxationKind kind,
                GrahamList grahamList, int unwinding)
{
  int32_t listCount = GrahamList_COUNT(grahamList);
  const uint8_t *availability = GrahamList_AVAILABILITY(grahamList);
  int16_t *_lambda = GrahamList__LAMBDA(grahamList);
  RelaxationItemTable_Ctor(Relaxation_ITEMTABLE(this),
                           memory, unwinding, listCount, _lambda);
  *Relaxation__AVAILABILITY(this) = availability;
  *Relaxation__UNWINDING(this) = unwinding;
  *Relaxation__KIND(this) = kind;
  return this;
}

#line 769 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
Relaxation_Dtor(Relaxation this)
{
  RelaxationItemTable_Dtor(Relaxation_ITEMTABLE(this));
}

#line 775 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
size_t
Relaxation_Size(Memory memory, RelaxationKind kind,
                GrahamList grahamList, int unwinding) {
	return sizeof(Relaxation_);
}

#line 790 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifndef Relaxation_POOLED
#define Relaxation_POOLED 128
#endif//Relaxation_POOLED

#line 794 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
Relaxation
Relaxation_new(Memory parent, RelaxationKind kind, GrahamList grahamList, int unwinding)
{
  Memory memory = Memory_new(parent, Relaxation_POOLED);
  size_t size = Relaxation_Size(memory, kind, grahamList, unwinding);
  Relaxation this = Memory_alloc(memory, size);
  Relaxation_Ctor(this, memory, kind, grahamList, unwinding);
  return this;
}

#line 808 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
Relaxation
Relaxation_delete(Relaxation this)
{
  if (this != NULL) {
    Memory memory = Relaxation_MEMORY(this);
#ifndef PRO_NDTOR
    Relaxation_Dtor(this);
    Memory_free_(memory, this);
#endif//PRO_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 857 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
Relaxation_enterGrahamList(Relaxation this, GrahamList grahamList)
{
  RelaxationItemTable itemTable = Relaxation_itemTable(this);
  int16_t lambda = RelaxationItemTable_LAMBDA(itemTable);
  int unwinding = Relaxation_UNWINDING(this), iteration = 0;
  do {
    if (Relaxation_KIND(this) == RelaxationKind_Backward) {
      GrahamList_FOREACH_GrahamListTask(grahamList, task) {
        IssueItem issueItem = GrahamListTask_ITEM(task);
        const uint8_t *requirements = IssueItem_REQUIREMENTS(issueItem);
        DDGraphNode early_node = IssueItem_EARLYNODE(issueItem);
        DDGraphNode late_node = IssueItem_LATENODE(issueItem);
        RelaxationItem item = RelaxationItemTable_makeItem(itemTable, requirements,
                                                           early_node, iteration);
        *RelaxationItem__DATES(item) = GrahamListTask_DATES(task) + iteration;
        *RelaxationItem__EARLYDATE(item) = DDGraphNode_DATE(early_node)
                                         + lambda*iteration;
        *RelaxationItem__LATEDATE(item) = DDGraph_EXTENT - DDGraphNode_DATE(late_node)
                                        + lambda*iteration;
      } GrahamList_ENDEACH_GrahamListTask;
    } else if (Relaxation_KIND(this) == RelaxationKind_Forward) {
      GrahamList_FORBACK_GrahamListTask(grahamList, task) {
        IssueItem issueItem = GrahamListTask_ITEM(task);
        const uint8_t *requirements = IssueItem_REQUIREMENTS(issueItem);
        DDGraphNode early_node = IssueItem_EARLYNODE(issueItem);
        DDGraphNode late_node = IssueItem_LATENODE(issueItem);
        RelaxationItem item = RelaxationItemTable_makeItem(itemTable, requirements,
                                                           late_node, iteration);
        *RelaxationItem__DATES(item) = GrahamListTask_DATES(task) + iteration;
        *RelaxationItem__EARLYDATE(item) = DDGraphNode_DATE(early_node)
                                         + lambda*iteration;
        *RelaxationItem__LATEDATE(item) = DDGraph_EXTENT - DDGraphNode_DATE(late_node)
                                        + lambda*iteration;
      } GrahamList_ENDBACK_GrahamListTask;
    } else Except_NEVER(true);
  } while (iteration++ < unwinding);
}

#line 902 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
Relaxation_readDates(Relaxation this)
{
  RelaxationKind kind = Relaxation_KIND(this);
  if (kind == RelaxationKind_Backward) {
    Relaxation_FOREACH_RelaxationItem(this, item) {
      int release = RelaxationItem_EARLYDATE(item);
      int deadline = RelaxationItem_LATEDATE(item) + 1;
      Except_CHECK(release >= 0 && release < deadline);
      *RelaxationItem__RELEASE(item) = release;
      *RelaxationItem__DEADLINE(item) = deadline;
    } Relaxation_ENDEACH_RelaxationItem;
  } else if (kind == RelaxationKind_Forward) {
    Relaxation_FOREACH_RelaxationItem(this, item) {
      int release = DDGraph_EXTENT - RelaxationItem_LATEDATE(item);
      int deadline = DDGraph_EXTENT - RelaxationItem_EARLYDATE(item) + 1;
      Except_CHECK(release >= 0 && release < deadline);
      *RelaxationItem__RELEASE(item) = release;
      *RelaxationItem__DEADLINE(item) = deadline;
    } Relaxation_ENDEACH_RelaxationItem;
  } else Except_NEVER(true);
}

#line 931 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
Relaxation_writeDates(Relaxation this)
{
  RelaxationKind kind = Relaxation_KIND(this);
  if (kind == RelaxationKind_Backward) {
    Relaxation_FOREACH_RelaxationItem(this, item) {
      int deadline = RelaxationItem_DEADLINE(item);
      *RelaxationItem__LATEDATE(item) = deadline - 1;
    } Relaxation_ENDEACH_RelaxationItem;
  } else if (kind == RelaxationKind_Forward) {
    Relaxation_FOREACH_RelaxationItem(this, item) {
      int deadline = RelaxationItem_DEADLINE(item);
      *RelaxationItem__EARLYDATE(item) = DDGraph_EXTENT - deadline + 1;
    } Relaxation_ENDEACH_RelaxationItem;
  } else Except_NEVER(true);
}

#line 955 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
bool
Relaxation_runRelaxationStep(Relaxation this, RelaxationStep relaxationStep,
                             int issueCount, const Resource *resources)
{
  RelaxationItemTable itemTable = Relaxation_ITEMTABLE(this);
  RelaxationItem baseItem = RelaxationItemTable_BASEITEM(itemTable);
  RelaxationItem firstItem = RelaxationStep_FIRSTITEM(relaxationStep);
  const uint8_t *requirements = RelaxationItem_REQUIREMENTS(firstItem);
  int unwinding = Relaxation_UNWINDING(this), issuePriority = 0;
  RelaxationItemTable_computeLatencies(itemTable, firstItem, unwinding);
  for (issuePriority = 0; issuePriority < issueCount; issuePriority++) {
    Resource resource = resources[issuePriority];
    if (firstItem == baseItem || requirements[resource] != 0) {
      int deadline = RelaxationStep_modifyDeadline(relaxationStep, resource);
      if (deadline < 0) return false;
      Except_CHECK(deadline <= RelaxationItem_DEADLINE(firstItem));
      if (deadline < RelaxationItem_DEADLINE(firstItem)) {
        RelaxationItemTable itemTable = Relaxation_ITEMTABLE(this);
        RelaxationItemTable_updateDeadlines(itemTable, firstItem, deadline);
      }
    }
  }
  return true;
}

#line 986 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
bool
Relaxation_testFeasibility(Relaxation this, int issueCount, const Resource *resources)
{
  int distance = 1;	//FIXME! should be MIN(MAX(OMEGA_i), OVERLAP+1)
  int unwinding = Relaxation_UNWINDING(this);
  const uint8_t *availability = Relaxation_AVAILABILITY(this);
  RelaxationItemTable itemTable = Relaxation_itemTable(this);
  int16_t lambda = RelaxationItemTable_LAMBDA(itemTable);
  RelaxationItemTable_FORBACK_RelaxationItem(itemTable, firstItem) {
    RelaxationStep relaxationStep = RelaxationStep_new(Relaxation_MEMORY(this), itemTable,
                                                       firstItem, availability);
    // Enter the RelaxationItem(s) and run the relaxationStep.
    bool feasible = RelaxationStep_enterItems(relaxationStep) &&
        Relaxation_runRelaxationStep(this, relaxationStep, issueCount, resources);
    if (feasible && distance > 0 &&
        RelaxationItem_ITERATION(firstItem) < unwinding &&
        RelaxationStep_isStationary(relaxationStep, distance)) {
      SCD_TRACE && fprintf(SCD_TRACE, "*** STATIONARY _%d[%d]\tdistance=%d\n",
          RelaxationItem_INDEX(firstItem), RelaxationItem_ITERATION(firstItem), distance);
      // Stationary RelaxationStep detected: complete the deadlines.
      RelaxationItemTable_completeDeadlines(itemTable, firstItem);
      RelaxationStep_delete(relaxationStep);
      return true;
    }
    RelaxationStep_delete(relaxationStep);
    if (!feasible) return false;
  } RelaxationItemTable_ENDBACK_RelaxationItem;
  return true;
}

#line 1022 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#ifndef ECL_NTRACE
bool
Relaxation_pretty(Relaxation this, FILE *file)
{
  RelaxationKind kind = Relaxation_KIND(this);
  int unwinding = Relaxation_UNWINDING(this);
  RelaxationItemTable itemTable = Relaxation_itemTable(this);
  int16_t lambda = RelaxationItemTable_LAMBDA(itemTable);
  if (file == NULL) file = stderr;
  fprintf(file, "Relaxation_%s\tunwinding=%d\tlambda=%d\n",
      RelaxationKind_name_(kind), unwinding, lambda);
  RelaxationItemTable_pretty(itemTable, file);
  return true;
}
#endif//ECL_NTRACE

#line 1117 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 1125 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
Relaxation_INIT(void)
{
}

#line 1136 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
void
Relaxation_FINI(void)
{
}

#line 1141 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
#if XCC__C
void
Relaxation_TEST(void)
{
#include "SCD/Relaxation_.c"
}

#line 1148 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Relaxation.xcc"
int
main(int argc, char **argv)
{
  SCD_Test(Relaxation_TEST);
  return 0;
}
#endif


#endif//Relaxation_c_INCLUDED
