#ifndef Relaxation__h_INCLUDED
#define Relaxation__h_INCLUDED

//
#include "SCD/Relaxation.h"

//
struct RelaxationItem_ {
  DDGraphNode NODE;		// The associated IssueItem.
  //@access INDEX	DDGraphNode_INDEX(RelaxationItem_NODE(this))
  const uint8_t *REQUIREMENTS;	// The Resource requirements.
  int16_t RELEASE;		// The release time of this RelaxationItem.
  int16_t DEADLINE;		// The release time of this RelaxationItem.
  int16_t LATENCY;		// Transitive latency from the first RelaxationItem.
  int16_t ITERATION;		// Iteration number in case of unwinding.
  FastListTask TASK;		// The associated FastListTask.
  GrahamListDates DATES;	// The EARLYDATE and LATEDATE of this RelaxationItem.
  //@access EARLYDATE	GrahamListDates_EARLYDATE(RelaxationItem_DATES(this))
  //@mutate EARLYDATE	GrahamListDates__EARLYDATE(RelaxationItem_DATES(this))
  //@access LATEDATE	GrahamListDates_LATEDATE(RelaxationItem_DATES(this))
  //@mutate LATEDATE	GrahamListDates__LATEDATE(RelaxationItem_DATES(this))
};

extern RelaxationItem
RelaxationItem_Ctor(RelaxationItem this, const uint8_t *requirements, DDGraphNode node, int iteration);

extern RelaxationItem
RelaxationItem_Copy(RelaxationItem this, const_RelaxationItem that);

extern void
RelaxationItem_Dtor(RelaxationItem this);

extern size_t
RelaxationItem_Size(const uint8_t *requirements, DDGraphNode node, int iteration);

#define RelaxationItem_NODE(this) ((this)->NODE)
#define RelaxationItem__NODE(this) (&(this)->NODE)
#define RelaxationItem_REQUIREMENTS(this) ((this)->REQUIREMENTS)
#define RelaxationItem__REQUIREMENTS(this) (&(this)->REQUIREMENTS)
#define RelaxationItem_RELEASE(this) ((this)->RELEASE)
#define RelaxationItem__RELEASE(this) (&(this)->RELEASE)
#define RelaxationItem_DEADLINE(this) ((this)->DEADLINE)
#define RelaxationItem__DEADLINE(this) (&(this)->DEADLINE)
#define RelaxationItem_LATENCY(this) ((this)->LATENCY)
#define RelaxationItem__LATENCY(this) (&(this)->LATENCY)
#define RelaxationItem_ITERATION(this) ((this)->ITERATION)
#define RelaxationItem__ITERATION(this) (&(this)->ITERATION)
#define RelaxationItem_TASK(this) ((this)->TASK)
#define RelaxationItem__TASK(this) (&(this)->TASK)
#define RelaxationItem_DATES(this) ((this)->DATES)
#define RelaxationItem__DATES(this) (&(this)->DATES)
#define RelaxationItem_LATEDATE(this) (GrahamListDates_LATEDATE(RelaxationItem_DATES(this)))
#define RelaxationItem_INDEX(this) (DDGraphNode_INDEX(RelaxationItem_NODE(this)))
#define RelaxationItem_EARLYDATE(this) (GrahamListDates_EARLYDATE(RelaxationItem_DATES(this)))
#define RelaxationItem__LATEDATE(this) (GrahamListDates__LATEDATE(RelaxationItem_DATES(this)))
#define RelaxationItem__EARLYDATE(this) (GrahamListDates__EARLYDATE(RelaxationItem_DATES(this)))


//
struct RelaxationItemTable_ {
  IStack_ ITEMS_;		// The inlined RelaxationItem(s).
  //@access ITEMS	RelaxationItemTable__ITEMS_(this)
  //@access MEMORY	IStack_memory(RelaxationItemTable_ITEMS(this))
  //@access BASEITEM	(RelaxationItem)IStack_base(RelaxationItemTable_ITEMS(this))
  //@access PASTITEM	(RelaxationItem)IStack_past(RelaxationItemTable_ITEMS(this))
  IArray_ IARRAY_;		// Mapping array from (INDEX, ITER) to ITEM.
  //@access IARRAY	RelaxationItemTable__IARRAY_(this)
  //@access MAPITEMS	(RelaxationItem*)IArray_base(RelaxationItemTable_IARRAY(this))
  int32_t LISTCOUNT;		// The number of items of the underlying list scheduler.
  //@access DIMENSION	(RelaxationItemTable_LISTCOUNT(this) + 1)
  int16_t *_LAMBDA;		// This Relaxation LAMBDA.
  //@access LAMBDA	(*RelaxationItemTable__LAMBDA(this))
};

extern RelaxationItemTable
RelaxationItemTable_Ctor(RelaxationItemTable this, Memory memory, int unwinding, int32_t listCount, int16_t *_lambda);

extern RelaxationItemTable
RelaxationItemTable_Copy(RelaxationItemTable this, const_RelaxationItemTable that);

extern void
RelaxationItemTable_Dtor(RelaxationItemTable this);

extern size_t
RelaxationItemTable_Size(Memory memory, int unwinding, int32_t listCount, int16_t *_lambda);

#define RelaxationItemTable_ITEMS_(this) ((this)->ITEMS_)
#define RelaxationItemTable__ITEMS_(this) (&(this)->ITEMS_)
#define RelaxationItemTable_IARRAY_(this) ((this)->IARRAY_)
#define RelaxationItemTable__IARRAY_(this) (&(this)->IARRAY_)
#define RelaxationItemTable_LISTCOUNT(this) ((this)->LISTCOUNT)
#define RelaxationItemTable__LISTCOUNT(this) (&(this)->LISTCOUNT)
#define RelaxationItemTable__LAMBDA(this) ((this)->_LAMBDA)
#define RelaxationItemTable___LAMBDA(this) (&(this)->_LAMBDA)
#define RelaxationItemTable_LAMBDA(this) ((*RelaxationItemTable__LAMBDA(this)))
#define RelaxationItemTable_IARRAY(this) (RelaxationItemTable__IARRAY_(this))
#define RelaxationItemTable_BASEITEM(this) ((RelaxationItem)IStack_base(RelaxationItemTable_ITEMS(this)))
#define RelaxationItemTable_DIMENSION(this) ((RelaxationItemTable_LISTCOUNT(this) + 1))
#define RelaxationItemTable_MAPITEMS(this) ((RelaxationItem*)IArray_base(RelaxationItemTable_IARRAY(this)))
#define RelaxationItemTable_ITEMS(this) (RelaxationItemTable__ITEMS_(this))
#define RelaxationItemTable_PASTITEM(this) ((RelaxationItem)IStack_past(RelaxationItemTable_ITEMS(this)))
#define RelaxationItemTable_MEMORY(this) (IStack_memory(RelaxationItemTable_ITEMS(this)))


/*
 * RelaxationItemTable_items --	For use by RelaxationItemTable_FOREACH_RelaxationItem.
 */
static inline IStack
RelaxationItemTable_items(RelaxationItemTable this)
{
  return RelaxationItemTable_ITEMS(this);
}

/*
 * RelaxationItemTable_FOREACH_RelaxationItem --	Iterate over this RelaxationItemTable RelaxationItems.
 */
#define RelaxationItemTable_FOREACH_RelaxationItem(this, item) { \
  IStack RelaxationItemTable_ITEMS = RelaxationItemTable_items(this); \
  IStack_FOREACH(RelaxationItemTable_ITEMS, RelaxationItem_, item)
#define RelaxationItemTable_ENDEACH_RelaxationItem \
  IStack_ENDEACH \
}

/*
 * RelaxationItemTable_FORBACK_RelaxationItem --	Iterate over this RelaxationItemTable RelaxationItems.
 */
#define RelaxationItemTable_FORBACK_RelaxationItem(this, item) { \
  IStack RelaxationItemTable_ITEMS = RelaxationItemTable_items(this); \
  IStack_FORBACK(RelaxationItemTable_ITEMS, RelaxationItem_, item)
#define RelaxationItemTable_ENDBACK_RelaxationItem \
  IStack_ENDBACK \
}

/*
 * RelaxationItemTable_nextItem --	Find the next RelaxationItem in case of unwinding.
 */
static inline RelaxationItem
RelaxationItemTable_nextItem(RelaxationItemTable this, RelaxationItem item)
{
  int32_t listCount = RelaxationItemTable_LISTCOUNT(this);
  RelaxationItem pastItem = RelaxationItemTable_PASTITEM(this);
  RelaxationItem nextItem = item + listCount;
  if (nextItem >= pastItem) return NULL;
  Except_CHECK(RelaxationItem_NODE(nextItem) == RelaxationItem_NODE(item));
  Except_CHECK(RelaxationItem_ITERATION(nextItem) == RelaxationItem_ITERATION(item) + 1);
  return nextItem;
}

/*
 * RelaxationItemTable_prevItem --	Find the previous RelaxationItem in case of unwinding.
 */
static inline RelaxationItem
RelaxationItemTable_prevItem(RelaxationItemTable this, RelaxationItem item)
{
  int32_t listCount = RelaxationItemTable_LISTCOUNT(this);
  RelaxationItem baseItem = RelaxationItemTable_BASEITEM(this);
  RelaxationItem nextItem = item - listCount;
  if (nextItem < baseItem) return NULL;
  Except_CHECK(RelaxationItem_NODE(nextItem) == RelaxationItem_NODE(item));
  Except_CHECK(RelaxationItem_ITERATION(nextItem) == RelaxationItem_ITERATION(item) - 1);
  return nextItem;
}

/*
 * RelaxationItemTable_mapItem --	RelaxationItemTable entry corresponding to index and iteration.
 */
static inline RelaxationItem *
RelaxationItemTable_mapItem(RelaxationItemTable this, int32_t index, int iteration)
{
  RelaxationItem *mapItems = RelaxationItemTable_MAPITEMS(this);
  int32_t dimension = RelaxationItemTable_DIMENSION(this);
  return mapItems + iteration*dimension + index;
}

/*
 * RelaxationItemTable_makeItem --	Make a new RelaxationItem for this RelaxationItemTable.
 */
RelaxationItem
RelaxationItemTable_makeItem(RelaxationItemTable this, const uint8_t *requirements,
                             DDGraphNode node, int iteration);

/*
 * RelaxationItemTable_computeLatencies --	Compute the transitive latencies from firstItem.
 */
void
RelaxationItemTable_computeLatencies(RelaxationItemTable this,
                                     RelaxationItem firstItem, int unwinding);

/*
 * RelaxationItemTable_updateDeadlines --	Update the deadlines to be consistent from firstItem.
 */
void
RelaxationItemTable_updateDeadlines(RelaxationItemTable this, RelaxationItem firstItem, int deadline);

//
struct RelaxationStep_ {
  RelaxationItemTable ITEMTABLE;	// The RelaxationItemTable of this RelaxationStep.
  //@access PASTITEM	RelaxationItemTable_PASTITEM(RelaxationStep_ITEMTABLE(this))
  RelaxationItem FIRSTITEM;	// The first RelaxationItem of this RelaxationStep.
  //@access ITEMCOUNT	(RelaxationStep_PASTITEM(this) - RelaxationStep_FIRSTITEM(this))
  FastList_ FASTLIST_;		// This RelaxationStep FastList scheduler.
  //@access FASTLIST	RelaxationStep__FASTLIST_(this)
  //@access MEMORY	FastList_MEMORY(RelaxationStep_FASTLIST(this))
  Resource RESOURCE;	// The Resource used in this RelaxationStep.
  RelaxationStepFlags FLAGS;	// This RelaxationStep flags.
  //@access isSorted	((RelaxationStep_FLAGS(this) & RelaxationStepFlag_Sorted) != 0)
};

extern RelaxationStep
RelaxationStep_Ctor(RelaxationStep this, Memory memory, RelaxationItemTable itemTable, RelaxationItem firstItem, const uint8_t *availability);

extern RelaxationStep
RelaxationStep_Copy(RelaxationStep this, const_RelaxationStep that);

extern void
RelaxationStep_Dtor(RelaxationStep this);

extern size_t
RelaxationStep_Size(Memory memory, RelaxationItemTable itemTable, RelaxationItem firstItem, const uint8_t *availability);

#define RelaxationStep_ITEMTABLE(this) ((this)->ITEMTABLE)
#define RelaxationStep__ITEMTABLE(this) (&(this)->ITEMTABLE)
#define RelaxationStep_FIRSTITEM(this) ((this)->FIRSTITEM)
#define RelaxationStep__FIRSTITEM(this) (&(this)->FIRSTITEM)
#define RelaxationStep_FASTLIST_(this) ((this)->FASTLIST_)
#define RelaxationStep__FASTLIST_(this) (&(this)->FASTLIST_)
#define RelaxationStep_RESOURCE(this) ((this)->RESOURCE)
#define RelaxationStep__RESOURCE(this) (&(this)->RESOURCE)
#define RelaxationStep_FLAGS(this) ((this)->FLAGS)
#define RelaxationStep__FLAGS(this) (&(this)->FLAGS)
#define RelaxationStep_ITEMCOUNT(this) ((RelaxationStep_PASTITEM(this) - RelaxationStep_FIRSTITEM(this)))
#define RelaxationStep_isSorted(this) (((RelaxationStep_FLAGS(this) & RelaxationStepFlag_Sorted) != 0))
#define RelaxationStep_PASTITEM(this) (RelaxationItemTable_PASTITEM(RelaxationStep_ITEMTABLE(this)))
#define RelaxationStep_MEMORY(this) (FastList_MEMORY(RelaxationStep_FASTLIST(this)))
#define RelaxationStep_FASTLIST(this) (RelaxationStep__FASTLIST_(this))


/*
 * RelaxationStep_firstItem --	For use by RelaxationStep_FOREACH_RelaxationItem.
 */
static inline RelaxationItem
RelaxationStep_firstItem(RelaxationStep this)
{
  return RelaxationStep_FIRSTITEM(this);
}

/*
 * RelaxationStep_pastItem --	For use by RelaxationStep_FOREACH_RelaxationItem.
 */
static inline RelaxationItem
RelaxationStep_pastItem(RelaxationStep this)
{
  return RelaxationStep_PASTITEM(this);
}

/*
 * RelaxationStep_FOREACH_RelaxationItem --	Iterate over this RelaxationStep RelaxationItems.
 */
#define RelaxationStep_FOREACH_RelaxationItem(this, item) { \
  RelaxationItem RelaxationStep_FIRSTITEM = RelaxationStep_firstItem(this); \
  RelaxationItem RelaxationStep_PASTITEM = RelaxationStep_pastItem(this), item; \
  for (item = RelaxationStep_FIRSTITEM; item < RelaxationStep_PASTITEM; item++) {
#define RelaxationStep_ENDEACH_RelaxationItem \
  } \
}

/*
 * RelaxationStep_enterItems --	Enter the RelaxationItem(s) of this RelaxationStep.
 */
bool
RelaxationStep_enterItems(RelaxationStep this);

/*
 * RelaxationStep_trySchedule --	Try to schedule for the given release and deadline dates.
 */
bool
RelaxationStep_trySchedule(RelaxationStep this, int p, int q);

/*
 * RelaxationStep_modifyDeadline --	Compute the modified deadline of the first RelaxationItem.
 * Return:	The modified deadline, or negative value if infeasible.
 */
int
RelaxationStep_modifyDeadline(RelaxationStep this, Resource resource);

/*
 * RelaxationStep_isStationary --	Test if this RelaxationStep is stationary.
 */
bool
RelaxationStep_isStationary(RelaxationStep this, int distance);

//
struct Relaxation_ {
  RelaxationItemTable_ ITEMTABLE_;
  //@access ITEMTABLE	Relaxation__ITEMTABLE_(this)
  //@access MEMORY	RelaxationItemTable_MEMORY(Relaxation_ITEMTABLE(this))
  const uint8_t *AVAILABILITY;	// This Relaxation resource availability.
  int16_t UNWINDING;		// This Relaxation unwinding level.
  RelaxationKind KIND;	// This RelaxationKind.
};

extern Relaxation
Relaxation_Ctor(Relaxation this, Memory memory, RelaxationKind kind, GrahamList grahamList, int unwinding);

extern Relaxation
Relaxation_Copy(Relaxation this, const_Relaxation that);

extern void
Relaxation_Dtor(Relaxation this);

extern size_t
Relaxation_Size(Memory memory, RelaxationKind kind, GrahamList grahamList, int unwinding);

#define Relaxation_ITEMTABLE_(this) ((this)->ITEMTABLE_)
#define Relaxation__ITEMTABLE_(this) (&(this)->ITEMTABLE_)
#define Relaxation_AVAILABILITY(this) ((this)->AVAILABILITY)
#define Relaxation__AVAILABILITY(this) (&(this)->AVAILABILITY)
#define Relaxation_UNWINDING(this) ((this)->UNWINDING)
#define Relaxation__UNWINDING(this) (&(this)->UNWINDING)
#define Relaxation_KIND(this) ((this)->KIND)
#define Relaxation__KIND(this) (&(this)->KIND)
#define Relaxation_ITEMTABLE(this) (Relaxation__ITEMTABLE_(this))
#define Relaxation_MEMORY(this) (RelaxationItemTable_MEMORY(Relaxation_ITEMTABLE(this)))


/*
 * Relaxation_itemTable --	For use by Relaxation_FOREACH_RelaxationItem.
 */
static inline RelaxationItemTable
Relaxation_itemTable(Relaxation this)
{
  return Relaxation_ITEMTABLE(this);
}

/*
 * Relaxation_FOREACH_RelaxationItem --	Iterate over this Relaxation RelaxationItems.
 */
#define Relaxation_FOREACH_RelaxationItem(this, item) { \
  RelaxationItemTable Relaxation_ITEMTABLE = Relaxation_itemTable(this); \
  RelaxationItemTable_FOREACH_RelaxationItem(Relaxation_ITEMTABLE, item)
#define Relaxation_ENDEACH_RelaxationItem \
  RelaxationItemTable_ENDEACH_RelaxationItem \
}

/*
 * Relaxation_FORBACK_RelaxationItem --	Iterate over this Relaxation RelaxationItems.
 */
#define Relaxation_FORBACK_RelaxationItem(this, item) { \
  RelaxationItemTable Relaxation_ITEMTABLE = Relaxation_itemTable(this); \
  RelaxationItemTable_FORBACK_RelaxationItem(Relaxation_ITEMTABLE, item)
#define Relaxation_ENDBACK_RelaxationItem \
  RelaxationItemTable_ENDBACK_RelaxationItem \
}

/*
 * Relaxation_enterGrahamList --	Enter a GrahamList for this Relaxation.
 */
void
Relaxation_enterGrahamList(Relaxation this, GrahamList grahamList);

/*
 * Relaxation_readDates --	Read the EARLYDATEs/LATEDATEs into the RELEASE/DEADLINE dates.
 */
void
Relaxation_readDates(Relaxation this);

/*
 * Relaxation_readDates --	Write the EARLYDATEs/LATEDATEs from the RELEASE/DEADLINE dates.
 */
void
Relaxation_writeDates(Relaxation this);

/*
 * Relaxation_runRelaxationStep --	Run a RelaxationStep of this Relaxation.
 */
bool
Relaxation_runRelaxationStep(Relaxation this, RelaxationStep relaxationStep,
                             int issueCount, const Resource *resources);

/*
 * Relaxation_testFeasibility --	Test this Relaxation feasibility status.
 */
bool
Relaxation_testFeasibility(Relaxation this, int issueCount, const Resource *resources);

/*
 * Relaxation_INIT --	Initialize the Relaxation module.
 */
void
Relaxation_INIT(void);

/*
 * Relaxation_FINI --	Finalize the Relaxation module.
 */
void
Relaxation_FINI(void);


#endif//Relaxation__h_INCLUDED
