#ifndef SCD_h_INCLUDED
#define SCD_h_INCLUDED

/*
 * !!!!	SCD.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CCL.h"
#include "CAL.h"
#include "MDS.h"
#include "AIR.h"
#include "LIR.h"
#include "PFA.h"
#include "PRO.h"
#include "XFA/XFA.h"

/*
 * The SCD (SCheDuling) library contains the following modules:
 * *	<FastList FastList.html> Fast List scheduling of Leung-Palem-Pnueli.
 * *	<GrahamList GrahamList.html> Graham List scheduling support.
 * *	<Relaxation Relaxation.html> Scheduling Relaxation based on  Leung-Palem-Pnueli.
 * *	<Scheduler Scheduler.html> Instruction Scheduler driver.
 * *	<Insertion Insertion.html> Insertion instruction scheduling.
 * *	<CycleBased CycleBased.html> Cycle-Based acyclic scheduling.
 * *	<Unwinding Unwinding.html> Unwinding modulo scheduling.
 * *	<Iterative Iterative.html> Iterative modulo scheduling.
 * *	<Acyclic Acyclic.html> Acyclic instruction scheduler.
 * *	<Formulation Formulation.html> RCMSP time-indexed Formulation(s).
 * *	<Construction Construction.html> Software pipeline Construction.
 * *	<RCMS RCMS.html> Interface to the RCMS component.
 */

//
#include "SCD/FastList.h"
#include "SCD/GrahamList.h"
#include "SCD/Relaxation.h"
#include "SCD/Scheduler.h"
#include "SCD/Insertion.h"
#include "SCD/CycleBased.h"
#include "SCD/Unwinding.h"
#include "SCD/Iterative.h"
#include "SCD/Acyclic.h"
#include "SCD/Formulation.h"
#include "SCD/Construction.h"
#include "SCD/RCMS.h"

/*
 * SCD_TRACE --	File handle used to trace the SCD.
 */
extern FILE *SCD_TRACE;
#ifdef ECL_NTRACE
#define SCD_TRACE NULL
#endif//ECL_NTRACE

/*
 * Variables used to control SCD in non-release mode.
 */
extern int LAO_STARTSCD;
extern int LAO_STARTSWP;
extern int LAO_IDENTITY;
extern int LAO_SERIALIZE;
extern int FORCE_LLPALEVEL;
extern int FORCE_UNWINDING;
extern int LAO_MAXITEMCOUNT;
extern int FORCE_SCANNING;
extern int FORCE_EXITWEIGHT;

/*
 * SCD_INIT --	SCD init function.
 */
void
SCD_INIT(void);

/*
 * SCD_FINI --	SCD fini function.
 */
void
SCD_FINI(void);


#endif//SCD_h_INCLUDED
