#ifndef Unwinding_c_INCLUDED
#define Unwinding_c_INCLUDED

#line 17 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
#include "SCD/SCD_.h"

#line 22 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
//#define TIME_STATS

#line 24 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
#ifdef TIME_STATS
#include <sys/timeb.h>
#endif

#line 28 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 30 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
//#define DETAILED_TRACE
#define BINARY_II_SEARCH

#line 33 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
static int32_t UnwindingInnerLoopCounter = 0;
static int32_t UnwindingNotInnerLoopCounter = 0;

#line 50 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
/*
 * Unwinding_KernelSearch -- Searches for a kernel in a schedule.
 *
 * Return: boolean success if kernel has been found.
 */
static bool
Unwinding_KernelSearch(CycleBasedScheduleContext context, CycleBasedCycle kernel[])
{
  UnwindingExternalContext externalContext = (UnwindingExternalContext)context->External;
  IDList patternDictionaryList = externalContext->PatternDictionaryList;
  int II = externalContext->II;
  CycleBasedNodeId nodeId;
  CycleBasedCycle cycle;
  CycleBasedCycle lastOccurCycle = 0;
  bool kernelFound = false;
  CycleBasedCycle *scheduledAt = context->ScheduledAt;
  int32_t cycleCount = context->CycleCount;
  int32_t nodeCount = context->NodeCount;
  int32_t iterationCount = externalContext->IterationCount;
  int32_t kernelNodeCount = nodeCount / iterationCount;
  CycleBasedNodeId *kernelArray = Memory_alloc(Memory_Root, 
                                               kernelNodeCount*sizeof(CycleBasedNodeId));
  Except_REQUIRE(kernelNodeCount * iterationCount == nodeCount);
  SCD_TRACE && fprintf(SCD_TRACE, "*** KERNEL SEARCH\n");
  IDList_empty(patternDictionaryList, NULL);
  for (cycle = 0; cycle < cycleCount - (II - 1); cycle += II) {
    int32_t quasiHash = 0;
    int32_t kernelArrayPos = 0;
    int i;
    int32_t j;
    SCD_TRACE && fprintf(SCD_TRACE, "*** SEARCH=%d\n", cycle/II);
    // get a pattern from schedule with a period length of II
    for (i = 0; i < II; ++i) {
      for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
        if (scheduledAt[nodeId] == cycle + i) {
          Except_CHECK(kernelArrayPos < kernelNodeCount);
          kernelArray[kernelArrayPos++] = nodeId;    
        } // end if
      } // end for: nodeId
    } // end for: i
    // check if the pattern contains appropriate number of nodes
    if (kernelArrayPos == kernelNodeCount) {
      bool allNodes = true;
      for (nodeId = 0; nodeId < kernelNodeCount; ++nodeId) {
        bool foundNode = false;
        kernelArrayPos = 0;
        while (kernelArrayPos < kernelNodeCount) {
          if (kernelArray[kernelArrayPos] % kernelNodeCount == nodeId) {
            foundNode = true;
            // calculate quasi hash value for simplier pattern checking
            quasiHash =   (quasiHash<<6) + (quasiHash<<16) - quasiHash 
                        + (nodeId%kernelNodeCount);
            break;
          }
          ++kernelArrayPos;
        }
        if (foundNode == false) {
          allNodes = false;
          break;
        } // end if
      } // end for
      // if all nodes in the pattern are uniqe
      if (allNodes == true) {
        bool patternExists = false;
        IDList_FOREACH(patternDictionaryList, 
                       UnwindingPatternListItem_, 
                       patternListItem) {
          CycleBasedNodeId *patternArray = patternListItem->patternArray;
          for (j = 0; j < kernelNodeCount; ++j) {
            SCD_TRACE && fprintf(SCD_TRACE, "\tcheck: %d <-> %d\n", 
                                            patternArray[j] % kernelNodeCount, 
                                            kernelArray[j] % kernelNodeCount);
            if (patternArray[j] % kernelNodeCount == kernelArray[j] % kernelNodeCount) {
              patternExists |= true;
            } else {
              patternExists = false;
              break;
            }
          }
          if (patternExists == true) {
            Except_CHECK(patternListItem->quasiHash == quasiHash);
            if (++patternListItem->num > 2) {
              kernelFound = true;
              lastOccurCycle = cycle;
            } // end if
            break;
          } // end if
        } IDList_ENDEACH;
        if (kernelFound == true ) {
          break;
        } // end if
        // if there is no such pattern word in the dictionary then put the current pattern to it
        if (patternExists == false) {
          UnwindingPatternListItem patternListItem;
          CycleBasedNodeId *patternArray = Memory_alloc(Memory_Root, 
                                                        kernelNodeCount*sizeof(CycleBasedNodeId));
          patternListItem = IDList_push(patternDictionaryList, sizeof(UnwindingPatternListItem_));
          patternListItem->quasiHash = quasiHash;
          patternListItem->num = 1;
          patternListItem->patternArray = patternArray;
          for (j = 0; j < kernelNodeCount; ++j) {
            patternArray[j] = kernelArray[j];
          }
        } // end if: patternExists == false
      } // end if: allNodes = true
    } // end if: count(kernelList) == nodes
  } // end for: cycle
  if (kernelFound == true) {
    Except_CHECK(lastOccurCycle > 0);
    SCD_TRACE && fprintf(SCD_TRACE, "*** KERNEL FOUND Cycle=%d\n", lastOccurCycle);
    for (cycle = lastOccurCycle;
         cycle < lastOccurCycle + II; ++cycle) {
      for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
        if (scheduledAt[nodeId] == cycle) {
          CycleBasedCycle offset = (nodeId/kernelNodeCount)*II;
          kernel[nodeId % kernelNodeCount] = cycle - offset;
          Except_CHECK(cycle >= offset);
        } // end if
      } // end for: nodeId
    } // end for: cycle
  } else {
    SCD_TRACE && fprintf(SCD_TRACE, "*** KERNEL NOT FOUND\n");
  } // end if
  return kernelFound;
} // end Unwinding

#line 176 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 190 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
Unwinding
Unwinding_Ctor(Unwinding this,
               Memory memory, Scheduler scheduler)
{
  DDGraph mcrlGraph = Scheduler_MCRLGRAPH(scheduler);
  GrahamList grahamList = Scheduler_GRAHAMLIST(scheduler);
  GrahamListTask firstTask = GrahamList_FIRSTTASK(grahamList);
  GrahamListTask lastTask = GrahamList_LASTTASK(grahamList);
  int32_t firstIndex = GrahamListTask_INDEX(firstTask);
  int32_t lastIndex = GrahamListTask_INDEX(lastTask);
  *Unwinding__MEMORY(this) = memory;
  *Unwinding__SCHEDULER(this) = scheduler;
  *Unwinding__FIRSTINDEX(this) = firstIndex;
  *Unwinding__LASTINDEX(this) = lastIndex;
   //
   return this;
}

#line 208 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
void
Unwinding_Dtor(Unwinding this)
{
}

#line 213 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
size_t
Unwinding_Size(Memory memory, Scheduler scheduler)
{
  return sizeof(Unwinding_);
}

#line 228 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
#ifndef Unwinding_POOLED
#define Unwinding_POOLED 128
#endif//Unwinding_POOLED

#line 232 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
Unwinding
Unwinding_new(Memory parent, Scheduler scheduler)
{
  Memory memory = Memory_new(parent, Unwinding_POOLED);
  size_t size = Unwinding_Size(memory, scheduler);
  Unwinding this = Memory_alloc(memory, size);
  return Unwinding_Ctor(this, memory, scheduler);
}

#line 245 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
Unwinding
Unwinding_delete(Unwinding this)
{
  if (this != NULL) {
    Memory memory = Unwinding_MEMORY(this);
#ifndef PRO_NDTOR
    Unwinding_Dtor(this);
    Memory_free_(memory, this);
#endif//PRO_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 259 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
/*
 * Unwinding_InnerLoopSchedule -- Prepares data for a an inner loop schedule creation and afterwards creates unwinding modulo schedule.
 *
 * Return: int II for which a schedule is a valid modulo schedule.
 */
static int
Unwinding_InnerLoopSchedule(Unwinding this, int scheduleDates[], int lambda_min, int lambda_max)
{
  int II = lambda_min;
  CycleBasedNodeId nodeId;
  CycleBasedArcId firstDependencyArcId;
  CycleBasedArcId arcId;
  CycleBasedArcId arcCounter = 0;
  CycleBasedCycle cycle;
  int32_t cycleCount;
  Scheduler scheduler = Unwinding_SCHEDULER(this);
  int32_t firstIndex = Unwinding_FIRSTINDEX(this);
  int32_t lastIndex = Unwinding_LASTINDEX(this);
  DDGraph mcrlGraph = Unwinding_MCRLGRAPH(this);
  int32_t nodeCount = 0;
  int32_t physicalNodeCount = 0;
  int32_t arcCount = 0;
  int32_t physicalArcCount = 0;
  int *latency;
  int *distance;
  CycleBasedNodeId *src;
  CycleBasedNodeId *dst;
  IDList inArcs;
  IDList outArcs;
  bool validSchedule = false;
  bool kernelFound = false;
  int minII = II;
  int maxII = 2*minII;
  int32_t iterationCount = 10;
  int32_t iteration;
  CycleBasedScheduleContext_ context[1];
  UnwindingExternalContext_ externalContext[1];
  CycleBasedResourceContext resourceContext = context->ResourceContext;
  CycleBasedScheduleInterface_ interface[1];
  int *resourceAvailabilityMap = resourceContext->ResourceAvailabilityMap;
  int *reservationsMap;
  CycleBasedCycle *kernel;
#ifdef DETAILED_TRACE
  CycleBasedCycle *scheduledAt;
  bool *isScheduled;
#endif
  externalContext->IterationCount = iterationCount;
  Except_REQUIRE(Scheduler_isInnerLoop(scheduler) == true);
  CycleBased_ResourceAvailabilityInit(scheduler, resourceContext);
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        ASSERT(tail_index==index);
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          int distanceValue = DDGraphArc_OMEGA(arc);
          int arcsNum = iterationCount - distanceValue;
          Except_CHECK(arcsNum > 0);
          arcCount += arcsNum;
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  nodeCount = lastIndex - firstIndex + 1;
  physicalNodeCount = nodeCount*iterationCount;
  arcCount += nodeCount*(iterationCount - 1);
  kernel = Memory_alloc(Memory_Root, nodeCount*sizeof(CycleBasedCycle));
  CycleBased_CreateScheduleContext(physicalNodeCount, arcCount, context);
  IDList_Ctor(externalContext->PatternDictionaryList, Memory_Root);
  latency = context->Latency;
  src = context->Src;
  dst = context->Dst;
  inArcs = context->InArcs;
  outArcs = context->OutArcs;
  reservationsMap = resourceContext->ReservationsMap;
#ifdef DETAILED_TRACE
  scheduledAt = context->ScheduledAt;
  isScheduled = context->IsScheduled;
#endif
  SCD_TRACE && fprintf(SCD_TRACE, "*** RESERVATION TABLE INIT\n");
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      Operation operation;
      IssueItem issueItem;
      Scheduling scheduling;
      operation = DDGraphNode_OPERATION(node);
      Except_REQUIRE(operation!=NULL);
      issueItem = Operation_issueItem(operation);
      scheduling = IssueItem_SCHEDULING(issueItem);
      nodeId = index - firstIndex;
      for (iteration = 0; iteration < iterationCount; ++iteration) {
        CycleBasedNodeId newNodeId = nodeId + iteration*nodeCount;
        CycleBased_ReservationTableInit(node, newNodeId, reservationsMap);
      } // end for: iteration
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  SCD_TRACE && fprintf(SCD_TRACE, "*** ARC EXTRACTION\n");
  DDGraph_FOREACH_DDGraphNode(mcrlGraph, node) {
    int32_t index = DDGraphNode_INDEX(node);
    if (index >= firstIndex && index <= lastIndex) {
      DDGraphNode_FORLEAVE_DDGraphArc(node, arc) {
        int32_t tail_index = DDGraphNode_INDEX(DDGraphArc_TAIL(arc));
        int32_t head_index = DDGraphNode_INDEX(DDGraphArc_HEAD(arc));
        ASSERT(tail_index==index);
        if (   (tail_index >= firstIndex && tail_index <= lastIndex)
            && (head_index >= firstIndex && head_index <= lastIndex)) {
          int distanceValue = DDGraphArc_OMEGA(arc);
          int newIterationCount = iterationCount - distanceValue;
          Except_CHECK(newIterationCount > 0);
          for (iteration = 0; iteration < newIterationCount; ++iteration) {
            CycleBasedArcId newArcId = arcCounter;
            CycleBasedNodeId srcNodeId = tail_index - firstIndex + iteration*nodeCount;
            CycleBasedNodeId dstNodeId =   head_index - firstIndex 
                                         + (iteration+distanceValue)*nodeCount;
            *(CycleBasedArcId *)IDList_push(inArcs + dstNodeId,
                                            sizeof(CycleBasedArcId)) = newArcId;
            *(CycleBasedArcId *)IDList_push(outArcs + srcNodeId,
                                            sizeof(CycleBasedArcId)) = newArcId;
            src[newArcId] = srcNodeId;
            dst[newArcId] = dstNodeId;
            latency[newArcId] = DDGraphArc_THETA(arc);
#ifdef DETAILED_TRACE
            SCD_TRACE && fprintf(SCD_TRACE, "\tarcId: %d\tsrc: %d\tdst: %d",
                                            newArcId, src[newArcId], dst[newArcId]);
            SCD_TRACE && fprintf(SCD_TRACE, "\tlatency: %d\n", latency[newArcId]);
#endif
            ++arcCounter;
          } // end for: iteration
        } // end if
      } DDGraphNode_ENDLEAVE_DDGraphArc;
    } // end if
  } DDGraph_ENDEACH_DDGraphNode;
  interface->callbackPreScheduleNode = CycleBased_callbackPreScheduleNode;
  interface->callbackPostScheduleNode = CycleBased_callbackPostScheduleNode;
  interface->callbackPreUnScheduleNode = CycleBased_callbackPreUnScheduleNode;
  interface->PickReadyAtCurrCycle = CycleBased_PickReadyAtCurrCycle;
  interface->callbackNewCycle = CycleBased_callbackNewCycle;
  firstDependencyArcId = arcCounter;
  for (iteration = 0; iteration < iterationCount-1; ++iteration) {
    for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
      CycleBasedNodeId srcNodeId = nodeId + iteration*nodeCount;
      CycleBasedNodeId dstNodeId = nodeId + (iteration+1)*nodeCount;
      arcId = arcCounter;
      *(CycleBasedArcId *)IDList_push(inArcs + dstNodeId,
                                      sizeof(CycleBasedArcId)) = arcId;
      *(CycleBasedArcId *)IDList_push(outArcs + srcNodeId,
                                      sizeof(CycleBasedArcId)) = arcId;
      src[arcId] = srcNodeId;
      dst[arcId] = dstNodeId;
      latency[arcId] = II;    
      ++arcCounter;
    } // end for: nodeId
  } // end for: iteration
  Except_CHECK(arcCounter == arcCount);
  CycleBased_CalculatePriorities(CycleBased_GetSinkNodes, context);
#ifdef BINARY_II_SEARCH
  // BEGIN: binary search for modulo schedule
  externalContext->II = minII;
  context->External = externalContext;
  for (arcId = firstDependencyArcId; arcId < arcCount; ++arcId) {
    latency[arcId] = minII;    
  } // end for: arcId
  validSchedule = CycleBased_Schedule(context, interface);
  Except_CHECK(validSchedule);
  kernelFound = Unwinding_KernelSearch(context, kernel);
  if (kernelFound == false) {
    II = minII + 1;
    externalContext->II = II;
    context->External = externalContext;
    for (arcId = firstDependencyArcId; arcId < arcCount; ++arcId) {
      latency[arcId] = minII+1;    
    } // end for: arcId
    validSchedule = CycleBased_Schedule(context, interface);
    Except_CHECK(validSchedule);
    kernelFound = Unwinding_KernelSearch(context, kernel);
    if (kernelFound == false) {
      int j = 1;
      II=minII+2*j;        
      while (II < maxII+1) {
        externalContext->II = II;
        context->External = externalContext;
        for (arcId = firstDependencyArcId; arcId < arcCount; ++arcId) {
          latency[arcId] = II;    
        } // end for: arcId
        validSchedule = CycleBased_Schedule(context, interface);
        Except_CHECK(validSchedule);
        kernelFound = Unwinding_KernelSearch(context, kernel);
        if (kernelFound == true) {
          break;
        } // end if
        ++j;
        II=minII+2*j;
      } // end while
      if (kernelFound == true) {
        if (II > minII+2) {
          kernelFound = false;
          for (II = minII + 2*(j-1); II <= minII + 2*j; ++II) {
            externalContext->II = II;
            context->External = externalContext;
            for (arcId = firstDependencyArcId; arcId < arcCount; ++arcId) {
              latency[arcId] = II;    
            } // end for: arcId
            validSchedule = CycleBased_Schedule(context, interface);
            Except_CHECK(validSchedule);
            kernelFound = Unwinding_KernelSearch(context, kernel);
            if (kernelFound == true) {
              break;
            } // end if
          } // end for: II
        } // end if
      } else {
        while (true) { 
          externalContext->II = II;
          context->External = externalContext;
          for (arcId = firstDependencyArcId; arcId < arcCount; ++arcId) {
            latency[arcId] = II;    
          } // end for: arcId
          validSchedule = CycleBased_Schedule(context, interface);
          Except_CHECK(validSchedule);
          kernelFound = Unwinding_KernelSearch(context, kernel);
          if (kernelFound == true) {
            break;
          } else {
            ++II;
          } // end if
        } // end while
      } // end if
    } // end if
  } // end if
  Except_CHECK(kernelFound);
  // END: binary search for modulo schedule
#else
  while (true) { 
    externalContext->II = II;
    context->External = externalContext;
    for (arcId = firstDependencyArcId; arcId < arcCount; ++arcId) {
      latency[arcId] = II;    
    } // end for: arcId
    validSchedule = CycleBased_Schedule(context, interface);
    Except_CHECK(validSchedule);
    kernelFound = Unwinding_KernelSearch(context, kernel);
    if (kernelFound == true) {
      break;
    } else {
      ++II;
    } // end if
  } // end while
#endif
#ifdef DETAILED_TRACE
  SCD_TRACE && fprintf(SCD_TRACE, "*** SCHEDULE\n");
  cycleCount = context->CycleCount;
  for (cycle = 0;
       cycle < cycleCount; 
       ++cycle) {
    SCD_TRACE && fprintf(SCD_TRACE, "\tcycle: %d", cycle);
    for (nodeId = 0; nodeId < physicalNodeCount; ++nodeId) {
      if (scheduledAt[nodeId] == cycle) {
        SCD_TRACE && fprintf(SCD_TRACE, "\t%d(%d)", 
                                        nodeId%nodeCount, 
                                        nodeId/nodeCount);
        Except_CHECK(isScheduled[nodeId]);
      } // end if
    } // end for: nodeId
    SCD_TRACE && fprintf(SCD_TRACE, "\n");
  } // end for: cycle
#endif
  for (nodeId = 0; nodeId < nodeCount; ++nodeId) {
    scheduleDates[nodeId+firstIndex] = kernel[nodeId];
  } // end for: nodeId
  IDList_Dtor(externalContext->PatternDictionaryList);
  CycleBased_ReleaseScheduleContext(context);
  SCD_TRACE && fprintf(SCD_TRACE, "STATS_PERF: InnerLoop loop: %d cycleCount: %d II: %d\n",
                                  UnwindingInnerLoopCounter, cycleCount, II);
  return II;
} // end of Unwinding_InnerLoopSchedule                

#line 539 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
/*
 * Unwinding_Schedule -- Creates a schedule for a particular part of code.
 *
 * Return: int lambda.
 */
static int
Unwinding_Schedule(Scheduler scheduler, int lambda_min, int lambda_max, int length)
{
#ifdef TIME_STATS
  struct timeb tStart;
  struct timeb tStop;
#endif
  int lambda = INT_MAX;
  bool isInnerLoop = Scheduler_isInnerLoop(scheduler);
  int pipelining = Scheduler_PIPELINING(scheduler);
  DDGraph mcrlGraph = Scheduler_MCRLGRAPH(scheduler);
  int32_t nodeCount = DDGraph_nodeCount(mcrlGraph);
  int* scheduleDates = Memory_alloc(Memory_Root, nodeCount*sizeof(int));
  int32_t firstIndex = Scheduler_FIRSTINDEX(scheduler);
  int32_t lastIndex = Scheduler_LASTINDEX(scheduler), nodeIndex;
  GrahamListTask *index2task = Scheduler_INDEX2TASK(scheduler);
  if (isInnerLoop == true) {
    Unwinding this = Unwinding_new(Scheduler_MEMORY(scheduler), scheduler);
    SCD_TRACE && fprintf(SCD_TRACE, "*** INNER_LOOP=%d\n", UnwindingInnerLoopCounter);
    SCD_TRACE && fprintf(SCD_TRACE, "Unwinding_Schedule(%d,%d)\n", lambda_min, lambda_max);
#ifdef TIME_STATS
    ftime(&tStart);
#endif
    lambda = Unwinding_InnerLoopSchedule(this, scheduleDates, lambda_min, lambda_max);
#ifdef TIME_STATS
    ftime(&tStop);
    SCD_TRACE && fprintf(SCD_TRACE, "STATS_TIME: InnerLoop loop: %d execTime: %ld [ms]\n",
                                    UnwindingInnerLoopCounter,
                                    1000 * (tStop.time - tStart.time) + tStop.millitm
                                    - tStart.millitm);
#endif
    ++UnwindingInnerLoopCounter;
    Unwinding_delete(this);
  } else {
    SCD_TRACE && fprintf(SCD_TRACE, "*** NOT_INNER_LOOP=%d\n", UnwindingNotInnerLoopCounter);
    SCD_TRACE && fprintf(SCD_TRACE, "CycleBased_Schedule(%d,%d)\n", lambda_min, lambda_max);
#ifdef TIME_STATS
    ftime(&tStart);
#endif
    lambda = CycleBased_NotInnerLoopSchedule(scheduler, scheduleDates, mcrlGraph, firstIndex, lastIndex);
#ifdef TIME_STATS
    ftime(&tStop);
    SCD_TRACE && fprintf(SCD_TRACE, "STATS_TIME: NotInnerLoop execTime: %ld [ms]\n",
                                    1000 * (tStop.time - tStart.time) + tStop.millitm
                                    - tStart.millitm);
#endif
    ++UnwindingNotInnerLoopCounter;
  } // end if
  // Commit back the schedule dates.
  if (lambda < INT_MAX) {
    for (nodeIndex = firstIndex; nodeIndex <= lastIndex; nodeIndex++) {
      int date = scheduleDates[nodeIndex];
      GrahamListTask task = index2task[nodeIndex];
      *GrahamListTask__TAU(task) = date%lambda;
      *GrahamListTask__PHI(task) = date/lambda;
    } // end for: nodeIndex
  } // end if
  return lambda;
} // end of Unwinding_Schedule

#line 604 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 612 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
void
Unwinding_INIT(void)
{
  //Scheduler_SetSchedule(OptimizeScheduling_Unwinding, Unwinding_Schedule);
}

#line 623 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
#if XCC__C
void
Unwinding_TEST(void)
{
#include "SCD/Unwinding_.c"
}

#line 630 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/SCD/Unwinding.xcc"
int
main(int argc, char **argv)
{
  SCD_Test(Unwinding_TEST);
  return 0;
}
#endif


#endif//Unwinding_c_INCLUDED
