#ifndef BlockTrace__h_INCLUDED
#define BlockTrace__h_INCLUDED

//
#include "XFA/BlockTrace.h"

//
int
BlockTrace_compare(const void *this_pointer, const void *that_pointer);

/*
 * BlockTraceSize --	Call-back for BlockTrace_SuperBlocks.
 */
size_t
BlockTraceSize(Memory memory, OptimizeRegionType regionType, BlockTraceFlags traceFlags);

/*
 * EntryCode --	Call-back for BlockTrace_SuperBlocks.
 */
void
EntryCode(BlockTrace this, Memory memory, OptimizeRegionType regionType,
          BlockTraceFlags traceFlags);

/*
 * BlockTraceGrow --	Call-back for BlockTrace_SuperBlocks.
 */
bool
BlockTraceGrow(BlockTrace this, CFGNode cfgNode);

/*
 * BlockTraceStop --	Call-back for BlockTrace_SuperBlocks.
 */
void
BlockTraceStop(BlockTrace this, RegionFlow regionFlow, LoopScope loopScope);

/*
 * BlockTrace_configure --	Complete this BlockTrace.
 */
void
BlockTrace_complete(BlockTrace this);

//
static inline IssueItem
BlockTrace_makeIssueItem(BlockTrace this, Operation operation)
{
  Processor processor = BlockTrace_PROCESSOR(this);
  Operator operator = Operation_operator(operation);
  Instance instance = Operator_selectInstance(operator, processor);
  IDList issueItems = BlockTrace_ISSUEITEMS(this);
  size_t size = IssueItem_Size(operation, instance);
  void *alloc = IDList_push(issueItems, size);
  IssueItem issueItem = IssueItem_Ctor(alloc, operation, instance);
  *IssueItem__ORDERING(issueItem) = IDList_count(issueItems);
  if (Operator_isControl(operator)) {
    *IssueItem__FLAGS(issueItem) |= InstructionFlag_Control;
  }
  Except_CHECK(   Operator_isPseudo(operator)
               || Instance_scheduling(instance) != Scheduling__UNDEF);
  return issueItem;
}

/*
 * BlockTrace_prepareBoosting --	Prepare for Operation boosting.
 */
void
BlockTrace_prepareBoosting(BlockTrace this);

//
struct RegionFlowCover_ {
  //@args	Memory memory, RegionFlow regionFlow,
  //@args	OptimizeRegionType regionType, BlockTraceFlags traceFlags
  RegionFlow REGIONFLOW;
  OptimizeRegionType REGIONTYPE;
  BlockTraceFlags TRACEFLAGS;
  IDList_ BLOCKTRACES_;
  //@access BLOCKTRACES	RegionFlowCover__BLOCKTRACES_(this)
  //@access MEMORY	IDList_memory(RegionFlowCover_BLOCKTRACES(this))
};

extern RegionFlowCover
RegionFlowCover_Ctor(RegionFlowCover this, Memory memory, RegionFlow regionFlow, OptimizeRegionType regionType, BlockTraceFlags traceFlags);

extern RegionFlowCover
RegionFlowCover_Copy(RegionFlowCover this, const_RegionFlowCover that);

extern void
RegionFlowCover_Dtor(RegionFlowCover this);

extern size_t
RegionFlowCover_Size(Memory memory, RegionFlow regionFlow, OptimizeRegionType regionType, BlockTraceFlags traceFlags);

#define RegionFlowCover_REGIONFLOW(this) ((this)->REGIONFLOW)
#define RegionFlowCover__REGIONFLOW(this) (&(this)->REGIONFLOW)
#define RegionFlowCover_REGIONTYPE(this) ((this)->REGIONTYPE)
#define RegionFlowCover__REGIONTYPE(this) (&(this)->REGIONTYPE)
#define RegionFlowCover_TRACEFLAGS(this) ((this)->TRACEFLAGS)
#define RegionFlowCover__TRACEFLAGS(this) (&(this)->TRACEFLAGS)
#define RegionFlowCover_BLOCKTRACES_(this) ((this)->BLOCKTRACES_)
#define RegionFlowCover__BLOCKTRACES_(this) (&(this)->BLOCKTRACES_)
#define RegionFlowCover_BLOCKTRACES(this) (RegionFlowCover__BLOCKTRACES_(this))
#define RegionFlowCover_MEMORY(this) (IDList_memory(RegionFlowCover_BLOCKTRACES(this)))


//
void
RegionFlowCover_processNode(RegionFlowCover this, OptimizeRegionType regionType,
                             BlockTraceFlags traceFlags, CFGNode cfgNode,
                             LoopScope loopScope);

//
void
RegionFlowCover_processHierarchyComponent(RegionFlowCover this,
                                          HierarchyComponent component,
                                          OptimizeRegionType regionType,
                                          BlockTraceFlags traceFlags);

/*
 * BlockTrace_INIT --	Initialize the BlockTrace module.
 */
#define BlockTrace_INIT()

/*
 * BlockTrace_FINI --	Finalize the BlockTrace module.
 */
#define BlockTrace_FINI()


#endif//BlockTrace__h_INCLUDED
