#ifndef Effect_c_INCLUDED
#define Effect_c_INCLUDED

#line 16 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
#include "XFA/XFA_.h"

#line 21 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 44 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
#ifndef ECL_NTRACE
ETableEntry_ EffectFlag_Entries[] = {
  ETableENTRY(EffectFlag, Write),
  ETableENTRY(EffectFlag, Kill),
  ETableENTRY(EffectFlag, Clobber),
  ETableENTRY(EffectFlag, Indirect),
  ETableENTRY(EffectFlag, Carried),
  ETableENTRY(EffectFlag, Supplied),
  ETableENTRY(EffectFlag, InTrace),
  ETableENTRY_NULL
};
ETable_ EffectFlag_Table[1];
#endif//ECL_NTRACE

#line 89 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
Effect
Effect_Ctor(Effect this,
            Operation operation, Temporary temporary, EffectFlags flags, int index)
{
  *Effect__OPERATION(this) = operation;
  *Effect__TEMPORARY(this) = temporary;
  *Effect__XXXINDEX(this) = index;
  *Effect__FLAGS(this) = flags;
  *Effect__RELAXABLE(this) = 0;
  *Effect__LIFETIME(this) = NULL;
  *Effect__PRED(this) = NULL;
  *Effect__SUCC(this) = NULL;
  return this;
}

#line 104 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
Effect
Effect_Copy(Effect this, const_Effect that)
{
  *Effect__OPERATION(this) = Effect_OPERATION(that);
  *Effect__TEMPORARY(this) = Effect_TEMPORARY(that);
  *Effect__XXXINDEX(this) = Effect_XXXINDEX(that);
  *Effect__FLAGS(this) = Effect_FLAGS(that);
  *Effect__RELAXABLE(this) = 0;
  *Effect__LIFETIME(this) = NULL;
  *Effect__PRED(this) = NULL;
  *Effect__SUCC(this) = NULL;
  return this;
}

#line 118 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
size_t
Effect_Size(Operation operation, Temporary temporary, EffectFlags flags, int index)
{
	return sizeof(Effect_);
}

#line 135 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
#ifndef ECL_NTRACE
bool
Effect_pretty(Effect this, FILE *file)
{
  Operation operation = Effect_OPERATION(this);
  Operator operator = Operation_operator(operation);
  EffectFlags flags = Effect_FLAGS(this);
  int relaxable = Effect_RELAXABLE(this);
  if (file == NULL) file = stderr;
  fprintf(file, "%d:%s_%d\t",
      Operation_identity(operation),
      Operator_mnemonic(operator),
      Operation_index(operation));
  if (flags & EffectFlag_Write) {
    fprintf(file, "Write(%d)\t", Effect_XXXINDEX(this));
  } else {
    fprintf(file, "Read(%d) \t", Effect_XXXINDEX(this));
  }
  ETable_prettyFlags(EffectFlag_Table, flags & ~EffectFlag_Write, "+", file);
  if (relaxable > 0) {
    fprintf(file, "\trelaxable=%d", relaxable);
  }
  return true;
}
#endif//ECL_NTRACE

#line 203 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
Effect_getRegisterRAWLatency(Effect this, Effect that, Processor processor, int extra)
{
  int theta = -1;
  int readStage = 2, writeStage = 2;	// HACK! Should be min readStage.
  int this_xxxIndex = Effect_XXXINDEX(this);
  int that_xxxIndex = Effect_XXXINDEX(that);
  Temporary temporary = Effect_TEMPORARY(this);
  Operation this_operation = Effect_OPERATION(this);
  Operation that_operation = Effect_OPERATION(that);
  Operator this_operator = Operation_operator(this_operation);
  Operator that_operator = Operation_operator(that_operation);
  const_RegisterAction writeAction = NULL, readAction = NULL;
  Instance write_instance = Instance__UNDEF, read_instance = Instance__UNDEF;
  if (Effect_isIndirect(this) || Effect_isIndirect(that)) {
    if (Processor_interlocks(processor)) return 0;
  }
  if (this_xxxIndex >= 0) {
    Instance this_instance = Operator_selectInstance(this_operator, processor); // FIXME!
    const_RegisterActionArray this_actionArray = Instance_actionArray(this_instance);
    RegisterActionArray_FORWRITE_RegisterAction(this_actionArray, action) {
      if (RegisterAction_xxxIndex(action) == this_xxxIndex) writeAction = action;
    } RegisterActionArray_ENDWRITE_RegisterAction;
    if (writeAction != NULL) writeStage = RegisterAction_stage(writeAction);
    write_instance = this_instance;
  }
  if (that_xxxIndex >= 0) {
    Instance that_instance = Operator_selectInstance(that_operator, processor); // FIXME!
    const_RegisterActionArray that_actionArray = Instance_actionArray(that_instance);
    RegisterActionArray_FORREAD_RegisterAction(that_actionArray, action) {
      if (RegisterAction_xxxIndex(action) == that_xxxIndex) readAction = action;
    } RegisterActionArray_ENDREAD_RegisterAction;
    if (readAction != NULL) readStage = RegisterAction_stage(readAction);
    read_instance = that_instance;
  }
  if (Temporary_hasRegister(temporary)) {
    Register registre = Temporary_REGISTER(temporary);
    if (Instance_fixupRAW(write_instance, read_instance)) {
      const_RegisterStorageCells storageCells = Register_storageCells(registre);
      StorageCell cell = RegisterStorageCells_access(storageCells, 0);
      readStage -= Instance_latencyRAW(write_instance, read_instance, cell);
    }
    if (Effect_isIndirect(that)) {
      readStage = Register_minReadStage(registre, processor);
    }
  }
  if (Effect_isIndirect(this)) writeStage = readStage;
  Except_CHECK(readStage >= 0 && writeStage >= 0);
  theta = writeStage - readStage;
  if (Effect_isIndirect(that)) {
    int indirectElapsed = Processor_minTaken(processor) + 1;
    if (Effect_isInTrace(that)) theta = 0;
    else if (indirectElapsed > theta) theta = 0;
    else theta -= indirectElapsed;
  } else theta += extra;
  Except_CHECK(theta >= 0);
  return theta;
}

#line 288 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
Effect_getRegisterWARLatency(Effect this, Effect that, Processor processor)
{
  if (Effect_isIndirect(this)) return 1;
  return Effect_getRegisterExtraWARLatency(this);
}

#line 301 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
Effect_getRegisterWAWLatency(Effect this, Effect that, Processor processor)
{
  if (Effect_isIndirect(this)) return 0;
  if (Effect_isIndirect(that)) return 0;
  return 1;
}

#line 349 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
Effect_getMemoryFlowLatency(Effect this, Effect that, Processor processor)
{
  // TODO
  return 1;
}

#line 362 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
Effect_getMemoryAntiLatency(Effect this, Effect that, Processor processor)
{
  // TODO
  return 1;
}

#line 375 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
Effect_getMemoryOutputLatency(Effect this, Effect that, Processor processor)
{
  // TODO
  return 1;
}

#line 382 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 395 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
EffectSeq
EffectSeq_Ctor(EffectSeq this,
               Memory memory, int32_t estimate)
{
  IStack_Ctor(EffectSeq_ISTACK(this), memory, sizeof(Effect), estimate);
  return this;
}

#line 403 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectSeq_Dtor(EffectSeq this)
{
  IStack_Dtor(EffectSeq_ISTACK(this));
}

#line 409 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
size_t
EffectSeq_Size(Memory memory, int32_t estimate) {
	return sizeof(EffectSeq_);
}

#line 474 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 497 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
EffectTable
EffectTable_Ctor(EffectTable this,
                 Memory memory, Procedure procedure, int32_t estimate)
{
  Temporary_ *dedicated = Procedure_dedicated(procedure);
  IStack_Ctor(EffectTable_OPERATIONS(this),
              memory, sizeof(Operation), estimate);
  HTable_Ctor(EffectTable_EFFECTSEQS(this), memory, estimate);
  IDList_Ctor(EffectTable_SIDEEFFECTS(this), memory);
  *EffectTable__PROCEDURE(this) = procedure;
  *EffectTable__DEDICATED_(this) = dedicated;
  return this;
}

#line 511 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_Dtor(EffectTable this)
{
  Memory memory = EffectTable_MEMORY(this);
  EffectTable_FORBACK_Operation(this, operation) {
    Memory_free(memory, Operation_mainEffects(operation));
  } EffectTable_ENDBACK_Operation;
  IStack_Dtor(EffectTable_OPERATIONS(this));
  HTable_empty(EffectTable_EFFECTSEQS(this),
               (HTableValueRelease)EffectSeq_Dtor);
  HTable_Dtor(EffectTable_EFFECTSEQS(this));
  IDList_Dtor(EffectTable_SIDEEFFECTS(this));
}

#line 525 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
size_t
EffectTable_Size(Memory memory, Procedure procedure, int32_t estimate)
{
  return sizeof(EffectTable_);
}

#line 540 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
#ifndef EffectTable_POOLED
#define EffectTable_POOLED 128
#endif//EffectTable_POOLED

#line 544 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
EffectTable
EffectTable_new(Memory parent, Procedure procedure, int32_t estimate)
{
  Memory memory = Memory_new(parent, EffectTable_POOLED);
  size_t size = EffectTable_Size(memory, procedure, estimate);
  EffectTable this = Memory_alloc(memory, size);
  return EffectTable_Ctor(this, memory, procedure, estimate);
}

#line 557 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
EffectTable
EffectTable_delete(EffectTable this)
{
  if (this != NULL) {
    Memory memory = EffectTable_MEMORY(this);
#ifndef PRO_NDTOR
    EffectTable_Dtor(this);
    Memory_free_(memory, this);
#endif//PRO_NDTOR
    Memory_delete(memory);
  }
  return NULL;
}

#line 628 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
EffectSeq
EffectTable_lookupEffectSeq(EffectTable this, Temporary temporary)
{
  Memory memory = EffectTable_MEMORY(this);
  Temporary dedicated = EffectTable_DEDICATED_(this);
  int32_t estimate = (EffectTable_COUNT(this) + 3)/4;
  size_t size = EffectSeq_Size(memory, estimate);
  HTable effectSeqs = EffectTable_EFFECTSEQS(this);
  int32_t count = HTable_count(effectSeqs);
  HTableKey key = (Temporary_isAssigned(temporary) || Temporary_isDedicated(temporary))?
      dedicated + Temporary_REGISTER(temporary): temporary;
  EffectSeq effectSeq = HTable_lookup_(effectSeqs, key, size);
  Except_CHECK(key != NULL);
  if (count != HTable_count(effectSeqs)) {
    if (Temporary_isControl(temporary)) estimate *= 8;
    if (estimate < 8) estimate = 8;
    EffectSeq_Ctor(effectSeq, memory, estimate);
  }
  return effectSeq;
}

#line 670 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_makeArgEffects(EffectTable this, Operation operation)
{
  EffectFlags flags = EffectFlag_Read;
  Effect_ *argEffects = Operation_argEffects(operation);
  int readCount = Operator_readCount(Operation_operator(operation));
  Operation_ARGUMENTS_FOREACH_Temporary(operation, temporary) {
    Effect effect = argEffects + Operation_ARGUMENTS_INDEX;
    if (Operation_ARGUMENTS_INDEX >= readCount) flags |= EffectFlag_Indirect;
    Effect_Ctor(effect, operation, temporary, flags, Operation_ARGUMENTS_INDEX);
    EffectTable_enterEffect(this, effect);
  } Operation_ARGUMENTS_ENDEACH_Temporary;
}

#line 694 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_makeResEffects(EffectTable this, Operation operation)
{
  EffectFlags flags = EffectFlag_Write;
  Effect_ *resEffects = Operation_resEffects(operation);
  int writeCount = Operator_writeCount(Operation_operator(operation));
  if (!Operation_isPartialDef(operation)) flags |= EffectFlag_Kill;
  Operation_RESULTS_FOREACH_Temporary(operation, temporary) {
    Effect effect = resEffects + Operation_RESULTS_INDEX;
    if (Operation_RESULTS_INDEX >= writeCount) flags |= EffectFlag_Indirect;
    Effect_Ctor(effect, operation, temporary, flags, Operation_RESULTS_INDEX);
    EffectTable_enterEffect(this, effect);
  } Operation_RESULTS_ENDEACH_Temporary;
}

#line 716 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
Effect
EffectTable_makeSideEffect(EffectTable this, Operation operation,
                           Temporary temporary, EffectFlags flags)
{
  Effect effect = (Effect)IDList_push(EffectTable_SIDEEFFECTS(this), sizeof(Effect_));
  Effect_Ctor(effect, operation, temporary, flags, -1);
  EffectTable_enterEffect(this, effect);
  return effect;
}

#line 734 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_enterRegisterSetSideEffects(EffectTable this, Operation operation,
                                        const_RegisterSet registerSet, EffectFlags flags)
{
  Temporary dedicated = EffectTable_DEDICATED_(this);
  RegisterSet_FOREACH_Register(registerSet, registre) {
    Temporary temporary = dedicated + registre;
    if (temporary != NULL) {
      EffectTable_makeSideEffect(this, operation, temporary, flags);
    }
  } RegisterSet_ENDEACH_Register
}

#line 755 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_enterTemporarySetSideEffects(EffectTable this, Operation operation,
                                         TemporarySet temporarySet, EffectFlags flags)
{
  PFA_TRACE && fprintf(PFA_TRACE, "    %d:%s_%d\t%s\t",
                       Operation_identity(operation),
                       Operator_mnemonic(Operation_operator(operation)),
                       Operation_index(operation),
                       flags&EffectFlag_Write? "Write": "Read");
  // Loop over the temporaries and enter their Effect.
  TemporarySet_FOREACH_Temporary(temporarySet, temporary) {
    bool pushEffect = true;
    EffectSeq effectSeq = EffectTable_lookupEffectSeq(this, temporary);
    if (!EffectSeq_isEmpty(effectSeq)) {
      // Avoid duplicate side effects.
      Effect lastEffect = EffectSeq_lastEffect(effectSeq);
      if (Effect_OPERATION(lastEffect) == operation &&
          Effect_isSide(lastEffect)) pushEffect = false;
    }
    if (pushEffect) {
      // Enter the indirect effect as a side-effect.
      EffectTable_makeSideEffect(this, operation, temporary, flags);
    }
  } TemporarySet_ENDEACH_Temporary;
  PFA_TRACE && TemporarySet_pretty(temporarySet, PFA_TRACE) && fprintf(PFA_TRACE, "\n");
}

#line 789 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_enterOperation(EffectTable this, Operation operation, bool sideEffects)
{
  Memory memory = EffectTable_MEMORY(this);
  Operator operator = Operation_operator(operation);
  RegisterSet clobberSet = Operation_clobberSet(operation);
  int32_t effectCount = Operation_argCount(operation) + Operation_resCount(operation);
  struct Temporary_ *dedicated = EffectTable_DEDICATED_(this);
  Operation_setMainEffects(operation,
                           Memory_alloc(memory, sizeof(Effect_)*(effectCount)));
  // Push the Operation.
  IStack_PUSH2(EffectTable_OPERATIONS(this), Operation, operation);
  // Argument effects.
  EffectTable_makeArgEffects(this, operation);
  // Side effects.
  if (sideEffects) {
    // Control effects.
    Temporary control_temporary = dedicated + Register__Control;
    if (Operator_isControl(operator)) {
      EffectTable_makeSideEffect(this, operation, control_temporary, EffectFlag_Write);
    } else {
      EffectTable_makeSideEffect(this, operation, control_temporary, 0);
    }
    // Memory effects.
    if (Operator_isMemory(operator) || Operation_isBarrier(operation)) {
      Temporary memory_temporary = dedicated + Register__Memory;
      if (Operator_isMemoryRead(operator) || Operation_isPrefetch(operation)) {
        EffectTable_makeSideEffect(this, operation, memory_temporary, 0);
      } else if (Operator_isMemoryWrite(operator)) {
        EffectTable_makeSideEffect(this, operation, memory_temporary, EffectFlag_Write);
      } else {
        EffectTable_makeSideEffect(this, operation, memory_temporary, 0);
        EffectTable_makeSideEffect(this, operation, memory_temporary, EffectFlag_Write);
      }
    }
    // Volatile effects.
    if (Operation_isVolatile(operation)) {
      Temporary volatile_temporary = dedicated + Register__Volatile;
      EffectTable_makeSideEffect(this, operation, volatile_temporary, EffectFlag_Write);
    }
    // Clobber effects.
    if (clobberSet != NULL) {
      EffectFlags flags = EffectFlag_Write | EffectFlag_Clobber;
      if (!Operation_isPartialDef(operation)) flags |= EffectFlag_Kill;
      EffectTable_enterRegisterSetSideEffects(this, operation, clobberSet, flags);
    }
  }
  // Result effects.
  EffectTable_makeResEffects(this, operation);
}

#line 844 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
EffectTable_releaseOperation(EffectTable this, Operation operation)
{
  Memory memory = EffectTable_MEMORY(this);
  Memory_free(memory, Operation_mainEffects(operation));
  Operation_setMainEffects(operation, NULL);
}

#line 852 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 860 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
Effect_INIT(void)
{
#ifndef ECL_NTRACE
  ETable_Ctor(EffectFlag_Table, Memory_Root, EffectFlag_Entries);
#endif//ECL_NTRACE
}

#line 874 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
void
Effect_FINI(void)
{
#ifndef ECL_NTRACE
  ETable_Dtor(EffectFlag_Table);
#endif//ECL_NTRACE
}

#line 882 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
#if XCC__C
void
Effect_TEST(void)
{
#include "XFA/Effect_.c"
}

#line 889 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/XFA/Effect.xcc"
int
main(int argc, char **argv)
{
  XFA_Test(Effect_TEST);
  return 0;
}
#endif


#endif//Effect_c_INCLUDED
