#ifndef Effect__h_INCLUDED
#define Effect__h_INCLUDED

//
#include "XFA/Effect.h"

/*
 * EffectTable_enterEffect --	Enter an Effect in this EffectTable.
 */
static inline void
EffectTable_enterEffect(EffectTable this, Effect effect)
{
  Temporary temporary = Effect_TEMPORARY(effect);
  EffectSeq effectSeq = EffectTable_lookupEffectSeq(this, temporary);
  EffectSeq_pushEffect(effectSeq, effect);
}

/*
 * EffectTable_makeArgEffects --	Make the argument Effect(s) of Operation and enter them.
 *
 * In case there are extra arguments beyond the number given by Operation
 * operator, flag the Effect(s) as indirect as these correspond to CALL and
 * RETURN arguments that have been made explicit.
 */
void
EffectTable_makeArgEffects(EffectTable this, Operation operation);

/*
 * EffectTable_makeResEffects --	Make the result Effect(s) of Operation and enter them.
 *
 * In case there are extra results beyond the number given by Operation
 * operator, flag the Effect(s) as indirect as these correspond to CALL and
 * RETURN arguments that have been made explicit.
 */
void
EffectTable_makeResEffects(EffectTable this, Operation operation);

/*
 * EffectTable_makeSideEffect --	Make a new side Effect for an Operation and enter it.
 */
Effect
EffectTable_makeSideEffect(EffectTable this, Operation operation,
                           Temporary temporary, EffectFlags flags);

/*
 * Effect_INIT --	Initialize Effect.
 */
void
Effect_INIT(void);

/*
 * Effect_FINI --	Finalize Effect.
 */
void
Effect_FINI(void);


#endif//Effect__h_INCLUDED
