#ifndef NativeBlock__h_INCLUDED
#define NativeBlock__h_INCLUDED

//
#include "AIR/NativeBlock.h"

/*
 * NativeBlock_assumeScheduled --	Assume this NativeBlock is scheduled.
 *
 * Require:	This NativeBlock has valid Instruction_startDate(s).
 * Return:	This NativeBlock updated NativeBlock_lastDate.
 */
void
NativeBlock_assumeScheduled(NativeBlock this);

/*
 * NativeBlock_assumeUnscheduled --	Assume this NativeBlock is not scheduled.
 */
void
NativeBlock_assumeUnscheduled(NativeBlock this);

/*
 * NativeBlock_bundleInstructions --	Bundle Instructions in this NativeBlock.
 *
 * Bundles contain Instruction(s) with identical non-negative start dates.
 */
void
NativeBlock_bundleInstructions(NativeBlock this, PlatformUInt nativePC);

/*
 * NativeBlock_getCodeSize --	Get the size required to encode this NativeBlock.
 *
 * =nativePC=:	A nativePC with same alignment as that will be used for encoding.
 * Return:	The size needed to encode this NativeBlock.
 */
PlatformUInt
NativeBlock_getCodeSize(NativeBlock this, PlatformUInt nativePC);

/*
 * NativeBlock_encode --	Encode this NativeBlock in the given buffer.
 */
void
NativeBlock_encode(NativeBlock this, PlatformUInt nativePC, void *buffer,
                   InstructionRelocationFixup instruction_fixup_function,
                   LiteralRelocationFixup literal_fixup_function,
                   void* user_data);

/*
 * NativeBlock_INIT --	Initialize the NativeBlock module.
 */
#ifndef ECL_NTRACE
void
NativeBlock_INIT(void);
#else//ECL_NTRACE
#define NativeBlock_INIT()
#endif//ECL_NTRACE

/*
 * NativeBlock_FINI --	Finalize the NativeBlock module.
 */
#ifndef ECL_NTRACE
void
NativeBlock_FINI(void);
#else//ECL_NTRACE
#define NativeBlock_FINI()
#endif//ECL_NTRACE

/*
  st200/NativeBlock.xcc -- Functions to create new Instruction(s).
//
  Automatically generated from the Machine Description System (MDS).
*/

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_src1_src2(NativeBlock this, Instance instance, Register dest, Register src1, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_src1_isrc2(NativeBlock this, Instance instance, Register idest, Register src1, Constant isrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_src1_xsrc2(NativeBlock this, Instance instance, Register idest, Register src1, Constant xsrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_bdest_src1_src2_scond(NativeBlock this, Instance instance, Register dest, Register bdest, Register src1, Register src2, Register scond);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest_src1_src2(NativeBlock this, Instance instance, Register bdest, Register src1, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_ibdest_src1_isrc2(NativeBlock this, Instance instance, Register ibdest, Register src1, Constant isrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_ibdest_src1_xsrc2(NativeBlock this, Instance instance, Register ibdest, Register src1, Constant xsrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_bcond_btarg(NativeBlock this, Instance instance, Register bcond, Constant btarg, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction__UNDEF(NativeBlock this, Instance instance);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_src1(NativeBlock this, Instance instance, Register idest, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_btarg(NativeBlock this, Instance instance, Constant btarg, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_imm(NativeBlock this, Instance instance, Constant imm);

//
extern Instruction
NativeBlock_makeInstruction_st200_nlidest_isrc2_src1(NativeBlock this, Instance instance, Register nlidest, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_nlidest_xsrc2_src1(NativeBlock this, Instance instance, Register nlidest, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_isrc2_src1(NativeBlock this, Instance instance, Register idest, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_xsrc2_src1(NativeBlock this, Instance instance, Register idest, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_nldest_src1_src2(NativeBlock this, Instance instance, Register nldest, Register src1, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_nlidest_src1_isrc2(NativeBlock this, Instance instance, Register nlidest, Register src1, Constant isrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_nlidest_src1_xsrc2(NativeBlock this, Instance instance, Register nlidest, Register src1, Constant xsrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_isrc2_src1(NativeBlock this, Instance instance, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_xsrc2_src1(NativeBlock this, Instance instance, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_sbrknum(NativeBlock this, Instance instance, Constant sbrknum);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_scond_src1_src2(NativeBlock this, Instance instance, Register dest, Register scond, Register src1, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_scond_src1_isrc2(NativeBlock this, Instance instance, Register idest, Register scond, Register src1, Constant isrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_scond_src1_xsrc2(NativeBlock this, Instance instance, Register idest, Register scond, Register src1, Constant xsrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_isrc2_src1_src2(NativeBlock this, Instance instance, Constant isrc2, Register src1, Register src2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_xsrc2_src1_src2(NativeBlock this, Instance instance, Constant xsrc2, Register src1, Register src2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_src2_src1(NativeBlock this, Instance instance, Register dest, Register src2, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_scond(NativeBlock this, Instance instance, Register idest, Register scond);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest_src1(NativeBlock this, Instance instance, Register bdest, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_src2(NativeBlock this, Instance instance, Register dest, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_isrc2(NativeBlock this, Instance instance, Register idest, Constant isrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_xsrc2(NativeBlock this, Instance instance, Register idest, Constant xsrc2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_src2(NativeBlock this, Instance instance, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest2_src1_src2(NativeBlock this, Instance instance, Register bdest2, Register src1, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_src1(NativeBlock this, Instance instance, Register dest, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_nldest_scond_src1_src2(NativeBlock this, Instance instance, Register nldest, Register scond, Register src1, Register src2);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest2_src1(NativeBlock this, Instance instance, Register bdest2, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_scond(NativeBlock this, Instance instance, Register dest, Register scond);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest2_bsrc1_bsrc2(NativeBlock this, Instance instance, Register bdest2, Register bsrc1, Register bsrc2);

//
extern Instruction
NativeBlock_makeInstruction_st200_nldest_src1(NativeBlock this, Instance instance, Register nldest, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_brknum(NativeBlock this, Instance instance, Constant brknum);

//
extern Instruction
NativeBlock_makeInstruction_st200_nlidest_pcond_isrc2_src1(NativeBlock this, Instance instance, Register nlidest, Register pcond, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_nlidest_pcond_xsrc2_src1(NativeBlock this, Instance instance, Register nlidest, Register pcond, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idestp_isrc2_src1(NativeBlock this, Instance instance, Register idestp, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idestp_xsrc2_src1(NativeBlock this, Instance instance, Register idestp, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idestp_pcond_isrc2_src1(NativeBlock this, Instance instance, Register idestp, Register pcond, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idestp_pcond_xsrc2_src1(NativeBlock this, Instance instance, Register idestp, Register pcond, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_pcond_isrc2_src1(NativeBlock this, Instance instance, Register idest, Register pcond, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_idest_pcond_xsrc2_src1(NativeBlock this, Instance instance, Register idest, Register pcond, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest2_bsrc1(NativeBlock this, Instance instance, Register bdest2, Register bsrc1);

//
extern Instruction
NativeBlock_makeInstruction_st200_pcond_isrc2_src1(NativeBlock this, Instance instance, Register pcond, Constant isrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_pcond_xsrc2_src1(NativeBlock this, Instance instance, Register pcond, Constant xsrc2, Register src1, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_isrc2_src1_pcond_src2(NativeBlock this, Instance instance, Constant isrc2, Register src1, Register pcond, Register src2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_xsrc2_src1_pcond_src2(NativeBlock this, Instance instance, Constant xsrc2, Register src1, Register pcond, Register src2, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_isrc2_src1_src2p(NativeBlock this, Instance instance, Constant isrc2, Register src1, Register src2p, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_xsrc2_src1_src2p(NativeBlock this, Instance instance, Constant xsrc2, Register src1, Register src2p, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_isrc2_src1_pcond_src2p(NativeBlock this, Instance instance, Constant isrc2, Register src1, Register pcond, Register src2p, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_xsrc2_src1_pcond_src2p(NativeBlock this, Instance instance, Constant xsrc2, Register src1, Register pcond, Register src2p, Relocation relocation);

//
extern Instruction
NativeBlock_makeInstruction_st200_bdest2_src2_src1(NativeBlock this, Instance instance, Register bdest2, Register src2, Register src1);

//
extern Instruction
NativeBlock_makeInstruction_st200_dest_scond_src2_src1(NativeBlock this, Instance instance, Register dest, Register scond, Register src2, Register src1);


#endif//NativeBlock__h_INCLUDED
