#ifndef CAL_h_INCLUDED
#define CAL_h_INCLUDED

/*
 * !!!!	CAL.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com)
 *
 * Copyright 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CDT.h"
#include "CCL/CCL.h"

/*
 * The CAL (Combinatorial Algorithm Library) contains the following modules:
 * *	<DGraph DGraph.html> Directed Graph implementation.
 * *	<Dominators Dominators.html> Dominators on DGraph (Lengauer Tarjan)
 * *	<Hierarchy Hierarchy.html> Hierarchy of SCCs in a DGraph (Bourdoncle).
 * *	<Network Network.html> Network simplex algorithms.
 * *	<LProgram LProgram.html> Linear Program wrapper for CPLEX and GLPK.
 */

/*
 * Include all the exported CAL headers.
 */
#include "CAL/DGraph.h"
#include "CAL/Dominators.h"
#include "CAL/Hierarchy.h"
#include "CAL/Network.h"
#include "CAL/LProgram.h"

/*
 * CAL_TRACE --	File handle used to print the CAL traces.
 */
extern FILE *CAL_TRACE;
#ifdef CDT_NTRACE
#define CAL_TRACE NULL
#endif//CDT_NTRACE

/*
 * IPF_TRACE --	File handle to output Integer Programming Formulations.
 */
extern FILE *IPF_TRACE;
#ifdef CDT_NTRACE
#define IPF_TRACE NULL
#endif//CDT_NTRACE

/*
 * CAL_INIT --	Initialize the CAL library.
 */
void
CAL_INIT(void);

/*
 * CAL_FINI --	Finalize CAL library.
 */
void
CAL_FINI(void);


#endif//CAL_h_INCLUDED
