#ifndef LProgram__h_INCLUDED
#define LProgram__h_INCLUDED

//
#include "CAL/LProgram.h"
#include <time.h>
#if CPLEX_ENABLED
#include <math.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <ilcplex/cplex.h>
#else//!CPLEX_ENABLED
#if GLPK_ENABLED
#include <time.h>
#include "glpk.h"
#endif//GLPK_ENABLED
#endif//CPLEX_ENABLED

//
struct LProgramConstraint_ {
  //@args	Memory memory, int32_t count, LProgramConstraintType type,
  //@args	LProgramValue bound, const char *name
  IStack_ TERMS_;	// The LProgramTerms of this LProgramConstraint.
  //@access TERMS	LProgramConstraint__TERMS_(this)
  //@access TERMCOUNT	(IStack_usedSize(LProgramConstraint_TERMS(this))/sizeof(LProgramTerm_))
  short_LProgramConstraintType TYPE;	// The type of this LProgramConstraint.
  //@access isLE	(LProgramConstraint_TYPE(this) == LProgramConstraintType_LE)
  //@access isGE	(LProgramConstraint_TYPE(this) == LProgramConstraintType_GE)
  //@access isEQ	(LProgramConstraint_TYPE(this) == LProgramConstraintType_EQ)
  LProgramValue BOUND;	// The bound (RHS) of this LProgramConstraint.
  const char *NAME;	// User name for this LProgramConstraint.
};

extern LProgramConstraint
LProgramConstraint_Ctor(LProgramConstraint this, Memory memory, int32_t count, LProgramConstraintType type, LProgramValue bound, const char *name);

extern LProgramConstraint
LProgramConstraint_Copy(LProgramConstraint this, const_LProgramConstraint that);

extern void
LProgramConstraint_Dtor(LProgramConstraint this);

extern size_t
LProgramConstraint_Size(Memory memory, int32_t count, LProgramConstraintType type, LProgramValue bound, const char *name);

#define LProgramConstraint_TERMS_(this) ((this)->TERMS_)
#define LProgramConstraint__TERMS_(this) (&(this)->TERMS_)
#define LProgramConstraint_TYPE(this) ((this)->TYPE)
#define LProgramConstraint__TYPE(this) (&(this)->TYPE)
#define LProgramConstraint_BOUND(this) ((this)->BOUND)
#define LProgramConstraint__BOUND(this) (&(this)->BOUND)
#define LProgramConstraint_NAME(this) ((this)->NAME)
#define LProgramConstraint__NAME(this) (&(this)->NAME)
#define LProgramConstraint_isGE(this) ((LProgramConstraint_TYPE(this) == LProgramConstraintType_GE))
#define LProgramConstraint_isEQ(this) ((LProgramConstraint_TYPE(this) == LProgramConstraintType_EQ))
#define LProgramConstraint_isLE(this) ((LProgramConstraint_TYPE(this) == LProgramConstraintType_LE))
#define LProgramConstraint_TERMCOUNT(this) ((IStack_usedSize(LProgramConstraint_TERMS(this))/sizeof(LProgramTerm_)))
#define LProgramConstraint_TERMS(this) (LProgramConstraint__TERMS_(this))


//
struct LProgram_ {
  //@args	Memory memory, LProgramType type, double timeOut
  HTable_ VARIABLEMAP_;
  //@access VARIABLEMAP	LProgram__VARIABLEMAP_(this)
  //@access MEMORY	HTable_memory(LProgram_VARIABLEMAP(this))
  IDList_ VARIABLES_;
  //@access VARIABLES	LProgram__VARIABLES_(this)
  //@access VARIABLECOUNT	IDList_count(LProgram_VARIABLES(this))
  IDList_ CONSTRAINTS_;
  //@access CONSTRAINTS	LProgram__CONSTRAINTS_(this)
  //@access CONSTRAINTCOUNT	IDList_count(LProgram_CONSTRAINTS(this))
  double TIMEOUT;
  double TIMESPENT;
  LProgramValue SOLUTION;
  short_LProgramType TYPE;
  //@access isFeasibility	(LProgram_TYPE(this) == LProgramType_Feasibility)
  //@access isMinimize	(LProgram_TYPE(this) == LProgramType_Minimize)
  //@access isMaximize	(LProgram_TYPE(this) == LProgramType_Maximize)
  short_LProgramStatus STATUS;
#if CPLEX_ENABLED
  struct cpxenv *CPXENV;
  struct cpxlp *CPXLP;
#else//!CPLEX_ENABLED
#if GLPK_ENABLED
  glp_prob *GLP;
#endif//GLPK_ENABLED
#endif//CPLEX_ENABLED
};

extern LProgram
LProgram_Ctor(LProgram this, Memory memory, LProgramType type, double timeOut);

extern LProgram
LProgram_Copy(LProgram this, const_LProgram that);

extern void
LProgram_Dtor(LProgram this);

extern size_t
LProgram_Size(Memory memory, LProgramType type, double timeOut);

#define LProgram_VARIABLEMAP_(this) ((this)->VARIABLEMAP_)
#define LProgram__VARIABLEMAP_(this) (&(this)->VARIABLEMAP_)
#define LProgram_VARIABLES_(this) ((this)->VARIABLES_)
#define LProgram__VARIABLES_(this) (&(this)->VARIABLES_)
#define LProgram_CONSTRAINTS_(this) ((this)->CONSTRAINTS_)
#define LProgram__CONSTRAINTS_(this) (&(this)->CONSTRAINTS_)
#define LProgram_TIMEOUT(this) ((this)->TIMEOUT)
#define LProgram__TIMEOUT(this) (&(this)->TIMEOUT)
#define LProgram_TIMESPENT(this) ((this)->TIMESPENT)
#define LProgram__TIMESPENT(this) (&(this)->TIMESPENT)
#define LProgram_SOLUTION(this) ((this)->SOLUTION)
#define LProgram__SOLUTION(this) (&(this)->SOLUTION)
#define LProgram_TYPE(this) ((this)->TYPE)
#define LProgram__TYPE(this) (&(this)->TYPE)
#define LProgram_STATUS(this) ((this)->STATUS)
#define LProgram__STATUS(this) (&(this)->STATUS)
#define LProgram_CPXENV(this) ((this)->CPXENV)
#define LProgram__CPXENV(this) (&(this)->CPXENV)
#define LProgram_CPXLP(this) ((this)->CPXLP)
#define LProgram__CPXLP(this) (&(this)->CPXLP)
#define LProgram_GLP(this) ((this)->GLP)
#define LProgram__GLP(this) (&(this)->GLP)
#define LProgram_CONSTRAINTCOUNT(this) (IDList_count(LProgram_CONSTRAINTS(this)))
#define LProgram_isMaximize(this) ((LProgram_TYPE(this) == LProgramType_Maximize))
#define LProgram_isFeasibility(this) ((LProgram_TYPE(this) == LProgramType_Feasibility))
#define LProgram_VARIABLECOUNT(this) (IDList_count(LProgram_VARIABLES(this)))
#define LProgram_CONSTRAINTS(this) (LProgram__CONSTRAINTS_(this))
#define LProgram_VARIABLEMAP(this) (LProgram__VARIABLEMAP_(this))
#define LProgram_MEMORY(this) (HTable_memory(LProgram_VARIABLEMAP(this)))
#define LProgram_VARIABLES(this) (LProgram__VARIABLES_(this))
#define LProgram_isMinimize(this) ((LProgram_TYPE(this) == LProgramType_Minimize))


//
bool
LProgram_enterVariables(LProgram this);

//
bool
LProgram_enterConstraints(LProgram this);

//
void
LProgram_setStatusCPLEX(LProgram this);

/*
 * LProgram_INIT --	Initialize the LProgram module.
 */
#define LProgram_INIT()

/*
 * LProgram_FINI --	Finalize the LProgram module.
 */
#define LProgram_FINI()


#endif//LProgram__h_INCLUDED
