#ifndef CGO_h_INCLUDED
#define CGO_h_INCLUDED

/*
 * !!!!	CGO.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CCL.h"
#include "ECL.h"
#include "MDS/MDS.h"
#include "AIR/AIR.h"
#include "LIR/LIR.h"
#include "PFA/PFA.h"

/*
 * The CGO (Code Generator Optimizations) library contains the following modules:
 * *	<ControlFlow ControlFlow.html> Control Flow optimizations.
 * *	<Predicator Predicator.html> Predication optimizations.
 * *	<Pipeliner Pipeliner.html> Instruction Pipeliner (to do).
 * *	<Allocator Allocator.html> Register Allocator base class (to do).
 * *	<ELinearScan ELinearScan.html> Extended Linear Scan allocator.
 * *	<DomTreeScan DomTreeScan.html> Dominance Tree Scan allocator.
 * *	<Scoreboarder Scoreboarder.html> Scoreboard scheduler.
 */

//
#include "CGO/ControlFlow.h"
#include "CGO/Predicator.h"
#include "CGO/Pipeliner.h"
#include "CGO/Allocator.h"
#include "CGO/ELinearScan.h"
#include "CGO/DomTreeScan.h"
#include "CGO/Scoreboarder.h"
#include "CGO/Spill.h"
#include "CGO/SSAReconstruct.h"
#include "CGO/SSASpill.h"
#include "CGO/ParallelCopy.h"

/*
 * CGO_TRACE --	File handle used to print the CGO traces.
 */
extern FILE *CGO_TRACE;
#ifdef ECL_NTRACE
#define CGO_TRACE NULL
#endif//ECL_NTRACE

/*
 * Variables used to control CGO in non-release mode.
 */

/*
 * CGO_INIT --	Initialize the CGO library.
 */
void
CGO_INIT(void);

/*
 * CGO_FINI --	Finalize the CGO library.
 */
void
CGO_FINI(void);


#endif//CGO_h_INCLUDED
