#ifndef ParallelCopy__h_INCLUDED
#define ParallelCopy__h_INCLUDED

//
#include "CGO/ParallelCopy.h"

//
extern bool
ParallelCopy_resolveDuplicationAtBBLevel(BasicBlock basicBlock, RegFile regFile);

//
struct PermutationNode_ {
  //@args Register color, Memory memory
  Register COLOR;
  IDList_ COPIES_;
  //@access COPIES PermutationNode__COPIES_(this)
  //@access MEMORY IDList_memory(PermutationNode_COPIES(this))
  struct PermutationNode_ *PARENT;
  int32_t CYCLEIDX;
  //@access isInCycle PermutationNode_CYCLEIDX(this) != -1
};
typedef struct PermutationNode_ PermutationNode_, *PermutationNode;
typedef const struct PermutationNode_ *const_PermutationNode;
typedef struct PermutationNode_ * restrict_PermutationNode;

extern PermutationNode
PermutationNode_Ctor(PermutationNode this, Register color, Memory memory);

extern PermutationNode
PermutationNode_Copy(PermutationNode this, const_PermutationNode that);

extern void
PermutationNode_Dtor(PermutationNode this);

extern size_t
PermutationNode_Size(Register color, Memory memory);

#define PermutationNode_COLOR(this) ((this)->COLOR)
#define PermutationNode__COLOR(this) (&(this)->COLOR)
#define PermutationNode_COPIES_(this) ((this)->COPIES_)
#define PermutationNode__COPIES_(this) (&(this)->COPIES_)
#define PermutationNode_PARENT(this) ((this)->PARENT)
#define PermutationNode__PARENT(this) (&(this)->PARENT)
#define PermutationNode_CYCLEIDX(this) ((this)->CYCLEIDX)
#define PermutationNode__CYCLEIDX(this) (&(this)->CYCLEIDX)
#define PermutationNode_isInCycle(this) (PermutationNode_CYCLEIDX(this) != -1)
#define PermutationNode_COPIES(this) (PermutationNode__COPIES_(this))
#define PermutationNode_MEMORY(this) (IDList_memory(PermutationNode_COPIES(this)))


//
extern bool
ParallelCopy_resolveDuplicationOnPhi(RegisterSet_ availableReg[][1],
                                     Operation phi, RegFile regFile);

//
// AvailableReg contains the registers which are avaible to solve duplication
// on edge.
// pcopy is the parallel copy with duplication that we will solve on edge
extern bool
ParallelCopy_resolveDuplicationOnParCopyOnEdge(const_RegisterSet availableReg,
                                               Operation pcopy,
                                               BasicBlockEdge edge);

//
extern bool
ParallelCopy_resolveDuplicationOnParCopy(const_RegisterSet availableReg,
                                         Operation parCopy);

//
extern bool
ParallelCopy_resolveDuplicationOnParCopy_(const_RegisterSet availableReg,
                                          Operation parCopy,
                                          BasicBlock basicBlock,
                                          Operation point, bool before);

//
struct ReachabilityNode_ {
  //@args Memory memory, int32_t size
  RegisterSet_ REACHES_;
  //@access REACHES ReachabilityNode__REACHES_(this)
  RegisterSet_ *REACHEDGES;
  Register *CHILDREN;
  int32_t SIZE;
  Memory MEMORY;
};
typedef struct ReachabilityNode_ ReachabilityNode_, *ReachabilityNode;
typedef const struct ReachabilityNode_ *const_ReachabilityNode;
typedef struct ReachabilityNode_ * restrict_ReachabilityNode;

extern ReachabilityNode
ReachabilityNode_Ctor(ReachabilityNode this, Memory memory, int32_t size);

extern ReachabilityNode
ReachabilityNode_Copy(ReachabilityNode this, const_ReachabilityNode that);

extern void
ReachabilityNode_Dtor(ReachabilityNode this);

extern size_t
ReachabilityNode_Size(Memory memory, int32_t size);

#define ReachabilityNode_REACHES_(this) ((this)->REACHES_)
#define ReachabilityNode__REACHES_(this) (&(this)->REACHES_)
#define ReachabilityNode_REACHEDGES(this) ((this)->REACHEDGES)
#define ReachabilityNode__REACHEDGES(this) (&(this)->REACHEDGES)
#define ReachabilityNode_CHILDREN(this) ((this)->CHILDREN)
#define ReachabilityNode__CHILDREN(this) (&(this)->CHILDREN)
#define ReachabilityNode_SIZE(this) ((this)->SIZE)
#define ReachabilityNode__SIZE(this) (&(this)->SIZE)
#define ReachabilityNode_MEMORY(this) ((this)->MEMORY)
#define ReachabilityNode__MEMORY(this) (&(this)->MEMORY)
#define ReachabilityNode_REACHES(this) (ReachabilityNode__REACHES_(this))


/*
 * ParallelCopy_INIT --	Initialize the ParallelCopy module.
 */
void
ParallelCopy_INIT(void);

/*
 * ParallelCopy_FINI --	Finalize the ParallelCopy module.
 */
void
ParallelCopy_FINI(void);


#endif//ParallelCopy__h_INCLUDED
