#ifndef GraphColoring__h_INCLUDED
#define GraphColoring__h_INCLUDED

//
#include "GRA/GraphColoring.h"

//
/*******************************
 * Friend members declarations *
 *******************************/
//@XCC_.h
/**
 * This module variables
 */
int GRA_ssa_split;
int GRA_ssa_nsplit;
int GRA_ssa_nback;
int GRA_ssa_cmove;
int GRA_ssa_best_nsplit;
int GRA_ssa_best_cmove;
int GRA_gcf;
bool GRA_redo_liveness;
float GRA_ssa_cost_split;
float GRA_ssa_cost_nsplit;
float GRA_ssa_cost_nback;
float GRA_ssa_cost_cmove;
float GRA_ssa_cost_best_nsplit;
float GRA_ssa_cost_best_cmove;
int GRA_ssa_move_count;
int GRA_ssa_use_new_rpc;

//
/************************
 * Our graph structures *
 ************************/

//
struct WorkLists_;
typedef struct WorkLists_ WorkLists_, *WorkLists;
typedef const struct WorkLists_ *const_WorkLists;
typedef struct WorkLists_ * restrict_WorkLists;

//
struct GraphColoring_;
typedef struct GraphColoring_ GraphColoring_, *GraphColoring;
typedef const struct GraphColoring_ *const_GraphColoring;
typedef struct GraphColoring_ * restrict_GraphColoring;

//
typedef enum {
  build,
  simplify,
  coalesce,
  freeze,
  selectSpill,
  briggsWins,
  briggsLoses,
  georgeWins,
  georgeLoses,
  inInitial,
  inSimplifyWorklist,
  outSimplifyWorklist,
  inFreezeWorklist,
  outFreezeWorklist,
  inSpillWorklist,
  outSpillWorklist,
  inSpilledNodes,
  inCoalescedNodes,
  inColoredNodes,
  inDiscardedNodes,
  inSelectStack,
  inCoalescedMoves,
  inConstrainedMoves,
  inFrozenMoves,
  inWorklistMoves,
  outWorklistMoves,
  inActiveMoves,
  outActiveMoves,
  Stat__
} Stat;

//
/**
 * The structure of the interference nodes lists.
 */
struct WorkLists_ {
  //! Machine registers, preassigned to a color.
  //@args Memory memory
  IDList_ precolored__;
  //@access precolored (IDList)WorkLists__precolored__(this)
  //! Temporary registers, not precolored and not yet processed.
  IDList_ initial__;
  //@access initial (IDList)WorkLists__initial__(this)
  //! Nodes marked for spilling during this round; initially empty.
  IDList_ spilledNodes__;
  //@access spilledNodes (IDList)WorkLists__spilledNodes__(this)
  //! Registers that have been coalesced; when u<-v is coalesced, v is added to this set and u put back on some work-list (or vice versa)
  IDList_ coalescedNodes__;
  //@access coalescedNodes (IDList)WorkLists__coalescedNodes__(this)
  //! Nodes successfully colored.
  IDList_ coloredNodes__;
  //@access coloredNodes (IDList)WorkLists__coloredNodes__(this)
  //! Nodes discarded for further passes, spilled nodes go there after each spill pass.
  IDList_ discardedNodes__;
  //@access discardedNodes (IDList)WorkLists__discardedNodes__(this)
  //! High-degree nodes.
  IDList_ spillWorklist__;
  //@access spillWorklist (IDList)WorkLists__spillWorklist__(this)
  //! Low-degree move-related nodes.
  IDList_ freezeWorklist__;
  //@access freezeWorklist (IDList)WorkLists__freezeWorklist__(this)
  //! Low-degree non-move-related nodes.
  IDList_ simplifyWorklist__;
  //@access simplifyWorklist (IDList)WorkLists__simplifyWorklist__(this)
  //! Stack containing temporaries removed from the graph.
  IDList_ selectStack__;
  //@access selectStack (IDList)WorkLists__selectStack__(this)
  //! Moves enabled for possible coalescing.
  IDList_ worklistMoves__;
  //@access worklistMoves (IDList)WorkLists__worklistMoves__(this)
  //! Moves not yet ready for coalescing.
  IDList_ activeMoves__;
  //@access activeMoves (IDList)WorkLists__activeMoves__(this)
  //! Moves that have been coalesced.
  IDList_ coalescedMoves__;
  //@access coalescedMoves (IDList)WorkLists__coalescedMoves__(this)
  //! Moves whose source ant target interfere.
  IDList_ constrainedMoves__;
  //@access constrainedMoves (IDList)WorkLists__constrainedMoves__(this)
  //! Moves that will no longer be considered for coalescing.
  IDList_ frozenMoves__;
  //@access frozenMoves (IDList)WorkLists__frozenMoves__(this)
};

extern WorkLists
WorkLists_Ctor(WorkLists this, Memory memory);

extern WorkLists
WorkLists_Copy(WorkLists this, const_WorkLists that);

extern void
WorkLists_Dtor(WorkLists this);

extern size_t
WorkLists_Size(Memory memory);

#define WorkLists_precolored__(this) ((this)->precolored__)
#define WorkLists__precolored__(this) (&(this)->precolored__)
#define WorkLists_initial__(this) ((this)->initial__)
#define WorkLists__initial__(this) (&(this)->initial__)
#define WorkLists_spilledNodes__(this) ((this)->spilledNodes__)
#define WorkLists__spilledNodes__(this) (&(this)->spilledNodes__)
#define WorkLists_coalescedNodes__(this) ((this)->coalescedNodes__)
#define WorkLists__coalescedNodes__(this) (&(this)->coalescedNodes__)
#define WorkLists_coloredNodes__(this) ((this)->coloredNodes__)
#define WorkLists__coloredNodes__(this) (&(this)->coloredNodes__)
#define WorkLists_discardedNodes__(this) ((this)->discardedNodes__)
#define WorkLists__discardedNodes__(this) (&(this)->discardedNodes__)
#define WorkLists_spillWorklist__(this) ((this)->spillWorklist__)
#define WorkLists__spillWorklist__(this) (&(this)->spillWorklist__)
#define WorkLists_freezeWorklist__(this) ((this)->freezeWorklist__)
#define WorkLists__freezeWorklist__(this) (&(this)->freezeWorklist__)
#define WorkLists_simplifyWorklist__(this) ((this)->simplifyWorklist__)
#define WorkLists__simplifyWorklist__(this) (&(this)->simplifyWorklist__)
#define WorkLists_selectStack__(this) ((this)->selectStack__)
#define WorkLists__selectStack__(this) (&(this)->selectStack__)
#define WorkLists_worklistMoves__(this) ((this)->worklistMoves__)
#define WorkLists__worklistMoves__(this) (&(this)->worklistMoves__)
#define WorkLists_activeMoves__(this) ((this)->activeMoves__)
#define WorkLists__activeMoves__(this) (&(this)->activeMoves__)
#define WorkLists_coalescedMoves__(this) ((this)->coalescedMoves__)
#define WorkLists__coalescedMoves__(this) (&(this)->coalescedMoves__)
#define WorkLists_constrainedMoves__(this) ((this)->constrainedMoves__)
#define WorkLists__constrainedMoves__(this) (&(this)->constrainedMoves__)
#define WorkLists_frozenMoves__(this) ((this)->frozenMoves__)
#define WorkLists__frozenMoves__(this) (&(this)->frozenMoves__)
#define WorkLists_freezeWorklist(this) ((IDList)WorkLists__freezeWorklist__(this))
#define WorkLists_activeMoves(this) ((IDList)WorkLists__activeMoves__(this))
#define WorkLists_selectStack(this) ((IDList)WorkLists__selectStack__(this))
#define WorkLists_coloredNodes(this) ((IDList)WorkLists__coloredNodes__(this))
#define WorkLists_frozenMoves(this) ((IDList)WorkLists__frozenMoves__(this))
#define WorkLists_worklistMoves(this) ((IDList)WorkLists__worklistMoves__(this))
#define WorkLists_constrainedMoves(this) ((IDList)WorkLists__constrainedMoves__(this))
#define WorkLists_coalescedMoves(this) ((IDList)WorkLists__coalescedMoves__(this))
#define WorkLists_discardedNodes(this) ((IDList)WorkLists__discardedNodes__(this))
#define WorkLists_spilledNodes(this) ((IDList)WorkLists__spilledNodes__(this))
#define WorkLists_coalescedNodes(this) ((IDList)WorkLists__coalescedNodes__(this))
#define WorkLists_initial(this) ((IDList)WorkLists__initial__(this))
#define WorkLists_precolored(this) ((IDList)WorkLists__precolored__(this))
#define WorkLists_spillWorklist(this) ((IDList)WorkLists__spillWorklist__(this))
#define WorkLists_simplifyWorklist(this) ((IDList)WorkLists__simplifyWorklist__(this))


//
/**
 * Structure of a graph coloring object.
 * A Graph coloring object is responsible for the allocation of a code region and one or several reg class
 * into the code region.
 * This object has an associated code region, and the associated reg class to allocate.
 * It maintains the temporary  set for the coderegion/regclass, the interference graph and the work lists
 * of nodes and moves.
 */
struct GraphColoring_ {
  //@args	Memory memory
  Memory MEMORY;
  //@access memory GraphColoring_MEMORY(this)
  CodeRegion CODEREGION; //!< Associated code region
  //@access codeRegion GraphColoring_CODEREGION(this)
  //@access procedure CodeRegion_procedure(GraphColoring_CODEREGION(this))
  //@access CONVENTION	CodeRegion_convention(GraphColoring_CODEREGION(this))
  Liveness LIVENESS;	// This CodeRegion Liveness.
  RegFile REGFILE; //!< Current register class
  //@access regFile GraphColoring_REGFILE(this)
  //@mutate regFile GraphColoring__REGFILE(this)
  int PASSNUM; //!< Current pass number for the algorithm
  //@access passNum GraphColoring_PASSNUM(this)
  TemporaryTable_ TMPTABLE_; //!< Set of temporaries to allocate
  //@access tmpTable GraphColoring__TMPTABLE_(this)
  int INITIALTMPCOUNT; //!< Initial count of register temporary
  //@access initialTmpCount GraphColoring_INITIALTMPCOUNT(this)
  InterfGraph_ INTERFGRAPH_; //!< The interference graph
  //@access interfGraph GraphColoring__INTERFGRAPH_(this)
  WorkLists_ WORKLISTS_; //!< The work lists
  //@access workLists GraphColoring__WORKLISTS_(this)
  Temporary *SPILLS; //!< The spill representative
  int OPERATIONCOUNTER;
  //@access operationCounter GraphColoring_operationCounter(this)
  uint32_t FLAGS;
  //@access flags GraphColoring_FLAGS(this)
  //@access tmpCount TemporaryTable_COUNT(GraphColoring_TMPTABLE(this))
};

extern GraphColoring
GraphColoring_Ctor(GraphColoring this, Memory memory);

extern GraphColoring
GraphColoring_Copy(GraphColoring this, const_GraphColoring that);

extern void
GraphColoring_Dtor(GraphColoring this);

extern size_t
GraphColoring_Size(Memory memory);

#define GraphColoring_MEMORY(this) ((this)->MEMORY)
#define GraphColoring__MEMORY(this) (&(this)->MEMORY)
#define GraphColoring_CODEREGION(this) ((this)->CODEREGION)
#define GraphColoring__CODEREGION(this) (&(this)->CODEREGION)
#define GraphColoring_LIVENESS(this) ((this)->LIVENESS)
#define GraphColoring__LIVENESS(this) (&(this)->LIVENESS)
#define GraphColoring_REGFILE(this) ((this)->REGFILE)
#define GraphColoring__REGFILE(this) (&(this)->REGFILE)
#define GraphColoring_PASSNUM(this) ((this)->PASSNUM)
#define GraphColoring__PASSNUM(this) (&(this)->PASSNUM)
#define GraphColoring_TMPTABLE_(this) ((this)->TMPTABLE_)
#define GraphColoring__TMPTABLE_(this) (&(this)->TMPTABLE_)
#define GraphColoring_INITIALTMPCOUNT(this) ((this)->INITIALTMPCOUNT)
#define GraphColoring__INITIALTMPCOUNT(this) (&(this)->INITIALTMPCOUNT)
#define GraphColoring_INTERFGRAPH_(this) ((this)->INTERFGRAPH_)
#define GraphColoring__INTERFGRAPH_(this) (&(this)->INTERFGRAPH_)
#define GraphColoring_WORKLISTS_(this) ((this)->WORKLISTS_)
#define GraphColoring__WORKLISTS_(this) (&(this)->WORKLISTS_)
#define GraphColoring_SPILLS(this) ((this)->SPILLS)
#define GraphColoring__SPILLS(this) (&(this)->SPILLS)
#define GraphColoring_OPERATIONCOUNTER(this) ((this)->OPERATIONCOUNTER)
#define GraphColoring__OPERATIONCOUNTER(this) (&(this)->OPERATIONCOUNTER)
#define GraphColoring_FLAGS(this) ((this)->FLAGS)
#define GraphColoring__FLAGS(this) (&(this)->FLAGS)
#define GraphColoring_workLists(this) (GraphColoring__WORKLISTS_(this))
#define GraphColoring_codeRegion(this) (GraphColoring_CODEREGION(this))
#define GraphColoring_flags(this) (GraphColoring_FLAGS(this))
#define GraphColoring_initialTmpCount(this) (GraphColoring_INITIALTMPCOUNT(this))
#define GraphColoring_tmpCount(this) (TemporaryTable_COUNT(GraphColoring_TMPTABLE(this)))
#define GraphColoring_passNum(this) (GraphColoring_PASSNUM(this))
#define GraphColoring_regFile(this) (GraphColoring_REGFILE(this))
#define GraphColoring_CONVENTION(this) (CodeRegion_convention(GraphColoring_CODEREGION(this)))
#define GraphColoring_memory(this) (GraphColoring_MEMORY(this))
#define GraphColoring_interfGraph(this) (GraphColoring__INTERFGRAPH_(this))
#define GraphColoring_tmpTable(this) (GraphColoring__TMPTABLE_(this))
#define GraphColoring_procedure(this) (CodeRegion_procedure(GraphColoring_CODEREGION(this)))
#define GraphColoring_operationCounter(this) (GraphColoring_operationCounter(this))
#define GraphColoring__regFile(this) (GraphColoring__REGFILE(this))


//
/*
 * GraphColoring flags.
 */
#define GRA_GCF_CHAINTIN_SPILL_HEURISTIC 1
#define GRA_GCF_SPILL_PHI_USES 2
#define GRA_GCF_SPILL_PHI_DEFS 4
#define GRA_GCF_SSA 8
#define GRA_GCF_NO_PHI_COALESCING 16
#define GRA_GCF_SPILL_PHI_ON_DEF 32
#define GRA_GCF_SPILL_PHI_ON_RESPILL 64
#define GRA_GCF_NO_MOVE_COALESCING 128
#define GRA_GCF_NO_BIASED 256
#define GRA_GCF_SPILL_POTENTIAL 512
#define GRA_GCF_USELESS_INITIAL 1024
#define GRA_GCF_ADD_MOVE_INTERFERENCE 2048

//
/*
 * Default flags set.
 */
#define GRA_GCF_DEFAULT (\
GRA_GCF_CHAINTIN_SPILL_HEURISTIC|\
0)
#define GRA_SSA_GCF_DEFAULT (\
GRA_GCF_SSA|\
GRA_GCF_SPILL_PHI_ON_RESPILL|\
0)

//
/*************************************************
 * GraphColoring methods.
 *************************************************/
//
/**
 * Allocates a code region.
 * \param this the graph coloring object
 * \param codeRegion Code region to be allocated
 * \return \e true in success, else \e false.
 */
static bool
GraphColoring_allocCodeRegion(GraphColoring this, CodeRegion codeRegion);

//
/**
 * Returns true for an allocatable register, regclass pair.
 */
bool
GRA_isRegisterAllocatable(Register reg, RegFile regFile);

//
/**
 * Returns true for an already assigned temporary.
 */
static inline bool
GRA_isAssigned(Temporary temporary)
{
  return Temporary_isDedicated(temporary) || Temporary_isAssigned(temporary)
    || Temporary_hasRegister(temporary);
}

//
/**
 * Returns true for an allocatable temporary, regclass pair.
 */
static inline bool
GRA_isAllocatable(Temporary temporary, RegFile regFile)
{
  bool res = Temporary_hasRegFile(temporary) &&
    Temporary_regFile(temporary) == regFile &&
    (!GRA_isAssigned(temporary) ||
     GRA_isRegisterAllocatable(Temporary_REGISTER(temporary), Temporary_regFile(temporary)));
  return res;
}

//
/**
 * Fill the temporary table allocatable for the given reg file.
 */
static void
GraphColoring_fillTemporaryTable(GraphColoring this, RegFile regFile);

//
static bool
Operation_usesTemporary(Operation operation, Temporary tmp);

//
static bool
Operation_definesTemporary(Operation operation, Temporary tmp);

//
static bool
InterfNode_hasPHI(InterfNode interfNode);

//
/**
 * Structure permettant de "spiller" celon l'heuristique de Chatin.
 */
struct ChaitinSpill_ {
  //@args Temporary origin
  Temporary ORIGIN; //<!variable d'origine que l'on cherche a "spiller"
  Temporary SUBSTITUTE; //<!variable utilisee pour le "spill"
  Operation LASTUSE; //<!derniere operation ou la variable fut utilisee
  bool RELOADED; //<!a-t-elle ete rechargee au moins une fois?
  bool RELOADED_LASTUSE; //<!a-t-elle ete rechargee a la derniere utilisation?
};
typedef struct ChaitinSpill_ ChaitinSpill_, *ChaitinSpill;
typedef const struct ChaitinSpill_ *const_ChaitinSpill;
typedef struct ChaitinSpill_ * restrict_ChaitinSpill;

extern ChaitinSpill
ChaitinSpill_Ctor(ChaitinSpill this, Temporary origin);

extern ChaitinSpill
ChaitinSpill_Copy(ChaitinSpill this, const_ChaitinSpill that);

extern void
ChaitinSpill_Dtor(ChaitinSpill this);

extern size_t
ChaitinSpill_Size(Temporary origin);

#define ChaitinSpill_ORIGIN(this) ((this)->ORIGIN)
#define ChaitinSpill__ORIGIN(this) (&(this)->ORIGIN)
#define ChaitinSpill_SUBSTITUTE(this) ((this)->SUBSTITUTE)
#define ChaitinSpill__SUBSTITUTE(this) (&(this)->SUBSTITUTE)
#define ChaitinSpill_LASTUSE(this) ((this)->LASTUSE)
#define ChaitinSpill__LASTUSE(this) (&(this)->LASTUSE)
#define ChaitinSpill_RELOADED(this) ((this)->RELOADED)
#define ChaitinSpill__RELOADED(this) (&(this)->RELOADED)
#define ChaitinSpill_RELOADED_LASTUSE(this) ((this)->RELOADED_LASTUSE)
#define ChaitinSpill__RELOADED_LASTUSE(this) (&(this)->RELOADED_LASTUSE)


//
/* Currently not ported */
#define BasicBlock_NUMBERING(b) (0)
#define BasicBlock__NUMBERING(b) ((int *)(0))

//
/**
 * Stats on edges and blocks.
 */
static void
GC_statMaxLive(GraphColoring this, bool after);

//
/**
 * Returns the (unweighted) cost of an operation in the given BB.
 * Return the cost of the operation only, not the branch penalty.
 * If 0, the operation must not be counted (pseudo).
 */
static float
BasicBlock_operationCost(BasicBlock this, Operation operation);

//
/**
 * Returns the (unweighted) branch cost penalty.
 * Return the banch penalty only, not the cost of the operation.
 */
static float
BasicBlock_operationBranchPenalty(BasicBlock this, Operation operation);

//
/**
 * Stats on edges and blocks.
 */
static void
CodeRegion_statSpill(GraphColoring this, const char *msg);

//
/**
 * Stats on edges and blocks.
 */
static void
CodeRegion_statCodeRegion(CodeRegion codeRegion);

//
/**
 * this = E = (A,B). Create C on E
 */
static void
CodeRegion_splitCriticalEdges(GraphColoring gc, CodeRegion codeRegion, bool split_false, bool split_weak, bool split_strong);

//
/**
 * Remove empty blocks that have a unique predecessor and a unique successor
 * with explicit branch operations linking the three blocks.
 * This includes the blocks inserted for spliting weak/strong critical arcs.
 */
static void
CodeRegion_removeEmptyBlocks(CodeRegion codeRegion);

/*
 * GraphColoring_INIT --	Initialize the GraphColoring module.
 */
#define GraphColoring_INIT()

/*
 * GraphColoring_FINI --	Finalize the GraphColoring module.
 */
#define GraphColoring_FINI()


#endif//GraphColoring__h_INCLUDED
