#ifndef InterfGraph_h_INCLUDED
#define InterfGraph_h_INCLUDED

/*
 * !!!!	InterfGraph.xcc
 *
 * Christophe Guillon (Christophe.Guillon@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
struct InterfNode_;
  //@args Memory memory, Temporary temporary
typedef struct InterfNode_ InterfNode_, *InterfNode;
typedef const struct InterfNode_ *const_InterfNode;
typedef struct InterfNode_ * restrict_InterfNode;

extern InterfNode
InterfNode_Ctor(InterfNode this, Memory memory, Temporary temporary);

extern InterfNode
InterfNode_Copy(InterfNode this, const_InterfNode that);

extern void
InterfNode_Dtor(InterfNode this);

extern size_t
InterfNode_Size(Memory memory, Temporary temporary);

//
struct InterfMove_;
//@args Operation operation, struct InterfNode_ *def,
//@args struct InterfNode_ *use, int32_t defIdx, int32_t useIdx, float cost
typedef struct InterfMove_ InterfMove_, *InterfMove;
typedef const struct InterfMove_ *const_InterfMove;
typedef struct InterfMove_ * restrict_InterfMove;

extern InterfMove
InterfMove_Ctor(InterfMove this, Operation operation, struct InterfNode_ *def, struct InterfNode_ *use, int32_t defIdx, int32_t useIdx, float cost);

extern InterfMove
InterfMove_Copy(InterfMove this, const_InterfMove that);

extern void
InterfMove_Dtor(InterfMove this);

extern size_t
InterfMove_Size(Operation operation, struct InterfNode_ *def, struct InterfNode_ *use, int32_t defIdx, int32_t useIdx, float cost);

//
extern Operation
InterfMove_operation(InterfMove this);

//
extern InterfNode
InterfMove_def(InterfMove this);

//
extern InterfNode
InterfMove_use(InterfMove this);

//
extern int32_t
InterfMove_useIdx(InterfMove this);

//
extern int32_t
InterfMove_defIdx(InterfMove this);

//
extern float
InterfMove_cost(InterfMove this);

//
extern IDList
InterfNode_uses(InterfNode this);

//
extern IDList
InterfNode_defs(InterfNode this);

//
extern Temporary
InterfNode_temporary(InterfNode this);

//
extern IDList
InterfNode_movelist(InterfNode this);

//
/**
 * Get the index (unique) of a node.
 * It simply gets the field IDENTITY of the related temporary.
 * \param interfNode The interference node.
 * \return The index (unique) of a node.
 */
extern uint32_t
InterfNode_index(InterfNode interfNode);				

//
extern void
InterfNode_addUse( InterfNode interfNode, Operation operation);

//
extern void
InterfNode_addDef( InterfNode interfNode, Operation operation);

//
/**
 * The structure of the interference graph.
 * The adjacency matrix is redundant, because the graph is undirected. So if the arc (a,b) is in the graph, so is (b,a).
 * The adjacency edge (a,a) is used to know if the node 'a' is in the graph.
 * Indices are between 0 and (size - 1).
 */
struct InterfGraph_ {
  //@args Memory memory
  Memory MEMORY; //!< Memory for allocation
  uint32_t SIZE; //!< Maximum number of nodes present in this graph.
  BitSet ADJSET; //!< Adjacency matrix.
  InterfNode *NODES;
};
typedef struct InterfGraph_ InterfGraph_, *InterfGraph;
typedef const struct InterfGraph_ *const_InterfGraph;
typedef struct InterfGraph_ * restrict_InterfGraph;

extern InterfGraph
InterfGraph_Ctor(InterfGraph this, Memory memory);

extern InterfGraph
InterfGraph_Copy(InterfGraph this, const_InterfGraph that);

extern void
InterfGraph_Dtor(InterfGraph this);

extern size_t
InterfGraph_Size(Memory memory);

#define InterfGraph_MEMORY(this) ((this)->MEMORY)
#define InterfGraph__MEMORY(this) (&(this)->MEMORY)
#define InterfGraph_SIZE(this) ((this)->SIZE)
#define InterfGraph__SIZE(this) (&(this)->SIZE)
#define InterfGraph_ADJSET(this) ((this)->ADJSET)
#define InterfGraph__ADJSET(this) (&(this)->ADJSET)
#define InterfGraph_NODES(this) ((this)->NODES)
#define InterfGraph__NODES(this) (&(this)->NODES)


//
/**
 * To know if two nodes are neighbors in the interference graph.
 * \param interfGraph The interference graph.
 * \param interfNode1 The first interference node.
 * \param interfNode2 The second interference node.
 * \return \e true if these two nodes are neighbors, else \e false.
 */
extern bool
InterfGraph_isAdjacent(InterfGraph interfGraph,
			InterfNode interfNode1,
			InterfNode interfNode2 );

/*
 * This helper is in fact a gathering of piece of code used in GraphColoring.xcc to
 * construct interference graph with affinities
 */
struct InterfGraphBuilder_ {
  //@args Memory memory, CodeRegion codeRegion, RegFile regFile,
  //@args bool clobbersDoNotInterfer, bool newHandlingOfNegAff,
  //@args bool addMoveInterference, bool noMoveAffinity, bool isSSA
  Memory MEMORY;
  bool CLOBBERSDONOTINTERFER;
  bool NEWHANDLINGOFNEGAFF;
  bool ADDMOVEINTERFERENCE;
  bool NOMOVEAFFINITY;
  bool ISSSA;
  CodeRegion CODEREGION;
  //@access CONVENTION	CodeRegion_convention(InterfGraphBuilder_CODEREGION(this))
  RegFile REGFILE;
  IDList_ NODES_;
  //@access NODES InterfGraphBuilder__NODES_(this)
  IDList_ MOVES_;
  //@access MOVES InterfGraphBuilder__MOVES_(this)
  InterfGraph_ INTERFGRAPH_;
  //@access INTERFGRAPH InterfGraphBuilder__INTERFGRAPH_(this)
  TemporaryTable_ TEMPORARIES_;
  //@access TEMPORARIES InterfGraphBuilder__TEMPORARIES_(this)
  Liveness LIVENESS;
  IDList_ SPECIALNODES_;
  //@access SPECIALNODES InterfGraphBuilder__SPECIALNODES_(this)
};
typedef struct InterfGraphBuilder_ InterfGraphBuilder_, *InterfGraphBuilder;
typedef const struct InterfGraphBuilder_ *const_InterfGraphBuilder;
typedef struct InterfGraphBuilder_ * restrict_InterfGraphBuilder;

extern InterfGraphBuilder
InterfGraphBuilder_Ctor(InterfGraphBuilder this, Memory memory, CodeRegion codeRegion, RegFile regFile, bool clobbersDoNotInterfer, bool newHandlingOfNegAff, bool addMoveInterference, bool noMoveAffinity, bool isSSA);

extern InterfGraphBuilder
InterfGraphBuilder_Copy(InterfGraphBuilder this, const_InterfGraphBuilder that);

extern void
InterfGraphBuilder_Dtor(InterfGraphBuilder this);

extern size_t
InterfGraphBuilder_Size(Memory memory, CodeRegion codeRegion, RegFile regFile, bool clobbersDoNotInterfer, bool newHandlingOfNegAff, bool addMoveInterference, bool noMoveAffinity, bool isSSA);

#define InterfGraphBuilder_MEMORY(this) ((this)->MEMORY)
#define InterfGraphBuilder__MEMORY(this) (&(this)->MEMORY)
#define InterfGraphBuilder_CLOBBERSDONOTINTERFER(this) ((this)->CLOBBERSDONOTINTERFER)
#define InterfGraphBuilder__CLOBBERSDONOTINTERFER(this) (&(this)->CLOBBERSDONOTINTERFER)
#define InterfGraphBuilder_NEWHANDLINGOFNEGAFF(this) ((this)->NEWHANDLINGOFNEGAFF)
#define InterfGraphBuilder__NEWHANDLINGOFNEGAFF(this) (&(this)->NEWHANDLINGOFNEGAFF)
#define InterfGraphBuilder_ADDMOVEINTERFERENCE(this) ((this)->ADDMOVEINTERFERENCE)
#define InterfGraphBuilder__ADDMOVEINTERFERENCE(this) (&(this)->ADDMOVEINTERFERENCE)
#define InterfGraphBuilder_NOMOVEAFFINITY(this) ((this)->NOMOVEAFFINITY)
#define InterfGraphBuilder__NOMOVEAFFINITY(this) (&(this)->NOMOVEAFFINITY)
#define InterfGraphBuilder_ISSSA(this) ((this)->ISSSA)
#define InterfGraphBuilder__ISSSA(this) (&(this)->ISSSA)
#define InterfGraphBuilder_CODEREGION(this) ((this)->CODEREGION)
#define InterfGraphBuilder__CODEREGION(this) (&(this)->CODEREGION)
#define InterfGraphBuilder_REGFILE(this) ((this)->REGFILE)
#define InterfGraphBuilder__REGFILE(this) (&(this)->REGFILE)
#define InterfGraphBuilder_NODES_(this) ((this)->NODES_)
#define InterfGraphBuilder__NODES_(this) (&(this)->NODES_)
#define InterfGraphBuilder_MOVES_(this) ((this)->MOVES_)
#define InterfGraphBuilder__MOVES_(this) (&(this)->MOVES_)
#define InterfGraphBuilder_INTERFGRAPH_(this) ((this)->INTERFGRAPH_)
#define InterfGraphBuilder__INTERFGRAPH_(this) (&(this)->INTERFGRAPH_)
#define InterfGraphBuilder_TEMPORARIES_(this) ((this)->TEMPORARIES_)
#define InterfGraphBuilder__TEMPORARIES_(this) (&(this)->TEMPORARIES_)
#define InterfGraphBuilder_LIVENESS(this) ((this)->LIVENESS)
#define InterfGraphBuilder__LIVENESS(this) (&(this)->LIVENESS)
#define InterfGraphBuilder_SPECIALNODES_(this) ((this)->SPECIALNODES_)
#define InterfGraphBuilder__SPECIALNODES_(this) (&(this)->SPECIALNODES_)
#define InterfGraphBuilder_NODES(this) (InterfGraphBuilder__NODES_(this))
#define InterfGraphBuilder_CONVENTION(this) (CodeRegion_convention(InterfGraphBuilder_CODEREGION(this)))
#define InterfGraphBuilder_INTERFGRAPH(this) (InterfGraphBuilder__INTERFGRAPH_(this))
#define InterfGraphBuilder_TEMPORARIES(this) (InterfGraphBuilder__TEMPORARIES_(this))
#define InterfGraphBuilder_SPECIALNODES(this) (InterfGraphBuilder__SPECIALNODES_(this))
#define InterfGraphBuilder_MOVES(this) (InterfGraphBuilder__MOVES_(this))


//
#define BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporary) \
  { \
    TemporarySet_ _liveOutSet_[1];                                \
    TemporarySet _liveOutSet = BasicBlock_liveOutSet(basicBlock); \
    BasicBlock _basicBlock = basicBlock;                                \
    if(BasicBlock_phiOutSet(basicBlock) != NULL) {                      \
      TemporarySet_Ctor(_liveOutSet_, TemporarySet_memory(_liveOutSet), \
                        TemporarySet_count(_liveOutSet) \
                        + TemporarySet_count(BasicBlock_phiOutSet(basicBlock))); \
      TemporarySet_union(_liveOutSet_, _liveOutSet); \
      TemporarySet_union(_liveOutSet_, BasicBlock_phiOutSet(basicBlock)); \
      _liveOutSet = _liveOutSet_; \
    }                                                                   \
    TemporarySet_FORSORT_Temporary(_liveOutSet, Temporary_compare, temporary)
#define BasicBlock_liveOutSet_ENDSORT_Temporary \
    TemporarySet_ENDSORT_Temporary                 \
    if(BasicBlock_phiOutSet(_basicBlock) != NULL) { \
      TemporarySet_Dtor(_liveOutSet_);             \
    }                                              \
  }
#define BasicBlock_liveInSet_FORSORT_Temporary(basicBlock, temporary) \
  TemporarySet_FORSORT_Temporary(BasicBlock_liveInSet(basicBlock), Temporary_compare, temporary)
#define BasicBlock_liveInSet_ENDSORT_Temporary \
  TemporarySet_ENDSORT_Temporary
#define PHI_arguments_FOREACH_temporary(operation, edge, idx, use) \
  Operation_ARGUMENTS_FOREACH_Temporary(operation, use) { \
    int idx = Operation_ARGUMENTS_INDEX; \
    BasicBlockEdge edge = BasicBlock_getPhiEnterEdge(Operation_basicBlock(operation),  Operation_ARGUMENTS_INDEX); \
    {
#define PHI_arguments_ENDEACH_temporary \
    } \
  } Operation_ARGUMENTS_ENDEACH_Temporary
#define BasicBlock_simpleLiveOutSet_FORSORT_Temporary(basicBlock, temporary) \
  TemporarySet_FORSORT_Temporary(BasicBlock_liveOutSet(basicBlock), Temporary_compare, temporary)
#define BasicBlock_simpleLiveOutSet_ENDSORT_Temporary \
  TemporarySet_ENDSORT_Temporary

//
extern Temporary
InterfGraphBuilder_getDedicated(InterfGraphBuilder this, Register registre);

//
extern InterfNode
InterfGraphBuilder_getNode(InterfGraphBuilder this, Temporary temporary);

//
extern void
InterfGraphBuilder_addPCOPYAffinity(InterfGraphBuilder this, Operation operation, int32_t idx);

//
extern void
InterfGraphBuilder_addMoveAffinity(InterfGraphBuilder this, Operation operation);

/*
 * Display mapping between index and identity for all nodes (temporaries)
 */
void
InterfGraphBuilder_displayMapping(InterfGraphBuilder this, FILE* file);

/*
 * File format is the following:
 * NbNodes NbColors
 * Interferences Edges: I J -1\n
 * Affinities Edges: I J W\n
 */
void
InterfGraphBuilder_exportGraph(InterfGraphBuilder this, FILE* file,
                               bool dispInterferenceOnSame,
                               bool dispAffinitiesOnInterference,
                               bool dispInterferenceOnDedicated);


#endif//InterfGraph_h_INCLUDED
