#ifndef OutOfColoredSSA_h_INCLUDED
#define OutOfColoredSSA_h_INCLUDED

/*
 * !!!!	OutOfColoredSSA.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 * Francois de Ferriere (Francois.de-Ferriere@st.com).
 * Christophe Guillon (Christophe.Guillon@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#define MAX_RPC 64

//
struct Rpc_;
typedef struct Rpc_ Rpc_, *Rpc;
typedef const struct Rpc_ *const_Rpc;
typedef struct Rpc_ * restrict_Rpc;

//
struct Rpc_ {
  //@args	int nregs
  int NREGS;
  //@access nRegs Rpc_NREGS(this)
  int TO_[MAX_RPC];
  //@access TO Rpc_TO_(this)
  int INV_[MAX_RPC];
  //@access INV Rpc_INV_(this)
};

extern Rpc
Rpc_Ctor(Rpc this, int nregs);

extern Rpc
Rpc_Copy(Rpc this, const_Rpc that);

extern void
Rpc_Dtor(Rpc this);

extern size_t
Rpc_Size(int nregs);

#define Rpc_NREGS(this) ((this)->NREGS)
#define Rpc__NREGS(this) (&(this)->NREGS)
#define Rpc_TO_(this) ((this)->TO_)
#define Rpc__TO_(this) (&(this)->TO_)
#define Rpc_INV_(this) ((this)->INV_)
#define Rpc__INV_(this) (&(this)->INV_)
#define Rpc_nRegs(this) (Rpc_NREGS(this))
#define Rpc_INV(this) (Rpc_INV_(this))
#define Rpc_TO(this) (Rpc_TO_(this))


//
struct RpcEvolved_;
typedef struct RpcEvolved_ RpcEvolved_, *RpcEvolved;
typedef const struct RpcEvolved_ *const_RpcEvolved;
typedef struct RpcEvolved_ * restrict_RpcEvolved;

//
// Evolution of Rpc structure defined in OutOfColoredSSA.xcc.
// The evolution consists in the adding of an array of temporary.
// We use this array to generate code with consistent variable naming (no
// variable with no defintion).
struct RpcEvolved_ {
  //@args int nregs
  Rpc_ __;
  //@access TO Rpc_TO(RpcEvolved____(this))
  //@access INV Rpc_INV(RpcEvolved____(this))
  //@access RPC RpcEvolved____(this)
  // Map colors to temporaries before the rpc
  Temporary BEFORE[MAX_RPC];
  // Map colors to temporaries after the rpc
  Temporary AFTER[MAX_RPC];
};

extern RpcEvolved
RpcEvolved_Ctor(RpcEvolved this, int nregs);

extern RpcEvolved
RpcEvolved_Copy(RpcEvolved this, const_RpcEvolved that);

extern void
RpcEvolved_Dtor(RpcEvolved this);

extern size_t
RpcEvolved_Size(int nregs);

#define RpcEvolved___(this) ((this)->__)
#define RpcEvolved____(this) (&(this)->__)
#define RpcEvolved_BEFORE(this) ((this)->BEFORE)
#define RpcEvolved__BEFORE(this) (&(this)->BEFORE)
#define RpcEvolved_AFTER(this) ((this)->AFTER)
#define RpcEvolved__AFTER(this) (&(this)->AFTER)
#define RpcEvolved_INV(this) (Rpc_INV(RpcEvolved____(this)))
#define RpcEvolved_RPC(this) (RpcEvolved____(this))
#define RpcEvolved_TO(this) (Rpc_TO(RpcEvolved____(this)))


//
extern void
RpcEvolved_clear(RpcEvolved this);

//
extern void
RpcEvolved_generateRpc(Operation point, RpcEvolved rpc, int32_t idxToReg[],
                       bool before);

//
// regToIdx and idxToReg arraies must have Register__ slot
extern void
RpcEvolved_initRegMap(int32_t regToIdx[], int32_t idxToReg[],
                      const_RegisterSet allocatableSet,
                      RegFile regFile);

//
/**
 * Discard PHI functions for the current register class.
 * Implment the out of colored SSA algorihm.
 *
 * WARNING: All arguments of a phi operation with the same register as the
 * result, must be in the same variable. Otherwise inconsistant code may
 * be generated
 *
 * NOTE: To initialize default options like in GraphColoring.xcc, one should
 * call GraphColoring_initialize
 */
extern void
OutOfColoredSSA_discardPHIs (CodeRegion codeRegion, RegFile regFile,
                             Liveness* pliveness, Memory memory);


#endif//OutOfColoredSSA_h_INCLUDED
