#ifndef LoopScope_h_INCLUDED
#define LoopScope_h_INCLUDED

/*
 * !!!!	LoopScope.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * LoopScope_Factory --	IndexedFactory for LoopScope(s).
 */
extern IndexedFactory_ LoopScope_Factory[1];

/*
 * LoopScopeFlag --	Enumerates the LoopScope flags.
 */
typedef enum {
  LoopScopeFlag_Inner = 0x1,		// LoopScope of inner loop.
  LoopScopeFlag_Simple = 0x2,		// LoopScope of simple loop.
  LoopScopeFlag_Hardware = 0x4,		// LoopScope of hardware loop.
  LoopScopeFlag_Remainder = 0x8,	// LoopScope of remainder loop.
} enum_LoopScopeFlag;
typedef uint8_t LoopScopeFlags;

//
extern struct ETable_ LoopScopeFlag_Table[1];

/*
 * LoopScope --	Persistent loop scope information.
 *
 * A LoopScope maintains loop scope information accross optimizations, in particular
 * dependences between memory access operations, user loop pragma, and diagnostics.
 */
struct LoopScope_ {
  //@args	Memory memory, BasicBlock headerBlock, Configure configure
  Indexed_ __;			// A LoopScope_ ISA Indexed_.
  //@access IDENTITY	Indexed_IDENTITY(LoopScope____(this))
  //@access INDEX	Indexed_INDEX(LoopScope____(this))
  //@mutate INDEX	Indexed__INDEX(LoopScope____(this))
  Memory MEMORY;			// This LoopScope Memory.
  BasicBlock HEADERBLOCK;		// The loop header BasicBlock.
  Temporary TRIPCOUNT;			// The Temporary that holds the loop trip count.
  struct LoopScope_ *PARENT;		// The parent of this LoopScope.
  DependenceTable DEPENDENCETABLE;	// The supplied dependences between operations.
  Configure_ CONFIGURE[1];		// Configure of this LoopScope.
  //@access PROCESSOR		Configure_PROCESSOR(LoopScope_configure(this))
  //@access COMPENSATION	Configure_COMPENSATION(LoopScope_configure(this))
  //@access SPECULATION		Configure_SPECULATION(LoopScope_configure(this))
  //@access RELAXATION		Configure_RELAXATION(LoopScope_configure(this))
  //@access RENAMING		Configure_RENAMING(LoopScope_configure(this))
  //@access BOOSTING		Configure_BOOSTING(LoopScope_configure(this))
  //@access ALIASING		Configure_ALIASING(LoopScope_configure(this))
  //@access TRIPMODULUS		Configure_TRIPMODULUS(LoopScope_configure(this))
  //@access TRIPRESIDUE		Configure_TRIPRESIDUE(LoopScope_configure(this))
  //@access LOGPREPAD		Configure_LOGPREPAD(LoopScope_configure(this))
  //@access LOGPOSTPAD		Configure_LOGPOSTPAD(LoopScope_configure(this))
  //@access PIPELINING		Configure_PIPELINING(LoopScope_configure(this))
  //@access PRELOADING		Configure_PRELOADING(LoopScope_configure(this))
  //@access L1MISSEXTRA		Configure_L1MISSEXTRA(LoopScope_configure(this))
  LoopScopeFlags FLAGS;		// This LoopScope flags.
  //@access isInner		((LoopScope_flags(this) & LoopScopeFlag_Inner) != 0)
  //@access isSimple		((LoopScope_flags(this) & LoopScopeFlag_Simple) != 0)
  //@access isHardware		((LoopScope_flags(this) & LoopScopeFlag_Hardware) != 0)
  //@access isRemainder		((LoopScope_flags(this) & LoopScopeFlag_Remainder) != 0)
  //@access isCounted		(LoopScope_tripCount(this) != NULL)
  int16_t NESTING;			// Nesting level of this LoopScope.
  int16_t BLOCKCOUNT;			// Count of BasicBlocks(s) with same HEADERBLOCK.
  int16_t PEELCOUNT;			// The amount of loop peeling applied.
  int16_t UNROLLED;			// The amount of loop unrolling applied.
  int16_t KUNROLL;			// The amount of kernel unrolling applied.
  int16_t LAMBDA;			// II after loop scheduling or pipelining.
  int16_t LENGTH;			// Makespan after loop scheduling or pipelining.
  int16_t LATENCY;			// Critical cycle latency in cycles.
  int16_t DISTANCE;			// Critical cycle distance in iterations.
};
typedef struct LoopScope_ LoopScope_, *LoopScope;
typedef const struct LoopScope_ *const_LoopScope;
typedef struct LoopScope_ * restrict_LoopScope;

extern LoopScope
LoopScope_Ctor(LoopScope this, Memory memory, BasicBlock headerBlock, Configure configure);

extern LoopScope
LoopScope_Copy(LoopScope this, const_LoopScope that);

extern void
LoopScope_Dtor(LoopScope this);

extern size_t
LoopScope_Size(Memory memory, BasicBlock headerBlock, Configure configure);

#define LoopScope___(this) ((this)->__)
#define LoopScope____(this) (&(this)->__)
#define LoopScope_MEMORY(this) ((this)->MEMORY)
#define LoopScope__MEMORY(this) (&(this)->MEMORY)
#define LoopScope_HEADERBLOCK(this) ((this)->HEADERBLOCK)
#define LoopScope__HEADERBLOCK(this) (&(this)->HEADERBLOCK)
#define LoopScope_TRIPCOUNT(this) ((this)->TRIPCOUNT)
#define LoopScope__TRIPCOUNT(this) (&(this)->TRIPCOUNT)
#define LoopScope_PARENT(this) ((this)->PARENT)
#define LoopScope__PARENT(this) (&(this)->PARENT)
#define LoopScope_DEPENDENCETABLE(this) ((this)->DEPENDENCETABLE)
#define LoopScope__DEPENDENCETABLE(this) (&(this)->DEPENDENCETABLE)
#define LoopScope_CONFIGURE(this) ((this)->CONFIGURE)
#define LoopScope__CONFIGURE(this) (&(this)->CONFIGURE)
#define LoopScope_FLAGS(this) ((this)->FLAGS)
#define LoopScope__FLAGS(this) (&(this)->FLAGS)
#define LoopScope_NESTING(this) ((this)->NESTING)
#define LoopScope__NESTING(this) (&(this)->NESTING)
#define LoopScope_BLOCKCOUNT(this) ((this)->BLOCKCOUNT)
#define LoopScope__BLOCKCOUNT(this) (&(this)->BLOCKCOUNT)
#define LoopScope_PEELCOUNT(this) ((this)->PEELCOUNT)
#define LoopScope__PEELCOUNT(this) (&(this)->PEELCOUNT)
#define LoopScope_UNROLLED(this) ((this)->UNROLLED)
#define LoopScope__UNROLLED(this) (&(this)->UNROLLED)
#define LoopScope_KUNROLL(this) ((this)->KUNROLL)
#define LoopScope__KUNROLL(this) (&(this)->KUNROLL)
#define LoopScope_LAMBDA(this) ((this)->LAMBDA)
#define LoopScope__LAMBDA(this) (&(this)->LAMBDA)
#define LoopScope_LENGTH(this) ((this)->LENGTH)
#define LoopScope__LENGTH(this) (&(this)->LENGTH)
#define LoopScope_LATENCY(this) ((this)->LATENCY)
#define LoopScope__LATENCY(this) (&(this)->LATENCY)
#define LoopScope_DISTANCE(this) ((this)->DISTANCE)
#define LoopScope__DISTANCE(this) (&(this)->DISTANCE)
#define LoopScope_COMPENSATION(this) (Configure_COMPENSATION(LoopScope_configure(this)))
#define LoopScope_isInner(this) (((LoopScope_flags(this) & LoopScopeFlag_Inner) != 0))
#define LoopScope_LOGPREPAD(this) (Configure_LOGPREPAD(LoopScope_configure(this)))
#define LoopScope_RENAMING(this) (Configure_RENAMING(LoopScope_configure(this)))
#define LoopScope_TRIPRESIDUE(this) (Configure_TRIPRESIDUE(LoopScope_configure(this)))
#define LoopScope_ALIASING(this) (Configure_ALIASING(LoopScope_configure(this)))
#define LoopScope_INDEX(this) (Indexed_INDEX(LoopScope____(this)))
#define LoopScope_RELAXATION(this) (Configure_RELAXATION(LoopScope_configure(this)))
#define LoopScope_isHardware(this) (((LoopScope_flags(this) & LoopScopeFlag_Hardware) != 0))
#define LoopScope_PRELOADING(this) (Configure_PRELOADING(LoopScope_configure(this)))
#define LoopScope_PIPELINING(this) (Configure_PIPELINING(LoopScope_configure(this)))
#define LoopScope_TRIPMODULUS(this) (Configure_TRIPMODULUS(LoopScope_configure(this)))
#define LoopScope_SPECULATION(this) (Configure_SPECULATION(LoopScope_configure(this)))
#define LoopScope_BOOSTING(this) (Configure_BOOSTING(LoopScope_configure(this)))
#define LoopScope_isRemainder(this) (((LoopScope_flags(this) & LoopScopeFlag_Remainder) != 0))
#define LoopScope_LOGPOSTPAD(this) (Configure_LOGPOSTPAD(LoopScope_configure(this)))
#define LoopScope_IDENTITY(this) (Indexed_IDENTITY(LoopScope____(this)))
#define LoopScope_isCounted(this) ((LoopScope_tripCount(this) != NULL))
#define LoopScope_PROCESSOR(this) (Configure_PROCESSOR(LoopScope_configure(this)))
#define LoopScope_L1MISSEXTRA(this) (Configure_L1MISSEXTRA(LoopScope_configure(this)))
#define LoopScope_isSimple(this) (((LoopScope_flags(this) & LoopScopeFlag_Simple) != 0))
#define LoopScope__INDEX(this) (Indexed__INDEX(LoopScope____(this)))


//
static inline Indexed
LoopScope__(LoopScope this) {
  return LoopScope____(this);
}
static inline const_Indexed
LoopScope__const(const_LoopScope this) {
  return LoopScope____(this);
}

/*
 * LoopScope_identity --	This LoopScope identity.
 */
static inline uint32_t
LoopScope_identity(const_LoopScope this)
{
  return LoopScope_IDENTITY(this);
}

/*
 * LoopScope_headerBlock --	This LoopScope loop head block.
 */
static inline BasicBlock
LoopScope_headerBlock(const_LoopScope this)
{
  return LoopScope_HEADERBLOCK(this);
}

/*
 * LoopScope_tripCount --	The loop tripcount Temporary when known, else NULL.
 */
static inline Temporary
LoopScope_tripCount(const_LoopScope this)
{
  return LoopScope_TRIPCOUNT(this);
}

/*
 * LoopScope_setTripCount --	Set the loop tripcount.
 */
static inline void
LoopScope_setTripCount(LoopScope this, Temporary tripCount)
{
  *LoopScope__TRIPCOUNT(this) = tripCount;
}

/*
 * LoopScope_configure --	The loop configure.
 */
static inline Configure
LoopScope_configure(LoopScope this)
{
  return LoopScope_CONFIGURE(this);
}

/*
 * LoopScope_flags --	This LoopScope flags.
 */
static inline LoopScopeFlags
LoopScope_flags(const_LoopScope this)
{
  return LoopScope_FLAGS(this);
}

/*
 * LoopScope_setFlags --	Set this LoopScope flags.
 */
static inline void
LoopScope_setFlags(LoopScope this, LoopScopeFlags flags)
{
  *LoopScope__FLAGS(this) |= flags;
}

/*
 * LoopScope_clearFlags --	Reset this LoopScope flags.
 */
static inline void
LoopScope_clearFlags(LoopScope this, LoopScopeFlags flags)
{
  *LoopScope__FLAGS(this) &= ~flags;
}

/*
 * LoopScope_nesting --	Nesting of this LoopScope.
 */
static inline int
LoopScope_nesting(const_LoopScope this)
{
  return LoopScope_NESTING(this);
}

/*
 * LoopScope_blockCount --	Count of BasicBlock(s) with same header block as this LoopScope.
 */
static inline int
LoopScope_blockCount(const_LoopScope this)
{
  return LoopScope_BLOCKCOUNT(this);
}

/*
 * LoopScope_peelCount --	The amount of loop peeling applied.
 */
static inline int
LoopScope_peelCount(const_LoopScope this)
{
  return LoopScope_PEELCOUNT(this);
}

/*
 * LoopScope_unrolled --	The amount of loop unrolling applied.
 */
static inline int
LoopScope_unrolled(const_LoopScope this)
{
  return LoopScope_UNROLLED(this);
}

/*
 * LoopScope_setUnrolled --	Set the amount of loop unrolling.
 */
static inline void
LoopScope_setUnrolled(LoopScope this, int unrolled)
{
  *LoopScope__UNROLLED(this) = unrolled;
}

/*
 * LoopScope_updateUnrolled --	Update the amount of loop unrolling.
 */
static inline void
LoopScope_updateUnrolled(LoopScope this, int kunroll)
{
  int unrolled = LoopScope_UNROLLED(this);
  Except_CHECK(kunroll > 0);
  if (unrolled > 0) {
    *LoopScope__UNROLLED(this) = unrolled*kunroll;
  } else {
    *LoopScope__UNROLLED(this) = kunroll;
  }
}

/*
 * LoopScope_parent --	This LoopScope parent.
 */
static inline LoopScope
LoopScope_parent(const_LoopScope this)
{
  return LoopScope_PARENT(this);
}

/*
 * LoopScope_makeDependenceTable --	Make a DependenceTable for this LoopScope.
 */
DependenceTable
LoopScope_makeDependenceTable(LoopScope this);

/*
 * LoopScope_killDependenceTable --	Kill the DependenceTable of this LoopScope.
 */
void
LoopScope_killDependenceTable(LoopScope this);

/*
 * LoopScope_dependenceTable --	This LoopScope dependenceTable.
 */
static inline DependenceTable
LoopScope_dependenceTable(const_LoopScope this)
{
  return LoopScope_DEPENDENCETABLE(this);
}

/*
 * LoopScope_transfer --	Transfer the contents of that LoopScope to this LoopScope.
 */
void
LoopScope_transfer(LoopScope this, LoopScope that);

/*
 * LoopScope_update --	Update this LoopScope after control-flow change.
 */
void
LoopScope_update(LoopScope this, LoopScope parent_scope,
                int blockCount, bool isSimpleLoop, bool isInnerLoop);

/*
 * LoopScope_pretty --	Pretty-print this LoopScope.
 */
bool
LoopScope_pretty(const_LoopScope this, FILE *file);

//
bool
LoopScope_DTD(FILE *file);

//
bool
LoopScope_dumpXML(LoopScope this, XMLGen gen);


#endif//LoopScope_h_INCLUDED
