#ifndef Variable_h_INCLUDED
#define Variable_h_INCLUDED

/*
 * !!!!	Variable.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 * Christophe Guillon (Christophe.Guillon@st.com).
 *
 * Copyright 2007 - 2008 STMicroelectronics.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * VariableFlags --	Enumerates the Variable flags.
 */
typedef enum {
  VariableFlag_Consuming = 0x1,		// Variable has a CONSUMERS list.
  VariableFlag_Coalesced = 0x2,		// Definition has been coalesced.
  VariableFlag_Unresolved = 0x4,	// Unresolved in Sreedhar.
  VariableFlag_PCOPYDef = 0x10,		// Defined in parallel COPY.
  VariableFlag_PCOPYUse = 0x20,		// Used in parallel COPY.
} enum_VariableFlag;
typedef uint8_t VariableFlags;
#define Variable_isConsuming(this) \
  ((Variable_FLAGS(this) & VariableFlag_Consuming) != 0)
#define Variable_raiseConsuming(this) \
  (*Variable__FLAGS(this) |= VariableFlag_Consuming)
#define Variable_clearConsuming(this) \
  (*Variable__FLAGS(this) &= ~VariableFlag_Consuming)
#define Variable_isCoalesced(this) \
  ((Variable_FLAGS(this) & VariableFlag_Coalesced) != 0)
#define Variable_raiseCoalesced(this) \
  (*Variable__FLAGS(this) |= VariableFlag_Coalesced)
#define Variable_clearCoalesced(this) \
  (*Variable__FLAGS(this) &= ~VariableFlag_Coalesced)
#define Variable_isUnresolved(this) \
  ((Variable_FLAGS(this) & VariableFlag_Unresolved) != 0)
#define Variable_raiseUnresolved(this) \
  (*Variable__FLAGS(this) |= VariableFlag_Unresolved)
#define Variable_clearUnresolved(this) \
  (*Variable__FLAGS(this) &= ~VariableFlag_Unresolved)
#define Variable_isPCOPYDef(this) \
  ((Variable_FLAGS(this) & VariableFlag_PCOPYDef) != 0)
#define Variable_raisePCOPYDef(this) \
  (*Variable__FLAGS(this) |= VariableFlag_PCOPYDef)
#define Variable_clearPCOPYDef(this) \
  (*Variable__FLAGS(this) &= ~VariableFlag_PCOPYDef)
#define Variable_isPCOPYUse(this) \
  ((Variable_FLAGS(this) & VariableFlag_PCOPYUse) != 0)
#define Variable_raisePCOPYUse(this) \
  (*Variable__FLAGS(this) |= VariableFlag_PCOPYUse)
#define Variable_clearPCOPYUse(this) \
  (*Variable__FLAGS(this) &= ~VariableFlag_PCOPYUse)

/*
 * Variable --	SSA form variable.
 */
struct Variable_ {
  //@args	Temporary origin
  Temporary_ __;			// Base of this Variable.
  //@access TEMPORARY	Variable____(this)
  //@access IDENTITY	Temporary_identity(Variable_TEMPORARY(this))
  //@access INDEX	Temporary_index(Variable_TEMPORARY(this))
  OperationResult_ PRODUCER[1];		// The producer OperationResult.
  //@access DEFUSES	((DefUses)Variable_PRODUCER(this))
  IDList_ CONSUMERS_;			// The consumers list of OperationArgument(s).
  //@access CONSUMERS	Variable__CONSUMERS_(this)
  Temporary ORIGIN;			// The origin Temporary.
  struct Variable_ *WEBLINK;		// Uplink for union-find in the SSA web.
  int32_t WEBRANK;			// Rank for union-find in the SSA web.
  VariableFlags FLAGS;			// This Variable flags.
  Register PINNED;			// Variable is pinned to this register.
  Register LOCALLY;			// Variable is locally assigned to this register.
  Register GLOBALLY;			// Variable is globally assigned to this register.
  struct Variable_ *CLASSNEXT;		// Next Variable in congruence class.
  struct VariableClass_ *CONGRUENCE;	// Coalescer congruence class.
  PtrSet_ *NEIGHBORS;			// Sreedhar unresolved neighbors.
  int32_t ORDERING;			// Ordering in VariableClass.
  struct Variable_ *IDOM;		// (Decoalescer)
  float WEIGHT;				// (Decoalescer)
  struct Variable_ *EQUIVALUE;		// Root of class of Variable(s) of same value.
  struct Variable_ *EQUALANCIN;		// For use by interference with equalities.
  struct Variable_ *EQUALANCOUT;	// For use by interference with equalities.
};
typedef struct Variable_ Variable_, *Variable;
typedef const struct Variable_ *const_Variable;
typedef struct Variable_ * restrict_Variable;

extern Variable
Variable_Ctor(Variable this, Temporary origin);

extern Variable
Variable_Copy(Variable this, const_Variable that);

extern void
Variable_Dtor(Variable this);

extern size_t
Variable_Size(Temporary origin);

#define Variable___(this) ((this)->__)
#define Variable____(this) (&(this)->__)
#define Variable_PRODUCER(this) ((this)->PRODUCER)
#define Variable__PRODUCER(this) (&(this)->PRODUCER)
#define Variable_CONSUMERS_(this) ((this)->CONSUMERS_)
#define Variable__CONSUMERS_(this) (&(this)->CONSUMERS_)
#define Variable_ORIGIN(this) ((this)->ORIGIN)
#define Variable__ORIGIN(this) (&(this)->ORIGIN)
#define Variable_WEBLINK(this) ((this)->WEBLINK)
#define Variable__WEBLINK(this) (&(this)->WEBLINK)
#define Variable_WEBRANK(this) ((this)->WEBRANK)
#define Variable__WEBRANK(this) (&(this)->WEBRANK)
#define Variable_FLAGS(this) ((this)->FLAGS)
#define Variable__FLAGS(this) (&(this)->FLAGS)
#define Variable_PINNED(this) ((this)->PINNED)
#define Variable__PINNED(this) (&(this)->PINNED)
#define Variable_LOCALLY(this) ((this)->LOCALLY)
#define Variable__LOCALLY(this) (&(this)->LOCALLY)
#define Variable_GLOBALLY(this) ((this)->GLOBALLY)
#define Variable__GLOBALLY(this) (&(this)->GLOBALLY)
#define Variable_CLASSNEXT(this) ((this)->CLASSNEXT)
#define Variable__CLASSNEXT(this) (&(this)->CLASSNEXT)
#define Variable_CONGRUENCE(this) ((this)->CONGRUENCE)
#define Variable__CONGRUENCE(this) (&(this)->CONGRUENCE)
#define Variable_NEIGHBORS(this) ((this)->NEIGHBORS)
#define Variable__NEIGHBORS(this) (&(this)->NEIGHBORS)
#define Variable_ORDERING(this) ((this)->ORDERING)
#define Variable__ORDERING(this) (&(this)->ORDERING)
#define Variable_IDOM(this) ((this)->IDOM)
#define Variable__IDOM(this) (&(this)->IDOM)
#define Variable_WEIGHT(this) ((this)->WEIGHT)
#define Variable__WEIGHT(this) (&(this)->WEIGHT)
#define Variable_EQUIVALUE(this) ((this)->EQUIVALUE)
#define Variable__EQUIVALUE(this) (&(this)->EQUIVALUE)
#define Variable_EQUALANCIN(this) ((this)->EQUALANCIN)
#define Variable__EQUALANCIN(this) (&(this)->EQUALANCIN)
#define Variable_EQUALANCOUT(this) ((this)->EQUALANCOUT)
#define Variable__EQUALANCOUT(this) (&(this)->EQUALANCOUT)
#define Variable_DEFUSES(this) (((DefUses)Variable_PRODUCER(this)))
#define Variable_INDEX(this) (Temporary_index(Variable_TEMPORARY(this)))
#define Variable_CONSUMERS(this) (Variable__CONSUMERS_(this))
#define Variable_TEMPORARY(this) (Variable____(this))
#define Variable_IDENTITY(this) (Temporary_identity(Variable_TEMPORARY(this)))


//
static inline Temporary
Variable__(Variable this)
{
  Temporary temporary = Variable____(this);
  Except_CHECK(Temporary_isVariable(temporary));
  return temporary;
}
static inline const_Temporary
Variable__const(const_Variable this)
{
  const_Temporary temporary = Variable____(this);
  Except_CHECK(Temporary_isVariable(temporary));
  return temporary;
}

//
static inline VariableFlags
Variable_flags(const_Variable this)
{
  return Variable_FLAGS(this);
}

/*
 * Variable_consumers --	Used by Variable_CONSUMERS_FOREACH_OperationArgument.
 */
static inline const_IDList
Variable_consumers(const_Variable this)
{
  return Variable_CONSUMERS(this);
}

//
#define Variable_CONSUMERS_FOREACH_OperationArgument(this, argument) \
  IDList_FOREACH(Variable_consumers(this), OperationArgument_, argument)
#define Variable_CONSUMERS_ENDEACH_OperationArgument \
  IDList_ENDEACH;

/*
 * Variable_addConsumer --	Add an OperationArgument to this Variable consumers.
 */
void
Variable_addConsumer(Variable this, Operation operation, int32_t argIndex);

/*
 * Variable_makeConsumers --	Make this Variable consumers list.
 */
void
Variable_makeConsumers(Variable this, Memory memory);

/*
 * Variable_killConsumers --	Kill this Variable consumers list.
 */
void
Variable_killConsumers(Variable this);

/*
 * Variable_setOperationResult --	Set an OperationResult for this Variable.
 *
 * In SSA form, should always use this function instead of Operation_setResult.
 */
void
Variable_setOperationResult(Variable this, Operation operation, int32_t resIndex);

/*
 * Variable_setOperationArgument --	Set an OperationArgument for this Variable.
 *
 * In SSA form, should always use this function instead of Operation_setArgument.
 */
void
Variable_setOperationArgument(Variable this, Operation operation, int32_t argIndex);

/*
 * Variable_findWebRoot --	Returns the root of a variable SSA web.
 */
Variable
Variable_findWebRoot(Variable this);

/*
 * Variable_linkWebRoots --	Link the roots of two variable SSA webs.
 */
Variable
Variable_linkWebRoots(Variable this, Variable that);

/*
 * Variable_compare --	Compare two Variable(s) for total ordering.
 */
int
Variable_compare(const_Variable this, const_Variable that);

/*
 * Variable_Sort --	Sort this array of Variable(s) in increasing Variable ORDERING.
 *
 * See the en.wikibook.org heapsort.
 */
void
Variable_Sort(Variable *base, int32_t count);

/*
 * Variable_pretty --  Pretty print of this Variable.
 */
bool
Variable_pretty(const_Variable this, FILE *file, int flags);

//
bool
Variable_prettyNeighbors(const_Variable this, FILE *file, int flags);

//
#define Variable_updateCongruence(this) \
  VariableClass_update(Variable_CONGRUENCE(this), this)

/*
 * VariableStack --	Stack of Variable(s).
 */
struct VariableStack_ {
  //@args	Memory memory, int maxCount
  PtrSeq_ __;		// Underlying PtrSeq.
};
typedef struct VariableStack_ VariableStack_, *VariableStack;
typedef const struct VariableStack_ *const_VariableStack;
typedef struct VariableStack_ * restrict_VariableStack;

extern VariableStack
VariableStack_Ctor(VariableStack this, Memory memory, int maxCount);

extern VariableStack
VariableStack_Copy(VariableStack this, const_VariableStack that);

extern void
VariableStack_Dtor(VariableStack this);

extern size_t
VariableStack_Size(Memory memory, int maxCount);

#define VariableStack___(this) ((this)->__)
#define VariableStack____(this) (&(this)->__)


//
static inline PtrSeq
VariableStack__(VariableStack this)
{
  return VariableStack____(this);
}
static inline const_PtrSeq
VariableStack__const(const_VariableStack this)
{
  return VariableStack____(this);
}

/*
 * VariableStack_FOREACH_Variable --	Iterate over this VariableStack Variable(s).
 */
#define VariableStack_FOREACH_Variable(this, variable) { \
  const_PtrSeq VariableStack____ = VariableStack__const(this); \
  PtrSeq_FOREACH(VariableStack____, Variable, variable) {
#define VariableStack_ENDEACH_Variable \
  } PtrSeq_ENDEACH \
}

/*
 * VariableStack_FORBACK_Variable --	Iterate over this VariableStack Variable(s).
 */
#define VariableStack_FORBACK_Variable(this, variable) { \
  const_PtrSeq VariableStack____ = VariableStack__const(this); \
  PtrSeq_FORBACK(VariableStack____, Variable, variable) {
#define VariableStack_ENDBACK_Variable \
  } PtrSeq_ENDBACK \
}

//
#define VariableStack_resize(this, maxCount) \
  PtrSeq_resize(VariableStack__(this), maxCount)
#define VariableStack_empty(this) \
  PtrSeq_empty(VariableStack__(this))
#define VariableStack_memory(this) \
  PtrSeq_memory(VariableStack__const(this))
#define VariableStack_base_(this) (Variable *) \
  PtrSeq_base_(VariableStack__(this))
#define VariableStack_base(this) (const Variable *) \
  PtrSeq_base(VariableStack__const(this))
#define VariableStack_past(this) (const Variable *) \
  PtrSeq_past(VariableStack__const(this))
#define VariableStack_isEmpty(this) \
  PtrSeq_isEmpty(VariableStack__const(this))
#define VariableStack_isSingle(this) \
  PtrSeq_isSingle(VariableStack__const(this))
#define VariableStack_isFull(this) \
  PtrSeq_isFull(VariableStack__const(this))
#define VariableStack_count(this) \
  PtrSeq_count(VariableStack__const(this))
#define VariableStack_access(this, index) (Variable) \
  PtrSeq_access(VariableStack__const(this), index)
#define VariableStack_bottom(this) (Variable) \
  PtrSeq_bottom(VariableStack__const(this))
#define VariableStack_top(this) (Variable) \
  PtrSeq_top(VariableStack__const(this))
#define VariableStack_pop(this) \
  PtrSeq_pop(VariableStack__(this))
#define VariableStack_push(this, variable) \
  PtrSeq_push(VariableStack__(this), variable)
#define VariableStack_push2(this, variable) \
  PtrSeq_push2(VariableStack__(this), variable)

/*
 * VariableClass --	Class of Variable(s) used for congruence classes.
 * Implemented as simple linked lists sorted by Variable_ORDERING, Variable_IDENTITY.
 */
struct VariableClass_ {
  //@args	Variable variable
  Variable FIRST;	// First Variable in this VariableClass.
  Variable LAST;	// Last Variable in this VariableClass.
  int32_t COUNT;	// Count of Variable(s) in this VariableClass.
  Register REGISTER;	// Used for register targeting constraints.
  Variable CURDOM;
  Operation CURPHI;
};
typedef struct VariableClass_ VariableClass_, *VariableClass;
typedef const struct VariableClass_ *const_VariableClass;
typedef struct VariableClass_ * restrict_VariableClass;

extern VariableClass
VariableClass_Ctor(VariableClass this, Variable variable);

extern VariableClass
VariableClass_Copy(VariableClass this, const_VariableClass that);

extern void
VariableClass_Dtor(VariableClass this);

extern size_t
VariableClass_Size(Variable variable);

#define VariableClass_FIRST(this) ((this)->FIRST)
#define VariableClass__FIRST(this) (&(this)->FIRST)
#define VariableClass_LAST(this) ((this)->LAST)
#define VariableClass__LAST(this) (&(this)->LAST)
#define VariableClass_COUNT(this) ((this)->COUNT)
#define VariableClass__COUNT(this) (&(this)->COUNT)
#define VariableClass_REGISTER(this) ((this)->REGISTER)
#define VariableClass__REGISTER(this) (&(this)->REGISTER)
#define VariableClass_CURDOM(this) ((this)->CURDOM)
#define VariableClass__CURDOM(this) (&(this)->CURDOM)
#define VariableClass_CURPHI(this) ((this)->CURPHI)
#define VariableClass__CURPHI(this) (&(this)->CURPHI)


//
static inline Variable
VariableClass_first(const_VariableClass this)
{
  return VariableClass_FIRST(this);
}

//
static inline Variable
VariableClass_last(const_VariableClass this)
{
  return VariableClass_LAST(this);
}

//
#define VariableClass_FOREACH_Variable(this, variable) { \
  Variable VariableClass_NEXT = NULL; \
  Variable variable = VariableClass_first(this); \
  for (; variable != NULL; variable = VariableClass_NEXT) { \
    VariableClass_NEXT = Variable_CLASSNEXT(variable);
#define VariableClass_ENDEACH_Variable \
  } \
}

//
static inline int32_t
VariableClass_count(const_VariableClass this)
{
  return VariableClass_COUNT(this);
}

//
static inline bool
VariableClass_isEmpty(const_VariableClass this)
{
  return VariableClass_COUNT(this) == 0;
}

//
bool
VariableClass_contains(const_VariableClass this, const_Variable variable);

//
bool
VariableClass_insert(VariableClass this, Variable variable);

//
bool
VariableClass_remove(VariableClass this, Variable variable);

//
void
VariableClass_merge(VariableClass this, VariableClass that);

//
void
VariableClass_update(VariableClass this, Variable variable);

//
bool
VariableClass_hasStrongInterference(VariableClass congruence0, VariableClass congruence1);

//
bool
VariableClass_check(const_VariableClass this);

//
bool
VariableClass_pretty(const_VariableClass this, FILE *file, bool flags);

//
bool
Variable_dumpXML(Variable this, FILE *file, int indent, XMLGen gen);


#endif//Variable_h_INCLUDED
