#ifndef Reservation_h_INCLUDED
#define Reservation_h_INCLUDED

/*
 * !!!!	Reservation.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * ReservationTable --	Reservation table.
 */
struct ReservationTable_ {
  int COLUMNCOUNT;
  ResourceVector_ *COLUMNS;
};
typedef struct ReservationTable_ ReservationTable_, *ReservationTable;
typedef const struct ReservationTable_ *const_ReservationTable;
typedef struct ReservationTable_ * restrict_ReservationTable;

#define ReservationTable_COLUMNCOUNT(this) ((this)->COLUMNCOUNT)
#define ReservationTable__COLUMNCOUNT(this) (&(this)->COLUMNCOUNT)
#define ReservationTable_COLUMNS(this) ((this)->COLUMNS)
#define ReservationTable__COLUMNS(this) (&(this)->COLUMNS)


/*
 * ReservationTable_columnCount --	Count of columns of this ReservationTable.
 */
static inline int
ReservationTable_columnCount(const_ReservationTable this)
{
  return ReservationTable_COLUMNCOUNT(this);
}

/*
 * ReservationTable_columns --	Array of columns of this ReservationTable.
 */
static inline const ResourceVector_ *
ReservationTable_columns(const_ReservationTable this)
{
  return ReservationTable_COLUMNS(this);
}

/*
 * Reservation_name_ --	Names for the Reservation enumeration.
 */
extern const char *
Reservation_name_(Reservation this);
extern const char *
Reservation_name(Reservation this);

/*
 * Reservation_table --	Access to this Reservation class ReservationTable
 */
const_ReservationTable
Reservation_table(Reservation this);

/*
 * Reservation_requirements --	The Resource requirements of this Reservation.
 */
static inline const_ResourceVector
Reservation_requirements(Reservation this)
{
  const_ReservationTable table = Reservation_table(this);
  return ReservationTable_COLUMNS(table);
}

/*
 * Reservation_criticality
 */
unsigned
Reservation_criticality(Reservation this);

/*
 * Reservation_odot --	The minimum number of cycles between two Reservation classes.
 */
int
Reservation_odot(Reservation this, Reservation that, Processor processor);

//
bool
Reservation_DTD(FILE *file);

//
bool
Reservation_XML(FILE *file);

/*
 * !!!!	st200-Reservation.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */


#endif//Reservation_h_INCLUDED
