#ifndef PFA_h_INCLUDED
#define PFA_h_INCLUDED

/*
 * !!!!	PFA.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CCL.h"
#include "ECL.h"
#include "MDS/MDS.h"
#include "AIR/AIR.h"
#include "LIR/LIR.h"

/*
 * The PFA (Program Flow Analyses) library contains the following modules:
 * *	<Liveness Liveness.html> Liveness analysis.
 * *	<Dominance Dominance.html> Dominance analysis.
 * *	<LoopForest LoopForest.html> Loop Forest analysis.
 * *	<ValueNumber ValueNumber.html> Value Number services.
 * *	<Interference Interference.html> Interference analysis.
 * *	<SSAForm SSAForm.html> SSA Form services.
 * *	<SSAConstruct SSAConstruct.html> SSA form Construction.
 * *	<LiveCheck LiveCheck.html> Variable Live Check.
 * *	<Coalescer Coalescer.html> Variable Coalescer base class.
 * *	<Decoalescer Decoalescer.html> Decoalescer out-of-SSA algorithm.
 * *	<Boissinot Boissinot.html> Boissinot out-of-SSA algorithm.
 * *	<Sreedhar Sreedhar.html> Sreedhar out-of-SSA algorithm.
 */

/*
 * Timer --	A multi valued timer
 */
struct Timer_;
typedef struct Timer_ Timer_, *Timer;
typedef const struct Timer_ *const_Timer;
typedef struct Timer_ * restrict_Timer;

//
extern int64_t Timer_start(Timer this, int timer_id);
extern int64_t Timer_stop(Timer this, int timer_id);
extern void Timer_reset(Timer this, int timer_id);
extern void Timer_resetAll(Timer this);
extern int64_t Timer_value(Timer this, int timer_id);
extern int64_t Timer_lastValue(Timer this, int timer_id);

//
#include "PFA/Liveness.h"
#include "PFA/Dominance.h"
#include "PFA/LoopForest.h"
#include "PFA/ValueNumber.h"
#include "PFA/Interference.h"
#include "PFA/SSAForm.h"
#include "PFA/SSAConstruct.h"
#include "PFA/LiveCheck.h"
#include "PFA/Coalescer.h"
#include "PFA/Decoalescer.h"
#include "PFA/Boissinot.h"
#include "PFA/Sreedhar.h"

/*
 * PFA_TRACE --	File handle used to print the PFA traces.
 */
extern FILE *PFA_TRACE;
#ifdef ECL_NTRACE
#define PFA_TRACE NULL
#endif//ECL_NTRACE

/*
 * MAXLIVE_TRACE --	File handle used to print the MAXLIVE traces.
 */
extern FILE *MAXLIVE_TRACE;
#ifdef ECL_NTRACE
#define MAXLIVE_TRACE NULL
#endif//ECL_NTRACE

/*
 * PFA_TIMER --	File handle used to print the PFA time information.
 */
extern FILE *PFA_TIMER;
#ifdef ECL_NTIMER
#define PFA_TIMER NULL
#endif//ECL_TIMER

/*
 * PFA_STATS --	File handle used to print the PFA stats.
 */
extern FILE *PFA_STATS;
#ifdef ECL_NTRACE
#define PFA_STATS NULL
#endif//ECL_NTRACE

/*
 * Variables used to control PFA in non-release mode.
 */
extern int FORCE_SUPERGROW;
extern int FORCE_REUSELIMIT;

/*
 * PFA_INIT --	Initialize the PFA library.
 */
void
PFA_INIT(void);

/*
 * PFA_FINI --	Finalize the PFA library.
 */
void
PFA_FINI(void);


#endif//PFA_h_INCLUDED
