#ifndef Coalescer__h_INCLUDED
#define Coalescer__h_INCLUDED

//
#include "PFA/Coalescer.h"

/*
 * Coalescer_INTERGRAPH --	If not zero, use explicit intersection graph.
 * 0 => ordering interference, 1 => intersection graph, 2 => both for debug.
 */
#if ECL_OPTIMIZE
//#define Coalescer_INTERGRAPH 0
#endif//ECL_OPTIMIZE
#ifndef Coalescer_INTERGRAPH
extern int Coalescer_INTERGRAPH;
#endif//Coalescer_INTERGRAPH

/*
 * Coalescer_SEQCOPIES --	If not zero, use sequential copies.
 */
#if ECL_OPTIMIZE
//#define Coalescer_SEQCOPIES 0
#endif//ECL_OPTIMIZE
#ifndef Coalescer_SEQCOPIES
extern int Coalescer_SEQCOPIES;
#endif//Coalescer_SEQCOPIES

/*
 * Coalescer_EQUIVALUE --	If not zero, use value equivalence between variables.
 */
#if ECL_OPTIMIZE
//#define Coalescer_EQUIVALUE 1
#endif//ECL_OPTIMIZE
#ifndef Coalescer_EQUIVALUE
extern int Coalescer_EQUIVALUE;
#endif//Coalescer_EQUIVALUE

/*
 * Coalescer_DOMFOREST --	If not zero, use lexical order for intersection between variables.
 */
#if ECL_OPTIMIZE
//#define Coalescer_DOMFOREST 1
#endif//ECL_OPTIMIZE
#ifndef Coalescer_DOMFOREST
extern int Coalescer_DOMFOREST;
#endif//Coalescer_DOMFOREST

//
struct Coalescer_ {
  //@args	Memory memory, CodeRegion codeRegion, TemporaryTable temporaryTable,
  //@args	Dominance dominance, OptimizeFlags coalescingFlags
  CodeRegion CODEREGION;		// The CodeRegion of this Coalescer.
  TemporaryTable TEMPORARYTABLE;	// The global Variable(s) considered.
  Dominance DOMINANCE;			// This SSAForm pre-dominance if requested.
  IBList_ CONGRUENCES[1];			// Container for the congruence classes.
  //@access MEMORY	IBList_memory(Coalescer_CONGRUENCES(this))
  VariableStack_ DOMSTACK[1];		// Stack used for DomForest interference checks.
  Interference INTERFERENCE;		// Interference graph of the PHI variables.
  Liveness LIVENESS;			// Liveness sets  of the PHI variables.
  LiveCheck LIVECHECK;                  // Live-check object for liveness checking
  //@access LIVEINSETS	LiveCheck_LIVEINSETS(Coalescer_LIVECHECK(this))
  //@access LIVEOUTSETS	LiveCheck_LIVEOUTSETS(Coalescer_LIVECHECK(this))
  //@access PHIINSETS	LiveCheck_PHIINSETS(Coalescer_LIVECHECK(this))
  //@access PHIOUTSETS	LiveCheck_PHIOUTSETS(Coalescer_LIVECHECK(this))
  bool CONSUMING;			// Whether Variable(s) are consuming.
  OptimizeFlags FLAGS;			// This OptimizeCoalescing flags.
  //@access isSreedhar	((Coalescer_FLAGS(this)&OptimizeCoalescing_Sreedhar) != 0)
  //@access isBoissinot	((Coalescer_FLAGS(this)&OptimizeCoalescing_Boissinot) != 0)
  //@access isDecoalesce	((Coalescer_FLAGS(this)&OptimizeCoalescing_Decoalesce) != 0)
  //@access isVirtualize	((Coalescer_FLAGS(this)&OptimizeCoalescing_Virtualize) != 0)
  //@access isCongruence	((Coalescer_FLAGS(this)&OptimizeCoalescing_Congruence) != 0)
};

extern Coalescer
Coalescer_Ctor(Coalescer this, Memory memory, CodeRegion codeRegion, TemporaryTable temporaryTable, Dominance dominance, OptimizeFlags coalescingFlags);

extern Coalescer
Coalescer_Copy(Coalescer this, const_Coalescer that);

extern void
Coalescer_Dtor(Coalescer this);

extern size_t
Coalescer_Size(Memory memory, CodeRegion codeRegion, TemporaryTable temporaryTable, Dominance dominance, OptimizeFlags coalescingFlags);

#define Coalescer_CODEREGION(this) ((this)->CODEREGION)
#define Coalescer__CODEREGION(this) (&(this)->CODEREGION)
#define Coalescer_TEMPORARYTABLE(this) ((this)->TEMPORARYTABLE)
#define Coalescer__TEMPORARYTABLE(this) (&(this)->TEMPORARYTABLE)
#define Coalescer_DOMINANCE(this) ((this)->DOMINANCE)
#define Coalescer__DOMINANCE(this) (&(this)->DOMINANCE)
#define Coalescer_CONGRUENCES(this) ((this)->CONGRUENCES)
#define Coalescer__CONGRUENCES(this) (&(this)->CONGRUENCES)
#define Coalescer_DOMSTACK(this) ((this)->DOMSTACK)
#define Coalescer__DOMSTACK(this) (&(this)->DOMSTACK)
#define Coalescer_INTERFERENCE(this) ((this)->INTERFERENCE)
#define Coalescer__INTERFERENCE(this) (&(this)->INTERFERENCE)
#define Coalescer_LIVENESS(this) ((this)->LIVENESS)
#define Coalescer__LIVENESS(this) (&(this)->LIVENESS)
#define Coalescer_LIVECHECK(this) ((this)->LIVECHECK)
#define Coalescer__LIVECHECK(this) (&(this)->LIVECHECK)
#define Coalescer_CONSUMING(this) ((this)->CONSUMING)
#define Coalescer__CONSUMING(this) (&(this)->CONSUMING)
#define Coalescer_FLAGS(this) ((this)->FLAGS)
#define Coalescer__FLAGS(this) (&(this)->FLAGS)
#define Coalescer_isCongruence(this) (((Coalescer_FLAGS(this)&OptimizeCoalescing_Congruence) != 0))
#define Coalescer_PHIOUTSETS(this) (LiveCheck_PHIOUTSETS(Coalescer_LIVECHECK(this)))
#define Coalescer_isBoissinot(this) (((Coalescer_FLAGS(this)&OptimizeCoalescing_Boissinot) != 0))
#define Coalescer_isVirtualize(this) (((Coalescer_FLAGS(this)&OptimizeCoalescing_Virtualize) != 0))
#define Coalescer_LIVEINSETS(this) (LiveCheck_LIVEINSETS(Coalescer_LIVECHECK(this)))
#define Coalescer_PHIINSETS(this) (LiveCheck_PHIINSETS(Coalescer_LIVECHECK(this)))
#define Coalescer_isSreedhar(this) (((Coalescer_FLAGS(this)&OptimizeCoalescing_Sreedhar) != 0))
#define Coalescer_MEMORY(this) (IBList_memory(Coalescer_CONGRUENCES(this)))
#define Coalescer_isDecoalesce(this) (((Coalescer_FLAGS(this)&OptimizeCoalescing_Decoalesce) != 0))
#define Coalescer_LIVEOUTSETS(this) (LiveCheck_LIVEOUTSETS(Coalescer_LIVECHECK(this)))


/*
 * Coalescer_initClasses --	Initialize congruence classes from pinned Variable(s).
 */
void
Coalescer_initClasses(Coalescer this);

/*
 * Coalescer_complete --	 Complete this Coalescer construction: creates liveness/livecheck in the right mode & interference if necessary
 */
void
Coalescer_complete(Coalescer this);

/*
 * Coalescer_coalesceCongruences --        Sreedhar congruence-based copy coalescing.
 *
 * Simpler than Sreedhar because did not nullify singleton congruence classes.
 */
/* fab: This is not Sreedhar's coalescing.
   This is a more general congruence-based coalescing under SSA.
   It can handle different interference modes: VALUE (quadratic | linear), Sreedhar.
   To be exhaustive we may add the intersection test (Interference_CC).
   Maybe the mode should be given as an argument with a cross check. */
void
Coalescer_coalesceCongruences(Coalescer this);

/*
 * Coalescer_markCoalescedCopies --        if the source & dest of a copy have the same web link then mark the source as isCoalesced
 */
void
Coalescer_markCoalescedCopies(Coalescer this);

/*
 * Coalescer_dominatesVV --	Dominance between Variable(s) a and b.
 */
bool
Coalescer_dominatesVV(const_Coalescer this, const_Variable a, const_Variable b);

/*
 * Coalescer_dominatesVB --	Dominance between Variable and BasicBlock.
 */
bool
Coalescer_dominatesVB(const_Coalescer this, const_Variable a, const_BasicBlock block_b);

/*
 * Coalescer_intersectsVV --	Intersection between Variable(s) a and b.
 */
bool
Coalescer_intersectsVV(Coalescer this, Variable a, Variable b);

/*
 * Coalescer_interferesVV --	Interference between Variable(s) a and b.
 */
bool
Coalescer_interferesVV(Coalescer this, Variable a, Variable b);

/*
 * Coalescer_interferesDomForestVV -- Interference between Variable(s) a and b or one ancestor.
 *
 * Require: b is null or a ORDERING >= b ORDERING
 */
static bool
Coalescer_interferesDomForestVV(Coalescer this, Variable a, Variable b);

/*
 * Coalescer_interferesDomForest --	Dominance forest interference between Variable(s).
 *
 * Require:	The Variable(s) have a valid lexicographic ORDERING.
 */
bool
Coalescer_interferesDomForest(Coalescer this,
			      VariableClass class_x,
			      VariableClass class_y);

/*
 * Coalescer_mergeCongruencesCC --	Merge classes after a succesful coalescing
 * Require: class_x and class_y two non intefering classes
 * Output: class_x and class_y merged into class_x.
 */
VariableClass
Coalescer_mergeCongruencesCC(Coalescer this, VariableClass class_x, VariableClass class_y);

/*
 * Coalescer_classesFromSSAwebs(coalescer) --	Coalesce phi closure and isCoalesced variables.
 *
 * Requires: initialized congruence classes.
 */
void
Coalescer_classesFromSSAwebs(Coalescer this);

/*
 * Coalescer_interferesCCRegisters --	Interference between class_x and class_y due to REGISTERs.
 */
static inline bool
Coalescer_interferesCCRegisters(Coalescer this, VariableClass class_x, VariableClass class_y)
{
  Register register_x = VariableClass_REGISTER(class_x);
  Register register_y = VariableClass_REGISTER(class_y);
  // There should be no two congruence classes with the same register.
  Except_CHECK(   register_x != register_y
               || register_x == Register__UNDEF);
  return   register_x != Register__UNDEF
        && register_y != Register__UNDEF;
}

/*
 * Coalescer_interferesCC --	Interference between class_x and class_y.
 */
bool
Coalescer_interferesCC(Coalescer this, VariableClass class_x, VariableClass class_y);

/*
 * Coalescer_interferesCP -- Interference between a class and a phi
 */
bool
Coalescer_interferesCP(Coalescer this, VariableClass X, Operation phi);

/*
 * Coalescer_addInterferenceVS --	Add interferences between {x} and set_y \ {x}.
 */
void
Coalescer_addInterferenceVS(Coalescer this,
                            Variable variable_x,
                            TemporarySet set_y);

/*
 * Coalescer_findInterferingClass --	Find an interfering congruence class.
 */
VariableClass
Coalescer_findInterferingClass(Coalescer this);

/*
 * Coalescer_findIntersectingClass --	Find an intersecting congruence class.
 */
VariableClass
Coalescer_findIntersectingClass(Coalescer this);

/*
 * Coalescer_isNeededPhiOutClass --    Check if a Temporary is needed in BasicBlock phi-out.
 */
bool
Coalescer_isNeededPhiOutClass(Coalescer this, BasicBlock basicBlock, Temporary temporary);

/*
 * Coalescer_explicitPhi --       Main processing for Generate_Copies
 */
void
Coalescer_explicitPhi(Coalescer this, Operation phi);

/*
 * Coalescer_INIT --	Initialize the Coalescer module.
 */
#if defined(Coalescer_INTERGRAPH) \
 && defined(Coalescer_SEQCOPIES) \
 && defined(Coalescer_EQUIVALUE) \
 && defined(Coalescer_DOMFOREST)
#define Coalescer_INIT()
#else
void
Coalescer_INIT(void);
#endif

/*
 * Coalescer_FINI --	Finalize the Coalescer module.
 */
#define Coalescer_FINI()


#endif//Coalescer__h_INCLUDED
