#ifndef Dominance__h_INCLUDED
#define Dominance__h_INCLUDED

//
#include "PFA/Dominance.h"

//
#ifndef Dominance_TDMSC
#define Dominance_TDMSC 2
#endif//Dominance_TDMSC

//
struct Dominance_ {
  //@args	Memory memory, CodeRegion codeRegion, DominanceFlags flags
  Memory MEMORY;
  CodeRegion CODEREGION;
  int32_t BLOCKCOUNT;		// Cache of CodeRegion_blockCount.
  const BasicBlock *BLOCKMAP;	// Map BasicBlock INDEX to BasicBlock.
  BasicBlock *POSTORDER;	// The BasicBlock(s) in control DFS post-order.
  BasicBlock *DOMDFS;		// BasicBlock(s) in DFS order of the Dominance tree
  int32_t POSTCOUNT;		// Counter used in control graph DFS.
  int32_t *POSTDFN;		// The post-order control DFS numbers.
  int32_t *IDOMIDX;		// Immediate dominator indices.
  BasicBlockStack_ *CHILDREN;	// Immediate dominance children sets.
  int32_t *PREDFN;		// The pre-order dominance DFS numbers.
  int32_t *LASTDFN;		// The pre-order dominance last DFS numbers.
  int32_t *TREELEVEL;		// Level in the dominator tree.
  BasicBlock *TREEORDER;	// The BasicBlock(s) ordered by non-decreasing TREELEVEL.
  PtrSet_ *FRONTIERS;		// The dominance frontiers.
  PtrSet_ *MERGESETS;		// The merge sets (iterated dominance frontiers).
  int32_t JOINTCOUNT;		// Count of Join edges.
  bool *VISITED;		// Visit flag for BasicBlock(s).
  DominanceFlags FLAGS;		// This Dominance flags.
};

extern Dominance
Dominance_Ctor(Dominance this, Memory memory, CodeRegion codeRegion, DominanceFlags flags);

extern Dominance
Dominance_Copy(Dominance this, const_Dominance that);

extern void
Dominance_Dtor(Dominance this);

extern size_t
Dominance_Size(Memory memory, CodeRegion codeRegion, DominanceFlags flags);

#define Dominance_MEMORY(this) ((this)->MEMORY)
#define Dominance__MEMORY(this) (&(this)->MEMORY)
#define Dominance_CODEREGION(this) ((this)->CODEREGION)
#define Dominance__CODEREGION(this) (&(this)->CODEREGION)
#define Dominance_BLOCKCOUNT(this) ((this)->BLOCKCOUNT)
#define Dominance__BLOCKCOUNT(this) (&(this)->BLOCKCOUNT)
#define Dominance_BLOCKMAP(this) ((this)->BLOCKMAP)
#define Dominance__BLOCKMAP(this) (&(this)->BLOCKMAP)
#define Dominance_POSTORDER(this) ((this)->POSTORDER)
#define Dominance__POSTORDER(this) (&(this)->POSTORDER)
#define Dominance_DOMDFS(this) ((this)->DOMDFS)
#define Dominance__DOMDFS(this) (&(this)->DOMDFS)
#define Dominance_POSTCOUNT(this) ((this)->POSTCOUNT)
#define Dominance__POSTCOUNT(this) (&(this)->POSTCOUNT)
#define Dominance_POSTDFN(this) ((this)->POSTDFN)
#define Dominance__POSTDFN(this) (&(this)->POSTDFN)
#define Dominance_IDOMIDX(this) ((this)->IDOMIDX)
#define Dominance__IDOMIDX(this) (&(this)->IDOMIDX)
#define Dominance_CHILDREN(this) ((this)->CHILDREN)
#define Dominance__CHILDREN(this) (&(this)->CHILDREN)
#define Dominance_PREDFN(this) ((this)->PREDFN)
#define Dominance__PREDFN(this) (&(this)->PREDFN)
#define Dominance_LASTDFN(this) ((this)->LASTDFN)
#define Dominance__LASTDFN(this) (&(this)->LASTDFN)
#define Dominance_TREELEVEL(this) ((this)->TREELEVEL)
#define Dominance__TREELEVEL(this) (&(this)->TREELEVEL)
#define Dominance_TREEORDER(this) ((this)->TREEORDER)
#define Dominance__TREEORDER(this) (&(this)->TREEORDER)
#define Dominance_FRONTIERS(this) ((this)->FRONTIERS)
#define Dominance__FRONTIERS(this) (&(this)->FRONTIERS)
#define Dominance_MERGESETS(this) ((this)->MERGESETS)
#define Dominance__MERGESETS(this) (&(this)->MERGESETS)
#define Dominance_JOINTCOUNT(this) ((this)->JOINTCOUNT)
#define Dominance__JOINTCOUNT(this) (&(this)->JOINTCOUNT)
#define Dominance_VISITED(this) ((this)->VISITED)
#define Dominance__VISITED(this) (&(this)->VISITED)
#define Dominance_FLAGS(this) ((this)->FLAGS)
#define Dominance__FLAGS(this) (&(this)->FLAGS)


//
void
Dominance_makeIDOMIDX(Dominance this, BasicBlock rootBlock, DominanceFlags flags);

//
void
Dominance_doForwardDFS(Dominance this, BasicBlock block,
                       int32_t *_postCount, BasicBlock *postOrder, int32_t *postDFN);

//
void
Dominance_doBackwardDFS(Dominance this, BasicBlock block,
                        int32_t *_postCount, BasicBlock *postOrder, int32_t *postDFN);

//
void
Dominance_makeCHILDREN(Dominance this);

//
void
Dominance_makePREDFN(Dominance this, BasicBlock rootBlock);

//
void
Dominance_doDominatorDFS(Dominance this, BasicBlock block, int32_t *_preCount,
                         int32_t *preDFN, int32_t *lastDFN, BasicBlock *domDFS);

//
void
Dominance_markJoinEdges(Dominance this);

//
void
Dominance_makeTREEORDER(Dominance this, BasicBlock rootBlock);

//
void
Dominance_makeFRONTIERS(Dominance this, DominanceFlags flags);

//
void
Dominance_makeMERGESETS(Dominance this, DominanceFlags flags);

//
bool
Dominance_doForwardTDMSC(Dominance this, bool visited[]);

/*
 * Dominance_INIT --	Initialize the Dominance module.
 */
#define Dominance_INIT()

/*
 * Dominance_FINI --	Finalize the Dominance module.
 */
#define Dominance_FINI()


#endif//Dominance__h_INCLUDED
