#ifndef Liveness_h_INCLUDED
#define Liveness_h_INCLUDED

/*
 * !!!!	Liveness.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * Liveness --	Liveness computation.
 *
 * The =phiSets= argument enables construction of Liveness PHIINSETS and LIVEOUTSETS.
 * These sets can be merged with the Liveness LIVEINSETS and LIVEOUTSETS by calling
 * Liveness_multiplex().
 */
struct Liveness_;
typedef struct Liveness_ Liveness_, *Liveness;
typedef const struct Liveness_ *const_Liveness;
typedef struct Liveness_ * restrict_Liveness;

//
Liveness
Liveness_make(Memory parent, CodeRegion codeRegion, TemporaryTable temporaryTable, bool phiSets);

//
Liveness
Liveness_kill(Liveness this);

/*
 * Liveness_memory --	This Liveness Memory.
 */
Memory
Liveness_memory(const_Liveness this);

/*
 * Liveness_blockCount --	For use in Liveness_FOREACH.
 */
int32_t
Liveness_blockCount(const_Liveness this);

/*
 * Liveness_liveInSets --	Array of live-in TemporarySet(s) indexed by BasicBlock.
 */
TemporarySet_ *
Liveness_liveInSets(const_Liveness this);

/*
 * Liveness_liveOutSets --	Array of live-out TemporarySet(s) indexed by BasicBlock.
 */
TemporarySet_ *
Liveness_liveOutSets(const_Liveness this);

/*
 * Liveness_phiInSets --	Array of phi-in TemporarySet(s) indexed by BasicBlock.
 */
TemporarySet_ *
Liveness_phiInSets(const_Liveness this);

/*
 * Liveness_phiOutSets --	Array of phi-out TemporarySet(s) indexed by BasicBlock.
 */
TemporarySet_ *
Liveness_phiOutSets(const_Liveness this);

/*
 Liveness_pretty -- Pretty-print this Liveness.
 */
bool
Liveness_pretty(Liveness this, FILE *file);

/*
 * Liveness_multiplex --	Convert Liveness to multiplex form (Sreedhar Phi effects).
 */
void
Liveness_multiplex(Liveness this);


#endif//Liveness_h_INCLUDED
