#ifndef Sreedhar__h_INCLUDED
#define Sreedhar__h_INCLUDED

//
#include "PFA/Sreedhar_.h"

//
#define Sreedhar_TIMER_T1 0
#define Sreedhar_TIMER_T2 1
#define Sreedhar_TIMER_T3 2
#define Sreedhar_TIMER_COUNT 3

//
struct Sreedhar_ {
  //@args	Memory memory, CodeRegion codeRegion, SSAForm ssaForm
  Coalescer_ COALESCER[1];		// SSA Coalescer.
  //@access CODEREGION		Coalescer_CODEREGION(Sreedhar_COALESCER(this))
  //@access TEMPORARYTABLE	Coalescer_TEMPORARYTABLE(Sreedhar_COALESCER(this))
  //@access LIVEINSETS		Coalescer_LIVEINSETS(Sreedhar_COALESCER(this))
  //@access LIVEOUTSETS		Coalescer_LIVEOUTSETS(Sreedhar_COALESCER(this))
  //@access PHIINSETS		Coalescer_PHIINSETS(Sreedhar_COALESCER(this))
  //@access PHIOUTSETS		Coalescer_PHIOUTSETS(Sreedhar_COALESCER(this))
  //@access CONGRUENCES		Coalescer_CONGRUENCES(Sreedhar_COALESCER(this))
  //@access SSAFORM		CodeRegion_ssaForm(Sreedhar_CODEREGION(this))
  OperationStack_ PHIOPERATIONS[1];	// Phi-Operations to process.
  //@access MEMORY	OperationStack_memory(Sreedhar_PHIOPERATIONS(this))
  int32_t MAXPHIARITY;			// Maximum PHI arity (operand count).
  PtrSet_ *NEIGHBORSETS;		// The neighbor sets, one per PHI operand.
  PtrSet_ CANDIDATESET[1];		// The current candidate resource set.
  PtrSeq_ UNRESOLVED[1];		// Set of unresolved resources.
};

extern Sreedhar
Sreedhar_Ctor(Sreedhar this, Memory memory, CodeRegion codeRegion, SSAForm ssaForm);

extern Sreedhar
Sreedhar_Copy(Sreedhar this, const_Sreedhar that);

extern void
Sreedhar_Dtor(Sreedhar this);

extern size_t
Sreedhar_Size(Memory memory, CodeRegion codeRegion, SSAForm ssaForm);

#define Sreedhar_COALESCER(this) ((this)->COALESCER)
#define Sreedhar__COALESCER(this) (&(this)->COALESCER)
#define Sreedhar_PHIOPERATIONS(this) ((this)->PHIOPERATIONS)
#define Sreedhar__PHIOPERATIONS(this) (&(this)->PHIOPERATIONS)
#define Sreedhar_MAXPHIARITY(this) ((this)->MAXPHIARITY)
#define Sreedhar__MAXPHIARITY(this) (&(this)->MAXPHIARITY)
#define Sreedhar_NEIGHBORSETS(this) ((this)->NEIGHBORSETS)
#define Sreedhar__NEIGHBORSETS(this) (&(this)->NEIGHBORSETS)
#define Sreedhar_CANDIDATESET(this) ((this)->CANDIDATESET)
#define Sreedhar__CANDIDATESET(this) (&(this)->CANDIDATESET)
#define Sreedhar_UNRESOLVED(this) ((this)->UNRESOLVED)
#define Sreedhar__UNRESOLVED(this) (&(this)->UNRESOLVED)
#define Sreedhar_PHIINSETS(this) (Coalescer_PHIINSETS(Sreedhar_COALESCER(this)))
#define Sreedhar_LIVEINSETS(this) (Coalescer_LIVEINSETS(Sreedhar_COALESCER(this)))
#define Sreedhar_CONGRUENCES(this) (Coalescer_CONGRUENCES(Sreedhar_COALESCER(this)))
#define Sreedhar_CODEREGION(this) (Coalescer_CODEREGION(Sreedhar_COALESCER(this)))
#define Sreedhar_TEMPORARYTABLE(this) (Coalescer_TEMPORARYTABLE(Sreedhar_COALESCER(this)))
#define Sreedhar_MEMORY(this) (OperationStack_memory(Sreedhar_PHIOPERATIONS(this)))
#define Sreedhar_SSAFORM(this) (CodeRegion_ssaForm(Sreedhar_CODEREGION(this)))
#define Sreedhar_PHIOUTSETS(this) (Coalescer_PHIOUTSETS(Sreedhar_COALESCER(this)))
#define Sreedhar_LIVEOUTSETS(this) (Coalescer_LIVEOUTSETS(Sreedhar_COALESCER(this)))


//
int32_t
Sreedhar_fillPhiOperations(Sreedhar this);

/*
 * Sreedhar_doMethod_III_3_4_5_6_7 --	Main processing of Sreedhar's Method III.
 */
void
Sreedhar_doMethod_III_3_4_5_6_7(Sreedhar this,
                                   Operation phi_operation,
                                   OptimizeFlags coalescingFlags);

/*
 * Sreedhar method III.4 --	Check for interference between the members of congruence
 * classes of resources in a phi operation and break interferences between them.
 */
void
Sreedhar_processInterferences(Sreedhar this, Operation phi_operation);

//
void
Sreedhar_breakInterferences(Sreedhar this,
                            Variable v_i, TemporarySet liveSet_i, TemporarySet phiSet_i,
                            Variable v_j, TemporarySet liveSet_j, TemporarySet phiSet_j);

/*
 * Sreedhar_insertCopy --	Inserts copies that are needed for conventional SSA.
 */
void
Sreedhar_insertCopy(Sreedhar this, Operation phi_operation, Temporary resource_i);

/*
 * Sreedhar_isNeededPhiOutClass --	Check if a Temporary is needed in BasicBlock phi-out.
 */
bool
Sreedhar_isNeededPhiOutClass(Sreedhar this, BasicBlock basicBlock, Temporary temporary);

/*
 * Sreedhar method III.5 --	Processes unresolved interferences among the neighbors.
 *
 * Problem is as follow: The NEIGHBORS represent an interference graph between variables
 * where an independent set is required. We want to maximize the size of the independent
 * set, since each vertex (interfering variable) not part of the independent set is put
 * into the candidateSet and this produces a COPY operation in the final code. Maximum
 * independent set is NP-complete, so we use an heuristic.
 *
 * Sreedhar's proposed heuristic is to remove largest degrees first ('mark as resolved')
 * until independent set is obtained, thent then re-insert the removed vertices that have
 * no neighbors in the independent set.
 *
 * We use instead the classic independet set heuristic, that is: find the lowest-degree
 * vertex, add it to the independent set, and delete it and all vertices adjacent to it;
 * repeat this process until the graph is empty. -- BD3
 */
void
Sreedhar_processUnresolvedResources(Sreedhar this, Operation phi_operation);

/*
 * Sreedhar_INIT --	Initialize the Sreedhar module.
 */
#define Sreedhar_INIT()

/*
 * Sreedhar_FINI --	Finalize the Sreedhar module.
 */
#define Sreedhar_FINI()


#endif//Sreedhar__h_INCLUDED
