#ifndef ValueNumber__h_INCLUDED
#define ValueNumber__h_INCLUDED

//
#include "PFA/ValueNumber.h"

//
typedef uint32_t VNTableKey;

/*
 * The VNTableEntry_ structure implements entries to Value Number table.
 */
struct VNTableEntry_ {
  //@args	VNTableKey key, Operation operation, struct VNTableEntry_ *link
  VNTableKey KEY;		// The key, which holds a pointer.
  Operation OPERATION;		// Value
  struct VNTableEntry_ *LINK;	// Link to the next VNTableEntry in bucket.
};
typedef struct VNTableEntry_ VNTableEntry_, *VNTableEntry;
typedef const struct VNTableEntry_ *const_VNTableEntry;
typedef struct VNTableEntry_ * restrict_VNTableEntry;

extern VNTableEntry
VNTableEntry_Ctor(VNTableEntry this, VNTableKey key, Operation operation, struct VNTableEntry_ *link);

extern VNTableEntry
VNTableEntry_Copy(VNTableEntry this, const_VNTableEntry that);

extern void
VNTableEntry_Dtor(VNTableEntry this);

extern size_t
VNTableEntry_Size(VNTableKey key, Operation operation, struct VNTableEntry_ *link);

#define VNTableEntry_KEY(this) ((this)->KEY)
#define VNTableEntry__KEY(this) (&(this)->KEY)
#define VNTableEntry_OPERATION(this) ((this)->OPERATION)
#define VNTableEntry__OPERATION(this) (&(this)->OPERATION)
#define VNTableEntry_LINK(this) ((this)->LINK)
#define VNTableEntry__LINK(this) (&(this)->LINK)


/*
 * The VNTableScope_ structure implements entries to Value Number table.
 *
 * This implementation is borowed from HTable and modified to fit the needs of Value
 * Numbering.
 *
 * FIXME! There is a problem with RootBlock which has no operations, but scope is created
 * for it. Currently, if estimate is 0, table of size 1 is created to avoid crashes.
 */
struct VNTableScope_ {
  //@args	Memory memory, VNTableScope next_scope, uint32_t estimate
  Memory MEMORY;
  uint32_t MASK;		// Mask for modulo clp2.
  int32_t COUNT;		// Count this VNTableScope ENTRIES.
  struct VNTableScope_ *NEXT;		// Pointer to next scope.
  VNTableEntry *BUCKETS;		// This VNTableScope buckets.
};
typedef struct VNTableScope_ VNTableScope_, *VNTableScope;
typedef const struct VNTableScope_ *const_VNTableScope;
typedef struct VNTableScope_ * restrict_VNTableScope;

extern VNTableScope
VNTableScope_Ctor(VNTableScope this, Memory memory, VNTableScope next_scope, uint32_t estimate);

extern VNTableScope
VNTableScope_Copy(VNTableScope this, const_VNTableScope that);

extern void
VNTableScope_Dtor(VNTableScope this);

extern size_t
VNTableScope_Size(Memory memory, VNTableScope next_scope, uint32_t estimate);

#define VNTableScope_MEMORY(this) ((this)->MEMORY)
#define VNTableScope__MEMORY(this) (&(this)->MEMORY)
#define VNTableScope_MASK(this) ((this)->MASK)
#define VNTableScope__MASK(this) (&(this)->MASK)
#define VNTableScope_COUNT(this) ((this)->COUNT)
#define VNTableScope__COUNT(this) (&(this)->COUNT)
#define VNTableScope_NEXT(this) ((this)->NEXT)
#define VNTableScope__NEXT(this) (&(this)->NEXT)
#define VNTableScope_BUCKETS(this) ((this)->BUCKETS)
#define VNTableScope__BUCKETS(this) (&(this)->BUCKETS)


//
VNTableScope
VNTableScope_new(Memory memory, VNTableScope next_scope, int32_t estimate);

//
void
VNTableScope_delete(VNTableScope this);

//
bool
VNTableScope_insert(VNTableScope this, VNTableKey key, Operation operation);

//
struct ValueNumber_ {
  //@args	Memory memory, SSAForm ssaForm
  Memory MEMORY;
  SSAForm SSAFORM;			// The SSAForm this ValueNumber applies to.
  CodeRegion CODEREGION;
  struct VNTableScope_ *CURRENT_SCOPE;
  Dominance DOMINANCE;
  bool DOMINANCE_BASED;
};

extern ValueNumber
ValueNumber_Ctor(ValueNumber this, Memory memory, SSAForm ssaForm);

extern ValueNumber
ValueNumber_Copy(ValueNumber this, const_ValueNumber that);

extern void
ValueNumber_Dtor(ValueNumber this);

extern size_t
ValueNumber_Size(Memory memory, SSAForm ssaForm);

#define ValueNumber_MEMORY(this) ((this)->MEMORY)
#define ValueNumber__MEMORY(this) (&(this)->MEMORY)
#define ValueNumber_SSAFORM(this) ((this)->SSAFORM)
#define ValueNumber__SSAFORM(this) (&(this)->SSAFORM)
#define ValueNumber_CODEREGION(this) ((this)->CODEREGION)
#define ValueNumber__CODEREGION(this) (&(this)->CODEREGION)
#define ValueNumber_CURRENT_SCOPE(this) ((this)->CURRENT_SCOPE)
#define ValueNumber__CURRENT_SCOPE(this) (&(this)->CURRENT_SCOPE)
#define ValueNumber_DOMINANCE(this) ((this)->DOMINANCE)
#define ValueNumber__DOMINANCE(this) (&(this)->DOMINANCE)
#define ValueNumber_DOMINANCE_BASED(this) ((this)->DOMINANCE_BASED)
#define ValueNumber__DOMINANCE_BASED(this) (&(this)->DOMINANCE_BASED)


/*
 * ValueNumber_INIT --	Initialize the ValueNumber module.
 */
#define ValueNumber_INIT()

/*
 * ValueNumber_FINI --	Finalize the ValueNumber module.
 */
#define ValueNumber_FINI()


#endif//ValueNumber__h_INCLUDED
