#ifndef Acyclic__h_INCLUDED
#define Acyclic__h_INCLUDED

//
#include "SCD/Acyclic.h"

/*
 * Acyclic_CRITICAL --	If zero, do not recompute CRITICAL from scratch on II increase.
 */
//#define Acyclic_CRITICAL 0
#ifndef Acyclic_CRITICAL
extern int Acyclic_CRITICAL;
#endif//Acyclic_CRITICAL

/*
 * Acyclic_CONVERGE --	If not zero, try to infer the next II
 */
//#define Acyclic_CONVERGE 0
#ifndef Acyclic_CONVERGE
extern int Acyclic_CONVERGE;
#endif//Acyclic_CONVERGE

//
struct AcyclicTask_ {
  //@args	DDGraphNode node, Instance instance, AcyclicTask pred, AcyclicTask succ
  InstructionTask_ __;			// Base class for this AcyclicTask.
  //@access SCHEDULING	InstructionTask_SCHEDULING(AcyclicTask____(this))
  //@access RESERVATION	InstructionTask_RESERVATION(AcyclicTask____(this))
  //@access PROCESSING	InstructionTask_PROCESSING(AcyclicTask____(this))
  //@access ITERATION	InstructionTask_ITERATION(AcyclicTask____(this))
  //@mutate ITERATION	InstructionTask__ITERATION(AcyclicTask____(this))
  //@access STARTDATE	InstructionTask_STARTDATE(AcyclicTask____(this))
  //@mutate STARTDATE	InstructionTask__STARTDATE(AcyclicTask____(this))
  //@access RELEASE	InstructionTask_RELEASE(AcyclicTask____(this))
  //@mutate RELEASE	InstructionTask__RELEASE(AcyclicTask____(this))
  //@access CRITICAL	InstructionTask_CRITICAL(AcyclicTask____(this))
  //@mutate CRITICAL	InstructionTask__CRITICAL(AcyclicTask____(this))
  //@access PREDCOUNT	InstructionTask_PREDCOUNT(AcyclicTask____(this))
  //@mutate PREDCOUNT	InstructionTask__PREDCOUNT(AcyclicTask____(this))
  DDGraphNode NODE;			// The associated DDGraphNode.
  //@access INDEX	DDGraphNode_INDEX(AcyclicTask_NODE(this))
  struct AcyclicTask_ *PRED;		// Predecessor AcyclicTask in regular chain.
  struct AcyclicTask_ *SUCC;		// Successor AcyclicTask in regular chain.
  struct AcyclicTask_ *NEXT;		// Next AcyclicTask in issue group.
};
typedef struct AcyclicTask_ AcyclicTask_, *AcyclicTask;
typedef const struct AcyclicTask_ *const_AcyclicTask;
typedef struct AcyclicTask_ * restrict_AcyclicTask;

extern AcyclicTask
AcyclicTask_Ctor(AcyclicTask this, DDGraphNode node, Instance instance, AcyclicTask pred, AcyclicTask succ);

extern AcyclicTask
AcyclicTask_Copy(AcyclicTask this, const_AcyclicTask that);

extern void
AcyclicTask_Dtor(AcyclicTask this);

extern size_t
AcyclicTask_Size(DDGraphNode node, Instance instance, AcyclicTask pred, AcyclicTask succ);

#define AcyclicTask___(this) ((this)->__)
#define AcyclicTask____(this) (&(this)->__)
#define AcyclicTask_NODE(this) ((this)->NODE)
#define AcyclicTask__NODE(this) (&(this)->NODE)
#define AcyclicTask_PRED(this) ((this)->PRED)
#define AcyclicTask__PRED(this) (&(this)->PRED)
#define AcyclicTask_SUCC(this) ((this)->SUCC)
#define AcyclicTask__SUCC(this) (&(this)->SUCC)
#define AcyclicTask_NEXT(this) ((this)->NEXT)
#define AcyclicTask__NEXT(this) (&(this)->NEXT)
#define AcyclicTask_CRITICAL(this) (InstructionTask_CRITICAL(AcyclicTask____(this)))
#define AcyclicTask_RESERVATION(this) (InstructionTask_RESERVATION(AcyclicTask____(this)))
#define AcyclicTask_RELEASE(this) (InstructionTask_RELEASE(AcyclicTask____(this)))
#define AcyclicTask_PREDCOUNT(this) (InstructionTask_PREDCOUNT(AcyclicTask____(this)))
#define AcyclicTask_ITERATION(this) (InstructionTask_ITERATION(AcyclicTask____(this)))
#define AcyclicTask_STARTDATE(this) (InstructionTask_STARTDATE(AcyclicTask____(this)))
#define AcyclicTask_INDEX(this) (DDGraphNode_INDEX(AcyclicTask_NODE(this)))
#define AcyclicTask_SCHEDULING(this) (InstructionTask_SCHEDULING(AcyclicTask____(this)))
#define AcyclicTask_PROCESSING(this) (InstructionTask_PROCESSING(AcyclicTask____(this)))
#define AcyclicTask__CRITICAL(this) (InstructionTask__CRITICAL(AcyclicTask____(this)))
#define AcyclicTask__RELEASE(this) (InstructionTask__RELEASE(AcyclicTask____(this)))
#define AcyclicTask__PREDCOUNT(this) (InstructionTask__PREDCOUNT(AcyclicTask____(this)))
#define AcyclicTask__ITERATION(this) (InstructionTask__ITERATION(AcyclicTask____(this)))
#define AcyclicTask__STARTDATE(this) (InstructionTask__STARTDATE(AcyclicTask____(this)))


//
bool
AcyclicTask_pretty(const_AcyclicTask this, FILE *file, const AcyclicTask_ *tasksBase);

//
struct AcyclicState_ {
  //@args	Memory memory, int bodyCount, int lambda, int barOmega
  PtrSeq_ GROUPS[1];		// Schedule groups as linked lists of AcyclicTask(s).
  //@access MEMORY	PtrSeq_memory(AcyclicState_GROUPS(this))
  //@access TIMESLOT	(PtrSeq_count(AcyclicState_GROUPS(this)) - 1)
  int16_t LAMBDA;		// Current initiation interval.
  int16_t BAROMEGA;		// Maximum omega in the dependence graph.
  int16_t BODYCOUNT;		// Count of Operation(s) in loop body.
  int16_t ERRORCOUNT;		// Count of AcyclicTask(s) inside the window minus bodyCount.
  int16_t BEGINDATE;		// Start date of the time window.
  int16_t ERRORDATE;		// Latest date with convergence error (ERRORCOUNT != 0).
  int16_t PREVSPAN;		// Previous makeSpan.
  int16_t DELTACOUNT;		// Count of makeSpan deltas.
  int32_t DELTATOTAL;		// Total of makeSpan deltas.
};
typedef struct AcyclicState_ AcyclicState_, *AcyclicState;
typedef const struct AcyclicState_ *const_AcyclicState;
typedef struct AcyclicState_ * restrict_AcyclicState;

extern AcyclicState
AcyclicState_Ctor(AcyclicState this, Memory memory, int bodyCount, int lambda, int barOmega);

extern AcyclicState
AcyclicState_Copy(AcyclicState this, const_AcyclicState that);

extern void
AcyclicState_Dtor(AcyclicState this);

extern size_t
AcyclicState_Size(Memory memory, int bodyCount, int lambda, int barOmega);

#define AcyclicState_GROUPS(this) ((this)->GROUPS)
#define AcyclicState__GROUPS(this) (&(this)->GROUPS)
#define AcyclicState_LAMBDA(this) ((this)->LAMBDA)
#define AcyclicState__LAMBDA(this) (&(this)->LAMBDA)
#define AcyclicState_BAROMEGA(this) ((this)->BAROMEGA)
#define AcyclicState__BAROMEGA(this) (&(this)->BAROMEGA)
#define AcyclicState_BODYCOUNT(this) ((this)->BODYCOUNT)
#define AcyclicState__BODYCOUNT(this) (&(this)->BODYCOUNT)
#define AcyclicState_ERRORCOUNT(this) ((this)->ERRORCOUNT)
#define AcyclicState__ERRORCOUNT(this) (&(this)->ERRORCOUNT)
#define AcyclicState_BEGINDATE(this) ((this)->BEGINDATE)
#define AcyclicState__BEGINDATE(this) (&(this)->BEGINDATE)
#define AcyclicState_ERRORDATE(this) ((this)->ERRORDATE)
#define AcyclicState__ERRORDATE(this) (&(this)->ERRORDATE)
#define AcyclicState_PREVSPAN(this) ((this)->PREVSPAN)
#define AcyclicState__PREVSPAN(this) (&(this)->PREVSPAN)
#define AcyclicState_DELTACOUNT(this) ((this)->DELTACOUNT)
#define AcyclicState__DELTACOUNT(this) (&(this)->DELTACOUNT)
#define AcyclicState_DELTATOTAL(this) ((this)->DELTATOTAL)
#define AcyclicState__DELTATOTAL(this) (&(this)->DELTATOTAL)
#define AcyclicState_TIMESLOT(this) ((PtrSeq_count(AcyclicState_GROUPS(this)) - 1))
#define AcyclicState_MEMORY(this) (PtrSeq_memory(AcyclicState_GROUPS(this)))


//
static inline bool
AcyclicState_converged(AcyclicState this, const_AcyclicTask firstTask)
{
  int lambda = AcyclicState_LAMBDA(this);
  int barOmega = AcyclicState_BAROMEGA(this);
  int bodyCount = AcyclicState_BODYCOUNT(this);
  int errorDate = AcyclicState_ERRORDATE(this);
  int iteration = AcyclicTask_ITERATION(firstTask);
  int firstDate = AcyclicTask_STARTDATE(firstTask);
  const_AcyclicTask lastTask = firstTask + bodyCount - 1;
  int processing = AcyclicTask_PROCESSING(lastTask);
  int lastDate = AcyclicTask_STARTDATE(lastTask);
  int makeSpan = lastDate + processing - firstDate;
  int stableDate = errorDate + 2 - lambda;
  int stabilized = lastDate - stableDate;
  // Strict convergence checks.
  if (stabilized >= makeSpan) {
    int successive = stabilized/lambda;
    if (successive >= barOmega) {
      int hatOmega = (makeSpan - 1)/lambda;
      if (successive >= hatOmega) {
        SCD_TRACE && fprintf(SCD_TRACE, "stableDate=%d\tmakeSpan=%d\tstabilized=%d\tbarOmega=%d\n",
                                        stableDate, makeSpan, stabilized, barOmega);
        return true;
      }
    }
  }
  return false;
}

//
bool
AcyclicState_continue(AcyclicState this, const_AcyclicTask firstTask);

//
bool
AcyclicState_pretty(const_AcyclicState this, FILE *file);

//
bool
AcyclicState_latex(const_AcyclicState this, FILE *file);

//
struct Acyclic_ {
  //@args Memory memory, Scheduler scheduler
  Scheduler SCHEDULER;
  ResourceTable_ *TABLE;
  //@access isInnerLoop	Scheduler_isInnerLoop(Acyclic_SCHEDULER(this))
  //@access PROCESSOR	Scheduler_PROCESSOR(Acyclic_SCHEDULER(this))
  //@access PIPELINING	Scheduler_PIPELINING(Acyclic_SCHEDULER(this))
  //@access MCRLGRAPH	Scheduler_MCRLGRAPH(Acyclic_SCHEDULER(this))
  //@access FIRSTINDEX	Scheduler_FIRSTINDEX(Acyclic_SCHEDULER(this))
  //@access LASTINDEX	Scheduler_LASTINDEX(Acyclic_SCHEDULER(this))
  //@access BAROMEGA	Scheduler_BAROMEGA(Acyclic_SCHEDULER(this))
  //@access UNWINDING	Scheduler_UNWINDING(Acyclic_SCHEDULER(this))
  PQueue_ WAITING[1];		// All predecessors are scheduled.
  PQueue_ READY[1];		// Data-ready tasks for scheduling.
  IStack_ TASKS[1];		// The set of unwinded tasks.
  //@access MEMORY	IStack_memory(Acyclic_TASKS(this))
  PtrSeq_ NODESTACK[1];		// DDGraphNode stack for priority computation.
  bool *INSTACK;		// Flag the DDGraphNode(s) that are on NODESTACK;
  int32_t BODYCOUNT;
  int32_t SCHEDULED;
};
typedef struct Acyclic_ Acyclic_, *Acyclic;
typedef const struct Acyclic_ *const_Acyclic;
typedef struct Acyclic_ * restrict_Acyclic;

extern Acyclic
Acyclic_Ctor(Acyclic this, Memory memory, Scheduler scheduler);

extern Acyclic
Acyclic_Copy(Acyclic this, const_Acyclic that);

extern void
Acyclic_Dtor(Acyclic this);

extern size_t
Acyclic_Size(Memory memory, Scheduler scheduler);

#define Acyclic_SCHEDULER(this) ((this)->SCHEDULER)
#define Acyclic__SCHEDULER(this) (&(this)->SCHEDULER)
#define Acyclic_TABLE(this) ((this)->TABLE)
#define Acyclic__TABLE(this) (&(this)->TABLE)
#define Acyclic_WAITING(this) ((this)->WAITING)
#define Acyclic__WAITING(this) (&(this)->WAITING)
#define Acyclic_READY(this) ((this)->READY)
#define Acyclic__READY(this) (&(this)->READY)
#define Acyclic_TASKS(this) ((this)->TASKS)
#define Acyclic__TASKS(this) (&(this)->TASKS)
#define Acyclic_NODESTACK(this) ((this)->NODESTACK)
#define Acyclic__NODESTACK(this) (&(this)->NODESTACK)
#define Acyclic_INSTACK(this) ((this)->INSTACK)
#define Acyclic__INSTACK(this) (&(this)->INSTACK)
#define Acyclic_BODYCOUNT(this) ((this)->BODYCOUNT)
#define Acyclic__BODYCOUNT(this) (&(this)->BODYCOUNT)
#define Acyclic_SCHEDULED(this) ((this)->SCHEDULED)
#define Acyclic__SCHEDULED(this) (&(this)->SCHEDULED)
#define Acyclic_BAROMEGA(this) (Scheduler_BAROMEGA(Acyclic_SCHEDULER(this)))
#define Acyclic_MCRLGRAPH(this) (Scheduler_MCRLGRAPH(Acyclic_SCHEDULER(this)))
#define Acyclic_isInnerLoop(this) (Scheduler_isInnerLoop(Acyclic_SCHEDULER(this)))
#define Acyclic_FIRSTINDEX(this) (Scheduler_FIRSTINDEX(Acyclic_SCHEDULER(this)))
#define Acyclic_PROCESSOR(this) (Scheduler_PROCESSOR(Acyclic_SCHEDULER(this)))
#define Acyclic_UNWINDING(this) (Scheduler_UNWINDING(Acyclic_SCHEDULER(this)))
#define Acyclic_MEMORY(this) (IStack_memory(Acyclic_TASKS(this)))
#define Acyclic_LASTINDEX(this) (Scheduler_LASTINDEX(Acyclic_SCHEDULER(this)))
#define Acyclic_PIPELINING(this) (Scheduler_PIPELINING(Acyclic_SCHEDULER(this)))


/*
 * Acyclic_tasks --	For use by Acyclic_FOREACH_AcyclicTask.
 */
static inline const_IStack
Acyclic_tasks(const_Acyclic this)
{
  return Acyclic_TASKS(this);
}

/*
 * Acyclic_FOREACH_AcyclicTask --	Iterate over this Acyclic AcyclicTask(s).
 */
#define Acyclic_FOREACH_AcyclicTask(this, task) { \
  const_IStack Acyclic_TASKS = Acyclic_tasks(this); \
  IStack_FOREACH(Acyclic_TASKS, AcyclicTask_, task) {
#define Acyclic_ENDEACH_AcyclicTask \
  } IStack_ENDEACH; \
}

/*
 * Acyclic_FORBACK_AcyclicTask --	Iterate backward over this Acyclic AcyclicTask(s).
 */
#define Acyclic_FORBACK_AcyclicTask(this, task) { \
  const_IStack Acyclic_TASKS = Acyclic_tasks(this); \
  IStack_FORBACK(Acyclic_TASKS, AcyclicTask_, task) {
#define Acyclic_ENDBACK_AcyclicTask \
  } IStack_ENDBACK; \
}

/*
 * Acyclic_enterTasks --	Enter the AcyclicTask(s) in this Acyclic.
 */
void
Acyclic_enterTasks(Acyclic this);

//
void
AcyclicTask_initPredCount(AcyclicTask this, int32_t firstIndex, int32_t lastIndex, IStack tasks);

//
void
Acyclic_reset(Acyclic this);

/*
 * Acyclic_makeCritical --	Compute AcyclicTask critical path length
 *
 * Take advantage from the fact that AcyclicTask(s) are entered in topological order.
 */

//
bool
Acyclic_checkCritical(Acyclic this, int lambda);

/*
 * Acyclic_trySchedule --	Schedule this Acyclic tasks based on their CRITICAL(ity).
 *
 * When all predecessors of an AcyclicTask are scheduled, it goes into a WAITING
 * queue in order to become data-ready. Once data-ready, it is transferred to a
 * READY queue so the highest CRITICAL(ity) are tried for scheduling first.
 */
AcyclicTask
Acyclic_trySchedule(Acyclic this, AcyclicState state);

//
void
Acyclic_notifySuccessors(Acyclic this, AcyclicTask task, int lambda);

//
bool
Acyclic_pretty(const_Acyclic this, FILE *file);

//
int
Acyclic_violatedDependencies(Acyclic this, AcyclicTask firstTask, int lambda);

//
void
Acyclic_commitSchedule(Acyclic this, AcyclicTask firstTask, int lambda);

/*
 * Acyclic_INIT --	Initialize the Acyclic module.
 */
void
Acyclic_INIT(void);

/*
 * Acyclic_FINI -- Finalize the Acyclic module.
 */
#define Acyclic_FINI()


#endif//Acyclic__h_INCLUDED
