#ifndef Construction_h_INCLUDED
#define Construction_h_INCLUDED

/*
 * !!!!	Construction.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

/*
 * Construction --	Implementation of software pipeline construction.
 */
struct Construction_;
  //@args	Memory memory, RegionFlow regionFlow, BlockTrace blockTrace
typedef struct Construction_ Construction_, *Construction;
typedef const struct Construction_ *const_Construction;
typedef struct Construction_ * restrict_Construction;

extern Construction
Construction_Ctor(Construction this, Memory memory, RegionFlow regionFlow, BlockTrace blockTrace);

extern Construction
Construction_Copy(Construction this, const_Construction that);

extern void
Construction_Dtor(Construction this);

extern size_t
Construction_Size(Memory memory, RegionFlow regionFlow, BlockTrace blockTrace);

//
Construction
Construction_make(Memory memory, RegionFlow regionFlow, BlockTrace blockTrace);

//
Construction
Construction_kill(Construction this);

/*
 * Construction_issueItems --	For use by Construction_FOREACH_IssueItem.
 */
IDList
Construction_issueItems(Construction this);

/*
 * Construction_FOREACH_IssueItem --	Iterate over this Construction IssueItems.
 */
#define Construction_FOREACH_IssueItem(this, issueItem) \
  IDList_FOREACH(Construction_issueItems(this), IssueItem_, issueItem) {
#define Construction_ENDEACH_IssueItem \
  } IDList_ENDEACH;

/*
 * Construction_FORBACK_IssueItem --	Iterate over this Construction IssueItems.
 */
#define Construction_FORBACK_IssueItem(this, issueItem) \
  IDList_FORBACK(Construction_issueItems(this), IssueItem_, issueItem) {
#define Construction_ENDBACK_IssueItem \
  } IDList_ENDBACK;

//
/*
 * Construction_prettyIssueItems --	Pretty-print this Construction IssueItems.
 */
bool
Construction_prettyIssueItems(Construction this, FILE *file);

/*
 * Construction_epilogs --	For use by Construction_FOREACH_ConstructionEpilog.
 */
IDList
Construction_epilogs(Construction this);

/*
 * Construction_FOREACH_ConstructionEpilog --	Iterate over this Construction ConstructionEpilogs.
 */
#define Construction_FOREACH_ConstructionEpilog(this, epilog) \
  IDList_FOREACH(Construction_epilogs(this), ConstructionEpilog_, epilog) {
#define Construction_ENDEACH_ConstructionEpilog \
  } IDList_ENDEACH;

/*
 * Construction_FORBACK_ConstructionEpilog --	Iterate over this Construction ConstructionEpilogs.
 */
#define Construction_FORBACK_ConstructionEpilog(this, epilog) \
  IDList_FORBACK(Construction_epilogs(this), ConstructionEpilog_, epilog) {
#define Construction_ENDBACK_ConstructionEpilog \
  } IDList_ENDBACK;

/*
 * ConstructionMasterFlag --      Enumerates the ConstructionMaster flags.
 */
typedef enum {
  ConstructionMasterFlag_Unscheduled = 0x1,    // ConstructionMaster contents are not scheduled.
} ConstructionMasterFlag;
typedef unsigned ConstructionMasterFlags;

/*
 * Construction_masters --	For use by Construction_FOREACH_ConstructionMaster.
 */
IDList
Construction_masters(Construction this);

/*
 * Construction_FOREACH_ConstructionMaster --	Iterate over this Construction ConstructionMasters.
 */
#define Construction_FOREACH_ConstructionMaster(this, master) \
  IDList_FOREACH(Construction_masters(this), ConstructionMaster_, master) {
#define Construction_ENDEACH_ConstructionMaster \
  } IDList_ENDEACH;

/*
 * Construction_FORBACK_ConstructionMaster --	Iterate over this Construction ConstructionMasters.
 */
#define Construction_FORBACK_ConstructionMaster(this, master) \
  IDList_FORBACK(Construction_masters(this), ConstructionMaster_, master) {
#define Construction_ENDBACK_ConstructionMaster \
  } IDList_ENDBACK;

/*
 * Construction_construct --	Main entry point.
 */
void
Construction_construct(Construction this, bool anyExit);

/*
 * Construction_fixSpeculated --	Fix the speculated Operations.
 */
void
Construction_fixSpeculated(Construction this);

/*
 * ConstructionEpilog --	Epilog block for the software pipeline Construction.
 */
struct ConstructionEpilog_;
  //@access NEXT		(ConstructionEpilog)(IDListItem_NEXT(this))
  //@access PREV		(ConstructionEpilog)(IDListItem_PREV(this))
typedef struct ConstructionEpilog_ ConstructionEpilog_, *ConstructionEpilog;
typedef const struct ConstructionEpilog_ *const_ConstructionEpilog;
typedef struct ConstructionEpilog_ * restrict_ConstructionEpilog;

#define ConstructionEpilog_NEXT(this) ((ConstructionEpilog)(IDListItem_NEXT(this)))
#define ConstructionEpilog_PREV(this) ((ConstructionEpilog)(IDListItem_PREV(this)))
//
/*
 * ConstructionEpilog_pretty --	Pretty-print this ConstructionEpilog.
 */
bool
ConstructionEpilog_pretty(ConstructionEpilog this, FILE *file);

/*
 * ConstructionMaster --	Master block for the software pipeline Construction.
 */
struct ConstructionMaster_;
  //@access NEXT		(ConstructionMaster)(IDListItem_NEXT(this))
  //@access PREV		(ConstructionMaster)(IDListItem_PREV(this))
typedef struct ConstructionMaster_ ConstructionMaster_, *ConstructionMaster;
typedef const struct ConstructionMaster_ *const_ConstructionMaster;
typedef struct ConstructionMaster_ * restrict_ConstructionMaster;

#define ConstructionMaster_NEXT(this) ((ConstructionMaster)(IDListItem_NEXT(this)))
#define ConstructionMaster_PREV(this) ((ConstructionMaster)(IDListItem_PREV(this)))
//
/*
 * ConstructionMaster_pretty --	Pretty-print this ConstructionMaster.
 */
bool
ConstructionMaster_pretty(ConstructionMaster this, FILE *file);


#endif//Construction_h_INCLUDED
