#ifndef Construction__h_INCLUDED
#define Construction__h_INCLUDED

//
#include "SCD/Construction.h"

//
struct Construction_ {
  RegionFlow REGIONFLOW;		// This Construction enclosing Procedure.
  BlockTrace BLOCKTRACE;	// This Construction BlockTrace.
  //@access PROCESSOR	BlockTrace_PROCESSOR(Construction_BLOCKTRACE(this))
  //@access LAMBDA	BlockTrace_LAMBDA(Construction_BLOCKTRACE(this))
  //@access LENGTH	BlockTrace_LENGTH(Construction_BLOCKTRACE(this))
  //@access KUNROLL	BlockTrace_KUNROLL(Construction_BLOCKTRACE(this))
  IDList_ ISSUEITEMS_;		// This Construction IssueItem list.
  //@access ISSUEITEMS	Construction__ISSUEITEMS_(this)
  //@access FIRSTITEM	(IssueItem)IDList_firstItem(Construction_ISSUEITEMS(this))
  //@access LASTITEM	(IssueItem)IDList_lastItem(Construction_ISSUEITEMS(this))
  //@access ITEMCOUNT	IDList_count(Construction_ISSUEITEMS(this))
  //@access MEMORY	IDList_memory(Construction_ISSUEITEMS(this))
  IssueItem_ ENDITEM_;		// The end IssueItem of this Construction.
  //@access ENDITEM	Construction__ENDITEM_(this)
  IDList_ MASTERS_;		// The list of ConstructionMasters of this Construction.
  //@access MASTERS	Construction__MASTERS_(this)
  //@access FIRSTMASTER	(ConstructionMaster)IDList_firstItem(Construction_MASTERS(this))
  //@access LASTMASTER	(ConstructionMaster)IDList_lastItem(Construction_MASTERS(this))
  //@access MASTERCOUNT	IDList_count(Construction_MASTERS(this))
  IDList_ EPILOGS_;		// The list of ConstructionEpilogs of this Construction.
  //@access EPILOGS	Construction__EPILOGS_(this)
  BasicBlock PREDBLOCK;		// Predecessor BasicBlock of the BlockTrace.
  BasicBlock HEADBLOCK;		// Head BasicBlock of the BlockTrace.
  intptr_t TRACEID;		// BasicBlock TRACEID of this Construction.
  int32_t REGIONID;		// REGIONID of the BlockTrace.
  int16_t UNROLLED;		// The unrolled of the constructed BasicBlock(s).
  float FREQUENCY;		// Frequency the BlockTrace is executed.
};

extern Construction
Construction_Ctor(Construction this, Memory memory, RegionFlow regionFlow, BlockTrace blockTrace);

extern Construction
Construction_Copy(Construction this, const_Construction that);

extern void
Construction_Dtor(Construction this);

extern size_t
Construction_Size(Memory memory, RegionFlow regionFlow, BlockTrace blockTrace);

#define Construction_REGIONFLOW(this) ((this)->REGIONFLOW)
#define Construction__REGIONFLOW(this) (&(this)->REGIONFLOW)
#define Construction_BLOCKTRACE(this) ((this)->BLOCKTRACE)
#define Construction__BLOCKTRACE(this) (&(this)->BLOCKTRACE)
#define Construction_ISSUEITEMS_(this) ((this)->ISSUEITEMS_)
#define Construction__ISSUEITEMS_(this) (&(this)->ISSUEITEMS_)
#define Construction_ENDITEM_(this) ((this)->ENDITEM_)
#define Construction__ENDITEM_(this) (&(this)->ENDITEM_)
#define Construction_MASTERS_(this) ((this)->MASTERS_)
#define Construction__MASTERS_(this) (&(this)->MASTERS_)
#define Construction_EPILOGS_(this) ((this)->EPILOGS_)
#define Construction__EPILOGS_(this) (&(this)->EPILOGS_)
#define Construction_PREDBLOCK(this) ((this)->PREDBLOCK)
#define Construction__PREDBLOCK(this) (&(this)->PREDBLOCK)
#define Construction_HEADBLOCK(this) ((this)->HEADBLOCK)
#define Construction__HEADBLOCK(this) (&(this)->HEADBLOCK)
#define Construction_TRACEID(this) ((this)->TRACEID)
#define Construction__TRACEID(this) (&(this)->TRACEID)
#define Construction_REGIONID(this) ((this)->REGIONID)
#define Construction__REGIONID(this) (&(this)->REGIONID)
#define Construction_UNROLLED(this) ((this)->UNROLLED)
#define Construction__UNROLLED(this) (&(this)->UNROLLED)
#define Construction_FREQUENCY(this) ((this)->FREQUENCY)
#define Construction__FREQUENCY(this) (&(this)->FREQUENCY)
#define Construction_ITEMCOUNT(this) (IDList_count(Construction_ISSUEITEMS(this)))
#define Construction_LAMBDA(this) (BlockTrace_LAMBDA(Construction_BLOCKTRACE(this)))
#define Construction_EPILOGS(this) (Construction__EPILOGS_(this))
#define Construction_LENGTH(this) (BlockTrace_LENGTH(Construction_BLOCKTRACE(this)))
#define Construction_LASTMASTER(this) ((ConstructionMaster)IDList_lastItem(Construction_MASTERS(this)))
#define Construction_ISSUEITEMS(this) (Construction__ISSUEITEMS_(this))
#define Construction_MASTERS(this) (Construction__MASTERS_(this))
#define Construction_LASTITEM(this) ((IssueItem)IDList_lastItem(Construction_ISSUEITEMS(this)))
#define Construction_KUNROLL(this) (BlockTrace_KUNROLL(Construction_BLOCKTRACE(this)))
#define Construction_FIRSTMASTER(this) ((ConstructionMaster)IDList_firstItem(Construction_MASTERS(this)))
#define Construction_PROCESSOR(this) (BlockTrace_PROCESSOR(Construction_BLOCKTRACE(this)))
#define Construction_MASTERCOUNT(this) (IDList_count(Construction_MASTERS(this)))
#define Construction_MEMORY(this) (IDList_memory(Construction_ISSUEITEMS(this)))
#define Construction_FIRSTITEM(this) ((IssueItem)IDList_firstItem(Construction_ISSUEITEMS(this)))
#define Construction_ENDITEM(this) (Construction__ENDITEM_(this))


/*
 * Construction_makeIssueItem --	Make a new IssueItem for this Construction.
 *
 * The Operation_issueItem is set to the new IssueItem.
 */
IssueItem
Construction_makeIssueItem(Construction this, IssueItem issueItem);

/*
 * Construction_makeEpilog --	Make a new ConstructionEpilog for this Construction.
 */
ConstructionEpilog
Construction_makeEpilog(Construction this, IssueItem issueItem);

//
void
Construction_sortEpilogs(Construction this);

/*
 * Construction_makeMaster --	Make a new ConstructionMaster for this Construction.
 */
void
Construction_makeMaster(Construction this, IssueItem issueItem, BasicBlock basicBlock);

/*
 * Construction_makeEndItem --	Step 1 of "A Unified Software Pipeline
 * Construction Scheme for Modulo Scheduler Loops", PaCT'97.
 *
 * The purpose of Step 1 is to compute endItem, the IssueItem at the end of the
 * master schedule. This is the earliest shifted copy of an exit branch that is
 * not before the last in the local schedule. In case of software pipelined
 * hardware loops, the kernel boundaries can be any operation  (+anyExit+ true).
 */
void
Construction_makeEndItem(Construction this, bool anyExit);

/*
 * Construction_buildMaster --	Step 2 of "A Unified Software Pipeline
 * Construction Scheme for Modulo Scheduler Loops", PaCT'97.
 *
 * The purpose of Step 2 is to create the master schedule. We omit the first and
 * the last operations from the replication process, as these operations are
 * scheduling anchors that should exist only once in the master schedule. Also
 * after Step2, no operation remains inside a BasicBlock, and the BasicBlock(s)
 * to be reused for the schedule kernel are into Construction_MASTERS.
 */
void
Construction_buildMaster(Construction this);

/*
 * Construction_splitMaster --	Split this Construction MASTERS.
 *
 * The limit item between the prolog and the kernel is the down-shifted copy by
 * kunroll of the ENDITEM.
 */
void
Construction_splitMaster(Construction this);

/*
 * Construction_fillEpilogs --	Step 3 of "A Unified Software Pipeline
 * Construction Scheme for Modulo Scheduler Loops", PaCT'97.
 *
 * The purpose of Step 3 is to fill the epilogs with operations.
 * The first difference with the paper is that we iterate over the master
 * schedule items, instead of bumping the ITERATION and PHI of a temporary
 * IssueItem. The second difference is that the two loops are inverted.
 */
void
Construction_fillEpilogs(Construction this);

/*
 * Construction_completeEpilogs --	Complete this Construction EPILOGS.
 */
void
Construction_completeEpilogs(Construction this);

/*
 * Construction_patchOperations --	Patch the operation arguments and results using
 * modulo renaming and inductive relaxation, as per "A Unified Software Pipeline
 * Construction Scheme for Modulo Scheduler Loops", PaCT'97.
 */
void
Construction_patchOperations(Construction this);

/*
 * Construction_moveRegisters --	Move the registers after modulo renaming, as per
 * "A Unified Software Pipeline Construction Scheme for Modulo Scheduler Loops", PaCT'97.
 *
 * Compared to the paper, we extended algorithm to work with counted hardware
 * loops. We also fixed two bugs:
 * * The renaming  register used by the MOVE operation into the register live
 * on exit is indexed using the iteration number of the exit branch minus its
 * Carried value.
 * * In case of a live on exit renamed register, only the lifetime that
 * contains the indirect use needs to generate the MOVE.
 */
void
Construction_moveRegisters(Construction this);

/*
 * Construction_fixCFGraph --	Fix the CFGraph for this Construction.
 *
 * Let headNode and tailNode be the head and the tail of the BlockTrace. Let
 * firstNode and lastNode be the first and the last nodes of the master schedule.
 * When not software pipelining: headNode == firstNode. When software pipelining:
 * firstNode is either the first prolog node, if firstNode != headNode, or is the
 * first kernel node in case there is no prolog; headNode is the first kernel
 * node; lastNode is the last kernel node. Note that, unlike headNode, tailNode
 * is not guaranteed to be used in the master schedule.
 */
void
Construction_fixCFGraph(Construction this);

/*
 * Construction_fixFrequencies --	Fix the CFGNode frequencies for this Construction.
 */
void
Construction_fixFrequencies(Construction this);

/*
 * Construction_fillBasicBlocks --	Fill the BasicBlock(s) of the master schedule.
 */
void
Construction_fillBasicBlocks(Construction this);

/*
 * Construction_fixBasicBlocks --	Fix the BasicBlock attributes.
 */
void
Construction_fixBasicBlocks(Construction this);

//
struct ConstructionEpilog_ {
  //@args	Construction construction, IssueItem issueItem
  Construction CONSTRUCTION;
  IssueItem ISSUEITEM;
  CFGNode CFGNODE;		// The CFGNode of this ConstructionEpilog.
  Selector_ SELECTOR_;
  //@access SELECTOR	ConstructionEpilog__SELECTOR_(this)
};

extern ConstructionEpilog
ConstructionEpilog_Ctor(ConstructionEpilog this, Construction construction, IssueItem issueItem);

extern ConstructionEpilog
ConstructionEpilog_Copy(ConstructionEpilog this, const_ConstructionEpilog that);

extern void
ConstructionEpilog_Dtor(ConstructionEpilog this);

extern size_t
ConstructionEpilog_Size(Construction construction, IssueItem issueItem);

#define ConstructionEpilog_CONSTRUCTION(this) ((this)->CONSTRUCTION)
#define ConstructionEpilog__CONSTRUCTION(this) (&(this)->CONSTRUCTION)
#define ConstructionEpilog_ISSUEITEM(this) ((this)->ISSUEITEM)
#define ConstructionEpilog__ISSUEITEM(this) (&(this)->ISSUEITEM)
#define ConstructionEpilog_CFGNODE(this) ((this)->CFGNODE)
#define ConstructionEpilog__CFGNODE(this) (&(this)->CFGNODE)
#define ConstructionEpilog_SELECTOR_(this) ((this)->SELECTOR_)
#define ConstructionEpilog__SELECTOR_(this) (&(this)->SELECTOR_)
#define ConstructionEpilog_SELECTOR(this) (ConstructionEpilog__SELECTOR_(this))


/*
 * ConstructionEpilog_FOREACH_Operation --	Iterate over this ConstructionEpilog Operations.
 */
#define ConstructionEpilog_FOREACH_Operation(this, operation) { \
  Selector ConstructionEpilog_SELECTOR = ConstructionEpilog_SELECTOR(this); \
  Selector_FOREACH_Operation(ConstructionEpilog_SELECTOR, operation)
#define ConstructionEpilog_ENDEACH_Operation \
  Selector_ENDEACH_Operation \
}

//
static inline void
ConstructionEpilog_pushOperation(ConstructionEpilog this, Operation operation)
{
  Selector selector = ConstructionEpilog_SELECTOR(this);
  Selector_attachLast(selector, operation);
}

//
int
ConstructionEpilog_compare(const void *this_pointer, const void *that_pointer);

//
struct ConstructionMaster_ {
  //@args	IssueItem issueItem, BasicBlock basicBlock
  IssueItem ISSUEITEM;		// The IssueItem that ends this Master.
  //@access EXITARC	IssueItem_EXITARC(ConstructionMaster_ISSUEITEM(this))
  //@mutate EXITARC	IssueItem__EXITARC(ConstructionMaster_ISSUEITEM(this))
  //@access FALLARC	IssueItem_FALLARC(ConstructionMaster_ISSUEITEM(this))
  //@mutate FALLARC	IssueItem__FALLARC(ConstructionMaster_ISSUEITEM(this))
  BasicBlock BASICBLOCK;	// The BasicBlock of this Master.
  CFGNode CFGNODE;		// The CFGNode of this Master.
  ConstructionEpilog EPILOG;
  ConstructionMasterFlags FLAGS; // This ConstructionMaster flags.
  //@access isUnscheduled ((ConstructionMaster_FLAGS(this)&ConstructionMasterFlag_Unscheduled) != 0)
};

extern ConstructionMaster
ConstructionMaster_Ctor(ConstructionMaster this, IssueItem issueItem, BasicBlock basicBlock);

extern ConstructionMaster
ConstructionMaster_Copy(ConstructionMaster this, const_ConstructionMaster that);

extern void
ConstructionMaster_Dtor(ConstructionMaster this);

extern size_t
ConstructionMaster_Size(IssueItem issueItem, BasicBlock basicBlock);

#define ConstructionMaster_ISSUEITEM(this) ((this)->ISSUEITEM)
#define ConstructionMaster__ISSUEITEM(this) (&(this)->ISSUEITEM)
#define ConstructionMaster_BASICBLOCK(this) ((this)->BASICBLOCK)
#define ConstructionMaster__BASICBLOCK(this) (&(this)->BASICBLOCK)
#define ConstructionMaster_CFGNODE(this) ((this)->CFGNODE)
#define ConstructionMaster__CFGNODE(this) (&(this)->CFGNODE)
#define ConstructionMaster_EPILOG(this) ((this)->EPILOG)
#define ConstructionMaster__EPILOG(this) (&(this)->EPILOG)
#define ConstructionMaster_FLAGS(this) ((this)->FLAGS)
#define ConstructionMaster__FLAGS(this) (&(this)->FLAGS)
#define ConstructionMaster_isUnscheduled(this) (((ConstructionMaster_FLAGS(this)&ConstructionMasterFlag_Unscheduled) != 0))
#define ConstructionMaster_FALLARC(this) (IssueItem_FALLARC(ConstructionMaster_ISSUEITEM(this)))
#define ConstructionMaster_EXITARC(this) (IssueItem_EXITARC(ConstructionMaster_ISSUEITEM(this)))
#define ConstructionMaster__FALLARC(this) (IssueItem__FALLARC(ConstructionMaster_ISSUEITEM(this)))
#define ConstructionMaster__EXITARC(this) (IssueItem__EXITARC(ConstructionMaster_ISSUEITEM(this)))


//
int
ConstructionMaster_compare(const void *this_pointer, const void *that_pointer);

/*
 * Construction_INIT --	Initialize the Construction module.
 */
void
Construction_INIT(void);

/*
 * Construction_FINI --	Finalize the Construction module.
 */
void
Construction_FINI(void);


#endif//Construction__h_INCLUDED
