#ifndef Insertion__h_INCLUDED
#define Insertion__h_INCLUDED

//
#include "SCD/Insertion.h"

/*
 * InsertionSlot --	Information for a particular Insertion slot.
 */
struct InsertionSlot_ {
  //@args	const uint8_t *availability, int tau, int phi
  IssueGroup_ GROUP_M_;	// IssueGroup_M from group splitting during insertion.
  //@access GROUP_M	InsertionSlot__GROUP_M_(this)
  IssueGroup_ GROUP_P_;	// IssueGroup_P from group splitting during insertion.
  //@access GROUP_P	InsertionSlot__GROUP_P_(this)
  int16_t TAU;			// STARTDATE mod LAMBDA.
  int16_t PHI;			// STARTDATE div LAMBDA.
  int8_t DELTA_M;		// DELTA_M value for Insertion.
  int8_t DELTA_P;		// DELTA_P value for Insertion.
};
typedef struct InsertionSlot_ InsertionSlot_, *InsertionSlot;
typedef const struct InsertionSlot_ *const_InsertionSlot;
typedef struct InsertionSlot_ * restrict_InsertionSlot;

extern InsertionSlot
InsertionSlot_Ctor(InsertionSlot this, const uint8_t *availability, int tau, int phi);

extern InsertionSlot
InsertionSlot_Copy(InsertionSlot this, const_InsertionSlot that);

extern void
InsertionSlot_Dtor(InsertionSlot this);

extern size_t
InsertionSlot_Size(const uint8_t *availability, int tau, int phi);

#define InsertionSlot_GROUP_M_(this) ((this)->GROUP_M_)
#define InsertionSlot__GROUP_M_(this) (&(this)->GROUP_M_)
#define InsertionSlot_GROUP_P_(this) ((this)->GROUP_P_)
#define InsertionSlot__GROUP_P_(this) (&(this)->GROUP_P_)
#define InsertionSlot_TAU(this) ((this)->TAU)
#define InsertionSlot__TAU(this) (&(this)->TAU)
#define InsertionSlot_PHI(this) ((this)->PHI)
#define InsertionSlot__PHI(this) (&(this)->PHI)
#define InsertionSlot_DELTA_M(this) ((this)->DELTA_M)
#define InsertionSlot__DELTA_M(this) (&(this)->DELTA_M)
#define InsertionSlot_DELTA_P(this) ((this)->DELTA_P)
#define InsertionSlot__DELTA_P(this) (&(this)->DELTA_P)
#define InsertionSlot_GROUP_M(this) (InsertionSlot__GROUP_M_(this))
#define InsertionSlot_GROUP_P(this) (InsertionSlot__GROUP_P_(this))


//
struct Insertion_ {
  //@args	Memory memory, Scheduler scheduler, int maxRange
  Memory MEMORY;
  Scheduler SCHEDULER;		// This Insertion Scheduler.
  int16_t MAXRANGE;		// Maximum range of start dates tried.
  int16_t SLOTCOUNT;		// The count of Insertion SLOTS.
  InsertionSlot_ SLOTS[2];	// The candidate InsertionSlots.
};

extern Insertion
Insertion_Ctor(Insertion this, Memory memory, Scheduler scheduler, int maxRange);

extern Insertion
Insertion_Copy(Insertion this, const_Insertion that);

extern void
Insertion_Dtor(Insertion this);

extern size_t
Insertion_Size(Memory memory, Scheduler scheduler, int maxRange);

#define Insertion_MEMORY(this) ((this)->MEMORY)
#define Insertion__MEMORY(this) (&(this)->MEMORY)
#define Insertion_SCHEDULER(this) ((this)->SCHEDULER)
#define Insertion__SCHEDULER(this) (&(this)->SCHEDULER)
#define Insertion_MAXRANGE(this) ((this)->MAXRANGE)
#define Insertion__MAXRANGE(this) (&(this)->MAXRANGE)
#define Insertion_SLOTCOUNT(this) ((this)->SLOTCOUNT)
#define Insertion__SLOTCOUNT(this) (&(this)->SLOTCOUNT)
#define Insertion_SLOTS(this) ((this)->SLOTS)
#define Insertion__SLOTS(this) (&(this)->SLOTS)


/*
 * Insertion_selectTask --	Select the next GrahamListTask to issue.
 */
GrahamListTask
Insertion_selectTask(Insertion this);

/*
 * Insertion_selectSlot --	Select an InsertionSlot for the task.
 */
InsertionSlot
Insertion_selectSlot(Insertion this, GrahamListTask task);

/*
 * Insertion_scheduleTaskAtSlot --	Schedule the previously selected GrahamListTask.
 */
int
Insertion_scheduleTaskAtSlot(Insertion this, GrahamListTask task, InsertionSlot slot);

/*
 * Insertion_schedule --	Main driver of the Insertion scheduling.
 *
 * The GrahamListTasks to issue are selected in heuristic order by
 * Insertion_selectTask.
 */
void
Insertion_schedule(Insertion this);

/*
 * Insertion_INIT --	Initialize the Insertion module.
 */
void
Insertion_INIT(void);

/*
 * Insertion_FINI --	Finalize the Insertion module.
 */
void
Insertion_FINI(void);


#endif//Insertion__h_INCLUDED
