#ifndef XFA_h_INCLUDED
#define XFA_h_INCLUDED

/*
 * !!!!	XFA.xcc
 *
 * Benoit Dupont de Dinechin (Benoit.Dupont-de-Dinechin@st.com).
 *
 * Copyright 2002 - 2007 STMicroelectronics.
 * Copyright 1995 - 1998 Commissariat a l'Energie Atomique.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of either (at your option): the GNU
 * General Public License (GPL) version 2; the GNU Lesser General
 * Public License (LGPL) version 2.1; any later version of these
 * licences as published by the Free Software Foundation.
 */

//
#include "CCL.h"
#include "CAL.h"
#include "MDS.h"
#include "AIR.h"
#include "LIR.h"
#include "PFA.h"
#include "PRO.h"

/*
 * The XFA (eXtra Flow Analyses) library contains the following modules:
 * *	<Issue Issue.html> Issue, IssueGroup and IssueTable.
 * *	<Effect Effect.html> Operation Effect and EffectTable.
 * *	<CFGraph CFGraph.html> Control-Flow Graph (to be replaced).
 * *	<RegionFlow RegionFlow.html> Region Flow (to be replaced).
 * *	<Lifetime Lifetime.html> Temporary Lifetime in a schedule.
 * *	<BlockTrace BlockTrace.html> Basic Block Trace (to be replaced).
 * *	<DDGraph DDGraph.html> Data Dependence Graph.
 */

//
#include "XFA/Issue.h"
#include "XFA/Effect.h"
#include "XFA/CFGraph.h"
#include "XFA/RegionFlow.h"
#include "XFA/Lifetime.h"
#include "XFA/BlockTrace.h"
#include "XFA/DDGraph.h"

/*
 * Variables used to control XFA in non-release mode.
 */
extern int FORCE_POSTCOMP;
extern int FORCE_POSTSPEC;
extern int FORCE_KUNROLL;
extern int FORCE_PRESERIAL;
extern int FORCE_POSTSERIAL;
extern unsigned DEBUG_BOOSTABLE;
extern unsigned DEBUG_SPECULABLE;
extern unsigned DEBUG_COMPENSABLE;

/*
 * XFA_INIT --	Initialize the XFA library.
 */
void
XFA_INIT(void);

/*
 * XFA_FINI --	Finalize the XFA library.
 */
void
XFA_FINI(void);


#endif//XFA_h_INCLUDED
