#ifndef GraphColoring_c_INCLUDED
#define GraphColoring_c_INCLUDED

#line 17 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * \file GraphColoring.xcc
 *
 * This module implements George/Appel's iterated register allocator.
 * - Input: Control-flow graph of a code region. Each node represents a basic block.
 * - Output: Spill code inserted and temporaries allocated into registers.
 *
 * I followed the same notation that we found in
 * "Modern Compiler Implementation in Java"
 *
 * To check the algorithm, run:
 * grep "\/\/ALGO:" GraphColoring.xcc |sed s/"\/\/ALGO:"/" "/g
 *
 * Note:
 * - \@XCC._h => Friend members declarations and xcc autocomplete structures.
 * - \@XCC.h => Public members declarations.
 * - \@XCC._c => Inline functions definitions.
 * - \@XCC.c => Private members declarations and members definitions.
 *
 */

#line 39 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/* TODO:
 *  - Tests.
 *  - Op Bool/Int
 *  - uint converted into int?
 */

#line 45 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#include "GRA/GRA_.h"

#line 50 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#include <time.h>

#line 52 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 153 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static float totalspillcost;
static int32_t statistiques[Stat__];

#line 210 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
WorkLists
WorkLists_Ctor(WorkLists this, Memory memory)
{
  IDList_Ctor(WorkLists_precolored(this), memory);
  IDList_Ctor(WorkLists_initial(this), memory);
  IDList_Ctor(WorkLists_spilledNodes(this), memory);
  IDList_Ctor(WorkLists_coalescedNodes(this), memory);
  IDList_Ctor(WorkLists_coloredNodes(this), memory);
  IDList_Ctor(WorkLists_discardedNodes(this), memory);
  IDList_Ctor(WorkLists_spillWorklist(this), memory);
  IDList_Ctor(WorkLists_freezeWorklist(this), memory);
  IDList_Ctor(WorkLists_simplifyWorklist(this), memory);
  IDList_Ctor(WorkLists_selectStack(this), memory);
  IDList_Ctor(WorkLists_worklistMoves(this), memory);
  IDList_Ctor(WorkLists_activeMoves(this), memory);
  IDList_Ctor(WorkLists_coalescedMoves(this), memory);
  IDList_Ctor(WorkLists_constrainedMoves(this), memory);
  IDList_Ctor(WorkLists_frozenMoves(this), memory);
  return this;
}

#line 231 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
void
WorkLists_Dtor(WorkLists this)
{
  IDList_empty(WorkLists_precolored(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_precolored(this));
  IDList_empty(WorkLists_initial(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_initial(this));
  IDList_empty(WorkLists_spilledNodes(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_spilledNodes(this));
  IDList_empty(WorkLists_coalescedNodes(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_coalescedNodes(this));
  IDList_empty(WorkLists_coloredNodes(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_coloredNodes(this));
  IDList_empty(WorkLists_discardedNodes(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_discardedNodes(this));
  IDList_empty(WorkLists_spillWorklist(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_spillWorklist(this));
  IDList_empty(WorkLists_freezeWorklist(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_freezeWorklist(this));
  IDList_empty(WorkLists_simplifyWorklist(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_simplifyWorklist(this));
  IDList_empty(WorkLists_selectStack(this), (void(*)(void*))InterfNode_Dtor);
  IDList_Dtor(WorkLists_selectStack(this));
  IDList_empty(WorkLists_worklistMoves(this), (void(*)(void*))InterfMove_Dtor);
  IDList_Dtor(WorkLists_worklistMoves(this));
  IDList_empty(WorkLists_activeMoves(this), (void(*)(void*))InterfMove_Dtor);
  IDList_Dtor(WorkLists_activeMoves(this));
  IDList_empty(WorkLists_coalescedMoves(this), (void(*)(void*))InterfMove_Dtor);
  IDList_Dtor(WorkLists_coalescedMoves(this));
  IDList_empty(WorkLists_constrainedMoves(this), (void(*)(void*))InterfMove_Dtor);
  IDList_Dtor(WorkLists_constrainedMoves(this));
  IDList_empty(WorkLists_frozenMoves(this), (void(*)(void*))InterfMove_Dtor);
  IDList_Dtor(WorkLists_frozenMoves(this));
}

#line 305 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
GraphColoring
GraphColoring_Ctor(GraphColoring this, Memory memory)
{
  *GraphColoring__MEMORY(this) = memory;
  *GraphColoring__SPILLS(this) = NULL;
  *GraphColoring__OPERATIONCOUNTER(this) = 0L;
  *GraphColoring__FLAGS(this) = GRA_gcf;
  return this;
}

#line 315 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
void
GraphColoring_Dtor(GraphColoring this)
{
}

#line 365 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//
/********************************
 * Private members declarations *
 ********************************/

#line 370 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Create and push (i.e., place at the end) a node into a worklist.
 * \param this object instance
 * \param worklist Add the node into this.
 * \param temporary The related temporary.
 * \param belongTo The list in which the node is added.
 * \return The InterfNode created.
 */
static InterfNode
AddNodeIntoWorklist(GraphColoring this, IDList worklist, Temporary temporary, NodeBelongTo belongTo);

#line 381 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Push (i.e., place at the end) a node into a worklist (no creation).
 * \param this object instance
 * \param worklist Push the node into this.
 * \param interfNode The related InterfNode
 * \param belongTo The list in which the node is added.
 */
static void
PushNodeIntoWorklist(GraphColoring this, IDList worklist, InterfNode interfNode, NodeBelongTo belongTo);

#line 391 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Push (i.e., place at the end) a move into a worklist.
 * \param worklist Add the move into this.
 * \param interfMove The related InterfMove
 * \param belongTo The list in which the move is added.
 * \return The InterfMove created.
 */
static void
PushMoveIntoWorklist(GraphColoring this, IDList worklist, InterfMove interfMove, MoveBelongTo belongTo);

#line 401 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns true if the temporary is associated with an
 * interfNode.
 * \param this object instance
 * \param temporary The related temporary.
 * \return true if the temporary has an associated node.
 */
static bool
GraphColoring_IsNode(GraphColoring this, Temporary temporary);

#line 411 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Get the node corresponding to a given temporary.
 * It is invalid to call this function if IsNode(temporary) == false.
 * \param this object instance
 * \param temporary The related temporary.
 * \return The node.
 */
static InterfNode
GraphColoring_GetNode(GraphColoring this, Temporary temporary);

#line 423 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Construct the interference graph.  It uses the control flow graph
 * to constructs the interference graph, and also initializes the
 * worklistMoves to contain all the related moves in the program.
 * \warning moveList, worklistMoves, and interfGraph must be
 * initialized and are destroyed!
 * \param this the graph coloring object
 * \return The number of temporary.
 */
static int32_t
Build(GraphColoring this);

#line 435 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Dummy function to calculate the number and size of multiplexing region
 */
static void
GC_multiplexRegion_traverse(BasicBlockEdge edge, int *num_edges, int *num_sources, int *num_dests, bool up );

#line 441 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Dummy function to calculate the number and size of multiplexing region
 */
bool
GC_multiplexRegion_pretty(GraphColoring this, FILE *file);

#line 447 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Add an edge to the interference graph.
 * \param this the graph coloring object
 * \param interfNode1 one end of the edge.
 * \param interfNode2 other end of the edge.
 * \return \e true in success, else \e false.
 */
static bool
AddEdge(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2);

#line 457 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Build initial worklists.
 * generate these (mutually disjoints) worklists:
 *  - spillWorklist
 *  - freezeWorklist
 *  - simplifyWorklist
 *  .
 * \param this the graph coloring object
 * \return Always \e true !?
 */
static bool
MakeWorklist(GraphColoring this);

#line 470 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * To know if a node is move-related.
 * \param interfNode The interference node.
 * \return \e true if the node is move-related, else \e false.
 */
static bool
MoveRelated(InterfNode interfNode);

#line 478 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Dump (human readable) Worklists.
 * Print all these worklists:
 *  - spillWorklist
 *  - freezeWorklist
 *  - simplifyWorklist
 *  - selectStack
 *  - .
 * \param file The file descriptor, where to print.
 */
bool
WorkLists_pretty(WorkLists this, FILE *file);

#line 491 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Remove a node from the graph.
 * It involves decrementing the degree of its current neighbors.
 * \param this the graph coloring object
 * \return Always \e true !?
 */
static bool
Simplify(GraphColoring this);

#line 500 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Decrement the degree of a node.  If the degree of a neighbor is
 * already less than K-1 then the neighbor must be move-related, and
 * is not added to the simplifyWorklist.  When the degree of a
 * neighbor transitions from K to K-1, moves associated with \e its
 * neighbors may be enable.
 * \param this the graph coloring object
 * \param interfNode The interference node.
 * \return Always \e true !?
 */
static bool
DecrementDegree(GraphColoring this, InterfNode interfNode);

#line 513 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
EnableMoves(GraphColoring this, InterfNode interfNode);

#line 516 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Perform conservative coalescing on the reduced graph obtained in
 * the simplification phase.  Since the degrees of many nodes have
 * been reduced by Simplify(), the conservative strategy is likely
 * to find many more moves to have been coalesced (and the move
 * instruction deleted), if the resulting node is no longer
 * move-related it will be available for the next round of
 * simplification. Simplify() and Coalesce() are repeated
 * until only significant-degree or move-related nodes remain.
 * \param this the graph coloring object
 * \return Always \e true !?
 */
static bool
Coalesce(GraphColoring this);

#line 531 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * When a move is coalesced, the related node may no longer be
 * move-related and can be added to the simplify work-list by the
 * procedure AddWorkList.
 * \param this the graph coloring object
 * \param interfNode The interference node.
 * \return Always \e true !?
 */
static bool
AddWorkList(GraphColoring this, InterfNode interfNode);

#line 542 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Combine two nodes for coalescing.
 * \param this the graph coloring object
 * \param interfNode1 The first interference node.
 * \param interfNode2 The second interference node.
 * \return Always \e true !?
 */
static bool
Combine(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2);

#line 552 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Implements the Briggs coalescing heuristic (conservative).
 * \param this the graph coloring object
 * \param interfNode1 The first interference node.
 * \param interfNode2 The second interference node.
 * \param interfMove The move to be coalesced.
 * \return \e true if the nodes can be conservative-coalesced, else \e false.
 */
static bool
Briggs(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2, InterfMove interfMove);

#line 563 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Implements the George coalescing heuristic (conservative -- note: only if no spill is necessary).
 * \param this the graph coloring object
 * \param interfNode1 The first interference node.
 * \param interfNode2 The second interference node, cannot be precolored.
 * \param interfMove The move to be coalesced.
 * \return \e true if the nodes can be conservative-coalesced, else \e false.
 */
static bool
George(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2, InterfMove interfMove);

#line 574 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * If neither Simplify() nor Coalesce() applies, we look for a move-related
 * node of low-degree. We Freeze() the moves in which this node is involved: that
 * is, we give up hope coalescing those moves. This causes the node (and perhaps
 * other nodes related to the frozen moves) to be considered non-move-related,
 * which should enable more simplification. Now Simplify() and Coalesce() are
 * resumed.
 * \param this the graph coloring object
 * \return Always true !?
 */
static bool
Freeze(GraphColoring this);

#line 587 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Freeze a move and update all related structures.
 * \param this the graph coloring object
 * \param interfNode_u The node in which the move is selected to be frozen.
 * \return Always true !?
 */
static bool
FreezeMoves(GraphColoring this, InterfNode interfNode_u);

#line 596 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * If there are no low-degree nodes, we select a significant-degree node
 * for potential spilling and push it on the stack.
 * \param this the graph coloring object
 */
static bool
SelectSpill(GraphColoring this);

#line 604 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Do the effective work of spilling a node,
 * push it into the stack
 * \param this the graph coloring object
 * \param interfNode the node selected for spill
 */
static void
SpillNode(GraphColoring this, InterfNode interfNode);

#line 613 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Push nodes forced to be spilled by ForceSpillNode in the spilled list.
 * Returns true is at least one node is forced to be spilled (for debug).
 */
static bool
ForceSpill(GraphColoring this);

#line 620 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns true is a node is forced to be spilled (for debug).
 */
static bool
ForceSpillNode(GraphColoring this, InterfNode interfNode);

#line 626 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Pop the entire stack, assign colors, update all related structures.
 * \param this the graph coloring object
 * \return Always true !?
 */
static bool
AssignColors(GraphColoring this);

#line 634 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
AssignRegisters(GraphColoring this);

#line 637 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Prune useless spills before the rewriting phases.
 */
static void
PruneUselessSpill(GraphColoring this, bool cisc_mode);

#line 643 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Allocates memory location for spilled (spill-everywhere strategy)
 * temporaries and inserts store and fetch instructions to access
 * them. These stores and fetches are newly created temporaries (with
 * tiny live-range), so the main loop must be performed on the altered
 * graph.
 * \param this the graph coloring object
 */
static bool
RewriteProgram(GraphColoring this);

#line 654 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Fonction pour appliquer l'heuristique de spill de Chaitin.
 * Si une variable est vivante en sortie du bloc de base, alors on lui cree
 * struct spill avec "spill.reloaded = true" de telle maniere que si l'on
 * rencontre une definition alors on la stocke en memoire (cf handleDefs).
 *
 * La boucle principale ne fait qu'appliquer les heuristiques de Chaitin.
 *
 * A la fin de l'iteration s'il reste des variables qui ont besoin d'etre
 * rechargees alors on les recharge (en prenant soin de ne pas charger des
 * variables qui ne font que traverser le bloc de base).
 *
 * \param this Graphe de coloration.
 */
static void
ChaitinSpillCode(GraphColoring this);

#line 671 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Discard all coalesced moves in the program.
 */
static bool
DiscardMoves(GraphColoring this);

#line 677 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns the spill representative of the temporary.
 * the return temporary maybe the temporary itself.
 * This function always return a valid temporary.
 * This is used to perform spill coalescing.
 */
static Temporary
GraphColoring_spillRepresentative(GraphColoring this, Temporary temporary)
{
  if (GraphColoring_SPILLS(this) == NULL) return temporary;
  if (Temporary_index(temporary) >= GraphColoring_INITIALTMPCOUNT(this))
    return temporary;
  if (GraphColoring_SPILLS(this)[Temporary_index(temporary)] != NULL) 
    return GraphColoring_SPILLS(this)[Temporary_index(temporary)];
  return temporary;
}

#line 694 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_setSpillRepresentative(GraphColoring this, Temporary temporary, Temporary representative)
{
  Except_CHECK (Temporary_INDEX(temporary) >= 0 &&
                Temporary_INDEX(temporary) < GraphColoring_INITIALTMPCOUNT(this));
  if (GraphColoring_SPILLS(this) == NULL) {
    *GraphColoring__SPILLS(this) = (Temporary *)malloc(GraphColoring_INITIALTMPCOUNT(this)*sizeof(Temporary));
    memset(GraphColoring_SPILLS(this), 0, GraphColoring_INITIALTMPCOUNT(this)*sizeof(Temporary));
  }
  GraphColoring_SPILLS(this)[Temporary_INDEX(temporary)] = representative;
}

#line 706 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_resetSpillRepresentatives(GraphColoring this)
{
  if (GraphColoring_SPILLS(this) != NULL) {
    free(GraphColoring_SPILLS(this));
  }
  *GraphColoring__SPILLS(this) = NULL;
}

#line 715 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Check the integrity of all structures.
 * \param file the output file for diagnostics.
 * \param this the graph coloring object
 * \return \e true on success, else \ e false.
 */
static bool
GraphColoring_check(GraphColoring this, FILE *file);

#line 724 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
GraphColoring_verification(GraphColoring this);

#line 727 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Get the queue of allocatable registers for a class.
 * \param regFile The class of registers.
 * \return The queue of Register allocatable for this class.
 */
static inline IQueue
AllocatableRegisters(RegFile regFile);

#line 735 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Get the number of registers allocatable for a class.
 * \param regFile The class of registers.
 * \return The number of register allocatable for this class.
 */
static inline uint32_t
NbRegisters(RegFile regFile);

#line 750 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static RegisterSet_
GRA_allocatable[RegFile__FIXME];
bool
GRA_isRegisterAllocatable(Register reg, RegFile regFile)
{
  return RegisterSet_contains(&GRA_allocatable[regFile], reg);
}

#line 783 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Temporary
GraphColoring_getDedicated(GraphColoring this, Register registre)
{
  return Procedure_makeDedicatedTemporary(GraphColoring_procedure(this), registre);
}

#line 796 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_fillTemporaryTable(GraphColoring this, RegFile regFile)
{
  CodeRegion codeRegion = GraphColoring_codeRegion(this);
  TemporaryTable table = GraphColoring_tmpTable(this);
  Register i;
  TemporaryTable_empty(table);
  for(i = RegFile_lowRegister(regFile); i <= RegFile_highRegister(regFile);
      ++i) {
    Temporary dedicatedTempo1 = GraphColoring_getDedicated(this, i);
    TemporaryTable_insert(table, dedicatedTempo1);
  }
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporary) {
      if(Temporary_regFile(temporary) == regFile) {
        TemporaryTable_insert(table, temporary);
      }
    } BasicBlock_liveOutSet_ENDSORT_Temporary;
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operation_ARGUMENTS_FOREACH_Temporary(operation, argument) {
	if (GRA_isAllocatable(argument, regFile))
	  TemporaryTable_insert(table, argument);
      } Operation_ARGUMENTS_ENDEACH_Temporary;
      Operation_RESULTS_FOREACH_Temporary(operation, result) {
	if (GRA_isAllocatable(result, regFile))
	  TemporaryTable_insert(table, result);
      } Operation_RESULTS_ENDEACH_Temporary;
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
}

#line 828 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define CEIL(x)		((float)((int)(x) < (x)? (int)(x) + 1: (int)(x)))

#line 830 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Private initialization for a code region.
 * Called from GraphColoring_allocateCodeRegion().
 * \param graph the graph coloring object
 * \param codeRegion the code region to allocate.
 */
static void
GraphColoring_initCodeRegion(GraphColoring this, CodeRegion codeRegion);

#line 839 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Private finalization for a code region.
 * Called from GraphColoring_allocateCodeRegion().
 * \param graph the graph coloring object
 */
static void
GraphColoring_finiCodeRegion(GraphColoring this);

#line 847 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Private initialization for a code region and a reg class.
 * Called from GraphColoring_allocateCodeRegion().
 * Must be called after GraphColoring_initCodeRegion().
 * \param graph the graph coloring object
 * \param regFile the register class to allocate.
 */
static void
GraphColoring_initRegFile(GraphColoring this, RegFile regFile);

#line 857 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Private finalization for a code region and a reg class.
 * Called from GraphColoring_allocateCodeRegion().
 * \param graph the graph coloring object
 */
static void
GraphColoring_finiRegFile(GraphColoring this);

#line 865 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Initialize a new pass for the coderegion/regclass.
 * Called from GraphColoring_allocateCodeRegion().
 * Must be called after GraphColoring_initPass().
 * \param graph the graph coloring object
 */
static void
GraphColoring_initPass(GraphColoring this);

#line 874 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Finalize a new pass for the coderegion/regclass.
 * Called from GraphColoring_allocateCodeRegion().
 * \param graph the graph coloring object
 */
static void
GraphColoring_finiPass(GraphColoring this);

#line 888 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Finalize for a program.
 */
static void
GraphColoring_finalize(void);

#line 894 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Statistiques_pretty(int32_t *statistiques, FILE *file);

#line 897 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
   * Export the control flow graph in \e .dot language
   *
   * \param file Where to write the graph
   * \param this GraphColoring
   * \param liveranges When set to true, will display the live-in/out of BB
   */
static void
exportGraphviz(FILE *file, GraphColoring this, bool liveranges);

#line 907 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
ComputeAllocationCost(GraphColoring this, FILE *file);

#line 910 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define MAX_USE 16
#define MAX_DEF 16

#line 913 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
// Some method shortcuts
#define WL_worklistMoves(this) WorkLists_worklistMoves(this)
#define WL_spillWorklist(this) WorkLists_spillWorklist(this)
#define WL_freezeWorklist(this) WorkLists_freezeWorklist(this)
#define WL_simplifyWorklist(this) WorkLists_simplifyWorklist(this)
#define WL_activeMoves(this) WorkLists_activeMoves(this)
#define WL_selectStack(this) WorkLists_selectStack(this)
#define WL_coalescedMoves(this) WorkLists_coalescedMoves(this)
#define WL_constrainedMoves(this) WorkLists_constrainedMoves(this)
#define WL_frozenMoves(this) WorkLists_frozenMoves(this)
#define WL_precolored(this) WorkLists_precolored(this)
#define WL_initial(this) WorkLists_initial(this)
#define WL_spilledNodes(this) WorkLists_spilledNodes(this)
#define WL_coalescedNodes(this) WorkLists_coalescedNodes(this)
#define WL_coloredNodes(this) WorkLists_coloredNodes(this)
#define WL_discardedNodes(this) WorkLists_discardedNodes(this)

#line 930 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define GC_regFile GraphColoring_regFile
#define GC_GetNode GraphColoring_GetNode
#define GC_IsNode GraphColoring_IsNode

#line 934 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define GC_workLists(this) GraphColoring_workLists(this)
#define GC_interfGraph(this) GraphColoring_interfGraph(this)

#line 937 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define GC_worklistMoves(this) WL_worklistMoves(GC_workLists(this))
#define GC_spillWorklist(this) WL_spillWorklist(GC_workLists(this))
#define GC_freezeWorklist(this) WL_freezeWorklist(GC_workLists(this))
#define GC_simplifyWorklist(this) WL_simplifyWorklist(GC_workLists(this))
#define GC_activeMoves(this) WL_activeMoves(GC_workLists(this))
#define GC_selectStack(this) WL_selectStack(GC_workLists(this))
#define GC_coalescedMoves(this) WL_coalescedMoves(GC_workLists(this))
#define GC_constrainedMoves(this) WL_constrainedMoves(GC_workLists(this))
#define GC_frozenMoves(this) WL_frozenMoves(GC_workLists(this))
#define GC_precolored(this) WL_precolored(GC_workLists(this))
#define GC_initial(this) WL_initial(GC_workLists(this))
#define GC_spilledNodes(this) WL_spilledNodes(GC_workLists(this))
#define GC_coalescedNodes(this) WL_coalescedNodes(GC_workLists(this))
#define GC_coloredNodes(this) WL_coloredNodes(GC_workLists(this))
#define GC_discardedNodes(this) WL_discardedNodes(GC_workLists(this))

#line 953 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Temporary
GraphColoring_newVirtual(GraphColoring this)
{
  return Procedure_makeVirtualTemporary(GraphColoring_procedure(this), GraphColoring_regFile(this));
}

#line 959 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Temporary
renameVariableToVirtual(CodeRegion this, Temporary temporary, bool isDef)
{
  Temporary ssareg;
  Variable variable = Temporary_VARIABLE(temporary);
  Except_CHECK(Temporary_isVariable(temporary));

#line 966 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (isDef) {
    RegFile regFile = Temporary_regFile(Variable_ORIGIN(variable));
    ssareg = Procedure_makeVirtualTemporary(CodeRegion_procedure(this), regFile);          
    *Variable__ORIGIN(variable) = ssareg;
  } else {
    // We already have renamed defs, thus the renaming is for the use.
    ssareg = Variable_ORIGIN(variable);
    Except_CHECK(Temporary_isVirtual(ssareg));
  }
  return ssareg;
}

#line 978 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
renameDefVariablesToVirtualDFS(CodeRegion this, Dominance dominance, BasicBlock basicBlock)
{
  if (basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operation_RESULTS_FOREACH_Temporary(operation, temporary) {
	if (!Temporary_isVariable(temporary))
	  continue;
	temporary = renameVariableToVirtual(this, temporary, true);
	Operation_setResult(operation, Operation_RESULTS_INDEX, temporary);
      } Operation_RESULTS_ENDEACH_Temporary;
    } BasicBlock_ENDEACH_Operation;
  }

#line 992 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Then process the children nodes in the dominator tree. */
  Dominance_BasicBlock_FORCHILDREN_BasicBlock(dominance, basicBlock, child) {
    renameDefVariablesToVirtualDFS(this, dominance, child);
  } Dominance_BasicBlock_ENDCHILDREN_BasicBlock;

#line 997 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
}

#line 1003 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
void
codeRegion_renameVariablesToVirtual(CodeRegion codeRegion)
{
  SSAForm ssaForm = CodeRegion_ssaForm(codeRegion);
  Dominance dominance = SSAForm_makeDominance(ssaForm, 0);

#line 1009 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // First rename defs in DFS order (not necessary but it gives a DFS numbering
  // of defs which is more readable).
  renameDefVariablesToVirtualDFS(codeRegion, dominance, CodeRegion_rootBlock(codeRegion));

#line 1013 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Then rename uses.
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operation_ARGUMENTS_FOREACH_Temporary(operation, temporary) {
	if (!Temporary_isVariable(temporary))
	  continue;
	temporary = renameVariableToVirtual(codeRegion, temporary, false);
	Operation_setArgument(operation, Operation_ARGUMENTS_INDEX, temporary);
      } Operation_ARGUMENTS_ENDEACH_Temporary;
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
}

#line 1026 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
codeRegion_makeSSAWebs(CodeRegion codeRegion, RegFile regFile, Equiv classes)
{
  Temporary web;

#line 1031 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_FOREACH_Operation(basicBlock, operation) {

#line 1034 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (Operation_operator(operation) != Operator_PHI)
	break;

#line 1037 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (!GRA_isAllocatable(Operation_getResult(operation, 0), regFile))
	continue;
      web = Equiv_find(classes, Operation_getResult(operation, 0));

#line 1041 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      Operation_ARGUMENTS_FOREACH_Temporary(operation, temporary) {
	if (web != Equiv_find(classes, temporary))
	  web = Equiv_union(classes, web, temporary);
      } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 1046 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (GRA_TRACE) {
	fprintf(GRA_TRACE, "Web for PHI:");
	Operation_pretty(operation, GRA_TRACE);
	fprintf(GRA_TRACE, "\tRepresentative: ");
	Temporary_pretty(web, GRA_TRACE);
	fprintf(GRA_TRACE, "\n");
      }
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;

#line 1056 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
}

#line 1058 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
GraphColoring_allocSSACodeRegion(GraphColoring this, CodeRegion codeRegion, SSAForm ssaForm)
{
  bool result;
  TemporaryTable globalTable;
  Liveness liveness;
  
  if (GRA_TRACE) {
    fprintf(GRA_TRACE, "*** GRA SSA Enter SSA\n");
    fprintf(GRA_TRACE, "<Enter> Construct_CodeRegionSSA\n");
    CodeRegion_pretty(codeRegion, GRA_TRACE);
  }
  
  /* Construct SSA form for the code region. */
  if(ssaForm == NULL) {
    SSAForm_construct(GraphColoring_memory(this), codeRegion, true);
  }
  if (GRA_TRACE) {
    fprintf(GRA_TRACE, "<Exit> Construct_CodeRegionSSA\n");
    CodeRegion_pretty(codeRegion, GRA_TRACE);
  }

#line 1080 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Rename variables to allocatable virtuals. */
  codeRegion_renameVariablesToVirtual(codeRegion);

#line 1083 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Do liveness over the SSA form. */
  globalTable = CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
  *GraphColoring__LIVENESS(this) = Liveness_make(GraphColoring_memory(this),
                                                 codeRegion, globalTable, true);
  Liveness_multiplex(GraphColoring_LIVENESS(this));

#line 1089 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GRA_TRACE) {
    fprintf(GRA_TRACE, "*** GRA SSA Before Coloring\n");
    CodeRegion_pretty(codeRegion, GRA_TRACE);
  }

#line 1094 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Do the graphcoloring. */
  result = GraphColoring_allocCodeRegion(this, codeRegion);

#line 1097 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Postcondition: if result is true, the code region is
     allocated and executable (out of colored SSA completed). */

#line 1100 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  Liveness_kill(GraphColoring_LIVENESS(this));
  *GraphColoring__LIVENESS(this) = NULL;

#line 1103 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return result;
}

#line 1106 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
bool
GraphColoring_allocateCodeRegion(CodeRegion codeRegion, SSAForm ssaForm)
{
  bool result;
  int gra_split = 0;
  GraphColoring_ allocator_;
  Memory memory;

#line 1114 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GraphColoring_initialize();

#line 1116 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  memory = Memory_new(Procedure_memory(CodeRegion_procedure(codeRegion)), 0);

#line 1118 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GraphColoring_Ctor(&allocator_, memory);

#line 1120 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  {
    Register reg;
    RegFile class;
    Convention convention = CodeRegion_convention(codeRegion);
    for (class = 0; class < RegFile__COUNT; class++) {
      GRA_allocatable[class] = Convention_allocatableSet(convention, class);
    }
    // Initialize arrays of allocatable registers.
    // Note that the order of the allocatable registers
    // in the array is the color assignation order.
    for (class = 0; class < RegFile__COUNT; class++) {
      IQueue_Ctor(AllocatableRegisters(class), Memory_Root, sizeof(Register), Register__COUNT);
      for (reg = RegFile_lowRegister(class); reg <= RegFile_highRegister(class); reg++) {
        if(RegisterSet_contains(&GRA_allocatable[class],reg))
          *((Register *)IQueue_push(AllocatableRegisters(class))) = reg;
      }
    }
  }

#line 1139 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Prespit edges
  // 1: false, 2:weak, 4:strong
  if (GETENV("GRA_SPLIT")) {
    sscanf(GETENV("GRA_SPLIT"), "%d", &gra_split);
    fprintf(stderr, "*** GRA_SPLIT=%08x\n", gra_split);
  }

#line 1146 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GRA_STAT) CodeRegion_statCodeRegion(codeRegion);
  if (GRA_STAT) {
    GRA_ssa_cost_nsplit = GRA_ssa_cost_split = GRA_ssa_cost_best_nsplit = GRA_ssa_cost_nback = GRA_ssa_cost_cmove = GRA_ssa_cost_best_cmove = 0.0;
    GRA_ssa_move_count = 0;
  }
  if (gra_split) {
    TemporaryTable globalTable;
    CodeRegion_splitCriticalEdges(&allocator_, codeRegion, gra_split&1, gra_split&2, gra_split&4);
    // We rebuild loopforest and liveness.
    LoopForest_delete(LoopForest_new(memory, codeRegion));
    Liveness_kill(GraphColoring_LIVENESS(&allocator_));
    globalTable = CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
    *GraphColoring__LIVENESS(&allocator_) =
        Liveness_make(GraphColoring_memory(&allocator_), codeRegion, globalTable, true);
    Liveness_multiplex(GraphColoring_LIVENESS(&allocator_));
  }

#line 1163 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GraphColoring_flags(&allocator_) & GRA_GCF_SSA) 
    result = GraphColoring_allocSSACodeRegion(&allocator_, codeRegion, ssaForm);
  else 
    result = GraphColoring_allocCodeRegion(&allocator_, codeRegion);

#line 1168 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GraphColoring_Dtor(&allocator_);
  Memory_delete(memory);

#line 1171 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GRA_TRACE) {
    fprintf(GRA_TRACE, "*** GRA After Coloring\n");
    CodeRegion_pretty(codeRegion, GRA_TRACE);
  }
  if (GRA_STAT) {
    fprintf(GRA_STAT, "<stat>\t\tN_move\tC_split\tC_nsplt\tC_best\tC_nback\tC_cmov\tC_best3\tP_name\n");
    fprintf(GRA_STAT, "<stat_move>\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	    GRA_ssa_move_count, GRA_ssa_cost_split, GRA_ssa_cost_nsplit, GRA_ssa_cost_best_nsplit, GRA_ssa_cost_nback, GRA_ssa_cost_cmove, GRA_ssa_cost_best_cmove, Procedure_name(CodeRegion_procedure(codeRegion)));
  }

#line 1181 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GraphColoring_finalize();

#line 1183 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return result;
}

#line 1186 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
GraphColoring_allocCodeRegion(GraphColoring this, CodeRegion codeRegion)
{
  bool change, forced_spill;
  int pass;
  RegFile regFile;
  int i = 0;

#line 1194 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GraphColoring_initCodeRegion(this, codeRegion);

#line 1196 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  for (pass = 0, regFile = 0; regFile < RegFile__COUNT; regFile++, pass++) {

#line 1198 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if(regFile == RegFile__UNDEF) {
      continue;
    }

#line 1202 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    totalspillcost = 0;

#line 1204 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    GraphColoring_initRegFile(this, regFile);

#line 1206 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (GRA_STAT) CodeRegion_statSpill(this,"before");

#line 1208 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    do {
      for(i=0; i<Stat__; i++) {
	if (i != build && i != inSpilledNodes && i != inCoalescedMoves && i != inDiscardedNodes)
	  statistiques[i] = 0;
      }
      change = false;

#line 1215 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      GraphColoring_initPass(this);

#line 1217 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (GRA_STAT) 
	GC_statMaxLive(this, false/*after*/);

#line 1220 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //fprintf( stderr, "Build...\n");
      Build(this);
      //fprintf( stderr, "done.\n");

#line 1224 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (GRA_MULTREGION)
        GC_multiplexRegion_pretty(this, GRA_MULTREGION);
      
      if (GRA_TRACE) InterfGraph_pretty(GC_interfGraph(this), GRA_TRACE);

#line 1229 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
      if (GETENV("GRA_INTERF")) 
      {
        char file[255];
        FILE *f;

#line 1235 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
        sprintf(file, "GRA_interf-%d.dot", pass);
        f = fopen(file, "w");
        //InterfGraph_dotty(GC_interfGraph(this), f);
        exportGraphviz(f, this, true);
        fclose(f);
      }
#endif//ECL_NTRACE

#line 1243 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // Check if some nodes are forced to be spilled (for debug).
      forced_spill = ForceSpill(this);

#line 1246 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (!forced_spill) {
	//fprintf( stderr, "MakeWorklist...\n");
	MakeWorklist(this);
      }

#line 1251 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //fprintf( stderr, "done.\n");
      if (GRA_TRACE) WorkLists_pretty(GC_workLists(this), GRA_TRACE); fflush(GRA_TRACE);
      if (GRA_TRACE) GraphColoring_check(this, GRA_TRACE);

#line 1255 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (!forced_spill) {
	do{
	  if( IDList_isEmpty(GC_simplifyWorklist(this)) == false ) {
	    //fprintf( stderr, "Simplify...\n");
	    Simplify(this);
	    //fprintf( stderr, "done.\n");
	  }
	  else if( IDList_isEmpty(GC_worklistMoves(this)) == false ) {
	    //fprintf( stderr, "Coalesce...\n");
	    Coalesce(this);
	    //fprintf( stderr, "done.\n");
	  }
	  else if( IDList_isEmpty(GC_freezeWorklist(this)) == false ) {
	    //fprintf( stderr, "Freeze...\n");
	    Freeze(this);
	    //fprintf( stderr, "done.\n");

#line 1272 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  } else if( IDList_isEmpty(GC_spillWorklist(this)) == false ) {
	    //fprintf( stderr, "Select...\n");
	    SelectSpill(this);
	    //fprintf( stderr, "done.\n");
	  }
	} while ( IDList_isEmpty(GC_simplifyWorklist(this)) != true ||
		  IDList_isEmpty(GC_worklistMoves(this)) != true ||
		  IDList_isEmpty(GC_freezeWorklist(this)) != true ||
		  IDList_isEmpty(GC_spillWorklist(this)) != true );

#line 1282 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	//fprintf( stderr, "Assign colors...\n");
	AssignColors(this);
	//fprintf( stderr, "done.\n");
      }

#line 1287 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if( IDList_isEmpty(GC_spilledNodes(this)) == false ) {
	//fprintf( stderr, "Rewrite program...\n");

#line 1290 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	PruneUselessSpill(this, true); // Stat Useless spill for cisc
	PruneUselessSpill(this, false); // Stat Useless spill for risc

#line 1294 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	GRA_redo_liveness = false;
	if (GraphColoring_flags(this) & GRA_GCF_CHAINTIN_SPILL_HEURISTIC)
	  ChaitinSpillCode(this);
	else
	  RewriteProgram(this);
	//fprintf( stderr, "done.\n");

#line 1301 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	// Currently we do not update correctly
	// liveness under SSA due to spill of PHI
	// operation, thus we redo the liveness
	// analysis if requested.
	if (GRA_redo_liveness) {
      // If we compute global table here, we will break the temporaries' index used in the
      // interference graph
          Liveness_kill(GraphColoring_LIVENESS(this));
          *GraphColoring__LIVENESS(this) =
            Liveness_make(GraphColoring_memory(this), codeRegion, GraphColoring_tmpTable(this),
                            true);
          Liveness_multiplex(GraphColoring_LIVENESS(this));
        }

#line 1315 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	change = true;
	GRA_TRACE &&  fprintf(GRA_TRACE, "*** GRA PASS SPILL\n");
	GRA_TRACE &&  CodeRegion_pretty(codeRegion, GRA_TRACE);
      }

#line 1320 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      GraphColoring_finiPass(this);

#line 1322 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    } while(change == true);

#line 1324 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (GRA_STAT) 
      GC_statMaxLive(this, true/*after*/);

#line 1327 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //fprintf( stderr, "Assign registers...\n");
    AssignRegisters(this);
    //fprintf( stderr, "done.\n");

#line 1331 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //fprintf( stderr, "Discard Moves...\n");
    DiscardMoves(this);
    //fprintf( stderr, "done.\n");

#line 1335 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //debug function
    /*                FILE *f = fopen("graphviz.dot","w");*/
    /*                fprintf(stderr, "Graphviz export\n","w");*/
    /*                fflush(stderr);*/
    /*                exportGraphviz(f, this, false);*/
    /*                fclose(f);*/

#line 1342 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // Discard phi kills indexes in tmpTable.
    // export solution before that
    GRA_exportSolution(GraphColoring_tmpTable(this), GRA_DISP_COLORING,
                       regFile);

#line 1347 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (GraphColoring_flags(this) & GRA_GCF_SSA) {
      if (GRA_STAT) CodeRegion_statSpill(this,"d_phi");
      GRA_TRACE &&  fprintf(GRA_TRACE, "*** BEFORE DISCARD PHI\n");
      if (GRA_TRACE) CodeRegion_pretty(codeRegion, GRA_TRACE);
      GRA_statisticOnCodeRegion(codeRegion, "Before discard phi", GRA_DISP_COST);
      OutOfColoredSSA_discardPHIs(codeRegion, GraphColoring_regFile(this),
                                  GraphColoring__LIVENESS(this),
                                  GraphColoring_memory(this));
    }

#line 1357 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (GRA_STAT) CodeRegion_statSpill(this,"after");

#line 1360 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //fprintf( stderr, "Verification...\n");
    if (GRA_TRACE) GraphColoring_verification(this);
    //fprintf( stderr, "done.\n");

#line 1364 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    GRA_TRACE &&  fprintf(GRA_TRACE, "*** GRA PASS\n");
    if (GRA_TRACE) CodeRegion_pretty(codeRegion, GRA_TRACE);
    if (GRA_TRACE) ComputeAllocationCost(this, GRA_TRACE);
    if (GRA_TRACE) Statistiques_pretty(statistiques, GRA_TRACE);

#line 1369 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    GraphColoring_finiRegFile(this);

#line 1371 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  }

#line 1373 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Remove empty blocks that may have been inserted by edge spliting
  if (GraphColoring_flags(this) & GRA_GCF_SSA) {
    CodeRegion_removeEmptyBlocks(codeRegion);
  }

#line 1378 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GraphColoring_finiCodeRegion(this);

#line 1380 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return true;
}

#line 1383 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_initCodeRegion(GraphColoring this, CodeRegion codeRegion)
{
  *GraphColoring__CODEREGION(this) = codeRegion;
  // Initialize spill module
  Spill_initCodeRegion(codeRegion);
}

#line 1391 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_finiCodeRegion(GraphColoring this)
{
  // Mark all bb as allocated.
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    BasicBlock_setFlags(basicBlock, BasicBlockFlag_Allocated);
  } CodeRegion_ENDEACH_BasicBlock;

#line 1399 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Finalize spill module.
  Spill_finiCodeRegion(GraphColoring_codeRegion(this));

#line 1402 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  *GraphColoring__CODEREGION(this) = NULL;
}

#line 1405 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_initRegFile(GraphColoring this, RegFile regFile)
{
  CodeRegion codeRegion = GraphColoring_codeRegion(this);
  *GraphColoring__regFile(this) = regFile;
  *GraphColoring__PASSNUM(this) = 0;

#line 1412 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  TemporaryTable_Ctor(GraphColoring_tmpTable(this), GraphColoring_memory(this), 1024);
  GraphColoring_fillTemporaryTable(this, regFile);
  *GraphColoring__INITIALTMPCOUNT(this) = TemporaryTable_count(GraphColoring_tmpTable(this));
  InterfGraph_Ctor(GraphColoring_interfGraph(this), GraphColoring_memory(this));
  WorkLists_Ctor(GraphColoring_workLists(this), GraphColoring_memory(this));

#line 1418 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // If allocating under CSSA, create ssa webs and
  // map temporary to spill representative.
  if (GraphColoring_flags(this) & GRA_GCF_SSA) {
    // Build SSA webs
    Equiv classes = Equiv_new(TemporaryTable_count(GraphColoring_tmpTable(this)), GraphColoring_memory(this));
    codeRegion_makeSSAWebs(codeRegion, regFile, classes);

#line 1425 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // Maps temporaries to spill representative with ssa webs
    GraphColoring_resetSpillRepresentatives(this);
    TemporaryTable_FOREACH_Temporary(GraphColoring_tmpTable(this), temporary) {
      Temporary representative;
      if (!GRA_isAllocatable(temporary, regFile))
	continue;

#line 1432 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if ((representative = Equiv_find(classes, temporary)) != NULL) {
	GraphColoring_setSpillRepresentative(this, temporary, representative);
      }
    } TemporaryTable_ENDEACH_Temporary;
    Equiv_del(classes);
  }
}

#line 1440 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_finiRegFile(GraphColoring this)
{
  WorkLists_Dtor(GraphColoring_workLists(this));
  InterfGraph_Dtor(GraphColoring_interfGraph(this));
  TemporaryTable_Dtor(GraphColoring_tmpTable(this));
  *GraphColoring__regFile(this) = 0;
  *GraphColoring__PASSNUM(this) = 0;
}

#line 1450 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_initPass(GraphColoring this)
{
  RegFile regFile = GraphColoring_regFile(this);
  (*GraphColoring__PASSNUM(this))++;

#line 1456 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // First pass.
  if (GraphColoring_passNum(this) == 1) {
    // Creates empty interference graph
    InterfGraph_createAdjSet(GraphColoring_interfGraph(this),
                             PtrSeq_maxCount(IndexedTable____(TemporaryTable____(GraphColoring_tmpTable(this)))));
  } else {
    // Update the interference graph
    InterfGraph_reallocAdjSet(GraphColoring_interfGraph(this), PtrSeq_maxCount(IndexedTable____(TemporaryTable____(GraphColoring_tmpTable(this)))));
  }

#line 1466 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Create and insert interfNodes into initial and precolored lists.
  TemporaryTable_FOREACH_Temporary(GraphColoring_tmpTable(this), temporary) {
    if(GRA_isAllocatable(temporary, regFile)) {
      if (!GC_IsNode(this, temporary))
	// New nodes are added only once for all passes.
	if (!GRA_isAssigned(temporary)) {
	  AddNodeIntoWorklist(this, GC_initial(this), temporary, _initial);
	  statistiques[inInitial]++;
	} else {
	  AddNodeIntoWorklist(this, GC_precolored(this), temporary, _precolored);
      InterfNode_setColor(GraphColoring_GetNode((this),temporary),
                          Temporary_register(temporary));
	}
      else
	// Old nodes are added to the adjset for checking purpose.
	InterfGraph_addToAdjSet(GraphColoring_interfGraph(this), GraphColoring_GetNode(this, temporary), GraphColoring_GetNode(this, temporary));

#line 1483 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    }
  } TemporaryTable_ENDEACH_Temporary;

#line 1486 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GraphColoring_passNum(this) == 1) return;

#line 1489 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Other pass (we have spilled). We reinsert nodes into initial lists.

#line 1491 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(GC_coloredNodes(this), InterfNode_, interfNode) {
    IDList_remove_(GC_coloredNodes(this), interfNode);
    PushNodeIntoWorklist(this, GC_initial(this), interfNode, _initial);
    *InterfNode__DEGREE(interfNode) = 0;
    *InterfNode__ALIAS(interfNode) = interfNode;
    IDList_empty(InterfNode_ADJLIST(interfNode), NULL);
    IDList_empty(InterfNode_MOVELIST(interfNode), NULL);
    IDList_empty(InterfNode_USES(interfNode), NULL);
    IDList_empty(InterfNode_DEFS(interfNode), NULL);
  } IDList_ENDEACH;

#line 1502 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(GC_coalescedNodes(this), InterfNode_, interfNode) {
    IDList_remove_(GC_coalescedNodes(this), interfNode);
    PushNodeIntoWorklist(this, GC_initial(this), interfNode, _initial);
    *InterfNode__DEGREE(interfNode) = 0;
    *InterfNode__ALIAS(interfNode) = interfNode;
    IDList_empty(InterfNode_ADJLIST(interfNode), NULL);
    IDList_empty(InterfNode_MOVELIST(interfNode), NULL);
    IDList_empty(InterfNode_USES(interfNode), NULL);
    IDList_empty(InterfNode_DEFS(interfNode), NULL);
  } IDList_ENDEACH;

#line 1513 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(GC_precolored(this), InterfNode_, interfNode) {
    *InterfNode__DEGREE(interfNode) = 0;
    *InterfNode__ALIAS(interfNode) = interfNode;
    IDList_empty(InterfNode_ADJLIST(interfNode), NULL);
    IDList_empty(InterfNode_MOVELIST(interfNode), NULL);
    IDList_empty(InterfNode_USES(interfNode), NULL);
    IDList_empty(InterfNode_DEFS(interfNode), NULL);
  } IDList_ENDEACH;

#line 1522 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Correct only for 'spill-everywhere' !
  // We move the spilled of the previous pass into the discarded nodes.
  IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {
    IDList_remove_(GC_spilledNodes(this), interfNode);
    PushNodeIntoWorklist(this, GC_discardedNodes(this), interfNode, _discardedNodes);
    statistiques[inDiscardedNodes]++;
    *InterfNode__DEGREE(interfNode) = 0;
    *InterfNode__ALIAS(interfNode) = interfNode;
    IDList_empty(InterfNode_ADJLIST(interfNode), NULL);
    IDList_empty(InterfNode_MOVELIST(interfNode), NULL);
    IDList_empty(InterfNode_USES(interfNode), NULL);
    IDList_empty(InterfNode_DEFS(interfNode), NULL);
  } IDList_ENDEACH;

#line 1536 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Reset all the graph structures.
/*   IDList_empty(GC_frozenMoves(this), (void(*)(void*))InterfMove_Dtor); */
/*   IDList_empty(GC_constrainedMoves(this), (void(*)(void*))InterfMove_Dtor); */
/*   IDList_empty(GC_coalescedMoves(this), (void(*)(void*))InterfMove_Dtor); */
/*   IDList_empty(GC_activeMoves(this), (void(*)(void*))InterfMove_Dtor); */
/*   IDList_empty(GC_worklistMoves(this), (void(*)(void*))InterfMove_Dtor); */
}

#line 1544 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_finiPass(GraphColoring this)
{
}

#line 1549 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
void
GraphColoring_initialize(void)
{
  int gra_ssa = 0;
  if (GETENV("GRA_SSA")) {
    sscanf(GETENV("GRA_SSA"), "%d", &gra_ssa);
    fprintf(stderr, "*** GRA_SSA=%d\n", gra_ssa);
  }

#line 1558 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (gra_ssa) GRA_gcf = GRA_SSA_GCF_DEFAULT;
  else GRA_gcf = GRA_GCF_DEFAULT;
  if (GETENV("LAO_GRA_GCF")) {
    sscanf(GETENV("LAO_GRA_GCF"), "%d", &GRA_gcf);
    fprintf(stderr, "*** LAO_GRA_GCF=%d\n", GRA_gcf);
  }

#line 1565 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  GRA_ssa_split = 0;
  GRA_ssa_nsplit = 0;
  GRA_ssa_cmove = 0;
  GRA_ssa_nback = 0;
  GRA_ssa_best_nsplit = 0;
  GRA_ssa_best_cmove = 0;
  GRA_ssa_use_new_rpc = 0;

#line 1573 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GETENV("GRA_SSA_SPLIT")) {
    sscanf(GETENV("GRA_SSA_SPLIT"), "%d", &GRA_ssa_split);
    fprintf(stderr, "*** GRA_SSA_SPLIT=%d\n", GRA_ssa_split);
  }
  if (GETENV("GRA_SSA_NSPLIT")) {
    sscanf(GETENV("GRA_SSA_NSPLIT"), "%d", &GRA_ssa_nsplit);
    fprintf(stderr, "*** GRA_SSA_NSPLIT=%d\n", GRA_ssa_nsplit);
  }
  if (GETENV("GRA_SSA_CMOVE")) {
    sscanf(GETENV("GRA_SSA_CMOVE"), "%d", &GRA_ssa_cmove);
    fprintf(stderr, "*** GRA_SSA_CMOVE=%d\n", GRA_ssa_cmove);
  }
  if (GETENV("GRA_SSA_NBACK")) {
    sscanf(GETENV("GRA_SSA_NBACK"), "%d", &GRA_ssa_nback);
    fprintf(stderr, "*** GRA_SSA_NBACK=%d\n", GRA_ssa_nback);
  }
  if (GETENV("GRA_SSA_BEST_NSPLIT")) {
    sscanf(GETENV("GRA_SSA_BEST_NSPLIT"), "%d", &GRA_ssa_best_nsplit);
    fprintf(stderr, "*** GRA_SSA_BEST_NSPLIT=%d\n", GRA_ssa_best_nsplit);
  }
  if (GETENV("GRA_SSA_BEST_CMOVE")) {
    sscanf(GETENV("GRA_SSA_BEST_CMOVE"), "%d", &GRA_ssa_best_cmove);
    fprintf(stderr, "*** GRA_SSA_BEST_CMOVE=%d\n", GRA_ssa_best_cmove);
  }
  if (GETENV("GRA_SSA_USE_NEW_RPC")) {
    sscanf(GETENV("GRA_SSA_USE_NEW_RPC"), "%d", &GRA_ssa_use_new_rpc);
    fprintf(stderr, "*** GRA_SSA_USE_NEW_RPC=%d\n", GRA_ssa_use_new_rpc);
  }
  Except_CHECK(GRA_ssa_split + GRA_ssa_nsplit + GRA_ssa_cmove + GRA_ssa_nback + GRA_ssa_best_nsplit + GRA_ssa_best_cmove <= 1);
  if (GRA_ssa_split + GRA_ssa_nsplit + GRA_ssa_cmove + GRA_ssa_nback + GRA_ssa_best_nsplit + GRA_ssa_best_cmove == 0) {
    /* The default. */
    GRA_ssa_best_nsplit = 1;
  }

#line 1607 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
}

#line 1610 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GraphColoring_finalize(void)
{
  RegFile class;
  for (class = 0; class < RegFile__COUNT; class++) {
    IQueue_Dtor(AllocatableRegisters(class));
  }
}

#line 1619 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static InterfNode
AddNodeIntoWorklist(GraphColoring this, IDList worklist, Temporary temporary, NodeBelongTo belongTo)
{
  IDListItem item;
  InterfGraph interfGraph = GC_interfGraph(this);

#line 1625 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  Except_CHECK(InterfGraph_temporaryIdentity(temporary) >= 0);
  Except_CHECK(InterfGraph_temporaryIdentity(temporary) < InterfGraph_SIZE(interfGraph));

#line 1628 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  item = IDList_push(worklist, sizeof(InterfNode_));
  InterfNode_Ctor((InterfNode)item, GraphColoring_memory(this), temporary);
  *InterfNode__BELONGTO((InterfNode)item) = belongTo;
  if (Temporary_index(temporary) >= GraphColoring_initialTmpCount(this))
    InterfNode_setFlag((InterfNode)item, InterfNode_ALREADYSPILLED);
  //fprintf(stderr, "index=%d\n", InterfGraph_temporaryIdentity(temporary));
  InterfGraph_addNode(interfGraph, (InterfNode)item);
  InterfGraph_addToAdjSet(interfGraph, (InterfNode)item, (InterfNode)item);
  return (InterfNode)item;
}

#line 1639 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
PushNodeIntoWorklist(GraphColoring this, IDList worklist, InterfNode interfNode, NodeBelongTo belongTo)
{
  IDList_push_(worklist, interfNode);
  *InterfNode__BELONGTO(interfNode) = belongTo;
}

#line 1646 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
PushMoveIntoWorklist(GraphColoring this, IDList worklist, InterfMove interfMove, MoveBelongTo belongTo)
{
  IDList_push_(worklist, interfMove);
  *InterfNode__BELONGTO(interfMove) = belongTo;
}

#line 1653 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
AddMoveAffinity(GraphColoring this, Operation operation)
{
  Temporary def = Operation_getResult(operation, 0);
  Temporary use = Operation_getArgument(operation, 0);
  float cost = BasicBlock_frequency(Operation_BASICBLOCK(operation));

#line 1660 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfMove affinity = (InterfMove)IDList_push(GC_worklistMoves(this), sizeof(InterfMove_));
  InterfMove_Ctor(affinity, operation, GC_GetNode(this, def), GC_GetNode(this, use), 0, 0, cost);
  *InterfMove__BELONGTO(affinity) = _worklistMoves;
  statistiques[inWorklistMoves]++;
  InterfNode_addToMoveList(GC_GetNode(this, Operation_getResult(operation, 0)), affinity);
  InterfNode_addToMoveList(GC_GetNode(this, Operation_getArgument(operation, 0)), affinity);
}

#line 1668 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
AddPhiAffinities(GraphColoring this, Operation operation)
{
  Temporary def = Operation_getResult(operation, 0);
  // For a PHI operation we add a move affinity between the def and each operand
  // the cost of the affinity is the frequency of the operand related edge.
  PHI_arguments_FOREACH_temporary(operation, edge, idx, use) {
    float cost = BasicBlock_frequency(BasicBlockEdge_origBlock(edge)) * BasicBlockEdge_probability(edge);
    InterfMove affinity = (InterfMove)IDList_push(GC_worklistMoves(this), sizeof(InterfMove_));
    InterfMove_Ctor(affinity, operation, GC_GetNode(this, def), GC_GetNode(this, use), 0, idx, cost);
    *InterfMove__BELONGTO(affinity) = _worklistMoves;
    statistiques[inWorklistMoves]++;
    InterfNode_addToMoveList(GC_GetNode(this, def), affinity);
    InterfNode_addToMoveList(GC_GetNode(this, use), affinity);
    //break;
  } PHI_arguments_ENDEACH_temporary;
}

#line 1686 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
// Warning: we store only temporaries and dedicated registers in which
// one can write.  Also, nodes in the graph that represent dedicated
// registers are slightly different, they will stay in the list of
// precolored nodes, they have an empty list of neighbors (their
// degree is not even defined), but they have a list of moves.
static int32_t
Build(GraphColoring this)
{
  /* procedure Build()
   *  forall b in 'blocks in program'
   *	let live = liveOut(b)
   *	forall I in instructions(b) in reverse order
   *		if isMoveInstruction(I) then
   *			live <- live/use(I)
   *			forall n in def(I) union use(I)
   *				moveList[n] <- moveList[n] union {I}
   *			worklistMoves <- worklistMoves union {I}
   *		live <- live union def(I)
   *		forall d in def(I)
   *			AddEdge(l, d)
   *		live <- Use(I) union (live\def(I))
   */

#line 1709 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  int i;

#line 1711 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  TemporarySet live, use, def;

#line 1713 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  RegFile regFile = GraphColoring_regFile(this);
  BasicBlock flowNode = NULL;
  InterfNode interfNode = NULL;
  InterfMove interfMove = NULL;
  InterfNode interfN2 = NULL;

#line 1719 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  uint32_t count = 0;
  uint32_t index = 0;

#line 1722 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  statistiques[build]++;

#line 1724 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  use = TemporarySet_make(GraphColoring_memory(this), 0);
  def = TemporarySet_make(GraphColoring_memory(this), 0);
  live = TemporarySet_make(GraphColoring_memory(this), 0);
//#if 0
  //ALGO: procedure Build()

#line 1730 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall b in 'blocks in program'
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    //ALGO: let live = liveOut(b)
    TemporarySet_empty(live);
    BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporary) {
      if(GRA_isAllocatable(temporary, regFile)) {
	TemporarySet_insert( live, temporary);
      }
    } BasicBlock_liveOutSet_ENDSORT_Temporary;

#line 1740 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: forall I in instructions(b) in reverse order
    BasicBlock_FORBACK_Operation( basicBlock, operation) {

#line 1743 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      TemporarySet_empty(use);
      TemporarySet_empty(def);

#line 1747 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // Construct def(I) and use(I)
      Operation_RESULTS_FOREACH_Temporary( operation, temporary) {
	if(GRA_isAllocatable(temporary, regFile)) {
	  if (TemporarySet_insert(def, temporary) == true) {
	    if (!GRA_isAssigned(temporary))
	      InterfNode_addDef(GC_GetNode(this, temporary), operation);
        // Add interference between definitions
        TemporarySet_FOREACH_Temporary(def, defTemp) {
          if(defTemp != temporary) {
            AddEdge(this, GC_GetNode(this, temporary), GC_GetNode(this, defTemp));
          }
        } TemporarySet_ENDEACH_Temporary;
      }
	}
      } Operation_RESULTS_ENDEACH_Temporary;			

#line 1763 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      Operation_ARGUMENTS_FOREACH_Temporary( operation, temporary) {
	if(GRA_isAllocatable(temporary, regFile)) {
	  if (TemporarySet_insert(use, temporary) == true)
	    if (!GRA_isAssigned(temporary))
	      InterfNode_addUse(GC_GetNode(this, temporary), operation);
	}
      } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 1771 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: if isMoveInstruction(I) then
      //Note: We assume in this implementation that a move has only one def and only one use.
      if( Operator_isSeqCopy( Operation_operator(operation)) == true &&
	  GRA_isAllocatable(Operation_getResult(operation, 0), regFile) &&
	  GRA_isAllocatable(Operation_getArgument(operation, 0), regFile) &&
	  Temporary_IDENTITY(Operation_getResult(operation, 0)) != Temporary_IDENTITY(Operation_getArgument(operation, 0)) &&
	  (!GRA_isAssigned(Operation_getResult(operation, 0)) ||
	   !GRA_isAssigned(Operation_getArgument(operation, 0)))) {
	//We assume here that a move has only one def and only one use.
	assert( Operation_argCount(operation) == 1 &&
		Operation_resCount(operation) == 1 );

#line 1783 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	//ALGO: live <- live \ use(I)
	if (!(GraphColoring_flags(this) & GRA_GCF_ADD_MOVE_INTERFERENCE))
	  TemporarySet_remove(live, Operation_getArgument(operation, 0));
	if (!(GraphColoring_flags(this) & GRA_GCF_NO_MOVE_COALESCING))
	  AddMoveAffinity(this, operation);
      }

#line 1790 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // ALGO. if is PHI(I) add move affinities
      if (Operation_operator(operation) == Operator_PHI &&
	  GRA_isAllocatable(Operation_getResult(operation, 0), regFile) &&
	  (!(GraphColoring_flags(this) & GRA_GCF_NO_PHI_COALESCING) ||
	   RegFile_width(regFile) == 1)) /* Force coalescing of BR (1 bit width), as BR copy is not correctly handled. */

#line 1796 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	AddPhiAffinities(this, operation);

#line 1798 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: live <- live union def(I)
// Do not unioned with def, since def are defined "after" operation execution,
// thus not affected by clobber
//      TemporarySet_union( live, def);
      TemporarySet_diff( live, def);

#line 1804 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: forall l in live
      TemporarySet_FORSORT_Temporary( live, Temporary_compare, temporary_live) {
	//ALGO: forall d in def(I)
	InterfNode interfNode_live = GC_GetNode(this, temporary_live);
	TemporarySet_FORSORT_Temporary( def, Temporary_compare, temporary_def) {
	  //ALGO: AddEdge(l, d)
	  AddEdge(this, interfNode_live, GC_GetNode(this, temporary_def));
	} TemporarySet_ENDSORT_Temporary;

#line 1813 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	if(Operation_hasClobber(operation)) { // If this operation corrupts some registers.
	  RegisterSet clobberSet = Operation_clobberSet(operation);
	  Temporary temporary_clobber = NULL;

#line 1817 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  RegisterSet_FOREACH_Register(clobberSet, clobber) {
	    temporary_clobber = GraphColoring_getDedicated(this, clobber);
	    if(GRA_isAllocatable(temporary_clobber, regFile)) {
	      AddEdge(this, interfNode_live, GC_GetNode(this, temporary_clobber));
	    }
	  } RegisterSet_ENDEACH_Register;
	}
      } TemporarySet_ENDSORT_Temporary;

#line 1826 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: live <- use(I) union (live \ def(I)) ignore uses on PHI
//      TemporarySet_diff( live, def);	
      if (Operation_operator(operation) != Operator_PHI)
	TemporarySet_union( live, use);
    } BasicBlock_ENDBACK_Operation;
    
    // for each PHI predecessor, add inteference between uses on PHI
    // this is necessary to keep C-SSA as we ignore interferences on moves
    // that otherwise do not interfere.
    if (GraphColoring_flags(this) & GRA_GCF_SSA) {
      int argIdx = 0;
      //ALGO: foreach argIdx in num_phi_args
      BasicBlock_ENTER_FOREACH_BasicBlockEdge(basicBlock, edge) {
	TemporarySet_empty(live);
	BasicBlock_FOREACH_Operation(basicBlock, operation) {
	  Temporary use;
	  InterfNode useNode;

#line 1844 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  if (Operation_operator(operation) != Operator_PHI) break;
	  use = Operation_getArgument(operation, argIdx);
	  if (!GRA_isAllocatable(use, regFile)) continue;
	  useNode = GC_GetNode(this, use);
	  
	  //ALGO: foreach i in live 
	  TemporarySet_FORSORT_Temporary(live, Temporary_compare, temporary_live) {
	    InterfNode interfNode_live = GC_GetNode(this, temporary_live);
	    //ALGO: AddEdge(use(argIdx), i)
	    AddEdge(this, interfNode_live, useNode);
	  } TemporarySet_ENDSORT_Temporary;

#line 1856 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  //ALGO: live <- live \ use(argIdx)
	  TemporarySet_insert(live, use);
	} BasicBlock_ENDEACH_Operation;
	argIdx++;
      } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
    }
  } CodeRegion_ENDEACH_BasicBlock;
//#endif
  TemporarySet_kill(use);
  TemporarySet_kill(def);
  TemporarySet_kill(live);

#line 1868 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return count;
  //ALGO:
}

#line 1873 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define GC_multiplexRegion_traverseUp(edge, e, s, d) \
  GC_multiplexRegion_traverse(edge, e, s, d, true)

#line 1876 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#define GC_multiplexRegion_traverseDown(edge, e, s, d) \
  GC_multiplexRegion_traverse(edge, e, s, d, false)

#line 1880 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
GC_multiplexRegion_traverse(BasicBlockEdge edge, int *num_edges, int *num_sources, int *num_dests, bool up )
{
  int tmp;
  *BasicBlockEdge__INDEX(edge) = 1;

#line 1886 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (NULL == (up? BasicBlockEdge_origBlock(edge): BasicBlockEdge_destBlock(edge))) {
    return;
  }

#line 1890 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if(!BasicBlockEdge_isCritical(edge)) return;

#line 1892 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf(GRA_MULTREGION, "<mult>\ttrav %u %s %u (freq %f) (back %d)\n", 
	  (unsigned int)BasicBlock_IDENTITY(BasicBlockEdge_origBlock(edge)),
	  up? "<-": "->",
	  (unsigned int)BasicBlock_IDENTITY(BasicBlockEdge_destBlock(edge)),
	  BasicBlock_frequency(BasicBlockEdge_origBlock(edge)) * BasicBlockEdge_probability(edge),
	  BasicBlockEdge_isBack(edge)
	  );

#line 1900 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  (*num_edges)++;

#line 1902 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (up) {
    BasicBlock basicBlock = BasicBlockEdge_origBlock(edge);
    if( ! (BasicBlock_NUMBERING(basicBlock) & 0x1) ) {
      (*BasicBlock__NUMBERING(basicBlock)) |= 0x1;
      (*num_sources)++;

#line 1908 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, a) {
        if (BasicBlockEdge_INDEX(a) == -1) {
          GC_multiplexRegion_traverseDown(a, num_edges, num_sources, num_dests);
        }
      } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
    }
  } else {
    BasicBlock basicBlock = BasicBlockEdge_destBlock(edge);
    if( ! (BasicBlock_NUMBERING(basicBlock) & 0x2)) {
      (*BasicBlock__NUMBERING(basicBlock)) |= 0x2;
      (*num_dests)++;

#line 1920 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      BasicBlock_ENTER_FOREACH_BasicBlockEdge(basicBlock, a) {
        if (BasicBlockEdge_INDEX(a) == -1) {
          GC_multiplexRegion_traverseUp(a, num_edges, num_sources, num_dests);
        }
      } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
    }
  }
}

#line 1930 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
int num_multiplex_pass = 0;

#line 1932 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
bool
GC_multiplexRegion_pretty(GraphColoring this, FILE *file)
{
  /*
   * The INDEX field of BasicBlockEdges will be used to mark edges
   */
  int max=0;
  int num_edges, num_sources, num_dests;

#line 1942 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (num_multiplex_pass > 0) return true;
  num_multiplex_pass++;

#line 1945 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if(file == NULL) file = stderr;

#line 1947 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf(file, "<mult> Running multiplexing region stats.\n");

#line 1949 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    *BasicBlock__NUMBERING(basicBlock) = 0;
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
      Except_CHECK(BasicBlockEdge_INDEX(edge) == -1);
      *BasicBlockEdge__INDEX(edge) = -1;
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  } CodeRegion_ENDEACH_BasicBlock;

#line 1957 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    if( ! (BasicBlock_NUMBERING(basicBlock) & 1)) {
      (*BasicBlock__NUMBERING(basicBlock)) |= 1;
      num_edges = num_dests = 0;
      num_sources = 1;

#line 1963 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
        if(BasicBlockEdge_INDEX(edge) == -1) {
          GC_multiplexRegion_traverseDown(edge, &num_edges, &num_sources, &num_dests);
        }
      } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
      if (num_edges > 0) {
        fprintf(file,"<mult> Leave Multiplexing region of %d: %d E, %d S, %d D\n", (unsigned int)BasicBlock_IDENTITY(basicBlock),num_edges, num_sources, num_dests);
      }
    }

#line 1973 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    /*BasicBlock_ENTER_FOREACH_BasicBlockEdge(basicBlock, edge) {
      if(BasicBlockEdge_INDEX(edge) == -1) {
        num_edges = num_sources = 0;
        num_dests = 1;
        GC_multiplexRegion_traverseUp(edge, &num_edges, &num_sources, &num_dests);
        fprintf(stderr,"Enter Multiplexing region: %d E, %d S, %d D\n",num_edges, num_sources, num_dests);
      }
    } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;*/
  } CodeRegion_ENDEACH_BasicBlock;

#line 1983 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    *BasicBlock__NUMBERING(basicBlock) = 0;
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
      *BasicBlockEdge__INDEX(edge) = -1;
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
    BasicBlock_ENTER_FOREACH_BasicBlockEdge(basicBlock, edge) {
      *BasicBlockEdge__INDEX(edge) = -1;
    } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
  } CodeRegion_ENDEACH_BasicBlock;
  //
  return true;
}
#endif//ECL_NTRACE

#line 1997 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
GraphColoring_IsNode(GraphColoring this, Temporary temporary)
{
  int index = InterfGraph_temporaryIdentity(temporary);
  return index >= 0 && index < InterfGraph_SIZE(GC_interfGraph(this)) &&
    InterfGraph_NODES(GC_interfGraph(this))[index] != NULL &&
    InterfNode_TEMPORARY(InterfGraph_NODES(GC_interfGraph(this))[index]) == temporary;
}

#line 2006 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static InterfNode
GraphColoring_GetNode(GraphColoring this, Temporary temporary)
{
  Except_CHECK(GraphColoring_IsNode(this, temporary));
  //Temporary_pretty(temporary, stderr);
  //fprintf(stderr, "\n");
  return InterfGraph_getNode(GC_interfGraph(this),InterfGraph_temporaryIdentity(temporary));
}

#line 2016 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
AddEdge(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2)
{
  /* procedure AddEdge(u,v)
   *   if((u,v) is not in adjSet) && (u != v) then
   *     adjSet <- adjSet union {(u,v), (v,u)}
   *     if u is not in precolored then
   *       adjList[u] <- adjList[u] union {v}
   *       degree[u] <- degree[u] + 1
   *     if v is not in precolored then
   *       adjList[v] <- adjList[v] union {u}
   *       degree[v] <- degree[v] + 1
   */
  InterfGraph interfGraph = GC_interfGraph(this);

#line 2031 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure AddEdge(u,v)

#line 2033 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if((u,v) is not in adjSet) && (u != v) then
  if( InterfGraph_isAdjacent(interfGraph, interfNode1, interfNode2) == false
      && (InterfNode_index(interfNode1) != InterfNode_index(interfNode2)) ) {

#line 2037 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: adjSet <- adjSet union {(u,v), (v,u)}
    InterfGraph_addToAdjSet(interfGraph, interfNode1, interfNode2);

#line 2040 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: if u is not in precolored then
    if( InterfNode_isPrecolored(interfNode1) == false ) {
      //ALGO: adjList[u] <- adjList[u] union {v}
      InterfNode_addToAdjList( interfNode1, interfNode2);
      //ALGO: degree[u] <- degree[u] + 1
      InterfNode_incDegree(interfNode1);
    }

#line 2048 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: if v is not in precolored then
    if( InterfNode_isPrecolored(interfNode2) == false ) {
      //ALGO: adjList[v] <- adjList[v] union {u}
      InterfNode_addToAdjList( interfNode2, interfNode1);
      //ALGO: degree[u] <- degree[u] + 1
      InterfNode_incDegree(interfNode2);
    }
  }
  return true;
  //ALGO:
}

#line 2061 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
MakeWorklist(GraphColoring this)
{
  /* procedure MakeWorklist()
   *   forall n in initial
   *     initial <- initial \ {n}
   *     if degree[n] >= K then
   *       spillWorklist <- spillWorklist union {n}
   *     else if MoveRelated(n) then
   *       freezeWorklist <- freezeWorklist union {n}
   *     else
   *       simplifyWorklist <- simplifyWorklist union {n}
   */

#line 2075 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure MakeWorklist()

#line 2077 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall n in initial
  IDList_FOREACH( GC_initial(this), InterfNode_, interfNode) {

#line 2080 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: initial <- initial \ {n}
    IDList_remove_( GC_initial(this), interfNode);

#line 2083 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: if degree[n] >= K then
    if( InterfNode_DEGREE(interfNode) >= NbRegisters(GraphColoring_regFile(this)) ) {
      //ALGO: spillWorklist <- spillWorklist union {n}
      PushNodeIntoWorklist(this, GC_spillWorklist(this), interfNode, _spillWorklist);
      statistiques[inSpillWorklist]++;
    }
    //ALGO: else if MoveRelated(n) then
    else if( MoveRelated(interfNode) == true ) {
      //ALGO: freezeWorklist <- freezeWorklist union {n}
      PushNodeIntoWorklist(this, GC_freezeWorklist(this), interfNode, _freezeWorklist);
      statistiques[inFreezeWorklist]++;
    }
    //ALGO: else
    else {
      //ALGO: simplifyWorklist <- simplifyWorklist union {n}
      PushNodeIntoWorklist(this, GC_simplifyWorklist(this), interfNode, _simplifyWorklist);
      statistiques[inSimplifyWorklist]++;
    }

#line 2102 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  } IDList_ENDEACH;
  //ALGO:
  return true;
}

#line 2107 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
// ALAIN: This could easily be improved with a counter. Same for the
// list of neighbors of a node, this would avoid consulting nodes that have
// already been "removed" from the graph.
static bool MoveRelated(InterfNode interfNode)
{
  /* procedure MoveRelated(n)
   *   moveList[n] inter (activeMoves union worlistMoves) != {}
   */

#line 2116 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure MoveRelated(n)

#line 2118 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: moveList[n] inter (activeMoves union worlistMoves) != {}
  IDList_FOREACH( InterfNode_MOVELIST(interfNode), InterfMove, interfMove_p) {
    if( InterfMove_BELONGTO(*interfMove_p) == _activeMoves ) return true;
    if( InterfMove_BELONGTO(*interfMove_p) == _worklistMoves ) return true;
  } IDList_ENDEACH;
  //ALGO:
  return false;
}

#line 2128 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Simplify(GraphColoring this)
{
  /* procedure Simplify()
   *   let n in simplifyWorklist [with n of lowest spill cost]
   *   simplifyWorklist <- simplifyWorklist \ {n}
   *   push(n, selectStack)
   *   forall m in Adjacent(n) (adjList[n] \ (selectStack union coalescedNodes))
   *     DecrementDegree(m)
   */
  float mincost = -1.0;

#line 2140 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure Simplify()
  InterfNode interfWith = NULL;

#line 2143 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: let n in simplifyWorklist [with n of lowest spill cost]
  // Taking the lowest spill cost will have the effect that nodes of
  // high cost will be selected first in the AssignColors, thus it should
  // favor less costly spill.
  InterfNode interfNode = IDList_firstItem(GC_simplifyWorklist(this));
  
#if 0
  IDList_FOREACH(GC_simplifyWorklist(this), InterfNode_, node) {
    float cost = InterfNode_SPILLCOST(node);
    if (cost < mincost || mincost == -1.0) {
      interfNode = node;
      mincost = cost;
    }
  } IDList_ENDEACH;
#endif
  
  statistiques[simplify]++;

#line 2161 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: simplifyWorklist <- simplifyWorklist \ {n}
  IDList_remove_( GC_simplifyWorklist(this), interfNode);
  statistiques[outSimplifyWorklist]++;

#line 2165 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: push(n, selectStack)
  PushNodeIntoWorklist(this, GC_selectStack(this), interfNode, _selectStack);
  statistiques[inSelectStack]++;

#line 2169 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall m in Adjacent(n) //(adjList[n] \ (selectStack union coalescedNodes))

#line 2171 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfWith_p) {
    interfWith = *interfWith_p;
    //ALGO: DecrementDegree(m)
    if( InterfNode_BELONGTO(interfWith) != _selectStack
	&& InterfNode_BELONGTO(interfWith) != _coalescedNodes
	&& InterfNode_BELONGTO(interfWith) != _precolored ) {
      DecrementDegree(this, interfWith);
    }
  } IDList_ENDEACH;
  //ALGO:	
  return true;
}

#line 2184 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool EnableMoves(GraphColoring this, InterfNode interfNode)
{
  // node must have lost a high degree neighbor, cf MLRISC implementation
  InterfMove interfMove = NULL;

#line 2189 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( InterfNode_MOVELIST(interfNode), InterfMove, interfMove_p) {
    interfMove = *interfMove_p;
    if( InterfMove_BELONGTO(interfMove) == _activeMoves ) {
      // hi(m):= hi(m) - 1
      *InterfMove__HIGH(interfMove) = InterfMove_HIGH(interfMove) - 1;
      // if hi(m) <= 0 then
      //   activeMoves:= activeMoves \ {m}
      //   worklistMoves:= worklistMoves U {m}
      if (InterfMove_HIGH(interfMove) <= 0) {
	IDList_remove_(  GC_activeMoves(this), interfMove);
	statistiques[outActiveMoves]++;
	PushMoveIntoWorklist(this, GC_worklistMoves(this), interfMove, _worklistMoves);
	statistiques[inWorklistMoves]++;
      }
    }
  } IDList_ENDEACH;

#line 2206 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return true;
}

#line 2210 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool DecrementDegree(GraphColoring this, InterfNode interfNode)
{
  /* procedure DecrementDegree(m)
   *   let d = degree[m]
   *   degree[m] = d-1
   *   if d = K then
   *     EnableMoves({m} union Adjacent(m))
   *     spillWorklist <- spillWorklist \ {m}
   *     if MoveRelated(m) then
   *       freezeWorklist <- freezeWorklist union {m}
   *     else
   *       simplifyWorklist <- simplifyWorklist union {m}
   */

#line 2224 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode interfWith = NULL;
  InterfMove interfMove = NULL;
  bool treated = false;

#line 2228 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure DecrementDegree(m)

#line 2230 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: let d = degree[m]
  uint32_t degree = InterfNode_DEGREE(interfNode);

#line 2233 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: degree[m] = d-1
  *InterfNode__DEGREE(interfNode) = degree -1;

#line 2236 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if d = K then
  if( degree == NbRegisters(GraphColoring_regFile(this)) ) {
    //ALGO: EnableMoves({m} union Adjacent(m))
    // Algo modified as in MLRISC report

#line 2241 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // EnableMoves(Adjacent(m))
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfWith_p) {
      interfWith = *interfWith_p;
      // n = Adjacent(m)
      // forall p in NodeMoves(n) ( = moveList[n] inter (activeMoves union worklistMoves)
      if( InterfNode_BELONGTO(interfWith) != _selectStack
	  && InterfNode_BELONGTO(interfWith) != _coalescedNodes )
	EnableMoves(this, interfWith);

#line 2250 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    } IDList_ENDEACH;

#line 2252 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: spillWorklist <- spillWorklist \ {m}
    IDList_remove_( GC_spillWorklist(this), interfNode);
    statistiques[outSpillWorklist]++;

#line 2256 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: if MoveRelated(m) then
    if( MoveRelated(interfNode) == true ) {
      //ALGO: freezeWorklist <- freezeWorklist union {m}
      PushNodeIntoWorklist(this, GC_freezeWorklist(this), interfNode, _freezeWorklist);
      statistiques[inFreezeWorklist]++;
    }
    else {
      //ALGO: simplifyWorklist <- simplifyWorklist union {m}
      PushNodeIntoWorklist(this, GC_simplifyWorklist(this), interfNode, _simplifyWorklist);
      statistiques[inSimplifyWorklist]++;
    }
  }
  //ALGO:
  return true;
}

#line 2273 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool Coalesce(GraphColoring this)
{
  /* procedure Coalesce()
   *   let m(=copy(x,y)) in worklistMoves
   *   x <- GetAlias(x)
   *   y <- GetAlias(y)
   *   if y is precolored then
   *     let (u,v) = (y,x)
   *   else
   *     let (u,v) = (x,y)
   *   worklistMoves <- worklistMoves \ {m}
   *   if (u = v) then
   *     coalescedMoves <- coalescedMove union {m}
   *     AddWorklist(u)
   *   else if v is precolored || (u,v) is in AdjSet then
   *     constrainedMoves <- constrainedMoves union {m}
   *     AddWorklist(u)
   *     AddWorklist(v)
   *   else if (u is precolored && conservative(u,v))
   *           coalescedMoves <- coalescedMoves union {m}
   *           Combine(u,v)
   *           AddWorklist(u)
   *  else
   *    ActiveMoves <- ActivesMoves  union {m}
   */

#line 2299 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  float cost;
  float mincost = -1;
  Operation operation;

#line 2303 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfMove interfMove = NULL;
  InterfMove bestinterfMove = NULL;
  InterfNode interfNode1 = NULL;
  InterfNode interfNode2 = NULL;

#line 2308 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure Coalesce()

#line 2310 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  statistiques[coalesce]++;

#line 2312 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: let m(=copy(x,y)) in worklistMoves
  IDList_FOREACH(GC_worklistMoves(this), InterfMove_, interfMove) {
    cost = InterfMove_COST(interfMove);
    if (mincost <= cost) {
      bestinterfMove = interfMove;
      mincost = cost;
    }
  } IDList_ENDEACH;

#line 2321 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //interfMove = IDList_firstItem(GC_worklistMoves(this));
  interfMove = bestinterfMove;

#line 2324 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  interfNode1 = InterfMove_DEF(interfMove);
  interfNode2 = InterfMove_USE(interfMove);

#line 2327 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: x <- GetAlias(x)
  interfNode1 = InterfNode_getAlias(interfNode1);

#line 2330 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: y <- GetAlias(y)
  interfNode2 = InterfNode_getAlias(interfNode2);

#line 2333 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if y is precolored then
  if( InterfNode_isPrecolored(interfNode2) == true ) {
    //ALGO: let (u,v) = (y,x)
    InterfNode tmp = interfNode1;
    interfNode1 = interfNode2;
    interfNode2 = tmp;
  }
  //ALGO: else
  //ALGO: let (u,v) = (x,y)

#line 2343 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: simplifyWorklist <- simplifyWorklist \ {n}
  IDList_remove_(GC_worklistMoves(this), interfMove);
  statistiques[outWorklistMoves]++;

#line 2347 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if (u = v) then
  if( InterfNode_index(interfNode1) == InterfNode_index(interfNode2)) {
    //ALGO: coalescedMoves <- coalescedMove union {m}
    PushMoveIntoWorklist(this, GC_coalescedMoves(this), interfMove, _coalescedMoves);
    statistiques[inCoalescedMoves]++;
    //ALGO: AddWorklist(u)
    AddWorkList(this, interfNode1);
  }
  //ALGO: else if v is precolored || (u,v) is in AdjSet then
  else if( InterfNode_isPrecolored(interfNode2) == true ||
	   InterfGraph_isAdjacent( GC_interfGraph(this), interfNode1, interfNode2 ) == true ) {
    //ALGO: constrainedMoves <- constrainedMoves union {m}
    PushMoveIntoWorklist(this, GC_constrainedMoves(this), interfMove, _constrainedMoves);
    statistiques[inConstrainedMoves]++;
    //ALGO: AddWorklist(u)
    AddWorkList(this, interfNode1);
    //ALGO: AddWorklist(u)
    AddWorkList(this, interfNode2);
  }
  //ALGO: else if (u is precolored && George(u,v) || u is not precolored && Briggs(u,v))
  else if((InterfNode_isPrecolored(interfNode1) == true &&
	   George(this, interfNode1, interfNode2, interfMove) == true) ||
	  (InterfNode_isPrecolored(interfNode1) == false &&
	   Briggs(this, interfNode1, interfNode2, interfMove) == true)) {
    //ALGO: coalescedMoves <- coalescedMoves union {m}
    PushMoveIntoWorklist(this, GC_coalescedMoves(this), interfMove, _coalescedMoves);
    statistiques[inCoalescedMoves]++;
    //ALGO: Combine(u,v)
    Combine(this, interfNode1, interfNode2);
    //ALGO: AddWorklist(u)
    AddWorkList(this, interfNode1);
  }
  //ALGO: else
  else {
    //ALGO: activeMoves <- activeMoves union {m}
    PushMoveIntoWorklist(this, GC_activeMoves(this), interfMove, _activeMoves);
    statistiques[inActiveMoves]++;
  }

#line 2386 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return true;
  //ALGO:
}

#line 2392 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool AddWorkList(GraphColoring this, InterfNode interfNode)
{
  /* procedure AddWorkList(u)
   *   if( u is not precolored && not(MoveRelated(u)) && degree[u]< K) then
   *     freezeWorklist <- freezeWorklist \ {u}
   *     simplifyWorklist <- simplifyWorklist union {n}
   */

#line 2401 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure AddWorkList(u)

#line 2403 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if( u is not precolored && not(MoveRelated(u)) && degree[u]< K) then
  if( InterfNode_isPrecolored(interfNode) == false
      && MoveRelated(interfNode) == false
      && InterfNode_DEGREE(interfNode) < NbRegisters(GC_regFile(this))) {
    //ALGO: freezeWorklist <- freezeWorklist \ {u}
    switch(InterfNode_BELONGTO(interfNode)) {
    case _precolored:
      IDList_remove_(GC_precolored(this), interfNode);
      break;
    case _initial:
      IDList_remove_(GC_initial(this), interfNode);
      break;
    case _simplifyWorklist:
      IDList_remove_(GC_simplifyWorklist(this), interfNode);
      break;
    case _freezeWorklist:
      IDList_remove_(GC_freezeWorklist(this), interfNode);
      statistiques[outFreezeWorklist]++;
      break;
    case _spillWorklist:
      IDList_remove_(GC_spillWorklist(this), interfNode);
      break;
    case _spilledNodes:
      IDList_remove_(GC_spillWorklist(this), interfNode);
      break;
    case _coalescedNodes:
      IDList_remove_(GC_coalescedNodes(this), interfNode);
      break;
    case _coloredNodes:
      IDList_remove_(GC_coloredNodes(this), interfNode);
      break;
    case _selectStack:
      IDList_remove_(GC_selectStack(this), interfNode);
      break;
    case _discardedNodes:
      IDList_remove_(GC_discardedNodes(this), interfNode);
      break;
    case noneNode:
      break;
    default:
      Except_CHECK(false);
    }
    //ALGO: simplifyWorklist <- simplifyWorklist union {n}
    PushNodeIntoWorklist(this, GC_simplifyWorklist(this), interfNode, _simplifyWorklist);
    statistiques[inSimplifyWorklist]++;
  }
  //ALGO:
  return true;
}

#line 2454 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Briggs(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2, InterfMove interfMove)
{
  /* function Briggs(u,v)
   *   let k = 0
   *   forall n in (Adjacent(u) union Adjacent(v))
   *     if degree[n] >= K then
   *       k<-k+1
   *   return( k < K )
   */

#line 2465 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO:
  //ALGO: let k = 0
  uint32_t k = 0;

#line 2469 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall n in (Adjacent(u) union Adjacent(v))
  IDList_FOREACH( InterfNode_ADJLIST(interfNode1), InterfNode, interfNode_p) {
    if(InterfNode_BELONGTO(*interfNode_p) != _selectStack &&
       InterfNode_BELONGTO(*interfNode_p) != _coalescedNodes &&
       (InterfNode_BELONGTO(*interfNode_p) == _precolored ||
	InterfNode_DEGREE(*interfNode_p) >= NbRegisters(GC_regFile(this))))
      //ALGO: if degree[n] >= K then
      //ALGO: k<-k+1
      k++;
  } IDList_ENDEACH;

#line 2480 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO2: forall n in (Adjacent(u) union Adjacent(v))

#line 2482 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALAIN: this could be slightly improved by doing k-- for each
  //common neighbor with exactly K neighbors, but then to be
  //sure that ActiveMoves only contains noncoalescable moves
  //will be more complicated. So, we implement the basic version.
  IDList_FOREACH( InterfNode_ADJLIST(interfNode2), InterfNode, interfNode_p) {
    if(InterfNode_BELONGTO(*interfNode_p) != _selectStack &&
       InterfNode_BELONGTO(*interfNode_p) != _coalescedNodes &&
       InterfGraph_isAdjacent(GC_interfGraph(this), *interfNode_p, interfNode1) == false &&
       (InterfNode_BELONGTO(*interfNode_p) == _precolored ||
	InterfNode_DEGREE(*interfNode_p) >= NbRegisters(GC_regFile(this))))
      k++;
  } IDList_ENDEACH;

#line 2496 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: return( k < K )
  if( k < NbRegisters(GC_regFile(this)) ) {
    //fprintf(stderr, "Briggs wins\n");
    statistiques[briggsWins]++;
    return true;
  }
  else {
    *InterfMove__HIGH(interfMove) = k - (NbRegisters(GC_regFile(this)) -1);		
    //fprintf(stderr, "Briggs loses\n");
    statistiques[briggsLoses]++;
    return false;
  }
  //ALGO:
}

#line 2512 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
George(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2, InterfMove interfMove)
{
  /*  function George(u,v)
   *   forall n in Adjacent(v)
   *     if( degree[n] >= K && (n,u) is not in adjSet && n is precolored) then
   *       return false
   *   return true;
   */

#line 2522 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Note1: George is called if interfNode1 is a precolored
  // node, so George test returns true iff interfNode2 has no
  // neighbor of high degree among the non precolored nodes that
  // is not already neighbor of interfNode1.  cf Code in MLRISC

#line 2527 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  uint32_t k = 0;

#line 2529 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( InterfNode_ADJLIST(interfNode2), InterfNode, interfNode_p) {
    if( InterfNode_BELONGTO(*interfNode_p) != _selectStack &&
	InterfNode_BELONGTO(*interfNode_p) != _coalescedNodes &&
	InterfNode_BELONGTO(*interfNode_p) != _precolored &&
	InterfNode_DEGREE(*interfNode_p) >= NbRegisters(GC_regFile(this)) &&
	InterfGraph_isAdjacent(GC_interfGraph(this), *interfNode_p, interfNode1) == false)
      k++;
  } IDList_ENDEACH;

#line 2538 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (k>0) {
    *InterfMove__HIGH(interfMove) = k;
    //fprintf(stderr, "George loses\n");
    statistiques[georgeLoses]++;
    return false;
  }
  else {
    statistiques[georgeWins]++;
    //fprintf(stderr, "George wins\n");
    return true;
  }
}

#line 2552 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Combine(GraphColoring this, InterfNode interfNode1, InterfNode interfNode2)
{
  /* procedure Combine(u,v)
   *   if v is in freezeWorklist then
   *     freezeWorklist <- freezeWorklist \ {v}
   *   else
   *     spillWorklist <- spillWorklist \ {v}
   *   coalescedNodes <- coalescedNodes union {v}
   *   alias[v] <- u
   *   movelist[u] <- movelist[u] union movelist[v]
   *   EnableMoves(v)
   *   forall t in Adjacent(v)
   *     AddEdge(t,u)
   *     DecrementDegree(t)
   *   if degree[u] >= K && u is in freezeWorklist then
   *     freezeWorklist <- freezeWorklist \ {u}
   *     spillWorklist <- spillWorklist union {u}
   */

#line 2572 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure Combine(u,v)

#line 2574 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //Note: at most one node can be precolored, this is interfNode1

#line 2577 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  bool loosingHiDegNode = false;
  bool HiDegNode1 = false;
  InterfNode interfNode3 = NULL;
  InterfMove interfMove = NULL;

#line 2582 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (InterfNode_BELONGTO(interfNode1) == _precolored ||
      (InterfNode_BELONGTO(interfNode1) != _precolored &&
       InterfNode_DEGREE(interfNode1) >= NbRegisters(GC_regFile(this)))) {
    HiDegNode1 = true;
    if (InterfNode_DEGREE(interfNode2) >= NbRegisters(GC_regFile(this))) {
      loosingHiDegNode = true;	
    }
  }

#line 2591 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if v is in freezeWorklist then
  if( InterfNode_BELONGTO(interfNode2) == _freezeWorklist ) {
    //ALGO: freezeWorklist <- freezeWorklist \ {v}
    IDList_remove_(GC_freezeWorklist(this), interfNode2);
    statistiques[outFreezeWorklist]++;
  }
  else {
    //ALGO: spillWorklist <- spillWorklist \ {v}
    IDList_remove_(GC_spillWorklist(this), interfNode2);
    statistiques[outSpillWorklist]++;
  }

#line 2603 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //coalescedNodes <- coalescedNodes union {v}
  PushNodeIntoWorklist(this, GC_coalescedNodes(this), interfNode2, _coalescedNodes);
  statistiques[inCoalescedNodes]++;

#line 2607 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: alias[v] <- u
  InterfNode_setAlias(interfNode2, interfNode1);

#line 2610 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: EnableMoves(v)
  //use the principles of MLRISC instead with slight modifications.
  if (InterfNode_isPrecolored(interfNode1) == true) { // George's test applied
    IDList_FOREACH( InterfNode_MOVELIST(interfNode2), InterfMove, interfMove_p) {
      interfMove = *interfMove_p;

#line 2616 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      interfNode3 = InterfMove_DEF(interfMove);

#line 2618 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // find the other end of the move
      if (interfNode3 == interfNode2) {
	interfNode3 = InterfMove_USE(interfMove);
	interfNode3 = InterfNode_getAlias(interfNode3);
      }

#line 2624 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (InterfNode_isPrecolored(interfNode3) == false ) { // neither interfNode1 nor another precolored

#line 2626 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	if (InterfMove_BELONGTO(interfMove) == _activeMoves ||
	    InterfMove_BELONGTO(interfMove) == _worklistMoves) {
	  InterfNode_addToMoveList(interfNode1, interfMove);

#line 2630 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  if( InterfMove_BELONGTO(interfMove) == _activeMoves ) {
	    // may be coalesced by George's test
	    // this is missing in MLRISC. Bug?
	    IDList_remove_( GC_activeMoves(this), interfMove);
	    statistiques[outActiveMoves]++;
	    PushMoveIntoWorklist(this, GC_worklistMoves(this), interfMove, _worklistMoves);
	    statistiques[inWorklistMoves]++;
	  }
	}
      }
    } IDList_ENDEACH;
  }
  else { //ALGO: movelist[u] <- movelist[u] union movelist[v]
    IDList_FOREACH( InterfNode_MOVELIST(interfNode2), InterfMove, interfMove_p) {
      interfMove = *interfMove_p;
      if (InterfMove_BELONGTO(interfMove) == _activeMoves ||
	  InterfMove_BELONGTO(interfMove) == _worklistMoves) {
	InterfNode_addToMoveList(interfNode1, interfMove);
      }
    } IDList_ENDEACH;
  }

#line 2652 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall t in Adjacent(v)
  IDList_FOREACH( InterfNode_ADJLIST(interfNode2), InterfNode, interfNode_p) {
    if( InterfNode_BELONGTO(*interfNode_p) != _selectStack &&
	InterfNode_BELONGTO(*interfNode_p) != _coalescedNodes ) {
      //ALGO: AddEdge(t,u)
      if (InterfGraph_isAdjacent(GC_interfGraph(this), *interfNode_p, interfNode1) == false) {
	AddEdge(this, *interfNode_p, interfNode1);
	// should not count the old neighbor
	if (InterfNode_BELONGTO(*interfNode_p) != _precolored) {
	  *InterfNode__DEGREE(*interfNode_p) = InterfNode_DEGREE(*interfNode_p) - 1;
	}
	else if (HiDegNode1 == true) {
	  // The node is precolored and his moves may be coalesced by George's rule
	  // because it may gain a high degree neighbor. This is missing in MLRISC. Bug?
	  // Actually, this could be limited to moves connected to a neighbor of interfNode1
	  EnableMoves(this, *interfNode_p);					
	}				
      }
      //ALGO: DecrementDegree(t)
      else {
	if (loosingHiDegNode == true) {
	  EnableMoves(this, *interfNode_p);
	}
	if (InterfNode_BELONGTO(*interfNode_p) != _precolored) {
	  DecrementDegree(this, *interfNode_p);
	}
      }
    }
  } IDList_ENDEACH;

#line 2682 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: if degree[u] >= K && u is in freezeWorklist then
  if( InterfNode_BELONGTO(interfNode1) == _freezeWorklist &&
      InterfNode_DEGREE(interfNode1) >= NbRegisters(GC_regFile(this))) {
    //ALGO: freezeWorklist <- freezeWorklist \ {u}
    IDList_remove_(GC_freezeWorklist(this), interfNode1);
    statistiques[outFreezeWorklist]++;

#line 2689 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: spillWorklist <- spillWorklist union {u}
    PushNodeIntoWorklist(this, GC_spillWorklist(this), interfNode1, _spillWorklist);
    statistiques[inSpillWorklist]++;
  }
  //ALGO:
  return true;
}

#line 2697 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Freeze(GraphColoring this)
{
  /* procedure Freeze()
   *   let u in freezeWorklist
   *   freezeWorklist <- freezeWorklist \ {u}
   *   simplifyWorklist <- simplifyWorklist union {u}
   *   FreezeMoves(u)
   */

#line 2707 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure Freeze()

#line 2709 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: let u in freezeWorklist

#line 2711 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  float cost;
  float mincost = -1;
  Operation operation;

#line 2715 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode interfNode = NULL;
  InterfNode bestinterfNode = NULL;
  InterfMove interfMove = NULL;

#line 2719 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  statistiques[freeze]++;

#line 2721 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(GC_freezeWorklist(this), InterfNode_, interfNode) {
    cost = 0;
    IDList_FOREACH( InterfNode_MOVELIST(interfNode), InterfMove, interfMove_p) {
      interfMove = *interfMove_p;
      if (InterfMove_BELONGTO(interfMove) == _activeMoves ||
	  InterfMove_BELONGTO(interfMove) == _worklistMoves) {
	cost = cost + InterfMove_COST(interfMove);
      }
    } IDList_ENDEACH;

#line 2731 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (mincost >= cost || mincost == -1) {
      bestinterfNode = interfNode;
      mincost = cost;
    }
  } IDList_ENDEACH;

#line 2737 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //InterfNode interfNode = IDList_firstItem(GC_freezeWorklist(this));
  interfNode = bestinterfNode;

#line 2740 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: freezeWorklist <- freezeWorklist \ {u}
  IDList_remove_(GC_freezeWorklist(this), interfNode);
  statistiques[outFreezeWorklist]++;

#line 2744 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: simplifyWorklist <- simplifyWorklist union {u}
  PushNodeIntoWorklist(this, GC_simplifyWorklist(this), interfNode, _simplifyWorklist);
  statistiques[inSimplifyWorklist]++;
  //ALGO: FreezeMoves(u)
  FreezeMoves(this, interfNode);
  return true;
}

#line 2753 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
FreezeMoves(GraphColoring this, InterfNode interfNode_u)
{
  /* procedure FreezeMoves
   *   forall m(=copy(x,y)) in NodeMoves(u)
   *     if GetAlias(y)=GetAlias(u) then
   *       v <- GetAlias(x)
   *     else
   *       v <- GetAlias(y)
   *     activeMoves <- activeMoves \ {m}
   *     frozenMoves <- frozenMoves union {m}
   *     if NodeMoves(v) = {} && degree[v]< K [&& v in freezeWorkList]Fix
   *       freezeWorklist <- freezeWorklist \ {v}
   *       simplifyWorklist <- simplifyWorklist union {v}
   */

#line 2769 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode interfNode_x = NULL;
  InterfNode interfNode_y = NULL;
  InterfNode interfNode_v = NULL;

#line 2773 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure FreezeMoves

#line 2775 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall m(=copy(x,y)) in NodeMoves(u)
  IDList_FOREACH( InterfNode_MOVELIST(interfNode_u), InterfMove, interfMove_p) {
    if( InterfMove_BELONGTO(*interfMove_p) == _activeMoves ||
	InterfMove_BELONGTO(*interfMove_p) == _worklistMoves ) {
      interfNode_x = InterfMove_DEF(*interfMove_p);
      interfNode_y = InterfMove_USE(*interfMove_p);

#line 2782 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: if GetAlias(y)=GetAlias(u) then
      if( InterfNode_getAlias(interfNode_y) == InterfNode_getAlias(interfNode_u) )
	//ALGO: v <- GetAlias(x)
	interfNode_v = InterfNode_getAlias(interfNode_x);
      else
	//ALGO: v <- GetAlias(y)
	interfNode_v = InterfNode_getAlias(interfNode_y);

#line 2790 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: activeMoves <- activeMoves \ {m}
      IDList_remove_(GC_activeMoves(this), *interfMove_p);
      statistiques[outActiveMoves]++;

#line 2794 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: frozenMoves <- frozenMoves union {m}
      PushMoveIntoWorklist(this, GC_frozenMoves(this), *interfMove_p, _frozenMoves);
      statistiques[inFrozenMoves]++;
      //ALGO: if NodeMoves(v) = {} && degree[v]< K

#line 2799 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if( InterfNode_BELONGTO(interfNode_v) != _precolored &&
	  InterfNode_DEGREE(interfNode_v) < NbRegisters(GC_regFile(this)) ) {

#line 2803 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	//ALGO: freezeWorklist <- freezeWorklist \ {v}
	//ALGO: simplifyWorklist <- simplifyWorklist union {v}
	if( MoveRelated(interfNode_v) == false && InterfNode_BELONGTO(interfNode_v) == _freezeWorklist) {
	  IDList_remove_(GC_freezeWorklist(this), interfNode_v);
	  statistiques[outFreezeWorklist]++;

#line 2809 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  PushNodeIntoWorklist(this, GC_simplifyWorklist(this), interfNode_v, _simplifyWorklist);
	  statistiques[inSimplifyWorklist]++;
	}
      }
    }
  } IDList_ENDEACH;
  return true;
}

#line 2818 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
SpillNode(GraphColoring this, InterfNode interfNode)
{
  /* Procedure SpillNode (m)
   * //  simplifyWorklist <- simplifyWorklist union {m}
   * //  FreezeMoves(m)
   * // Modified to handle spill immediately
   * push m in selectStack
   * FreezeMoves(m)
   * EnableMoves(Adjacent(n))
   * DecrementDegree(Adjacent(n))
   */

#line 2831 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode interfWith = NULL;

#line 2833 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Modification to handle spill immediately
  PushNodeIntoWorklist(this, GC_selectStack(this), interfNode, _selectStack);
  InterfNode_setFlag(interfNode, InterfNode_POTENTIALSPILL);
  
  statistiques[inSelectStack];
  FreezeMoves(this, interfNode);

#line 2840 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // EnableMoves(Adjacent(n))
  IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfWith_p) {
    interfWith = *interfWith_p;
    if( InterfNode_BELONGTO(interfWith) != _selectStack
	&& InterfNode_BELONGTO(interfWith) != _coalescedNodes ) {
      EnableMoves(this, interfWith);
    }		
  } IDList_ENDEACH;

#line 2849 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // DecrementDegree(Adjacent(n))

#line 2851 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfWith_p) {
    interfWith = *interfWith_p;
    if( InterfNode_BELONGTO(interfWith) != _selectStack
	&& InterfNode_BELONGTO(interfWith) != _coalescedNodes
	&& InterfNode_BELONGTO(interfWith) != _precolored ) {
      DecrementDegree(this, interfWith);
    }
  } IDList_ENDEACH;
}

#line 2861 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
SelectSpill(GraphColoring this)
{
  /* procedure SelectSpill
   *   let m in spillWorklist
   *   spillWorklist <- spillWorklist \ {m}
   *   SpillNode(m)
   */

#line 2870 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode interfNode = NULL;	
  float min_cost = -1;
  float cost;

#line 2874 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  statistiques[selectSpill]++;

#line 2876 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(GC_spillWorklist(this), InterfNode_, interfN) {

#line 2878 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    cost = InterfNode_SPILLCOST(interfN)/InterfNode_DEGREE(interfN);

#line 2880 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALAIN: to be improved once we will reuse objects in the second phase
    if( (min_cost >= cost || min_cost == -1) &&
	(!InterfNode_hasFlag(interfN, InterfNode_ALREADYSPILLED) ||
	 /* For PHI related nodes, we allow respill. */
	 ((GraphColoring_flags(this) & GRA_GCF_SPILL_PHI_ON_RESPILL) && InterfNode_hasPHI(interfN)))) {
      interfNode = interfN;
      min_cost = cost;
    }
  } IDList_ENDEACH;

#line 2890 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (interfNode == NULL) {
    //ALAIN: this is a fatal error today because we use a
    //spill-everywhere strategy, so there should always be
    //something else to spill.
    fprintf(stderr, "Fatal error: needs to spill an already spilled temporary!!\n");
    abort();
  }

#line 2898 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_remove_( GC_spillWorklist(this), interfNode);
  statistiques[outSpillWorklist]++;

#line 2901 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  SpillNode(this, interfNode);

#line 2903 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return true;
}

#line 2907 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
AssignRegisters(GraphColoring this)
{
  Register reg;

#line 2912 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coloredNodes(this), InterfNode_, interfNode) {
    reg = InterfNode_COLOR(interfNode);
    if (!GRA_isAssigned(InterfNode_TEMPORARY(interfNode)))
      Temporary_assignRegister(InterfNode_TEMPORARY(interfNode), reg);
  } IDList_ENDEACH;

#line 2918 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coalescedNodes(this), InterfNode_, interfNode) {
    reg = InterfNode_COLOR(InterfNode_getAlias(interfNode));
    if (!GRA_isAssigned(InterfNode_TEMPORARY(interfNode)))
      Temporary_assignRegister(InterfNode_TEMPORARY(interfNode), reg);
  } IDList_ENDEACH;

#line 2924 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return true;
}

#line 2928 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
DiscardMoves(GraphColoring this)
{

#line 2932 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Rename all use/defs sites of coalesced nodes. */
  IDList_FOREACH(GC_coalescedNodes(this), InterfNode_, interfNode) {
    InterfNode alias = InterfNode_getAlias(interfNode);

#line 2936 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    IDList_FOREACH(InterfNode_USES(interfNode), Operation, operation_p) {
      Operation_ARGUMENTS_FOREACH_Temporary(*operation_p, temporary) {
	if(temporary == InterfNode_TEMPORARY(interfNode)) {
	  // Update the argument.
	  Operation_setArgument(*operation_p, Operation_ARGUMENTS_INDEX,
				InterfNode_TEMPORARY(alias));
	}
      } Operation_ARGUMENTS_ENDEACH_Temporary;
    } IDList_ENDEACH;

#line 2946 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    IDList_FOREACH(InterfNode_DEFS(interfNode), Operation, operation_p) {
      Operation_RESULTS_FOREACH_Temporary(*operation_p, temporary) {
	if(temporary == InterfNode_TEMPORARY(interfNode)) {
	  // Update the result.
	  Operation_setResult(*operation_p, Operation_RESULTS_INDEX,
			      InterfNode_TEMPORARY(alias));
	}
      } Operation_RESULTS_ENDEACH_Temporary;
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 2957 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /* Remove coalesced nodes from live in live out sets and insert aliases. */
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    BasicBlock_liveInSet_FORSORT_Temporary(basicBlock, temporaryIn) {
      if(GC_IsNode(this, temporaryIn) &&
	 InterfNode_BELONGTO(GC_GetNode(this, temporaryIn)) == _coalescedNodes) {
	InterfNode alias = InterfNode_getAlias(GC_GetNode(this, temporaryIn));
	TemporarySet_remove(BasicBlock_liveInSet(basicBlock), temporaryIn);
	TemporarySet_insert(BasicBlock_liveInSet(basicBlock), InterfNode_TEMPORARY(alias));
      }
    } BasicBlock_liveInSet_ENDSORT_Temporary;

#line 2968 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporaryOut) {
      if(GC_IsNode(this, temporaryOut) &&
	 InterfNode_BELONGTO(GC_GetNode(this, temporaryOut)) == _coalescedNodes) {
	InterfNode alias = InterfNode_getAlias(GC_GetNode(this, temporaryOut));
	TemporarySet_remove(BasicBlock_liveOutSet(basicBlock), temporaryOut);
	TemporarySet_insert(BasicBlock_liveOutSet(basicBlock), InterfNode_TEMPORARY(alias));
      }
    } BasicBlock_liveOutSet_ENDSORT_Temporary;

#line 2977 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    /* Remove coalesced move operations. */
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      if (Operator_isSeqCopy(Operation_operator(operation)) &&
	  Operation_getResult(operation, 0) == Operation_getArgument(operation, 0))
	BasicBlock_detachOperation(Operation_basicBlock(operation), operation);
    } BasicBlock_ENDEACH_Operation;
  } CodeRegion_ENDEACH_BasicBlock;
  return true;
}

#line 2987 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
ForceSpillNode(GraphColoring this, InterfNode interfNode)
{
  static int n;
  if (GraphColoring_flags(this) & GRA_GCF_SPILL_PHI_USES) {
    // We force spill of initial temporaries in PHI arguments 
    if (Temporary_index(InterfNode_TEMPORARY(interfNode)) < 
	GraphColoring_INITIALTMPCOUNT(this)) {
      IDList_FOREACH(InterfNode_USES(interfNode), Operation, operation_p) {
	if (Operation_operator(*operation_p) == Operator_PHI) {
	  return true;
	}
      } IDList_ENDEACH;
    }
  }
  if (GraphColoring_flags(this) & GRA_GCF_SPILL_PHI_DEFS) {
    // We force spill of initial temporaries in PHI result
    if (Temporary_index(InterfNode_TEMPORARY(interfNode)) < 
	GraphColoring_INITIALTMPCOUNT(this)) {
      IDList_FOREACH(InterfNode_DEFS(interfNode), Operation, operation_p) {
	if (Operation_operator(*operation_p) == Operator_PHI) {
	  return true;
	}
      } IDList_ENDEACH;
    }
  }
  if (n != -1) {
    if (n == 0) {
      char *env = GETENV("GRA_FORCE");
      if (env != NULL) n = atoi(env);
      else n = -1;
    }
    if (Temporary_IDENTITY(InterfNode_TEMPORARY(interfNode)) == n) {
      return true;
    }
  }
  return false;
}

#line 3026 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
ForceSpill(GraphColoring this)
{
  /* procedure ForceSpill()
   *   forall n in initial
   *   if force_spill(n)
   *     initial <- initial \  {n}
   *     spilledNodes <- spilledNodes union {n}
   */

#line 3036 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  bool forced_spill = false;

#line 3038 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall n in initial
  IDList_FOREACH( GC_initial(this), InterfNode_, interfNode) {

#line 3041 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: if force_spill(n)
    if (ForceSpillNode(this, interfNode)) {
      //ALGO: initial <- initial \ {n}
      IDList_remove_( GC_initial(this), interfNode);
      //ALGO: spilledNodes <- spilledNodes union {n}
      PushNodeIntoWorklist(this, GC_spilledNodes(this), interfNode, _spilledNodes);
      statistiques[inSpilledNodes]++;
      totalspillcost += InterfNode_SPILLCOST(interfNode);
      forced_spill = true;
    }
  } IDList_ENDEACH;
  // Put all other nodes in colored list
  if (forced_spill) {
    IDList_FOREACH( GC_initial(this), InterfNode_, interfNode) {
      IDList_remove_( GC_initial(this), interfNode);
      PushNodeIntoWorklist(this, GC_coloredNodes(this), interfNode, _coloredNodes);
      statistiques[inColoredNodes]++;
    } IDList_ENDEACH;
  }
  return forced_spill;
}

#line 3063 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Register
SimpleColoring(GraphColoring this, InterfNode node, BitSet okColors)
{
  Register reg = Register__UNDEF;
  IQueue_FOREACH(AllocatableRegisters(GC_regFile(this)), Register, reg_p) {
    if (BitSet_contains(okColors, *reg_p)) {
      reg = *reg_p;
      break;
    }
  } IQueue_ENDEACH;
  Except_CHECK(reg != Register__UNDEF);
  return reg;
}

#line 3077 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Register
BiasedColoring(GraphColoring this, InterfNode node, BitSet okColors)
{
  float *reg_benefit = alloca(sizeof(float)*Register__COUNT);
  RegFile regFile = GraphColoring_regFile(this);
  float max;
  Register max_reg = Register__UNDEF;

#line 3085 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IQueue_FOREACH(AllocatableRegisters(GC_regFile(this)), Register, reg_p) {
      reg_benefit[*reg_p] = 0.0;
  } IQueue_ENDEACH;

#line 3089 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(InterfNode_MOVELIST(node), InterfMove, interfMove_p) {
    InterfNode x = InterfMove_USE(*interfMove_p);
    InterfNode y = InterfMove_DEF(*interfMove_p);
    if (x == node) x = y; /* Select the other one. */
    if( InterfNode_BELONGTO(x) == _coloredNodes ||
	InterfNode_BELONGTO(x) == _precolored ) {
      Register color = InterfNode_COLOR(x);
      reg_benefit[color] += InterfMove_COST(*interfMove_p);
    }
  } IDList_ENDEACH;

#line 3100 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  max_reg = Register__UNDEF;
  max = -1.0;
  IQueue_FOREACH(AllocatableRegisters(GC_regFile(this)), Register, reg_p) {
    if (BitSet_contains(okColors, *reg_p)) {
      if (reg_benefit[*reg_p] > max) {
	max = reg_benefit[*reg_p];
	max_reg = *reg_p;
      }
    }
  } IQueue_ENDEACH;
  Except_CHECK(max_reg != Register__UNDEF);
  return max_reg;
}

#line 3114 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
AssignColors(GraphColoring this)
{
  /* procedure AssignColors()
   *   while SelectStack is not empty
   *     let n = pop(SelectStack)
   *     okColors <- {0, ..., K-1}
   *     forall w in adjList[n]
   *        if GetAlias(w) is in (coloredNode union precolored) then
   *           okColors <- okColors \ {color[GetAlias(w)]}
   *     if okColors = {} then
   *       spilledNodes <- spilledNodes union {n}
   *     else
   *       coloredNodes <- coloredNodes union {n}
   *       let c in okColors
   *       color[n] <- c
   *   forall n in coalescedNodes
   *     color[n] <- color[GetAlias(n)]
   */

#line 3134 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: procedure AssignColors()

#line 3136 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode interfNode = NULL;
  BitSet_ okColors_, *okColors = &okColors_;
  Register reg;

#line 3140 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  InterfNode alias = NULL;
  BitSet_Ctor(okColors, GraphColoring_memory(this), Register__COUNT);

#line 3143 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GRA_TRACE) {
    fprintf(GRA_TRACE, "Assign Colors:\n");
  }
  //ALGO: while SelectStack is not empty
  while( IDList_isEmpty(GC_selectStack(this)) == false ) {

#line 3149 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: let n = pop(SelectStack)
    interfNode = IDList_lastItem(GC_selectStack(this));
    IDList_remove_(GC_selectStack(this), interfNode);

#line 3153 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: okColors <- {0, ..., K-1}
    BitSet_empty(okColors);
    IQueue_FOREACH(AllocatableRegisters(GC_regFile(this)), Register, reg) {
      BitSet_insert(okColors, *reg);
    } IQueue_ENDEACH;

#line 3160 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: forall w in adjList[n]
    IDList_FOREACH(InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      //ALGO: if GetAlias(w) is in (coloredNode union precolored) then
      alias = InterfNode_getAlias(*interfNode_p);
      if( InterfNode_BELONGTO(alias) == _coloredNodes ||
	  InterfNode_BELONGTO(alias) == _precolored ) {
	//ALGO: okColors <- okColors \ {color[GetAlias(w)]}
	BitSet_remove(okColors, InterfNode_COLOR(alias));
      }
    } IDList_ENDEACH;

#line 3171 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: if okColors = {} then
    if (((GraphColoring_flags(this) & GRA_GCF_SPILL_POTENTIAL) &&
	 InterfNode_hasFlag(interfNode, InterfNode_POTENTIALSPILL)) ||
	BitSet_isEmpty(okColors) ) {
      //ALGO: spilledNodes <- spilledNodes union {n}
      //NOTE: The spilledNodes set is in fact ordered.
      //NOTE: A list from highest to lowest spill cost, lowest is last. 
      //NOTE: This order is used by the PruneUselessSpill function.
      PushNodeIntoWorklist(this, GC_spilledNodes(this), interfNode, _spilledNodes);
      statistiques[inSpilledNodes]++;
      InterfNode_setFlag(interfNode, InterfNode_ACTUALSPILL);
      totalspillcost += InterfNode_SPILLCOST(interfNode);
      if (GRA_TRACE) {
	fprintf(GRA_TRACE, "spill ");
	Temporary_pretty(InterfNode_TEMPORARY(interfNode), GRA_TRACE);
	fprintf(GRA_TRACE, " (cost: %.2f)\n", InterfNode_SPILLCOST(interfNode));
      }
    }
    else {
      Register color;
      //ALGO: coloredNodes <- coloredNodes union {n}
      PushNodeIntoWorklist(this, GC_coloredNodes(this), interfNode, _coloredNodes);
      statistiques[inColoredNodes]++;
      //ALGO: let c in okColors
      //ALGO:	color[n] <- c
      if (!(GraphColoring_flags(this) & GRA_GCF_NO_BIASED)) {
	color = BiasedColoring(this, interfNode, okColors);
      } else {
	color = SimpleColoring(this, interfNode, okColors);
      }
      InterfNode_setColor(interfNode, color);
      if (GRA_TRACE) {
	fprintf(GRA_TRACE, "assign %s to ", Register_name(color));
	Temporary_pretty(InterfNode_TEMPORARY(interfNode), GRA_TRACE);
	fprintf(GRA_TRACE, " (cost:%.2f)\n",  InterfNode_SPILLCOST(interfNode));
      }
    }
  }
  
  BitSet_Dtor(okColors);

#line 3212 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: forall n in coalescedNodes

#line 3214 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coalescedNodes(this), InterfNode_, interfNode2) {
    //ALGO: color[n] <- color[GetAlias(n)]
    InterfNode_setColor(interfNode2, InterfNode_COLOR(InterfNode_getAlias(interfNode2)));
  } IDList_ENDEACH;

#line 3219 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return true;
  //ALGO:
}

#line 3223 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
SpillUseOnPHI(GraphColoring this, Temporary spilled, Operation operation)
{
  Except_CHECK(Operation_operator(operation) == Operator_PHI);
  PHI_arguments_FOREACH_temporary(operation, edge, idx, use) {
    if (use == spilled) {
      BasicBlock pred = BasicBlockEdge_origBlock(edge);
      Temporary representative = GraphColoring_spillRepresentative(this, spilled);
      Temporary tmp = GraphColoring_newVirtual(this);
      TemporaryTable_insert(GraphColoring_tmpTable(this), tmp);

#line 3234 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if (GRA_TRACE) {
	fprintf(GRA_TRACE, "Inserting reload for PHI argument: ");
	Temporary_pretty(use, GRA_TRACE);
	fprintf(GRA_TRACE, " of PHI: ");
	Operation_pretty(operation, GRA_TRACE);
      }
      // Insert reload at the end of the predecessor block
      Spill_insertReload(pred, NULL, tmp, representative);
      // Update the argument.
      Operation_setArgument(operation, idx, tmp);
      // Update the live out of the predecessor block
      TemporarySet_insert(BasicBlock_liveOutSet(BasicBlockEdge_origBlock(edge)), tmp);
    }
  } PHI_arguments_ENDEACH_temporary;
}

#line 3250 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
SpillDefOnPHI(GraphColoring this, Temporary spilled, Operation operation)
{
  Except_CHECK(Operation_operator(operation) == Operator_PHI);
  {
    Temporary representative = GraphColoring_spillRepresentative(this, spilled);
    BasicBlock basicBlock = Operation_basicBlock(operation);
    Temporary tmp = GraphColoring_newVirtual(this);
    TemporaryTable_insert(GraphColoring_tmpTable(this), tmp);

#line 3260 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (GRA_TRACE) {
      fprintf(GRA_TRACE, "Inserting spill for PHI result of PHI: ");
      Operation_pretty(operation, GRA_TRACE);
    }
    // Insert spill at the start of basic block (after PHI operations).
    Spill_appendSpill(basicBlock, NULL, tmp, representative);
    // Update the PHI result.
    Operation_setResult(operation, 0, tmp);
    // Update the live in of the block.
    TemporarySet_insert(BasicBlock_liveInSet(basicBlock), tmp);
  }
}

#line 3273 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Operation
InterfNode_uniqueDef(InterfNode node)
{
  int count = 0;
  IDList_FOREACH(InterfNode_DEFS(node), Operation, operation_p) {
    Except_CHECK(count == 0);
    return *operation_p;
  } IDList_ENDEACH;
  return NULL;
}

#line 3284 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static Operation
InterfNode_uniqueUse(InterfNode node)
{
  int count = 0;
  IDList_FOREACH(InterfNode_USES(node), Operation, operation_p) {
    Except_CHECK(count == 0);
    return *operation_p;
  } IDList_ENDEACH;
  return NULL;
}

#line 3295 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
SpillDefAndPHI(GraphColoring this, Temporary spilled, Operation operation)
{
  Except_CHECK(Operation_operator(operation) == Operator_PHI);
  {
    Temporary representative = GraphColoring_spillRepresentative(this, spilled);
    if (GRA_TRACE) {
      fprintf(GRA_TRACE, "Spill DefAndPHI: ");
      Operation_pretty(operation, GRA_TRACE);
    }
    // for each operand node, store the definition
    PHI_arguments_FOREACH_temporary(operation, edge, idx, use) {
      InterfNode useNode = GC_GetNode(this, use);
      Operation defOp = InterfNode_uniqueDef(useNode);
      if (InterfNode_BELONGTO(useNode) != _spilledNodes) {
	// If the use is not itself spilled in this round
	// Generate store after the definition
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Inserting Spill for PHI use: ");
	  Operation_pretty(defOp, GRA_TRACE);
	}
	Spill_appendSpill(Operation_basicBlock(defOp), defOp, use, representative);      
      } else {
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Skiping Spill for spilled PHI use: ");
	  Operation_pretty(defOp, GRA_TRACE);
	}
      }

#line 3324 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    } PHI_arguments_ENDEACH_temporary;
    // Now discard the PHI operation
    BasicBlock_detachOperation(Operation_basicBlock(operation), operation);

#line 3328 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // Update live information.
    // Note that we don't know how to update simply the live analysis
    // as discarding the PHI may have non local effects on the live 
    // information. This will require to redo live analysis after this round of
    // rewrite.
    GRA_redo_liveness = true;
  }

#line 3336 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
}

#line 3338 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
SpillPHIOnRespill(GraphColoring this, Operation operation)
{
  Except_CHECK(Operation_operator(operation) == Operator_PHI);
  {
    BasicBlock basicBlock = Operation_basicBlock(operation);

#line 3345 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if (GRA_TRACE) {
      fprintf(GRA_TRACE, "Spill PHI operation on respill: ");
      Operation_pretty(operation, GRA_TRACE);
    }

#line 3350 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // For each PHI argument, store at the definition point if the
    // argument is not itself spilled in this round of rewrite.
    PHI_arguments_FOREACH_temporary(operation, edge, idx, use) {
      InterfNode useNode = GC_GetNode(this, use);
      if (InterfNode_hasFlag(useNode, InterfNode_ALREADYSPILLED)) {
	// If the use has been spilled previously remove the reload 
	// operation for the PHI. 
	Operation reload = InterfNode_uniqueDef(useNode);
	Except_CHECK(Operation_flags(reload) & OperationFlag_SpillCode);
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Removing reload on spilled PHI argument: ");
	  Operation_pretty(reload, GRA_TRACE);
	}
	BasicBlock_detachOperation(Operation_basicBlock(reload), reload);
      } else if (InterfNode_BELONGTO(useNode) != _spilledNodes) {
	Operation defOp = InterfNode_uniqueDef(useNode);
	Temporary representative = GraphColoring_spillRepresentative(this, use);
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Inserting Spill for PHI argument: ");
	  Operation_pretty(defOp, GRA_TRACE);
	}
	/* The defining operation my have been removed if it is itself a PHI.
	   In this case there is nothing to do.
	   Otherwize we store the definition. */
	if (Operation_basicBlock(defOp) != NULL) 
	  Spill_appendSpill(Operation_basicBlock(defOp), defOp, use, representative);      
      } else {
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Skiping Spill for spilled PHI argument: ");
	  Temporary_pretty(use, GRA_TRACE);
	  fprintf(GRA_TRACE, "\n");
	}
      }
    } PHI_arguments_ENDEACH_temporary;

#line 3385 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // If the PHI result is not itself spilled in this round of rewrite,
    // reload it before the first use in the block or at the end of the block.
    // We could find a better reload point, for instance the nearest dominance
    // point of all uses.
    {
      Temporary def = Operation_getResult(operation, 0);
      InterfNode defNode = GC_GetNode(this, def);
      if (InterfNode_hasFlag(defNode, InterfNode_ALREADYSPILLED)) {
	// If the def has been spilled previously remove the spill 
	// operation for the PHI result. 
	Operation spill = InterfNode_uniqueUse(defNode);
	Except_CHECK(Operation_flags(spill) & OperationFlag_SpillCode);
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Removing spill on spilled PHI result: ");
	  Operation_pretty(spill, GRA_TRACE);
	}
	BasicBlock_detachOperation(Operation_basicBlock(spill), spill);
      } else if (InterfNode_BELONGTO(defNode) != _spilledNodes) {
	Temporary representative = GraphColoring_spillRepresentative(this, def);
	Operation point = NULL;
	BasicBlock_FOREACH_Operation(basicBlock, operation) {
	  if (Operation_usesTemporary(operation, def)) {
	    point = operation;
	    break;
	  }
	} BasicBlock_ENDEACH_Operation;
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Inserting Reload for PHI result on use: ");
	  if (point != NULL) Operation_pretty(point, GRA_TRACE);
	  else fprintf(GRA_TRACE, "End of PHI Block\n");
	}
	Spill_insertReload(basicBlock, point, def, representative);      
      } else {
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "Skiping reload of spilled PHI result: ");
	  Operation_pretty(operation, GRA_TRACE);
	}
      }
    }
    // Now discard the PHI operation
    BasicBlock_detachOperation(Operation_basicBlock(operation), operation);

#line 3427 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    // Update live information.
    // Note that we don't know how to update simply the live analysis
    // as discarding the PHI may have non local effects on the live 
    // information. This will require to redo live analysis after this round of
    // rewrite.
    GRA_redo_liveness = true;
  }
}

#line 3441 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Operation_usesTemporary(Operation operation, Temporary tmp)
{
  Operation_ARGUMENTS_FOREACH_Temporary(operation, use) {
    if (use == tmp) return true;
  } Operation_ARGUMENTS_ENDEACH_Temporary;
  return false;
}

#line 3454 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
Operation_definesTemporary(Operation operation, Temporary tmp)
{
  Operation_RESULTS_FOREACH_Temporary(operation, def) {
    if (def == tmp) return true;
  } Operation_RESULTS_ENDEACH_Temporary;
  return false;
}

#line 3463 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
SpillPHIOperation(GraphColoring this, Operation operation) 
{
  bool removed;
  BasicBlock basicBlock;
  Except_CHECK(Operation_operator(operation) == Operator_PHI);
  basicBlock = Operation_basicBlock(operation);
  Except_CHECK(basicBlock != NULL);
  // If a PHI is spilled, we remove the PHI and all the loads/store
  //corresponding to the uses/def. This is possible as all uses/defs
  // are spilled to the same web location
  PHI_arguments_FOREACH_temporary(operation, edge, idx, use) {
    BasicBlock pred = BasicBlockEdge_origBlock(edge);
    removed = false;
    BasicBlock_FORBACK_Operation(pred, pred_op) {
      if (Operation_definesTemporary(pred_op, use)) {
	Except_CHECK(Operation_flags(pred_op) & OperationFlag_SpillCode);
	BasicBlock_detachOperation(pred, pred_op);
	removed = true;
	break;
      }
    } BasicBlock_ENDBACK_Operation;
    Except_CHECK(removed);
    TemporarySet_remove(BasicBlock_liveOutSet(BasicBlockEdge_origBlock(edge)), use);
  } PHI_arguments_ENDEACH_temporary;
  removed = false;
  BasicBlock_FOREACH_Operation(basicBlock, succ_op) {
    if (Operation_usesTemporary(succ_op, Operation_getResult(operation, 0))) {
      Except_CHECK(Operation_flags(succ_op) & OperationFlag_SpillCode);
      BasicBlock_detachOperation(basicBlock, succ_op);
      removed = true;
      break;
    }
  } BasicBlock_ENDEACH_Operation;
  TemporarySet_remove(BasicBlock_liveInSet(basicBlock),
                      Operation_getResult(operation, 0));
  BasicBlock_detachOperation(basicBlock, operation);
}

#line 3506 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
InterfNode_hasPHI(InterfNode interfNode)
{
  IDList_FOREACH(InterfNode_DEFS(interfNode), Operation, operation_p) {
    if (Operation_operator(*operation_p) == Operator_PHI)
      return true;
  } IDList_ENDEACH;
  IDList_FOREACH(InterfNode_USES(interfNode), Operation, operation_p) {
    if (Operation_operator(*operation_p) == Operator_PHI)
      return true;
  } IDList_ENDEACH;
  return false;
}

#line 3520 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static bool
RewriteProgram(GraphColoring this)
{
  Temporary tmp = NULL;
  IDListItem item = NULL;
  InterfNode interfNode2 = NULL;

#line 3527 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    // Update live-in and live-out
    BasicBlock_liveInSet_FORSORT_Temporary(basicBlock, temporaryIn) {
      if(GC_IsNode(this, temporaryIn) &&
	 InterfNode_BELONGTO(GC_GetNode(this, temporaryIn)) == _spilledNodes)
	TemporarySet_remove(BasicBlock_liveInSet(basicBlock), temporaryIn);
    } BasicBlock_liveInSet_ENDSORT_Temporary;

#line 3535 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporaryOut) {
      if(GC_IsNode(this, temporaryOut) &&
	 InterfNode_BELONGTO(GC_GetNode(this, temporaryOut)) == _spilledNodes)
	TemporarySet_remove(BasicBlock_liveOutSet(basicBlock), temporaryOut);
    } BasicBlock_liveOutSet_ENDSORT_Temporary;

#line 3541 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  } CodeRegion_ENDEACH_BasicBlock;

#line 3543 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // When we spill PHI operations on def, we process PHI first
  if (GraphColoring_flags(this) & GRA_GCF_SPILL_PHI_ON_DEF) {
    IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {
      IDList_FOREACH(InterfNode_DEFS(interfNode), Operation, operation_p) {
	Temporary spilled = InterfNode_TEMPORARY(interfNode);
	if (Operation_operator(*operation_p) == Operator_PHI) {
	  SpillDefAndPHI(this, spilled, *operation_p);
	}
      } IDList_ENDEACH;
    } IDList_ENDEACH;
  }

#line 3555 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // When we spill PHI operations on respill, we process PHI first.
  if (GraphColoring_flags(this) & GRA_GCF_SPILL_PHI_ON_RESPILL) {
    IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {
      // For each node that is respilled, we spill the related PHI operations.
      if (InterfNode_hasFlag(interfNode, InterfNode_ALREADYSPILLED)) {
	Except_CHECK(InterfNode_hasPHI(interfNode));
	IDList_FOREACH(InterfNode_DEFS(interfNode), Operation, operation_p) {
	  Operation operation = *operation_p;
	  // The operation may have been already discarded in this phase.
	  if (Operation_basicBlock(operation) != NULL &&
	      Operation_operator(operation) == Operator_PHI) {
	    SpillPHIOnRespill(this, operation);
	  }
	} IDList_ENDEACH;
	IDList_FOREACH(InterfNode_USES(interfNode), Operation, operation_p) {
	  Operation operation = *operation_p;
	  // The operation may have been already discarded in this phase.
	  if (Operation_basicBlock(operation) != NULL &&
	      Operation_operator(operation) == Operator_PHI) {
	    SpillPHIOnRespill(this, operation);
	  }
	} IDList_ENDEACH;
      }
    } IDList_ENDEACH;
  }

#line 3581 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Rewrite program with inserted spill operations.
  IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {

#line 3584 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    IDList_FOREACH(InterfNode_USES(interfNode), Operation, operation_p) {
      Temporary spilled = InterfNode_TEMPORARY(interfNode);
      // The operation may have been removed by a PHI spill.
      if (Operation_basicBlock(*operation_p) == NULL) continue;
      if (Operation_operator(*operation_p) == Operator_PHI) {
	SpillUseOnPHI(this, spilled, *operation_p);
      } else {
	// A temporary may appear in more than one argument per operation.
	tmp = NULL;
	Operation_ARGUMENTS_FOREACH_Temporary(*operation_p, temporary) {

#line 3595 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	  if(temporary == spilled) {
	    // Create the new temporary for the reload.
	    if (tmp == NULL) {
	      Temporary representative = GraphColoring_spillRepresentative(this, InterfNode_TEMPORARY(interfNode));
	      tmp = GraphColoring_newVirtual(this);
	      TemporaryTable_insert(GraphColoring_tmpTable(this), tmp);
	      Spill_insertReload(Operation_basicBlock(*operation_p), *operation_p, tmp, representative);
	    }
	    // Update the argument.
	    Operation_setArgument(*operation_p, Operation_ARGUMENTS_INDEX, tmp);
	  }
	} Operation_ARGUMENTS_ENDEACH_Temporary;
      }
    } IDList_ENDEACH;

#line 3610 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    IDList_FOREACH(InterfNode_DEFS(interfNode), Operation, operation_p) {
      Temporary spilled = InterfNode_TEMPORARY(interfNode);
      if (Operation_basicBlock(*operation_p) == NULL) continue;
      if (Operation_operator(*operation_p) == Operator_PHI) {
	SpillDefOnPHI(this, spilled, *operation_p);
      } else {
	// A temporary may appear in more than one result per operation.
	tmp = NULL;
	Operation_RESULTS_FOREACH_Temporary(*operation_p, temporary) {
	  if(temporary == spilled) {
	    // Create the new temporary for the spill.
	    if (tmp == NULL) {
	      Temporary representative = GraphColoring_spillRepresentative(this, InterfNode_TEMPORARY(interfNode));
	      tmp = GraphColoring_newVirtual(this);
	      TemporaryTable_insert(GraphColoring_tmpTable(this), tmp);
	      Spill_appendSpill(Operation_basicBlock(*operation_p), *operation_p, tmp, representative);
	    }
	    // Update the result.
	    Operation_setResult(*operation_p, Operation_RESULTS_INDEX, tmp);
	  }
	}
      } Operation_RESULTS_ENDEACH_Temporary;
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 3635 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (!(GraphColoring_flags(this) & (GRA_GCF_SPILL_PHI_ON_DEF|GRA_GCF_SPILL_PHI_ON_RESPILL))) {
    // Check PHI operations for which all def/uses are spilled
    CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
      bool spilled;

#line 3640 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      do {
	spilled = false;
	BasicBlock_FOREACH_Operation(basicBlock, operation) {
	  if (Operation_operator(operation) != Operator_PHI) break;
	  spilled = true;
	  Operation_ARGUMENTS_FOREACH_Temporary(operation, temporary) {
	    if (Temporary_index(temporary) < GraphColoring_INITIALTMPCOUNT(this)) {
	      spilled = false;
	    }
	  } Operation_ARGUMENTS_ENDEACH_Temporary;
	  Operation_RESULTS_FOREACH_Temporary(operation, temporary) {
	    if (Temporary_index(temporary) < GraphColoring_INITIALTMPCOUNT(this)) {
	      spilled = false;
	    }
	  } Operation_RESULTS_ENDEACH_Temporary;
	  if (spilled) {
	    if (GRA_TRACE) {
	      fprintf(GRA_TRACE, "Spilled PHI: ");
	      Operation_pretty(operation, GRA_TRACE);
	    }
	    SpillPHIOperation(this, operation);
	    break; // Need to break the loop has we removed the current operation
	  }
	} BasicBlock_ENDEACH_Operation;
      } while (spilled);
    } CodeRegion_ENDEACH_BasicBlock;
  }
  return true;
}

#line 3670 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**********************************
 * Prune Useless Spill BEGIN      *
 **********************************/

#line 3674 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//ALGO: PruneUselessSpill
static void
PruneUselessSpill(GraphColoring this, bool cisc_mode)
{
  //ALGO: input: P: procedure
  //ALGO: input: cisc_mode: true to estimae for cisc machine spill
  //ALGO: input: spill: set of liveranges to be spilled
  //ALGO: output: useless: set of liverange

#line 3683 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion codeRegion = GraphColoring_codeRegion(this);
  RegFile regFile = GraphColoring_regFile(this);
  TemporarySet live, use, def, clobber, useless, useful_p, tmp, spill;
  const char *procedureName = Procedure_name(CodeRegion_procedure(codeRegion));
  int n_useless, n_useless_p, n_spilled;
  float w_useless, w_useless_p, w_spilled;
  float max;
  InterfNode maxNode;
  int round = 0;
  int use_initial_order = GraphColoring_flags(this) & GRA_GCF_USELESS_INITIAL;

#line 3694 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // Do not do it under SSA.
  if (GraphColoring_flags(this) & GRA_GCF_SSA) return;

#line 3697 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  // currently only for stats
  if (!GRA_STAT) return;

#line 3700 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  use = TemporarySet_make(GraphColoring_memory(this), 0);
  def = TemporarySet_make(GraphColoring_memory(this), 0);
  clobber = TemporarySet_make(GraphColoring_memory(this), 0);
  live = TemporarySet_make(GraphColoring_memory(this), 0);
  useless = TemporarySet_make(GraphColoring_memory(this), 0);
  useful_p = TemporarySet_make(GraphColoring_memory(this), 0);
  spill = TemporarySet_make(GraphColoring_memory(this), 0);
  tmp = TemporarySet_make(GraphColoring_memory(this), 0);

#line 3709 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  n_spilled = 0;
  w_spilled = 0;
  n_useless = 0;
  w_useless = 0;
  n_useless_p = 0;
  w_useless_p = 0;

#line 3716 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {
    Temporary spilled = InterfNode_TEMPORARY(interfNode);
    TemporarySet_insert(spill, spilled);
    n_spilled++;
    w_spilled += InterfNode_SPILLCOST(interfNode);
  } IDList_ENDEACH;

#line 3724 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: repeat 
  do {
    round++;
    //ALGO: useful_p <- empty
     TemporarySet_empty(useful_p);
    
    //ALGO: foreach B in basicblocks(P)
    CodeRegion_FOREACH_BasicBlock( GraphColoring_codeRegion(this), basicBlock) {
      //ALGO: live <- liveout(B)
      TemporarySet_empty(live);
      BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporary) {
	if(GRA_isAllocatable(temporary, regFile)) {
	  TemporarySet_insert(live, temporary);
	}
      } BasicBlock_liveOutSet_ENDSORT_Temporary;
      
      //ALGO: foreach I in instructions(B) in reverse order
      BasicBlock_FORBACK_Operation(basicBlock, operation) {
	
	TemporarySet_empty(use);
	TemporarySet_empty(def);
	TemporarySet_empty(clobber);
	
	Operation_RESULTS_FOREACH_Temporary(operation, temporary) {
	  if(GRA_isAllocatable(temporary, regFile)) {
	    TemporarySet_insert(def, temporary);
	  }
	} Operation_RESULTS_ENDEACH_Temporary;			
	
	Operation_ARGUMENTS_FOREACH_Temporary(operation, temporary) {
	  if(GRA_isAllocatable(temporary, regFile)) {
	    TemporarySet_insert(use, temporary);
	  }
	} Operation_ARGUMENTS_ENDEACH_Temporary;
	
	if(Operation_hasClobber(operation)) {
	  RegisterSet clobberSet = Operation_clobberSet(operation);
	  RegisterSet_FOREACH_Register(clobberSet, reg_clobber) {
	    Temporary temporary_clobber = GraphColoring_getDedicated(this, reg_clobber);
	    if(GRA_isAllocatable(temporary_clobber, regFile))
	      TemporarySet_insert(clobber, temporary_clobber);
	  } RegisterSet_ENDEACH_Register;
	}
	
	//ALGO: live <- live U clobbers(I) U defs(I)  /* single point live ranges */
	TemporarySet_union( live, def);
	TemporarySet_union( live, clobber);
	
	//ALGO: if (cisc_mode)  
	if (cisc_mode) {
	  //ALGO: if (size(live\spill) == K)  
	  TemporarySet_empty(tmp);
	  TemporarySet_union(tmp, live);
	  TemporarySet_diff(tmp, spill);
	  if (TemporarySet_count(tmp) == NbRegisters(regFile)) {
	    //ALGO: useful_p <- useful_p U (live^spill)
	    TemporarySet_empty(tmp);
	    TemporarySet_union(tmp,live);
	    TemporarySet_inter(tmp, spill);
	    TemporarySet_union(useful_p, tmp);
	  }
	} else {
	  //ALGO: if (size(live\spill U defs(I)) >= K)  
	  TemporarySet_empty(tmp);
	  TemporarySet_union(tmp, live);
	  TemporarySet_diff(tmp, spill);
	  TemporarySet_union(tmp, def);
	  if (TemporarySet_count(tmp) >= NbRegisters(regFile)) {
	    //ALGO: useful_p <- useful_p U (live^spill \ defs)
	    TemporarySet_empty(tmp);
	    TemporarySet_union(tmp,live);
	    TemporarySet_inter(tmp, spill);
	    TemporarySet_diff(tmp, def);
	    TemporarySet_union(useful_p, tmp);
	  }
   	}
      
	//ALGO: live <- live \ (defs(I) U clobbers(I))
	TemporarySet_diff( live, def);
	TemporarySet_diff( live, clobber);
	
	//ALGO: live <- live U uses(I)
	TemporarySet_union( live, use);
	
	if (cisc_mode) {
	  //ALGO: if (size(live\spill) == K)  
	  TemporarySet_empty(tmp);
	  TemporarySet_union(tmp, live);
	  TemporarySet_diff(tmp, spill);
	  if (TemporarySet_count(tmp) == NbRegisters(regFile)) {
	    //ALGO: useful_p <- useful_p U (live^spill)
	    TemporarySet_empty(tmp);
	    TemporarySet_union(tmp,live);
	    TemporarySet_inter(tmp, spill);
	    TemporarySet_union(useful_p, tmp);
	  }
	} else {
	  //ALGO: if (size(live\spill U uses(I)) >= K)  
	  TemporarySet_empty(tmp);
	  TemporarySet_union(tmp, live);
	  TemporarySet_diff(tmp, spill);
	  TemporarySet_union(tmp, use);
	  if (TemporarySet_count(tmp) >= NbRegisters(regFile)) {
	    //ALGO: useful_p <- useful_p U (live^spill \ uses(I))
	    TemporarySet_empty(tmp);
	    TemporarySet_union(tmp,live);
	    TemporarySet_inter(tmp, spill);
	    TemporarySet_diff(tmp, use);
	    TemporarySet_union(useful_p, tmp);
	  }
	}
      } BasicBlock_ENDBACK_Operation;
    } CodeRegion_ENDEACH_BasicBlock;

#line 3838 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    max = -1;
    if (use_initial_order) {
      //ALGO: useless_p <- spill \ useful_p
      //ALGO: node <- first node spilled list and in useless_p.
      //NOTE: The spilled list is ordered from top to bottom of the select stack.
      //NOTE: Thus the useless selection here is done in the same order as the 
      //NOTE: node coloring in the AssignColors phase.
      IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {
	Temporary spilled = InterfNode_TEMPORARY(interfNode);
	if (!TemporarySet_contains(useful_p, spilled) &&
	    TemporarySet_contains(spill, spilled)) {
	  if (round == 1) {
	    n_useless_p++;
	    w_useless_p += InterfNode_SPILLCOST(interfNode);
	  }
	  if (max == -1) {
	    maxNode = interfNode;
	    max = InterfNode_SPILLCOST(interfNode);
	  }
	}
      } IDList_ENDEACH;
    } else {
      //ALGO: useless_p <- spill \ useful_p
      //ALGO: node <- node of greatest spill cost in useless_p
      //NOTE: Doing that way is false and should become obsolete
      //NOTE: as the spill cost here is not the same as the
      //NOTE: spill_cost/degree used to order the select stack.
      IDList_FOREACH(GC_spilledNodes(this), InterfNode_, interfNode) {
	Temporary spilled = InterfNode_TEMPORARY(interfNode);
	if (!TemporarySet_contains(useful_p, spilled) &&
	    TemporarySet_contains(spill, spilled)) {
	  if (round == 1) {
	    n_useless_p++;
	    w_useless_p += InterfNode_SPILLCOST(interfNode);
	  }
	  if (InterfNode_SPILLCOST(interfNode) > max) {
	    maxNode = interfNode;
	    max = InterfNode_SPILLCOST(interfNode);
	  }
	}
      } IDList_ENDEACH;
    }
    //ALGO: if (!empty(useless_p))
    if (max != -1) {
      //ALGO: spill <- spill \ node
      //ALGO: useless <- useless U node
      Temporary temp = InterfNode_TEMPORARY(maxNode);
      n_useless++;
      w_useless += max;
      TemporarySet_insert(useless, temp);
      Except_CHECK(TemporarySet_contains(spill, temp));
      TemporarySet_remove(spill, temp);
#if 0
      if (GRA_STAT) {
	fprintf(GRA_STAT, "<useless> ");
	Temporary_pretty(temp, GRA_STAT);
	fprintf(GRA_STAT, " (cost:%.2f)\n",  InterfNode_SPILLCOST(maxNode));
      }
#endif
    }
  } while(max != -1);
  //ALGO: until(empty(useless_p))
      
  if (GRA_STAT) {
    fprintf(GRA_STAT, "<stat>\t\trclass\tpass\tK\tn_spilled\tw_spilled\tn_useless_p\tw_useless_p\tn_useless\tw_useless\tP_name\n");
    fprintf(GRA_STAT, "<stat_prune%s>\t %s\t %d\t %d\t %d\t %.2f\t %d\t %.2f\t %d\t %.2f\t %s\n",
	    cisc_mode? "": "2",
	    RegFile_name(regFile), GraphColoring_PASSNUM(this), 
	    NbRegisters(regFile), 

#line 3908 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	    n_spilled, w_spilled, 
	    n_useless_p, w_useless_p,
	    n_useless, w_useless,
	    procedureName);
  }
  
  TemporarySet_kill(use);
  TemporarySet_kill(clobber);
  TemporarySet_kill(def);
  TemporarySet_kill(live);
  TemporarySet_kill(useful_p);
  TemporarySet_kill(useless);
  TemporarySet_kill(spill);
  TemporarySet_kill(tmp);
}
//ALGO:

#line 3925 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**********************************
 * Prune Useless Spill END        *
 **********************************/

#line 3929 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**********************************
 * Chaintin Spill Heuristic BEGIN *
 **********************************/

#line 3946 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
ChaitinSpill
ChaitinSpill_Ctor(ChaitinSpill this, Temporary origin)
{
  *ChaitinSpill__ORIGIN(this) = origin;
  *ChaitinSpill__SUBSTITUTE(this) = NULL;
  *ChaitinSpill__LASTUSE(this) = NULL;
  *ChaitinSpill__RELOADED(this) = false;
  *ChaitinSpill__RELOADED_LASTUSE(this) = false;
  return this;
}

#line 3957 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
void
ChaitinSpill_Dtor(ChaitinSpill this)
{
  *ChaitinSpill__ORIGIN(this) = NULL;
  *ChaitinSpill__SUBSTITUTE(this) = NULL;
  *ChaitinSpill__LASTUSE(this) = NULL;
  *ChaitinSpill__RELOADED(this) = false;
  *ChaitinSpill__RELOADED_LASTUSE(this) = false;
}

#line 3967 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Cette fonction traite les variables utilisees dans une operation
 * Si la variable recontree doit etre spillee, cette fonction distingue
 * trois cas:
 *  - cette variable n'est pas encore dans la liste needLoad. Dans ce cas
 * on initialise un nouveau struct spill et on remplace la variable a "spiller"
 * par une nouvelle;
 *  - cette variable est dans la liste needLoad et a ete rechargee depuis
 * sa derniere utilisation. Dans ce cas on remplace la variable a spiller
 * par une nouvelle;
 *  - cette variable est dans la liste needLoad mais n'a pas ete rechargee
 * depuis sa derniere utilisation. Dans ce cas on remplace la variable
 * a spiller par celle utilisee pour le remplacement lors de sa derniere
 * utilisation.
 *
 * \param this Graphe de coloration.
 * \param op Operation ou l'on verifie les variables utilisees
 * \param needLoad Liste des variables a recharger
 * \param live Ensemble des variables en vie.
 */
static void
handleUses(GraphColoring this, Operation op, IDList needLoad, TemporarySet live);

#line 3990 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Cette fonction traite la mort d'une variable dans une operation.
 * Si une variable meurt et que les variables a "spiller" n'ont pas deja ete
 * rechargees depuis leur derniere utilisation, alors on insert un "load"
 * pour chacune d'entre elles.
 *
 * \param this Graphe de coloration.
 * \param op Operation ou l'on verifie les variables utilisees.
 * \param needLoad Liste des variables a recharger.
 * \param live Ensemble des variables en vie.
 */
static void
checkDeaths(GraphColoring this, Operation op, IDList needLoad, TemporarySet live);

#line 4004 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Cette fonction traite les variables definies dans une operation.
 * Quoi qu'il arrive si la variable rencontree doit etre "spillee" et
 * qu'elle vit en sortie de ce bloc de base, alors on doit la sauver en memoire
 * (puisque l'on ne sait pas se passe dans les blocs suivants).
 * Ce genre de variables sont tout de suite inserees dans needLoad avec
 * "reloaded = true" (cf ChaitinSpillCode).
 *
 * Si la variable rencontree a ete rechargee au moins une fois, alors on la
 * sauvegarde en memoire.
 *
 * S'il existe une variable telle qu'elle soit definie mais jamais utilise
 * apres cette definition, alors on ne fait que remplacer cette variable...
 *
 * Sinon, c'est que toutes les utilisations de cette variables sont "closes" a
 * sa definition, et qu'elle a ete "spillee" a tord (elle interferait ailleurs
 * et par hazard on a choisi celle-ci pour liberer de la pression alors que
 * c'etait une autre qui genait). On la remplace donc par la variable de
 * remplacement utilise dans toutes les utilisations precedentes.
 *
 * \param this Graphe de coloration.
 * \param op Operation ou l'on verifie les variables utilisees.
 * \param needLoad Liste des varaibles a recharger.
 * \param live Ensemble des variables en vie.
 */
static void
handleDefs(GraphColoring this, Operation op, IDList needLoad, TemporarySet live);

#line 4032 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Fonction de comparaison pour l'heuristique de spill de Chaitin.
 */
static int
ChaitinSpill_cmp(void *interfNode, void *temporary);

#line 4038 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Affiche de maniere visible une structure de spill
 * utilisee dans l'heuristique de Chaitin.
 */
bool
ChaitinSpill_pretty(ChaitinSpill spill, FILE* file);

#line 4046 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//ALGO: ChaitinSpillCode(code)
static void
ChaitinSpillCode(GraphColoring this)
{
  TemporarySet live;
  RegFile regFile = GraphColoring_regFile(this);
  IDList_ needLoad__;
  IDList needLoad = (IDList)&needLoad__;

#line 4055 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  live = TemporarySet_make(GraphColoring_memory(this), 0);
  IDList_Ctor(needLoad, GraphColoring_memory(this));

#line 4058 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: foreachBasicBlock(code, blk)
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), blk) {
    //ALGO: copySet(live, blk->liveOut)
    TemporarySet_empty(live);
    BasicBlock_liveOutSet_FORSORT_Temporary(blk, var) {
      if(GRA_isAllocatable(var, regFile)) {
	TemporarySet_insert(live, var);
      }
    } BasicBlock_liveOutSet_ENDSORT_Temporary;

#line 4068 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef PRO_NDTOR
    //ALGO: clearSet(needLoad)
    IDList_empty(needLoad, (void(*)(void*))ChaitinSpill_Dtor);
#endif//PRO_NDTOR

#line 4073 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: foreachMember(live, var)
    TemporarySet_FORSORT_Temporary(live, Temporary_compare, var) {
      //ALGO: if(true == isMember(spilled, var))
      if(NULL != IDList_containsSame(GC_spilledNodes(this), ChaitinSpill_cmp, var)) {
	//ALGO: spill = createMember(needLoad)
	ChaitinSpill spill = (ChaitinSpill)IDList_push(needLoad, sizeof(ChaitinSpill_));
	//ALGO: spill.origin = var
	ChaitinSpill_Ctor(spill, var);
	//ALGO: spill.substitute = NULL
	*ChaitinSpill__SUBSTITUTE(spill) = NULL;
	//ALGO: spill.lastUse = NULL
	*ChaitinSpill__LASTUSE(spill) = NULL;
	//ALGO: spill.reloaded = true
	*ChaitinSpill__RELOADED(spill) = true;
	//ALGO: spill.reloaded_lastUse = true
	*ChaitinSpill__RELOADED_LASTUSE(spill) = true;
      }
    } TemporarySet_ENDSORT_Temporary;

#line 4092 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: forbackOp(blk, op)
    BasicBlock_FORBACK_Operation(blk, op) {
      //ALGO: handleDefs(op)
      handleDefs(this, op, needLoad, live);
      //ALGO: checkDeaths(op)
      checkDeaths(this, op, needLoad, live);
      //ALGO: handleUses(op)
      handleUses(this, op, needLoad, live);
    } BasicBlock_ENDBACK_Operation;

#line 4102 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: foreachMember(needLoad, spill)
    IDList_FOREACH(needLoad, ChaitinSpill_, spill) {
      //ALGO: if(NULL != spill.substitute && false == spill.reloaded_lastuse)
      if(NULL != ChaitinSpill_SUBSTITUTE(spill) && false == ChaitinSpill_RELOADED_LASTUSE(spill)) {
	//ALGO: insertLoad(spill.origin, spill.substitute, spill.lastUse)
	Temporary representative = GraphColoring_spillRepresentative(this, ChaitinSpill_ORIGIN(spill));

#line 4109 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	Spill_insertReload(Operation_basicBlock(ChaitinSpill_LASTUSE(spill)), ChaitinSpill_LASTUSE(spill), ChaitinSpill_SUBSTITUTE(spill), representative);
      }
    } IDList_ENDEACH;

#line 4113 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: update(blk->liveOut) TODO: don't remove liveout if operand of PHI in one successor
    BasicBlock_liveOutSet_FORSORT_Temporary(blk, temporaryOut) {
      if(GC_IsNode(this, temporaryOut) &&
	 InterfNode_BELONGTO(GC_GetNode(this, temporaryOut)) == _spilledNodes)
	TemporarySet_remove(BasicBlock_liveOutSet(blk), temporaryOut);
    } BasicBlock_liveOutSet_ENDSORT_Temporary;

#line 4120 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: update(blk->liveIn)
    BasicBlock_liveInSet_FORSORT_Temporary(blk, temporaryIn) {
      if(GC_IsNode(this, temporaryIn) &&
	 InterfNode_BELONGTO(GC_GetNode(this, temporaryIn)) == _spilledNodes)
	TemporarySet_remove(BasicBlock_liveInSet(blk), temporaryIn);
    } BasicBlock_liveInSet_ENDSORT_Temporary;

#line 4127 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  } CodeRegion_ENDEACH_BasicBlock;

#line 4129 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  TemporarySet_kill(live);
#ifndef PRO_NDTOR
  IDList_empty(needLoad, (void(*)(void*))ChaitinSpill_Dtor);
#endif//PRO_NDTOR
  IDList_Dtor(needLoad);
}
//ALGO:

#line 4138 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//ALGO: handleUses(op)
static void
handleUses(GraphColoring this, Operation op, IDList needLoad, TemporarySet live)
{
  //ALGO: forEachUsedVar(op, var)
  Operation_ARGUMENTS_FOREACH_Temporary(op, var) {
    //ALGO: if(true == isMember(spilled, var))
    if(NULL != IDList_containsSame(GC_spilledNodes(this), ChaitinSpill_cmp, var)) {
      //ALGO: get spill in needLoad where spill.origin == var
      ChaitinSpill spill = NULL;
      IDList_FOREACH(needLoad, ChaitinSpill_, tmp) {
	if(ChaitinSpill_ORIGIN(tmp) == var) {
	  spill = tmp;
	  break;
	}
      } IDList_ENDEACH;

#line 4155 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: if(undef == spill)
      if(NULL == spill) {
	//ALGO: spill = createMember(needLoad)
	spill = (ChaitinSpill)IDList_push(needLoad, sizeof(ChaitinSpill_));
	//ALGO: spill.origin = var
	ChaitinSpill_Ctor(spill, var);
	//ALGO: spill.substitute = NULL
	*ChaitinSpill__SUBSTITUTE(spill) = NULL;
	//ALGO: spill.lastUse = NULL
	*ChaitinSpill__LASTUSE(spill) = NULL;
	//ALGO: spill.reloaded = false
	*ChaitinSpill__RELOADED(spill) = false;
	//ALGO: spill.reloaded_lastUse = true
	*ChaitinSpill__RELOADED_LASTUSE(spill) = true;
      }

#line 4171 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: if(true == spill.reloaded_lastUse)
      if(true == ChaitinSpill_RELOADED_LASTUSE(spill))
	//ALGO: if (spill.substitute == NULL) spill.substitute = new Var
	if(NULL == ChaitinSpill_SUBSTITUTE(spill)) {
	  *ChaitinSpill__SUBSTITUTE(spill) = GraphColoring_newVirtual(this);
	  TemporaryTable_insert(GraphColoring_tmpTable(this), ChaitinSpill_SUBSTITUTE(spill));
	}
      //ALGO: spill.reloaded_lastUse = false
      *ChaitinSpill__RELOADED_LASTUSE(spill) = false;
      //ALGO: spill.lastUse = op
      *ChaitinSpill__LASTUSE(spill) = op;
      //ALGO: replacesUses(var, spill.substitute, op)
      Operation_setArgument(op, Operation_ARGUMENTS_INDEX, ChaitinSpill_SUBSTITUTE(spill));
    }

#line 4186 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: addMember(live, var)
    TemporarySet_insert(live, var);
  } Operation_ARGUMENTS_ENDEACH_Temporary;
}
//ALGO:

#line 4192 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//ALGO: checkDeaths(op)
static void
checkDeaths(GraphColoring this, Operation op, IDList needLoad, TemporarySet live)
{
  //ALGO: foreachUsedVar(op, var)
  Operation_ARGUMENTS_FOREACH_Temporary(op, var) {
    //ALGO: if(false == isMember(live, var)
    if(false == TemporarySet_contains(live, var)) {
      //ALGO: foreachMember(needLoad, spill)
      IDList_FOREACH(needLoad, ChaitinSpill_, spill) {
	//ALGO: if(false == spill.reloaded_lastUse)
	if(false == ChaitinSpill_RELOADED_LASTUSE(spill)) {
	  //ALGO: insertLoad(spill.origin, spill.substitute, spill.lastUse)
	  Temporary representative = GraphColoring_spillRepresentative(this, ChaitinSpill_ORIGIN(spill));
	  Spill_insertReload(Operation_basicBlock(ChaitinSpill_LASTUSE(spill)), ChaitinSpill_LASTUSE(spill), ChaitinSpill_SUBSTITUTE(spill), representative);
	  //ALGO: spill.reloaded = true
	  *ChaitinSpill__RELOADED(spill) = true;
	  //ALGO: spill.reloaded_lastUse = true
	  *ChaitinSpill__RELOADED_LASTUSE(spill) = true;
	  //ALGO: spill.substitute = NULL
	  *ChaitinSpill__SUBSTITUTE(spill) = NULL;
	}
      } IDList_ENDEACH;
    }
  } Operation_ARGUMENTS_ENDEACH_Temporary;
}
//ALGO:

#line 4220 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//ALGO: handleDefs(op)
static void
handleDefs(GraphColoring this, Operation op, IDList needLoad, TemporarySet live)
{
  //ALGO: foreachDefinedVar(op, var)
  Operation_RESULTS_FOREACH_Temporary(op, var) {
    //ALGO: if(true == isMember(spilled, var)
    if(NULL != IDList_containsSame(GC_spilledNodes(this), ChaitinSpill_cmp, var)) {
      //ALGO: get spill from needLoad where spill.origin == var
      ChaitinSpill spill = NULL;
      IDList_FOREACH(needLoad, ChaitinSpill_, tmp) {
	if(ChaitinSpill_ORIGIN(tmp) == var) {
	  spill = tmp;
	  break;
	}
      } IDList_ENDEACH;

#line 4237 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //ALGO: if(undef != spill)
      if(NULL != spill) {
	//ALGO: if(true == spill.reloaded)
	if(true == ChaitinSpill_RELOADED(spill)) {
	  Temporary representative;
	  //ALGO: if (spill.substitute != NULL) spill.substiture = new Var
	  if(NULL == ChaitinSpill_SUBSTITUTE(spill)) {
	    *ChaitinSpill__SUBSTITUTE(spill) = GraphColoring_newVirtual(this);
	    TemporaryTable_insert(GraphColoring_tmpTable(this), ChaitinSpill_SUBSTITUTE(spill));
	  }
	  //ALGO: replaceDefs(var, spill.substitute, op)
	  Operation_setResult(op, Operation_RESULTS_INDEX, ChaitinSpill_SUBSTITUTE(spill));
	  //ALGO: insertStore(spill.origin, spill.substitute, op)
	  representative = GraphColoring_spillRepresentative(this, ChaitinSpill_ORIGIN(spill));
	  Spill_appendSpill(Operation_basicBlock(op), op, ChaitinSpill_SUBSTITUTE(spill), representative);
	}
	//ALGO: else
	else {
	  //ALGO: replaceDefs(var, spill.substitute, op)
	  Operation_setResult(op, Operation_RESULTS_INDEX, ChaitinSpill_SUBSTITUTE(spill));
	}

#line 4259 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	//ALGO: deleteMember(needLoad, spill)
	IDList_remove_(needLoad, spill);
      }
      //ALGO: else
      else {
	//ALGO: replaceDefs(var, new Var, op)
	Temporary new_var = GraphColoring_newVirtual(this);
	TemporaryTable_insert(GraphColoring_tmpTable(this), new_var);
	Operation_setResult(op, Operation_RESULTS_INDEX, new_var);
      }

#line 4270 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    }
    //ALGO: deleteMember(live, var)
    TemporarySet_remove(live, var);
  } Operation_RESULTS_ENDEACH_Temporary;
}
//ALGO:		

#line 4277 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static int
ChaitinSpill_cmp(void *interfNode, void *temporary)
{
  return (Temporary_IDENTITY(InterfNode_TEMPORARY((InterfNode)interfNode)) - Temporary_IDENTITY((Temporary)temporary));
}

#line 4283 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
bool
ChaitinSpill_pretty(ChaitinSpill spill, FILE* file)
{
  if(NULL == file) file = stderr;
  fprintf(file, "Spill:\n");
  fprintf(file, "\tTemporary origin\t= ");
  if(NULL == ChaitinSpill_ORIGIN(spill))
    fprintf(file, "NULL");
  else Temporary_pretty(ChaitinSpill_ORIGIN(spill), file);
  fprintf(file, "\n\tTemporary substitute\t= ");
  if(NULL == ChaitinSpill_SUBSTITUTE(spill))
    fprintf(file, "NULL");
  else Temporary_pretty(ChaitinSpill_SUBSTITUTE(spill), file);
  fprintf(file, "\n\tOperation lastUse\t= ");
  if(NULL == ChaitinSpill_LASTUSE(spill))
    fprintf(file, "NULL\n");
  else Operation_pretty(ChaitinSpill_LASTUSE(spill), file);
  fprintf(file, "\tbool reloaded\t\t= ");
  switch(ChaitinSpill_RELOADED(spill)) {
  case true:
    fprintf(file, "true\n");
    break;
  case false:
    fprintf(file, "false\n");
    break;
  default:
    fprintf(file, "unknown\n");
    break;
  }
  fprintf(file, "\tbool reloaded_lastUse\t= ");
  switch(ChaitinSpill_RELOADED_LASTUSE(spill)) {
  case true:
    fprintf(file, "true\n");
    break;
  case false:
    fprintf(file, "false\n");
    break;
  default:
    fprintf(file, "unknown\n");
    break;
  }
  fprintf(file, "\n");
  //
  return true;
}
#endif//ECL_NTRACE

#line 4331 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/********************************
 * Chaintin Spill Heuristic END *
 ********************************/

#line 4335 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/*************************
 * Export Graphviz BEGIN *
 *************************/

#line 4339 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//ALGO: exportGraphviz(file, code, var)
#ifndef ECL_NTRACE
static void
exportGraphviz(FILE *file, GraphColoring this, bool liveranges)
{
  int count = 0;

#line 4346 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if(NULL == file)
    file = stderr;

#line 4349 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: createGraphvizHead(file)
  fprintf(file, "digraph controlflow {\n" /*\}*/ );

#line 4352 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  /*  if(NULL != var) {*/
  /*    fprintf(file, "\tlabel=");*/
  /*    Temporary_pretty(var, file);*/
  /*    fprintf(file, ";\n");*/
  /*  }*/

#line 4359 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: foreachBasicBlock(code, blk)
  CodeRegion_FOREACH_BasicBlock(GraphColoring_codeRegion(this), basicBlock) {
    int32_t id = BasicBlock_INDEX(basicBlock);

#line 4363 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    fprintf(file, "\t%d [shape=record, label = \"{.+*= BB %d =*+.|",
	    id,
	    BasicBlock_identity(basicBlock));

#line 4367 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO:  print livein
    if (liveranges) {
      int count = 0;
      BasicBlock_liveInSet_FORSORT_Temporary(basicBlock, a) {
	count++;
	if (GRA_isAssigned(a)) count++;
	if (count > 9) {
	  fprintf(file, "\n");
	  count = 0;
	}
	Temporary_pretty (a, file);
	fprintf(file, " ");
      } BasicBlock_liveInSet_ENDSORT_Temporary;
      fprintf(file, "|");
    }

#line 4383 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    BasicBlock_pretty(basicBlock, file);

#line 4385 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO:  print liveout
    if (liveranges) {
      int count = 0;
      fprintf(file, "|");
      BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, a) {
	count++;
	if (GRA_isAssigned(a)) count++;
	if (count > 9) {
	  fprintf(file, "\n");
	  count = 0;
	}
	Temporary_pretty (a, file);
	fprintf(file, " ");
      } BasicBlock_liveOutSet_ENDSORT_Temporary;
    }

#line 4402 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    fprintf(file,"}\"];\n");

#line 4404 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    //ALGO: foreachPreviousBlock(blk, prev)
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
      BasicBlock dest_node = BasicBlockEdge_destBlock(edge);
      //ALGO: createGraphvizArc(file, blk, prev)
      if (NULL != dest_node) {
	fprintf(file, "\t%d -> %d ", BasicBlock_INDEX(basicBlock), BasicBlock_INDEX(dest_node));
	count = 0;

#line 4412 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	/*        if(NULL != var) {*/
	/*          BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, tmp) {*/
	/*            if(var == tmp) {*/
	/*              count++;*/
	/*              break;*/
	/*            }*/
	/*          } BasicBlock_liveOutSet_ENDSORT_Temporary;*/

#line 4420 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	/*          BasicBlock_liveInSet_FORSORT_Temporary(dest_node, tmp) {*/
	/*            if(var == tmp) {*/
	/*              count++;*/
	/*              break;*/
	/*            }*/
	/*          } BasicBlock_liveInSet_ENDSORT_Temporary;*/
	/*        }*/

#line 4428 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	if(2 == count)
	  fprintf(file, "[style=bold];\n");
	else fprintf(file, ";\n");
      }
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  } CodeRegion_ENDEACH_BasicBlock;

#line 4435 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  //ALGO: createGraphvizFoot(file)
  fprintf(file, /*\{*/ "}\n");
}
#endif//ECL_NTRACE
//ALGO:

#line 4441 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/***********************
 * Export Graphviz END *
 ***********************/

#line 4445 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static IQueue_ allocatableRegisters_[RegFile__FIXME];

#line 4447 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static inline IQueue
AllocatableRegisters(RegFile regFile)
{
  Except_CHECK(regFile < RegFile__COUNT);
  return &allocatableRegisters_[regFile];
}

#line 4454 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static inline uint32_t
NbRegisters(RegFile regFile)
{
  return IQueue_count(AllocatableRegisters(regFile));
}

#line 4460 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
bool
GraphColoring_check(GraphColoring this, FILE *file)
{
  InterfGraph interfGraph = GC_interfGraph(this);
  int i;
  bool *treated = Memory_alloc(GraphColoring_memory(this), InterfGraph_SIZE(interfGraph)*sizeof(bool));

#line 4468 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  for( i=0; i<InterfGraph_SIZE(interfGraph); i++) {
    if(InterfGraph_isAdjacentIndex(interfGraph, i, i))
      treated[i] = true;
    else treated[i] = false;
  }

#line 4474 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (file == NULL) file = stderr;

#line 4476 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf(file,"Checking...\n");
  IDList_FOREACH( GC_precolored(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _precolored ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'precolored' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4500 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_initial(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _initial ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'initial' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4523 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_simplifyWorklist(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _simplifyWorklist ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'simplifyWorklist' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4546 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_freezeWorklist(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _freezeWorklist ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'freezeWorklist' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	//fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4569 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_spillWorklist(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _spillWorklist ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'spillWorklist' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4592 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_spilledNodes(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _spilledNodes ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'spilledNodes' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4615 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coalescedNodes(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _coalescedNodes ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'coalescedNodes' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4638 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coloredNodes(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _coloredNodes ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'coloredNodes' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4661 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_discardedNodes(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _discardedNodes ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'discardedNodes' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
  } IDList_ENDEACH;

#line 4675 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_selectStack(this), InterfNode_, interfNode) {
    treated[InterfNode_index(interfNode)] = false;
    if( InterfNode_BELONGTO(interfNode) != _selectStack ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the 'selectStack' list:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    if( !InterfGraph_isAdjacent(interfGraph, interfNode, interfNode) ) {
      fprintf(file, "ERROR: The following InterfNode is not declared into the adjSet:");
      InterfNode_pretty(interfNode, file);
      return false;
    }
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( !InterfGraph_isAdjacent(interfGraph, interfNode, *interfNode_p) ||
	  !InterfGraph_isAdjacent(interfGraph, *interfNode_p, interfNode)) {
	fprintf(file, "ERROR: The following InterfNodes are not declared adjacent into the adjSet:");
	InterfNode_pretty(interfNode, file);
	InterfNode_pretty(*interfNode_p, file);
	return false;
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4698 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  for( i=0; i<InterfGraph_SIZE(interfGraph); i++) {
    if(treated[i] == true) {
      fprintf(file, "ERROR: The following InterfNode is not in a list: (index) %d", i);
      Temporary_pretty(TemporaryTable_BASE(GraphColoring_tmpTable(this))[i], file);
      fprintf(file, "\n");
      return false;
    }
  }

#line 4707 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  Memory_free(GraphColoring_memory(this), treated);
  //fprintf(file,"done.\n");
  return true;
}
#endif//ECL_NTRACE

#line 4713 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
static void
ComputeAllocationCost(GraphColoring this, FILE *file)
{
  InterfMove interfMove;
  InterfNode interfNode1, interfNode2;
  Operation operation;
  int32_t MoveNbr = 0;
  int32_t MoveRemaining = 0;
  float MoveCost = 0;

#line 4724 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "Total Spill Cost: = %f\n", totalspillcost);

#line 4726 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "\ncoalescedMoves:\n");
  IDList_FOREACH( GC_coalescedMoves(this), InterfMove_, interfMove) {
    interfNode1 = InterfMove_DEF(interfMove);
    interfNode2 = InterfMove_USE(interfMove);

#line 4731 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    assert( InterfNode_COLOR(interfNode1) == InterfNode_COLOR(interfNode2) );
    MoveNbr++;
  } IDList_ENDEACH;

#line 4735 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "\nconstrainedMoves:\n");
  IDList_FOREACH( GC_constrainedMoves(this), InterfMove_, interfMove) {
    interfNode1 = InterfMove_DEF(interfMove);
    interfNode2 = InterfMove_USE(interfMove);
    assert( InterfNode_COLOR(interfNode1) != InterfNode_COLOR(interfNode2) );
    MoveNbr++;
    MoveRemaining++;
    MoveCost += InterfMove_COST(interfMove);
  } IDList_ENDEACH;

#line 4745 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "\nfrozenMoves:\n");
  IDList_FOREACH( GC_frozenMoves(this), InterfMove_, interfMove) {
    operation = InterfMove_OPERATION(interfMove);
    interfNode1 = InterfMove_DEF(interfMove);
    interfNode2 = InterfMove_USE(interfMove);
    MoveNbr++;

#line 4752 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if ( InterfNode_COLOR(interfNode1) != InterfNode_COLOR(interfNode2) ) {
      MoveRemaining++;
      MoveCost += InterfMove_COST(interfMove);
    }
    else {
      fprintf(file, "Frozen move is coalesced.\n");
      InterfMove_pretty(interfMove, file);
    }
  } IDList_ENDEACH;

#line 4762 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "\nworklistMoves:\n");
  assert(IDList_isEmpty(GC_worklistMoves(this)) == true);

#line 4765 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "\nactiveMoves:\n");
  assert(IDList_isEmpty( GC_activeMoves(this)) == true);

#line 4769 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  fprintf( file, "Total Move Cost: = %f\n", MoveCost);
  fprintf( file, "Remains %d Moves among %d\n", MoveRemaining, MoveNbr);

#line 4772 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
}
#endif//ECL_NTRACE

#line 4775 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
static bool
Statistiques_pretty(int32_t *statistiques, FILE *file)
{
  if (file == NULL) file = stderr;
  //
  fprintf( file, "\nNumber of call(s) to:\n");
  fprintf( file, "Build()\t\t= %d\n", statistiques[build]);
  fprintf( file, "Simplify()\t= %d\n", statistiques[simplify]);
  fprintf( file, "Coalesce()\t= %d\n", statistiques[coalesce]);
  fprintf( file, "Freeze()\t= %d\n", statistiques[freeze]);
  fprintf( file, "SelectSpill()\t= %d\n", statistiques[selectSpill]);
  fprintf( file, "Briggs() wins\t= %d\n", statistiques[briggsWins]);
  fprintf( file, "Briggs() loses\t= %d\n", statistiques[briggsLoses]);
  fprintf( file, "George() wins\t= %d\n", statistiques[georgeWins]);
  fprintf( file, "George() loses\t= %d\n", statistiques[georgeLoses]);	
  //
  fprintf( file, "\nNumber of temporaries pushed into:\n");
  fprintf( file, "initial\t\t = %d\n", statistiques[inInitial]);
  fprintf( file, "simplifyWorklist = %d\n", statistiques[inSimplifyWorklist]);
  fprintf( file, "freezeWorklist\t = %d\n",  statistiques[inFreezeWorklist]);
  fprintf( file, "spillWorklist\t = %d\n", statistiques[inSpillWorklist]);
  fprintf( file, "spilledNodes\t = %d\n", statistiques[inSpilledNodes]);
  fprintf( file, "coalescedNodes\t = %d\n", statistiques[inCoalescedNodes]);
  fprintf( file, "coloredNodes\t = %d\n", statistiques[inColoredNodes]);
  fprintf( file, "discardedNodes\t = %d\n", statistiques[inDiscardedNodes]);
  fprintf( file, "selectStack\t = %d\n", statistiques[inSelectStack]);
  fprintf( file, "coalescedMoves\t = %d\n", statistiques[inCoalescedMoves]);
  fprintf( file, "constrainedMoves = %d\n", statistiques[inConstrainedMoves]);
  fprintf( file, "frozenMoves\t = %d\n", statistiques[inFrozenMoves]);
  fprintf( file, "worklistMoves\t = %d\n", statistiques[inWorklistMoves]);
  fprintf( file, "activeMoves\t = %d\n", statistiques[inActiveMoves]);
  //
  fprintf( file, "\nNumber of Temporaries removed from:\n");
  fprintf( file, "simplifyWorklist = %d\n", statistiques[outSimplifyWorklist]);
  fprintf( file, "freezeWorklist\t = %d\n",  statistiques[outFreezeWorklist]);
  fprintf( file, "spillWorklist\t = %d\n", statistiques[outSpillWorklist]);
  fprintf( file, "worklistMoves\t = %d\n", statistiques[outWorklistMoves]);
  fprintf( file, "activeMoves\t = %d\n", statistiques[outActiveMoves]);
  //
  return true;
}
#endif//ECL_NTRACE

#line 4819 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
bool
WorkLists_pretty(WorkLists this, FILE *file)
{
  if (file == NULL) file = stderr;
  //
  fprintf( file, "\n********************************************\n");
  //
  fprintf( file, "\ninitial:\n");
  IDList_FOREACH( WL_initial(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode), file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\nprecolored:\n");
  IDList_FOREACH( WL_precolored(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\nspillWorklist:\n");
  IDList_FOREACH( WL_spillWorklist(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\nfreezeWorklist:\n");
  IDList_FOREACH( WL_freezeWorklist(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\nsimplifyWorklist:\n");
  IDList_FOREACH( WL_simplifyWorklist(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode), file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\nselectStack:\n");
  IDList_FOREACH( WL_selectStack(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\nspilledNodes:\n");
  IDList_FOREACH( WL_spilledNodes(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\ncoloredNodes:\n");
  IDList_FOREACH( WL_coloredNodes(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\ndiscardedNodes:\n");
  IDList_FOREACH( WL_discardedNodes(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\ncoalescedNodes:\n");
  IDList_FOREACH( WL_coalescedNodes(this), InterfNode_, interfNode) {
    Temporary_pretty(InterfNode_TEMPORARY(interfNode),  file);
    fprintf( file, ", ");
  } IDList_ENDEACH;
  //
  fprintf( file, "\ncoalescedMoves:\n");
  IDList_FOREACH( WL_coalescedMoves(this), InterfMove_, interfMove) {
    InterfMove_pretty(interfMove, file);
  } IDList_ENDEACH;
  //
  fprintf( file, "\nconstrainedMoves:\n");
  IDList_FOREACH( WL_constrainedMoves(this), InterfMove_, interfMove) {
    InterfMove_pretty(interfMove, file);
  } IDList_ENDEACH;
  //
  fprintf( file, "\nfrozenMoves:\n");
  IDList_FOREACH( WL_frozenMoves(this), InterfMove_, interfMove) {
    InterfMove_pretty(interfMove, file);
  } IDList_ENDEACH;
  //
  fprintf( file, "\nworklistMoves:\n");
  IDList_FOREACH( WL_worklistMoves(this), InterfMove_, interfMove) {
    InterfMove_pretty(interfMove, file);
  } IDList_ENDEACH;
  //
  fprintf( file, "\nactiveMoves:\n");
  IDList_FOREACH(  WL_activeMoves(this), InterfMove_, interfMove) {
    InterfMove_pretty(interfMove, file);
  } IDList_ENDEACH;
  //
  return true;
}
#endif//ECL_NTRACE

#line 4916 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
static bool
GraphColoring_verification(GraphColoring this)
{
  bool ok = true;

#line 4922 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_initial(this), InterfNode_, interfNode) {
    fprintf(stderr, "\nERROR: initial worklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfNode_pretty(interfNode, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 4930 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_spillWorklist(this), InterfNode_, interfNode) {
    fprintf(stderr, "\nERROR: spill worklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfNode_pretty(interfNode, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 4938 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_freezeWorklist(this), InterfNode_, interfNode) {
    fprintf(stderr, "\nERROR: freeze worklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfNode_pretty(interfNode, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 4946 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_simplifyWorklist(this), InterfNode_, interfNode) {
    fprintf(stderr, "\nERROR: simplify worklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfNode_pretty(interfNode, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 4954 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_selectStack(this), InterfNode_, interfNode) {
    fprintf(stderr, "\nERROR: select stack has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfNode_pretty(interfNode, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 4962 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_spilledNodes(this), InterfNode_, interfNode) {
    fprintf(stderr, "\nERROR: spilled woklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfNode_pretty(interfNode, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 4970 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_precolored(this), InterfNode_, interfNode) {
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( InterfNode_COLOR(interfNode) == InterfNode_COLOR(*interfNode_p)) {
	fprintf(stderr, "\nERROR: bad interference in precolored worklist !!!\n");
	ok = false;
	fprintf( stderr, "\n############################################\n");
	InterfNode_pretty(interfNode, stderr);
	fprintf( stderr, "\n");
	InterfNode_pretty(*interfNode_p, stderr);
	fgetc(stdin);
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4984 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coloredNodes(this), InterfNode_, interfNode) {
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( InterfNode_COLOR(interfNode) == InterfNode_COLOR(*interfNode_p)) {
	fprintf(stderr, "\nERROR: bad interference in colored worklist !!!\n");
	ok = false;
	fprintf( stderr, "\n############################################\n");
	InterfNode_pretty(interfNode, stderr);
	fprintf( stderr, "\n");
	InterfNode_pretty(*interfNode_p, stderr);
	fgetc(stdin);
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 4998 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_coalescedNodes(this), InterfNode_, interfNode) {
    IDList_FOREACH( InterfNode_ADJLIST(interfNode), InterfNode, interfNode_p) {
      if( InterfNode_COLOR(interfNode) == InterfNode_COLOR(*interfNode_p)) {
	fprintf(stderr, "\nERROR: bad interference in coalesced worklist !!!\n");
	ok = false;
	fprintf( stderr, "\n############################################\n");
	InterfNode_pretty(interfNode, stderr);
	fprintf( stderr, "\n");
	InterfNode_pretty(*interfNode_p, stderr);
	fgetc(stdin);
      }
    } IDList_ENDEACH;
  } IDList_ENDEACH;

#line 5012 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(  GC_activeMoves(this), InterfMove_, interfMove) {
    fprintf(stderr, "\nERROR: activeMoves worklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfMove_pretty(interfMove, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 5020 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH( GC_worklistMoves(this), InterfMove_, interfMove) {
    fprintf(stderr, "\nERROR: worklistMoves worklist has to be empty !!!\n");
    ok = false;
    fprintf( stderr, "\n############################################\n");
    InterfMove_pretty(interfMove, stderr);
    fgetc(stdin);
  } IDList_ENDEACH;

#line 5028 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  return ok;
}
#endif//ECL_NTRACE

#line 5032 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/************************************************************
 * Support for weak/sweet edges BEGIN                       *
 ************************************************************/

#line 5049 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
Temporary 
Operation_getCondition(Operation operation)
{
  Except_CHECK(Operator_isGoTo(Operation_operator(operation)) &&
	 Operator_isConditional(Operation_operator(operation)));
  return Operation_getArgument(operation, Operator_predicateIndex(Operation_operator(operation)));
}

#line 5064 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#ifndef ECL_NTRACE
static void
GC_statMaxLive(GraphColoring this, bool after)
{
  CodeRegion codeRegion = GraphColoring_codeRegion(this);
  RegFile regFile = GraphColoring_regFile(this);
  TemporarySet live, use, def, clobber;
  int n_point, n_maxlive, n_fat;
  float w_point, usage, w_maxlive, w_fat;
  int n_edge, n_edge_maxlive, n_edge_fat;
  float w_edge, usage_edge, w_edge_maxlive, w_edge_fat;
  int n_bedge, n_bedge_maxlive, n_bedge_fat;
  float w_bedge, usage_bedge, w_bedge_maxlive, w_bedge_fat;
  int n_fedge, n_fedge_maxlive, n_fedge_fat;
  float w_fedge, usage_fedge, w_fedge_maxlive, w_fedge_fat;
  int n_live;
  const char *procedureName = Procedure_name(CodeRegion_procedure(codeRegion));
  int pass_num = GraphColoring_PASSNUM(this);

#line 5083 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  n_point = n_maxlive = n_fat = 0;
  w_point = usage =  w_maxlive = w_fat = 0.0;
  n_edge = n_edge_maxlive = n_edge_fat = 0;
  w_edge = usage_edge = w_edge_maxlive = w_edge_fat = 0.0;
  n_bedge = n_bedge_maxlive = n_bedge_fat = 0;
  w_bedge = usage_bedge = w_bedge_maxlive = w_bedge_fat = 0.0;
  n_fedge = n_fedge_maxlive = n_fedge_fat = 0;
  w_fedge = usage_fedge = w_fedge_maxlive = w_fedge_fat = 0.0;

#line 5092 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  use = TemporarySet_make(GraphColoring_memory(this), 0);
  def = TemporarySet_make(GraphColoring_memory(this), 0);
  clobber = TemporarySet_make(GraphColoring_memory(this), 0);
  live = TemporarySet_make(GraphColoring_memory(this), 0);

#line 5097 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock( GraphColoring_codeRegion(this), basicBlock) {
    float weight = BasicBlock_frequency(basicBlock);

#line 5100 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
      BasicBlock destBB = BasicBlockEdge_destBlock(edge);
      float edge_weight = weight * BasicBlockEdge_probability(edge);
      bool is_back;
      if (destBB == NULL) continue;
      is_back = BasicBlockEdge_isBack (edge);
      TemporarySet_empty(live);
      BasicBlock_liveInSet_FORSORT_Temporary(destBB, temporary) {
	if(GRA_isAllocatable(temporary, regFile)) {
	  Temporary tmp = after? GraphColoring_getDedicated(this, Temporary_REGISTER(temporary)): temporary;
	  TemporarySet_insert( live, tmp);
	}
      } BasicBlock_liveInSet_ENDSORT_Temporary;
      n_live = TemporarySet_count(live);
      n_edge++;
      w_edge += edge_weight;
      usage_edge += (float)n_live/NbRegisters(regFile);
      if (n_live == NbRegisters(regFile)) {
	n_edge_maxlive++;
	w_edge_maxlive += edge_weight;
      }
      if (n_live > NbRegisters(regFile)) {
	n_edge_fat++;
	w_edge_fat += edge_weight;
      }
      if (is_back) {
	n_bedge++;
	w_bedge += edge_weight;
	usage_bedge += (float)n_live/NbRegisters(regFile);
	if (n_live == NbRegisters(regFile)) {
	  n_bedge_maxlive++;
	  w_bedge_maxlive += edge_weight;
	}
	if (n_live > NbRegisters(regFile)) {
	  n_bedge_fat++;
	  w_bedge_fat += edge_weight;
	}
      } else {
	n_fedge++;
	w_fedge += edge_weight;
	usage_fedge += (float)n_live/NbRegisters(regFile);
	if (n_live == NbRegisters(regFile)) {
	  n_fedge_maxlive++;
	  w_fedge_maxlive += edge_weight;
	}
	if (n_live > NbRegisters(regFile)) {
	  n_fedge_fat++;
	  w_fedge_fat += edge_weight;
	}
      }
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
	
    TemporarySet_empty(live);
    BasicBlock_liveOutSet_FORSORT_Temporary(basicBlock, temporary) {
      if(GRA_isAllocatable(temporary, regFile)) {
	Temporary tmp = after? GraphColoring_getDedicated(this, Temporary_REGISTER(temporary)): temporary;
	TemporarySet_insert( live, tmp);
      }
    } BasicBlock_liveOutSet_ENDSORT_Temporary;
    
    BasicBlock_FORBACK_Operation( basicBlock, operation) {

#line 5162 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // We do not count PHI operations
      if (Operation_operator(operation) == Operator_PHI) goto end_block;

#line 5165 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      TemporarySet_empty(use);
      TemporarySet_empty(def);
      TemporarySet_empty(clobber);

#line 5169 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      Operation_RESULTS_FOREACH_Temporary( operation, temporary) {
	if(GRA_isAllocatable(temporary, regFile)) {
	  Temporary tmp = after? GraphColoring_getDedicated(this, Temporary_REGISTER(temporary)): temporary;
	  TemporarySet_insert(def, tmp);
	}
      } Operation_RESULTS_ENDEACH_Temporary;			

#line 5176 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      Operation_ARGUMENTS_FOREACH_Temporary( operation, temporary) {
	if(GRA_isAllocatable(temporary, regFile)) {
	  Temporary tmp = after? GraphColoring_getDedicated(this, Temporary_REGISTER(temporary)): temporary;
	  TemporarySet_insert(use, tmp);
	}
      } Operation_ARGUMENTS_ENDEACH_Temporary;

#line 5183 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      if(Operation_hasClobber(operation)) {
	RegisterSet clobberSet = Operation_clobberSet(operation);
	RegisterSet_FOREACH_Register(clobberSet, reg_clobber) {
	  Temporary temporary_clobber = GraphColoring_getDedicated(this, reg_clobber);
	  if(GRA_isAllocatable(temporary_clobber, regFile))
	    TemporarySet_insert(clobber, temporary_clobber);
	} RegisterSet_ENDEACH_Register;
      }

#line 5192 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      //live <- live union def(I) union clobber(I)
      TemporarySet_union( live, def);
      TemporarySet_union( live, clobber);

#line 5196 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // We consider live at def point.
      n_live = TemporarySet_count(live);
      if (!(!after || n_live <= NbRegisters(regFile))) {
	if (GRA_TRACE) {
	  fprintf(GRA_TRACE, "ERROR: %d Maxlive not true at def of operation: %d > %d: ", after, n_live, NbRegisters(regFile));
	  Operation_pretty(operation, GRA_TRACE);
	}
	TemporarySet_pretty(live, GRA_TRACE);
	CodeRegion_pretty(codeRegion, GRA_TRACE);
	fflush(GRA_TRACE);
	Except_CHECK(!after || n_live <= NbRegisters(regFile));
      }

#line 5209 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // Do not stat spill operations
      if (!(Operation_flags(operation) & OperationFlag_SpillCode)) {
	n_point++;
	w_point += weight;
	usage += (float)n_live/NbRegisters(regFile);

#line 5215 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
	if (n_live == NbRegisters(regFile)) {
	  n_maxlive++;
	  w_maxlive += weight;
	}
	if (n_live > NbRegisters(regFile)) {
	  n_fat++;
	  w_fat += weight;
	}
      }
      // live <- live \ (def(I) union clobber(I)
      TemporarySet_diff( live, def);
      TemporarySet_diff( live, clobber);

#line 5228 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
      // live <- live union use
      TemporarySet_union( live, use);

#line 5231 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    } BasicBlock_ENDBACK_Operation;
  end_block:
    (void)0;
  } CodeRegion_ENDEACH_BasicBlock;

#line 5236 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  TemporarySet_kill(use);
  TemporarySet_kill(clobber);
  TemporarySet_kill(def);
  TemporarySet_kill(live);

#line 5241 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GRA_STAT) {
    fprintf(GRA_STAT, "<stat>\t\trclass\tpass\tmaxlive\tN_point\tR_max\tWR_max\tR_fat\tWR_fat\tusage\tW_point\tP_name\n");
    fprintf(GRA_STAT, "<stat_live>\t %s\t %d\t %d\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	    RegFile_name(regFile), after?-1:pass_num, NbRegisters(regFile), n_point, (float)n_maxlive/n_point, w_maxlive/w_point, (float)n_fat/n_point, w_fat/w_point, usage/n_point, w_point, procedureName);
    fprintf(GRA_STAT, "<stat>\t\trclass\tpass\tmaxlive\tN_edge\tR_max\tWR_max\tR_fat\tWR_fat\tusage\tW_edge\tP_name\n");
    fprintf(GRA_STAT, "<stat_ledge>\t %s\t %d\t %d\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	    RegFile_name(regFile), after?-1:pass_num, NbRegisters(regFile), n_edge, (float)n_edge_maxlive/n_edge, w_edge_maxlive/w_edge, (float)n_edge_fat/n_edge, w_edge_fat/w_edge, usage_edge/n_edge, w_edge, procedureName);
    fprintf(GRA_STAT, "<stat>\t\trclass\tpass\tmaxlive\tN_edge\tR_max\tWR_max\tR_fat\tWR_fat\tusage\tW_edge\tP_name\n");
    fprintf(GRA_STAT, "<stat_lbedge>\t %s\t %d\t %d\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	    RegFile_name(regFile), after?-1:pass_num, NbRegisters(regFile), n_bedge, (float)n_bedge_maxlive/n_bedge, w_bedge_maxlive/w_bedge, (float)n_bedge_fat/n_bedge, w_bedge_fat/w_bedge, usage_bedge/n_bedge, w_bedge, procedureName);
    fprintf(GRA_STAT, "<stat>\t\trclass\tpass\tmaxlive\tN_edge\tR_max\tWR_max\tR_fat\tWR_fat\tusage\tW_edge\tP_name\n");
    fprintf(GRA_STAT, "<stat_lfedge>\t %s\t %d\t %d\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	    RegFile_name(regFile), after?-1:pass_num, NbRegisters(regFile), n_fedge, (float)n_fedge_maxlive/n_fedge, w_fedge_maxlive/w_fedge, (float)n_fedge_fat/n_fedge, w_fedge_fat/w_fedge, usage_fedge/n_fedge, w_fedge, procedureName);
  }
}
#endif//ECL_NTRACE

#line 5267 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static float
BasicBlock_operationCost(BasicBlock this, Operation operation)
{
  Operator operator = Operation_operator(operation);
  float cost;
  if (operator == Operator_PHI) {
    cost = 0.0;
  } else if (Operation_flags(operation) & OperationFlag_SpillCode) {
    cost = 1.0;
  } else if (Operator_isMemoryWrite(operator) || Operator_isMemoryRead(operator)) {
    cost = 1.0;
  } else if (Operator_isGoTo(operator)) {
    BasicBlockEdge edge = BasicBlock_getTargetEdge(this);
    if (Operation_isConditional(operation)) {
      cost = 1.0/4/* operation cost*/;
    } else {
      cost = 1.0/4/* operation cost */;
    }
  } else if (Operator_isSeqCopy(operator)) {
    cost = 1.0 / 4;
  } else {
    cost = 1.0 / 4;
  }
  return cost;
}

#line 5302 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static float
BasicBlock_operationBranchPenalty(BasicBlock this, Operation operation)
{
  Operator operator = Operation_operator(operation);
  float cost = 0;
  if (Operator_isGoTo(operator)) {
    BasicBlockEdge edge = BasicBlock_getTargetEdge(this);
    if (Operation_isConditional(operation)) {
      cost = BasicBlockEdge_probability(edge) * 1.0;
    } else {
      cost = 1.0;
    }
  }
  return cost;
}

#line 5325 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
CodeRegion_statSpill(GraphColoring this, const char *msg)
{
  CodeRegion codeRegion = GraphColoring_codeRegion(this);
  const char *procedureName = Procedure_name(CodeRegion_procedure(codeRegion));
  RegFile regFile = GraphColoring_regFile(this);
  int n_op = 0, n_s = 0, n_move = 0, n_goto = 0, n_mem = 0, n_cond = 0, n_other = 0;
  int n_bb_op;
  float w_bb_op = 0;
  float w_op = 0, w_s = 0, w_move = 0, w_goto = 0, w_mem = 0, w_cond = 0, w_other = 0;;
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    float weight = BasicBlock_frequency(basicBlock);
    float cost = 0.0;
    float bb_ops_cost, bb_br_cost;

#line 5340 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    bb_ops_cost = 0;
    bb_br_cost = 0;
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      Operator operator = Operation_operator(operation);
      float op_cost = BasicBlock_operationCost(basicBlock, operation);
      float br_cost = BasicBlock_operationBranchPenalty(basicBlock, operation);
      if (op_cost == 0.0) continue; /* skip pseudo (includes PHIs) */
      cost = (op_cost + br_cost) * weight;
      
      if (Operation_flags(operation) & OperationFlag_SpillCode) {
	w_s += cost;
	n_s++;
      } else if (Operator_isMemoryWrite(operator) || Operator_isMemoryRead(operator)) {
	w_mem += cost;
	n_mem++;
      }	else if (Operator_isGoTo(operator)) {
	if (Operation_isConditional(operation)) {
	  n_cond++;
	  w_cond += cost;
	} else {
	  n_goto++;
	  w_goto += cost;
	}
      } else if (Operator_isSeqCopy(operator)) {
	n_move++;
	w_move += cost;
      } else {
	n_other++;
	w_other += cost;
      }
      n_op++;
      w_op += cost;
      bb_ops_cost += op_cost;
      bb_br_cost += br_cost;
    } BasicBlock_ENDEACH_Operation;
    w_bb_op += (CEIL(bb_ops_cost) + bb_br_cost) * weight;
  } CodeRegion_ENDEACH_BasicBlock;
  fprintf(GRA_STAT, "<stat>\t\ttype\tclass\tn_op\tw_op\tn_spill\tw_spill\tn_goto\tw_goto\tn_cond\tw_cond\tn_move\tw_move\tn_mem\tw_mem\tn_other\tw_other\tw_bbop\tP_name\n");
  fprintf(GRA_STAT, "<stat_spill>\t %s\t %s\t %d\t %.2f\t %d\t %.2f\t %d\t %.2f\t %d\t %.2f\t %d\t %.2f\t %d\t %.2f\t %d\t %.2f\t %.2f\t %s\n",
	  RegFile_name(regFile), msg, n_op, w_op, n_s, w_s, 
	  n_goto, w_goto, n_cond, w_cond, n_move, w_move,
	  n_mem, w_mem, n_other, w_other, w_bb_op,
	  procedureName);
  
}

#line 5393 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
CodeRegion_statCodeRegion(CodeRegion codeRegion)
{
  const char *procedureName = Procedure_name(CodeRegion_procedure(codeRegion));

#line 5398 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  int n_e, n_c, n_f, n_w , n_s, n_b;
  float w_e, w_c, w_f, w_w, w_s, w_b, w_g;
  int n_back_e, n_back_c, n_back_f, n_back_w , n_back_s;
  float w_back_e, w_back_c, w_back_f, w_back_w, w_back_s;
  n_e = n_c = n_f = n_w = n_s = n_b = 0;
  w_e = w_c = w_f = w_w = w_s = w_b = w_g = 0.0;
  n_back_e = n_back_c = n_back_f = n_back_w = n_back_s = 0;
  w_back_e = w_back_c = w_back_f = w_back_w = w_back_s = 0.0;

#line 5407 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (!GRA_STAT) return;
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    float weight = BasicBlock_frequency(basicBlock);
    float bb_op_cost = 0.0;
    float bb_br_cost = 0.0;
    Operation goto_op;

#line 5414 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    BasicBlock_FOREACH_Operation(basicBlock, operation) {
      float op_cost = BasicBlock_operationCost(basicBlock, operation);
      float br_cost = BasicBlock_operationBranchPenalty(basicBlock, operation);
      bb_op_cost += op_cost * weight;
      bb_br_cost += br_cost * weight;
    } BasicBlock_ENDEACH_Operation;
    n_b++;
    w_b += bb_op_cost;
    w_g += bb_br_cost;
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
      BasicBlock bbHead = BasicBlockEdge_destBlock(edge);
      // Ensure we do not have a dummy flowarc.
      if (bbHead != NULL) {
	float tmp;
	float weight = BasicBlock_frequency(basicBlock) * BasicBlockEdge_probability(edge);
	bool is_back = BasicBlockEdge_isBack(edge);
	n_e++;
	w_e += weight;
	if (is_back) {
	  n_back_e++;
	  w_back_e += weight;
	}
	if (BasicBlockEdge_isCritical(edge)) {
	  n_c++;
	  w_c += weight;
	  if (is_back) {
	    n_back_c++;
	    w_back_c += weight;
	  }
	  if (BasicBlockEdge_isFalse(edge, &tmp)) {
	    n_f++;
	    w_f += weight;
	    if (is_back) {
	      n_back_f++;
	      w_back_f += weight;
	    }
	  } else if (BasicBlockEdge_isStrong(edge)) {
	    n_s++;
	    w_s += weight;
	    if (is_back) {
	      n_back_s++;
	      w_back_s += weight;
	    }
	  } else {
	    // Weak
	    n_w++;
	    w_w += weight;
	    if (is_back) {
	      n_back_w++;
	      w_back_w += weight;
	    }
	  }
	}
      }
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  } CodeRegion_ENDEACH_BasicBlock;
  fprintf(GRA_STAT, "<stat>\t\tN_edge\tN_crit\tN_false\tN_weak\tN_strng\tW_edge\tW_crit\tW_false\tW_weak\tW_strng\tP_name\n");
  fprintf(GRA_STAT, "<stat_edge>\t %d\t %d\t %d\t %d\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	  n_e, n_c, n_f, n_w, n_s, w_e, w_c, w_f, w_w, w_s, procedureName);
  fprintf(GRA_STAT, "<stat_edgeb>\t %d\t %d\t %d\t %d\t %d\t %.2f\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	  n_back_e, n_back_c, n_back_f, n_back_w, n_back_s, w_back_e, w_back_c, w_back_f, w_back_w, w_back_s, procedureName);
  fprintf(GRA_STAT, "<stat>\t\tp_crit\tp_false\tp_weak\tp_strng\tP_name\n");
  fprintf(GRA_STAT, "<stat_edgep>\t %.2f\t %.2f\t %.2f\t %.2f\t %s\n",
	  (float)n_c/n_e*100, (float)n_f/n_c*100, (float)n_w/n_c*100, (float)n_s/n_c*100, procedureName);
  fprintf(GRA_STAT, "<stat>\t\tN_block\tW_block\tW_branch\tP_name\n");
  fprintf(GRA_STAT, "<stat_block>\t %d\t %.2f\t %.2f\t %s\n",
	  n_b, w_b, w_g, procedureName);
}

#line 5490 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
BasicBlock
BasicBlockEdge_splitFalseWithLiveness (Liveness* liveness, Memory memory, CodeRegion codeRegion, BasicBlockEdge this) {
  BasicBlock result = BasicBlockEdge_splitFalse(codeRegion, this);
  /* Create live information on C. */
  // No more possible with new handle of liveness!
#if 0
  PtrSet_union(TemporarySet__(BasicBlock_liveOutSet(bbC)),
               TemporarySet__const(BasicBlock_liveInSet(bbB)));
  PtrSet_union(TemporarySet__(BasicBlock_liveInSet(bbC)),
               TemporarySet__const(BasicBlock_liveOutSet(bbC)));
#else
  if(liveness != NULL)  {
    TemporaryTable globalTable = 
      CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
    Liveness_kill(*liveness);
    *liveness =
      Liveness_make(memory, codeRegion, globalTable, true);
  }
#endif
  return result;
}

#line 5519 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
BasicBlock
BasicBlockEdge_splitCriticalWithLiveness (Liveness* liveness, Memory memory, CodeRegion codeRegion, BasicBlockEdge this) {
  BasicBlock result = BasicBlockEdge_splitCritical(codeRegion, this);
  /* Create live information on C. */
  // No more possible with new handle of liveness!
#if 0
  PtrSet_union(TemporarySet__(BasicBlock_liveOutSet(bbC)),
               TemporarySet__const(BasicBlock_liveInSet(bbB)));
  PtrSet_union(TemporarySet__(BasicBlock_liveInSet(bbC)),
               TemporarySet__const(BasicBlock_liveOutSet(bbC)));
#else
  if(liveness != NULL)  {
    TemporaryTable globalTable = 
      CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
    Liveness_kill(*liveness);
    *liveness =
      Liveness_make(memory, codeRegion, globalTable, true);
  }
#endif
  return result;
}

#line 5549 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
BasicBlock
BasicBlockEdge_splitWithLiveness (Liveness* liveness, Memory memory, CodeRegion codeRegion, BasicBlockEdge this, bool canReuseBB, Operation *point) {
  BasicBlock result = BasicBlockEdge_splitAndAction(this, canReuseBB, point, NULL,
                                                    NULL);
  /* Create live information on C. */
  // No more possible with new handle of liveness!
#if 0
  PtrSet_union(TemporarySet__(BasicBlock_liveOutSet(bbC)),
               TemporarySet__const(BasicBlock_liveInSet(bbB)));
  PtrSet_union(TemporarySet__(BasicBlock_liveInSet(bbC)),
               TemporarySet__const(BasicBlock_liveOutSet(bbC)));
#else
  if(liveness != NULL)  {
    TemporaryTable globalTable = 
      CodeRegion_makeGlobalTable(codeRegion, TemporaryFlag_Constant, 0);
    Liveness_kill(*liveness);
    *liveness =
      Liveness_make(memory, codeRegion, globalTable, true);
  }
#endif
  return result;
}

#line 5580 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
CodeRegion_splitCriticalEdges(GraphColoring gc, CodeRegion codeRegion, bool split_false, bool split_weak, bool split_strong)
{
  int nSplit = 0, nStrong = 0, nWeak = 0, nFalse = 0, nTotal = 0;
  IDList_ criticals_;
  IDList criticals = (IDList)&criticals_;
  Memory memory = Procedure_memory(CodeRegion_procedure(codeRegion));
  
  IDList_Ctor(criticals, memory);

#line 5590 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock_LEAVE_FOREACH_BasicBlockEdge(basicBlock, edge) {
      BasicBlock bbHead = BasicBlockEdge_destBlock(edge);
      // Ensure we do not have a dummy flowarc.
      if (bbHead != NULL) {
	if (BasicBlockEdge_isCritical(edge)) {
	  if (GRA_TRACE) {
	    const char *kind;
	    float devnull;
	    if (BasicBlockEdge_isFalse(edge, &devnull)) kind = "false";
	    else if (BasicBlockEdge_isStrong(edge)) kind = "strong";
	    else kind = "weak";
	    fprintf(GRA_TRACE, "Critical: %d -> %d %s\n", 
		    BasicBlock_INDEX(BasicBlockEdge_origBlock(edge)),
		    BasicBlock_INDEX(BasicBlockEdge_destBlock(edge)),
		    kind);
	  }
	  IDList_PUSH(criticals, BasicBlockEdge, edge);
	}
      }
    } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  } CodeRegion_ENDEACH_BasicBlock;

#line 5613 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(criticals, BasicBlockEdge, arc_p) {
    BasicBlockEdge edge = *arc_p;
    float cost;
    Except_CHECK(BasicBlockEdge_isCritical(edge));
    nTotal++;
    if (BasicBlockEdge_isFalse(edge, &cost)) {
      nFalse++;
      if (split_false) {
        BasicBlockEdge_splitFalseWithLiveness(GraphColoring__LIVENESS(gc), GraphColoring_memory(gc), codeRegion, edge);
	nSplit++;
      }
    } else if (BasicBlockEdge_isStrong(edge)) {
      nStrong++;
      if (split_strong) {
        BasicBlockEdge_splitCriticalWithLiveness(GraphColoring__LIVENESS(gc), GraphColoring_memory(gc), codeRegion, edge);
	nSplit++;
      }
    } else {
      // Weak
      nWeak++;
      if (split_weak) {
        BasicBlockEdge_splitCriticalWithLiveness(GraphColoring__LIVENESS(gc), GraphColoring_memory(gc), codeRegion, edge);
	nSplit++;
      }
    }
  } IDList_ENDEACH;
  IDList_Dtor(criticals);

#line 5641 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  if (GRA_TRACE) {
    fprintf(GRA_TRACE, "CodeRegion_splitCriticalEdges: split: %d total: %d false: %d weak: %d strong: %d\n", nSplit, nTotal, nFalse, nWeak, nStrong);
  }

#line 5645 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
}

#line 5648 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns the unique enter edge if it exist and is the single enter edge for the block.
 * Otherwise return NULL.
 */
static BasicBlockEdge
BasicBlock_getUniqueEnterEdge(BasicBlock this)
{
  BasicBlockEdge only = NULL;
  BasicBlock_ENTER_FOREACH_BasicBlockEdge(this, edge) {
    if (only != NULL) return NULL;
    only = edge;
  } BasicBlock_ENTER_ENDEACH_BasicBlockEdge;
  return only;
}

#line 5663 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns the unique prdecessor if is exist and is the single predecessor block.
 * Otherwise return NULL.
 */
static BasicBlock
BasicBlock_getUniquePredecessor(BasicBlock this)
{
  BasicBlockEdge only = BasicBlock_getUniqueEnterEdge(this);
  return only == NULL? NULL: BasicBlockEdge_origBlock(only);
}

#line 5674 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns the unique leave edge if is exist and is the single leave edge for the block.
 * Otherwise return NULL.
 */
static BasicBlockEdge
BasicBlock_getUniqueLeaveEdge(BasicBlock this)
{
  BasicBlockEdge only = NULL;
  BasicBlock_LEAVE_FOREACH_BasicBlockEdge(this, edge) {
    if (only != NULL) return NULL;
    only = edge;
  } BasicBlock_LEAVE_ENDEACH_BasicBlockEdge;
  return only;
}

#line 5689 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/**
 * Returns the unique successor if is exist and is the single successor block.
 * Otherwise return NULL.
 */
static BasicBlock
BasicBlock_getUniqueSuccessor(BasicBlock this)
{
  BasicBlockEdge only = BasicBlock_getUniqueLeaveEdge(this);
  return only == NULL? NULL: BasicBlockEdge_destBlock(only);
}

#line 5709 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
static void
CodeRegion_removeEmptyBlocks(CodeRegion codeRegion)
{
  IDList_ empty_blocks_;
  IDList empty_blocks = (IDList)&empty_blocks_;
  Memory memory = Procedure_memory(CodeRegion_procedure(codeRegion));

#line 5716 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_Ctor(empty_blocks, memory);

#line 5718 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  CodeRegion_FOREACH_BasicBlock(codeRegion, basicBlock) {
    BasicBlock prev, succ;
    BasicBlockEdge edge;

#line 5722 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
    if ((prev = BasicBlock_getUniquePredecessor(basicBlock)) != NULL &&
	(succ = BasicBlock_getUniqueSuccessor(basicBlock)) != NULL &&
	prev != CodeRegion_rootBlock(codeRegion) &&  // Skip dummy entry block
	succ != CodeRegion_rootBlock(codeRegion) &&  // Skip dummy exit block
	(edge = BasicBlock_getTargetEdge(prev)) != NULL &&
	(BasicBlockEdge_destBlock(edge) == basicBlock) &&
	(edge = BasicBlock_getTargetEdge(basicBlock)) != NULL &&
	(BasicBlockEdge_destBlock(edge) == succ) &&
	BasicBlock_countOperations(basicBlock) == 1) {
      /* The only operation in the basic block is the goto operation. */
      IDList_PUSH(empty_blocks, BasicBlock, basicBlock);
    }
  } CodeRegion_ENDEACH_BasicBlock;

#line 5736 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
  IDList_FOREACH(empty_blocks, BasicBlock, bb_p) {
    BasicBlock basicBlock = *bb_p;
    BasicBlock prev_bb = BasicBlock_getUniquePredecessor(basicBlock);
    BasicBlock succ_bb = BasicBlock_getUniqueSuccessor(basicBlock);
    Operation prev_branch = BasicBlock_getControlOperation(prev_bb);
    BasicBlockEdge target_edge = BasicBlock_getTargetEdge(prev_bb);
    Label succ_label = BasicBlock_makeLabel(succ_bb);
    Procedure_updateTargetedOperation(CodeRegion_procedure(codeRegion), prev_branch, succ_label);
    BasicBlock_killEdge(prev_bb, target_edge);
    BasicBlock_makeEdge(prev_bb, succ_bb, 1.0);
    Procedure_unchainBasicBlock(CodeRegion_procedure(codeRegion), basicBlock);
    //TODO MERGE: makeEnterEdges. Is it ok to do this??
    BasicBlock_makeEnterEdges(succ_bb, Procedure_memory(CodeRegion_procedure(codeRegion)),
    0);
  } IDList_ENDEACH;
  IDList_Dtor(empty_blocks);
}

#line 5754 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
/************************************************************
 * Support for weak/sweet edges END                         *
 ************************************************************/

#line 5758 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
// Support of out of colored ssa moved into OutOfColoredSSA.xcc

#line 5760 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
//////////////////////////////////////////////////////////////////////////////////////////

#line 5772 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
#if XCC__C
static void
GraphColoring_TEST(void)
{
#include "GRA/GraphColoring_.c"
}

#line 5779 "/shares/group/ce-rvex/tools/LAO/LAO-rvex/PRO/GRA/GraphColoring.xcc"
int
main(int argc, char **argv)
{
  GRA_Test(GraphColoring_TEST);
  return 0;
}
#endif


#endif//GraphColoring_c_INCLUDED
